# *****************************************************************************
# *****************************************************************************

# standard

import json
import urllib.request
from typing import Tuple  # not needed with python 3.9

# local, external

from geopandas import GeoDataFrame, GeoSeries
from pandas import DataFrame, merge
from shapely.geometry import Point

# *****************************************************************************
# *****************************************************************************

# URLs

# url prefix to find data using the building entrance id

# url_prefix_entrance = 'https://api.dataforsyningen.dk/adgangsadresser/'

url_prefix_entrance = (
    "https://api.dataforsyningen.dk/bbrlight/opgange?adgangsadresseid="
)

# url prefix to find BBR building data

# url_prefix_buildings = 'https://api.dataforsyningen.dk/bbrlight/bygninger?id='

url_prefix_buildings = "https://api.dataforsyningen.dk/bbrlight/bygninger/"

# url prefix to find the building location data

url_prefix_building_point = "https://api.dataforsyningen.dk/bbrlight/bygningspunkter/"

# *****************************************************************************
# *****************************************************************************

# labels

BBR_BDG_ENTR_LABELS = [
    "ois_id",
    "ois_ts",
    "Opgang_id",
    "AdgAdr_id",
    "ObjStatus",
    "NyByg",
    "Crud_id",
    "ByggeSag_id",
    "OpgStam_id",
    "Bygning_id",
    "Elevator",
    "DataFelt1",
    "DataFelt2",
    "DataFelt3",
    "SagsType",
    "KomFelt1",
    "KomFelt2",
    "KomFelt3",
    "KomFelt4",
    "KomFelt5",
    "KomFelt6",
    "JourNr",
    "ESDH_Ref",
    "SikKlas",
    "OPRET_TS",
    "AENDR_TS",
    "Aendr_Funk",
    "Ophoert_ts",
    "Gyldighedsdato",
    "href",
]

SELECT_BBR_BDG_ENTR_LABELS = ["Opgang_id", "AdgAdr_id", "Bygning_id"]

BBR_BDG_POINT_LABELS = [
    "ois_id",
    "ois_ts",
    "BygPkt_id",
    "Kommune_id",
    "PktRevDato",
    "KoorOest",
    "KoorNord",
    "KoorSystem",
    "TxtRetn",
    "TxtPlacer",
    "DDKNcelle100m",
    "DDKNcelle1km",
    "DDKNcelle10km",
    "OPRET_TS",
    "AENDR_TS",
    "Aendr_Funk",
    "Ophoert_ts",
    "BygPktKilde",
    "koordinater",  # 'koordinater' returns a list
    "href",
]

SELECT_BBR_BDG_POINT_LABELS = [
    "KoorOest",
    "KoorNord",
    "KoorSystem",
    "koordinater",  # 'koordinater' returns a list
]

BBR_BDG_LABELS = [
    "ois_id",
    "ois_ts",
    "Bygning_id",
    "Bygningsnr",
    "ObjStatus",
    "NyByg",
    "Crud_id",
    "Grund_id",
    "Byggesag_id",
    "BygStam_id",
    "AdgAdr_id",
    "BYG_ANVEND_KODE",
    "Klassifikation",
    "AnvKonvKilde",
    "AntLejMKoekken",
    "AntLejUKoekken",
    "OPFOERELSE_AAR",
    "OMBYG_AAR",
    "MIDL_OPRET_KODE",
    "UdloebDatoMidl",
    "BYG_VANDFORSY_KODE",
    "BYG_AFLOEB_KODE",
    "BYG_AFLOEB_TILL",
    "YDERVAEG_KODE",
    "TAG_KODE",
    "SuppYderVaegMat",
    "SuppTagDaekMat",
    "AsbestMateriale",
    "KILDE_MATR_KODE",
    "BYG_ARL_SAML",
    "BYG_BOLIG_ARL_SAML",
    "ERHV_ARL_SAML",
    "BYG_BEBYG_ARL",
    "GARAGE_INDB_ARL",
    "CARPORT_INDB_ARL",
    "UDHUS_INDB_ARL",
    "UDESTUE_ARL",
    "LukOverdaekAreal",
    "AFFALDSRUM_ARL",
    "ANDET_ARL",
    "OverdaekAreal",
    "AabenOverdaekAreal",
    "AdgangsAreal",
    "Carport_Princip",
    "KILDE_ARL_KODE",
    "ETAGER_ANT",
    "ETAGER_AFVIG_KODE",
    "VARMEINSTAL_KODE",
    "OPVARMNING_KODE",
    "VARME_SUPPL_KODE",
    "SIKRINGSRUM_ANT",
    "FREDNING_KODE",
    "BevarVaerdig",
    "UdlejForhold1",
    "HusLeje",
    "HuslejeOplysDato",
    "SagsType",
    "BygPktNoejagtigKls",
    "BygPkt_id",
    "FOT_id",
    "StormRaadPaalaeg",
    "StormRaadPaalaegDato",
    "BygSkadeForsikSelskab",
    "BygSkadeForsikSelskabDato",
    "KomFelt1",
    "KomFelt2",
    "KomFelt3",
    "KomFelt4",
    "KomFelt5",
    "KomFelt6",
    "JourNr",
    "ESDH_Ref",
    "SikKlas",
    "OPRET_TS",
    "AENDR_TS",
    "Aendr_Funk",
    "Ophoert_ts",
    "Kommune_id",
    "KomKode",
    "eRef",
    "Landsejerlavkode",
    "KomEjerlavKode",
    "MatrNr",
    "ESREjdNr",
    "Matr_art_kode",
    "Delnr",
    "Opdelingsnr",
    "BygSkadeOmfatFors",
    "Gyldighedsdato",
    "href",
    "ejerskaber",  # 'ejerskaber' returns a list
    "bygningspunkt",
]  # 'bygningspunkt' returns a dict

SELECT_BBR_BDG_LABELS = [
    "BYG_ANVEND_KODE",
    "BYG_ARL_SAML",
    "BYG_BOLIG_ARL_SAML",
    "ERHV_ARL_SAML",
    "BYG_BEBYG_ARL",
    "GARAGE_INDB_ARL",
    "CARPORT_INDB_ARL",
    "UDHUS_INDB_ARL",
    "UDESTUE_ARL",
    "LukOverdaekAreal",
    "AFFALDSRUM_ARL",
    "ANDET_ARL",
    "OverdaekAreal",
    "AabenOverdaekAreal",
    "AdgangsAreal",
    "ETAGER_ANT",
    "ETAGER_AFVIG_KODE",
    "VARMEINSTAL_KODE",
    "OPVARMNING_KODE",
    "VARME_SUPPL_KODE",
    "BygPkt_id",
]

BBR_CONTAINER_LABELS = {
    "bygningspunkt": dict,
    "ejerskaber": list,
    "koordinater": list,
}

# *****************************************************************************
# *****************************************************************************

# BBR codes

# Source: https://datamodel.lifa.dk/element/BBR%201.7/NyBBR_BygningView/dbo#

# OPVARMNING_KODE: Arten af det opvarmningsmiddel, der anvendes i eget anlæg
# Translation: Nature of the heating agent used in own system

# Codes:
# OPVARMNING_KODE 	1 	Elektricitet
# OPVARMNING_KODE 	2 	Gasværksgas
# OPVARMNING_KODE 	3 	Flydende brændsel
# OPVARMNING_KODE 	4 	Fast brændsel
# OPVARMNING_KODE 	6 	Halm
# OPVARMNING_KODE 	7 	Naturgas
# OPVARMNING_KODE 	9 	Andet

# VARME_SUPPL_KODE: Angives når der udover den hovedsagelige varmeinstallation
# tillige opvarmes af en supplerende varmekilde
# Translation: To be indicated when, in addition to the main heating installation,
# it is also heated by a supplementary heat source

# VARME_SUPPL_KODE 	0 	Ikke oplyst
# VARME_SUPPL_KODE 	1 	Varmepumpeanlæg
# VARME_SUPPL_KODE 	10 	Biogasanlæg
# VARME_SUPPL_KODE 	2 	Ovne til fast eller flydende brændsel
# VARME_SUPPL_KODE 	3 	Ovne til flydende brændsel
# VARME_SUPPL_KODE 	4 	Solpaneler
# VARME_SUPPL_KODE 	5 	Pejs
# VARME_SUPPL_KODE 	6 	Gasradiator
# VARME_SUPPL_KODE 	7 	Elvarme
# VARME_SUPPL_KODE 	80 	Andet
# VARME_SUPPL_KODE 	90 	(UDFASES) Bygningen har ingen supplerende varme

# VARMEINSTAL_KODE: Hvis der forefindes flere forskellige varmeinstallationer,
# anføres koden for den installation, der opvarmer den største delen. Supplerende
# omflyttelige ovne registreres ikke
# Translation: If there are several different heating installations, enter the
# code of the installation that heats the largest part. Additional relocating
# ovens are not registered

# VARMEINSTAL_KODE 	1 	Fjernvarme/blokvarme
# VARMEINSTAL_KODE 	2 	Centralvarme med én fyringsenhed
# VARMEINSTAL_KODE 	3 	Ovn til fast og flydende brændsel
# VARMEINSTAL_KODE 	5 	Varmepumpe
# VARMEINSTAL_KODE 	6 	Centralvarme med to fyringsenheder
# VARMEINSTAL_KODE 	7 	Elvarme
# VARMEINSTAL_KODE 	8 	Gasradiator
# VARMEINSTAL_KODE 	9 	Ingen varmeinstallation
# VARMEINSTAL_KODE 	99 	Blandet

# fuel type

label_bbr_fuel_type = "OPVARMNING_KODE"

dict_bbr_fuel_type_codes = {
    "1": "Elektricitet",
    "2": "Gasværksgas",
    "3": "Flydende brændsel",
    "4": "Fast brændsel",
    "6": "Halm",
    "7": "Naturgas",
    "9": "Andet",
}

# supplementary heating system

label_bbr_extra_heating = "VARME_SUPPL_KODE"

dict_bbr_extra_heating_codes = {
    "0": "Ikke oplyst",
    "1": "Varmepumpeanlæg",
    "10": "Biogasanlæg",
    "2": "Ovne til fast eller flydende brændsel",
    "3": "Ovne til flydende brændsel",
    "4": "Solpaneler",
    "5": "Pejs",
    "6": "Gasradiator",
    "7": "Elvarme",
    "80": "Andet",
    "90": "(UDFASES) Bygningen har ingen supplerende varme",
}

# main heating system

label_bbr_heating_system = "VARMEINSTAL_KODE"

dict_bbr_heating_system_codes = {
    "1": "Fjernvarme/blokvarme",
    "2": "Centralvarme med én fyringsenhed",
    "3": "Ovn til fast og flydende brændsel",
    "5": "Varmepumpe",
    "6": "Centralvarme med to fyringsenheder",
    "7": "Elvarme",
    "8": "Gasradiator",
    "9": "Ingen varmeinstallation",
    "99": "Blandet",
}

# coordinate system

label_bbr_bygningpunkt_koorsys = "KoorSystem"

label_bbr_bygningpunkt_koorsys_codes = {
    "1": "System 34",
    "2": "System 45",
    "3": "KP2000 (System 2000)",
    "4": "UTM ED50",
    "5": "WGS 84",
}

# building use

label_bbr_building_uses = "BYG_ANVEND_KODE"  # byganvendelse

dict_bbr_building_use_codes = {
    "110": "Stuehus til landbrugsejendom",
    "120": "Fritliggende enfamiliehus",
    "121": "Sammenbygget enfamiliehus",
    "122": "Fritliggende enfamiliehus i tæt-lav bebyggelse",
    "130": "(UDFASES) Række-, kæde-, eller dobbelthus (lodret adskillelse mellem enhederne).",
    "131": "Række-, kæde- og klyngehus",
    "132": "Dobbelthus",
    "140": "Etagebolig-bygning, flerfamiliehus eller to-familiehus",
    "150": "Kollegium",
    "160": "Boligbygning til døgninstitution",
    "185": "Anneks i tilknytning til helårsbolig.",
    "190": "Anden bygning til helårsbeboelse",
    "210": "(UDFASES) Bygning til erhvervsmæssig produktion vedrørende landbrug, gartneri, råstofudvinding o. lign",
    "211": "Stald til svin",
    "212": "Stald til kvæg, får mv.",
    "213": "Stald til fjerkræ",
    "214": "Minkhal",
    "215": "Væksthus",
    "216": "Lade til foder, afgrøder mv.",
    "217": "Maskinhus, garage mv.",
    "218": "Lade til halm, hø mv.",
    "219": "Anden bygning til landbrug mv.",
    "220": "(UDFASES) Bygning til erhvervsmæssig produktion vedrørende industri, håndværk m.v. (fabrik, værksted o.lign.)",
    "221": "Bygning til industri med integreret produktionsapparat",
    "222": "Bygning til industri uden integreret produktionsapparat",
    "223": "Værksted",
    "229": "Anden bygning til produktion",
    "230": "(UDFASES) El-, gas-, vand- eller varmeværk, forbrændingsanstalt m.v.",
    "231": "Bygning til energiproduktion",
    "232": "Bygning til forsyning- og energidistribution",
    "233": "Bygning til vandforsyning",
    "234": "Bygning til håndtering af affald og spildevand",
    "239": "Anden bygning til energiproduktion og -distribution",
    "290": "(UDFASES) Anden bygning til landbrug, industri etc.",
    "310": "(UDFASES) Transport- og garageanlæg (fragtmandshal, lufthavnsbygning, banegårdsbygning, parkeringshus). Garage med plads til et eller to køretøjer registreres med anvendelseskode 910",
    "311": "Bygning til jernbane- og busdrift",
    "312": "Bygning til luftfart",
    "313": "Bygning til parkering- og transportanlæg",
    "314": "Bygning til parkering af flere end to køretøjer i tilknytning til boliger",
    "315": "Havneanlæg",
    "319": "Andet transportanlæg",
    "320": "(UDFASES) Bygning til kontor, handel, lager, herunder offentlig administration",
    "321": "Bygning til kontor",
    "322": "Bygning til detailhandel",
    "323": "Bygning til lager",
    "324": "Butikscenter",
    "325": "Tankstation",
    "329": "Anden bygning til kontor, handel og lager",
    "330": "(UDFASES) Bygning til hotel, restaurant, vaskeri, frisør og anden servicevirksomhed",
    "331": "Hotel, kro eller konferencecenter med overnatning",
    "332": "Bed & breakfast mv.",
    "333": "Restaurant, café og konferencecenter uden overnatning",
    "334": "Privat servicevirksomhed som frisør, vaskeri, netcafé mv.",
    "339": "Anden bygning til serviceerhverv",
    "390": "(UDFASES) Anden bygning til transport, handel etc",
    "410": "(UDFASES) Bygning til biograf, teater, erhvervsmæssig udstilling, bibliotek, museum, kirke o. lign.",
    "411": "Biograf, teater, koncertsted mv.",
    "412": "Museum",
    "413": "Bibliotek",
    "414": "Kirke eller anden bygning til trosudøvelse for statsanerkendte trossamfund",
    "415": "Forsamlingshus",
    "416": "Forlystelsespark",
    "419": "Anden bygning til kulturelle formål",
    "420": "(UDFASES) Bygning til undervisning og forskning (skole, gymnasium, forskningslabratorium o.lign.).",
    "421": "Grundskole",
    "422": "Universitet",
    "429": "Anden bygning til undervisning og forskning",
    "430": "(UDFASES) Bygning til hospital, sygehjem, fødeklinik o. lign.",
    "431": "Hospital og sygehus",
    "432": "Hospice, behandlingshjem mv.",
    "433": "Sundhedscenter, lægehus, fødeklinik mv.",
    "439": "Anden bygning til sundhedsformål",
    "440": "(UDFASES) Bygning til daginstitution",
    "441": "Daginstitution",
    "442": "Servicefunktion på døgninstitution",
    "443": "Kaserne",
    "444": "Fængsel, arresthus mv.",
    "449": "Anden bygning til institutionsformål",
    "490": "(UDFASES) Bygning til anden institution, herunder kaserne, fængsel o. lign.",
    "510": "Sommerhus",
    "520": "(UDFASES) Bygning til feriekoloni, vandrehjem o.lign. bortset fra sommerhus",
    "521": "Feriecenter, center til campingplads mv.",
    "522": "Bygning med ferielejligheder til erhvervsmæssig udlejning",
    "523": "Bygning med ferielejligheder til eget brug",
    "529": "Anden bygning til ferieformål",
    "530": "(UDFASES) Bygning i forbindelse med idrætsudøvelse (klubhus, idrætshal, svømmehal o. lign.)",
    "531": "Klubhus i forbindelse med fritid og idræt",
    "532": "Svømmehal",
    "533": "Idrætshal",
    "534": "Tribune i forbindelse med stadion",
    "535": "Bygning til træning og opstaldning af heste",
    "539": "Anden bygning til idrætformål",
    "540": "Kolonihavehus",
    "585": "Anneks i tilknytning til fritids- og sommerhus",
    "590": "Anden bygning til fritidsformål",
    "910": "Garage (med plads til et eller to køretøjer)",
    "920": "Carport",
    "930": "Udhus",
    "940": "Drivhus",
    "950": "Fritliggende overdækning",
    "960": "Fritliggende udestue",
    "970": "Tiloversbleven landbrugsbygning",
    "990": "Faldefærdig bygning",
    "999": "Ukendt bygning",
}

# floor types

label_bbr_floor_types = "ETAGER_AFVIG_KODE"

dict_bbr_floor_type_codes = {
    "0": "Bygningen har ikke afvigende etager",
    "10": "Bygningen har afvigende etager",
    "11": "Bygningen indeholder hems",
    "12": "Bygningen indeholder dobbelt højt rum",
    "13": "Bygningen indeholder indskudt etage",
}

# all codes

bbr_codes = {
    label_bbr_fuel_type: dict_bbr_fuel_type_codes,
    label_bbr_extra_heating: dict_bbr_extra_heating_codes,
    label_bbr_heating_system: dict_bbr_heating_system_codes,
    label_bbr_bygningpunkt_koorsys: label_bbr_bygningpunkt_koorsys_codes,
    label_bbr_building_uses: dict_bbr_building_use_codes,
    label_bbr_floor_types: dict_bbr_floor_type_codes,
}

# BBR labels

# label under which the building id can be found in the building entrance obj.

label_bbr_opgang_id = "Opgang_id"

label_bbr_entrance_id = "AdgAdr_id"

label_bbr_building_id = "Bygning_id"

label_bbr_bygningpunkt = "bygningspunkt"

label_bbr_bygningpunkt_coord = "koordinater"

label_bbr_opgang_id = "Opgang_id"

label_bbr_entrance_id = "AdgAdr_id"

label_bbr_building_id = "Bygning_id"

label_bbr_building_area = "BYG_BOLIG_ARL_SAML"

label_bbr_housing_area = "BYG_BEBYG_ARL"

label_bbr_number_floors = "ETAGER_ANT"

list_labels_bbr = [
    label_bbr_building_id,
    label_bbr_building_area,
    label_bbr_housing_area,
    label_bbr_fuel_type,
    label_bbr_building_uses,
    label_bbr_number_floors,
    label_bbr_floor_types,
    label_bbr_extra_heating,
]

# *****************************************************************************
# *****************************************************************************

def get_bbr_building_data_geodataframe(
    building_entrance_ids: list,
    selected_bbr_bdg_entrance_labels: list = SELECT_BBR_BDG_ENTR_LABELS,
    selected_bbr_building_labels: list = SELECT_BBR_BDG_LABELS,
    selected_bbr_building_point_labels: list = SELECT_BBR_BDG_POINT_LABELS,
) -> Tuple[GeoDataFrame, list]:

    # get data about building entrances
    dict_building_entrances, list_failures = fetch_building_entrance_data(
        building_entrance_ids
    )
    if selected_bbr_bdg_entrance_labels == None:
        # includes all labels
        selected_bbr_bdg_entrance_labels = BBR_BDG_ENTR_LABELS

    list_entries = [
        [
            value[bbr_key]
            for bbr_key in value
            if bbr_key in selected_bbr_bdg_entrance_labels
        ]
        for key, value in dict_building_entrances.items()
    ]
    df_building_entrances = DataFrame(
        data=list_entries,
        columns=selected_bbr_bdg_entrance_labels,
        index=dict_building_entrances.keys(),
    )

    # *************************************************************************
    # *************************************************************************

    # get data about buildings

    dict_buildings = fetch_building_data(df_building_entrances.index)
    if selected_bbr_building_labels == None:
        # includes all labels
        selected_bbr_building_labels = BBR_BDG_LABELS

    # create dataframe with building data
    list_entries = [
        [value[bbr_key] for bbr_key in value if bbr_key in selected_bbr_building_labels]
        for key, value in dict_buildings.items()
    ]
    df_buildings = DataFrame(
        data=list_entries,
        columns=selected_bbr_building_labels,
        index=dict_buildings.keys(),
    )

    # *************************************************************************
    # *************************************************************************

    # get building point data
    if selected_bbr_building_point_labels == None:
        # includes all labels

        selected_bbr_building_point_labels = BBR_BDG_POINT_LABELS
    dict_buildings_points = {
        building_entrance_id: dict_buildings[building_entrance_id][  # (
            label_bbr_bygningpunkt
        ]
        # if building_entrance_id in dict_buildings else None)
        for building_entrance_id in dict_building_entrances
        if building_entrance_id in dict_buildings  # excludes failures
    }

    # create dataframe with building point data
    list_entries = [
        [
            value[bbr_key]
            for bbr_key in value
            if bbr_key in selected_bbr_building_point_labels
        ]
        for key, value in dict_buildings_points.items()
    ]
    df_building_points = DataFrame(
        data=list_entries,
        columns=selected_bbr_building_point_labels,
        index=dict_buildings_points.keys(),
    )

    # merge all three, two at a time

    df_buildings = merge(
        df_buildings,
        df_building_points,
        right_index=True,
        left_index=True,
        suffixes=(None, "_x"),
    )  # adds "_x" to duplicate columns

    df_buildings = merge(
        df_buildings,
        df_building_entrances,
        right_index=True,
        left_index=True,
        suffixes=(None, "_y"),
    )  # adds "_y" to duplicate columns

    # *************************************************************************
    # *************************************************************************

    # create a geodataframe whose geometry is that of building points
    # specify the coordinate system
    coordinate_system = "EPSG:4326"  # latitude, longitude
    key_bbr_coordinate_system = 5  # WGS 84 = EPSG:4326

    # raise an error if different coordinates systems are being used
    for building_entrance_id in dict_building_entrances:
        if (
            dict_buildings[building_entrance_id][label_bbr_bygningpunkt][
                label_bbr_bygningpunkt_koorsys
            ]
            != key_bbr_coordinate_system
        ):
            raise NotImplementedError("Only WGS 84 coordinates can be used.")

    # create a dictionary with the building point geometries (i.e. points)
    dict_building_point_geometry = {
        building_entrance_id: Point(
            dict_buildings[building_entrance_id][label_bbr_bygningpunkt][
                label_bbr_bygningpunkt_coord
            ]
        )
        for building_entrance_id in dict_building_entrances
        if dict_buildings[building_entrance_id][label_bbr_bygningpunkt][
            label_bbr_bygningpunkt_koorsys
        ]
        == key_bbr_coordinate_system
    }

    # create geodataframe
    gdf_buildings = GeoDataFrame(
        data=df_buildings,
        geometry=GeoSeries(data=dict_building_point_geometry),
        crs=coordinate_system,
    )
    
    return gdf_buildings, list_failures

# *****************************************************************************
# *****************************************************************************

def fetch_building_entrance_data(building_entrance_ids: list) -> Tuple[dict, list]:

    # retrieve data about each node identified through OSM
    dict_building_entrances = {}
    list_failures = []
    # # determine the number of osm entries
    # number_building_entrance_ids = len(building_entrance_ids)
    # for each building entrance id
    for building_entrance_id in building_entrance_ids:
        # compose the url from which to get bbr data associated with the id
        _url = url_prefix_entrance + building_entrance_id
        try:
            # retrieve the building entrance data
            with urllib.request.urlopen(_url) as response:
                # parse the data
                bbr_entrance_data_json = json.loads(response.read().decode("utf-8"))
                # store the data
                if len(bbr_entrance_data_json) != 0:
                    for bbr_entry in bbr_entrance_data_json:
                        dict_building_entrances[
                            bbr_entry[label_bbr_building_id]
                        ] = bbr_entry
                else:
                    list_failures.append(building_entrance_id)
            response.close()
        except Exception:
            response.close()

    return dict_building_entrances, list_failures

# *****************************************************************************
# *****************************************************************************

def fetch_building_data(building_codes: list):

    # get data about each specific building
    dict_buildings = {}
    # for each building id
    for building_id in building_codes:
        # compose a url with it
        _url = url_prefix_buildings + building_id
        # try statement
        try:
            # retrieve that data
            with urllib.request.urlopen(_url) as response:
                # parse the data
                bbr_data = json.loads(response.read().decode("utf-8"))
                dict_buildings[building_id] = bbr_data
            response.close()
        except Exception:
            response.close()

    return dict_buildings

# *****************************************************************************
# *****************************************************************************
