# *****************************************************************************
# *****************************************************************************

import numpy as np
from geopandas import GeoDataFrame
from ...misc.utils import discrete_sinusoid_matching_integral
from ...misc.utils import create_profile_using_time_weighted_state
from ...misc.utils import generate_manual_state_correlated_profile
from ...misc.utils import generate_state_correlated_profile, generate_profile
from .bbr import label_bbr_entrance_id, label_bbr_housing_area

# *****************************************************************************
# *****************************************************************************

# labels

selected_bbr_adgang_labels = ["Opgang_id", "AdgAdr_id", "Bygning_id"]

selected_bbr_building_point_labels = [
    "KoorOest",
    "KoorNord",
    "KoorSystem",
    # "koordinater" # corresponds to a list, which cannot be written to a file
]

selected_bbr_building_labels = [
    "BYG_ANVEND_KODE",
    "OPFOERELSE_AAR",  # new
    "OMBYG_AAR",  # new
    "BYG_ARL_SAML",
    "BYG_BOLIG_ARL_SAML",
    "ERHV_ARL_SAML",
    "BYG_BEBYG_ARL",
    "GARAGE_INDB_ARL",
    "CARPORT_INDB_ARL",
    "UDHUS_INDB_ARL",
    "UDESTUE_ARL",
    "LukOverdaekAreal",
    "AFFALDSRUM_ARL",
    "ANDET_ARL",
    "OverdaekAreal",
    "AabenOverdaekAreal",
    "AdgangsAreal",
    "ETAGER_ANT",
    "ETAGER_AFVIG_KODE",
    "VARMEINSTAL_KODE",
    "OPVARMNING_KODE",
    "VARME_SUPPL_KODE",
    "BygPkt_id",
]


# label under which building entrance ids can be found in OSM
label_osm_entrance_id = "osak:identifier"

# *****************************************************************************
# *****************************************************************************

def heat_demand_dict_by_building_entrance(
    gdf_osm: GeoDataFrame,
    gdf_buildings: GeoDataFrame,
    number_intervals: int,
    time_interval_durations: list,
    bdg_specific_demand: dict,
    bdg_min_max_ratio: dict,
    bdg_demand_phase_shift: dict = None,
    key_osm_entr_id: str = label_osm_entrance_id,
    key_bbr_entr_id: str = label_bbr_entrance_id,
    avg_state: list = None,
    state_correlates_with_output: bool = False,
    deviation_gain: float = 1.0,
    solver: str = 'glpk',
    **kwargs
) -> dict:
    
    # initialise dict for each building entrance
    demand_dict = {}

    # for each building entrance
    for osm_index in gdf_osm.index:
        # initialise dict for each building consumption point
        heat_demand_profiles = []

        # find the indexes for each building leading to the curr. cons. point
        building_indexes = gdf_buildings[
            gdf_buildings[key_bbr_entr_id] == gdf_osm.loc[osm_index][key_osm_entr_id]
        ].index

        # for each building
        for building_index in building_indexes:
            # get relevant data
            area = gdf_buildings.loc[building_index][label_bbr_housing_area]
            # generate the profile            
            if type(avg_state) == type(None):
                # ignore states
                heat_demand_profiles.append(
                    np.array(
                        discrete_sinusoid_matching_integral(
                            bdg_specific_demand[building_index] * area,
                            time_interval_durations=time_interval_durations,
                            min_max_ratio=bdg_min_max_ratio[building_index],
                            phase_shift_radians=(
                                bdg_demand_phase_shift[building_index]
                            ),
                        )
                    )
                )

            elif type(deviation_gain) == type(None):
                # states matter but the gain must be determined
                # heat_demand_profiles.append(
                #     np.array(
                #         create_profile_using_time_weighted_state(
                #             integration_result=(
                #                 bdg_specific_demand[building_index] * area
                #             ),
                #             avg_state=avg_state,
                #             time_interval_durations=time_interval_durations,
                #             min_max_ratio=bdg_min_max_ratio[building_index],
                #             state_correlates_with_output=state_correlates_with_output,
                #         )
                #     )
                # )
                heat_demand_profiles.append(
                    np.array(
                        generate_state_correlated_profile(
                            integration_result=(
                                bdg_specific_demand[building_index] * area
                            ), 
                            states=avg_state, 
                            time_interval_durations=time_interval_durations,
                            states_correlate_profile=state_correlates_with_output,
                            min_max_ratio=bdg_min_max_ratio[building_index],
                            solver=solver
                            )
                        )
                    )
            else:
                # states matter and the gain is predefined
                heat_demand_profiles.append(
                    np.array(
                        generate_manual_state_correlated_profile(
                            integration_result=(
                                bdg_specific_demand[building_index] * area
                            ), 
                            states=avg_state, 
                            time_interval_durations=time_interval_durations, 
                            deviation_gain=deviation_gain
                            )
                        )
                    )

        # add the profiles, time step by time step
        if len(heat_demand_profiles) == 0:
            final_profile = []
        else:
            final_profile = sum(profile for profile in heat_demand_profiles)

        # store the demand profile
        demand_dict[osm_index] = final_profile

    # return
    return demand_dict

# *****************************************************************************
# *****************************************************************************

# TODO: allow reusing the gain

def heat_demand_profiles(
    gdf_osm: GeoDataFrame,
    gdf_buildings: GeoDataFrame,
    time_interval_durations: list,
    assessments: list,
    annual_heat_demand: dict,
    air_temperature: dict = None, 
    reuse_deviation_gain: bool = True,
    **kwargs
) -> dict:    
    
    # calculate the total area
    total_area = total_heating_area(gdf_osm, gdf_buildings)
    # initialise data dict
    heat_demand_dict = {}
    # for each building entrance
    for osm_index in gdf_osm.index:
        # initialise dict for each building entrance
        bdg_entr_dict = {}
        # find the indexes for each building leading to the curr. cons. point
        building_indexes = gdf_buildings[
            gdf_buildings[label_bbr_entrance_id] == gdf_osm.loc[osm_index][label_osm_entrance_id]
        ].index
        for q in assessments:
            # define the specific heat demand
            specific_demand = annual_heat_demand[q]/total_area
            # 
            # initialise dict for each building consumption point
            bdg_profile_list = []
            # for each building
            for building_index in building_indexes:
                # get relevant data
                area = gdf_buildings.loc[building_index][label_bbr_housing_area]
                # handle states
                if type(air_temperature) != type(None):
                    kwargs['states'] = air_temperature[q]
                # append the profile for each building to the list
                _profile = generate_profile(
                    integration_result=specific_demand*area, 
                    time_interval_durations=time_interval_durations, 
                    **kwargs
                    # min_max_ratio, 
                    # states, 
                    # states_correlate_profile, 
                    # solver, 
                    # deviation_gain
                    )
                bdg_profile_list.append(np.array(_profile))
            # aggregate profiles for the same building entrance
            bdg_entr_profile = (
                [] 
                if len(bdg_profile_list) == 0 else 
                sum(profile for profile in bdg_profile_list)
                )
            # store the demand profile
            bdg_entr_dict[q] = bdg_entr_profile
        # add to the main dict
        heat_demand_dict[osm_index] = bdg_entr_dict
        
    return heat_demand_dict, total_area

# *****************************************************************************
# *****************************************************************************

def total_heating_area(
    gdf_osm: GeoDataFrame,
    gdf_buildings: GeoDataFrame,
    key_osm_entr_id: str = label_osm_entrance_id,
    key_bbr_entr_id: str = label_bbr_entrance_id,
) -> float:
    area = 0
    for osm_index in gdf_osm.index:
        # find the indexes for each building leading to the curr. cons. point
        building_indexes = gdf_buildings[
            gdf_buildings[label_bbr_entrance_id]
            == gdf_osm.loc[osm_index][label_osm_entrance_id]
        ].index
        # for each building
        for building_index in building_indexes:
            # get relevant data
            area += gdf_buildings.loc[building_index][label_bbr_housing_area]
    return area

# *****************************************************************************
# *****************************************************************************