# imports

# standard imports

from numbers import Real

# local, external imports

# from topupheat.pipes.trenches import SupplyReturnPipeTrenchWithIdenticalPipes
from topupheat.pipes.trenches import SupplyReturnPipeTrench

# local, internal imports
from ...problems.esipp.network import ArcsWithoutProportionalLosses

# TODO: add way to define polarity of cash flows

from ...data.finance.utils import ArcInvestments

# *****************************************************************************
# *****************************************************************************

# scenarios: different datasets/trench objects
# options: fully specified trenches, including prices

# constants

KEY_DHT_OPTIONS_OBJ = "trench"
KEY_DHT_LENGTH = "length"
KEY_DHT_UCF = "capacity_unit_conversion_factor"
KEY_HHT_DHT_PIPES = "pipes"
KEY_HHT_STD_PIPES = "pipe_tuple"

# *****************************************************************************
# *****************************************************************************


class PipeTrenchOptions(ArcsWithoutProportionalLosses):
    "A class for defining investments in district heating trenches."

    def __init__(
        self,
        trench: SupplyReturnPipeTrench,
        name: str,
        length: float,
        specific_capacity_cost: float or list = None,
        minimum_cost: list or tuple = None,  # default: pipes
        capacity_is_instantaneous: bool = False,
        unit_conversion_factor: float = 1.0,
    ):
        # store the unit conversion
        self.unit_conversion_factor = unit_conversion_factor
        # keep the trench object
        self.trench = trench
        # keep the trench length
        self.length = (
            [length for i in range(trench.number_options())]
            if trench.vector_mode
            else length
        )
        # determine the rated heat capacity
        rhc = trench.rated_heat_capacity(unit_conversion_factor=unit_conversion_factor)
        # initialise the object using the mother class
        ArcsWithoutProportionalLosses.__init__(
            self,
            name=name,
            static_loss=None,
            capacity=[rhc] if isinstance(rhc, Real) else rhc,
            minimum_cost=minimum_cost,
            specific_capacity_cost=(
                0
                if type(specific_capacity_cost) == type(None)
                else specific_capacity_cost
            ),
            capacity_is_instantaneous=False,
        )
        # initialise the minimum cost
        if type(minimum_cost) == type(None):
            self.set_minimum_cost()

    # *************************************************************************
    # *************************************************************************

    def set_minimum_cost(self, minimum_cost=None):
        # minimum arc cost
        # if no external minimum cost list was provided, calculate it
        if type(minimum_cost) == type(None):
            # use the specific pipe cost that features in the database
            if self.trench.vector_mode:
                # multiple options
                self.minimum_cost = tuple(
                    (
                        pipe.sp * length  # twin pipes: one twin pipe
                        if self.trench.twin_pipes
                        else pipe.sp * length * 2
                    )  # single pipes: two single pipes
                    for pipe, length in zip(self.trench.supply_pipe, self.length)
                )
            else:  # only one option
                self.minimum_cost = (self.trench.supply_pipe.sp * self.length,)
        else:  # use an external minimum cost
            self.minimum_cost = tuple(minimum_cost)

    # *************************************************************************
    # *************************************************************************

    def set_capacity(self, **kwargs):
        # retrieve the rated heat capacity
        rhc = self.trench.rated_heat_capacity(**kwargs)
        if self.trench.vector_mode:
            # multiple options, rhc is already a list of values
            self.capacity = rhc
        else:
            # one option, rhc is one value
            self.capacity = (rhc,)

    # *************************************************************************
    # *************************************************************************

    def set_static_losses(
        self,
        scenario_key,
        ground_thermal_conductivity: float or list,
        ground_air_heat_transfer_coefficient: float or list,
        time_interval_duration: float or list,
        temperature_surroundings: float or list,
        length: float or list = None,
        unit_conversion_factor: float = None,
        **kwargs
    ):
        hts = self.trench.heat_transfer_surroundings(
            ground_thermal_conductivity=ground_thermal_conductivity,
            ground_air_heat_transfer_coefficient=(ground_air_heat_transfer_coefficient),
            time_interval_duration=time_interval_duration,
            temperature_surroundings=temperature_surroundings,
            length=(self.length if type(length) == type(None) else length),
            unit_conversion_factor=(
                self.unit_conversion_factor
                if type(unit_conversion_factor) == type(None)
                else unit_conversion_factor
            ),
            **kwargs
        )

        if self.trench.vector_mode:
            # multiple options: hts is a vector
            if hasattr(self, "static_loss") and type(self.static_loss) != type(None):
                # update the static loss dictionary
                if type(hts[0]) == list:
                    # multiple time intervals
                    self.static_loss.update(
                        {
                            (h, scenario_key, k): hts[h][k]
                            for h, hts_h in enumerate(hts)
                            for k, hts_hk in enumerate(hts_h)
                        }
                    )
                else:  # not a list: one time interval
                    self.static_loss.update(
                        {(h, scenario_key, 0): hts[h] for h, hts_h in enumerate(hts)}
                    )
            else:
                # no static loss dictionary, create it
                if type(hts[0]) == list:
                    # multiple time intervals
                    self.static_loss = {
                        (h, scenario_key, k): hts[h][k]
                        for h, hts_h in enumerate(hts)
                        for k, hts_hk in enumerate(hts_h)
                    }
                else:  # not a list: one time interval
                    self.static_loss = {
                        (h, scenario_key, 0): hts[h] for h, hts_h in enumerate(hts)
                    }
        else:
            # one option: hts might be a number
            if hasattr(self, "static_loss") and type(self.static_loss) != type(None):
                # update the static loss dictionary
                if not isinstance(hts, Real):
                    # multiple time intervals
                    self.static_loss.update(
                        {(0, scenario_key, k): hts[k] for k, hts_k in enumerate(hts)}
                    )
                else:  # not a list: one time interval
                    self.static_loss.update({(0, scenario_key, 0): hts})
            else:
                # no static loss dictionary, create it
                if not isinstance(hts, Real):
                    # multiple time intervals
                    self.static_loss = {
                        (0, scenario_key, k): hts_k for k, hts_k in enumerate(hts)
                    }
                else:  # not a list: one time interval
                    self.static_loss = {(0, scenario_key, 0): hts}


# *****************************************************************************
# *****************************************************************************


class PipeTrenchInvestments(ArcInvestments, PipeTrenchOptions):
    "A class for defining investments in district heating trenches."

    def __init__(
        self,
        trench: SupplyReturnPipeTrench,
        name: str,
        length: float,
        investments: tuple,
        static_loss: dict = None,
        specific_capacity_cost: float or list = None,
        capacity_is_instantaneous: bool = False,
        unit_conversion_factor: float = 1.0,
        **kwargs
    ):
        # store the unit conversion
        self.unit_conversion_factor = unit_conversion_factor
        # keep the trench object
        self.trench = trench
        # keep the trench length
        self.length = (
            [length for i in range(trench.number_options())]
            if trench.vector_mode
            else length
        )
        # determine the rated heat capacity
        rhc = trench.rated_heat_capacity(unit_conversion_factor=unit_conversion_factor)
        # initialise the object using the mother class
        ArcInvestments.__init__(
            self,
            investments=investments,
            name=name,
            efficiency=None,
            efficiency_reverse=None,
            static_loss=static_loss,
            capacity=[rhc] if isinstance(rhc, Real) else rhc,
            specific_capacity_cost=(
                0
                if type(specific_capacity_cost) == type(None)
                else specific_capacity_cost
            ),
            capacity_is_instantaneous=False,
            validate=False,
        )

    # # *************************************************************************
    # # *************************************************************************

    # def set_minimum_cost(self, minimum_cost = None):

    #     # minimum arc cost
    #     # if no external minimum cost list was provided, calculate it
    #     if type(minimum_cost) == type(None):
    #         # use the specific pipe cost that features in the database
    #         if self.trench.vector_mode:
    #             # multiple options
    #             self.minimum_cost = tuple(
    #                 (pipe.sp*length # twin pipes: one twin pipe
    #                  if self.trench.twin_pipes else
    #                  pipe.sp*length*2) # single pipes: two single pipes
    #                 for pipe, length in zip(
    #                         self.trench.supply_pipe,
    #                         self.length
    #                         )
    #                 )
    #         else: # only one option
    #             self.minimum_cost = (self.trench.supply_pipe.sp*self.length,)
    #     else: # use an external minimum cost
    #         self.minimum_cost = tuple(minimum_cost)

    # # *************************************************************************
    # # *************************************************************************

    # def set_capacity(self, **kwargs):
    #     # retrieve the rated heat capacity
    #     rhc = self.trench.rated_heat_capacity(**kwargs)
    #     if self.trench.vector_mode:
    #         # multiple options, rhc is already a list of values
    #         self.capacity = rhc
    #     else:
    #         # one option, rhc is one value
    #         self.capacity = (rhc,)

    # # *************************************************************************
    # # *************************************************************************

    # def set_static_losses(
    #         self,
    #         scenario_key,
    #         ground_thermal_conductivity: float or list,
    #         ground_air_heat_transfer_coefficient: float or list,
    #         time_interval_duration: float or list,
    #         temperature_surroundings: float or list,
    #         length: float or list = None,
    #         unit_conversion_factor: float = None,
    #         **kwargs):

    #     hts = self.trench.heat_transfer_surroundings(
    #         ground_thermal_conductivity=ground_thermal_conductivity,
    #         ground_air_heat_transfer_coefficient=(
    #             ground_air_heat_transfer_coefficient),
    #         time_interval_duration=time_interval_duration,
    #         temperature_surroundings=temperature_surroundings,
    #         length=(
    #             self.length
    #             if type(length) == type(None) else
    #             length
    #             ),
    #         unit_conversion_factor=(
    #             self.unit_conversion_factor
    #             if type(unit_conversion_factor) == type(None) else
    #             unit_conversion_factor
    #             ),
    #         **kwargs)

    #     if self.trench.vector_mode:
    #         # multiple options: hts is a vector
    #         if (hasattr(self, "static_loss") and
    #             type(self.static_loss) != type(None)):
    #             # update the static loss dictionary
    #             if type(hts[0]) == list:
    #                 # multiple time intervals
    #                 self.static_loss.update({
    #                     (h, scenario_key, k): hts[h][k]
    #                     for h, hts_h in enumerate(hts)
    #                     for k, hts_hk in enumerate(hts_h)
    #                     })
    #             else: # not a list: one time interval
    #                 self.static_loss.update({
    #                     (h, scenario_key, 0): hts[h]
    #                     for h, hts_h in enumerate(hts)
    #                     })
    #         else:
    #             # no static loss dictionary, create it
    #             if type(hts[0]) == list:
    #                 # multiple time intervals
    #                 self.static_loss = {
    #                     (h, scenario_key, k): hts[h][k]
    #                     for h, hts_h in enumerate(hts)
    #                     for k, hts_hk in enumerate(hts_h)
    #                     }
    #             else: # not a list: one time interval
    #                 self.static_loss = {
    #                     (h, scenario_key, 0): hts[h]
    #                     for h, hts_h in enumerate(hts)
    #                     }
    #     else:
    #         # one option: hts might be a number
    #         if (hasattr(self, "static_loss") and
    #             type(self.static_loss) != type(None)):
    #             # update the static loss dictionary
    #             if not isinstance(hts, Real):
    #                 # multiple time intervals
    #                 self.static_loss.update({
    #                     (0, scenario_key, k): hts[k]
    #                     for k, hts_k in enumerate(hts)
    #                     })
    #             else: # not a list: one time interval
    #                 self.static_loss.update({
    #                     (0, scenario_key, 0): hts
    #                     })
    #         else:
    #             # no static loss dictionary, create it
    #             if not isinstance(hts, Real):
    #                 # multiple time intervals
    #                 self.static_loss = {
    #                     (0, scenario_key, k): hts_k
    #                     for k, hts_k in enumerate(hts)
    #                     }
    #             else: # not a list: one time interval
    #                 self.static_loss = {
    #                     (0, scenario_key, 0): hts
    #                     }


# *****************************************************************************
# *****************************************************************************


class ExistingPipeTrench(PipeTrenchOptions):
    "A class for existing pipe trenches."

    def __init__(self, option_selected: int, **kwargs):
        # initialise
        PipeTrenchOptions.__init__(
            self,
            minimum_cost=[0 for i in range(kwargs["trench"].number_options())],
            **kwargs
        )
        # define the option that already exists
        self.options_selected[option_selected] = True


# *****************************************************************************
# *****************************************************************************
