# imports

# standard

# local, external

import osmnx as ox
import contextily as cx
import matplotlib.pyplot as plt
import numpy as np

# local, internal

from ...problems.esipp.network import Network
from .network import PipeTrenchOptions
from topupheat.pipes.trenches import SupplyReturnPipeTrench
from numbers import Real

# *****************************************************************************
# *****************************************************************************


def cost_pipes(trench: SupplyReturnPipeTrench, length: float or tuple) -> tuple:
    """
    Returns the costs of each trench option for a given trench length.

    Parameters
    ----------
    trench : SupplyReturnPipeTrench
        The object describing the trench options.
    length : float or tuple
        The trench length in meters.

    Raises
    ------
    ValueError
        This error is raised if unrecognised inputs are inserted.

    Returns
    -------
    tuple
        A tuple of the pipe costs for each option.

    """
    # use the specific pipe cost that features in the database
    if trench.vector_mode:
        # multiple options
        if type(length) == tuple and len(length) == trench.number_options():
            # multiple trench lengths
            return tuple(
                (
                    pipe.sp * length  # twin pipes: one twin pipe
                    if trench.twin_pipes
                    else pipe.sp * length * 2
                )  # single pipes: two single pipes
                for pipe, length in zip(trench.supply_pipe, length)
            )
        elif isinstance(length, Real):
            # one trench length
            return tuple(
                (
                    pipe.sp * length  # twin pipes: one twin pipe
                    if trench.twin_pipes
                    else pipe.sp * length * 2
                )  # single pipes: two single pipes
                for pipe in trench.supply_pipe
            )
        else:
            raise ValueError("Unrecognised input combination.")
    elif not trench.vector_mode and isinstance(length, Real):
        # only one option
        return (trench.supply_pipe.sp * length,)
    else:  # only one option
        raise ValueError("Unrecognised input combination.")

    # # keep the trench length
    # self.length = (
    #     [length for i in range(trench.number_options())]
    #     if trench.vector_mode else
    #     length
    #     )


# *****************************************************************************
# *****************************************************************************


def summarise_network_by_pipe_technology(
    network: Network, print_output: bool = False
) -> dict:
    "A method to summarise a network by pipe technology."

    # *************************************************************************
    # *************************************************************************

    # create a dictionary that compiles the lengths of each arc technology

    length_dict = {}

    # *************************************************************************
    # *************************************************************************

    # for each arc
    for arc_key in network.edges(keys=True):
        # check if it is a PipeTrench object
        if not isinstance(
            network.edges[arc_key][Network.KEY_ARC_TECH], PipeTrenchOptions
        ):
            # if not, skip arc
            continue

        # for each arc technology option
        for h, tech_option in enumerate(
            network.edges[arc_key][Network.KEY_ARC_TECH].options_selected
        ):
            # check if the tech option was selected
            if tech_option:
                # technology option was selected
                # get the length of the arc
                arc_length = (
                    network.edges[arc_key][Network.KEY_ARC_TECH].length[h]
                    if type(network.edges[arc_key][Network.KEY_ARC_TECH].length) == list
                    else network.edges[arc_key][Network.KEY_ARC_TECH].length
                )
                # identify the option
                tech_option_label = network.edges[arc_key][
                    Network.KEY_ARC_TECH
                ].trench.printable_description(h)
                # if the arc technology has been previously selected...
                if tech_option_label in length_dict:
                    # ...increment the total length
                    length_dict[tech_option_label] += arc_length
                else:
                    # if not, add a new arc technology to the dictionary
                    length_dict[tech_option_label] = arc_length

    # *************************************************************************
    # *************************************************************************

    if print_output:
        print("printing the arc technologies selected by pipe size...")
        for key, value in sorted(
            (tech, length) for tech, length in length_dict.items()
        ):
            print(str(key) + ": " + str(value))
        print("total: " + str(sum(length_dict.values())))

    return length_dict

    # *************************************************************************
    # *************************************************************************


# *****************************************************************************
# *****************************************************************************


def plot_network_layout(
    network: Network,
    include_basemap: bool = False,
    figure_size: tuple = (25, 25),
    min_linewidth: float = 1.0,
    max_linewidth: float = 3.0,
    legend_fontsize: float = 20.0,
    basemap_zoom_level: float = 15,
    legend_location: str = "lower left",
    legend_with_brand_model: bool = False,
    legend_transparency: float = None,
):
    # convert graph object to GDF

    _, my_gdf_arcs = ox.graph_to_gdfs(network)

    # convert to final plot CRS

    my_gdf = my_gdf_arcs.to_crs(epsg=3857)

    # dict: keys are the pipe tuples and the values are lists of edge keys

    arc_tech_summary_dict = {}

    # for each edge
    for arc_key in my_gdf.index:
        # check if it is a PipeTrenchOptions object
        if not isinstance(
            network.edges[arc_key][Network.KEY_ARC_TECH], PipeTrenchOptions
        ):
            # if not, skip arc
            continue

        # find the trench's description, if it was selected

        try:
            selected_option = (
                my_gdf[Network.KEY_ARC_TECH]
                .loc[arc_key]
                .trench.printable_description(
                    my_gdf[Network.KEY_ARC_TECH]
                    .loc[arc_key]
                    .options_selected.index(True)
                )
            )
        except ValueError:
            continue

        # if the pipe tuple already exists as a key in the dict
        if selected_option in arc_tech_summary_dict:
            # append the edge_key to the list obtained via that pipe tuple key
            arc_tech_summary_dict[selected_option].append(arc_key)
        else:  # if not
            # add a new dict entry whose key is the pipe tuple and create a list
            arc_tech_summary_dict[selected_option] = [arc_key]

    list_sorted = sorted(
        (int(printable_description[2:]), printable_description)
        for printable_description in arc_tech_summary_dict.keys()
    )
    (list_sorted_dn, list_sorted_descriptions) = list(map(list, zip(*list_sorted)))

    list_arc_widths = (
        [
            min_linewidth
            + (max_linewidth - min_linewidth) * iteration / (len(list_sorted_dn) - 1)
            for iteration, _ in enumerate(list_sorted_dn)
        ]
        if len(list_sorted_dn) != 1
        else [(max_linewidth + min_linewidth) / 2]
    )

    # *************************************************************************
    # *************************************************************************

    fig, ax = plt.subplots(1, 1)

    fig.set_size_inches(*figure_size)

    for description, arc_width in zip(list_sorted_descriptions, list_arc_widths):
        # prepare plot

        my_gdf.loc[arc_tech_summary_dict[description]].plot(
            edgecolor="k", legend=True, linewidth=arc_width, ax=ax
        )

    # adjust legend labels

    ax.legend(
        list_sorted_descriptions,
        fontsize=legend_fontsize,
        loc=legend_location,
        framealpha=(
            legend_transparency if type(legend_transparency) != type(None) else None
        ),
    )

    # add base map

    if include_basemap:
        cx.add_basemap(
            ax,
            zoom=basemap_zoom_level,
            source=cx.providers.OpenStreetMap.Mapnik,
            # crs=gdf_map.crs,
        )

    # *************************************************************************
    # *************************************************************************


# *****************************************************************************
# *****************************************************************************


def plot_heating_demand(
    losses: list,
    end_use_demand: list,
    labels: list,
    ylabel: str = "Heating demand [MWh]",
    title: str = "Heat demand by month",
):
    energy_totals = {
        "Losses (optimised)": np.array(losses),
        "End use (estimated)": np.array(end_use_demand),
    }
    colors = {
        "Losses (optimised)": "tab:orange",
        "End use (estimated)": "tab:blue",
    }
    # width = 0.8  # the width of the bars: can also be len(x) sequence

    # make sure the grid lines are behind the bars
    zorder_bars = 3
    zorder_grid = 0

    fig, ax = plt.subplots()
    bottom = np.zeros(len(labels))
    figure_size = (8, 4)
    fig.set_size_inches(figure_size[0], figure_size[1])
    for energy_category, energy_total in energy_totals.items():
        p = ax.bar(
            labels,
            energy_total,
            label=energy_category,
            bottom=bottom,
            color=colors[energy_category],
            zorder=zorder_bars,
        )
        bottom += energy_total

        ax.bar_label(p, fmt="{:,.0f}", label_type="center")
        # ax.bar_label(p, fmt='{:,.0f}')

    ax.grid(zorder=zorder_grid)  # zorder=0 to make the grid
    ax.set(ylabel=ylabel, title=title)
    ax.legend()

    plt.show()


# *****************************************************************************
# *****************************************************************************
