# *****************************************************************************
# *****************************************************************************

from math import prod

from statistics import mean

# *****************************************************************************
# *****************************************************************************

# TODO: enable swapping the polarity


class Investment:
    """This class is meant to enable analysis of specific investments."""

    # self.discount_rates: N samples
    # self.net_cash_flows: N+1 samples

    # TODO: consider using dicts to make things more intuitive, time-wise

    def __init__(
        self,
        discount_rates: list,
        net_cash_flows: list = None,
        discount_rate: float = None,
        analysis_period_span: int = None,
    ):
        """
        Create an object for investment analysis using typical information.

        Parameters
        ----------
        discount_rates : list
            A list specifying the discount rates to use in the calculation.
        net_cash_flows : list
            A list specifying the net cash flows to use in the calculation.

        Returns
        -------
        None.

        """
        # validate the inputs

        if type(discount_rates) != type(None):
            # discount_rates is not None:

            if type(discount_rates) != tuple:
                raise TypeError("The discount rates must be provided as a tuple.")

            self.discount_rates = tuple(discount_rates)

            self.analysis_period_span = len(self.discount_rates)

            if self.analysis_period_span <= 0:
                raise ValueError(
                    "The duration of the period under analysis must be " + "positive."
                )

        else:
            # discount_rates is None:
            # discount rate must be positive real under 1
            # analysis_period_span must be an int

            if type(discount_rate) != float:
                raise TypeError("The discount rate must be provided as a float.")

            if discount_rate <= 0 or discount_rate >= 1:
                raise ValueError(
                    "The discount rate must be in the open interval between 0"
                    + " and 1."
                )

            if type(analysis_period_span) != int:
                raise TypeError(
                    "The duration of the period under consideration must be "
                    + "provided as an integer."
                )

            if analysis_period_span <= 0:
                raise ValueError(
                    "The duration of the period under analysis must be " + "positive."
                )

            self.analysis_period_span = analysis_period_span

            self.discount_rates = tuple(
                discount_rate for i in range(self.analysis_period_span)
            )

        # check the net cash flows

        if type(net_cash_flows) != type(None):
            if type(net_cash_flows) != list:
                raise TypeError("The net cash flows must be provided as a list.")

            if len(net_cash_flows) != self.analysis_period_span + 1:
                raise ValueError("The inputs are consistent in terms of length.")

            self.net_cash_flows = list(net_cash_flows)

        else:
            # net_cash_flows is None: initialise it as a list of zeros

            self.net_cash_flows = list(0 for i in range(self.analysis_period_span + 1))

        # discount factors

        self.discount_factors = tuple(
            discount_factor(self.discount_rates[:i])
            for i in range(self.analysis_period_span + 1)
        )

    # *************************************************************************
    # *************************************************************************

    def add_investment(
        self,
        investment: float,
        investment_period: int,
        investment_longevity: int,
        commissioning_delay_after_investment: int = 0,
        salvage_value_method: str = "annuity",
    ):
        if salvage_value_method == "annuity":
            mean_discount_rate = mean(self.discount_rates)

            residual_value = salvage_value_annuity(
                investment=investment,
                investment_longevity=investment_longevity,
                investment_period=investment_period,
                discount_rate=mean_discount_rate,
                analysis_period_span=self.analysis_period_span,
            )

            self.net_cash_flows[investment_period] += investment
            self.net_cash_flows[self.analysis_period_span] += -residual_value

        else:
            residual_value = salvage_value_linear_depreciation(
                investment=investment,
                investment_period=investment_period,
                investment_longevity=investment_longevity,
                analysis_period_span=self.analysis_period_span,
                commissioning_delay_after_investment=(
                    commissioning_delay_after_investment
                ),
            )

            self.net_cash_flows[investment_period] += investment
            self.net_cash_flows[self.analysis_period_span] += -residual_value

    # *************************************************************************
    # *************************************************************************

    def add_operational_cash_flows(
        self, cash_flow: float or int, start_period: int, longevity: int = None
    ):
        """Adds a sequence of cash flows to the analysis."""

        if type(longevity) == type(None):
            # until the planning horizon

            for i in range(self.analysis_period_span - start_period + 1):
                # add operational cash flows

                self.net_cash_flows[i + start_period] += cash_flow

        else:
            # limited longevity

            for i in range(longevity):
                if i + start_period >= self.analysis_period_span + 1:
                    break

                # add operational cash flows

                self.net_cash_flows[i + start_period] += cash_flow

    # *************************************************************************
    # *************************************************************************

    def net_present_value(self):
        """Returns the net present value for the investment under analysis."""

        return npv(self.discount_rates, self.net_cash_flows)

    # *************************************************************************
    # *************************************************************************


# *****************************************************************************
# *****************************************************************************


def npv(
    discount_rates: list, net_cash_flows: list, return_discount_factors: bool = False
) -> float or tuple:
    """
    Calculates the net present value using the information provided.

    Parameters
    ----------
    discount_rates : list
        A list with the discount rates for each period after the present.
    net_cash_flows : list
        A list with the net cash flows for each period including the present.
    return_discount_factors : bool, optional
        If True, return the discount factors too. The default is False.

    Raises
    ------
    ValueError
        This error is raised if the lists for the discount rates and the
        net cash flows do not have the same size.

    Returns
    -------
    float or tuple
        If return_discount_factors is False, returns the net present value.
        If True, returns the net present value in addition to a list with
        the discount factors used in the calculation.

    """
    # check sizes

    if len(discount_rates) != len(net_cash_flows) - 1:
        # the inputs do not match, return None

        raise ValueError("The inputs are inconsistent.")

    discount_factors = [
        discount_factor(discount_rates[:t]) for t in range(len(discount_rates) + 1)
    ]

    if return_discount_factors:
        return (
            sum(
                ncf_t * df_t for (ncf_t, df_t) in zip(net_cash_flows, discount_factors)
            ),
            discount_factors,
        )

    else:
        return sum(
            ncf_t * df_t for (ncf_t, df_t) in zip(net_cash_flows, discount_factors)
        )


# *****************************************************************************
# *****************************************************************************


def discount_factor(discount_rates: list) -> float:
    """
    Return the discount factor consistent with the discount rates provided.

    To calculate the net present value, we need to quantify the effect time
    has on the net cash flows. This amounts to a factor, which depends on
    the discount rates between the time of the cash flow and the present.

    Parameters
    ----------
    discount_rates : list
        A list with the discount rates for each time interval between a
        given time interval and the present. The order is irrelevant.

    Returns
    -------
    float
        The discount factor consistent with the discount rates provided. It
        uses all discount rates in the list.

    """
    return prod([1 / (1 + i) for i in discount_rates])


# *****************************************************************************
# *****************************************************************************


def salvage_value_linear_depreciation(
    investment: int or float,
    investment_period: int,
    investment_longevity: int,
    analysis_period_span: int,
    commissioning_delay_after_investment: int = 1,
) -> float:
    """
    Determine an asset\'s salvage value by the end of an analysis period.

    The depreciation is assumed to be linear: the asset is initially rated at
    100% of the investment made and then 0% by the time it is no longer usable.

    The salvage value is the asset\'s value after the analysis period.

    Parameters
    ----------
    investment : int or float
        The sum initially invested in the asset.
    investment_period : int
        The period during which the investment is made or reported. It has to
        be between zero (present) and analysis_period_span+1 (planning horizon).
    investment_longevity : int
        The investment\'s expected longevity expressed as a number of periods.
    commissioning_delay_after_investment : int, optional
        The delay between the investment and the period within which the asset
        is used, specified in reporting periods. The default is one (1).

    Raises
    ------
    ValueError
        This error is raised if the investment takes place outside the planning
        period under analysis.

    Returns
    -------
    float
        The salvage value.

    """
    if investment_period >= analysis_period_span + 1:
        raise ValueError(
            "The investment has to be made within the period being analysed."
        )

    # calculate the salvage value

    return (
        (
            investment_longevity
            + investment_period
            + commissioning_delay_after_investment
            - 1
            - analysis_period_span
        )
        * investment
        / investment_longevity
    )


# *****************************************************************************
# *****************************************************************************


def salvage_value_annuity(
    investment: int or float,
    discount_rate: float,
    investment_longevity: int,
    investment_period: int,
    analysis_period_span: int,
) -> float:
    npv_salvage = present_salvage_value_annuity(
        investment=investment,
        investment_longevity=investment_longevity,
        investment_period=investment_period,
        discount_rate=discount_rate,
        analysis_period_span=analysis_period_span,
        return_annuity=False,
    )

    return npv_salvage / discount_factor(
        tuple(discount_rate for i in range(analysis_period_span))
    )


# *****************************************************************************
# *****************************************************************************


def annuity(
    investment: int or float, investment_longevity: int, discount_rate: float
) -> float:
    "Returns the annuity value for a given investment sum and longevity."

    return (
        investment
        * discount_rate
        / (1 - (1 + discount_rate) ** (-investment_longevity))
    )


# *****************************************************************************
# *****************************************************************************


def present_salvage_value_annuity(
    investment: int or float,
    investment_longevity: int,
    investment_period: int,
    discount_rate: float,
    analysis_period_span: int,
    return_annuity: bool = False,
) -> float:
    """
    Calculates the present value of an asset after a given analysis period.

    The calculation is based on the annuity method, which assumes that the
    investment could produce a steady revenue stream for a given period after
    the investment is made -- the investment longevity.

    This method assumes that the investment precedes the annuity payments.

    Parameters
    ----------
    investment : int or float
        The sum initially invested.
    investment_longevity : int
        The investment\'s expected longevity expressed as a number of periods.
    investment_period : int
        The period during which the investment is made or reported. It has to
        be between 0 (present) and analysis_period_span+1 (planning horizon).
    discount_rate : float
        The discount rate for the periods under considerations.
    analysis_period_span : int
        The span of time from the present until the planning horizon, expressed
        as a number of periods.
    return_annuity : bool, optional
        If True, returns the annuity as well. The default is False.

    Raises
    ------
    ValueError
        This error is raised if the inputs are inconsistent.

    Returns
    -------
    float
        The present salvage value.
    float
        The annuity.

    """
    if investment_period >= analysis_period_span + 1:
        raise ValueError(
            "The investment has to be made within the period being analysed."
        )

    # the present salvage value requires the lifetime to extend beyond the hor.

    if analysis_period_span >= investment_longevity + investment_period:
        if return_annuity:
            return 0, annuity(
                investment=investment,
                investment_longevity=investment_longevity,
                discount_rate=discount_rate,
            )

        else:
            return 0

    # the annuity has to consider the asset longevity and the commission. delay

    value_annuity = annuity(
        investment=investment,
        investment_longevity=investment_longevity,
        discount_rate=discount_rate,
    )

    discount_rates = tuple(
        discount_rate for i in range(investment_longevity + investment_period)
    )

    net_cash_flows = list(
        value_annuity for i in range(investment_longevity + investment_period + 1)
    )

    for year_index in range(analysis_period_span + 1):
        net_cash_flows[year_index] = 0

    if return_annuity:
        return (
            npv(discount_rates=discount_rates, net_cash_flows=net_cash_flows),
            value_annuity,
        )

    else:
        return npv(discount_rates=discount_rates, net_cash_flows=net_cash_flows)


# *****************************************************************************
# *****************************************************************************
