# imports

from math import inf

# local, external

from networkx import MultiDiGraph, is_path
from geopandas import GeoDataFrame
from pandas import isna
from shapely.geometry import LineString, Point
from shapely import length
from osmnx.distance import great_circle
from osmnx.stats import count_streets_per_node
from osmnx.projection import is_projected

# local, internal

from ..gis import osm as osm
from ..gis import identify as ident

# *****************************************************************************
# *****************************************************************************


def edge_lengths(network: MultiDiGraph, edge_keys: tuple = None) -> dict:
    """
    Calculate edge lengths in a OSMnx-formatted MultiDiGraph network object.

    The calculation method changes depending on whether the coordinates are
    projected and depending on whether the edges are simplified.

    Parameters
    ----------
    network : MultiDiGraph
        The object describing the network.
    edge_keys : tuple, optional
        The keys for the edges under consideration. The default is None, which
        means all edges in the network will be considered.

    Returns
    -------
    dict
        A dictionary with the lengths for each edge.

    """
    # determine if the graph is projected or not
    graph_is_projected = is_projected(network.graph["crs"])
    # check if edge keys were specified
    if type(edge_keys) == type(None):
        # no particular edge keys were provided: consider all edges (default)
        edge_keys = network.edges(keys=True)  # tuple(network.edges(keys=True))
    # initialise length dict
    length_dict = {}
    # for each edge on the graph
    for edge_key in edge_keys:
        # calculate it using the library
        if graph_is_projected:
            # calculate it using projected coordinates
            if osm.KEY_OSMNX_GEOMETRY in network.edges[edge_key]:
                # use geometry
                length_dict[edge_key] = length(
                    network.edges[edge_key][osm.KEY_OSMNX_GEOMETRY]
                )
            else:
                # use (projected) coordinates
                start_point = Point(
                    (
                        network.nodes[edge_key[0]][osm.KEY_OSMNX_X],
                        network.nodes[edge_key[0]][osm.KEY_OSMNX_Y],
                    )
                )
                end_point = Point(
                    (
                        network.nodes[edge_key[1]][osm.KEY_OSMNX_X],
                        network.nodes[edge_key[1]][osm.KEY_OSMNX_Y],
                    )
                )
                length_dict[edge_key] = start_point.distance(end_point)

        else:
            # calculate it using unprojected coordinates (lat/long)
            if osm.KEY_OSMNX_GEOMETRY in network.edges[edge_key]:
                # use geometry
                length_dict[edge_key] = great_circle_distance_along_path(
                    network.edges[edge_key][osm.KEY_OSMNX_GEOMETRY]
                )
            else:
                # use (unprojected) coordinates
                length_dict[edge_key] = great_circle(
                    lat1=network.nodes[edge_key[0]][osm.KEY_OSMNX_Y],
                    lon1=network.nodes[edge_key[0]][osm.KEY_OSMNX_X],
                    lat2=network.nodes[edge_key[1]][osm.KEY_OSMNX_Y],
                    lon2=network.nodes[edge_key[1]][osm.KEY_OSMNX_X],
                )
    # return the dict with lengths of each edge
    return length_dict


# *****************************************************************************
# *****************************************************************************


def great_circle_distance_along_path(path: LineString) -> float:
    """
    Computes the great circle distance along a given path.

    The distance is to be calculated using a shapely LineString object made of
    (longitude, latitude) coordinate tuples. The calculation is vectorised.

    Parameters
    ----------
    path : LineString
        An object describing the path whose distance is to be calculated.

    Returns
    -------
    float
        The sum of the individual great circle distances along the path.

    """
    # get coordinates
    lon = tuple(path.coords.xy[0])
    lat = tuple(path.coords.xy[1])
    # sum individual distances and return
    return sum(
        great_circle(
            lat[:-1],  # latitudes of starting points
            lon[:-1],  # longitudes of starting points
            lat[1:],  # latitudes of ending points
            lon[1:],  # longitudes of ending points
        )
    )


# *****************************************************************************
# *****************************************************************************


def update_street_count(network: MultiDiGraph):
    """
    Updates the street count attributes of nodes in a MultiDiGraph object.

    Parameters
    ----------
    network : MultiDiGraph
        The object describing the network.

    Returns
    -------
    None.

    """
    # update street count
    street_count_dict = count_streets_per_node(network)
    network.add_nodes_from(
        (
            (key, {osm.KEY_OSMNX_STREET_COUNT: value})
            for key, value in street_count_dict.items()
        )
    )


# *****************************************************************************
# *****************************************************************************


def node_path_length(
    network: MultiDiGraph, path: list, return_minimum_length_only: bool = True
) -> list or float:
    """
    Returns the length or lengths of a path defined using nodes.

    If more than one edge connects adjacent nodes along the path, a length value
    will be returned for each possible path combination.

    Parameters
    ----------
    network : MultiDiGraph
        The object describing the network.
    path : list
        A list of node keys or edge keys describing the path.
    return_minimum_length_only : bool, optional
        If True, returns the minimum path length only. The default is True.

    Returns
    -------
    list or float
        The path\'s length or all lengths consistent with the path provided.

    """

    # direction matters
    path_length = len(path)
    if path_length == 0:
        return inf

    # if the path is given as a list of node keys, then it is subjective
    # i.e., it may refer to many paths, namely if parallel edges exist

    # check if the path object qualifies as such
    if not is_path(network, path):
        # it does not, exit
        if return_minimum_length_only:
            return inf
        else:
            return [inf]

    # prepare a list with all possible paths given as lists of edge keys
    list_of_edge_key_paths = [[]]  # a list of edge key lists

    # for each pair of nodes in the path
    for node_pair in range(path_length - 1):
        # get the edges between these two nodes
        edge_keys = ident.get_edges_from_a_to_b(
            network, path[node_pair], path[node_pair + 1]
        )
        number_edge_keys = len(edge_keys)
        if number_edge_keys == 1:
            # only one edge exists: append its key to all existing lists/paths
            for edge_key_path in list_of_edge_key_paths:
                edge_key_path.append(edge_keys[0])
        else:  # multiple edges exist: each path identified so far has to be
            # replicated a total of number_edge_keys times and then updated
            number_paths = len(list_of_edge_key_paths)
            # for each parallel edge
            for edge_key_index in range(number_edge_keys - 1):
                # replicate all paths
                for path_index in range(number_paths):
                    list_of_edge_key_paths.append(
                        list(list_of_edge_key_paths[path_index])
                    )
            # paths have been replicated, now add the edges
            for edge_key_index in range(number_edge_keys):
                for path_index in range(number_paths):
                    # add the new edge
                    list_of_edge_key_paths[
                        path_index + edge_key_index * number_paths
                    ].append(edge_keys[edge_key_index])

    # *************************************************************************

    path_lenths = [
        sum(network.edges[edge_key][osm.KEY_OSMNX_LENGTH] for edge_key in edge_key_path)
        for edge_key_path in list_of_edge_key_paths
    ]
    if return_minimum_length_only:
        return min(path_lenths)
    else:
        return path_lenths

    # *************************************************************************


# *****************************************************************************
# *****************************************************************************


def edge_path_length(network: MultiDiGraph, path: list, **kwargs) -> float:
    """
    Returns the total length of a path defined using edges.

    If the path does not exist, or if no path is provided, the result will be
    infinity (math.inf).

    Parameters
    ----------
    network : MultiDiGraph
        The object describing the network.
    path : list
        A list of edge keys describing the path.

    Returns
    -------
    list or float
        The path\'s length or all lengths consistent with the path provided.

    """

    # check the number of
    path_length = len(path)
    if path_length == 0:
        return inf
    if ident.is_edge_path(network, path, **kwargs):
        return sum(network.edges[edge_key][osm.KEY_OSMNX_LENGTH] for edge_key in path)
    else:
        # no path provided
        return inf


# *****************************************************************************
# *****************************************************************************


def count_ocurrences(
    gdf: GeoDataFrame, column: str, column_entries: list = None
) -> dict:
    """
    Counts the number of occurrences per entry in a DataFrame object's column.

    If a list is provided, only the entries that match those in the list are
    counted. If no list is provided, all unique entries are counted.

    Parameters
    ----------
    gdf : GeoDataFrame
        The object holding the data.
    column : str
        A string with the name of the column.
    column_entries : list, optional
        A list with the entries that are to be counted. The default is None, in
        which case all the unique entries will be counted.

    Returns
    -------
    dict
        A dictionary with the counts whose keys are the values counted.

    """

    if type(column_entries) == list:
        # find entries also present in the dict
        # initialise dict
        count_dict = {}
        # for each key in the dict
        for key in column_entries:
            # # store the number of rows
            # count_dict[key] = gdf[gdf[column]==key].shape[0]
            # count the number of rows with this key
            if isna(key):
                count_dict[key] = gdf[gdf[column].isnull()].shape[0]
            else:
                count_dict[key] = gdf[gdf[column] == key].shape[0]
    else:
        # find all unique entries
        # initialise dict
        count_dict = {}
        for entry in gdf[column]:
            # check if it is already in the dict
            if entry in count_dict:
                # it is, skip
                continue
            # it is not, count and store the number of rows with said entry
            if isna(entry):  # type(entry) == type(None):
                count_dict[entry] = gdf[gdf[column].isnull()].shape[0]
            else:
                count_dict[entry] = gdf[gdf[column] == entry].shape[0]
    # return statement
    return count_dict


# *****************************************************************************
# *****************************************************************************
