# standard library imports

from typing import Tuple
from numbers import Real

# local, external

import networkx as nx
from shapely.geometry import Point, LineString
from osmnx.utils_graph import remove_isolated_nodes
from osmnx.distance import nearest_nodes, nearest_edges
from osmnx.projection import project_graph, is_projected

# local, internal

from .calculate import node_path_length

from ..gis import osm

# *****************************************************************************
# *****************************************************************************


def is_edge_consistent_with_geometry(network: nx.MultiDiGraph, edge_key):
    """
    Returns True if a given edge in an OSMnx-formatted graph is declared in the
    same order as its geometry. That is, if the source node corresponds to the
    first position in the geometry attribute and so forth.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The object describing the graph.
    edge_key : TYPE
        The key for the edge under consideration.

    Raises
    ------
    ValueError
        Is raised if no matching edge can be found.

    Returns
    -------
    bool
        Returns True if the edge is declared consistently or False otherwise.

    """
    # three cases:
    # no geometry: the edge is okay
    # geometry appears in the same order as the edge: the edge is okay
    # geometry does not appear in the same order as the edge: the edge is not okay
    if not network.has_edge(*edge_key):
        # the edge does not exist
        raise ValueError(
            "No edge was found matching the key provided: " + str(edge_key)
        )
    elif osm.KEY_OSMNX_GEOMETRY in network.edges[edge_key]:
        # edge exists and has a geometry attribute: check the geometry
        # check if the first point on the geometry matches the first node
        return tuple(network.edges[edge_key][osm.KEY_OSMNX_GEOMETRY].coords)[0] == (
            network.nodes[edge_key[0]][osm.KEY_OSMNX_X],
            network.nodes[edge_key[0]][osm.KEY_OSMNX_Y],
        )
    else:
        # edge exists but has no geometry attribute: it is consistent
        return True


# *****************************************************************************
# *****************************************************************************


def find_edges_in_reverse(network: nx.MultiDiGraph) -> dict:
    """
    Finds edges in reverse within an OSMnx-formatted MultiDiGraph object.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The object describing the network.

    Returns
    -------
    dict
        A dictionary whose keys and values are keys for reversed edges.

    """
    return {
        edge_key: [
            other_edge_key
            for other_edge_key in get_edges_from_a_to_b(
                network, node_start=edge_key[1], node_end=edge_key[0]
            )
            if edges_are_in_reverse(network, edge_key, other_edge_key)
        ]
        for edge_key in network.edges(keys=True)
    }


# *****************************************************************************
# *****************************************************************************


def is_edge_osmnx_compliant(network: nx.MultiDiGraph, edge_key) -> bool:
    """Returns True if a given edge is osmnx compliant and False otherwise."""

    # check if the edge exists
    if not network.has_edge(*edge_key):
        raise ValueError("Edge not found.")
    # check compatibility with osmnx
    return is_edge_data_osmnx_compliant(network.get_edge_data(*edge_key))


# *****************************************************************************
# *****************************************************************************


def is_edge_data_osmnx_compliant(edge_data: dict):
    """Returns True a given MultiDiGraph edge data dict is OSMnx-compliant."""
    # check if all the essential attributes are in
    for key in osm.KEYS_OSMNX_EDGES_ESSENTIAL:
        if key not in edge_data.keys():
            return False
    # check each attribute
    for key, value in edge_data.items():
        # check if the keys are standard
        if key in osm.KEYS_OSMNX_EDGES:
            # known edge attribute: predefined types
            if key == osm.KEY_OSMNX_GEOMETRY:
                # geometry: it must be a LineString
                if type(value) != LineString:
                    return False
            elif key == osm.KEY_OSMNX_OSMID:
                # osmid: it can be an integer or a list of integers
                if type(value) == list:
                    for u in value:
                        if type(u) != int:
                            return False
                elif type(value) != int:
                    return False
            elif key == osm.KEY_OSMNX_LENGTH:
                # length: it must be a non-negative real number
                if not isinstance(value, Real):
                    return False
            elif key == osm.KEY_OSMNX_REVERSED:
                # reversed: it must be a boolean or a list of booleans
                if type(value) == list:
                    for u in value:
                        if type(u) != bool:
                            return False
                elif type(value) != bool:
                    return False
            elif key == osm.KEY_OSMNX_ONEWAY:
                # oneway: it must be a boolean or a list of booleans
                if type(value) == list:
                    for u in value:
                        if type(u) != bool:
                            return False
                elif type(value) != bool:
                    return False
            # else: # no action
    return True


# *****************************************************************************
# *****************************************************************************


def edges_are_in_reverse(
    network: nx.MultiDiGraph,
    edge_a: tuple,
    edge_b: tuple,
    tolerance: float or int = 1e-3,
) -> bool:
    """
    Returns True if two edges in a graph represent the same one but in reverse.

    The conditions the two edges must observe are the following:
        - the set of each attribute appearing as a list must match;
        - the geometries must be identical but in reverse;
        - the lengths must be within a given tolerance of one another;
        - the reversed flags must be opposite, unless they are a list;
        - every other attribute must be identical.

    The graph should be formatted by OSMnx standards.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The object describing the network.
    edge_a : tuple
        The key to one of the two edges.
    edge_b : tuple
        The key to the other edge.
    tolerance : float or int
        The tolerance with which to determine closeness.

    Raises
    ------
    ValueError
        This error is raised if at least one of the edges is not recognised.

    Returns
    -------
    bool
        If True, the edges are in reverse.

    """

    # the edges are the same but in reverse:
    # - all attributes have to be the same or lists with
    # the same content, as in a set, except for
    # the geometry and reversed attributes

    # must involve the same nodes in reverse order
    if edge_a[0] != edge_b[1] or edge_a[1] != edge_b[0]:
        # the nodes do not match
        return False
    # make sure both edges exist and comply with osmnx
    if not is_edge_osmnx_compliant(network, edge_a) or not is_edge_osmnx_compliant(
        network, edge_b
    ):
        raise ValueError("One or more of the edges is not OSMnx-compliant.")

    fw_dict = network.get_edge_data(*edge_a)
    rv_dict = network.get_edge_data(*edge_b)

    # check if any of the non-mandatory attributes exist in one dict but not
    # the other
    key_attr = set([osm.KEY_OSMNX_GEOMETRY, osm.KEY_OSMNX_REVERSED])
    for _attr in key_attr:
        if (_attr in fw_dict.keys() and _attr not in rv_dict.keys()) or (
            _attr not in fw_dict.keys() and _attr in rv_dict.keys()
        ):
            # incoherent inputs
            return False

    # for each key, value pair in the forward edge's dict
    for attr_key, attr_value in fw_dict.items():
        if type(attr_value) == list and (
            (
                type(rv_dict[attr_key]) == list
                and set(attr_value) != set(rv_dict[attr_key])
            )
            or type(rv_dict[attr_key]) != list
        ):
            # the sets of list arguments do not match
            # or, the arguments are not equivalent
            return False
        elif (
            type(attr_value) == list
            and type(rv_dict[attr_key]) == list
            and set(attr_value) == set(rv_dict[attr_key])
        ):
            # the sets of list arguments match
            continue
        elif attr_key == osm.KEY_OSMNX_GEOMETRY and (
            (
                type(rv_dict[attr_key]) == LineString
                and tuple(attr_value.coords)
                != tuple(rv_dict[attr_key].reverse().coords)
            )
            or attr_key not in rv_dict
            or type(rv_dict[attr_key]) != LineString
        ):
            # either the geometries are not reversed
            # or, there is no geometry attribute in the reverse dict
            # or, the geometry in the reverse edge is not for a LineString
            return False
        elif (
            attr_key == osm.KEY_OSMNX_GEOMETRY
            and type(rv_dict[attr_key]) == LineString
            and tuple(attr_value.coords) == tuple(rv_dict[attr_key].reverse().coords)
        ):
            # the geometries are reversed
            continue
        elif attr_key == osm.KEY_OSMNX_REVERSED and (
            (attr_key in rv_dict and attr_value == rv_dict[attr_key])
            or attr_key not in rv_dict
            or type(rv_dict[attr_key]) != bool
        ):
            # either the reversed flags match
            # or, there is no reversed flag in the reverse dict
            return False
        elif (
            attr_key == osm.KEY_OSMNX_REVERSED
            and attr_key in rv_dict
            and not attr_value == rv_dict[attr_key]
        ):
            # the reversed flags are logical opposites
            continue
        elif attr_key == osm.KEY_OSMNX_LENGTH and (
            (
                attr_key in rv_dict
                and
                # isinstance(rv_dict[attr_key], Real) and
                abs(attr_value - rv_dict[attr_key]) > tolerance
            )
            or attr_key not in rv_dict
        ):
            # either the lengths differ too much
            # or, there is no length attribute in the reverse dict
            # or it is not a numeric type
            return False
        elif (
            attr_key == osm.KEY_OSMNX_LENGTH
            and attr_key in rv_dict
            and
            # isinstance(rv_dict[attr_key], Real) and
            abs(attr_value - rv_dict[attr_key]) <= tolerance
        ):
            # the lengths are within the tolerance
            continue
        elif attr_key in rv_dict and attr_value != rv_dict[attr_key]:
            # either the attributes do not match
            return False
        # else: # the argument does not exist

    # all other possibilities have been exhausted: return True
    return True


# *****************************************************************************
# *****************************************************************************


def close_to_extremities(
    line: LineString,
    points: tuple,
    tolerance: float = 7 / 3 - 4 / 3 - 1,
    use_start_point_equidistant: bool = True,
    return_distances: bool = False,
) -> tuple:
    """
    Determines which points are close to a line\'s start and end points.

    Closeness between points is defined by being within a given tolerance of
    each other.

    Parameters
    ----------
    line : LineString
        The object describing the line.
    points : tuple
        A tuple of shapely.Point objects that are to be considered.
    tolerance : float, optional
        The tolerance with which closeness is defined. By default, the
        tolerance is set to machine precision (epsilon), that is, 7/3-4/3-1.
    use_start_point_equidistant : bool, optional
        If True, the start point is adopted as the closest point when the start
        and end points are equidistant from the main one. If False, the end
        point is adopted instead. The default is True.
    return_distances : bool, optional
        If True, distances are also returned as an output, namely the distances
        between the line and the points, the distances between the start point
        and the points, and the distances between the end point and the points.

    Returns
    -------
    tuple
        A tuple composed of two lists. The first has the indices for the points
        that are close to the start point. The second has the indices for the
        points that are closest to the end point.

    """

    # calculate the distances to the line
    line_distances = line.distance(points)

    # identify the start and end points
    start_point = Point(line.coords[0])
    end_point = Point(line.coords[-1])

    # calculate the distances to the start and end points
    start_distances = start_point.distance(points)
    end_distances = end_point.distance(points)

    # for each point
    _start = []
    _end = []
    for i, (line_distance, start_distance, end_distance) in enumerate(
        zip(line_distances, start_distances, end_distances)
    ):
        if start_distance < end_distance:
            # the point is closer to the start point than to the end point
            if abs(start_distance - line_distance) <= tolerance:
                # the point is within range of the start point
                _start.append(i)
        elif start_distance > end_distance:
            # the point is closer to the end point than to the start point
            if abs(end_distance - line_distance) <= tolerance:
                # the point is within range of the end point
                _end.append(i)
        else:
            # the point is equidistant to the start and end points
            if line_distance < end_distance:
                # the point is closer to the line than to the start/end points
                continue
            # reach these statements
            if use_start_point_equidistant:
                _start.append(i)
            else:
                _end.append(i)

    # return statement
    if return_distances:
        return _start, _end, line_distances, start_distances, end_distances
    else:
        return _start, _end


# *****************************************************************************
# *****************************************************************************


def find_roundabouts(
    network: nx.MultiDiGraph,
    maximum_perimeter: float = None,
    minimum_perimeter: float = None,
    maximum_number_nodes: int = None,
    minimum_number_nodes: int = None,
) -> list:
    """
    Finds sequences of nodes in a network that constitute roundabouts.

    A roundabout is defined as a sequence of nodes connected through one-way
    edges to form an endless loop. One-way edges are identified by the presence
    of the 'oneway' attribute equal to True in the edge dictionary. The minimum
    and maximum roundabout perimeter can be used to filter out roundabouts that
    are too big or too small, as can minimum and maximum number of nodes.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The object describing the network.
    maximum_perimeter : float, optional
        The maximum perimeter for an ordered sequence of nodes to qualify as a
        roundabout. The units are those of OSMnx: meters. The default is None,
        which leads to the maximum perimeter being ignored.
    minimum_perimeter : float, optional
        The minimum perimeter for an ordered sequence of nodes to qualify as a
        roundabout. The units are those of OSMNX: meters. The default is None,
        which leads to the minimum perimeter being ignored.
    maximum_number_nodes : int, optional
        The maximum number of nodes that can be in what is defined to be a
        roundabout for the purposes of this method. The default is None, which
        means there will no maximum number of nodes used to filter roundabouts.
    minimum_number_nodes : int, optional
        The minimum number of nodes that can be in what is defined to be a
        roundabout for the purposes of this method. The default is None, which
        means there will no minimum number of nodes used to filter roundabouts.

    Returns
    -------
    list
        A list of lists with node keys representing a roundabout's nodes. The
        nodes are ordered in the manner needed to go around the roundabout.
    """

    # copy the network object

    new_network = network.copy()

    # remove edges that do not qualify:
    # 1) self-loops
    # 2) edges that are not oneway
    # 3) edges that do not have the necessary attributes

    # node number limits
    there_are_upper_node_number_limits = True if maximum_number_nodes != None else False
    there_are_lower_node_number_limits = True if minimum_number_nodes != None else False
    there_are_node_number_limits = (
        there_are_upper_node_number_limits or there_are_lower_node_number_limits
    )

    # perimeter limits
    there_are_upper_perimeter_limits = True if maximum_perimeter != None else False
    there_are_lower_perimeter_limits = True if minimum_perimeter != None else False
    there_are_perimeter_limits = (
        there_are_upper_perimeter_limits or there_are_lower_perimeter_limits
    )

    # find edges that are not one way
    list_removable_edges = []
    for edge_key in new_network.edges(keys=True):
        edge_data_dict = new_network.get_edge_data(*edge_key)
        # if osm.KEY_OSMNX_ONEWAY not in edge_data_dict:
        #     # no information about being oneway or not: not a oneway edge
        #     # add edge to list
        #     list_removable_edges.append(edge_key)
        #     continue
        # else:
        #     # there is information about the edge being oneway or not
        #     if not edge_data_dict[osm.KEY_OSMNX_ONEWAY]:
        #         # oneway attr is False: not a oneway edge
        #         # add edge to list
        #         list_removable_edges.append(edge_key)
        #         continue
        # there is information about the edge being oneway or not
        if not edge_data_dict[osm.KEY_OSMNX_ONEWAY]:
            # oneway attr is False: not a oneway edge
            # add edge to list
            list_removable_edges.append(edge_key)
            continue
        # it is a one way node...
    # remove the edges
    new_network.remove_edges_from(list_removable_edges)
    # remove isolated nodes
    remove_isolated_nodes(new_network)
    # exclude self loops
    list_selflooping_nodes = find_self_loops(new_network)
    for node_key in list_selflooping_nodes:
        while new_network.has_edge(u=node_key, v=node_key):
            new_network.remove_edge(u=node_key, v=node_key)

    # *************************************************************************
    # *************************************************************************

    # find loops
    node_paths = nx.simple_cycles(new_network)

    # *************************************************************************
    # *************************************************************************

    # exclude paths based on the perimeter and the number of nodes, if set
    if not there_are_node_number_limits and not there_are_perimeter_limits:
        return list(node_paths)
    else:  # each potential candidate needs to be checked
        final_node_paths = []
        # for each node group
        for node_path in node_paths:
            if there_are_perimeter_limits:
                # compute the total length for each node
                total_length = node_path_length(
                    network, node_path, return_minimum_length_only=True
                )
                if there_are_lower_perimeter_limits:
                    if total_length < minimum_perimeter:
                        continue
                if there_are_upper_perimeter_limits:
                    if total_length > maximum_perimeter:
                        continue
            if there_are_node_number_limits:
                number_nodes_in_path = len(node_path)
                if there_are_lower_node_number_limits:
                    if number_nodes_in_path < minimum_number_nodes:
                        continue
                if there_are_upper_node_number_limits:
                    if number_nodes_in_path > maximum_number_nodes:
                        continue
            # if none of the conditions was violated, add it the final list
            final_node_paths.append(node_path)
        # return the final list
        return final_node_paths

        # *********************************************************************
        # *********************************************************************


# *****************************************************************************
# *****************************************************************************


def is_roundabout(
    network: nx.MultiDiGraph, path: list, path_as_node_keys: bool = True
) -> bool:
    """
    Returns True if a given path constitutes a roundabout in a directed graph.

    A roundabout is defined as a sequence of nodes connected through one-way
    edges to form an endless loop. One-way edges are identified by the presence
    of the 'oneway' attribute equal to True in the edge dictionary.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The object describing the network.
    path : list
        A list with node keys forming the roundabout. The node keys must be
        ordered so as to traverse the roundabout without skipping segments. The
        list should not contain redundancies (i.e., no repeated nodes).

    Returns
    -------
    bool
        True, if the path constitutes a roundabout and False otherwise.

    """

    # paths are given as node lists

    # roundabouts require at least two nodes

    if len(path) <= 1:
        raise ValueError("Node paths require at least two nodes.")

    # for each node in path

    for node_key in path:
        # check if it exists in the network

        if not network.has_node(node_key):
            return False

        # there should be no repetitions

        if path.count(node_key) > 1:
            return False

    # check if the last node connects to the first

    edge_keys = get_edges_from_a_to_b(network, path[-1], path[0])

    if len(edge_keys) == 0:
        return False

    else:
        # among the edges between them, find at least one compatible

        compatible_edge_exists = False

        for edge_key in edge_keys:
            # get its data

            edge_data_dict = network.get_edge_data(
                u=edge_key[0], v=edge_key[1], key=edge_key[2]
            )

            # ensure that this edge has the oneway attribute

            if osm.KEY_OSMNX_ONEWAY in edge_data_dict:
                # ensure that it is true

                if edge_data_dict[osm.KEY_OSMNX_ONEWAY]:
                    compatible_edge_exists = True

                    break

        # check for compatible edges

        if not compatible_edge_exists:
            # no compatible edges exist between these two nodes

            return False

    # for each other node pair

    for node_pair in range(len(path) - 1):
        # for each edge between them, find at least one compatible edge

        compatible_edge_exists = False

        for edge_key in get_edges_from_a_to_b(
            network, path[node_pair], path[node_pair + 1]
        ):
            # get its data

            edge_data_dict = network.get_edge_data(
                u=edge_key[0], v=edge_key[1], key=edge_key[2]
            )

            # ensure that this edge has the oneway attribute

            if osm.KEY_OSMNX_ONEWAY in edge_data_dict:
                # ensure that it is true

                if edge_data_dict[osm.KEY_OSMNX_ONEWAY]:
                    compatible_edge_exists = True

                    break

        # check for compatible edges

        if not compatible_edge_exists:
            # no compatible edges exist between these two nodes

            return False

    # otherwise, it is a roundabout

    return True


# *****************************************************************************
# *****************************************************************************


def get_edges_from_a_to_b(network: nx.MultiDiGraph, node_start, node_end) -> list:
    """
    Retrieve the keys for edges from one node to another.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The object describing the network.
    node_start : hashable-type
        The start node.
    node_end : hashable-type
        The end node.

    Returns
    -------
    list
        A list of edge keys from the start to the end node.

    """
    if network.has_edge(u=node_start, v=node_end):
        return [
            (node_start, node_end, key) for key in network._adj[node_start][node_end]
        ]
    else:
        return []


# *****************************************************************************
# *****************************************************************************


def get_edges_between_two_nodes(network: nx.MultiDiGraph, u, v) -> list:
    """
    Retrieve the keys for all edges involving two specific nodes.

    The keys concern edges in both directions. For a single direction, consider
    using the method get_edges_from_a_to_b instead.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The object describing the network.
    u : hashable-type
        The key to one of the nodes.
    v : hashable-type
        The key to the other node.

    Returns
    -------
    list
        A list of edge keys involving both nodes, in both directions.

    """

    if network.has_edge(u, v):
        # edges exist from u to v
        _out = [(u, v, k) for k in network._adj[u][v]]
        try:
            # try finding out if edges exist from v to u
            _out.extend([(v, u, k) for k in network._adj[v][u]])
        except KeyError:
            # edges do not exist from v to u
            pass
        # return what was obtained
        return _out
    elif network.has_edge(v, u):
        # edges do not exist from u to v but exist from v to u
        return [(v, u, k) for k in network._adj[v][u]]
    else:
        # no edges found
        return []


# *****************************************************************************
# *****************************************************************************


def get_edges_involving_node(
    network: nx.MultiDiGraph,
    node_key,
    include_outgoing_edges: bool = True,
    include_incoming_edges: bool = True,
    include_self_loops: bool = True,
) -> list:
    """
    Retrieve the keys for all edges involving a specific node.

    The keys concern incoming and outgoing edges. Optionally, the keys retrieved
    can concern only incoming or outgoing edges, self-loops included or not.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The object describing the network.
    node_key : hashable-type
        The key to the node under consideration.
    include_outgoing_edges : bool, optional
        If True, outgoing edges are considered. The default is True.
    include_incoming_edges : bool, optional
        If True, incoming edges are considered. The default is True.
    include_self_loops : bool, optional
        If True, self-loops are considered. The default is True.

    Returns
    -------
    list
        A list of edge keys involving the specified node.

    """

    return [
        edge_key
        for edge_key in network.edges(keys=True)
        if node_key in edge_key[0:2]
        # outgoing edges
        if (
            (node_key != edge_key[0] and not include_outgoing_edges)
            or include_outgoing_edges
        )
        # incoming edges
        if (
            (node_key != edge_key[1] and not include_incoming_edges)
            or include_incoming_edges
        )
        # self-loops
        if (
            (edge_key[0] != edge_key[1] and not include_self_loops)
            or include_self_loops
        )
    ]


# *****************************************************************************
# *****************************************************************************


def neighbours(
    network: nx.MultiDiGraph or nx.MultiGraph, node_key, ignore_self_loops: bool = True
):
    """
    Return a given node\'s neighbours.

    This method relies on networkx\'s neighbours method but adds the option to
    ignore self-loops.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The network object.
    node_key : hashable-type
        The key for the node whose neighbours the method is to identify.
    ignore_self_loops : bool, optional
        If True, self-loops are ignored. The default is True.

    Returns
    -------
    dict_keyiterator
        An object to iterate over the neighbours of the node specified.

    """

    if network.has_edge(node_key, node_key) and ignore_self_loops:
        return (
            _node_key
            for _node_key in nx.all_neighbors(network, node_key)
            if _node_key != node_key
        )

    else:
        return nx.all_neighbors(network, node_key)


# *****************************************************************************
# *****************************************************************************


def is_node_path(
    network: nx.MultiDiGraph, path: list, consider_reversed_edges: bool = False
) -> bool:
    """
    Indicates if a given path qualifies as a node path in a directed network.

    A node path consists of a sequence of nodes connected by directed edges.

    The sequence must include at least two nodes.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The network object.
    path : list
        The tentative node path.
    consider_reversed_edges : bool, optional
        If True, reversed edges can be used to identify a path. If False, only
        edges in the travel direction will be considered. The default is False.

    Returns
    -------
    bool
        Returns True if path is a node path and False otherwise.

    """
    if len(path) <= 1:
        return False
    elif consider_reversed_edges:
        # all nodes must exist in the network
        for node in path:
            if not network.has_node(node):
                return False
        # each pair of nodes must be connected, in one way or another
        for n1, n2 in nx.utils.pairwise(path):
            if network.has_edge(n1, n2) or network.has_edge(n2, n1):
                continue
            else:
                return False
        return True
    else:
        return nx.is_path(network, path)


# *****************************************************************************
# *****************************************************************************


def is_edge_path(
    network: nx.MultiDiGraph,
    path: list,
    ignore_edge_direction: bool = False,
    allow_multiple_formats: bool = False,
) -> bool:
    """
    Indicates if a given path qualifies as an edge path in a directed network.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The network object.
    path : list
        The tentative edge path.
    ignore_edge_direction : bool, optional
        If True, edges on the path may be reversed, as long as they can still
        form a sequence of nodes via existing edges. The default is False.
    allow_multiple_formats : bool, optional
        If True, edge keys can be supplied as 2- and 3-tuples. If False, only
        one of those can be accepted. The default is False.

    Returns
    -------
    bool
        Returns True if the path qualifies as an edge path and False otherwise.

    """

    if len(path) == 0:
        # empty path

        return False

    else:
        # all the edges have to exist

        previous_edge_key_length = len(path[0])

        for edge_i, tentative_edge_key in enumerate(path):
            edge_key_length = len(tentative_edge_key)

            if not allow_multiple_formats:
                if previous_edge_key_length != edge_key_length:
                    # the edge key format changes: not a path

                    raise ValueError(
                        "The path must be provided using only one edge format."
                    )

            # find out if the edge exists

            if edge_key_length == 3:
                # 3-tuple format

                if not network.has_edge(
                    u=tentative_edge_key[0],
                    v=tentative_edge_key[1],
                    key=tentative_edge_key[2],
                ):
                    # the edge does not exist as specified

                    return False

            elif edge_key_length == 2:
                # 2-tuple format

                if not network.has_edge(
                    u=tentative_edge_key[0], v=tentative_edge_key[1]
                ):
                    # the edge does not exist as specified

                    return False

            else:
                # unknown format

                return False

            # the edge exists: check if it forms a sequence

            if edge_i != 0:  # skip the first iteration
                # if none of the current edge's nodes is mentioned in the
                # previous edge, then no sequence is formed

                if (
                    tentative_edge_key[0] not in path[edge_i - 1][0:2]
                    and tentative_edge_key[1] not in path[edge_i - 1][0:2]
                    and ignore_edge_direction
                ):
                    return False

                # if the previous edge's end node is not the current edge's
                # start node, then it is not a valid edge path

                if (
                    path[edge_i - 1][1] != tentative_edge_key[0]
                    and not ignore_edge_direction
                ):
                    return False

                # # check the formats, if necessary

                # if (not allow_multiple_formats and
                #     len(path[edge_i-1]) != len(tentative_edge_key)):

                #     return False

            previous_edge_key_length = edge_key_length

        return True


# *****************************************************************************
# *****************************************************************************


def is_path_straight(
    network: nx.MultiDiGraph,
    path: list,
    consider_reversed_edges: bool = False,
    ignore_self_loops: bool = False,
) -> bool:
    """
    Returns True if the path is straight and False otherwise.

    A path is defined to be straight if it presents no options along it.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The objet describing the network.
    path : list
        The list of nodes along the path.
    consider_reversed_edges : bool, optional
        If True, reversed edges can also be used to form paths. If False, only
        edges in the stated direction will be considered. The default is False.
    ignore_self_loops : bool, optional
        If True, paths with self-loops can still be straight. If False, paths
        containing self-loops cannot be straight. The default is False.

    Returns
    -------
    bool
        A boolean indicating whether the path is straight or not.

    """
    # confirm that it is a path
    if not is_node_path(network, path, consider_reversed_edges):
        return False
    # a straight path requires at least two nodes
    path_length = len(path)
    if path_length == 2:
        return True  # path with two nodes is always straight

    # check if the intermediate nodes have the right number of neighbours
    for intermediate_node in path[1:-1]:
        if (
            len(
                set(
                    neighbours(
                        network, intermediate_node, ignore_self_loops=ignore_self_loops
                    )
                )
            )
            != 2
        ):
            # the path is not straight if the intermediate nodes do not have
            # two distinct neighbours
            return False

    # if all intermediate nodes have two neighbours, return True
    return True


# *****************************************************************************
# *****************************************************************************


def find_simplifiable_paths(
    network: nx.MultiDiGraph,
    protected_nodes: list,
    ignore_self_loops: bool = False,
    consider_reversed_edges: bool = False,
    include_both_directions: bool = False,
) -> list:
    """
    Enumerates the simplifiable paths found in a given graph.

    A path is defined to be simplifiable if it presents no options along it
    and involves at least three different nodes: two-node paths are straight
    but are not simplifiable, with or without reversed edges.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The object describing the graph.
    protected_nodes : list
        A list of keys for nodes that cannot be in any straight path.
    ignore_self_loops : bool, optional
        If True, paths including self-loops can still be straight. If False,
        paths including self-loops cannot be straight. The default is False.
    consider_reversed_edges : bool, optional
        If True, a straight path can include nodes connected in reverse. If
        False, only edges in the stated direction will be considered. The
        default is False.
    include_both_directions : bool, optional
        If True, and if reverse edges are allowed, simplifiable paths will
        appear once per direction, otherwise just once. The default is False.

    Returns
    -------
    list
        A list of the straight paths in the graph.

    """

    # a straight path is a path where the intermediate nodes (all excluding the
    # first and the last) have to exist and have exactly 2 distinct neighbours

    # *************************************************************************
    # *************************************************************************

    # locate all the non-excluded nodes that can form straight paths

    intermediate_candidate_nodes = set(
        [
            node_key
            for node_key in network.nodes()
            # the node cannot be among those excluded
            if node_key not in protected_nodes
            # the node has to be linked to two other nodes other than itself
            if len(set(neighbours(network, node_key, ignore_self_loops=True))) == 2
            # exclude nodes with self-loops if desired:
            # 1) self-loops are tolerated (no need to check)
            # 2) self-loops are not tolerated and they do not exist
            if (
                ignore_self_loops
                or (not ignore_self_loops and not network.has_edge(node_key, node_key))
            )
        ]
    )

    # *************************************************************************

    # find out paths around the nodes identified

    list_paths = []
    list_paths_nodes = []
    list_nodes_joined = set([])

    # try to form paths around the candidate nodes
    for candidate_node in intermediate_candidate_nodes:
        # skip if the node is already in a path
        if candidate_node in list_nodes_joined:
            continue

        # select method
        if consider_reversed_edges:
            # reversed edges are accepted
            new_sequence = _find_path_direction_insensitive(
                network,
                list_valid_nodes=intermediate_candidate_nodes - list_nodes_joined,
                start_node=candidate_node,
                ignore_self_loops=ignore_self_loops,
            )
        else:
            # reversed edges are not accepted
            new_sequence = _find_path_direction_sensitive(
                network,
                list_valid_nodes=intermediate_candidate_nodes - list_nodes_joined,
                start_node=candidate_node,
                ignore_self_loops=ignore_self_loops,
            )
        # make sure the sequence is not redundant
        if len(new_sequence) <= 2 or new_sequence in list_paths:
            # path is just one edge or has already been included
            continue

        # add the path
        list_paths.append(new_sequence)
        list_paths_nodes.append(set(new_sequence))
        if consider_reversed_edges and include_both_directions:
            # directions do not matter:
            list_paths.append(new_sequence[::-1])
        # update the list of intermediate nodes already on paths
        list_nodes_joined.update(set(new_sequence[1:-1]))

    # *************************************************************************
    # *************************************************************************

    return list_paths

    # *************************************************************************
    # *************************************************************************


# *****************************************************************************
# *****************************************************************************


def _find_path_direction_sensitive(
    network: nx.MultiDiGraph,
    list_valid_nodes: list,
    start_node,
    ignore_self_loops: bool,
) -> list:
    def find_path_forward(network: nx.MultiDiGraph, current_node, path: list):
        # identify the last node's neighbours
        current_neighbours = set(
            neighbours(network, current_node, ignore_self_loops=True)
        )
        # check each neighbour
        for a_neighbour in current_neighbours:
            # check the direction of edge towards the neighbour
            if network.has_edge(current_node, a_neighbour):
                # forward edge
                # 1) node is not on the path
                # 1.1) is a valid node >> path continues
                # 1.2) is not a valid node >> end of path
                # 2) node is on the path already
                # 2.1) matches the start node = cycle
                # 2.2) does not match the start node (reversed edge) = ignore
                if a_neighbour not in path:
                    # neighbour is not on the path
                    if a_neighbour in list_valid_nodes:
                        # is a valid node: path continues
                        # add the neighbour to the end of the path
                        path.append(a_neighbour)
                        # recursive call with extended path
                        return find_path_forward(network, path[-1], path)
                    else:  # is not a valid node: path ends
                        # add the neighbour to the end of the path:
                        path.append(a_neighbour)
                        # return the path
                        return path
                elif a_neighbour == path[0] and len(path) >= 3:
                    # the path length is greater than or equal to 3,
                    # neighbour is already on the path, matches the start,
                    # and has two neighbours other than itself:
                    # close the loop and return the path
                    if (
                        len(
                            set(
                                neighbours(
                                    network,
                                    a_neighbour,
                                    ignore_self_loops=ignore_self_loops,
                                )
                            )
                        )
                        == 2
                    ):
                        # add the neighbour to the end of the path:
                        path.append(a_neighbour)
                        # return the path
                        return path
        # all neighbours have been visited: return the current path
        return path

    def find_path_backward(network: nx.MultiDiGraph, current_node, path: list):
        # identify the last node's neighbours
        current_neighbours = set(
            neighbours(network, current_node, ignore_self_loops=True)
        )
        # check each neighbour
        # 1) if the neighbour is ahead and is a valid node:
        # >> recursion w/ neighbour and then return the final path
        # 2) if the neighbour is ahead and is not a valid node:
        # >> add it to the path and then return it
        # 3) if the neighbour is not ahead and is on the path:
        # >> continue, namely to check the other neighbour
        # 4) if the neighbour is not ahead and is not on the path:
        # >> add it to the beginning of the path and continue
        # check each neighbour
        for a_neighbour in current_neighbours:
            # check the direction of edge towards the neighbour
            if network.has_edge(a_neighbour, current_node):
                # backward edge
                # 1) node is not on the path
                # 1.1) is a valid node >> path continues
                # 1.2) is not a valid node >> end of path
                # 2) node is on the path already
                # 2.1) matches the start node = cycle
                # 2.2) does not match the start node (reversed edge) = ignore
                if a_neighbour not in path:
                    # neighbour is not on the path
                    if a_neighbour in list_valid_nodes:
                        # is a valid node: path continues
                        # add the neighbour to the start of the path
                        path.insert(0, a_neighbour)
                        # recursive call with extended path
                        return find_path_backward(network, path[0], path)
                    else:  # is not a valid node: path ends
                        # add the neighbour to the start of the path
                        path.insert(0, a_neighbour)
                        # return the path
                        return path
                elif a_neighbour == path[-1] and len(path) >= 3:
                    # the path length is greater than or equal to 3,
                    # neighbour is already on the path, matches the end,
                    # and has two neighbours other than itself:
                    # close the loop and return the path
                    # if (len(set(neighbours(
                    #         network,
                    #         a_neighbour,
                    #         ignore_self_loops=ignore_self_loops))) == 2):
                    #     # add the neighbour to the start of the path
                    #     path.insert(0, a_neighbour)
                    #     # return the path
                    #     return path
                    path.insert(0, a_neighbour)
                    # return the path
                    return path
        # all neighbours have been visited: return the current path
        return path

    # *************************************************************************

    # find the path forward, check for cycles and then find the path backwards
    # find the forward path segment
    path = find_path_forward(network, start_node, [start_node])
    # cycles have to be detected on the first try
    if len(path) >= 3 and path[0] == path[-1]:
        # it is a cycle: no need to search backwards
        return path
    # find the backward path segment
    return find_path_backward(network, path[0], path)

    # *************************************************************************


# *****************************************************************************
# *****************************************************************************


def _find_path_direction_insensitive(
    network: nx.MultiDiGraph,
    list_valid_nodes: list,
    start_node,
    ignore_self_loops: bool,
) -> list:
    def find_path_forward(network: nx.MultiDiGraph, current_node, path: list):
        # identify the last node's neighbours
        current_neighbours = set(
            neighbours(network, current_node, ignore_self_loops=True)
        )
        # check each neighbour
        for a_neighbour in current_neighbours:
            # check the direction of edge towards the neighbour:
            # 1) node is not on the path
            # 1.1) is a valid node >> path continues
            # 1.2) is not a valid node >> end of path
            # 2) node is on the path already
            # 2.1) matches the start node = cycle
            # 2.2) does not match the start node (reversed edge) = ignore
            if a_neighbour not in path:
                # neighbour is not on the path
                if a_neighbour in list_valid_nodes:
                    # is a valid node: path continues
                    # add the neighbour to the end of the path
                    path.append(a_neighbour)
                    # recursive call with extended path
                    return find_path_forward(network, path[-1], path)
                else:  # is not a valid node: path ends
                    # add the neighbour to the end of the path:
                    path.append(a_neighbour)
                    # return the path
                    return path
            elif a_neighbour == path[0] and len(path) >= 3:
                # the path length is greater than or equal to 3,
                # neighbour is already on the path, matches the start,
                # and has two neighbours other than itself:
                # close the loop and return the path
                if (
                    len(
                        set(
                            neighbours(
                                network,
                                a_neighbour,
                                ignore_self_loops=ignore_self_loops,
                            )
                        )
                    )
                    == 2
                ):
                    # add the neighbour to the end of the path:
                    path.append(a_neighbour)
                    # return the path
                    return path
        # all neighbours have been visited: return the current path
        return path

    def find_path_backward(network: nx.MultiDiGraph, current_node, path: list):
        # identify the last node's neighbours
        current_neighbours = set(
            neighbours(network, current_node, ignore_self_loops=True)
        )
        # check each neighbour
        # 1) if the neighbour is ahead and is a valid node:
        # >> recursion w/ neighbour and then return the final path
        # 2) if the neighbour is ahead and is not a valid node:
        # >> add it to the path and then return it
        # 3) if the neighbour is not ahead and is on the path:
        # >> continue, namely to check the other neighbour
        # 4) if the neighbour is not ahead and is not on the path:
        # >> add it to the beginning of the path and continue

        # check each neighbour
        for a_neighbour in current_neighbours:
            # check the direction of edge towards the neighbour
            # 1) node is not on the path
            # 1.1) is a valid node >> path continues
            # 1.2) is not a valid node >> end of path
            # 2) node is on the path already
            # 2.1) matches the start node = cycle
            # 2.2) does not match the start node (reversed edge) = ignore
            if a_neighbour not in path:
                # neighbour is not on the path
                if a_neighbour in list_valid_nodes:
                    # is a valid node: path continues
                    # add the neighbour to the start of the path
                    path.insert(0, a_neighbour)
                    # recursive call with extended path
                    return find_path_backward(network, path[0], path)
                else:  # is not a valid node: path ends
                    # add the neighbour to the start of the path
                    path.insert(0, a_neighbour)
                    # return the path
                    return path
            elif a_neighbour == path[-1] and len(path) >= 3:
                # the path length is greater than or equal to 3,
                # neighbour is already on the path, matches the end,
                # and has two neighbours other than itself:
                # close the loop and return the path
                # if (len(set(neighbours(
                #         network,
                #         a_neighbour,
                #         ignore_self_loops=ignore_self_loops))) == 2):
                #     # add the neighbour to the start of the path
                #     path.insert(0, a_neighbour)
                #     # return the path
                #     return path
                # add the neighbour to the start of the path
                path.insert(0, a_neighbour)
                # return the path
                return path
        # all neighbours have been visited: return the current path
        return path

    # *************************************************************************

    # find the path forward, check for cycles and then find the path backwards
    # explore paths in the forward sense
    path = find_path_forward(network, start_node, [start_node])
    # check for cycles
    if len(path) >= 3 and path[0] == path[-1]:
        # it is a cycle: no need to search backwards
        return path
    # explore paths in the backward sense and return the path
    return find_path_backward(network, path[0], path)

    # *************************************************************************
    # *************************************************************************


# *****************************************************************************
# *****************************************************************************


def find_self_loops(network: nx.MultiDiGraph) -> list:
    """
    Returns a list with the nodes that connect to themselves.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The network object.

    Returns
    -------
    generator-expression
        The keys to the nodes with self-loops.

    """

    return (
        node_key
        for node_key in network.nodes()
        if network.has_edge(u=node_key, v=node_key)
    )


# *****************************************************************************
# *****************************************************************************


def find_unconnected_nodes(network: nx.MultiDiGraph) -> list:
    """
    Returns a list with the nodes that are not connected whilst in the network.

    The method is meant to be used with MultiDiGraph objects.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The network object.

    Returns
    -------
    list
        The list of unconnected nodes.

    """

    return [
        node_key
        for node_key in network.nodes()
        if len(tuple(network.neighbors(node_key))) == 0
    ]


# *****************************************************************************
# *****************************************************************************


def nearest_nodes_other_than_themselves(
    network: nx.MultiDiGraph, node_keys: list, return_dist: bool = False
) -> list:
    """
    Returns a list with the keys of nodes closest to another set of nodes.

    The method relies on osmnx\'s nearest_nodes method.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The network object.
    node_keys : list
        A list of keys for the nodes whose closest nodes we want to identify.
    return_dist : bool, optional
        A boolean variable indicating distances should be returned too.
        The default is False.

    Returns
    -------
    nn or (nn, dist) : int/list or tuple
        nearest node IDs or optionally a tuple where dist contains distances
        between the points and their nearest nodes (note: from nearest_nodes).

    """

    # create a copy of the network

    network_copy = network.copy()

    # remove selected node keys from the copy

    network_copy.remove_nodes_from(node_keys)

    # use nearest_nodes from osmnx to find the nearest nodes in the copy

    nearest_node_keys = nearest_nodes(
        network_copy,
        [network.nodes[node_key]["x"] for node_key in node_keys],
        [network.nodes[node_key]["y"] for node_key in node_keys],
        return_dist=return_dist,
    )

    return nearest_node_keys


# *****************************************************************************
# *****************************************************************************


def is_start_or_end_point_or_close(
    line: LineString, point: Point, tolerance: float = 1e-3
) -> bool:
    """
    Returns True if a given point is near the start or end points of a line.

    Parameters
    ----------
    line : LineString
        The object describing the line.
    point : Point
        The point under consideration.
    tolerance : float, optional
        The tolerance with which closeness is defined. The default is 1e-3.

    Returns
    -------
    bool
        A boolean indicating whether the point is close to the line\'s start or
        end point.

    """

    # get the start and end points

    start_coords, *_, end_coords = line.coords

    # # compare the coordinates

    # if (tuple(point.coords)[0] == start_coords or
    #     tuple(point.coords)[0] == end_coords):

    #     return True

    # compare with the start point

    start_point = Point(start_coords)

    if start_point.distance(point) <= tolerance:
        return True

    # compare with the end point

    end_point = Point(end_coords)

    if end_point.distance(point) <= tolerance:
        return True

    # return statement

    return False


# *****************************************************************************
# *****************************************************************************


def is_start_or_end_point(line: LineString, point: Point) -> bool:
    """
    Returns True if a given point is the start or end point of a line.

    Parameters
    ----------
    line : LineString
        The object describing the line.
    point : Point
        The point under consideration.

    Returns
    -------
    bool
        A boolean indicating whether the point is the start or end points.

    """

    # get the start and end points

    start_coords, *_, end_coords = line.coords

    # compare the coordinates

    if tuple(point.coords)[0] == start_coords or tuple(point.coords)[0] == end_coords:
        return True

    # return statement

    return False


# *****************************************************************************
# *****************************************************************************


def identify_edge_closest_to_node(
    network: nx.MultiDiGraph, node_keys: list, crs: str = None
) -> Tuple[list, nx.MultiDiGraph]:
    """
    Identify the edges that are closest to a given set of nodes.

    The network object should formatted according to OSMnx standards.

    Distances are calculated using projected coordinates, unless a specific
    coordinate system is given.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The network object.
    node_keys : list
        A list of keys corresponding to the nodes under consideration.
    crs : str, optional
        The CRS under which the operation is to be done. The default is None.
        If None, the CRS is determined automatically via the OSMnx library.

    Returns
    -------
    nearest_edge_keys : list
        The list of keys for the edges that are closest to the nodes (1:1).
    network : nx.MultiDiGraph
        The object for the projected network.

    """

    # *************************************************************************

    # 1) ensure that the network crs is correct and convert if not
    # 2) identify the edges that are nearest to the nodes
    # 3) revert network crs back to the original, if necessary

    # *************************************************************************

    # if it is a geographic CRS, convert it to a projected CRS

    if not is_projected(network.graph["crs"]) or type(crs) != type(None):
        # convert to a projected CRS (including if crs=None)

        projected_network = project_graph(network, to_crs=crs)

    else:
        projected_network = network

    # *************************************************************************

    # 2) identify the edges that are nearest to the nodes

    nearest_edge_keys = nearest_edges(
        projected_network,
        X=[
            projected_network.nodes[node_key][osm.KEY_OSMNX_X] for node_key in node_keys
        ],
        Y=[
            projected_network.nodes[node_key][osm.KEY_OSMNX_Y] for node_key in node_keys
        ],
        return_dist=False,
    )

    # return statement

    return nearest_edge_keys, projected_network


# *****************************************************************************
# *****************************************************************************
