# standard library imports

# from typing import Tuple

# local, external

import networkx as nx
from shapely.geometry import Point, LineString
from shapely.ops import nearest_points, linemerge, snap
from shapely import length
from osmnx.distance import great_circle, add_edge_lengths
from osmnx.projection import is_projected

# from pyproj import CRS

# local, internal

from ...problems.esipp.utils import unused_node_key
from ..misc.utils import generate_pseudo_unique_key
from ..gis import osm
from ..gis import identify as gis_iden

# from ..gis import identify as gis_calc
from .identify import close_to_extremities
from .calculate import update_street_count, edge_lengths

# *****************************************************************************
# *****************************************************************************


def remove_self_loops(network: nx.MultiDiGraph) -> list:
    """
    Removes self-loops from a directed graph defined in a MultiDiGraph object.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The object describing the graph.

    Returns
    -------
    list
        The keys to the nodes whose self-loops were removed.

    """
    
    selflooping_nodes = list(gis_iden.find_self_loops(network))
    for node in selflooping_nodes:
        while network.has_edge(u=node, v=node):
            network.remove_edge(u=node, v=node)
    return selflooping_nodes


# *****************************************************************************
# *****************************************************************************


def transform_roundabouts_into_crossroads(
    network: nx.MultiDiGraph, roundabouts: list
) -> list:
    """
    Transform roundabouts into crossroads.

    If there are roundabouts that encompass other roundabouts, the latter will
    be ignored.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The object describing the network.
    roundabouts : list
        A list of lists with node keys representing the nodes in a roundabout.

    Returns
    -------
    list
        A list of the node keys for the intersections created.

    """

    # declare the output list
    list_roundabout_centroids = []

    # for each roundabout
    for roundabout in roundabouts:
        # make sure roundabout qualifies as a roundabout
        if not gis_iden.is_roundabout(network, roundabout):
            # not a roundabout, skip
            list_roundabout_centroids.append(None)
            continue
        # if any node in the roundabout also appears in another one
        roundabout_overlaps = False
        # for each node forming this roundabout
        for node_roundabout in roundabout:
            # check every other roundabout
            for other_roundabout in roundabouts:
                if other_roundabout == roundabout:
                    continue
                # if the node exists in another roundabout
                if node_roundabout in other_roundabout:
                    # roundabouts overlap
                    roundabout_overlaps = True
                    break
            if roundabout_overlaps:
                # break out of the loop
                break
        if roundabout_overlaps:
            # the roundabout overlaps with some other one, skip it
            list_roundabout_centroids.append(None)
            continue

        # *********************************************************************
        # *********************************************************************
        # create a new node whose location is the roundabout's centroid
        list_point_coordinates = [
            (
                network.nodes[node_key][osm.KEY_OSMNX_X],
                network.nodes[node_key][osm.KEY_OSMNX_Y],
            )
            for node_key in roundabout
        ]
        new_geo = LineString(list_point_coordinates)
        roundabout_centroid_key = generate_pseudo_unique_key(network)
        network.add_node(
            roundabout_centroid_key,
            **{
                osm.KEY_OSMNX_X: new_geo.centroid.coords.xy[0][0],
                osm.KEY_OSMNX_Y: new_geo.centroid.coords.xy[1][0],
            }
        )
        list_roundabout_centroids.append(roundabout_centroid_key)
        # *********************************************************************
        # *********************************************************************
        # create new edges to link each node leading to the roundabout to the
        # node just created (new_node_key)
        # find the edges leading to the roundabout
        list_edges_leading_to_roundabout = [
            edge_key
            # for each node in the roundabout
            for node_key in roundabout
            # for each neighbouring nodes
            for other_node_key in gis_iden.neighbours(network, node_key)
            # if it is not in the roundabout itself
            if other_node_key not in roundabout
            # for each edge between the two nodes
            for edge_key in gis_iden.get_edges_between_two_nodes(
                network, node_key, other_node_key
            )
        ]
        # for each edge leading to the roundabout
        for edge_key in list_edges_leading_to_roundabout:
            # replace it with a new edge to the new node
            # get edge dict
            edge_dict = network.get_edge_data(edge_key[0], edge_key[1], edge_key[2])
            if osm.KEY_OSMNX_GEOMETRY in edge_dict:
                # geometry exists
                old_geometry = edge_dict[osm.KEY_OSMNX_GEOMETRY]
            else:
                # geometry does not exist
                # create it
                old_geometry = LineString(
                    [
                        (
                            network.nodes[edge_key[0]][osm.KEY_OSMNX_X],
                            network.nodes[edge_key[0]][osm.KEY_OSMNX_Y],
                        ),
                        (
                            network.nodes[edge_key[1]][osm.KEY_OSMNX_X],
                            network.nodes[edge_key[1]][osm.KEY_OSMNX_Y],
                        ),
                    ]
                )
            # if osm.KEY_OSMNX_LENGTH in edge_dict:
            #     # length exists
            #     old_length = edge_dict[osm.KEY_OSMNX_LENGTH]
            # else:
            #     # length does not exist
            #     old_length = edge_lengths(
            #         network,
            #         edge_keys=[edge_key])[edge_key]
            # the old length has to exist
            old_length = edge_dict[osm.KEY_OSMNX_LENGTH]

            # *****************************************************************
            # *****************************************************************

            # find closest point
            if edge_key[0] in roundabout:
                # this edge starts from the roundabout
                new_edge_start_node = roundabout_centroid_key
                new_edge_end_node = edge_key[1]
                # create geometry object between old roundabout point to the
                # roundabout's centroid
                extra_geometry = LineString(
                    [
                        (
                            network.nodes[roundabout_centroid_key][osm.KEY_OSMNX_X],
                            network.nodes[roundabout_centroid_key][osm.KEY_OSMNX_Y],
                        ),
                        (
                            network.nodes[edge_key[0]][osm.KEY_OSMNX_X],
                            network.nodes[edge_key[0]][osm.KEY_OSMNX_Y],
                        ),
                    ]
                )
                if is_projected(network.graph["crs"]):
                    # projected graph: use direct method
                    extra_length = length(extra_geometry)
                else:  # unprojected graph: use great circle method
                    extra_length = great_circle(
                        lat1=network.nodes[roundabout_centroid_key][osm.KEY_OSMNX_Y],
                        lon1=network.nodes[roundabout_centroid_key][osm.KEY_OSMNX_X],
                        lat2=network.nodes[edge_key[0]][osm.KEY_OSMNX_Y],
                        lon2=network.nodes[edge_key[0]][osm.KEY_OSMNX_X],
                    )
            elif edge_key[1] in roundabout:
                # this edge ends in the roundabout
                new_edge_start_node = edge_key[0]
                new_edge_end_node = roundabout_centroid_key
                # create geometry object between old roundabout point to the
                # roundabout's centroid
                extra_geometry = LineString(
                    [
                        (
                            network.nodes[roundabout_centroid_key][osm.KEY_OSMNX_X],
                            network.nodes[roundabout_centroid_key][osm.KEY_OSMNX_Y],
                        ),
                        (
                            network.nodes[edge_key[1]][osm.KEY_OSMNX_X],
                            network.nodes[edge_key[1]][osm.KEY_OSMNX_Y],
                        ),
                    ]
                )
                if is_projected(network.graph["crs"]):
                    # projected graph, use direct method
                    extra_length = length(extra_geometry)
                else:
                    # unprojected graph, use great circle method
                    extra_length = great_circle(
                        lat1=network.nodes[roundabout_centroid_key][osm.KEY_OSMNX_Y],
                        lon1=network.nodes[roundabout_centroid_key][osm.KEY_OSMNX_X],
                        lat2=network.nodes[edge_key[1]][osm.KEY_OSMNX_Y],
                        lon2=network.nodes[edge_key[1]][osm.KEY_OSMNX_X],
                    )

            # *****************************************************************
            # *****************************************************************

            edge_dict[osm.KEY_OSMNX_GEOMETRY] = linemerge(
                [old_geometry, extra_geometry]
            )
            edge_dict[osm.KEY_OSMNX_LENGTH] = old_length + extra_length
            network.add_edge(new_edge_start_node, new_edge_end_node, **edge_dict)

    # *************************************************************************
    # *************************************************************************

    # remove the roundabout nodes

    for roundabout_index, roundabout in enumerate(roundabouts):
        # if the transformation of the roundabout was successful...
        if list_roundabout_centroids[roundabout_index] != None:
            # remove the roundabout nodes
            network.remove_nodes_from(roundabout)
    # return
    return list_roundabout_centroids

    # *************************************************************************
    # *************************************************************************


# *****************************************************************************
# *****************************************************************************

# TODO: develop algorithm to traverse the graph in search of dead ends


def remove_dead_ends(
    network: nx.MultiDiGraph, 
    protected_nodes: list = None, 
    max_iterations: int = 1
) -> list:
    """
    Removes dead ends (non-cyclical branches) from a directed graph.

    The removal process is iterative and nodes that are initially not dead ends
    may be removed in subsequent iterations.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The object describing the directed graph.
    protected_nodes : list, optional
        A list of keys for the nodes that are not to be considered for removal.
        The default is None, which means all nodes are under consideration.
    max_iterations : int, optional
        The maximum number of iterations. The default is 1.

    Returns
    -------
    list
        A list of the keys for the nodes that were removed.

    """

    if type(protected_nodes) == type(None):
        protected_nodes = []

    # while true
    nodes_removed = []
    iteration_counter = 0
    while iteration_counter < max_iterations:
        # find nodes that can be excluded and that connect to only 1 other node
        # it is necessary to rule out that more one 1 node is involved
        target_nodes = [
            node_key
            for node_key in network.nodes()
            if node_key not in protected_nodes
            # if it has at most one neighbour other than itself
            if len(set(gis_iden.neighbours(network, node_key, ignore_self_loops=True)))
            <= 1
        ]
        # if there no nodes meeting those conditions, break out of loop
        if len(target_nodes) == 0:
            break
        # otherwise, find straight paths from these nodes
        # remove all the nodes in these paths
        network.remove_nodes_from(target_nodes)
        # increment iteration counter
        iteration_counter += 1
        # store the nodes removed
        nodes_removed.extend(target_nodes)
    # return the list of nodes removed
    return nodes_removed


# *****************************************************************************
# *****************************************************************************


def replace_path(network: nx.MultiDiGraph, path: list) -> tuple:
    """
    Replaces a simplifiable path with one equivalent edge linking both ends.

    If there are parallel or anti-parallel edges along the path, only one will
    be used to create the new edge.

    There should only be one edge between each of the nodes on the path, since
    only one will be used to create the new edge. By default, the edges between
    the nodes should be in the forward direction, but this restriction can be
    lifted. In that case, the edges between the nodes can be in any direction.

    The intermediate nodes on the path will be removed.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The object describing the network graph.
    path : list
        A list of keys for the nodes on the path.

    Raises
    ------
    ValueError
        This error is raised if the path is invalid.

    Returns
    -------
    tuple
        The key for the new edge.

    """

    # *************************************************************************

    # make sure path it is a simplifiable path
    if not gis_iden.is_path_straight(
        network, path, consider_reversed_edges=True, ignore_self_loops=True
    ):
        raise ValueError("The path cannot be simplified.")

    # *************************************************************************

    # create the new edge

    # create the geometry/linestring
    list_oneway = []
    list_reversed = []
    list_osmid = []
    list_geometries = []

    edge_length = 0

    for node_pair_index in range(len(path) - 1):
        # get one edge for this node pair
        edge_key = list(
            gis_iden.get_edges_between_two_nodes(
                network, path[node_pair_index], path[node_pair_index + 1]
            )
        )
        edge_key = sorted(
            (network.edges[_key][osm.KEY_OSMNX_LENGTH], _key) for _key in edge_key
        )[0][1]
        if osm.KEY_OSMNX_GEOMETRY in network.edges[edge_key]:
            # geometry exists: possibly a composite geometry
            # check if the geometry is consistent with the edge declaration
            if gis_iden.is_edge_consistent_with_geometry(network, edge_key):
                # the geometry is not reversed
                list_geometries.append(network.edges[edge_key][osm.KEY_OSMNX_GEOMETRY])
            else:  # the geometry is reversed
                list_geometries.append(
                    network.edges[edge_key][osm.KEY_OSMNX_GEOMETRY].reverse()
                )
        else:
            # geometry does not exist: direct path
            # the edge is not reversed: use it as is
            list_geometries.append(
                LineString(
                    [
                        (
                            network.nodes[path[node_pair_index]][osm.KEY_OSMNX_X],
                            network.nodes[path[node_pair_index]][osm.KEY_OSMNX_Y],
                        ),
                        (
                            network.nodes[path[node_pair_index + 1]][osm.KEY_OSMNX_X],
                            network.nodes[path[node_pair_index + 1]][osm.KEY_OSMNX_Y],
                        ),
                    ]
                )
            )

        # osmid
        if type(network.edges[edge_key][osm.KEY_OSMNX_OSMID]) == list:
            list_osmid.extend(network.edges[edge_key][osm.KEY_OSMNX_OSMID])
        else:
            list_osmid.append(network.edges[edge_key][osm.KEY_OSMNX_OSMID])
        # reversed
        if type(network.edges[edge_key][osm.KEY_OSMNX_REVERSED]) == list:
            list_reversed.extend(network.edges[edge_key][osm.KEY_OSMNX_REVERSED])
        else:
            list_reversed.append(network.edges[edge_key][osm.KEY_OSMNX_REVERSED])
        # oneway
        if type(network.edges[edge_key][osm.KEY_OSMNX_ONEWAY]) == list:
            list_oneway.extend(network.edges[edge_key][osm.KEY_OSMNX_ONEWAY])
        else:
            list_oneway.append(network.edges[edge_key][osm.KEY_OSMNX_ONEWAY])

        # update the edge length
        edge_length += network.edges[edge_key][osm.KEY_OSMNX_LENGTH]

    # merge the geometries
    new_geo = linemerge(list_geometries)

    # verify that it led to the creation of a linestring object
    if type(new_geo) != LineString:
        # TODO: make sure this is still relevant and add tests
        # if not a linestring, it is a multistring
        # assert type(new_geo) == MultiLineString
        list_geometries = []
        # snap each consecutive geometry pair in the MultiLineString object
        # since linemerge separates linestrings that are not contiguous
        for geo_pair_index in range(len(new_geo.geoms) - 1):
            list_geometries.append(
                snap(
                    new_geo.geoms[geo_pair_index],
                    new_geo.geoms[geo_pair_index + 1],
                    tolerance=1e-3,
                )
            )
        new_geo = linemerge(list_geometries)

    # prepare edge data dict
    list_osmid = list(set(list_osmid))
    list_oneway = list(set(list_oneway))
    list_reversed = list(set(list_reversed))
    # create the dict
    edge_dict = {
        osm.KEY_OSMNX_LENGTH: edge_length,
        osm.KEY_OSMNX_GEOMETRY: new_geo,
        osm.KEY_OSMNX_ONEWAY: (
            list_oneway if len(list_oneway) != 1 else list_oneway[0]
        ),
        osm.KEY_OSMNX_REVERSED: (
            list_reversed if len(list_reversed) != 1 else list_reversed[0]
        ),
        osm.KEY_OSMNX_OSMID: (list_osmid if len(list_osmid) != 1 else list_osmid[0]),
    }

    # *************************************************************************

    # add edges
    start_node = path[0]
    end_node = path[-1]

    # create the forward edge
    for_k = network.add_edge(start_node, end_node, **edge_dict)

    # delete all intermediate nodes
    network.remove_nodes_from(path[1:-1])

    # return the edge key
    return (start_node, end_node, for_k)


# *****************************************************************************
# *****************************************************************************


def remove_longer_parallel_edges(
    network: nx.MultiDiGraph, 
    distance_key: str = osm.KEY_OSMNX_LENGTH, 
    ignore_edge_directions: bool = True,
    protected_edges: list = None
) -> list:
    """
    Removes longer parallel edges from the network.
    
    Parallel edges refer to multiple edges connecting the same nodes. If there 
    are parallel edges between any given pair of nodes, only the shortest one 
    will be retained. If desired, edge directions can be ignored. By default,
    edge directions are taken into account when selecting parallel edges.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The object describing the graph.
    distance_key : str, optional
        The key used to obtain distances. The default is osm.KEY_OSMNX_LENGTH.
    ignore_edge_directions : bool, optional
        If True, edge directions are ignored. The default is True.
    protected_edges : list, optional
        A list of edges that should be retained. The default is None, which
        means all edges are elligible.

    Returns
    -------
    list
        A list of the edges removed.

    """
    if type(protected_edges) == type(None):
        # default: empty list
        protected_edges = []
    edges_removed = []
    for node_one in network.nodes():
        for node_two in network.nodes():
            # skip self-loops
            if node_one == node_two:
                continue
            # get the edges between the two nodes
            if ignore_edge_directions:  # both directions
                list_edges = gis_iden.get_edges_between_two_nodes(
                    network, node_one, node_two
                )
            else:  # one direction
                list_edges = gis_iden.get_edges_from_a_to_b(
                    network, node_start=node_one, node_end=node_two
                )
            # if none exist, skip
            if len(list_edges) == 0:
                continue
            # otherwise, sort them by distance (shorter distances first)
            # note: protected edges are considered in the sorting too
            sorted_edges = sorted(
                (network.edges[edge_key][distance_key], edge_key)
                for edge_key in list_edges
            )
            # edges to be removed (drop protected edges here)
            edges_for_removal = tuple(
                edge_tuple[1] 
                for edge_tuple in sorted_edges[1:]
                if edge_tuple[1] not in protected_edges
                )
            # remove all but the shortest edge
            network.remove_edges_from(edges_for_removal)
            # update the list of edges removed
            edges_removed.extend(edges_for_removal)
    # return statement
    return edges_removed

# *****************************************************************************
# *****************************************************************************

def merge_points_into_linestring(
    line: LineString,
    points: tuple or list,
    tolerance: float = 7.0 / 3 - 4.0 / 3 - 1,
    fixed_extremities: bool = True,
    use_start_point_equidistant: bool = True,
) -> LineString:
    """
    Merge points into a line where they are closest to it.

    The points are merged in succession and the line keeps changing as the
    points are merged but the original points remain on it.

    The points added do not need to be close to the line. The tolerance
    parameter is only used to determine closeness to the points already on the
    line.

    If a point is closest to a point on the line shared by two line segments,
    then the former is placed after the latter, and not before.

    If a point is closest to multiple line segments, then the point is placed
    between the start and end points of the first among the closest segments.

    Parameters
    ----------
    line : LineString
        The LineString object into which Point objects are to be merged.
    points : tuple or list
        A list of Point objects that are to be merged.
    tolerance : float, optional
        The tolerance with which closeness to points on the line is defined.
        The default is 7/3-4/3-1 (2.220446049250313e-16).
    fixed_extremities : bool, optional
        If False, the line can be extended beyond the original start and end
        points. If not, points closest to the extremities are not merged and
        are instead linked to the closest extremities. The default is True.
    use_start_point_equidistant : bool, optional
        If True, the start point is adopted as the closest point when the start
        and end points are equidistant from the main one. If False, the end
        point is adopted instead. The default is True.

    Returns
    -------
    new_line : LineString
        The LineString object with the points merged.
    close_to_start : list
        A list with the indices of the points closest to the start point.
    close_to_end: list
        A list with the indices of the points closest to the end point.

    """

    if fixed_extremities:
        # the line cannot be extended

        # identify which points are close to the start and end points
        # note: these points will not be merged

        (
            close_to_start,
            close_to_end,
            line_distances,
            start_distances,
            end_distances,
        ) = close_to_extremities(
            line,
            points,
            tolerance=tolerance,
            use_start_point_equidistant=use_start_point_equidistant,
            return_distances=True,
        )

        # for each mew point

        for i in range(len(points)):
            if i in close_to_start or i in close_to_end:
                # the point is close to the start or end nodes: skip iteration

                continue

            if points[i].coords[0] in line.coords:
                # this point is already on the line: skip iteration

                continue

            # merge the points that are between the start and end points...

            # create line segments for each pair of points on the line

            line_segments = [
                LineString([line.coords[j], line.coords[j + 1]])
                for j in range(len(line.coords) - 1)
            ]

            # calculate the distances between point i and each point on the li.

            line_segment_distances = points[i].distance(line_segments)

            # locate the closest pair of points

            sorted_distances = sorted(
                (line_segment_distance, j)
                for j, line_segment_distance in enumerate(line_segment_distances)
            )
            # prepare new line coordinates with the new point

            line_coords = list(line.coords)

            if (
                len(sorted_distances) >= 2
                and sorted_distances[0][0] == sorted_distances[1][0]
            ):
                # there are 2(+) segments that are equally close to the point

                # if the closest points are end/start points of a segment, then
                # place the point after the second point of the first segment

                if (
                    abs(
                        Point(
                            line_segments[sorted_distances[0][1]].coords[-1]
                        ).distance(points[i])
                        - line_distances[i]
                    )
                    <= tolerance
                ):
                    line_coords.insert(
                        # sorted_distances[0][1]+1,
                        sorted_distances[0][1] + 2,
                        tuple(points[i].coords[0]),
                    )

                else:
                    line_coords.insert(
                        sorted_distances[0][1] + 1,
                        # sorted_distances[0][1]+2,
                        tuple(points[i].coords[0]),
                    )

            else:
                # there is only segment with the minimum distance:
                # place the new point where the end point of the segment is

                line_coords.insert(
                    sorted_distances[0][1] + 1,  # i.e., the segment number + 1
                    tuple(points[i].coords[0]),
                )

            # create new line

            line = LineString(line_coords)

    else:
        # the line can be extended

        raise NotImplementedError

    # return statement

    return line, close_to_start, close_to_end


# *****************************************************************************
# *****************************************************************************


def split_linestring(
    line: LineString, points: list, tolerance: float = 7.0 / 3 - 4.0 / 3 - 1
):
    """
    Split a line into segments according to a set of cutting points.

    The cutting points should be close to the original line, since they will be
    merged into it before the cutting takes place. If they are not close to the
    line, the resulting segments will not resemble the original line.

    Parameters
    ----------
    line : LineString
        The object describing the original line.
    points : list
        A list with the points with which the line is to be split.
    tolerance : float, optional
        The tolerance with which closeness to other points is defined.
        The default is 7./3-4./3-1.

    Returns
    -------
    line_segments : list
        A list with the line segments.
    close_to_start : list
        A list with the indices of the points closest to the start point.
    close_to_end : list
        A list with the indices of the points closest to the end point.

    """

    # add the points to the linestring
    new_line, close_to_start, close_to_end = merge_points_into_linestring(
        line=line, points=points, tolerance=tolerance, fixed_extremities=True
    )

    if len(close_to_end) + len(close_to_start) == len(points):
        # no changes to the line (assuming no swaps)
        return [], close_to_start, close_to_end

    # split the linestring object (new_line)
    line_segments = []
    previous_split_index = 0
    # for each point on the new line (they should be ordered)
    for coords_index, coords in enumerate(new_line.coords):
        # if it matches one of the input points
        if Point(coords) in points:
            # this is a cutting point
            if new_line.coords[0] == coords or new_line.coords[-1] == coords:
                # it is a start or end point: skip the iteration
                # line_segments.append(None)
                continue

            # if it is not a start nor an end point, and it is on the original
            # line, then
            # if not a start nor an end point, build the segment between the
            # previous split point and the current input point
            line_segments.append(
                LineString(new_line.coords[previous_split_index : coords_index + 1])
            )

            # store new end/start point
            previous_split_index = coords_index

        # else:

        #     # the point is not on the original line: split point

        #     pass

        # next iteration

    # add the last segment
    line_segments.append(LineString(new_line.coords[previous_split_index:]))

    # return the geometries for each segment and the relevant points by order
    return line_segments, close_to_start, close_to_end

    # *************************************************************************


# *****************************************************************************
# *****************************************************************************


def recreate_edges(
    network: nx.MultiDiGraph, points: dict, tolerance: float = 7.0 / 3 - 4.0 / 3 - 1
) -> tuple:
    """
    Recreates OSMnx-type edges by splitting them into multiple smaller edges,
    which are defined by points along the original edge.

    If the points are closest to the extremities than other parts of an edge,
    no changes are introduced and the points become synonymous with the closest
    extremity (i.e., the start or end points).

    If the points are closest to other parts of an edge, the edge is split there
    with new nodes and edges being created to replace the original edge.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The object describing the network with the edges and nodes.
    points : dict
        A dictionary keyed by edge and whose values are the Point geometries
        with which to split and then recreate the edge.

    Returns
    -------
    dict
        A dictionary keyed by edge and holding the node keys for the points that
        were provided initially to split the edge. These node keys are for nodes
        that already existed (start or end nodes) or newly created ones.
    dict
        A dictionary keyed by edge and holding the keys for the edges that were
        created to replace the original edge. If a given edge was not recreated,
        its key does not appear in the dictionary.

    """

    # declare outputs

    connection_node_keys_per_edge = {}

    recreated_edges = {}

    # for each edge that is to be split

    for edge_key, points_in_edge in points.items():
        # check if there is a geometry already
        if osm.KEY_OSMNX_GEOMETRY in network.edges[edge_key]:
            # get the geometry
            line = network.edges[edge_key][osm.KEY_OSMNX_GEOMETRY]
            # check if the geometry is consistent with the edge declaration
            if not gis_iden.is_edge_consistent_with_geometry(network, edge_key):
                # the geometry must be reversed
                line = LineString(line.reverse())
        else:
            # there is not geometry, create it
            line = LineString(
                [
                    (
                        network.nodes[edge_key[0]][osm.KEY_OSMNX_X],
                        network.nodes[edge_key[0]][osm.KEY_OSMNX_Y],
                    ),
                    (
                        network.nodes[edge_key[1]][osm.KEY_OSMNX_X],
                        network.nodes[edge_key[1]][osm.KEY_OSMNX_Y],
                    ),
                ]
            )

        # split the line into segments using the intermediate points
        # note: points that are close to the start and end points are handled
        # separately

        line_segments, close_to_start, close_to_end = split_linestring(
            line=line, points=points_in_edge
        )

        # link each point to a node key:
        # those closer to the start node: edge_key[0]
        # those closer to the end node: edge_key[1]
        # intermediate points: new node key via unused_node_key

        # _node_keys_by_point = {
        #     points_in_edge[i]: (
        #         edge_key[0] if i in close_to_start else edge_key[1]
        #         ) if (
        #             i in close_to_start or i in close_to_end
        #             ) else unused_node_key(network)
        #     for i in range(len(points_in_edge))
        #     }

        _node_keys_by_point = {}

        for i in range(len(points_in_edge)):
            if i in close_to_start or i in close_to_end:
                # point i is close to the extremities: use start/end node key
                _node_keys_by_point[points_in_edge[i]] = (
                    edge_key[0] if i in close_to_start else edge_key[1]
                )
            else:
                # point i is not close to the extremities: new node key
                _node_keys_by_point[points_in_edge[i]] = unused_node_key(network)
                network.add_node(_node_keys_by_point[points_in_edge[i]])

        # _node_keys = [
        #     (edge_key[0] if i in close_to_start else edge_key[1]) if (
        #         i in close_to_start or i in close_to_end
        #         ) else unused_node_key(network)
        #     for i in range(len(points_in_edge))
        #     ]

        # should be the same order as in the inputs

        _node_keys = [_node_keys_by_point[point] for point in points_in_edge]

        # *********************************************************************

        # create new edges between the points to rebuild the edge

        segment_keys = []

        edge_dict = dict(network.get_edge_data(*edge_key))

        for line_index, line_segment in enumerate(line_segments):
            edge_dict[osm.KEY_OSMNX_GEOMETRY] = line_segment

            if line_index == 0:
                # initial segment

                v_key = _node_keys_by_point[Point(line_segment.coords[-1])]

                k_key = network.add_edge(
                    u_for_edge=edge_key[0], v_for_edge=v_key, **edge_dict
                )

                network.add_node(
                    v_key,
                    **{
                        osm.KEY_OSMNX_X: line_segment.coords[-1][0],
                        osm.KEY_OSMNX_Y: line_segment.coords[-1][1],
                    }
                )

                segment_keys.append((edge_key[0], v_key, k_key))

            elif line_index == len(line_segments) - 1:
                # final segment

                u_key = _node_keys_by_point[Point(line_segment.coords[0])]

                k_key = network.add_edge(
                    u_for_edge=u_key, v_for_edge=edge_key[1], **edge_dict
                )

                network.add_node(
                    u_key,
                    **{
                        osm.KEY_OSMNX_X: line_segment.coords[0][0],
                        osm.KEY_OSMNX_Y: line_segment.coords[0][1],
                    }
                )

                segment_keys.append((u_key, edge_key[1], k_key))

            else:  # intermediate segment
                u_key = _node_keys_by_point[Point(line_segment.coords[0])]

                v_key = _node_keys_by_point[Point(line_segment.coords[-1])]

                k_key = network.add_edge(
                    u_for_edge=u_key, v_for_edge=v_key, **edge_dict
                )

                network.add_node(
                    u_key,
                    **{
                        osm.KEY_OSMNX_X: line_segment.coords[0][0],
                        osm.KEY_OSMNX_Y: line_segment.coords[0][1],
                    }
                )

                network.add_node(
                    v_key,
                    **{
                        osm.KEY_OSMNX_X: line_segment.coords[-1][0],
                        osm.KEY_OSMNX_Y: line_segment.coords[-1][1],
                    }
                )

                segment_keys.append((u_key, v_key, k_key))

        # TODO: use network.add_edges_from() for performance?

        # TODO: try to create all the edges (with lengths included) in one go

        # calculate the lengths
        edge_lengths_by_dict = edge_lengths(network, edge_keys=segment_keys)
        network.add_edges_from(
            tuple(
                (
                    *segment_key,
                    {osm.KEY_OSMNX_LENGTH: edge_lengths_by_dict[segment_key]},
                )
                for segment_key in segment_keys
            )
        )

        # update the outputs
        if len(line_segments) > 0:
            recreated_edges[edge_key] = segment_keys
        connection_node_keys_per_edge[edge_key] = _node_keys
    # return statement
    return connection_node_keys_per_edge, recreated_edges


# *****************************************************************************
# *****************************************************************************


def connect_nodes_to_edges(
    network: nx.MultiDiGraph,
    node_keys: list,
    edge_keys: list,
    store_unsimplified_geometries: bool = False,
    use_one_edge_per_direction: bool = False,
) -> tuple:
    """
    Connects nodes to edges using additional edges in an OSMnx-formatted graph.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The object describing the network.
    node_keys : list
        The keys of the nodes that are to be connected.
    edge_keys : list
        The keys of the edges that are to be connected to the nodes. It must be
        ordered based on node_keys and must have the same size.
    store_unsimplified_geometries : bool, optional
        If True, straight line geometries that are created are also preserved.
        If False, they are not preserved. The default is False.
    use_one_edge_per_direction : bool, optional
        If True, two edges are used for each new edge created to connect a node.
        If False, only one (forward) edge will be created. The default is False.

    Returns
    -------
    network : nx.MultiDiGraph
        A network graph object where the node and edge pairs are connected.
    new_edge_keys : list
        An ordered list containing the keys for the new edges created to connect
        each node.
    connection_node_keys_per_edge : dict
        A dictionary keyed by edge and holding the node keys for the points that
        were provided initially to split the edge. These node keys are for nodes
        that already existed (start or end nodes) or newly created ones.
    recreated_edges : dict
        A dictionary keyed by edge and holding the keys for the edges that were
        created to replace the original edge. If a given edge was not recreated,
        its key does not appear in the dictionary.

    """

    # *************************************************************************

    # 1) group nodes by the edge they are closest to
    # 2) for each edge, and node that is to be connected to it, find its closest
    # point on the edge
    # 3) recreate each edge after dividing it at the specified points
    # 4) connect the nodes to the edges
    # 5) delete the original edges, if they have been split
    # 6) calculate or update the edge attributes

    # *************************************************************************
    # *************************************************************************

    # 1) group nodes by the edge they are closest to

    nodes_to_connect_to_edge = {
        edge_key: tuple(
            node_key
            for other_edge_key, node_key in zip(edge_keys, node_keys)
            if other_edge_key == edge_key
        )
        for edge_key in set(edge_keys)
    }

    # *************************************************************************
    # *************************************************************************

    # 2) for each edge, and node that is to be connected to it, find its closest
    # point on the edge
    points_per_edge = {}
    for edge_key, _node_keys in nodes_to_connect_to_edge.items():
        # check if the geometry exists
        if osm.KEY_OSMNX_GEOMETRY in network.edges[edge_key]:
            # the geometry object exists, get it
            edge_geo = network.edges[edge_key][osm.KEY_OSMNX_GEOMETRY]
        else:
            # the geometry object does not exist, make it
            edge_geo = LineString(
                [
                    (
                        network.nodes[edge_key[0]][osm.KEY_OSMNX_X],
                        network.nodes[edge_key[0]][osm.KEY_OSMNX_Y],
                    ),
                    (
                        network.nodes[edge_key[1]][osm.KEY_OSMNX_X],
                        network.nodes[edge_key[1]][osm.KEY_OSMNX_Y],
                    ),
                ]
            )
            # store the geometry
            if store_unsimplified_geometries:
                # update the edge
                network.add_edge(*edge_key, **{osm.KEY_OSMNX_GEOMETRY: edge_geo})
        # use nearest_points to locate the closest points on the edge
        points_per_edge[edge_key] = [
            nearest_points(
                edge_geo,
                Point(
                    network.nodes[node_key][osm.KEY_OSMNX_X],
                    network.nodes[node_key][osm.KEY_OSMNX_Y],
                ),
            )[
                0
            ]  # [0] to get the point on the edge
            for node_key in _node_keys
        ]
        # TIP: exclude the points that can be considered close to the start or end nodes
    # TIP: use the shortest line method to obtain the line geometry

    # *************************************************************************
    # *************************************************************************

    # 3) recreate each edge after dividing it at the specified points

    connection_node_keys_per_edge, recreated_edges = recreate_edges(
        network, points=points_per_edge
    )

    # put the keys for the connection nodes

    connection_node_keys = [
        connection_node_keys_per_edge[edge_key][
            nodes_to_connect_to_edge[edge_key].index(node_key)
        ]
        for node_key, edge_key in zip(node_keys, edge_keys)
    ]

    # delete the original edges, if they have been split

    network.remove_edges_from(recreated_edges)

    # *************************************************************************
    # *************************************************************************

    # 4) connect the nodes to the edges

    connection_edge_containers = []

    for node_key, connection_node_key in zip(node_keys, connection_node_keys):
        # skip self-loops
        if node_key == connection_node_key:
            continue
        # proceed with other types of edges
        if use_one_edge_per_direction:
            # add one directed edge per direction
            connection_edge_containers.append((node_key, connection_node_key))
            connection_edge_containers.append((connection_node_key, node_key))
        else:
            # add one directed edge starting from the edge and ending in the node
            connection_edge_containers.append((connection_node_key, node_key))
    edge_keys = network.add_edges_from(connection_edge_containers)

    # *************************************************************************
    # *************************************************************************

    # 5) calculate or update the edge attributes

    # calculate edge lengths and street counts for the new edges
    if len(edge_keys) != 0:
        # there are new edges: calculate the lengths and add them
        new_edge_keys = [
            (*edge_tuple[0:2], edge_key)  # apply it only to specific edges
            for edge_tuple, edge_key in zip(connection_edge_containers, edge_keys)
        ]

        if is_projected(network.graph["crs"]):
            # projected crs: use own method
            lengths_dict = edge_lengths(network, edge_keys=new_edge_keys)

            network.add_edges_from(
                tuple(
                    (
                        *edge_key,
                        {
                            osm.KEY_OSMNX_LENGTH: lengths_dict[edge_key],
                            osm.KEY_OSMNX_ONEWAY: False,
                            osm.KEY_OSMNX_REVERSED: False,
                            osm.KEY_OSMNX_OSMID: None,
                        },
                    )
                    for edge_key in new_edge_keys
                )
            )

        else:
            # unprojected crs: use the osmnx method
            network = add_edge_lengths(network, edges=new_edge_keys)

        # update the street count
        update_street_count(network)
    else:
        new_edge_keys = []

    # *************************************************************************
    # *************************************************************************

    return network, new_edge_keys, connection_node_keys_per_edge, recreated_edges


# *****************************************************************************
# *****************************************************************************


def remove_reversed_edges(network: nx.MultiDiGraph, reversed_attr: bool = True) -> list:
    """
    Removes reversed edges from an OSMnx-formatted multi directed edge graph.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The object describing the graph.
    reversed_attr : bool, optional
        If True, the edges removed are those whose \'reversed\' attribute
        equals True. The default is True.

    Returns
    -------
    list
        A list with keys of the edges removed.

    """
    reversed_edges = gis_iden.find_edges_in_reverse(network)
    # conditions to remove an edge:
    # 1) it must have at least one reversed edge
    # 2) its reversed edge(s) cannot be removed
    # 3) it must satisfy the reversed attribute value
    edges_removed = [
        edge_key
        for edge_key, reverse_edges in reversed_edges.items()
        # at least one reversed edge
        if len(reverse_edges) >= 1
        # must satisfy the reversed attribute value
        if (
            network.edges[edge_key][osm.KEY_OSMNX_REVERSED] == reversed_attr
            or type(network.edges[edge_key][osm.KEY_OSMNX_REVERSED]) == list
        )
    ]
    # filter
    for edge_key in edges_removed:
        for other_edge_key in reversed_edges[edge_key]:
            if other_edge_key in edges_removed:
                edges_removed.remove(other_edge_key)
    # remove the edges
    network.remove_edges_from(edges_removed)
    # return
    return edges_removed


# *****************************************************************************
# *****************************************************************************


def create_reverse_edges(
    network: nx.MultiDiGraph,
    edge_keys: list = None,
) -> list:
    """
    Creates reversed edges for all or select edges in a network.

    A reversed edge has the same attributes as the original except that it is
    declared in the opposite direction and has reversed geometry and related
    attributes. The edges created are compliant with OSMnx.

    Parameters
    ----------
    network : nx.MultiDiGraph
        The object describing the network.
    edge_keys : list, optional
        The keys to the edges that are to be recreated in reverse. The default
        is None, which means all existing edges will be recreated.

    Raises
    ------
    ValueError
        This error is raised if one or more edges do not exist in the network.

    Returns
    -------
    list
        A list with the keys for the new edges.

    """

    edges_created = []

    if type(edge_keys) == type(None):
        edge_keys = tuple(network.edges(keys=True))

    for edge_key in edge_keys:
        # make sure the edge exists
        if not network.has_edge(*edge_key):
            raise ValueError("Unknown edge: " + str(edge_key))
        # get its data
        edge_dict = network.get_edge_data(*edge_key)
        # create a dict for the reversed edge
        rev_edge_dict = dict(edge_dict)
        # check for the reversed keyword
        if type(edge_dict[osm.KEY_OSMNX_REVERSED]) == bool:
            # boolean: negate it
            rev_edge_dict[osm.KEY_OSMNX_REVERSED] = not edge_dict[
                osm.KEY_OSMNX_REVERSED
            ]
        elif (
            type(edge_dict[osm.KEY_OSMNX_REVERSED]) == list
            and len(edge_dict[osm.KEY_OSMNX_REVERSED]) == 2
            and len(set(edge_dict[osm.KEY_OSMNX_REVERSED])) == 2
        ):
            # list:
            rev_edge_dict[osm.KEY_OSMNX_REVERSED] = [True, False]
        else:
            raise ValueError(
                "The edge " + str(edge_key) + "is not compliant with OSMnx."
            )
        # check for the geometry keyword
        if osm.KEY_OSMNX_GEOMETRY in edge_dict:
            # a geometry exists, reverse it for the reverse edge dict
            rev_edge_dict[osm.KEY_OSMNX_GEOMETRY] = edge_dict[
                osm.KEY_OSMNX_GEOMETRY
            ].reverse()
        # add the edge
        rev_k = network.add_edge(edge_key[1], edge_key[0], **rev_edge_dict)

        edges_created.append((edge_key[1], edge_key[0], rev_k))
    # return the keys for the edges created
    return edges_created


# *****************************************************************************
# *****************************************************************************
