# imports

import pyomo.environ as pyo
    
# *****************************************************************************
# *****************************************************************************

def price_delta_block(
    model: pyo.AbstractModel,
    enable_default_values: bool = True,
    enable_validation: bool = True,
    enable_initialisation: bool = True
    ):
    
    # auxiliary set for pyomo
    model.set_GLQPK = model.set_GL_exp_imp*model.set_QPK
    
    # create a block for a transshipment node during a given time interval
    def rule_block_prices(b, g, l, q, p, k):
        
        # *********************************************************************
        # *********************************************************************
        
        # sets
        
        # set of price segments
        b.set_S = pyo.Set()
    
        # *********************************************************************
        # *********************************************************************
        
        # parameters

        # resource prices
        b.param_p_s = pyo.Param(b.set_S, within=pyo.NonNegativeReals)
        
        # price function convexity
        b.param_price_function_is_convex = pyo.Param(within=pyo.Boolean)

        # maximum resource volumes for each prices
        b.param_v_max_s = pyo.Param(b.set_S, within=pyo.NonNegativeReals)
    
        # *********************************************************************
        # *********************************************************************
    
        # variables
    
        # *********************************************************************
        # *********************************************************************
        
        # TODO: consider replacing None in the bounds with inf in the parameter
    
        # import and export flows
        def bounds_var_trans_flows(b):
            try:
                return (0, sum(b.param_v_max_s[s] for s in b.set_S))
            except Exception:
                return (0, None)
        b.var_trans_flows = pyo.Var(
            # within=pyo.NonNegativeReals,
            bounds=bounds_var_trans_flows
        )
        
        # segment usage variables
        b.var_segment_usage_s = pyo.Var(
            b.set_S, 
            within=pyo.UnitInterval, 
        )
        
        # *********************************************************************
        # *********************************************************************
    
        # import flow costs and export flow revenues (y function)
        def rule_constr_trans_monetary_flows(b):
            if (g,l) in b.parent_block().set_GL_imp:
                return (
                    sum(
                        b.var_segment_usage_s[s]*b.param_p_s[s]*b.param_v_max_s[s]
                        for s in b.set_S
                    )
                    == b.parent_block().var_ifc_glqpk[(g,l,q,p,k)]
                )
            else:
                return (
                    sum(
                        b.var_segment_usage_s[s]*b.param_p_s[s]*b.param_v_max_s[s]
                        for s in b.set_S
                    )
                    == b.parent_block().var_efr_glqpk[(g,l,q,p,k)]
                )
        b.constr_trans_monetary_flows = pyo.Constraint(
            rule=rule_constr_trans_monetary_flows
        )        
        
        # imported and exported flows (x function)
        def rule_constr_trans_flows_seg(b):
            return sum(
                b.var_segment_usage_s[s]*b.param_v_max_s[s]
                for s in b.set_S
            ) == b.var_trans_flows
        b.constr_trans_flows_seg = pyo.Constraint(rule=rule_constr_trans_flows_seg)
        
        # imported and exported flows: energy system defines the x value
        def rule_constr_trans_flows_sys(b):
            if (g,l) in b.parent_block().set_GL_imp:
                return sum(
                    b.parent_block().var_v_glljqk[(g,l,l_star,j,q,k)]
                    for l_star in b.parent_block().set_L[g]
                    if l_star not in b.parent_block().set_L_imp[g]
                    for j in b.parent_block().set_J[(g,l,l_star)]  # only directed arcs
                ) == b.var_trans_flows
            else:
                return sum(
                    b.parent_block().var_v_glljqk[(g,l_star,l,j,q,k)]
                    * b.parent_block().param_eta_glljqk[(g,l_star,l,j,q,k)]
                    for l_star in b.parent_block().set_L[g]
                    if l_star not in b.parent_block().set_L_exp[g]
                    for j in b.parent_block().set_J[(g,l_star,l)]  # only directed arcs
                ) == b.var_trans_flows
        b.constr_trans_flows_sys = pyo.Constraint(rule=rule_constr_trans_flows_sys)
    
        # *********************************************************************
        # *********************************************************************
        
        # non-convex price functions
        if not b.param_price_function_is_convex:
            # this means the segment usage variables are binary
            
            # variables to indicate if the segments are active
            b.var_segment_is_full = pyo.Var(
                b.set_S, 
                within=pyo.Binary
            )
            
            # if next level is active, bin. var. for current level must be one
            # if bin. var. for current level is one, segment must be used fully
            # if previous level is not full, bin. var for current level must be zero
            # if bin. var. for current level is zero, 
            def rule_constr_nonconvex_p1(b, s):
                if s == len(b.set_S)-1:
                    # last segment, skip
                    return pyo.Constraint.Skip
                return (
                    b.var_segment_usage_s[s+1] <= 
                    b.var_segment_is_full[s]
                    )
            b.constr_nonconvex_p1 = pyo.Constraint(
                b.set_S, rule=rule_constr_nonconvex_p1
                )
            
            def rule_constr_nonconvex_p2(b, s):
                if s == len(b.set_S)-1:
                    # last segment, skip
                    return pyo.Constraint.Skip
                return (
                    b.var_segment_usage_s[s] >= 
                    b.var_segment_is_full[s]
                    )
            b.constr_nonconvex_p2 = pyo.Constraint(
                b.set_S, rule=rule_constr_nonconvex_p2
                )
            
            def rule_constr_nonconvex_p3(b, s):
                if s == len(b.set_S)-1:
                    # last segment, skip
                    return pyo.Constraint.Skip
                return (
                    b.var_segment_usage_s[s] >= 
                    b.var_segment_is_full[s+1]
                    )
            b.constr_nonconvex_p3 = pyo.Constraint(
                b.set_S, rule=rule_constr_nonconvex_p3
                )
            
            def rule_constr_nonconvex_p4(b, s):
                if s == len(b.set_S)-1:
                    # last segment, skip
                    return pyo.Constraint.Skip
                return (
                    b.var_segment_usage_s[s+1] <= 
                    b.var_segment_is_full[s+1]
                    )
            b.constr_nonconvex_p4 = pyo.Constraint(
                b.set_S, rule=rule_constr_nonconvex_p4
                )
            
        
        # *********************************************************************
        # ********************************************************************* 
        
    model.block_prices = pyo.Block(model.set_GLQPK, rule=rule_block_prices)

    # *************************************************************************
    # *************************************************************************

# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************

def price_delta_no_block(
    model: pyo.AbstractModel,
    # convex_price_function: bool = True,
    enable_default_values: bool = True,
    enable_validation: bool = True,
    enable_initialisation: bool = True
    ):
    
    # auxiliary set for pyomo
    model.set_GLQPK = model.set_GL_exp_imp*model.set_QPK

    # set of price segments
    model.set_S = pyo.Set(model.set_GLQPK)

    # set of GLQKS tuples
    def init_set_GLQPKS(m):
        return (
            (g, l, q, p, k, s)
            # for (g,l) in m.set_GL_exp_imp
            # for (q,k) in m.set_QK
            for (g, l, q, p, k) in m.set_S
            for s in m.set_S[(g, l, q, p, k)]
        )

    model.set_GLQPKS = pyo.Set(
        dimen=6, initialize=(init_set_GLQPKS if enable_initialisation else None)
    )

    # *************************************************************************
    # *************************************************************************

    # parameters

    # resource prices

    model.param_p_glqpks = pyo.Param(model.set_GLQPKS, within=pyo.NonNegativeReals)
    
    # price function convexity

    model.param_price_function_is_convex = pyo.Param(
        model.set_GLQPK, 
        within=pyo.Boolean
        )

    # maximum resource volumes for each prices

    model.param_v_max_glqpks = pyo.Param(
        model.set_GLQPKS, 
        within=pyo.NonNegativeReals,
        # default=math.inf
        )

    # *************************************************************************
    # *************************************************************************

    # variables

    # *************************************************************************
    # *************************************************************************

    # import and export flows
    def bounds_var_trans_flows_glqpk(m, g, l, q, p, k):
        try:
            return (0, sum(m.param_v_max_glqpks[s] for s in m.set_S[(g, l, q, p, k)]))
        except Exception:
            return (0, None)
    model.var_trans_flows_glqpk = pyo.Var(
        model.set_GLQPK, within=pyo.NonNegativeReals, bounds=bounds_var_trans_flows_glqpk
    )
            
    # segment usage variables
    model.var_segment_usage_glqpks = pyo.Var(
        model.set_GLQPKS, 
        within=pyo.UnitInterval, 
    )

    # *************************************************************************
    # *************************************************************************
    
    
    # import flow costs and export flow revenues (y function)
    def rule_constr_trans_monetary_flows(m, g, l, q, p, k):
        if (g,l) in m.set_GL_imp:
            return (
                sum(
                    m.var_segment_usage_glqpks[(g, l, q, p, k, s)]*
                    m.param_p_glqpks[(g, l, q, p, k, s)]*
                    m.param_v_max_glqpks[(g, l, q, p, k, s)]
                    for s in m.set_S[(g, l, q, p, k)]
                )
                == m.var_ifc_glqpk[(g,l,q,p,k)]
            )
        else:
            return (
                sum(
                    m.var_segment_usage_glqpks[(g, l, q, p, k, s)]*
                    m.param_p_glqpks[(g, l, q, p, k, s)]*
                    m.param_v_max_glqpks[(g, l, q, p, k, s)]
                    for s in m.set_S[(g, l, q, p, k)]
                )
                == m.var_efr_glqpk[(g,l,q,p,k)]
            )
    model.constr_trans_monetary_flows = pyo.Constraint(
        model.set_GLQPK,
        rule=rule_constr_trans_monetary_flows
    )        
    
    # imported and exported flows (x function)
    def rule_constr_trans_flows_seg(m, g, l, q, p, k):
        return sum(
            m.var_segment_usage_glqpks[(g, l, q, p, k, s)]*m.param_v_max_glqpks[(g, l, q, p, k, s)]
            for s in m.set_S[(g, l, q, p, k)]
        ) == m.var_trans_flows_glqpk[(g, l, q, p, k)]
    model.constr_trans_flows_seg = pyo.Constraint(
        model.set_GLQPK,
        rule=rule_constr_trans_flows_seg
        )
    
    # imported and exported flows: energy system defines the x value
    def rule_constr_trans_flows_sys(m, g, l, q, p, k):
        if (g,l) in m.set_GL_imp:
            return sum(
                m.var_v_glljqk[(g,l,l_star,j,q,k)]
                for l_star in m.set_L[g]
                if l_star not in m.set_L_imp[g]
                for j in m.set_J[(g,l,l_star)]  # only directed arcs
            ) == m.var_trans_flows_glqpk[(g, l, q, p, k)]
        else:
            return sum(
                m.var_v_glljqk[(g,l_star,l,j,q,k)]
                * m.param_eta_glljqk[(g,l_star,l,j,q,k)]
                for l_star in m.set_L[g]
                if l_star not in m.set_L_exp[g]
                for j in m.set_J[(g,l_star,l)]  # only directed arcs
            ) == m.var_trans_flows_glqpk[(g, l, q, p, k)]
    model.constr_trans_flows_sys = pyo.Constraint(
        model.set_GLQPK,
        rule=rule_constr_trans_flows_sys
        )

    # *************************************************************************
    # *************************************************************************
    
    # non-convex price functions
    
    # variables to indicate if the segments are active
    model.var_segment_is_full = pyo.Var(
        model.set_GLQPKS, 
        within=pyo.Binary
    )
    
    # if next level is active, bin. var. for current level must be one
    # if bin. var. for current level is one, segment must be used fully
    # if previous level is not full, bin. var for current level must be zero
    # if bin. var. for current level is zero, 
    def rule_constr_nonconvex_p1(m, g, l, q, p, k, s):
        if s == len(m.set_S[(g, l, q, p, k)])-1:
            # last segment, skip
            return pyo.Constraint.Skip
        return (
            m.var_segment_usage_glqpks[(g,l,q,p,k,s+1)] <= 
            m.var_segment_is_full[(g,l,q,p,k,s)]
            )
    model.constr_nonconvex_p1 = pyo.Constraint(
        model.set_GLQPKS, rule=rule_constr_nonconvex_p1
        )
    
    def rule_constr_nonconvex_p2(m, g, l, q, p, k, s):
        if s == len(m.set_S[(g, l, q, p, k)])-1:
            # last segment, skip
            return pyo.Constraint.Skip
        return (
            m.var_segment_usage_glqpks[(g,l,q,p,k,s)] >= 
            m.var_segment_is_full[(g,l,q,p,k,s)]
            )
    model.constr_nonconvex_p2 = pyo.Constraint(
        model.set_GLQPKS, rule=rule_constr_nonconvex_p2
        )
    
    def rule_constr_nonconvex_p3(m, g, l, q, p, k, s):
        if s == len(m.set_S[(g, l, q, p, k)])-1:
            # last segment, skip
            return pyo.Constraint.Skip
        return (
            m.var_segment_usage_glqpks[(g,l,q,p,k,s)] >= 
            m.var_segment_is_full[(g,l,q,p,k,s+1)]
            )
    model.constr_nonconvex_p3 = pyo.Constraint(
        model.set_GLQPKS, rule=rule_constr_nonconvex_p3
        )
    
    def rule_constr_nonconvex_p4(m, g, l, q, p, k, s):
        if s == len(m.set_S[(g, l, q, p, k)])-1:
            # last segment, skip
            return pyo.Constraint.Skip
        return (
            m.var_segment_usage_glqpks[(g,l,q,p,k,s+1)] <= 
            m.var_segment_is_full[(g,l,q,p,k,s+1)]
            )
    model.constr_nonconvex_p4 = pyo.Constraint(
        model.set_GLQPKS, rule=rule_constr_nonconvex_p4
        )

# *****************************************************************************
# *****************************************************************************