# imports

import pyomo.environ as pyo
    
# *****************************************************************************
# *****************************************************************************

def price_lambda_block(
    model: pyo.AbstractModel,
    enable_default_values: bool = True,
    enable_validation: bool = True,
    enable_initialisation: bool = True
    ):
    
    # auxiliary set for pyomo
    model.set_GLQPK = model.set_GL_exp_imp*model.set_QPK
    
    # create a block for a transshipment node during a given time interval
    def rule_block_prices(b, g, l, q, p, k):
        
        # *********************************************************************
        # *********************************************************************
        
        # sets
        
        # set of price segments
        b.set_S = pyo.Set()
    
        # *********************************************************************
        # *********************************************************************
        
        # parameters

        # resource prices
        b.param_p_s = pyo.Param(b.set_S, within=pyo.NonNegativeReals)
        
        # price function convexity
        b.param_price_function_is_convex = pyo.Param(within=pyo.Boolean)

        # maximum resource volumes for each prices
        b.param_v_max_s = pyo.Param(b.set_S, within=pyo.NonNegativeReals)
    
        # *********************************************************************
        # *********************************************************************
    
        # variables
    
        # *********************************************************************
        # *********************************************************************
        
        # set of points
        b.set_O = pyo.Set(initialize=[i for i in range(len(b.set_S)+1)])
        
        # set of x points
        def init_param_x_o(b, o):
            return 0 if o == 0 else sum(b.param_v_max_s[i] for i in range(o))
        b.param_x_o = pyo.Param(
            b.set_O, 
            within=pyo.NonNegativeReals,
            initialize=init_param_x_o
            )
        
        # set of y points
        def init_param_y_o(b, o):
            return 0 if o == 0 else (b.param_x_o[o]-b.param_x_o[o-1])*b.param_p_s[o-1]+b.param_y_o[o-1]
        # sum(b.param_p_s[i] for i in range(p+1))
        b.param_y_o = pyo.Param(
            b.set_O, 
            within=pyo.NonNegativeReals,
            initialize=init_param_y_o
            )
    
        # interpoint weights
        b.var_weights_o = pyo.Var(b.set_O, within=pyo.UnitInterval)
        
        # TODO: consider replacing None in the bounds with inf in the parameter
        
        # transshipment flows        
        def bounds_var_trans_flows(b):
            try:
                return (0, sum(b.param_v_max_s[s] for s in b.set_S))
            except Exception:
                return (0, None)
        b.var_trans_flows = pyo.Var(bounds=bounds_var_trans_flows)
        
        # *********************************************************************
        # *********************************************************************
        
        def rule_constr_stick_to_line(b):
            return sum(b.var_weights_o[p] for p in b.set_O) == 1
        b.constr_stick_to_line = pyo.Constraint(rule=rule_constr_stick_to_line)
        
        # *********************************************************************
        # *********************************************************************
    
        # import flow costs and export flow revenues (y equation)
        def rule_constr_y_equation(b):
            if (g,l) in b.parent_block().set_GL_imp:
                return (
                    sum(b.var_weights_o[p]*b.param_y_o[p] for p in b.set_O)
                    == b.parent_block().var_ifc_glqpk[(g,l,q,p,k)]
                )
            else:
                return (
                    sum(b.var_weights_o[p]*b.param_y_o[p] for p in b.set_O)
                    == b.parent_block().var_efr_glqpk[(g,l,q,p,k)]
                )
        b.constr_y_equation = pyo.Constraint(rule=rule_constr_y_equation)
    
        # imported and exported flows (x equation)
        def rule_constr_x_equation(b):
            if (g,l) in b.parent_block().set_GL_imp:
                return (
                    sum(b.var_weights_o[p]*b.param_x_o[p] for p in b.set_O)
                    == b.var_trans_flows
                )
            else:
                return (
                    sum(b.var_weights_o[p]*b.param_x_o[p] for p in b.set_O)
                    == b.var_trans_flows
                )
        b.constr_x_equation = pyo.Constraint(rule=rule_constr_x_equation)
        
        # imported and exported flows (system equation)
        def rule_constr_sys_equation(b):
            if (g,l) in b.parent_block().set_GL_imp:
                return sum(
                    b.parent_block().var_v_glljqk[(g,l,l_star,j,q,k)]
                    for l_star in b.parent_block().set_L[g]
                    if l_star not in b.parent_block().set_L_imp[g]
                    for j in b.parent_block().set_J[(g,l,l_star)]  # only directed arcs
                ) == b.var_trans_flows
            else:
                return sum(
                    b.parent_block().var_v_glljqk[(g,l_star,l,j,q,k)]
                    * b.parent_block().param_eta_glljqk[(g,l_star,l,j,q,k)]
                    for l_star in b.parent_block().set_L[g]
                    if l_star not in b.parent_block().set_L_exp[g]
                    for j in b.parent_block().set_J[(g,l_star,l)]  # only directed arcs
                ) == b.var_trans_flows
        b.constr_sys_equation = pyo.Constraint(rule=rule_constr_sys_equation)
    
        # *********************************************************************
        # *********************************************************************
        
        # non-convex price functions
        if not b.param_price_function_is_convex:
            
            # declare SOS2
            def rule_constr_sos2_weights(b):
                return [b.var_weights_o[p] for p in b.set_O]            
            b.constr_sos2_weights = pyo.SOSConstraint(rule=rule_constr_sos2_weights, sos=2)
        
        # *********************************************************************
        # ********************************************************************* 
        
    model.block_prices = pyo.Block(model.set_GLQPK, rule=rule_block_prices)

    # *************************************************************************
    # *************************************************************************

# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************

def price_lambda_no_block(
    model: pyo.AbstractModel,
    # convex_price_function: bool = True,
    enable_default_values: bool = True,
    enable_validation: bool = True,
    enable_initialisation: bool = True
    ):
    
    # auxiliary set for pyomo
    model.set_GLQPK = model.set_GL_exp_imp*model.set_QPK

    # set of price segments
    model.set_S = pyo.Set(model.set_GLQPK)

    # set of GLQKS tuples
    def init_set_GLQPKS(m):
        return (
            (g, l, q, p, k, s)
            # for (g,l) in m.set_GL_exp_imp
            # for (q,k) in m.set_QK
            for (g, l, q, p, k) in m.set_S
            for s in m.set_S[(g, l, q, p, k)]
        )

    model.set_GLQPKS = pyo.Set(
        dimen=6, initialize=(init_set_GLQPKS if enable_initialisation else None)
    )

    # *************************************************************************
    # *************************************************************************

    # parameters

    # resource prices
    model.param_p_glqpks = pyo.Param(model.set_GLQPKS, within=pyo.NonNegativeReals)
    
    # price function convexity
    model.param_price_function_is_convex = pyo.Param(
        model.set_GLQPK, 
        within=pyo.Boolean
        )

    # maximum resource volumes for each prices
    model.param_v_max_glqpks = pyo.Param(
        model.set_GLQPKS, 
        within=pyo.NonNegativeReals,
        # default=math.inf
        )
    
    # *************************************************************************
    # *************************************************************************

    # variables

    # *************************************************************************
    # *************************************************************************
    
    # set of points
    def init_set_O(m, g, l, q, p, k):
        return [i for i in range(len(m.set_S[(g,l,q,p,k)])+1)]
        # return 0 if o == 0 else sum(m.param_v_max_glqpks[i] for i in range(o))
    model.set_O = pyo.Set(
        model.set_GLQPK,
        initialize=(init_set_O if enable_initialisation else None)
        )
    
    
    # set of GLQKS tuples
    def init_set_GLQPKO(m):
        return (
            (g, l, q, p, k, o)
            # for (g,l) in m.set_GL_exp_imp
            # for (q,k) in m.set_QK
            for (g, l, q, p, k) in m.set_O
            for o in m.set_O[(g, l, q, p, k)]
        )
    model.set_GLQPKO = pyo.Set(
        dimen=6, initialize=(init_set_GLQPKO if enable_initialisation else None)
    )
    
    # set of x points
    def init_param_x_glqpko(m, g, l, q, p, k, o):
        return 0 if o == 0 else sum(m.param_v_max_glqpks[(g,l,q,p,k,i)] for i in range(o))
    model.param_x_glqpko = pyo.Param(
        model.set_GLQPKO, 
        within=pyo.NonNegativeReals,
        initialize=init_param_x_glqpko
        )
    
    # set of y points
    def init_param_y_glqpko(m, g, l, q, p, k, o):
        return (
            0 
            if o == 0 else 
            (m.param_x_glqpko[(g,l,q,p,k,o)]
             -m.param_x_glqpko[(g,l,q,p,k,o-1)])*
            m.param_p_glqpks[(g,l,q,p,k,o-1)]+
            m.param_y_glqpko[(g,l,q,p,k,o-1)]
            )
    model.param_y_glqpko = pyo.Param(
        model.set_GLQPKO, 
        within=pyo.NonNegativeReals,
        initialize=init_param_y_glqpko
        )

    # interpoint weights
    model.var_weights_glqpko = pyo.Var(
        model.set_GLQPKO, 
        within=pyo.UnitInterval
        )
    
    # TODO: consider replacing None in the bounds with inf in the parameter
    
    # transshipment flows        
    def bounds_var_trans_flows_glqpk(m, g, l, q, p, k):
        try:
            return (0, sum(m.param_v_max_glqpks[(g, l, q, p, k, s)]
                           for s in m.set_S[(g, l, q, p, k)]))
        except Exception:
            return (0, None)
    model.var_trans_flows_glqpk = pyo.Var(
        model.set_GLQPK, 
        bounds=bounds_var_trans_flows_glqpk
        )

    # *************************************************************************
    # *************************************************************************
    
    def rule_constr_stick_to_line(m, g, l, q, p, k):
        return sum(m.var_weights_glqpko[(g,l,q,p,k,o)]
                   for o in m.set_O[(g,l,q,p,k)]) == 1
    model.constr_stick_to_line = pyo.Constraint(
        model.set_GLQPK, 
        rule=rule_constr_stick_to_line
        )
    
    # *************************************************************************
    # *************************************************************************

    # import flow costs and export flow revenues (y equation)
    def rule_constr_y_equation(m, g, l, q, p, k):
        if (g,l) in m.set_GL_imp:
            return (
                sum(m.var_weights_glqpko[(g,l,q,p,k,o)]*
                    m.param_y_glqpko[(g,l,q,p,k,o)] 
                    for o in m.set_O[(g,l,q,p,k)])
                == m.var_ifc_glqpk[(g,l,q,p,k)]
            )
        else:
            return (
                sum(m.var_weights_glqpko[(g,l,q,p,k,o)]*
                    m.param_y_glqpko[(g,l,q,p,k,o)] 
                    for o in m.set_O[(g,l,q,p,k)])
                == m.var_efr_glqpk[(g,l,q,p,k)]
            )
    model.constr_y_equation = pyo.Constraint(
        model.set_GLQPK,
        rule=rule_constr_y_equation
        )

    # imported and exported flows (x equation)
    def rule_constr_x_equation(m, g, l, q, p, k):
        if (g,l) in m.set_GL_imp:
            return (
                sum(m.var_weights_glqpko[(g,l,q,p,k,o)]*m.param_x_glqpko[(g,l,q,p,k,o)] for o in m.set_O[(g,l,q,p,k)])
                == m.var_trans_flows_glqpk[(g,l,q,p,k)]
            )
        else:
            return (
                sum(m.var_weights_glqpko[(g,l,q,p,k,o)]*m.param_x_glqpko[(g,l,q,p,k,o)] for o in m.set_O[(g,l,q,p,k)])
                == m.var_trans_flows_glqpk[(g,l,q,p,k)]
            )
    model.constr_x_equation = pyo.Constraint(
        model.set_GLQPK,
        rule=rule_constr_x_equation
        )
    
    # imported and exported flows (system equation)
    def rule_constr_sys_equation(m, g, l, q, p, k):
        if (g,l) in m.set_GL_imp:
            return sum(
                m.var_v_glljqk[(g,l,l_star,j,q,k)]
                for l_star in m.set_L[g]
                if l_star not in m.set_L_imp[g]
                for j in m.set_J[(g,l,l_star)]  # only directed arcs
            ) == m.var_trans_flows_glqpk[(g,l,q,p,k)]
        else:
            return sum(
                m.var_v_glljqk[(g,l_star,l,j,q,k)]
                * m.param_eta_glljqk[(g,l_star,l,j,q,k)]
                for l_star in m.set_L[g]
                if l_star not in m.set_L_exp[g]
                for j in m.set_J[(g,l_star,l)]  # only directed arcs
            ) == m.var_trans_flows_glqpk[(g,l,q,p,k)]
    model.constr_sys_equation = pyo.Constraint(
        model.set_GLQPK,
        rule=rule_constr_sys_equation
        )

    # *************************************************************************
    # *************************************************************************
    
    # non-convex price functions
        
    # declare SOS2
    def rule_constr_sos2_weights(m, g, l, q, p, k): 
        if m.param_price_function_is_convex[(g,l,q,p,k)]:
            return pyo.SOSConstraint.Skip
        return [m.var_weights_glqpko[(g,l,q,p,k,o)] for o in m.set_O[(g,l,q,p,k)]]            
    model.constr_sos2_weights = pyo.SOSConstraint(
        model.set_GLQPK,
        rule=rule_constr_sos2_weights, 
        sos=2) 

    # *************************************************************************
    # *************************************************************************        

# *****************************************************************************
# *****************************************************************************