# imports

import pyomo.environ as pyo

from math import isfinite, inf

# *****************************************************************************
# *****************************************************************************


def add_network_restrictions(
        model: pyo.AbstractModel,
        enable_default_values: bool = True,
        enable_validation: bool = True,
        enable_initialisation: bool = True,
):
    
    # *************************************************************************
    # *************************************************************************
    
    model.set_L_max_in_g = pyo.Set(
        model.set_G, within=model.set_L
    )  # should inherently exclude import nodes

    model.set_L_max_out_g = pyo.Set(
        model.set_G, within=model.set_L
    )  # should inherently exclude export nodes

    # maximum number of arcs per node pair

    model.param_max_number_parallel_arcs = pyo.Param(
        model.set_GLL,
        # within=pyo.PositiveIntegers,
        within=pyo.PositiveReals,
        default=inf,
    )

    def init_set_GLL_arc_max(m):
        return (
            (g, l1, l2)
            for (g, l1, l2) in m.param_max_number_parallel_arcs
            if isfinite(m.param_max_number_parallel_arcs[(g, l1, l2)])
        )

    model.set_GLL_arc_max = pyo.Set(
        dimen=3, within=model.set_GLL, initialize=init_set_GLL_arc_max
    )

    # *************************************************************************
    # *************************************************************************

    # limit number of directed arcs per direction

    def rule_constr_limited_parallel_arcs_per_direction(m, g, l1, l2):
        # cases:
        # 1) the number of options is lower than or equal to the limit (skip)
        # 2) the number of preexisting and new mandatory arcs exceeds
        # the limit (infeasible: pyo.Constraint.Infeasible)
        # 3) all other cases (constraint)

        # number of preexisting arcs going from l1 to l2

        number_arcs_pre_nom = (
            len(m.set_J_pre[(g, l1, l2)]) if (g, l1, l2) in m.set_J_pre else 0
        )

        number_arcs_pre_rev = (
            sum(1 for j in m.set_J_pre[(g, l2, l1)] if j in m.set_J_und[(g, l2, l1)])
            if (g, l2, l1) in m.set_J_pre
            else 0
        )

        # number of mandatory arcs going from l1 to l2

        number_arcs_mdt_nom = (
            len(m.set_J_mdt[(g, l1, l2)]) if (g, l1, l2) in m.set_J_mdt else 0
        )

        number_arcs_mdt_rev = (
            sum(1 for j in m.set_J_mdt[(g, l2, l1)] if j in m.set_J_und[(g, l2, l1)])
            if (g, l2, l1) in m.set_J_mdt
            else 0
        )

        # number of optional arcs going from l1 to l2

        number_arcs_opt_nom = (
            sum(
                1
                for j in m.set_J[(g, l1, l2)]
                if j not in m.set_J_pre[(g, l1, l2)]
                if j not in m.set_J_mdt[(g, l1, l2)]
            )
            if (g, l1, l2) in m.set_J
            else 0
        )

        number_arcs_opt_rev = (
            sum(
                1
                for j in m.set_J[(g, l2, l1)]
                if j not in m.set_J_pre[(g, l2, l1)]
                if j not in m.set_J_mdt[(g, l2, l1)]
                if j in m.set_J_und[(g, l2, l1)]
            )
            if (g, l2, l1) in m.set_J
            else 0
        )

        # build the constraints

        if (
            number_arcs_mdt_nom
            + number_arcs_mdt_rev
            + number_arcs_pre_nom
            + number_arcs_pre_rev
            > m.param_max_number_parallel_arcs[(g, l1, l2)]
        ):
            # the number of unavoidable arcs already exceeds the limit
            return pyo.Constraint.Infeasible

        elif (
            number_arcs_opt_nom
            + number_arcs_opt_rev
            + number_arcs_mdt_nom
            + number_arcs_mdt_rev
            + number_arcs_pre_nom
            + number_arcs_pre_rev
            > m.param_max_number_parallel_arcs[(g, l1, l2)]
        ):
            # the number of potential arcs exceeds the limit: cannot be skipped
            return (
                # preexisting arcs
                number_arcs_pre_nom + number_arcs_pre_rev +
                # mandatory arcs
                number_arcs_mdt_nom + number_arcs_mdt_rev +
                # arcs within an (optional) group that uses interfaces
                sum(
                    (
                        sum(
                            1
                            for j in m.set_J_col[(g, l1, l2)]
                            if (g, l1, l2, j) in m.set_GLLJ_col_t[t]
                        )
                        if (g, l1, l2) in m.set_J_col
                        else 0
                        + sum(
                            1
                            for j in m.set_J_col[(g, l2, l1)]
                            if j in m.set_J_und[(g, l2, l1)]
                            if (g, l2, l1, j) in m.set_GLLJ_col_t[t]
                        )
                        if ((g, l2, l1) in m.set_J_col and (g, l2, l1) in m.set_J_und)
                        else 0
                    )
                    * m.var_xi_arc_inv_t[t]
                    for t in m.set_T_int
                )
                +
                # arcs within an (optional) group that does not use interfaces
                sum(
                    (
                        sum(
                            1
                            for j in m.set_J_col[(g, l1, l2)]
                            if (g, l1, l2, j) in m.set_GLLJ_col_t[t]
                        )
                        if (g, l1, l2) in m.set_J_col
                        else 0
                        + sum(
                            1
                            for j in m.set_J_col[(g, l2, l1)]
                            if j in m.set_J_und[(g, l2, l1)]
                            if (g, l2, l1, j) in m.set_GLLJ_col_t[t]
                        )
                        if ((g, l2, l1) in m.set_J_col and (g, l2, l1) in m.set_J_und)
                        else 0
                    )
                    * sum(m.var_delta_arc_inv_th[(t, h)] for h in m.set_H_t[t])
                    for t in m.set_T  # new
                    if t not in m.set_T_mdt  # optional
                    if t not in m.set_T_int  # not interfaced
                )
                +
                # optional individual arcs using interfaces, nominal direction
                sum(
                    m.var_xi_arc_inv_gllj[(g, l1, l2, j)]
                    for j in m.set_J_int[(g, l1, l2)]  # interfaced
                    if j not in m.set_J_col[(g, l1, l2)]  # individual
                )
                if (g, l1, l2) in m.set_J_int
                else 0 +
                # optional individual arcs using interfaces, reverse direction
                sum(
                    m.var_xi_arc_inv_gllj[(g, l2, l1, j)]
                    for j in m.set_J_int[(g, l2, l1)]  # interfaced
                    if j in m.set_J_und[(g, l2, l1)]  # undirected
                    if j not in m.set_J_col[(g, l1, l2)]  # individual
                )
                if ((g, l2, l1) in m.set_J_int and (g, l2, l1) in m.set_J_und)
                else 0 +
                # optional individual arcs not using interfaces, nominal dir.
                sum(
                    sum(
                        m.var_delta_arc_inv_glljh[(g, l1, l2, j, h)]
                        for h in m.set_H_gllj[(g, l1, l2, j)]
                    )
                    for j in m.set_J[(g, l1, l2)]
                    if j not in m.set_J_pre[(g, l1, l2)]  # not preexisting
                    if j not in m.set_J_mdt[(g, l1, l2)]  # not mandatory
                    if j not in m.set_J_int[(g, l1, l2)]  # not interfaced
                    if j not in m.set_J_col[(g, l1, l2)]  # individual
                )
                if (g, l1, l2) in m.set_J
                else 0 +
                # optional individual arcs not using interfaces, reverse dir.
                sum(
                    sum(
                        m.var_delta_arc_inv_glljh[(g, l2, l1, j, h)]
                        for h in m.set_H_gllj[(g, l2, l1, j)]
                    )
                    for j in m.set_J_opt[(g, l2, l1)]
                    if j in m.set_J_und[(g, l2, l1)]
                    if j not in m.set_J_pre[(g, l2, l1)]  # not preexisting
                    if j not in m.set_J_mdt[(g, l2, l1)]  # not mandatory
                    if j not in m.set_J_int[(g, l2, l1)]  # not interfaced
                    if j not in m.set_J_col[(g, l2, l1)]  # individual
                )
                if (g, l2, l1) in m.set_J
                else 0 <= m.param_max_number_parallel_arcs[(g, l1, l2)]
            )

        else:  # the number of options is lower than or equal to the limit: skip
            return pyo.Constraint.Skip

    model.constr_limited_parallel_arcs_per_direction = pyo.Constraint(
        model.set_GLL_arc_max, rule=rule_constr_limited_parallel_arcs_per_direction
    )

    # *************************************************************************
    # *************************************************************************

    # there can only one incoming arc at most, if there are no outgoing arcs

    def rule_constr_max_incoming_directed_arcs(m, g, l):
        # check if the node is not among those subject to a limited number of incoming arcs
        if l not in m.set_L_max_in_g[g]:
            # it is not, skip this constraint
            return pyo.Constraint.Skip

        # max number of directed incoming arcs
        n_max_dir_in = sum(
            sum(
                1
                for j in m.set_J[(g, l_line, l)]
                if j not in m.set_J_und[(g, l_line, l)]
            )  # directed
            for l_line in m.set_L[g] # for every node
            if l_line != l # cannot be the same node
            # if l_line not in m.set_L_imp[g] # why?
            if (g, l_line, l) in m.set_J
        )
        
        # check the maximum number of incoming arcs
        if n_max_dir_in <= 1:
            # there can only be one incoming arc at most: redundant constraint
            return pyo.Constraint.Skip
        else:  # more than one incoming arc is possible
        
            # number of (new) incoming directed arcs in a group
            b_max_in_gl = 0

            # the big m

            M_gl = n_max_dir_in - 1  # has to be positive since n_max_dir_in > 1
            # TODO: put parenthesis to avoid funny results
            temp_constr = (
                sum(
                    # *********************************************************
                    # interfaced groups
                    sum(
                        sum(
                            1
                            for j in m.set_J_col[(g, l_circ, l)]  # part of group
                            if j not in m.set_J_und[(g, l_circ, l)]  # directed
                            if (g, l_circ, l, j) in m.set_GLLJ_col_t[t]
                        )
                        * m.var_xi_arc_inv_t[t]  # in t
                        for t in m.set_T_int
                    )
                    +
                    # *********************************************************
                    # optional non-interfaced groups
                    sum(
                        sum(
                            sum(
                                1
                                for j in m.set_J_col[(g, l_circ, l)]  # part of group
                                if j not in m.set_J_und[(g, l_circ, l)]  # directed
                                if (g, l_circ, l, j) in m.set_GLLJ_col_t[t]
                            )
                            * m.var_delta_arc_inv_th[(t, h)]
                            for h in m.set_H_t[t]
                        )
                        for t in m.set_T
                        if t not in m.set_T_mdt  # optional
                        if t not in m.set_T_int  # not interfaced
                    )
                    +
                    # *********************************************************
                    # interfaced arcs
                    (sum(
                        m.var_xi_arc_inv_gllj[(g, l_circ, l, j_circ)]
                        for j_circ in m.set_J[(g, l_circ, l)]
                        if j_circ not in m.set_J_und[(g, l_circ, l)]  # directed
                        if j_circ in m.set_J_int[(g, l_circ, l)]  # interfaced
                        if j_circ not in m.set_J_col[(g, l_circ, l)]  # individual
                    )
                    if (g, l_circ, l) in m.set_J
                    else 0) +
                    # *********************************************************
                    # optional non-interfaced arcs
                    (sum(
                        sum(
                            m.var_delta_arc_inv_glljh[(g, l_circ, l, j_dot, h_dot)]
                            for h_dot in m.set_H_gllj[(g, l_circ, l, j_dot)]
                        )
                        for j_dot in m.set_J[(g, l_circ, l)]
                        if j_dot not in m.set_J_und[(g, l_circ, l)]  # directed
                        if j_dot not in m.set_J_int[(g, l_circ, l)]  # not interfaced
                        if j_dot not in m.set_J_col[(g, l_circ, l)]  # individual
                        if j_dot not in m.set_J_pre[(g, l_circ, l)]  # new
                        if j_dot not in m.set_J_mdt[(g, l_circ, l)]  # optional
                    )
                    if (g, l_circ, l) in m.set_J
                    else 0) +
                    # *********************************************************
                    # preexisting directed arcs
                    (sum(
                        1
                        for j_pre_dir in m.set_J_pre[(g, l_circ, l)]  # preexisting
                        if j_pre_dir not in m.set_J_und[(g, l_circ, l)]  # directed
                    )
                    if (g, l_circ, l) in m.set_J_pre
                    else 0) +
                    # *********************************************************
                    # mandatory directed arcs
                    (sum(
                        1
                        for j_mdt_dir in m.set_J_mdt[(g, l_circ, l)]
                        if j_mdt_dir not in m.set_J_und[(g, l_circ, l)]  # directed
                    )
                    if (g, l_circ, l) in m.set_J_mdt
                    else 0)
                    # *********************************************************
                    for l_circ in m.set_L[g]
                    if l_circ not in m.set_L_exp[g]
                    if l_circ != l
                )
                <= 1  # +
                # M_gl*sum(
                #     # *********************************************************
                #     # outgoing arcs in interfaced groups, nominal direction
                #     sum(sum(1
                #             for j in m.set_J_col[(g,l,l_diamond)]
                #             #if j in m.set_J_int[(g,l,l_diamond)]
                #             if (g,l,l_diamond,j) in m.set_GLLJ_col_t[t]
                #             )*m.var_xi_arc_inv_t[t]
                #         for t in m.set_T_int
                #         ) if (g,l,l_diamond) in m.set_J_col else 0
                #     +
                #     # outgoing arcs in interfaced groups, reverse direction
                #     sum(sum(1
                #             for j in m.set_J_col[(g,l_diamond,l)]
                #             #if j in m.set_J_int[(g,l_diamond,l)]
                #             if j in m.set_J_und[(g,l_diamond,l)]
                #             if (g,l_diamond,l,j) in m.set_GLLJ_col_t[t]
                #             )*m.var_xi_arc_inv_t[t]
                #         for t in m.set_T_int
                #         ) if (g,l_diamond,l) in m.set_J_col else 0
                #     +
                #     # *********************************************************
                #     # TODO: outgoing arcs in non-interfaced optional groups, nominal
                #     sum(sum(1
                #             for j in m.set_J_col[(g,l,l_diamond)]
                #             #if j in m.set_J_int[(g,l,l_diamond)]
                #             if (g,l,l_diamond,j) in m.set_GLLJ_col_t[t]
                #             )*sum(
                #                 m.var_delta_arc_inv_th[(t,h)]
                #                 for h in m.set_H_t[t]
                #                 )
                #         for t in m.set_T
                #         if t not in m.set_T_mdt
                #         if t not in m.set_T_int
                #         ) if (g,l,l_diamond) in m.set_J_col else 0
                #     +
                #     # TODO: outgoing arcs in non-interfaced optional groups, reverse
                #     sum(sum(1
                #             for j in m.set_J_col[(g,l_diamond,l)]
                #             #if j in m.set_J_int[(g,l_diamond,l)]
                #             if j in m.set_J_und[(g,l_diamond,l)]
                #             if (g,l_diamond,l,j) in m.set_GLLJ_col_t[t]
                #             )*sum(
                #                 m.var_delta_arc_inv_th[(t,h)]
                #                 for h in m.set_H_t[t]
                #                 )
                #         for t in m.set_T
                #         if t not in m.set_T_mdt
                #         if t not in m.set_T_int
                #         ) if (g,l_diamond,l) in m.set_J_col else 0
                #     +
                #     # *********************************************************
                #     # interfaced individual outgoing arcs, nominal direction
                #     sum(m.var_xi_arc_inv_gllj[(g,l,l_diamond,j)]
                #         for j in m.set_J_int[(g,l,l_diamond)] # interfaced
                #         if j not in m.set_J_col[(g,l,l_diamond)] # individual
                #         ) if (g,l,l_diamond) in m.set_J_int else 0
                #     +
                #     # *********************************************************
                #     # interfaced individual undirected arcs, reverse direction
                #     sum(m.var_xi_arc_inv_gllj[(g,l,l_diamond,j)]
                #         for j in m.set_J_und[(g,l_diamond,l)] # undirected
                #         if j in m.set_J_int[(g,l_diamond,l)] # interfaced
                #         if j not in m.set_J_col[(g,l_diamond,l)] # individual
                #         ) if (g,l_diamond,l) in m.set_J_und else 0
                #     +
                #     # *********************************************************
                #     # outgoing non-interfaced individual optional arcs
                #     sum(
                #         sum(m.var_delta_arc_inv_glljh[(g,l,l_diamond,j,h)]
                #             for h in m.set_H_gllj[(g,l,l_diamond,j)])
                #         for j in m.set_J[(g,l,l_diamond)]
                #         if j not in m.set_J_col[(g,l,l_diamond)] # individual
                #         if j not in m.set_J_mdt[(g,l,l_diamond)] # optional
                #         if j not in m.set_J_int[(g,l,l_diamond)] # interfaced
                #         ) if (g,l,l_diamond) in m.set_J else 0
                #     +
                #     # *********************************************************
                #     # individual non-interfaced undirected arcs, reverse dir.
                #     sum(
                #         sum(m.var_delta_arc_inv_glljh[(g,l_diamond,l,j,h)]
                #             for h in m.set_H_gllj[(g,l_diamond,l,j)])
                #         for j in m.set_J_und[(g,l_diamond,l)] # undirected
                #         if j not in m.set_J_col[(g,l_diamond,l)] # individual
                #         if j not in m.set_J_mdt[(g,l_diamond,l)] # optional
                #         if j not in m.set_J_int[(g,l_diamond,l)] # interfaced
                #         ) if (g,l_diamond,l) in m.set_J_und else 0
                #     +
                #     # *********************************************************
                #     # preselected outgonig arcs, nominal direction
                #     len(m.set_J_pre[(g,l,l_diamond)]
                #         ) if (g,l,l_diamond) in m.set_J_pre else 0
                #     +
                #     # *********************************************************
                #     # mandatory outgoing arcs, nominal direction
                #     len(m.set_J_mdt[(g,l,l_diamond)]
                #         ) if (g,l,l_diamond) in m.set_J_mdt else 0
                #     +
                #     # *********************************************************
                #     # undirected preselected arcs, reverse direction
                #     sum(1
                #         for j in m.set_J_pre[(g,l_diamond,l)]
                #         if j in m.set_J_und[(g,l_diamond,l)]
                #         ) if (g,l_diamond,l) in m.set_J_pre else 0
                #     +
                #     # *********************************************************
                #     # undirected mandatory arcs, reverse direction
                #     sum(1
                #         for j in m.set_J_mdt[(g,l_diamond,l)]
                #         if j in m.set_J_und[(g,l_diamond,l)]
                #         ) if (g,l_diamond,l) in m.set_J_mdt else 0
                #     # *********************************************************
                #     for l_diamond in m.set_L[g]
                #     if l_diamond not in m.set_L_imp[g]
                #     if l_diamond != l
                #     )
            )
            if type(temp_constr) == bool:
                # trivial outcome
                return pyo.Constraint.Feasible if temp_constr else pyo.Constraint.Infeasible
            else:
                # constraint is relevant
                return temp_constr

    model.constr_max_incoming_directed_arcs = pyo.Constraint(
        model.set_GL, rule=rule_constr_max_incoming_directed_arcs
    )

    # *************************************************************************
    # *************************************************************************

    def rule_constr_max_outgoing_directed_arcs(m, g, l):
        # check if the node is not among those subject to a limited number of outgoing arcs
        if l not in m.set_L_max_out_g[g]:
            # it is not, skip this constraint
            return pyo.Constraint.Skip

        # max number of directed outgoing arcs
        n_max_dir_out = sum(
            sum(
                1
                for j in m.set_J[(g, l, l_line)]
                if j not in m.set_J_und[(g, l, l_line)]
            )  # directed
            for l_line in m.set_L[g]
            if l_line != l
            # if l_line not in m.set_L_exp[g] # cannot be an export: why?
            if (g, l, l_line) in m.set_J
        )
        
        # check the maximum number of incoming arcs
        if n_max_dir_out <= 1:
            # there can only be one outgoing arc at most: redundant constraint
            # TODO: consider this condition when defining the set
            return pyo.Constraint.Skip
        else:  # more than one outgoing arc is possible

            # number of (new) incoming directed arcs in a group
            b_max_out_gl = 0

            # the big m
            M_gl = n_max_dir_out - 1  # has to be positive since n_max_dir_out > 1
            # TODO: put parenthesis to avoid funny results
            temp_constr = (
                sum(
                    # *********************************************************
                    # interfaced groups
                    sum(
                        sum(
                            1
                            for j in m.set_J_col[(g, l, l_circ)]  # part of group
                            if j not in m.set_J_und[(g, l, l_circ)]  # directed
                            if (g, l, l_circ, j) in m.set_GLLJ_col_t[t]
                        )
                        * m.var_xi_arc_inv_t[t]  # in t
                        for t in m.set_T_int
                    )
                    +
                    # *********************************************************
                    # optional non-interfaced groups
                    sum(
                        sum(
                            sum(
                                1
                                for j in m.set_J_col[(g, l, l_circ)]  # part of group
                                if j not in m.set_J_und[(g, l, l_circ)]  # directed
                                if (g, l, l_circ, j) in m.set_GLLJ_col_t[t]
                            )
                            * m.var_delta_arc_inv_th[(t, h)]
                            for h in m.set_H_t[t]
                        )
                        for t in m.set_T
                        if t not in m.set_T_mdt  # optional
                        if t not in m.set_T_int  # not interfaced
                    )
                    +
                    # *********************************************************
                    # interfaced arcs
                    (sum(
                        m.var_xi_arc_inv_gllj[(g, l, l_circ, j_circ)]
                        for j_circ in m.set_J[(g, l, l_circ)]
                        if j_circ not in m.set_J_und[(g, l, l_circ)]  # directed
                        if j_circ in m.set_J_int[(g, l, l_circ)]  # interfaced
                        if j_circ not in m.set_J_col[(g, l, l_circ)]  # individual
                    )
                    if (g, l, l_circ) in m.set_J
                    else 0) +
                    # *********************************************************
                    # optional non-interfaced arcs
                    (sum(
                        sum(
                            m.var_delta_arc_inv_glljh[(g, l, l_circ, j_dot, h_dot)]
                            for h_dot in m.set_H_gllj[(g, l, l_circ, j_dot)]
                        )
                        for j_dot in m.set_J[(g, l, l_circ)]
                        if j_dot not in m.set_J_und[(g, l, l_circ)]  # directed
                        if j_dot not in m.set_J_int[(g, l, l_circ)]  # not interfaced
                        if j_dot not in m.set_J_col[(g, l, l_circ)]  # individual
                        if j_dot not in m.set_J_pre[(g, l, l_circ)]  # new
                        if j_dot not in m.set_J_mdt[(g, l, l_circ)]  # optional
                    )
                    if (g, l, l_circ) in m.set_J
                    else 0) +
                    # *********************************************************
                    # preexisting directed arcs
                    (sum(
                        1
                        for j_pre_dir in m.set_J_pre[(g, l, l_circ)]  # preexisting
                        if j_pre_dir not in m.set_J_und[(g, l, l_circ)]  # directed
                    )
                    if (g, l, l_circ) in m.set_J_pre
                    else 0) +
                    # *********************************************************
                    # mandatory directed arcs
                    (sum(
                        1
                        for j_mdt_dir in m.set_J_mdt[(g, l, l_circ)]
                        if j_mdt_dir not in m.set_J_und[(g, l, l_circ)]  # directed
                    )
                    if (g, l, l_circ) in m.set_J_mdt
                    else 0)
                    # *********************************************************
                    for l_circ in m.set_L[g]
                    if l_circ not in m.set_L_imp[g]
                    if l_circ != l
                )
                <= 1  # + 
                # TODO: what is below has copy&pasted, must be completely revised
                # M_gl*sum(
                #     # *********************************************************
                #     # outgoing arcs in interfaced groups, nominal direction
                #     sum(sum(1
                #             for j in m.set_J_col[(g,l,l_diamond)]
                #             #if j in m.set_J_int[(g,l,l_diamond)]
                #             if (g,l,l_diamond,j) in m.set_GLLJ_col_t[t]
                #             )*m.var_xi_arc_inv_t[t]
                #         for t in m.set_T_int
                #         ) if (g,l,l_diamond) in m.set_J_col else 0
                #     +
                #     # outgoing arcs in interfaced groups, reverse direction
                #     sum(sum(1
                #             for j in m.set_J_col[(g,l_diamond,l)]
                #             #if j in m.set_J_int[(g,l_diamond,l)]
                #             if j in m.set_J_und[(g,l_diamond,l)]
                #             if (g,l_diamond,l,j) in m.set_GLLJ_col_t[t]
                #             )*m.var_xi_arc_inv_t[t]
                #         for t in m.set_T_int
                #         ) if (g,l_diamond,l) in m.set_J_col else 0
                #     +
                #     # *********************************************************
                #     # TODO: outgoing arcs in non-interfaced optional groups, nominal
                #     sum(sum(1
                #             for j in m.set_J_col[(g,l,l_diamond)]
                #             #if j in m.set_J_int[(g,l,l_diamond)]
                #             if (g,l,l_diamond,j) in m.set_GLLJ_col_t[t]
                #             )*sum(
                #                 m.var_delta_arc_inv_th[(t,h)]
                #                 for h in m.set_H_t[t]
                #                 )
                #         for t in m.set_T
                #         if t not in m.set_T_mdt
                #         if t not in m.set_T_int
                #         ) if (g,l,l_diamond) in m.set_J_col else 0
                #     +
                #     # TODO: outgoing arcs in non-interfaced optional groups, reverse
                #     sum(sum(1
                #             for j in m.set_J_col[(g,l_diamond,l)]
                #             #if j in m.set_J_int[(g,l_diamond,l)]
                #             if j in m.set_J_und[(g,l_diamond,l)]
                #             if (g,l_diamond,l,j) in m.set_GLLJ_col_t[t]
                #             )*sum(
                #                 m.var_delta_arc_inv_th[(t,h)]
                #                 for h in m.set_H_t[t]
                #                 )
                #         for t in m.set_T
                #         if t not in m.set_T_mdt
                #         if t not in m.set_T_int
                #         ) if (g,l_diamond,l) in m.set_J_col else 0
                #     +
                #     # *********************************************************
                #     # interfaced individual outgoing arcs, nominal direction
                #     sum(m.var_xi_arc_inv_gllj[(g,l,l_diamond,j)]
                #         for j in m.set_J_int[(g,l,l_diamond)] # interfaced
                #         if j not in m.set_J_col[(g,l,l_diamond)] # individual
                #         ) if (g,l,l_diamond) in m.set_J_int else 0
                #     +
                #     # *********************************************************
                #     # interfaced individual undirected arcs, reverse direction
                #     sum(m.var_xi_arc_inv_gllj[(g,l,l_diamond,j)]
                #         for j in m.set_J_und[(g,l_diamond,l)] # undirected
                #         if j in m.set_J_int[(g,l_diamond,l)] # interfaced
                #         if j not in m.set_J_col[(g,l_diamond,l)] # individual
                #         ) if (g,l_diamond,l) in m.set_J_und else 0
                #     +
                #     # *********************************************************
                #     # outgoing non-interfaced individual optional arcs
                #     sum(
                #         sum(m.var_delta_arc_inv_glljh[(g,l,l_diamond,j,h)]
                #             for h in m.set_H_gllj[(g,l,l_diamond,j)])
                #         for j in m.set_J[(g,l,l_diamond)]
                #         if j not in m.set_J_col[(g,l,l_diamond)] # individual
                #         if j not in m.set_J_mdt[(g,l,l_diamond)] # optional
                #         if j not in m.set_J_int[(g,l,l_diamond)] # interfaced
                #         ) if (g,l,l_diamond) in m.set_J else 0
                #     +
                #     # *********************************************************
                #     # individual non-interfaced undirected arcs, reverse dir.
                #     sum(
                #         sum(m.var_delta_arc_inv_glljh[(g,l_diamond,l,j,h)]
                #             for h in m.set_H_gllj[(g,l_diamond,l,j)])
                #         for j in m.set_J_und[(g,l_diamond,l)] # undirected
                #         if j not in m.set_J_col[(g,l_diamond,l)] # individual
                #         if j not in m.set_J_mdt[(g,l_diamond,l)] # optional
                #         if j not in m.set_J_int[(g,l_diamond,l)] # interfaced
                #         ) if (g,l_diamond,l) in m.set_J_und else 0
                #     +
                #     # *********************************************************
                #     # preselected outgonig arcs, nominal direction
                #     len(m.set_J_pre[(g,l,l_diamond)]
                #         ) if (g,l,l_diamond) in m.set_J_pre else 0
                #     +
                #     # *********************************************************
                #     # mandatory outgoing arcs, nominal direction
                #     len(m.set_J_mdt[(g,l,l_diamond)]
                #         ) if (g,l,l_diamond) in m.set_J_mdt else 0
                #     +
                #     # *********************************************************
                #     # undirected preselected arcs, reverse direction
                #     sum(1
                #         for j in m.set_J_pre[(g,l_diamond,l)]
                #         if j in m.set_J_und[(g,l_diamond,l)]
                #         ) if (g,l_diamond,l) in m.set_J_pre else 0
                #     +
                #     # *********************************************************
                #     # undirected mandatory arcs, reverse direction
                #     sum(1
                #         for j in m.set_J_mdt[(g,l_diamond,l)]
                #         if j in m.set_J_und[(g,l_diamond,l)]
                #         ) if (g,l_diamond,l) in m.set_J_mdt else 0
                #     # *********************************************************
                #     for l_diamond in m.set_L[g]
                #     if l_diamond not in m.set_L_imp[g]
                #     if l_diamond != l
                #     )
            )
            if type(temp_constr) == bool:
                # trivial outcome
                return pyo.Constraint.Feasible if temp_constr else pyo.Constraint.Infeasible
            else:
                # constraint is relevant
                return temp_constr

    model.constr_max_outgoing_directed_arcs = pyo.Constraint(
        model.set_GL, rule=rule_constr_max_outgoing_directed_arcs
    )

    # *************************************************************************
    # *************************************************************************

    return model

    # *************************************************************************
    # *************************************************************************

# *****************************************************************************
# *****************************************************************************
