# imports

import pyomo.environ as pyo

# *****************************************************************************
# *****************************************************************************

# description: variation of the delta formulation

# *****************************************************************************
# *****************************************************************************

def price_other_block(
    model: pyo.AbstractModel,
    enable_default_values: bool = True,
    enable_validation: bool = True,
    enable_initialisation: bool = True
    ):
    
    # auxiliary set for pyomo
    model.set_GLQPK = model.set_GL_exp_imp*model.set_QPK
    
    # create a block for a transshipment node during a given time interval
    def rule_block_prices(b, g, l, q, p, k):
        
        # *********************************************************************
        # *********************************************************************
        
        # sets
        
        # set of price segments
        b.set_S = pyo.Set()
        
        # TODO: introduce a set of price segments for non-convex tariffs
        # def init_set_S_nonconvex(b, s):
        #     return (s for s in b.set_S if b.param_price_function_is_convex)
        # b.set_S_nonconvex = pyo.Set(within=b.set_S, initialize=init_set_S_nonconvex)
    
        # *********************************************************************
        # *********************************************************************
        
        # parameters

        # resource prices
        b.param_p_s = pyo.Param(b.set_S, within=pyo.NonNegativeReals)
        
        # price function convexity
        b.param_price_function_is_convex = pyo.Param(within=pyo.Boolean)

        # maximum resource volumes for each prices
        b.param_v_max_s = pyo.Param(b.set_S, within=pyo.NonNegativeReals)
    
        # *********************************************************************
        # *********************************************************************
    
        # variables
    
        # *********************************************************************
        # *********************************************************************
        
        # TODO: consider replacing None in the bounds with inf in the parameter
    
        # import and export flows
        def bounds_var_trans_flows_s(b, s):
            if s in b.param_v_max_s:
                # predefined finite capacity
                return (0, b.param_v_max_s[s])
            else:
                # infinite capacity
                return (0, None)
        b.var_trans_flows_s = pyo.Var(
            b.set_S, 
            within=pyo.NonNegativeReals, 
            bounds=bounds_var_trans_flows_s
        )
        
        # *********************************************************************
        # *********************************************************************
    
        # import flow costs and export flow revenues
        def rule_constr_trans_monetary_flows(b):
            if (g,l) in b.parent_block().set_GL_imp:
                return (
                    sum(b.var_trans_flows_s[s]*b.param_p_s[s]
                        for s in b.set_S)
                    == b.parent_block().var_ifc_glqpk[(g,l,q,p,k)]
                )
            else:
                return (
                    sum(b.var_trans_flows_s[s]*b.param_p_s[s]
                        for s in b.set_S)
                    == b.parent_block().var_efr_glqpk[(g,l,q,p,k)]
                )
        b.constr_trans_monetary_flows = pyo.Constraint(
            rule=rule_constr_trans_monetary_flows
        )
        
        # imported and exported flows
        def rule_constr_trans_flows(b):
            if (g,l) in b.parent_block().set_GL_imp:
                return sum(
                    b.parent_block().var_v_glljqk[(g,l,l_star,j,q,k)]
                    for l_star in b.parent_block().set_L[g]
                    if l_star not in b.parent_block().set_L_imp[g]
                    for j in b.parent_block().set_J[(g,l,l_star)]  # only directed arcs
                ) == sum(b.var_trans_flows_s[s] for s in b.set_S)
            else:
                return sum(
                    b.parent_block().var_v_glljqk[(g,l_star,l,j,q,k)]
                    * b.parent_block().param_eta_glljqk[(g,l_star,l,j,q,k)]
                    for l_star in b.parent_block().set_L[g]
                    if l_star not in b.parent_block().set_L_exp[g]
                    for j in b.parent_block().set_J[(g,l_star,l)]  # only directed arcs
                ) == sum(b.var_trans_flows_s[s] for s in b.set_S)
        b.constr_trans_flows = pyo.Constraint(rule=rule_constr_trans_flows)
    
        # *********************************************************************
        # *********************************************************************
        
        # non-convex price functions
        if not b.param_price_function_is_convex:
            
            # delta variables
            b.var_active_segment_s = pyo.Var(b.set_S, within=pyo.Binary)
            
            # segments must be empty if the respective delta variable is zero
            def rule_constr_empty_segment_if_delta_zero(b, s):
                if len(b.set_S) == 1 or b.param_price_function_is_convex:
                    # single segment, skip
                    # convex, skip
                    return pyo.Constraint.Skip
                return (
                    b.var_trans_flows_s[s] <= 
                    b.param_v_max_s[s]*b.var_active_segment_s[s]
                    )
            b.constr_empty_segment_if_delta_zero = pyo.Constraint(
                b.set_S, rule=rule_constr_empty_segment_if_delta_zero
                )
            
            # if delta var is one, previous ones must be one too
            # if delta var is zero, the next ones must also be zero
            def rule_constr_delta_summing_logic(b, s):
                if s == len(b.set_S)-1 or b.param_price_function_is_convex:
                    # last segment, skip
                    # convex, skip
                    return pyo.Constraint.Skip
                return (
                    b.var_active_segment_s[s] >= 
                    b.var_active_segment_s[s+1]
                    )
            b.constr_delta_summing_logic = pyo.Constraint(
                b.set_S, rule=rule_constr_delta_summing_logic
                )
            
            # if a segment is not completely used, the next ones must remain empty
            def rule_constr_fill_up_segment_before_next(b, s):
                if s == len(b.set_S)-1 or b.param_price_function_is_convex:
                    # last segment, skip
                    # convex, skip
                    return pyo.Constraint.Skip
                return (
                    b.var_trans_flows_s[s] >= 
                    b.var_active_segment_s[s+1]*
                    b.param_v_max_s[s]
                    )
            b.constr_fill_up_segment_before_next = pyo.Constraint(
                b.set_S, rule=rule_constr_fill_up_segment_before_next
                )
        
        # *********************************************************************
        # ********************************************************************* 
        
    model.block_prices = pyo.Block(model.set_GLQPK, rule=rule_block_prices)

    # *************************************************************************
    # *************************************************************************
    

# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************

def price_other_no_block(
    model: pyo.AbstractModel,
    # convex_price_function: bool = True,
    enable_default_values: bool = True,
    enable_validation: bool = True,
    enable_initialisation: bool = True
    ):
    
    # auxiliary set for pyomo
    model.set_GLQPK = model.set_GL_exp_imp*model.set_QPK

    # set of price segments
    model.set_S = pyo.Set(model.set_GLQPK)

    # set of GLQKS tuples
    def init_set_GLQPKS(m):
        return (
            (g, l, q, p, k, s)
            # for (g,l) in m.set_GL_exp_imp
            # for (q,k) in m.set_QK
            for (g, l, q, p, k) in m.set_S
            for s in m.set_S[(g, l, q, p, k)]
        )

    model.set_GLQPKS = pyo.Set(
        dimen=6, initialize=(init_set_GLQPKS if enable_initialisation else None)
    )

    # *************************************************************************
    # *************************************************************************

    # parameters

    # resource prices

    model.param_p_glqpks = pyo.Param(model.set_GLQPKS, within=pyo.NonNegativeReals)
    
    # price function convexity

    model.param_price_function_is_convex = pyo.Param(
        model.set_GLQPK, 
        within=pyo.Boolean
        )

    # maximum resource volumes for each prices

    model.param_v_max_glqpks = pyo.Param(
        model.set_GLQPKS, 
        within=pyo.NonNegativeReals,
        # default=math.inf
        )

    # *************************************************************************
    # *************************************************************************

    # variables

    # *************************************************************************
    # *************************************************************************

    # import and export flows
    def bounds_var_trans_flows_glqpks(m, g, l, q, p, k, s):
        # return (0, m.param_v_max_glqpks[(g, l, q, p, k, s)])
        if (g, l, q, p, k, s) in m.param_v_max_glqpks:
            # predefined finite capacity
            return (0, m.param_v_max_glqpks[(g, l, q, p, k, s)])
        else:
            # infinite capacity
            return (0, None)
    model.var_trans_flows_glqpks = pyo.Var(
        model.set_GLQPKS, within=pyo.NonNegativeReals, bounds=bounds_var_trans_flows_glqpks
    )

    # *************************************************************************
    # *************************************************************************

    # import flow costs and export flow revenues
    def rule_constr_trans_monetary_flows(m, g, l, q, p, k):
        if (g,l) in m.set_GL_imp:
            return (
                sum(
                    m.var_trans_flows_glqpks[(g, l, q, p, k, s)]
                    * m.param_p_glqpks[(g, l, q, p, k, s)]
                    for s in m.set_S[(g, l, q, p, k)]
                )
                == m.var_ifc_glqpk[(g, l, q, p, k)]
            )
        else:
            return (
                sum(
                    m.var_trans_flows_glqpks[(g, l, q, p, k, s)]
                    * m.param_p_glqpks[(g, l, q, p, k, s)]
                    for s in m.set_S[(g, l, q, p, k)]
                )
                == m.var_efr_glqpk[(g, l, q, p, k)]
            )
    model.constr_trans_monetary_flows = pyo.Constraint(
        model.set_GLQPK, rule=rule_constr_trans_monetary_flows
    )

    # imported and exported flows
    def rule_constr_trans_flows(m, g, l, q, p, k):
        if (g,l) in m.set_GL_imp:
            return sum(
                m.var_v_glljqk[(g, l, l_star, j, q, k)]
                for l_star in m.set_L[g]
                if l_star not in m.set_L_imp[g]
                for j in m.set_J[(g, l, l_star)]  # only directed arcs
            ) == sum(m.var_trans_flows_glqpks[(g, l, q, p, k, s)] for s in m.set_S[(g, l, q, p, k)])
        else:
            return sum(
                m.var_v_glljqk[(g, l_star, l, j, q, k)]
                * m.param_eta_glljqk[(g, l_star, l, j, q, k)]
                for l_star in m.set_L[g]
                if l_star not in m.set_L_exp[g]
                for j in m.set_J[(g, l_star, l)]  # only directed arcs
            ) == sum(m.var_trans_flows_glqpks[(g, l, q, p, k, s)] for s in m.set_S[(g, l, q, p, k)])

    model.constr_trans_flows = pyo.Constraint(
        model.set_GLQPK, rule=rule_constr_trans_flows
    )

    # *************************************************************************
    # *************************************************************************
    
    # non-convex price functions
    # TODO: remove these variables from the model if they are not needed
    # delta variables
    model.var_active_segment_glqpks = pyo.Var(
        model.set_GLQPKS, within=pyo.Binary
    )
    
    # segments must be empty if the respective delta variable is zero
    def rule_constr_empty_segment_if_delta_zero(m, g, l, q, p, k, s):
        if len(m.set_S[(g,l,q,p,k)]) == 1 or m.param_price_function_is_convex[(g,l,q,p,k)]:
            # single segment, skip
            # convex, skip
            return pyo.Constraint.Skip
        return (
            m.var_trans_flows_glqpks[(g,l,q,p,k,s)] <= 
            m.param_v_max_glqpks[(g,l,q,p,k,s)]*
            m.var_active_segment_glqpks[(g,l,q,p,k,s)]
            )
    model.constr_empty_segment_if_delta_zero = pyo.Constraint(
        model.set_GLQPKS, rule=rule_constr_empty_segment_if_delta_zero
        )
    
    # if delta var is one, previous ones must be one too
    # if delta var is zero, the next ones must also be zero
    def rule_constr_delta_summing_logic(m, g, l, q, p, k, s):
        if s == len(m.set_S[(g,l,q,p,k)])-1 or m.param_price_function_is_convex[(g,l,q,p,k)]:
            # last segment, skip
            # convex, skip
            return pyo.Constraint.Skip
        return (
            m.var_active_segment_glqpks[(g,l,q,p,k,s)] >= 
            m.var_active_segment_glqpks[(g,l,q,p,k,s+1)]
            )
    model.constr_delta_summing_logic = pyo.Constraint(
        model.set_GLQPKS, rule=rule_constr_delta_summing_logic
        )
    
    # if a segment is not completely used, the next ones must remain empty
    def rule_constr_fill_up_segment_before_next(m, g, l, q, p, k, s):
        if s == len(m.set_S[(g,l,q,p,k)])-1 or m.param_price_function_is_convex[(g,l,q,p,k)]:
            # last segment, skip
            # convex, skip
            return pyo.Constraint.Skip
        return (
            m.var_trans_flows_glqpks[(g,l,q,p,k,s)] >= 
            m.var_active_segment_glqpks[(g,l,q,p,k,s+1)]*
            m.param_v_max_glqpks[(g,l,q,p,k,s)]
            )
        # return (
        #     m.var_if_glqpks[(g,l,q,p,k,s)]/m.param_v_max_glqpks[(g,l,q,p,k,s)] >= 
        #     m.var_active_segment_glqpks[(g,l,q,p,k,s+1)]
        #     )
        # return (
        #     m.param_v_max_glqpks[(g,l,q,p,k,s)]-m.var_if_glqpks[(g,l,q,p,k,s)] <= 
        #     m.param_v_max_glqpks[(g,l,q,p,k,s)]*(1- m.var_active_segment_glqpks[(g,l,q,p,k,s+1)])
        #     )
    model.constr_fill_up_segment_before_next = pyo.Constraint(
        model.set_GLQPKS, rule=rule_constr_fill_up_segment_before_next
        )

# *****************************************************************************
# *****************************************************************************