# standard libraries

# local libraries, external
import numpy as np
from scipy.linalg import expm, inv

# local libraries, internal

# *****************************************************************************
# *****************************************************************************

# TODO: integrate only selected outputs on demand, not all of them

# TODO: ensure it is compatible with stateless systems

class DynamicSystem:
    """A class for dynamic systems described using A, B, C and D matrices."""
    
    # use cases:
    # 1) simulate
    # 2) optimise

    # input options:
    # 1) the A, B, C and D matrices are provided: inputs, states and outputs (general case)
    # 2) the A and B matrices are provided: inputs and states (no outputs)
    # 3) the D matrix is provided: inputs and outputs (no states)

    # formats:
    # 1) the (A, B, C and D) matrices are provided directly as numpy arrays (1 matrix each at most)
    # 2) the (A, B, C and D) matrices are provided within lists, whose length
    # is the same as the number of time intervals, to represent time variation
    # 3) the matrices are provided independently of each other, and may vary
    # in  time or not, independently of the others (not ready yet)

    # note: matrices must be provided, even if they are zero matrices

    # other options:
    # i) time intervals: can be supplied as a numeric type (single interval) or
    # as a list, for multiple intervals (regular or irregular)
    # note: if the time interval is provided as a numeric type, the number of
    # intervals may still be more than one, in accordance with the number of
    # matrices, in which case an internal list of time intervals will be created
    # ii) integrate outputs: it may not be relevant to obtain instantaneous
    # information, in which case the outputs can be integrated (all or none, for now)
    # note: this option requires the D matrix

    # TODO: define whether this class holds information about the number of
    # time intervals or if that information comes from somewhere else

    # *************************************************************************
    # *************************************************************************

    def __init__(
        self,
        time_interval_durations: list or float or int,
        A: list or np.array,
        B: list or np.array,
        C: list or np.array,
        D: list or np.array,
        integrate_outputs: bool = True,
    ):
        # steps:
        # 1) validate the input data
        # 2) compute and/or adjust the data in preparation for discretisation
        # 3) discretise the model

        # *********************************************************************
        # *********************************************************************

        # cases:
        # 1) time invariant system, regular time steps
        # >> one continuous SS model discretised once and reused every step
        # 2) time invariant system, irregular time steps
        # >> one continuous SS model discretised once for each time step
        # 3) time-varying system, regular time steps
        # >> multiple continous SS models, each discretised once
        # 4) time-varying system, irregular time steps
        # >> multiple continous SS models, each discretised once
        # note: case 3 and 4 are effectively the same

        # *********************************************************************
        # *********************************************************************

        # if time_interval_durations is a number, transform it into a list

        if (
            type(time_interval_durations) == int
            or type(time_interval_durations) == float
        ):
            # a single time interval duration has been provided

            if time_interval_durations <= 0:
                # zero or negative time interval duration

                raise ValueError("Time interval durations cannot be zero nor negative.")

            # self.number_time_intervals = len(time_interval_durations)

            self.time_interval_durations = [time_interval_durations]

            # validate the inputs

            (
                self.number_inputs,
                self.number_states,
                self.number_outputs,
                self.A_is_time_varying,
                self.B_is_time_varying,
                self.C_is_time_varying,
                self.D_is_time_varying,
            ) = self.validate(
                time_interval_durations=[time_interval_durations], A=A, B=B, C=C, D=D
            )

        else:  # time_interval_durations is not a number
            self.time_interval_durations = time_interval_durations

            # validate the inputs

            (
                self.number_inputs,
                self.number_states,
                self.number_outputs,
                self.A_is_time_varying,
                self.B_is_time_varying,
                self.C_is_time_varying,
                self.D_is_time_varying,
            ) = self.validate(
                time_interval_durations=time_interval_durations, A=A, B=B, C=C, D=D
            )

        # *********************************************************************

        # determine the number of models

        if self.A_is_time_varying:
            self.number_models = len(A)

        else:
            self.number_models = 1

        if self.B_is_time_varying:
            if len(B) > self.number_models:
                self.number_models = len(B)  # TODO: reach this statement

        if self.C_is_time_varying:
            if len(C) > self.number_models:
                self.number_models = len(C)  # TODO: reach this statement

        if self.D_is_time_varying:
            if len(D) > self.number_models:
                self.number_models = len(D)  # TODO: reach this statement

        # extend the time_interval_durations for convenience

        if len(self.time_interval_durations) == 1 and (
            self.A_is_time_varying
            or self.B_is_time_varying
            or self.C_is_time_varying
            or self.D_is_time_varying
        ):
            # time_interval_durations is a list with a single element and
            # at least one of the matrices is time-varying

            # if self.A_is_time_varying:

            #     number_models = len(A)

            # else:

            #     number_models = 1

            # if self.B_is_time_varying:

            #     if len(B) > number_models:

            #         number_models = len(B)

            # if self.C_is_time_varying:

            #     if len(C) > number_models:

            #         number_models = len(C)

            # if self.D_is_time_varying:

            #     if len(D) > number_models:

            #         number_models = len(D)

            self.time_interval_durations = [  # TODO: reach this statement
                self.time_interval_durations[0] for _ in range(self.number_models)
            ]

        # else:

        #     self.time_interval_durations = list(time_interval_durations)

        # determine the number of models

        # *********************************************************************
        # *********************************************************************

        # get the data into lists

        # matrices

        if type(A) == list or A is None:
            self.A = A

        else:
            self.A = [A]

        if type(B) == list or B is None:
            self.B = B

        else:
            self.B = [B]

        if type(C) == list or C is None:
            self.C = C

        else:
            self.C = [C]

        if type(D) == list or D is None:
            self.D = D

        else:
            self.D = [D]

        # *********************************************************************
        # *********************************************************************

        # discretise the model

        self.discretise(integrate_outputs=integrate_outputs)

        # *********************************************************************
        # *********************************************************************

    # *************************************************************************
    # *************************************************************************

    def validate(
        self,
        time_interval_durations: list or np.array,
        A: list or np.array,
        B: list or np.array,
        C: list or np.array,
        D: list or np.array,
    ) -> tuple:
        # *********************************************************************
        # *********************************************************************

        # time interval duration

        if type(time_interval_durations) == list:
            # a list has been provided

            if len(time_interval_durations) == 0:
                # empty list

                raise ValueError("No time interval durations were provided.")

            else:
                # check each element

                for elem in time_interval_durations:
                    # check if they are numbers

                    if type(elem) == int or type(elem) == float:
                        # elem is an integer or a float, all is fine so far

                        if elem <= 0:
                            # zero or negative time interval duration

                            raise ValueError(
                                "Time interval durations cannot be zero nor "
                                + "negative."
                            )

                    else:  # not an integer nor a float, something is up
                        raise TypeError("Unrecognised time interval duration format.")

        else:  # unknown input
            raise TypeError("Unrecognised time interval duration format.")

        # *********************************************************************
        # *********************************************************************

        # if multiple matrices are provided for a given position, then they
        # must all have the same shape (rows and columns)

        # if multiple matrices are provided for a given position, and more than
        # one time interval is provided, then their sizes must match

        # start by assuming the matrices are invariant

        A_is_time_varying = False

        B_is_time_varying = False

        C_is_time_varying = False

        D_is_time_varying = False

        # A matrix

        if type(A) == list:
            # A is a list

            if isinstance(A[0], np.ndarray):
                A_m, A_n = A[0].shape

            else:
                raise TypeError("Unrecognised A matrix format.")

            for a in A:
                if a.shape != (A_m, A_n):
                    raise ValueError("The A matrices are inconsistent.")

            if type(time_interval_durations) == list:
                if (
                    len(time_interval_durations) != len(A)
                    and len(time_interval_durations) > 1
                    and len(A) > 1
                ):
                    # the number of time intervals and matrices do not match &
                    # they are larger than one: mismatch

                    raise ValueError(
                        "The number of A matrices is inconsistent with the number of time intervals."
                    )

            if len(A) > 1:
                A_is_time_varying = True

        elif isinstance(A, np.ndarray):
            # A is not a list

            A_m, A_n = A.shape

        elif A is None:
            A_m, A_n = 0, 0

        else:
            raise TypeError("Unrecognised A matrix format.")

        # B matrix

        if type(B) == list:
            # B is a list

            if isinstance(B[0], np.ndarray):
                B_m, B_n = B[0].shape

            else:
                raise TypeError("Unrecognised B matrix format.")

            for b in B:
                if b.shape != (B_m, B_n):
                    raise ValueError("The B matrices are inconsistent.")

            if type(time_interval_durations) == list:
                if (
                    len(time_interval_durations) != len(B)
                    and len(time_interval_durations) > 1
                    and len(B) > 1
                ):
                    # the number of time intervals and matrices do not match &
                    # they are larger than one: mismatch

                    raise ValueError(
                        "The number of B matrices is inconsistent with the number of time intervals."
                    )

            if len(B) > 1:
                B_is_time_varying = True

        elif isinstance(B, np.ndarray):
            # B is not a list

            B_m, B_n = B.shape

        elif A is None:
            B_m, B_n = 0, 0

        else:
            raise TypeError("Unrecognised B matrix format.")

        # C matrix

        if type(C) == list:
            # C is a list

            if isinstance(C[0], np.ndarray):
                C_m, C_n = C[0].shape

            else:
                raise TypeError("Unrecognised C matrix format.")

            for c in C:
                if c.shape != (C_m, C_n):
                    raise ValueError("The C matrices are inconsistent.")

            if type(time_interval_durations) == list:
                if (
                    len(time_interval_durations) != len(C)
                    and len(time_interval_durations) > 1
                    and len(C) > 1
                ):
                    # the number of time intervals and matrices do not match &
                    # they are larger than one: mismatch

                    raise ValueError(
                        "The number of C matrices is inconsistent with the number of time intervals."
                    )

            if len(C) > 1:
                C_is_time_varying = True

        elif isinstance(C, np.ndarray):
            # C is not a list

            C_m, C_n = C.shape

        elif C is None:
            C_m, C_n = 0, 0

        else:
            raise TypeError("Unrecognised C matrix format.")

        # D matrix

        if type(D) == list:
            # D is a list

            if isinstance(D[0], np.ndarray):
                D_m, D_n = D[0].shape

            else:
                raise TypeError("Unrecognised D matrix format.")

            for d in D:
                if d.shape != (D_m, D_n):
                    raise ValueError("The D matrices are inconsistent.")

            if type(time_interval_durations) == list:
                if (
                    len(time_interval_durations) != len(D)
                    and len(time_interval_durations) > 1
                    and len(D) > 1
                ):
                    # the number of time intervals and matrices do not match &
                    # they are larger than one: mismatch

                    raise ValueError(
                        "The number of D matrices is inconsistent with the number of time intervals."
                    )

            if len(D) > 1:
                D_is_time_varying = True

        elif isinstance(D, np.ndarray):
            # D is not a list

            D_m, D_n = D.shape

        elif D is None:
            D_m, D_n = 0, 0

        else:
            raise TypeError("Unrecognised D matrix format.")

        # *********************************************************************

        # make sure the matrices have appropriate shapes

        # the A matrix has to be square: A_m == A_n

        if A_m != A_n:
            raise ValueError("The A matrix is not square.")

        # the A and B matrices have to have the same number of rows: A_m == B_m

        #

        if A_m != B_m:
            raise ValueError(
                "The A and B matrices do not have the same number of rows."
            )

        # the C and D matrices have to have the same number of rows: C_m == D_m
        # except if A, B and C have zero rows: no states

        if C_m != D_m and (C_m != 0 and A_m != 0 and B_m != 0):
            raise ValueError(
                "The C and D matrices do not have the same number of rows."
            )

        # the A and C matrices have to have the same number of columns: A_n == C_n

        if A_n != 0 and C_m != 0 and C_n != 0:  # Note: A_m == A_n
            if A_n != C_n:
                raise ValueError(
                    "The A and C matrices do not have the same number of columns."
                )

        # the B and D matrices have to have the same number of columns: B_n == D_n

        if (
            B_m != 0 and B_n != 0 and D_m != 0 and D_n != 0
        ):  # both matrices need to exist
            if B_n != D_n:
                raise ValueError(
                    "The B and D matrices do not have the same number of columns."
                )

        # *********************************************************************

        # number of inputs, states and outputs

        # number_inputs = B_n # B_n if D_n does not exist or D_n

        # number_states = A_m # or A_n or C_n

        # number_outputs = D_m # C_m or D_m (better, since C is optional)

        # *********************************************************************

        return (
            B_n if D_n == 0 else D_n,
            A_m,
            D_m,
            A_is_time_varying,
            B_is_time_varying,
            C_is_time_varying,
            D_is_time_varying,
        )

    # *************************************************************************
    # *************************************************************************

    def discretise(self, integrate_outputs: bool = True):
        # cases:
        # *********************************************************************
        # 1) time invariant system, regular time steps
        # >> one continuous SS model discretised once and reused every step
        # *********************************************************************
        # 2) time invariant system, irregular time steps
        # >> one continuous SS model discretised once for each time step
        # *********************************************************************
        # 3) time-varying system, regular time steps
        # >> multiple continous SS models, each discretised once
        # *********************************************************************
        # 4) time-varying system, irregular time steps
        # >> multiple continous SS models, each discretised once
        # *********************************************************************
        # note: case 3 and 4 are effectively the same
        # *********************************************************************

        # TODO: do not assume that A, B, C, D lists have the same lengths if time-varying

        # TODO: add warning to notify users that A, B, C, D lists need to have the same lengths if time-varying

        # A, B, C and D are matrices that cannot simultaenously be None
        # if C and D are None: okay, but no outputs
        # if A, B and C are None: okay, but that means no states (only input-to-output functions)
        # if A, B, C and D are None: error

        # A matrix

        if self.A is None:
            self.A_line_k = None

        elif len(self.A) == len(self.time_interval_durations):
            # 1) time invariant system, regular time steps
            # >> only if there is only one time interval
            # 3) time-varying system, regular time steps
            # 4) time-varying system, irregular time steps

            self.A_line_k = [
                expm(dt * A_matrix)
                for (A_matrix, dt) in zip(self.A, self.time_interval_durations)
            ]

        elif len(self.A) == 1:
            # there is only one state space model and more time intervals

            # 1) time invariant system, regular time steps
            # >> one model and multiple time steps
            # 2) time invariant system, irregular time steps

            self.A_line_k = [
                expm(dt * self.A[0]) for dt in self.time_interval_durations
            ]

        else:  # the number of time steps and models do not match
            # 3) time-varying system, regular time steps
            # >> multiple models and one time step duration (the case with mul-
            # tiple time steps with the same duration is tackled above already)

            self.A_line_k = [
                expm(self.time_interval_durations[0] * A_matrix) for A_matrix in self.A
            ]

        # *********************************************************************

        # B matrix

        if self.B is None or self.A is None:
            self.B_line_k = None

        elif len(self.B) == len(self.time_interval_durations):
            # 1) time invariant system, regular time steps
            # >> only if there is only one time interval
            # 3) time-varying system, regular time steps
            # 4) time-varying system, irregular time steps

            self.B_line_k = [
                np.matmul(
                    np.matmul(
                        expm(dt * self.A[i if self.A_is_time_varying else 0])
                        - np.eye(self.number_states),
                        inv(self.A[i if self.A_is_time_varying else 0]),
                    ),
                    self.B[i if self.B_is_time_varying else 0],
                )
                for i, dt in enumerate(self.time_interval_durations)
            ]

        elif len(self.B) == 1:
            # there is only one state space model and more time intervals

            # 1) time invariant system, regular time steps
            # >> one model and multiple time steps
            # 2) time invariant system, irregular time steps

            self.B_line_k = [
                np.matmul(
                    np.matmul(
                        expm(dt * self.A[i if self.A_is_time_varying else 0])
                        - np.eye(self.number_states),
                        inv(self.A[i if self.A_is_time_varying else 0]),
                    ),
                    self.B[0],
                )
                for i, dt in enumerate(self.time_interval_durations)
            ]

        else:  # the number of time steps and models do not match
            # 3) time-varying system, regular time steps
            # >> multiple models and one time step duration (the case with mul-
            # tiple time steps with the same duration is tackled above already)

            self.B_line_k = [  # TODO: reach this statement
                np.matmul(
                    np.matmul(
                        expm(self.time_interval_durations[0] * A_matrix)
                        - np.eye(self.number_states),
                        inv(A_matrix),
                    ),
                    B_matrix,
                )
                for (A_matrix, B_matrix) in zip(self.A, self.B)
            ]

        # *********************************************************************

        # C matrix

        if self.C is None or self.A is None or self.B is None or self.D is None:
            self.C_line_k = None

        elif len(self.C) == len(self.time_interval_durations):
            # 1) time invariant system, regular time steps
            # >> only if there is only one time interval
            # 3) time-varying system, regular time steps
            # 4) time-varying system, irregular time steps

            if integrate_outputs:
                # integrate the outputs

                self.C_line_k = [
                    np.dot(
                        self.C[i if self.C_is_time_varying else 0],
                        np.dot(
                            expm(dt * self.A[i if self.A_is_time_varying else 0])
                            - np.eye(self.number_states),
                            inv(self.A[i if self.A_is_time_varying else 0]),
                        ),
                    )
                    for i, dt in enumerate(self.time_interval_durations)
                ]

            else:
                # do not integrate the outputs

                self.C_line_k = [C_matrix for C_matrix in self.C]

        elif len(self.C) == 1:
            # there is only one state space model and more time intervals

            # 1) time invariant system, regular time steps
            # >> one model and multiple time steps

            if integrate_outputs:
                # integrate the outputs

                self.C_line_k = [
                    np.dot(
                        self.C[i if self.C_is_time_varying else 0],
                        np.dot(
                            expm(dt * self.A[i if self.A_is_time_varying else 0])
                            - np.eye(self.number_states),
                            inv(self.A[i if self.A_is_time_varying else 0]),
                        ),
                    )
                    for i, dt in enumerate(self.time_interval_durations)
                ]

            else:
                # do not integrate the outputs

                self.C_line_k = [
                    self.C[i if self.C_is_time_varying else 0]
                    for i, dt in enumerate(self.time_interval_durations)
                ]

        else:  # the number of time steps and models do not match
            # 3) time-varying system, regular time steps
            # >> multiple models and one time step duration (the case with mul-
            # tiple time steps with the same duration is tackled above already)

            if integrate_outputs:  # TODO: reach this statement
                # integrate the outputs

                self.C_line_k = [
                    np.dot(
                        C_matrix,
                        np.dot(
                            expm(
                                self.time_interval_durations[0]
                                * self.A[i if self.A_is_time_varying else 0]
                            )
                            - np.eye(self.number_states),
                            inv(self.A[i if self.A_is_time_varying else 0]),
                        ),
                    )
                    for i, C_matrix in enumerate(self.C)
                ]

            else:
                # do not integrate the outputs

                self.C_line_k = [C_matrix for C_matrix in self.C]

        # *********************************************************************

        # D matrix

        if self.D is None:
            self.D_line_k = None

        elif len(self.D) == len(self.time_interval_durations):
            # 1) time invariant system, regular time steps
            # >> only if there is only one time interval
            # 3) time-varying system, regular time steps
            # 4) time-varying system, irregular time steps

            if integrate_outputs:
                if self.C is None or self.A is None or self.B is None:
                    # no states, outputs only

                    self.D_line_k = [
                        dt * self.D[i if self.D_is_time_varying else 0]
                        for i, dt in enumerate(self.time_interval_durations)
                    ]

                else:
                    self.D_line_k = [
                        np.dot(
                            np.dot(
                                self.C[i if self.C_is_time_varying else 0],
                                np.dot(
                                    expm(
                                        dt * self.A[i if self.A_is_time_varying else 0]
                                    )
                                    - np.eye(self.number_states),
                                    inv(self.A[i if self.A_is_time_varying else 0]),
                                )
                                - dt * np.eye(self.number_states),
                            ),
                            np.dot(
                                inv(self.A[i if self.A_is_time_varying else 0]),
                                self.B[i if self.B_is_time_varying else 0],
                            ),
                        )
                        + dt * self.D[i if self.D_is_time_varying else 0]
                        for i, dt in enumerate(self.time_interval_durations)
                    ]

            else:
                # do not integrate the outputs

                self.D_line_k = [
                    self.D[i if self.D_is_time_varying else 0]
                    for i, dt in enumerate(self.time_interval_durations)
                ]

            # *****************************************************************
            # *****************************************************************

        elif len(self.D) == 1:
            # there is only one state space model and more time intervals

            # 1) time invariant system, regular time steps
            # >> one model and multiple time steps

            # output equations

            if integrate_outputs:
                # integrate the outputs

                if self.C is None or self.A is None or self.B is None:
                    # no states, outputs only

                    self.D_line_k = [
                        dt * self.D[i if self.D_is_time_varying else 0]
                        for i, dt in enumerate(self.time_interval_durations)
                    ]

                else:
                    self.D_line_k = [
                        np.dot(
                            np.dot(
                                self.C[i if self.C_is_time_varying else 0],
                                np.dot(
                                    expm(
                                        dt * self.A[i if self.A_is_time_varying else 0]
                                    )
                                    - np.eye(self.number_states),
                                    inv(self.A[i if self.A_is_time_varying else 0]),
                                )
                                - dt * np.eye(self.number_states),
                            ),
                            np.dot(
                                inv(self.A[i if self.A_is_time_varying else 0]),
                                self.B[i if self.B_is_time_varying else 0],
                            ),
                        )
                        + dt * self.D[i if self.D_is_time_varying else 0]
                        for i, dt in enumerate(self.time_interval_durations)
                    ]

            else:
                # do not integrate the outputs

                self.D_line_k = [
                    self.D[i if self.D_is_time_varying else 0]
                    for i, dt in enumerate(self.time_interval_durations)
                ]

            # *****************************************************************
            # *****************************************************************

        else:  # the number of time steps and models do not match
            # 3) time-varying system, regular time steps
            # >> multiple models and one time step duration (the case with mul-
            # tiple time steps with the same duration is tackled above already)

            # output equations
            # TODO: reach this statement
            if integrate_outputs:
                # integrate the outputs

                if self.C is None or self.A is None or self.B is None:
                    # no states, outputs only

                    self.D_line_k = [
                        self.time_interval_durations[0] * D_matrix
                        for i, D_matrix in enumerate(self.D)
                    ]

                else:
                    self.D_line_k = [
                        np.dot(
                            np.dot(
                                self.C[i if self.C_is_time_varying else 0],
                                np.dot(
                                    expm(
                                        self.time_interval_durations[0]
                                        * self.A[i if self.A_is_time_varying else 0]
                                    )
                                    - np.eye(self.number_states),
                                    inv(self.A[i if self.A_is_time_varying else 0]),
                                )
                                - self.time_interval_durations[0]
                                * np.eye(self.number_states),
                            ),
                            np.dot(
                                inv(self.A[i if self.A_is_time_varying else 0]),
                                self.B[i if self.B_is_time_varying else 0],
                            ),
                        )
                        + self.time_interval_durations[0] * D_matrix
                        for i, D_matrix in enumerate(self.D)
                    ]

            else:
                # do not integrate the outputs

                self.D_line_k = [D_matrix for D_matrix in self.D]

            # *****************************************************************
            # *****************************************************************

        # *********************************************************************
        # *********************************************************************

        # store information about the matrices

        if self.A_line_k is None:
            self.A_line_is_time_varying = None

        else:
            self.A_line_is_time_varying = True if len(self.A_line_k) > 1 else False

        if self.B_line_k is None:
            self.B_line_is_time_varying = None

        else:
            self.B_line_is_time_varying = True if len(self.B_line_k) > 1 else False

        if self.C_line_k is None:
            self.C_line_is_time_varying = None

        else:
            self.C_line_is_time_varying = True if len(self.C_line_k) > 1 else False

        if self.D_line_k is None:
            self.D_line_is_time_varying = None

        else:
            self.D_line_is_time_varying = True if len(self.D_line_k) > 1 else False

        # integrated or not

        if integrate_outputs:
            self.outputs_integrated = [y for y in range(self.number_outputs)]

        else:
            self.outputs_integrated = []

    # *************************************************************************
    # *************************************************************************

    def simulate(self, U: np.ndarray, X0: np.ndarray):
        # inputs:
        # U: m*t, for each time interval, where t is the number of time steps
        # X0: n*1, only once

        # outputs:
        # X: n*(t+1)
        # Y: r*(t+1), if no integration is done, or
        # Y: r*t, if integration is done

        # cases:
        # 1) A_line_k and B_line_k are not None, but C_line_k and D_line_k may be
        # >> the method will return X (states exist) but maybe not Y (no outputs)
        # 2) D_line_k is not None, but A_line_k, B_line_k and C_line_k are
        # >> the method will return Y (outputs exist) but not X (no states)
        # 3)

        if len(self.time_interval_durations) == 1 and self.number_models == 1:
            # if there is only one time interval duration and one ss model, the
            # discretised model can be reused multiple times with the same time
            # interval duration, provided the inputs make sense

            # determine how many time intervals the input is for

            m, number_steps = U.shape

            if m != self.number_inputs:
                raise ValueError(
                    "The U signal is not consistent with the system matrices."
                )

        else:
            number_steps = len(self.time_interval_durations)

        n = self.number_states
        # m = self.number_inputs
        r = self.number_outputs

        # *********************************************************************

        if (
            self.A_line_is_time_varying is not None
            and self.B_line_is_time_varying is not None
        ):
            # state equations

            X = np.zeros((n, number_steps + 1))

            X[:, 0] = X0

            for i in range(number_steps):
                if self.A_line_is_time_varying:
                    i_a = i

                else:
                    i_a = 0

                if self.B_line_is_time_varying:
                    i_b = i

                else:
                    i_b = 0

                X[:, i + 1] = np.dot(self.A_line_k[i_a], X[:, i]) + np.dot(
                    self.B_line_k[i_b], U[:, i]
                )

            # there may be output equations

            if (
                self.C_line_is_time_varying is not None
                and self.D_line_is_time_varying is not None
            ):
                # there are output equations

                Y = np.zeros((r, number_steps + 1))

                if len(self.outputs_integrated) == 0:
                    # no integration

                    # assumes zoh inputs and the same initial C and D matrices

                    Y[:, 0] = np.dot(self.C_line_k[0], X[:, 0]) + np.dot(
                        self.D_line_k[0], U[:, 0]
                    )

                # calculate

                for i in range(number_steps):
                    if self.C_line_is_time_varying:
                        i_c = i

                    else:
                        i_c = 0  # TODO: reach this statement

                    if self.D_line_is_time_varying:
                        i_d = i

                    else:
                        i_d = 0  # TODO: reach this statement

                    Y[:, i + 1] = np.dot(self.C_line_k[i_c], X[:, i]) + np.dot(
                        self.D_line_k[i_d], U[:, i]
                    )

                # return statement

                if len(self.outputs_integrated) == 0:
                    # no integration: all Y slots

                    return X, Y

                else:
                    # integration: skip the first Y slot

                    return X, Y[:, 1:]

            # there are no outputs

            # return statement

            return X, None

        elif (
            self.C_line_is_time_varying is None
            and self.D_line_is_time_varying is not None
        ):
            # there are no state equations, but there may be output equations

            Y = np.zeros((r, number_steps + 1))

            if len(self.outputs_integrated) == 0:
                # no integration

                # assumes zoh inputs and the same initial C and D matrices

                Y[:, 0] = np.dot(self.D_line_k[0], U[:, 0])

            # simulate

            for i in range(number_steps):
                if self.D_line_is_time_varying:
                    i_d = i

                else:
                    i_d = 0

                Y[:, i + 1] = np.dot(self.D_line_k[i_d], U[:, i])

            # return statement

            if len(self.outputs_integrated) == 0:
                # no integration: all Y slots

                return None, Y

            else:
                # integration: skip the first Y slot

                return None, Y[:, 1:]

    # *************************************************************************
    # *************************************************************************


# *****************************************************************************
# *****************************************************************************


class StatelessSystem(DynamicSystem):
    "A class for dynamic systems without states."

    # a stateless system is a dynamic system without states

    def __init__(
        self,
        time_interval_durations: list or float or int,
        D: list or np.array,
        integrate_outputs: bool = True,
    ):
        DynamicSystem.__init__(
            self,
            time_interval_durations=time_interval_durations,
            A=None,
            B=None,
            C=None,
            D=D,
            integrate_outputs=integrate_outputs,
        )  # outputs can be integrated

    # *************************************************************************
    # *************************************************************************

    def simulate(self, U: np.ndarray):
        "Simulate how the system responds to a set of input signals."

        return DynamicSystem.simulate(self, U=U, X0=None)


# *****************************************************************************
# *****************************************************************************


class OutputlessSystem(DynamicSystem):
    "A class for dynamic systems without outputs."

    def __init__(
        self,
        time_interval_durations: list or float or int,
        A: list or np.array,
        B: list or np.array,
    ):
        DynamicSystem.__init__(
            self,
            time_interval_durations=time_interval_durations,
            A=A,
            B=B,
            C=None,
            D=None,
            integrate_outputs=False,
        )  # no outputs to integrate

    # *************************************************************************
    # *************************************************************************

    def simulate(self, U: np.ndarray, X0: np.ndarray):
        "Simulate how the system responds to a set of input signals and initial conditions."

        return DynamicSystem.simulate(self, U=U, X0=X0)

# *****************************************************************************
# *****************************************************************************

class DiscretisedDynamicSystem(DynamicSystem):
    
    # workflow:
    # 1) define the system
    # 2) discretise
    # 3) simulate response
    
    # what should it do?
    # 1) discretise based on (q,k) tuples
    # 2) simulate response
    
    # 
    
    def __init__(self):
        
        pass

# *****************************************************************************
# *****************************************************************************
