# imports

import pyomo.environ as pyo

from .blocks.networks import add_network_restrictions
from .blocks.prices import add_price_functions

# *****************************************************************************
# *****************************************************************************


def create_model(
    name: str,
    use_prices_block: bool,
    node_price_model: str,
    enable_default_values: bool = True,
    enable_validation: bool = True,
    enable_initialisation: bool = True
):
    # TODO: test initialisation

    # *************************************************************************
    # *************************************************************************

    # create model object

    model = pyo.AbstractModel(name)
    
    # *************************************************************************
    # *************************************************************************

    # naming convention:
    # variables start with "var_"
    # sets start with "set_"
    # parameters start with "param_"
    # constraints start with "constr_"
    # constraints have short, tentatively descriptive names
    # variables, parameters and sets use indexes as suffixes as in the model
    # methods to build constraints start with "rule_"
    # methods to initialise sets/.../parameters start with "init_"
    # methods to validate sets/.../parameters start with "validate_"

    # TODO: migrate to kernel pyomo

    # TODO: try to use blocks to improve readability and modularity

    # *************************************************************************
    # *************************************************************************
    # *************************************************************************
    # *************************************************************************

    # sets

    # *************************************************************************
    # *************************************************************************

    # input sets

    # *************************************************************************

    # set of assessments
    model.set_Q = pyo.Set()

    # TODO: use rangesets for time-related sets

    # set of time step intervals for each assessment
    model.set_K_q = pyo.Set(model.set_Q)

    # set of representative evaluation periods for each assessment
    model.set_P_q = pyo.Set(model.set_Q)

    # set of networks
    model.set_G = pyo.Set()

    # set of nodes on each network
    model.set_L = pyo.Set(model.set_G)

    # set of importing nodes on each network
    model.set_L_imp = pyo.Set(model.set_G, within=model.set_L)

    # set of exporting nodes on each network
    model.set_L_exp = pyo.Set(model.set_G, within=model.set_L)
    
    # *************************************************************************
    # *************************************************************************

    # sparse sets

    # (q,k) tuples

    def init_set_QK(m):
        return ((q, k) for q in m.set_Q for k in m.set_K_q[q])

    model.set_QK = pyo.Set(
        dimen=2, initialize=init_set_QK if enable_initialisation else None
    )

    # (q,p) tuples

    def init_set_QP(m):
        return ((q, p) for q in m.set_Q for p in m.set_P_q[q])

    model.set_QP = pyo.Set(
        dimen=2, initialize=init_set_QP if enable_initialisation else None
    )

    # (q,p,k) tuples

    def init_set_QPK(m):
        return ((q, p, k) for (q, p) in m.set_QP for k in m.set_K_q[q])

    model.set_QPK = pyo.Set(
        dimen=3, initialize=init_set_QPK if enable_initialisation else None
    )

    # set of GL tuples for the various nodes

    def init_set_GL(m):
        return (
            (g, l) for g in m.set_G for l in m.set_L[g]  # GL tuple  # for each network
        )  # for each location

    model.set_GL = pyo.Set(
        dimen=2, initialize=(init_set_GL if enable_initialisation else None)
    )

    # set of GL tuples for import nodes

    def init_set_GL_imp(m):
        return ((g, l) for (g, l) in m.set_GL if l in m.set_L_imp[g])

    model.set_GL_imp = pyo.Set(
        dimen=2, initialize=(init_set_GL_imp if enable_initialisation else None)
    )

    # set of GL tuples for export nodes

    def init_set_GL_exp(m):
        return ((g, l) for (g, l) in m.set_GL if l in m.set_L_exp[g])

    model.set_GL_exp = pyo.Set(
        dimen=2, initialize=(init_set_GL_exp if enable_initialisation else None)
    )

    # # set of GL tuples for import or export nodes

    # model.set_GL_exp_imp = model.set_GL_imp | model.set_GL_exp

    # # set of GL tuples for non-import, non-export nodes

    # model.set_GL_not_exp_imp = model.set_GL - model.set_GL_exp_imp

    # set of GL tuples for import or export nodes

    model.set_GL_exp_imp = pyo.Set(
        dimen=2,
        initialize=(
            model.set_GL_imp | model.set_GL_exp if enable_initialisation else None
        ),
    )

    # set of GL tuples for non-import, non-export nodes

    model.set_GL_not_exp_imp = pyo.Set(
        dimen=2,
        initialize=(
            model.set_GL - model.set_GL_exp_imp if enable_initialisation else None
        ),
    )

    # set of GLL tuples for all valid node pair and direction combinations

    def init_set_GLL(m):
        return (
            (g, l1, l2)
            for g in m.set_G
            for l1 in m.set_L[g]
            if l1 not in m.set_L_exp[g]
            for l2 in m.set_L[g]
            if l2 not in m.set_L_imp[g]
            if l1 != l2
        )

    model.set_GLL = pyo.Set(
        dimen=3, initialize=(init_set_GLL if enable_initialisation else None)
    )

    # *************************************************************************
    # *************************************************************************

    # input sets

    # set for all arcs

    model.set_J = pyo.Set(model.set_GLL)

    # set for all collectively-decided arcs

    model.set_J_col = pyo.Set(model.set_GLL)

    # set for all undirected arcs

    # def validate_J_und(m, j):
    #     # j is valid if it is in J[g,l1,l2] and not J[g,l2,l1]
    #     return False
    model.set_J_und = pyo.Set(
        model.set_GLL,
        # validate=validate_J_und,
        within=model.set_J,
    )

    # set for all preexisting arcs

    model.set_J_pre = pyo.Set(model.set_GLL, within=model.set_J)

    # set for all preexisting arcs with infinite capacity

    model.set_J_pre_inf = pyo.Set(
        model.set_GLL,
        within=model.set_J_pre,
    )

    # set for all mandatory arcs

    model.set_J_mdt = pyo.Set(model.set_GLL, within=model.set_J)

    # set of arcs whose investment decision is to be modelled using sos1

    model.set_J_arc_sos1 = pyo.Set(model.set_GLL, within=model.set_J)

    # set of arcs whose investment decision is to be modelled using nnr

    model.set_J_arc_nnr = pyo.Set(model.set_GLL, within=model.set_J)

    # set of undirected arcs whose flow sense is determined using sos1

    model.set_J_sns_sos1 = pyo.Set(model.set_GLL, within=model.set_J_und)

    # set of undirected arcs whose flow sense is determined using nnr

    model.set_J_sns_nnr = pyo.Set(model.set_GLL, within=model.set_J_sns_sos1)

    # set of (new, optional) arcs that are to use interface variables

    model.set_J_int = pyo.Set(model.set_GLL, within=model.set_J)

    # set of arcs implying static losses

    model.set_J_stt = pyo.Set(model.set_GLL, within=model.set_J)

    # set of arcs whose static losses are placed in the start node

    model.set_J_stt_dep = pyo.Set(model.set_GLL, within=model.set_J_stt)

    # set of arcs whose static losses are placed in the end node

    model.set_J_stt_arr = pyo.Set(model.set_GLL, within=model.set_J_stt)

    # set of arcs whose static losses are placed upstream

    model.set_J_stt_us = pyo.Set(model.set_GLL, within=model.set_J_stt)

    # set of arcs whose static losses are placed downstream

    model.set_J_stt_ds = pyo.Set(model.set_GLL, within=model.set_J_stt)

    # *************************************************************************
    # *************************************************************************

    # systems

    # set of all systems

    model.set_I = pyo.Set()

    # set of optional systems

    model.set_I_new = pyo.Set(within=model.set_I)

    # *************************************************************************

    # inputs

    # set of inputs (indexed by system)

    model.set_M = pyo.Set(model.set_I)

    # set of inputs modelled using non-negative real variables

    model.set_M_nnr = pyo.Set(model.set_I, within=model.set_M)

    # set of inputs modelled using binary variables

    model.set_M_bin = pyo.Set(model.set_I, within=model.set_M)

    # set of amplitude-constrained inputs

    model.set_M_dim = pyo.Set(model.set_I_new, within=model.set_M)

    # set of amplitude-constrained inputs

    model.set_M_fix = pyo.Set(model.set_I, within=model.set_M)

    # set of externality-inducing inputs

    model.set_M_ext = pyo.Set(model.set_I, within=model.set_M)

    # *************************************************************************

    # outputs

    # set of outputs (indexed by system)

    model.set_R = pyo.Set(model.set_I)

    # set of outputs with fixed bounds

    model.set_R_fix = pyo.Set(model.set_I, within=model.set_R)

    # set of positive amplitude-constrained outputs

    model.set_R_dim_pos = pyo.Set(model.set_I, within=model.set_R)

    # set of negative amplitude-constrained outputs

    model.set_R_dim_neg = pyo.Set(model.set_I, within=model.set_R)

    # set of amplitude-limited outputs with matching pos. and neg. amplitudes

    model.set_R_dim_eq = pyo.Set(model.set_I, within=model.set_R)

    # set of outputs (indexed by system) inducing externalities

    model.set_R_ext = pyo.Set(model.set_I)

    # *************************************************************************

    # states

    # set of states

    model.set_N = pyo.Set(model.set_I)

    # set of states with fixed bounds

    model.set_N_fix = pyo.Set(model.set_I, within=model.set_N)

    # set of positive amplitude-constrained states

    model.set_N_dim_pos = pyo.Set(model.set_I, within=model.set_N)

    # set of negative amplitude-constrained states

    model.set_N_dim_neg = pyo.Set(model.set_I, within=model.set_N)

    # set of amplitude-limited states with matching pos. and neg. amplitudes

    model.set_N_dim_eq = pyo.Set(model.set_I, within=model.set_N)

    # set of states (indexed by system) inducing externalities

    model.set_N_ext = pyo.Set(model.set_I, within=model.set_N)

    # set of positive state variation-penalised states

    model.set_N_pos_var = pyo.Set(model.set_I, within=model.set_N)

    # set of negative state variation-penalised states

    model.set_N_neg_var = pyo.Set(model.set_I, within=model.set_N)

    # set of upper reference violation-penalised states

    model.set_N_ref_u = pyo.Set(model.set_I, within=model.set_N)

    # set of lower reference violation-penalised states

    model.set_N_ref_d = pyo.Set(model.set_I, within=model.set_N)

    # *************************************************************************
    # *************************************************************************

    # sparse index sets

    # *************************************************************************

    # all arcs

    # set of GLLJ tuples for all arcs (undirected arcs appear twice)

    def init_set_GLLJ(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2) in m.set_GLL
            if l1 != l2  # redundant, as it is also a condition for set_GLL
            for j in m.set_J[(g, l1, l2)]
        )

    model.set_GLLJ = pyo.Set(
        dimen=4, initialize=(init_set_GLLJ if enable_initialisation else None)
    )

    # sets of GLLJ tuples for all arcs with static losses

    def init_set_GLLJ_static(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ
            for j in m.set_J_stt[(g, l1, l2)]
        )

    model.set_GLLJ_static = pyo.Set(
        dimen=4, initialize=(init_set_GLLJ_static if enable_initialisation else None)
    )

    # sets of GLLJ tuples for all pre-existing arcs with static losses

    def init_set_GLLJ_static_pre(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ_static
            if j in m.set_J_pre[(g, l1, l2)]
        )

    model.set_GLLJ_static_pre = pyo.Set(
        dimen=4,
        initialize=(init_set_GLLJ_static_pre if enable_initialisation else None),
    )

    # sets of GLLJ tuples for all new arcs with static losses

    def init_set_GLLJ_static_new(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ_static
            if j not in m.set_J_pre[(g, l1, l2)]
        )

    model.set_GLLJ_static_new = pyo.Set(
        dimen=4,
        initialize=(init_set_GLLJ_static_new if enable_initialisation else None),
    )

    # *************************************************************************

    # sets of GLLJ tuples for directed arcs

    # set of GLLJ tuples for directed arcs

    def init_set_GLLJ_dir(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ
            if j not in m.set_J_und[(g, l1, l2)]
        )

    model.set_GLLJ_dir = pyo.Set(dimen=4, initialize=init_set_GLLJ_dir)

    # set of GLLJ tuples for preexisting directed arcs

    def init_set_GLLJ_dir_pre(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ_dir
            if j in m.set_J_pre[(g, l1, l2)]
        )

    model.set_GLLJ_dir_pre = pyo.Set(
        dimen=4,
        # within=model.set_GLLJ_dir,
        initialize=init_set_GLLJ_dir_pre,
    )

    # set of GLLJ tuples for preexisting directed arcs with infinite capacity

    def init_set_GLLJ_dir_pre_inf(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ_dir_pre
            if j in m.set_J_pre_inf[(g, l1, l2)]
        )

    model.set_GLLJ_dir_pre_inf = pyo.Set(
        dimen=4,
        # within=model.set_GLLJ_dir_pre,
        initialize=init_set_GLLJ_dir_pre_inf,
    )

    # set of GLLJ tuples preexisting directed arcs with finite capacity

    model.set_GLLJ_dir_pre_fin = model.set_GLLJ_dir_pre - model.set_GLLJ_dir_pre_inf

    # set of GLLJ tuples for new directed arcs

    model.set_GLLJ_dir_new = model.set_GLLJ_dir - model.set_GLLJ_dir_pre

    # set of GLLJ tuples for new directed arcs modelled using SOS1

    def init_set_GLLJ_dir_new_sos1(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ_dir_new
            if j in m.set_J_arc_sos1[(g, l1, l2)]
        )

    model.set_GLLJ_dir_new_sos1 = pyo.Set(
        dimen=4,
        # within=model.set_GLLJ_dir_new,
        initialize=init_set_GLLJ_dir_new_sos1,
    )

    # set of GLLJ tuples for new yet mandatory directed arcs

    def init_set_GLLJ_dir_new_mdt(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ_dir_new
            if j in m.set_J_mdt[(g, l1, l2)]
        )

    model.set_GLLJ_dir_new_mdt = pyo.Set(
        dimen=4,
        # within=model.set_GLLJ_dir_new,
        initialize=init_set_GLLJ_dir_new_mdt,
    )

    # set of GLLJ tuples for new yet mandatory directed arcs

    model.set_GLLJ_dir_new_opt = model.set_GLLJ_dir_new - model.set_GLLJ_dir_new_mdt

    # *************************************************************************

    # sets of GLLJ tuples for undirected arcs

    # set of GLLJ tuples for undirected arcs (one tuple per undirected arc)

    def init_set_GLLJ_und(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ
            if (g, l1, l2) in m.set_J_und
            if j in m.set_J_und[(g, l1, l2)]
        )

    model.set_GLLJ_und = pyo.Set(
        dimen=4, initialize=(init_set_GLLJ_und if enable_initialisation else None)
    )

    # set of GLLJ tuples for preexisting undirected arcs (one per arc)

    def init_set_GLLJ_und_pre(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ_und
            if (g, l1, l2) in m.set_J_pre
            if j in m.set_J_pre[(g, l1, l2)]
        )

    model.set_GLLJ_und_pre = pyo.Set(
        dimen=4,
        # within=model.set_GLLJ_und,
        initialize=init_set_GLLJ_und_pre,
    )

    # set of GLLJ tuples for preexisting undirected arcs with infinite capacity

    def init_set_GLLJ_und_pre_inf(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ_und_pre
            if (g, l1, l2) in m.set_J_pre_inf
            if j in m.set_J_pre_inf[(g, l1, l2)]
        )

    model.set_GLLJ_und_pre_inf = pyo.Set(
        dimen=4,
        # within=model.set_GLLJ_und_pre,
        initialize=init_set_GLLJ_und_pre_inf,
    )

    # set of GLLJ tuples for preexisting undirected arcs with finite capacity

    model.set_GLLJ_und_pre_fin = model.set_GLLJ_und_pre - model.set_GLLJ_und_pre_inf

    # set of GLLJ tuples for new undirected arcs (one per arc)

    model.set_GLLJ_und_new = model.set_GLLJ_und - model.set_GLLJ_und_pre

    # set of GLLJ tuples for new undirected arcs modelled using SOS1

    def init_set_GLLJ_und_new_sos1(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ_und_new
            if (g, l1, l2) in m.set_J_arc_sos1
            if j in m.set_J_arc_sos1[(g, l1, l2)]
        )

    model.set_GLLJ_und_new_sos1 = pyo.Set(
        dimen=4,
        # within=model.set_GLLJ_und_new,
        initialize=init_set_GLLJ_und_new_sos1,
    )

    # set of GLLJ tuples for new yet mandatory undirected arcs

    def init_set_GLLJ_und_new_mdt(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ_und_new
            if (g, l1, l2) in m.set_J_mdt
            if j in m.set_J_mdt[(g, l1, l2)]
        )

    model.set_GLLJ_und_new_mdt = pyo.Set(
        dimen=4,
        # within=model.set_GLLJ_und_new,
        initialize=init_set_GLLJ_und_new_mdt,
    )

    # set of GLLJ tuples for new yet mandatory undirected arcs

    model.set_GLLJ_und_new_opt = model.set_GLLJ_und_new - model.set_GLLJ_und_new_mdt

    # set of GLLJ tuples for arcs using interfaces

    def init_set_GLLJ_int(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ
            if (g, l1, l2) in m.set_J_int
            if j in m.set_J_int[(g, l1, l2)]
        )

    model.set_GLLJ_int = pyo.Set(
        dimen=4, initialize=(init_set_GLLJ_int if enable_initialisation else None)
    )

    # *************************************************************************

    # set of complementary GLLJ tuples for undirected arcs

    def init_set_GLLJ_und_ext(m):
        return ((g, l2, l1, j) for (g, l1, l2, j) in m.set_GLLJ_und)

    model.set_GLLJ_und_ext = pyo.Set(
        dimen=4, initialize=(init_set_GLLJ_und_ext if enable_initialisation else None)
    )

    # set of complementary GLLJ tuples for preexisting undirected arcs

    def init_set_GLLJ_und_pre_ext(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ_und_ext
            if j in m.set_J_pre[(g, l2, l1)]
        )

    model.set_GLLJ_und_pre_ext = pyo.Set(
        dimen=4,
        # within=model.set_GLLJ_und_ext,
        initialize=init_set_GLLJ_und_pre_ext,
    )

    # set of complementary GLLJ tuples for new undirected arcs

    model.set_GLLJ_und_new_ext = model.set_GLLJ_und_ext - model.set_GLLJ_und_pre_ext

    # set of complementary GLLJ tuples for inf. cap. presel. undirected arcs

    def init_set_GLLJ_und_pre_inf_ext(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ_und_pre_ext
            if j in m.set_J_pre_inf[(g, l2, l1)]
        )

    model.set_GLLJ_und_pre_inf_ext = pyo.Set(
        dimen=4,
        # within=model.set_GLLJ_und_pre_ext,
        initialize=init_set_GLLJ_und_pre_inf_ext,
    )

    # set of complementary GLLJ tuples for fin. cap. presel. undirected arcs

    model.set_GLLJ_und_pre_fin_ext = (
        model.set_GLLJ_und_pre_ext - model.set_GLLJ_und_pre_inf_ext
    )

    # set of complementary GLLJ tuples for sel. und. arcs modelled using SOS1

    def init_set_GLLJ_und_new_sos1_ext(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ_und_new_ext
            if j in m.set_J_arc_sos1[(g, l2, l1)]
        )

    model.set_GLLJ_und_new_sos1_ext = pyo.Set(
        dimen=4,
        # within=model.set_GLLJ_und_new_ext,
        initialize=init_set_GLLJ_und_new_sos1_ext,
    )

    # set of complementary GLLJ tuples for sel. yet mandatory undirected arcs

    def init_set_GLLJ_und_new_mdt_ext(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ_und_new_ext
            if j in m.set_J_mdt[(g, l2, l1)]
        )

    model.set_GLLJ_und_new_mdt_ext = pyo.Set(
        dimen=4,
        # within=model.set_GLLJ_und_new_ext,
        initialize=init_set_GLLJ_und_new_mdt,
    )

    # set of GLLJ tuples for new yet mandatory undirected arcs

    model.set_GLLJ_und_new_opt_ext = (
        model.set_GLLJ_und_new_ext - model.set_GLLJ_und_new_mdt_ext
    )

    def init_set_GLLJ_static_und(m):
        return (
            (g, l1, l2, j)
            # for (g,l1,l2) in m.set_GLL # or m.J_und
            # if l1!=l2 # redundant, as it is also a condition for set_GLL
            # for (g,l1,l2,j) in m.set_GLLJ
            # if j in m.set_J_und[(g,l1,l2)]
            for (g, l1, l2, j) in m.set_GLLJ_und
            if (g, l1, l2) in m.set_J_stt
            if j in m.set_J_stt[(g, l1, l2)]
        )

    model.set_GLLJ_static_und = pyo.Set(dimen=4, initialize=init_set_GLLJ_static_und)

    def init_set_GLLJ_static_und_ext(m):
        return (
            (g, l2, l1, j)
            # for (g,l1,l2) in m.set_GLL # or m.J_und
            # if l1!=l2 # redundant, as it is also a condition for set_GLL
            # for (g,l1,l2,j) in m.set_GLLJ
            # if j in m.set_J_und[(g,l1,l2)]
            for (g, l1, l2, j) in m.set_GLLJ_und
            if (g, l1, l2) in m.set_J_stt
            if j in m.set_J_stt[(g, l1, l2)]
        )

    model.set_GLLJ_static_und_ext = pyo.Set(
        dimen=4, initialize=init_set_GLLJ_static_und_ext
    )

    # model.set_GLLJ_static_und_red = (
    #     model.set_GLLJ_static_und | model.set_GLLJ_static_und_ext
    #     )

    def init_set_GLLJ_static_und_red(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ_und_red
            if (j in m.set_J_stt[(g, l1, l2)] or j in m.set_J_stt[(g, l2, l1)])
        )

    model.set_GLLJ_static_und_red = pyo.Set(
        dimen=4,
        initialize=(init_set_GLLJ_static_und_red if enable_initialisation else None),
    )

    # sets of GLLJ tuples for pre-existing undirected arcs w/ static losses

    model.set_GLLJ_static_und_pre = model.set_GLLJ_static & model.set_GLLJ_und_pre

    # sets of GLLJ tuples for new undirected arcs w/ static losses

    model.set_GLLJ_static_und_new = model.set_GLLJ_static & model.set_GLLJ_und_new

    # *************************************************************************

    # set of GLLJ tuples for undirected arcs with redundancies

    # set of GLLJ tuples for undirected arcs

    def init_set_GLLJ_und_red(m):
        return m.set_GLLJ_und | m.set_GLLJ_und_ext

    model.set_GLLJ_und_red = pyo.Set(
        dimen=4, initialize=(init_set_GLLJ_und_red if enable_initialisation else None)
    )

    # set of GLLJ tuples for preexisting undirected arcs

    model.set_GLLJ_und_pre_red = model.set_GLLJ_und_pre | model.set_GLLJ_und_pre_ext

    # # set of GLLJ tuples for new undirected arcs

    # model.set_GLLJ_und_new_red = (
    #     model.set_GLLJ_und_red - model.set_GLLJ_und_pre_red)

    # set of GLLJ tuples for preexisting undirected arcs with infinite capacity

    model.set_GLLJ_und_pre_inf_red = (
        model.set_GLLJ_und_pre_inf | model.set_GLLJ_und_pre_inf_ext
    )

    # set of GLLJ tuples for preexisting undirected arcs with finite capacity

    model.set_GLLJ_und_pre_fin_red = (
        model.set_GLLJ_und_pre_red - model.set_GLLJ_und_pre_inf_red
    )

    # *************************************************************************

    # sets of GLLJ tuples for directed and undirected arcs (no redundancies)

    # set of GLLJ tuples for preexisting arcs with finite capacity

    model.set_GLLJ_pre_fin = model.set_GLLJ_dir_pre_fin | model.set_GLLJ_und_pre_fin

    # set of GLLJ tuples for new arcs

    def init_set_GLLJ_new(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ
            if j not in m.set_J_pre[(g, l1, l2)]
        )

    model.set_GLLJ_new = pyo.Set(
        dimen=4, initialize=(init_set_GLLJ_new if enable_initialisation else None)
    )

    # set of GLLJ tuples for new individual arcs

    def init_set_GLLJ_sgl(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ_new
            if j not in m.set_J_col[(g, l1, l2)]
        )

    model.set_GLLJ_sgl = pyo.Set(
        dimen=4, initialize=(init_set_GLLJ_sgl if enable_initialisation else None)
    )

    # *************************************************************************

    # sets of GLLJ tuples for directed and undirected arcs (with redundancies)

    # set of GLLJ tuples for all arcs (two tuples per undirected arc)

    # model.set_GLLJ_red = (
    #     model.set_GLLJ_dir | model.set_GLLJ_und_red
    #     )

    def init_set_GLLJ_red(m):
        # return model.set_GLLJ_dir | model.set_GLLJ_und_red
        return m.set_GLLJ | m.set_GLLJ_und_ext
        # return ((g,l1,l2,j)
        #         for (g,l1,l2,j) in m.set_GLLJ
        #         if j not in m.set_J_pre[(g,l1,l2)])

    model.set_GLLJ_red = pyo.Set(
        dimen=4, initialize=(init_set_GLLJ_red if enable_initialisation else None)
    )

    # set of GLLJ tuples for preexisting arcs with finite capacity

    # model.set_GLLJ_pre_fin_red = (
    #     model.set_GLLJ_dir_pre_fin | model.set_GLLJ_und_pre_fin_red
    #     )

    def init_set_GLLJ_pre_fin_red(m):
        # return m.set_GLLJ_dir_pre_fin | m.set_GLLJ_und_pre_fin_red
        return m.set_GLLJ_pre_fin | m.set_GLLJ_und_pre_fin_ext

    model.set_GLLJ_pre_fin_red = pyo.Set(
        dimen=4,
        initialize=(init_set_GLLJ_pre_fin_red if enable_initialisation else None),
    )

    # *************************************************************************

    # set of arc options

    model.set_H_gllj = pyo.Set(model.set_GLLJ_sgl)

    # TODO: set_GLLJH

    # sets of GLLJH tuples for arc options

    def init_set_GLLJH_sgl(m):
        return (
            (g, l1, l2, j, h)
            for (g, l1, l2, j) in m.set_GLLJ_sgl
            for h in m.set_H_gllj[(g, l1, l2, j)]
        )

    model.set_GLLJH_sgl = pyo.Set(
        dimen=5, initialize=(init_set_GLLJH_sgl if enable_initialisation else None)
    )
    # set of GLLJH tuples for new arcs with static losses

    def init_set_GLLJH_static_new(m):
        _temp = [
            (g, l1, l2, j, h)
            for (g, l1, l2, j, h) in m.set_GLLJH_sgl
            if j in m.set_J_stt[(g, l1, l2)]
        ]
        _temp.extend(
            [
                (g, l1, l2, j, h)
                for (t, g, l1, l2, j) in m.set_TGLLJ
                if j in m.set_J_stt[(g, l1, l2)]
                for h in m.set_H_t[t]
            ]
        )
        return _temp

    model.set_GLLJH_static_new = pyo.Set(
        dimen=5,
        initialize=(init_set_GLLJH_static_new if enable_initialisation else None),
    )

    # *************************************************************************

    # arc selection using SOS1

    # set of GLLJ tuples for all new arcs using sos1

    def init_set_GLLJ_arc_inv_sos1(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ_sgl
            if (g, l1, l2) in m.set_J_arc_sos1
            if j in m.set_J_arc_sos1[(g, l1, l2)]
        )

    model.set_GLLJ_arc_inv_sos1 = pyo.Set(
        dimen=4,
        initialize=(init_set_GLLJ_arc_inv_sos1 if enable_initialisation else None),
    )

    # set of GLLJH tuples for new arcs modelled using SOS1

    def init_set_GLLJH_arc_inv_sos1(m):
        return (
            (g, l1, l2, j, h)
            for (g, l1, l2, j) in m.set_GLLJ_arc_inv_sos1
            for h in m.set_H_gllj[(g, l1, l2, j)]
        )

    model.set_GLLJH_arc_inv_sos1 = pyo.Set(
        dimen=5,
        within=model.set_GLLJH_sgl,
        initialize=(init_set_GLLJH_arc_inv_sos1 if enable_initialisation else None),
    )

    # *************************************************************************

    # flow sense determination using SOS1

    # set of GLLJ tuples for undirected arcs whose flow sense is det. via SOS1

    def init_set_GLLJ_und_sns_sos1(m):
        return (
            (g, l1, l2, j)
            for (g, l1, l2, j) in m.set_GLLJ_und
            if (g, l1, l2) in m.set_J_sns_sos1
            if j in m.set_J_sns_sos1[(g, l1, l2)]
        )

    model.set_GLLJ_und_sns_sos1 = pyo.Set(
        dimen=4,
        # within=model.set_GLLJ_und,
        initialize=(init_set_GLLJ_und_sns_sos1 if enable_initialisation else None),
    )

    # set of GLLJQK tuples for flow sense determination using SOS1

    model.set_GLLJQK_und_sns_sos1_red = pyo.Set(
        dimen=6,
        initialize=(
            model.set_GLLJ_und_red * model.set_QK if enable_initialisation else None
        ),
    )

    # set of GLLJQK tuples for both directions indexed by the ref. GLLJQK tuple
    # note: this set includes tuples for both directions

    def init_set_GLLJQK_und_sns_sos1_red_gllj(m, g, l1, l2, j, q, k):
        return ((g, l1, l2, j, q, k), (g, l2, l1, j, q, k))

    model.set_GLLJQK_und_sns_sos1_red_gllj = pyo.Set(
        model.set_GLLJ_und_sns_sos1,
        model.set_QK,
        dimen=6,
        initialize=init_set_GLLJQK_und_sns_sos1_red_gllj,
    )

    # *************************************************************************

    # inputs

    # set of IM tuples

    def init_set_IM(m):
        return ((i, m_i) for i in m.set_I for m_i in m.set_M[i])

    model.set_IM = pyo.Set(dimen=2, initialize=init_set_IM)

    # set of IM tuples for systems with binary signals

    def init_set_IM_bin(m):
        return ((i, m_i) for (i, m_i) in m.set_IM if m_i in m.set_M_bin[i])

    model.set_IM_bin = pyo.Set(dimen=2, initialize=init_set_IM_bin, within=model.set_IM)

    # set of IM tuples for tech. with dimensionable reference mode levels

    def init_set_IM_dim(m):
        return ((i, m_i) for (i, m_i) in m.set_IM if m_i in m.set_M_dim[i])

    model.set_IM_dim = pyo.Set(dimen=2, initialize=init_set_IM_dim, within=model.set_IM)

    # set of IM tuples for fixed amplitude inputs

    def init_set_IM_fix(m):
        return ((i, m_i) for (i, m_i) in m.set_IM if m_i in m.set_M_fix[i])

    model.set_IM_fix = pyo.Set(dimen=2, initialize=init_set_IM_fix, within=model.set_IM)

    # set of IM tuples for technologies whose modes can induce externalities

    def init_set_IM_ext(m):
        return ((i, m_i) for (i, m_i) in m.set_IM if m_i in m.set_M_ext[i])

    model.set_IM_ext = pyo.Set(dimen=2, initialize=init_set_IM_ext, within=model.set_IM)

    # *************************************************************************

    # states

    # set of IN tuples

    def init_set_IN(m):
        return (
            (i, n_i) for i in m.set_I for n_i in m.set_N[i]  # IN tuple
        )  # for each state

    model.set_IN = pyo.Set(dimen=2, initialize=init_set_IN)

    # set of IN tuples for states with fixed bounds

    def init_set_IN_fix(m):
        return ((i, n_i) for i in m.set_I for n_i in m.set_N_fix[i])

    model.set_IN_fix = pyo.Set(dimen=2, initialize=init_set_IN_fix)

    # set of IN tuples for converters with amplitude-constrained states

    def init_set_IN_dim_eq(m):
        return ((i, n_i) for (i, n_i) in m.set_IN if n_i in m.set_N_dim_eq[i])

    model.set_IN_dim_eq = pyo.Set(
        dimen=2, initialize=init_set_IN_dim_eq, within=model.set_IN
    )

    # set of IN tuples for converters with pos. amplitude-constrained states

    def init_set_IN_dim_pos(m):
        return ((i, n_i) for (i, n_i) in m.set_IN if n_i in m.set_N_dim_pos[i])

    model.set_IN_dim_pos = pyo.Set(
        dimen=2, initialize=init_set_IN_dim_pos, within=model.set_IN
    )

    # set of IN tuples for converters with neg. amplitude-constrained states

    def init_set_IN_dim_neg(m):
        return ((i, n_i) for (i, n_i) in m.set_IN if n_i in m.set_N_dim_neg[i])

    model.set_IN_dim_neg = pyo.Set(
        dimen=2, initialize=init_set_IN_dim_neg, within=model.set_IN
    )

    # set of IN tuples for converters with externality-inducing states

    def init_set_IN_ext(m):
        return ((i, n_i) for (i, n_i) in m.set_IN if n_i in m.set_N_ext[i])

    model.set_IN_ext = pyo.Set(dimen=2, initialize=init_set_IN_ext, within=model.set_IN)

    # set of IN tuples for positive variation-penalised states

    def init_set_IN_pos_var(m):
        return ((i, n_i) for (i, n_i) in m.set_IN if n_i in m.set_N_pos_var[i])

    model.set_IN_pos_var = pyo.Set(
        dimen=2, initialize=init_set_IN_pos_var, within=model.set_IN
    )

    # set of IN tuples for negative variation-penalised states

    def init_set_IN_neg_var(m):
        return ((i, n_i) for (i, n_i) in m.set_IN if n_i in m.set_N_neg_var[i])

    model.set_IN_neg_var = pyo.Set(
        dimen=2, initialize=init_set_IN_neg_var, within=model.set_IN
    )

    # set of IN tuples for upper reference violation penalised states

    def init_set_IN_ref_u(m):
        return ((i, n_i) for (i, n_i) in m.set_IN if n_i in m.set_N_ref_u[i])

    model.set_IN_ref_u = pyo.Set(
        dimen=2, initialize=init_set_IN_ref_u, within=model.set_IN
    )

    # set of IN tuples for lower reference violation penalised states

    def init_set_IN_ref_d(m):
        return ((i, n_i) for (i, n_i) in m.set_IN if n_i in m.set_N_ref_d[i])

    model.set_IN_ref_d = pyo.Set(
        dimen=2, initialize=init_set_IN_ref_d, within=model.set_IN
    )

    # *************************************************************************

    # outputs

    # set of IR tuples

    def init_set_IR(m):
        return ((i, r_i) for i in m.set_I for r_i in m.set_R[i])

    model.set_IR = pyo.Set(dimen=2, initialize=init_set_IR)

    # set of IR tuples for outputs with fixed bounds

    def init_set_IR_fix(m):
        return ((i, r_i) for i in m.set_I for r_i in m.set_R_fix[i])

    model.set_IR_fix = pyo.Set(dimen=2, initialize=init_set_IR_fix)

    # set of IR tuples for converters with matching pos. and neg. out. amp. limits

    def init_set_IR_dim_eq(m):
        return ((i, r_i) for (i, r_i) in m.set_IR if r_i in m.set_R_dim_eq[i])

    model.set_IR_dim_eq = pyo.Set(dimen=2, initialize=init_set_IR_dim_eq)

    # set of IR tuples for converters with amplitude-penalised outputs

    def init_set_IR_dim_neg(m):
        return ((i, r_i) for (i, r_i) in m.set_IR if r_i in m.set_R_dim_neg[i])

    model.set_IR_dim_neg = pyo.Set(dimen=2, initialize=init_set_IR_dim_neg)

    # set of IR tuples for converters with amplitude-penalised outputs

    def init_set_IR_dim(m):
        return ((i, r_i) for (i, r_i) in m.set_IR if r_i in m.set_R_dim[i])

    model.set_IR_dim = pyo.Set(dimen=2, initialize=init_set_IR_dim)

    # set of IR tuples for converters with pos. amplitude-constrained outputs

    def init_set_IR_dim_pos(m):
        return ((i, r_i) for (i, r_i) in m.set_IR if r_i in m.set_R_dim_pos[i])

    model.set_IR_dim_pos = pyo.Set(dimen=2, initialize=init_set_IR_dim_pos)

    # set of IR tuples for converters with externality-inducing outputs

    def init_set_IR_ext(m):
        return ((i, r_i) for (i, r_i) in m.set_IR if r_i in m.set_R_ext[i])

    model.set_IR_ext = pyo.Set(dimen=2, initialize=init_set_IR_ext)

    # *************************************************************************

    # combined inputs/states/outputs

    # TODO: narrow down these sets if possible

    # set of INN tuples

    def init_set_INN(m):
        return ((i, n1, n2) for (i, n1) in m.set_IN for n2 in m.set_N[i])

    model.set_INN = pyo.Set(dimen=3, initialize=init_set_INN)

    # set of INM tuples

    def init_set_INM(m):
        return ((i, n_i, m_i) for (i, n_i) in m.set_IN for m_i in m.set_M[i])

    model.set_INM = pyo.Set(dimen=3, initialize=init_set_INM)

    # set of IRM tuples

    def init_set_IRM(m):
        return (
            (i, r_i, m_i) for (i, r_i) in m.set_IR for m_i in m.set_M[i]
        )  # can be further constrained

    model.set_IRM = pyo.Set(dimen=3, initialize=init_set_IRM)
    # set of IRN tuples

    def init_set_IRN(m):
        return (
            (i, r_i, n_i) for (i, r_i) in m.set_IR for n_i in m.set_N[i]
        )  # can be further constrained

    model.set_IRN = pyo.Set(dimen=3, initialize=init_set_IRN)

    # *************************************************************************
    # *************************************************************************

    # set of arc groups

    model.set_T = pyo.Set()

    # set of mandatory arc groups

    model.set_T_mdt = pyo.Set(within=model.set_T)

    # set of arc groups requiring interface variables

    model.set_T_int = pyo.Set(within=model.set_T)

    # set of arc groups relying on SOS1

    model.set_T_sos1 = pyo.Set(within=model.set_T)

    # set of arc groups relying on binary variables

    model.set_T_bin = pyo.Set(within=model.set_T)

    # set of arg groups relying on non-negative real variables

    model.set_T_nnr = pyo.Set(within=model.set_T_sos1)  # or set_T

    # # set of GLLJ tuples for arcs in arc groups

    # model.set_GLLJ_col = pyo.Set(within=model.set_GLLJ_new)

    # set of arcs in the various arc groups

    model.set_GLLJ_col_t = pyo.Set(model.set_T, within=model.set_GLLJ_new)

    # set of arc options for arc groups

    model.set_H_t = pyo.Set(model.set_T)

    # set of (t,h) tuples

    def init_set_TH(m):
        return ((t, h) for t in m.set_T for h in m.set_H_t[t])

    model.set_TH = pyo.Set(dimen=2, initialize=init_set_TH)

    # set of (t,g,l1,l2,j) tuples

    def init_set_TGLLJ(m):
        return (
            (t, g, l1, l2, j) for t in m.set_T for (g, l1, l2, j) in m.set_GLLJ_col_t[t]
        )

    model.set_TGLLJ = pyo.Set(dimen=5, initialize=init_set_TGLLJ)

    # set of (t,g,l1,l2,j) tuples for undirected arcs

    def init_set_TGLLJ_und(m):
        return (
            (t, g, l1, l2, j)
            for (t, g, l1, l2, j) in m.set_TGLLJ
            if j in m.set_J_und[(g, l1, l2)]
        )

    model.set_TGLLJ_und = pyo.Set(dimen=5, initialize=init_set_TGLLJ_und)

    # set of TH tuples for groups using SOS1

    def init_set_TH_arc_inv_sos1(m):
        return ((t, h) for (t, h) in m.set_TH if t in m.set_T_sos1)
        # for t in m.set_T_sos1
        # for h in m.set_H_t[t])

    model.set_TH_arc_inv_sos1 = pyo.Set(
        dimen=2, within=model.set_TH, initialize=init_set_TH_arc_inv_sos1
    )

    # minimum cost of a group of arcs

    model.param_c_arc_min_th = pyo.Param(model.set_TH, within=pyo.NonNegativeReals)

    # unit flow amplitude cost

    model.param_c_arc_var_t = pyo.Param(model.set_T, within=pyo.NonNegativeReals)

    # maximum nominal amplitude

    model.param_v_amp_max_th = pyo.Param(model.set_TH, within=pyo.PositiveReals)

    # sos1 weights for arc group options

    model.param_arc_inv_sos1_weights_th = pyo.Param(
        model.set_TH_arc_inv_sos1, within=pyo.NonNegativeReals
    )

    # capital cost of group

    model.var_capex_arc_col_t = pyo.Var(model.set_T, within=pyo.NonNegativeReals)

    # investment decision for group (can be binary or nnr)

    def domain_var_delta_arc_inv_th(m, t, h):
        # has to be in set_J_sos1 and set_J_mdt
        if t in m.set_T_nnr:
            return pyo.UnitInterval  # [0,1]; alternatively: [0,inf]
        else:
            return pyo.Binary  # {0,1}

    model.var_delta_arc_inv_th = pyo.Var(
        model.set_TH, domain=domain_var_delta_arc_inv_th
    )

    # interface variable for group

    model.var_xi_arc_inv_t = pyo.Var(model.set_T_int, within=pyo.UnitInterval)

    # nominal amplitude

    model.var_v_amp_t = pyo.Var(model.set_T)

    # *************************************************************************
    # *************************************************************************
    # *************************************************************************
    # *************************************************************************

    # parameters

    # *************************************************************************
    # *************************************************************************

    # objective function

    # *************************************************************************

    # general parameters

    # assessment weight (default: 1)

    model.param_c_wgt_q = pyo.Param(model.set_Q, within=pyo.PositiveReals, default=1)

    # discount factors

    model.param_c_df_qp = pyo.Param(model.set_QP, within=pyo.PositiveReals)

    # relative weight of time step k within representative period p

    model.param_c_time_qpk = pyo.Param(
        model.set_QPK, within=pyo.PositiveReals, default=1
    )

    # converters

    # externality cost per input unit

    model.param_c_ext_u_imqk = pyo.Param(
        model.set_IM_ext, model.set_QK, within=pyo.NonNegativeReals, default=0
    )

    # externality cost per output unit

    model.param_c_ext_y_irqk = pyo.Param(
        model.set_IR_ext, model.set_QK, within=pyo.NonNegativeReals, default=0
    )

    # externality cost per state unit

    model.param_c_ext_x_inqk = pyo.Param(
        model.set_IN_ext, model.set_QK, within=pyo.NonNegativeReals, default=0
    )

    # unit cost of positive state variations

    model.param_c_pos_var_in = pyo.Param(
        model.set_IN_pos_var, within=pyo.NonNegativeReals, default=0
    )

    # unit cost of negative state variations

    model.param_c_neg_var_in = pyo.Param(
        model.set_IN_neg_var, within=pyo.NonNegativeReals, default=0
    )

    # unit cost of upper state reference violations

    model.param_c_ref_u_inqk = pyo.Param(
        model.set_IN_ref_u, model.set_QK, within=pyo.NonNegativeReals, default=0
    )

    # unit cost of lower state reference violations

    model.param_c_ref_d_inqk = pyo.Param(
        model.set_IN_ref_d, model.set_QK, within=pyo.NonNegativeReals, default=0
    )

    # minimum converter cost

    model.param_c_cvt_min_i = pyo.Param(
        model.set_I_new, within=pyo.NonNegativeReals, default=0
    )

    # unit (positive) input amplitude cost

    model.param_c_cvt_u_im = pyo.Param(
        model.set_IM_dim, within=pyo.NonNegativeReals, default=0
    )

    # unit output amplitude cost

    model.param_c_cvt_y_ir = pyo.Param(
        model.set_IR_dim, within=pyo.NonNegativeReals, default=0
    )

    # unit positive state amplitude cost

    model.param_c_cvt_x_pos_in = pyo.Param(
        model.set_IN_dim_pos, within=pyo.NonNegativeReals, default=0
    )

    # unit negative state amplitude cost

    model.param_c_cvt_x_neg_in = pyo.Param(
        model.set_IN_dim_neg, within=pyo.NonNegativeReals, default=0
    )

    # unit positive output amplitude cost

    model.param_c_cvt_y_pos_ir = pyo.Param(
        model.set_IR_dim_pos, within=pyo.NonNegativeReals, default=0
    )

    # unit negative output amplitude cost

    model.param_c_cvt_y_neg_ir = pyo.Param(
        model.set_IR_dim_neg, within=pyo.NonNegativeReals, default=0
    )

    # *************************************************************************

    # arcs

    # minimum arc cost

    model.param_c_arc_min_glljh = pyo.Param(
        model.set_GLLJH_sgl, within=pyo.NonNegativeReals
    )

    # specific arc flow amplitude cost

    model.param_c_arc_var_gllj = pyo.Param(
        model.set_GLLJ_sgl, within=pyo.NonNegativeReals  # inclusive
    )

    # arc transmission efficiency

    model.param_eta_glljqk = pyo.Param(
        model.set_GLLJ_red, model.set_QK, within=pyo.PositiveReals
    )

    # maximum arc flow amplitude (new arcs only)

    model.param_v_amp_max_glljh = pyo.Param(
        model.set_GLLJH_sgl, within=pyo.PositiveReals
    )

    # upper bounds for arc flows

    model.param_v_ub_glljqk = pyo.Param(
        model.set_GLLJ_pre_fin_red, model.set_QK, within=pyo.PositiveReals
    )

    # arc flow amplitude adjustment coefficient (can be zero)

    model.param_f_amp_v_glljqk = pyo.Param(
        model.set_GLLJ_new, model.set_QK, within=pyo.NonNegativeReals
    )

    # sos1 weights for arc options

    model.param_arc_inv_sos1_weights_glljh = pyo.Param(
        model.set_GLLJH_arc_inv_sos1, within=pyo.NonNegativeReals
    )

    # sos1 weights for flow senses
    # TODO: define these coefficients here using an init method
    model.param_arc_sns_sos1_weights_glljqk = pyo.Param(
        model.set_GLLJQK_und_sns_sos1_red, within=pyo.NonNegativeReals
    )

    # *************************************************************************

    # relative variation of arc-dependent losses

    # arc-dependent losses (new arcs)

    model.param_w_new_glljhqk = pyo.Param(
        model.set_GLLJH_static_new, model.set_QK, within=pyo.NonNegativeReals
    )

    # arc-dependent losses (preexisting arcs)

    model.param_w_pre_glljqk = pyo.Param(
        model.set_GLLJ_static_pre, model.set_QK, within=pyo.NonNegativeReals
    )

    # *************************************************************************

    # network

    # static flow needs

    model.param_v_base_glqk = pyo.Param(
        model.set_GL_not_exp_imp, model.set_QK, within=pyo.Reals, default=0
    )

    # effect of system inputs on specific network and node pairs

    model.param_a_nw_glimqk = pyo.Param(
        model.set_GL_not_exp_imp,
        model.set_IM,
        model.set_QK,
        default=0,  # default: no effect
        within=pyo.Reals,
    )

    # effect of system outputs on specific network and node pairs

    model.param_a_nw_glirqk = pyo.Param(
        model.set_GL_not_exp_imp,
        model.set_IR,
        model.set_QK,
        default=0,  # default: no effect
        within=pyo.Reals,
    )

    # *************************************************************************

    # inputs

    # upper bounds for (non-binary, non-dimensionable) inputs

    model.param_u_ub_imqk = pyo.Param(
        model.set_IM_fix, model.set_QK, within=pyo.PositiveReals
    )

    # maximum input limits

    model.param_u_amp_max_im = pyo.Param(
        model.set_IM_dim, within=pyo.PositiveReals, default=1
    )

    # time interval-dependent adjustment coefficients for input limits

    model.param_f_amp_u_imqk = pyo.Param(
        model.set_IM_dim, model.set_QK, within=pyo.PositiveReals, default=1
    )

    # *************************************************************************

    # states

    # initial conditions

    model.param_x_inq0 = pyo.Param(model.set_IN, model.set_Q, within=pyo.Reals)

    # fixed upper bounds for state variables

    model.param_x_ub_irqk = pyo.Param(model.set_IN_fix, model.set_QK, within=pyo.Reals)

    # fixed lower bounds for state variables

    model.param_x_lb_irqk = pyo.Param(model.set_IN_fix, model.set_QK, within=pyo.Reals)

    # maximum positive amplitude for states

    model.param_x_amp_pos_max_in = pyo.Param(
        model.set_IN_dim_pos, within=pyo.PositiveReals
    )

    # maximum negative amplitude for states

    model.param_x_amp_neg_max_in = pyo.Param(
        model.set_IN_dim_neg, within=pyo.PositiveReals
    )

    # adjustment of positive state amplitude limits

    model.param_f_amp_pos_x_inqk = pyo.Param(
        model.set_IN_dim_pos, model.set_QK, within=pyo.PositiveReals, default=1
    )

    # adjustment of negative state amplitude limits

    model.param_f_amp_neg_x_inqk = pyo.Param(
        model.set_IN_dim_neg, model.set_QK, within=pyo.PositiveReals, default=1
    )

    # state equations: coefficients from C matrix

    model.param_a_eq_x_innqk = pyo.Param(
        model.set_INN, model.set_QK, default=0, within=pyo.Reals  # default: no effect
    )

    # state equations: coefficients from D matrix

    model.param_b_eq_x_inmqk = pyo.Param(
        model.set_INM, model.set_QK, default=0, within=pyo.Reals  # default: no effect
    )

    # state equations: constant term

    model.param_e_eq_x_inqk = pyo.Param(
        model.set_IN, model.set_QK, default=0, within=pyo.Reals  # default: no effect
    )

    # *************************************************************************

    # outputs

    # fixed upper bounds for output variables

    model.param_y_ub_irqk = pyo.Param(model.set_IR_fix, model.set_QK, within=pyo.Reals)

    # fixed lower bounds for output variables

    model.param_y_lb_irqk = pyo.Param(model.set_IR_fix, model.set_QK, within=pyo.Reals)

    # adjustment of positive output amplitude limits

    model.param_f_amp_y_pos_irqk = pyo.Param(
        model.set_IR_dim_pos, model.set_QK, within=pyo.PositiveReals, default=1
    )

    # adjustment of negative output amplitude limits

    model.param_f_amp_y_neg_irqk = pyo.Param(
        model.set_IR_dim_neg, model.set_QK, within=pyo.PositiveReals, default=1
    )

    # maximum positive amplitude limit for outputs

    model.param_y_amp_pos_max_ir = pyo.Param(
        model.set_IR_dim_pos, within=pyo.PositiveReals
    )

    # maximum negative amplitude limit for outputs

    model.param_y_amp_neg_max_ir = pyo.Param(
        model.set_IR_dim_neg, within=pyo.PositiveReals
    )

    # output equation coefficients from C matrix

    model.param_c_eq_y_irnqk = pyo.Param(
        model.set_IRN, model.set_QK, default=0, within=pyo.Reals  # default: no effect
    )

    # output equation coefficients from D matrix

    model.param_d_eq_y_irmqk = pyo.Param(
        model.set_IRM, model.set_QK, default=0, within=pyo.Reals  # default: no effect
    )

    # output equation constant

    model.param_e_eq_y_irqk = pyo.Param(
        model.set_IR, model.set_QK, default=0, within=pyo.Reals  # default: no effect
    )

    # *************************************************************************
    # *************************************************************************
    # *************************************************************************
    # *************************************************************************

    # variables

    # *************************************************************************
    # *************************************************************************

    # objective function

    # capex

    model.var_capex = pyo.Var(within=pyo.NonNegativeReals)

    # sum of discounted net cash flows

    model.var_sdncf_q = pyo.Var(model.set_Q, within=pyo.Reals)

    # sum of discounted externalities

    model.var_sdext_q = pyo.Var(model.set_Q, within=pyo.Reals)

    # capex for individual arcs

    model.var_capex_arc_gllj = pyo.Var(model.set_GLLJ_sgl, within=pyo.NonNegativeReals)

    # capex for installing individual converters

    model.var_capex_cvt_i = pyo.Var(model.set_I_new, within=pyo.NonNegativeReals)

    # exported flow revenue

    model.var_efr_glqpk = pyo.Var(
        model.set_GL_exp, model.set_QPK, within=pyo.NonNegativeReals
    )

    # imported flow cost

    model.var_ifc_glqpk = pyo.Var(
        model.set_GL_imp, model.set_QPK, within=pyo.NonNegativeReals
    )

    # *************************************************************************

    # arcs

    # arc flow variables

    def bounds_var_v_glljqk(m, g, l1, l2, j, q, k):
        # if ((g, l1, l2, j) in m.set_GLLJ_pre_fin_red and
        #     (g, l1, l2, j) not in m.set_GLLJ_und and
        #     (g, l2, l1, j) not in m.set_GLLJ_und):
        if (
            (g, l1, l2, j) in m.set_GLLJ_pre_fin_red
            and (g, l1, l2) in m.set_J
            and j not in m.set_J_und[(g, l1, l2)]
        ):
            # directed, preexisting and finite capacity
            return (0, m.param_v_ub_glljqk[(g, l1, l2, j, q, k)])
        else:
            # infinite capacity or dynamic finite capacity (set elsewhere)
            return (0, None)

    model.var_v_glljqk = pyo.Var(
        model.set_GLLJ_red,
        model.set_QK,
        within=pyo.NonNegativeReals,
        bounds=bounds_var_v_glljqk,
    )

    # arc flow amplitude variables

    model.var_v_amp_gllj = pyo.Var(model.set_GLLJ_sgl, within=pyo.NonNegativeReals)

    # static losses per arc

    model.var_w_glljqk = pyo.Var(
        model.set_GLLJ_static_new, model.set_QK, within=pyo.NonNegativeReals
    )

    # static losses modulated by the flow sense

    model.var_w_sns_glljqk = pyo.Var(
        model.set_GLLJ_static_und_red, model.set_QK, within=pyo.NonNegativeReals
    )

    # *************************************************************************

    # decision to install a given arc

    def domain_var_delta_arc_inv_glljh(m, g, l1, l2, j, h):
        if j in m.set_J_arc_nnr[(g, l1, l2)]:
            return pyo.UnitInterval  # [0,1]; alternatively: [0,inf]
        else:
            return pyo.Binary  # {0,1}

    model.var_delta_arc_inv_glljh = pyo.Var(
        model.set_GLLJH_sgl, domain=domain_var_delta_arc_inv_glljh
    )

    # flow direction variables for undirected arcs

    def domain_var_zeta_sns_glljqk(m, g, l1, l2, j, q, k):
        try:
            if j in m.set_J_sns_nnr[(g, l1, l2)]:
                return pyo.UnitInterval  # [0,1]; alternatively: [0,inf]
            else:
                return pyo.Binary  # default: {0,1}
        except KeyError:
            return pyo.Binary  # default: {0,1}

    model.var_zeta_sns_glljqk = pyo.Var(
        model.set_GLLJ_und_red,  # undirected arcs, both preexisting and new
        model.set_QK,  # one per time interval
        domain=domain_var_zeta_sns_glljqk,  # can be binary or nnr
    )

    # interface variables to separate the flow sense selection and arc invest.

    model.var_xi_arc_inv_gllj = pyo.Var(model.set_GLLJ_int, within=pyo.UnitInterval)

    # *************************************************************************

    # converters

    # decision to install converter i

    model.var_cvt_inv_i = pyo.Var(model.set_I_new, within=pyo.Binary)

    # inputs

    # input variables

    def bounds_var_u_imqk(m, i, m_i, q, k):
        if (i, m_i) in m.param_u_ub_imqk:
            # predefined limit
            return (0, m.param_u_ub_imqk[(i, m_i, q, k)])
        else:
            # dynamic limit (set elsewhere)
            return (0, None)

    def domain_var_u_imqk(m, i, m_i, q, k):
        try:
            if m_i in m.set_M_bin[i]:
                return pyo.Binary  # binary: {0,1}
            else:
                return pyo.NonNegativeReals  # nonnegative real: [0,inf]
        except KeyError:
            return pyo.NonNegativeReals  # nonnegative real: [0,inf]

    model.var_u_imqk = pyo.Var(
        model.set_IM,
        model.set_QK,
        domain=domain_var_u_imqk,
        # within=pyo.NonNegativeReals,
        bounds=bounds_var_u_imqk,
    )

    # input amplitude variables (only one per sign is needed, as vars. are nnr)

    model.var_u_amp_im = pyo.Var(model.set_IM_dim, within=pyo.NonNegativeReals)

    # *************************************************************************

    # outputs

    # output variables

    def bounds_var_y_irqk(m, i, r, q, k):
        if r in m.set_R_fix:
            # predefined limit
            return (m.param_u_lb_irqk[(i, r, q, k)], m.param_u_ub_irqk[(i, r, q, k)])
        else:
            # do not enforce any limits
            return (None, None)

    # def domain_var_y_irqk(m, i, r, k):
    #     try:
    #         if m_i in m.set_M_bin[i]:
    #             return pyo.Binary # binary: {0,1}
    #         else:
    #             return pyo.NonNegativeReals # nonnegative real: [0,inf]
    #     except KeyError:
    #         return pyo.NonNegativeReals # nonnegative real: [0,inf]

    model.var_y_irqk = pyo.Var(
        model.set_IR, model.set_QK, bounds=bounds_var_y_irqk, within=pyo.Reals
    )

    # positive output amplitudes

    model.var_y_amp_pos_ir = pyo.Var(model.set_IR_dim_pos, within=pyo.Reals)

    # output amplitudes

    model.var_y_amp_neg_ir = pyo.Var(model.set_IR_dim_neg, within=pyo.Reals)

    # *************************************************************************

    # states

    # state variables

    model.var_x_inqk = pyo.Var(model.set_IN, model.set_QK, within=pyo.Reals)

    # positive amplitude variables

    model.var_x_amp_pos_in = pyo.Var(model.set_IN_dim_pos, within=pyo.NonNegativeReals)

    # negative amplitude variables

    model.var_x_amp_neg_in = pyo.Var(model.set_IN_dim_neg, within=pyo.NonNegativeReals)

    # positive state variation

    model.var_delta_x_pos_var_in = pyo.Var(
        model.set_IN_pos_var, within=pyo.NonNegativeReals
    )

    # negative state variation

    model.var_delta_x_neg_var_in = pyo.Var(
        model.set_IN_neg_var, within=pyo.NonNegativeReals
    )

    # positive reference state violation

    model.var_delta_x_ref_u_inqk = pyo.Var(
        model.set_IN_ref_u, model.set_QK, within=pyo.NonNegativeReals
    )

    # negative reference state violation

    model.var_delta_x_ref_d_inqk = pyo.Var(
        model.set_IN_ref_d, model.set_QK, within=pyo.NonNegativeReals
    )

    # *************************************************************************
    # *************************************************************************
    # *************************************************************************
    # *************************************************************************

    # objective function

    # *************************************************************************
    # *************************************************************************

    # maximise npv

    def obj_f_rule(m):
        return (
            sum(
                (m.var_sdncf_q[q] + m.var_sdext_q[q]) * m.param_c_wgt_q[q]
                for q in m.set_Q
            )
            - m.var_capex
        )

    model.obj_f = pyo.Objective(rule=obj_f_rule, sense=pyo.maximize)

    # *************************************************************************
    # *************************************************************************
    # *************************************************************************
    # *************************************************************************

    # Constraints

    # *************************************************************************
    # *************************************************************************

    # objective function

    # *************************************************************************

    # opex

    # sum of discounted net cash flows

    def rule_sdncf_q(m, q):
        return (
            sum(
                m.param_c_df_qp[(q, p)]
                * sum(
                    m.param_c_time_qpk[(q, p, k)]
                    * sum(
                        sum(
                            m.var_efr_glqpk[(g, l_exp, q, p, k)]
                            for l_exp in m.set_L_exp[g]
                        )
                        - sum(
                            m.var_ifc_glqpk[(g, l_imp, q, p, k)]
                            for l_imp in m.set_L_imp[g]
                        )
                        for g in m.set_G
                    )
                    for k in m.set_K_q[q]
                )
                for p in m.set_P_q[q]
            )
            == m.var_sdncf_q[q]
        )

    model.constr_sdncf_q = pyo.Constraint(model.set_Q, rule=rule_sdncf_q)

    # *************************************************************************

    # sum of discounted externalities

    def rule_sdext_q(m, q):
        return (
            sum(
                m.param_c_df_qp[(q, p)]
                * (
                    sum(
                        m.param_c_time_qpk[(q, p, k)]
                        * sum(
                            m.var_u_imqk[(i, m_i, q, k)]
                            * m.param_c_ext_u_imqk[(i, m_i, q, k)]
                            for (i, m_i) in m.set_IM_ext
                        )
                        + sum(
                            m.var_x_inqk[(i, n_i, q, k)]
                            * m.param_c_ext_x_inqk[(i, n_i, q, k)]
                            for (i, n_i) in m.set_IN_ext
                        )
                        + sum(
                            m.var_y_irqk[(i, r_i, q, k)]
                            * m.param_c_ext_y_irqk[(i, r_i, q, k)]
                            for (i, r_i) in m.set_IR_ext
                        )
                        + sum(
                            m.var_delta_x_ref_d_inqk[(i, n_i, q, k)]
                            * m.param_c_ref_d_inqk[(i, n_i, q, k)]
                            for (i, n_i) in m.set_IN_ref_d
                        )
                        + sum(
                            m.var_delta_x_ref_u_inqk[(i, n_i, q, k)]
                            * m.param_c_ref_u_inqk[(i, n_i, q, k)]
                            for (i, n_i) in m.set_IN_ref_u
                        )
                        for k in m.set_K_q[q]
                    )
                    + sum(
                        m.param_c_pos_var_in[(i, n_i)]
                        * m.var_delta_x_pos_var_in[(i, n_i)]
                        for (i, n_i) in m.set_IN_pos_var
                    )
                    + sum(
                        m.param_c_neg_var_in[(i, n_i)]
                        * m.var_delta_x_neg_var_in[(i, n_i)]
                        for (i, n_i) in m.set_IN_neg_var
                    )
                )
                for p in m.set_P_q[q]
            )
            == m.var_sdext_q[q]
        )

    model.constr_sdext_q = pyo.Constraint(model.set_Q, rule=rule_sdext_q)

    # *************************************************************************

    # capex

    def rule_capex(m):
        return (
            sum(
                m.var_capex_arc_gllj[(g, l1, l2, j)]
                for (g, l1, l2, j) in m.set_GLLJ_sgl
            )
            + sum(m.var_capex_arc_col_t[t] for t in m.set_T)
            + sum(m.var_capex_cvt_i[i] for i in m.set_I_new)
            <= m.var_capex
        )

    model.constr_capex = pyo.Constraint(rule=rule_capex)

    # capex for arcs

    def rule_constr_capex_arcs(m, g, l1, l2, j):
        return (
            m.var_v_amp_gllj[(g, l1, l2, j)] * m.param_c_arc_var_gllj[(g, l1, l2, j)]
            + sum(
                m.var_delta_arc_inv_glljh[(g, l1, l2, j, h)]
                * m.param_c_arc_min_glljh[(g, l1, l2, j, h)]
                for h in m.set_H_gllj[(g, l1, l2, j)]
            )
            <= m.var_capex_arc_gllj[(g, l1, l2, j)]
        )

    model.constr_capex_arcs = pyo.Constraint(
        model.set_GLLJ_sgl, rule=rule_constr_capex_arcs  #
    )

    # capex for groups

    def rule_constr_capex_arcs_col(m, t):
        return (
            m.var_v_amp_t[t] * m.param_c_arc_var_t[t]
            + sum(
                m.var_delta_arc_inv_th[(t, h)] * m.param_c_arc_min_th[(t, h)]
                for h in m.set_H_t[t]
            )
            <= m.var_capex_arc_col_t[t]
        )

    model.constr_capex_arcs_col = pyo.Constraint(
        model.set_T, rule=rule_constr_capex_arcs_col
    )

    # capex for individual systems

    def rule_capex_converter(m, i):
        return (
            m.var_cvt_inv_i[i] * m.param_c_cvt_min_i[i]
            + sum(
                m.var_u_amp_im[(i, m_i)] * m.param_c_cvt_u_im[(i, m_i)]
                for m_i in m.set_M_dim_i[i]
            )
            + sum(
                m.var_x_amp_pos_in[(i, n)] * m.param_c_cvt_x_pos_in[(i, n)]
                for n in m.set_N_dim_pos[i]
            )
            + sum(
                m.var_x_amp_neg_in[(i, n)] * m.param_c_cvt_x_neg_in[(i, n)]
                for n in m.set_N_dim_neg[i]
            )
            + sum(
                m.var_y_amp_pos_ir[(i, r)] * m.param_c_cvt_y_pos_ir[(i, r)]
                for r in m.set_N_dim_pos[i]
            )
            + sum(
                m.var_y_amp_neg_ir[(i, r)] * m.param_c_cvt_y_neg_ir[(i, r)]
                for r in m.set_N_dim_neg[i]
            )
            <= m.var_capex_cvt_i[i]
        )

    model.constr_capex_system = pyo.Constraint(
        model.set_I_new, rule=rule_capex_converter
    )
    
    # prices
    add_price_functions(
        model,
        use_blocks=use_prices_block, 
        node_price_model=node_price_model
        )

    # *************************************************************************
    # *************************************************************************

    # network

    # *************************************************************************
    # *************************************************************************

    # flow equilibrium equation

    def rule_node_balance(m, g, l, q, k):
        return (
            m.param_v_base_glqk[(g, l, q, k)],
            # incoming arcs, default direction
            sum(
                sum(
                    m.param_eta_glljqk[(g, l2, l, j, q, k)]
                    * m.var_v_glljqk[(g, l2, l, j, q, k)]
                    for j in m.set_J[(g, l2, l)]
                )
                for l2 in m.set_L[g] - m.set_L_exp[g]
                if l != l2
            )
            +
            # incoming undirected arcs, reverse direction
            sum(
                sum(
                    m.param_eta_glljqk[(g, l2, l, j, q, k)]
                    * m.var_v_glljqk[(g, l2, l, j, q, k)]
                    for j in m.set_J_und[(g, l, l2)]
                )
                for l2 in m.set_L[g] - m.set_L_exp[g] - m.set_L_imp[g]
                if l != l2
            )
            -
            # outgoing directed arcs, default direction
            sum(
                sum(m.var_v_glljqk[(g, l, l2, j, q, k)] for j in m.set_J[(g, l, l2)])
                for l2 in m.set_L[g] - m.set_L_imp[g]
                if l != l2
            )
            -
            # outgoing directed arcs, reverse direction
            sum(
                sum(
                    m.var_v_glljqk[(g, l, l2, j, q, k)] for j in m.set_J_und[(g, l2, l)]
                )
                for l2 in m.set_L[g] - m.set_L_imp[g] - m.set_L_exp[g]
                if l != l2
            )
            # static losses placed downstream for new directed incoming arcs
            - sum(
                sum(
                    m.param_eta_glljqk[(g, l2, l, j, q, k)]
                    * (
                        m.var_w_glljqk[(g, l2, l, j, q, k)]
                        if j not in m.set_J_pre[(g, l2, l)]
                        else m.param_w_pre_glljqk[(g, l2, l, j, q, k)]
                    )
                    for j in m.set_J_stt_arr[(g, l2, l)]
                    if j not in m.set_J_und[(g, l2, l)]  # directed arcs only
                )
                for l2 in m.set_L[g] - m.set_L_exp[g]
                if l != l2
            )
            # static losses placed upstream for new directed outgoing arcs
            - sum(
                sum(
                    (
                        m.var_w_glljqk[(g, l, l2, j, q, k)]
                        if j not in m.set_J_pre[(g, l, l2)]
                        else m.param_w_pre_glljqk[(g, l, l2, j, q, k)]
                    )
                    for j in m.set_J_stt[(g, l, l2)]
                    if j not in m.set_J_stt_arr[(g, l, l2)]
                    if j not in m.set_J_und[(g, l, l2)]  # directed arcs only
                )
                for l2 in m.set_L[g] - m.set_L_imp[g]
                if l != l2
            )
            # static losses for undirected arcs, incoming (a and ds)
            # static losses for undirected arcs, outgoing (a and us)
            - sum(
                # nominal direction
                sum(
                    m.param_eta_glljqk[(g, l2, l, j, q, k)]
                    * m.var_w_sns_glljqk[(g, l2, l, j, q, k)]
                    for j in m.set_J_und[(g, l, l2)]
                    if (
                        j in m.set_J_stt_dep[(g, l, l2)]
                        or j in m.set_J_stt_ds[(g, l, l2)]
                    )
                )
                +
                # reverse direction
                sum(
                    m.var_w_sns_glljqk[(g, l, l2, j, q, k)]
                    for j in m.set_J_und[(g, l, l2)]
                    if (
                        j in m.set_J_stt_dep[(g, l, l2)]
                        or j in m.set_J_stt_us[(g, l, l2)]
                    )
                )
                for l2 in m.set_L[g] - m.set_L_exp[g] - m.set_L_imp[g]
                if l != l2
            )
            # static losses for undirected arcs, incoming (b and ds)
            # static losses for undirected arcs, outgoing (b and us)
            - sum(
                # nominal direction
                sum(
                    m.param_eta_glljqk[(g, l2, l, j, q, k)]
                    * m.var_w_sns_glljqk[(g, l2, l, j, q, k)]
                    for j in m.set_J_und[(g, l2, l)]
                    if (
                        j in m.set_J_stt_arr[(g, l2, l)]
                        or j in m.set_J_stt_ds[(g, l2, l)]
                    )
                )
                +
                # reverse direction
                sum(
                    m.var_w_sns_glljqk[(g, l, l2, j, q, k)]
                    for j in m.set_J_und[(g, l2, l)]
                    if (
                        j in m.set_J_stt_arr[(g, l2, l)]
                        or j in m.set_J_stt_us[(g, l2, l)]
                    )
                )
                for l2 in m.set_L[g] - m.set_L_exp[g] - m.set_L_imp[g]
                if l != l2
            )
            # converters
            + sum(
                m.var_y_irqk[(i, r, q, k)] * m.param_a_nw_glirqk[(g, l, i, r, q, k)]
                for (i, r) in m.set_IR
            )
            + sum(
                m.var_u_imqk[(i, m_i, q, k)] * m.param_a_nw_glimqk[(g, l, i, m_i, q, k)]
                for (i, m_i) in m.set_IM
            ),
            m.param_v_base_glqk[(g, l, q, k)],
        )

    model.constr_node_balance = pyo.Constraint(
        model.set_GL_not_exp_imp, model.set_QK, rule=rule_node_balance
    )

    # *************************************************************************
    # *************************************************************************
    
    add_network_restrictions(model)

    # *************************************************************************
    # *************************************************************************

    # arcs

    # *************************************************************************
    # *************************************************************************

    # maximum flow amplitude per time interval

    # nominal direction, directed and undirected arcs

    def rule_constr_arc_max_flow_time_nom(m, g, l1, l2, j, q, k):
        return m.var_v_glljqk[(g, l1, l2, j, q, k)] <= m.param_f_amp_v_glljqk[
            (g, l1, l2, j, q, k)
        ] * m.var_v_amp_gllj[(g, l1, l2, j)] - (
            m.var_w_glljqk[(g, l1, l2, j, q, k)]
            if (  # j in m.set_J_stt[(g,l1,l2)] and
                j in m.set_J_stt_dep[(g, l1, l2)] or j in m.set_J_stt_us[(g, l1, l2)]
            )
            else 0
        )

    model.constr_arc_max_flow_time_nom = pyo.Constraint(
        model.set_GLLJ_sgl,  # new directed and undirected arcs
        model.set_QK,  # once per time interval
        rule=rule_constr_arc_max_flow_time_nom,
    )

    # reverse direction, for undirected arcs

    def rule_constr_arc_max_flow_time_rev(m, g, l1, l2, j, q, k):
        if j in m.set_J_col[(g, l1, l2)]:
            # group arc # TODO: remove this part
            return pyo.Constraint.Skip
        else:  # individual arc
            return m.var_v_glljqk[(g, l2, l1, j, q, k)] <= m.param_f_amp_v_glljqk[
                (g, l1, l2, j, q, k)
            ] * m.var_v_amp_gllj[(g, l1, l2, j)] - (
                m.var_w_glljqk[(g, l1, l2, j, q, k)]
                if (  # j in m.set_J_stt[(g,l1,l2)] and
                    j in m.set_J_stt_arr[(g, l1, l2)]
                    or j in m.set_J_stt_us[(g, l1, l2)]
                )
                else 0
            )

    model.constr_arc_max_flow_time_rev = pyo.Constraint(
        model.set_GLLJ_und_new,  # new and undirected
        model.set_QK,  # once per time interval
        rule=rule_constr_arc_max_flow_time_rev,
    )

    # *************************************************************************

    # maximum nominal flow amplitude (applies to both directions)

    def rule_constr_max_nominal_flow_amplitude(m, g, l1, l2, j):
        return m.var_v_amp_gllj[(g, l1, l2, j)] <= sum(
            m.var_delta_arc_inv_glljh[(g, l1, l2, j, h)]
            * m.param_v_amp_max_glljh[(g, l1, l2, j, h)]
            for h in m.set_H_gllj[(g, l1, l2, j)]
        )

    model.constr_max_nominal_flow_amplitude = pyo.Constraint(
        model.set_GLLJ_sgl,  # directed and undirected arcs, new only
        rule=rule_constr_max_nominal_flow_amplitude,
    )

    # *************************************************************************

    # one option per arc
    # note: one constraint per (undirected) arc

    def rule_constr_one_arc_option_per_arc(m, g, l1, l2, j):
        if j in m.set_J_mdt[(g, l1, l2)]:
            return (
                sum(
                    m.var_delta_arc_inv_glljh[(g, l1, l2, j, h)]
                    for h in m.set_H_gllj[(g, l1, l2, j)]
                )
                == 1
            )  # one arc is mandatory
        elif j in m.set_J_int[(g, l1, l2)]:
            # not mandatory and interfaced´
            return m.var_xi_arc_inv_gllj[(g, l1, l2, j)] == sum(
                m.var_delta_arc_inv_glljh[(g, l1, l2, j, h)]
                for h in m.set_H_gllj[(g, l1, l2, j)]
            )
        else:  # optional and not interfaced
            return (
                sum(
                    m.var_delta_arc_inv_glljh[(g, l1, l2, j, h)]
                    for h in m.set_H_gllj[(g, l1, l2, j)]
                )
                <= 1
            )  # arcs are optional

    model.constr_one_arc_option_per_arc = pyo.Constraint(
        model.set_GLLJ_sgl,  # directed and undirected arcs, new only
        rule=rule_constr_one_arc_option_per_arc,
    )

    # *************************************************************************

    # # interface equations (unnecessary if the domain is UnitInterval)

    # def rule_constr_single_arc_interfaces(m, g, l1, l2, j):
    #     return (
    #         m.var_xi_arc_inv_gllj[(g,l1,l2,j)] <= 1
    #         )
    # model.constr_single_arc_interfaces = pyo.Constraint(
    #     model.set_GLLJ_int,
    #     rule=rule_constr_single_arc_interfaces)

    # *************************************************************************

    # SOS1 constraints for arc selection    
    def rule_constr_arc_sos1(m, g, l1, l2, j):
        var_list = [
            m.var_delta_arc_inv_glljh[(g, l1, l2, j, h)]
            for h in m.set_H_gllj[(g, l1, l2, j)]
            ]
        weight_list = [
            m.param_arc_inv_sos1_weights_glljh[(g, l1, l2, j, h)]
            for h in m.set_H_gllj[(g, l1, l2, j)]
            ]
        return (var_list, weight_list)
    model.constr_arc_sos1 = pyo.SOSConstraint(
        model.set_GLLJ_arc_inv_sos1,
        rule=rule_constr_arc_sos1,
        sos=1
    )

    # *************************************************************************

    # one flow direction per time interval in undirected preexisting arcs

    def rule_constr_one_sns_per_time_interval(m, g, l1, l2, j, q, k):
        if j in m.set_J_col[(g, l1, l2)]:
            return pyo.Constraint.Skip  # TODO:  remove this part
        if j in m.set_J_int[(g, l1, l2)]:
            # using interfaces
            return (
                m.var_zeta_sns_glljqk[(g, l1, l2, j, q, k)]
                + m.var_zeta_sns_glljqk[(g, l2, l1, j, q, k)]
                == m.var_xi_arc_inv_gllj[(g, l1, l2, j)]
            )
        elif j in m.set_J_mdt[(g, l1, l2)] or j in m.set_J_pre[(g, l1, l2)]:
            # mandatory or pre-existing
            return (
                m.var_zeta_sns_glljqk[(g, l1, l2, j, q, k)]
                + m.var_zeta_sns_glljqk[(g, l2, l1, j, q, k)]
                == 1
            )
        else:
            # optional
            return m.var_zeta_sns_glljqk[(g, l1, l2, j, q, k)] + m.var_zeta_sns_glljqk[
                (g, l2, l1, j, q, k)
            ] == sum(
                m.var_delta_arc_inv_glljh[(g, l1, l2, j, h)]
                for h in m.set_H_gllj[(g, l1, l2, j)]
            )

    model.constr_one_sns_per_time_interval = pyo.Constraint(
        model.set_GLLJ_und,  # once per undirected arc
        model.set_QK,  # once per time interval
        rule=rule_constr_one_sns_per_time_interval,
    )

    # *************************************************************************

    # no flow except in the flow direction, for new undirected arcs

    # nominal direction

    def rule_constr_no_flow_except_in_sns_nom(m, g, l1, l2, j, q, k):
        if j in m.set_J_col[(g, l1, l2)]:
            return pyo.Constraint.Skip  # TODO: remove this
        if j in m.set_J_pre[(g, l1, l2)]:  # pre-existing arc
            return (
                m.var_v_glljqk[(g, l1, l2, j, q, k)]
                <= m.var_zeta_sns_glljqk[(g, l1, l2, j, q, k)]
                * m.param_v_ub_glljqk[(g, l1, l2, j, q, k)]
            )
        else:  # new arc
            return m.var_v_glljqk[(g, l1, l2, j, q, k)] <= m.var_zeta_sns_glljqk[
                (g, l1, l2, j, q, k)
            ] * (
                max(
                    m.param_f_amp_v_glljqk[(g, l1, l2, j, q, k)]
                    * m.param_v_amp_max_glljh[(g, l1, l2, j, h)]
                    - (
                        m.param_w_new_glljhqk[(g, l1, l2, j, h, q, k)]
                        if (
                            (
                                (g, l1, l2) in m.set_J_stt_dep
                                and j in m.set_J_stt_dep[(g, l1, l2)]
                            )
                            or (
                                (g, l1, l2) in m.set_J_stt_us
                                and j in m.set_J_stt_us[(g, l1, l2)]
                            )
                        )
                        else 0
                    )
                    for h in m.set_H_gllj[(g, l1, l2, j)]
                )
            )

    model.constr_no_flow_except_in_sns_nom = pyo.Constraint(
        model.set_GLLJ_und,  # new undirected arcs
        model.set_QK,  # once per time interval
        rule=rule_constr_no_flow_except_in_sns_nom,
    )

    # reverse direction

    def rule_constr_no_flow_except_in_sns_rev(m, g, l1, l2, j, q, k):
        if j in m.set_J_col[(g, l1, l2)]:
            return pyo.Constraint.Skip  # TODO: remove this
        if j in m.set_J_pre[(g, l1, l2)]:  # pre-existing arc
            return (
                m.var_v_glljqk[(g, l2, l1, j, q, k)]
                <= m.var_zeta_sns_glljqk[(g, l2, l1, j, q, k)]
                * m.param_v_ub_glljqk[(g, l2, l1, j, q, k)]
            )
        else:  # new arc
            return m.var_v_glljqk[(g, l2, l1, j, q, k)] <= m.var_zeta_sns_glljqk[
                (g, l2, l1, j, q, k)
            ] * (
                max(
                    m.param_f_amp_v_glljqk[(g, l1, l2, j, q, k)]
                    * m.param_v_amp_max_glljh[(g, l1, l2, j, h)]
                    - (
                        m.param_w_new_glljhqk[(g, l1, l2, j, h, q, k)]
                        if (
                            (
                                (g, l1, l2) in m.set_J_stt_arr
                                and j in m.set_J_stt_arr[(g, l1, l2)]
                            )
                            or (
                                (g, l1, l2) in m.set_J_stt_us
                                and j in m.set_J_stt_us[(g, l1, l2)]
                            )
                        )
                        else 0
                    )
                    for h in m.set_H_gllj[(g, l1, l2, j)]
                )
            )

    model.constr_no_flow_except_in_sns_rev = pyo.Constraint(
        model.set_GLLJ_und,  # new undirected arcs
        model.set_QK,  # once per time interval
        rule=rule_constr_no_flow_except_in_sns_rev,
    )

    # *************************************************************************

    # SOS1 constraints for flow sense determination (undirected arcs)
    # model.constr_sns_sos1 = pyo.SOSConstraint(
    #     model.set_GLLJ_und,  # one constraint per undirected arc
    #     model.set_QK,  # and time interval
    #     var=model.var_zeta_sns_glljqk,  # set_GLLJ_und_red and set_K
    #     index=model.set_GLLJQK_und_sns_sos1_red_gllj,  #
    #     weights=model.param_arc_sns_sos1_weights_glljqk,
    #     sos=1,
    # )
    # TODO: make the weight list dependent in model options rather than literal
    def rule_constr_sns_sos1(m, g, l1, l2, j, q, k):
        var_list = [
            m.var_zeta_sns_glljqk[(g, l1, l2, j, q, k)],
            m.var_zeta_sns_glljqk[(g, l2, l1, j, q, k)]
            ]
        weight_list = [1, 2] if True else [2, 1]
        return (var_list, weight_list)
    model.constr_sns_sos1 = pyo.SOSConstraint(
        model.set_GLLJ_und,
        model.set_QK,
        rule=rule_constr_sns_sos1,
        sos=1
    )
    
    # *************************************************************************

    # static losses

    # static losses only exist if the arc exists

    def rule_constr_static_losses_existence(m, g, l1, l2, j, q, k):
        if j in m.set_J_col[(g, l1, l2)]:  # group
            return m.var_w_glljqk[(g, l1, l2, j, q, k)] == sum(
                m.var_delta_arc_inv_th[(t, h)]
                * m.param_w_new_glljhqk[(g, l1, l2, j, h, q, k)]
                for t in m.set_T
                if (g, l1, l2, j) in m.set_GLLJ_col_t[t]
                for h in m.set_H_t[t]
            )
        else:  # individual
            return m.var_w_glljqk[(g, l1, l2, j, q, k)] == sum(
                m.var_delta_arc_inv_glljh[(g, l1, l2, j, h)]
                * m.param_w_new_glljhqk[(g, l1, l2, j, h, q, k)]
                for h in m.set_H_gllj[(g, l1, l2, j)]
            )

    model.constr_static_losses_existence = pyo.Constraint(
        model.set_GLLJ_static_new,
        model.set_QK,
        rule=rule_constr_static_losses_existence,
    )

    # *************************************************************************

    # static losses placed downstream

    # nominal direction

    def rule_constr_static_losses_downstream_nom(m, g, l1, l2, j, q, k):
        if (
            j not in m.set_J_stt_arr[(g, l1, l2)]
            and j not in m.set_J_stt_ds[(g, l1, l2)]
        ):
            return pyo.Constraint.Skip

        if j in m.set_J_und[(g, l1, l2)]:
            # undirected arc

            # TODO: adjust if losses for pre. undirected arcs are handled diff.

            # arrival node or downstream

            return (
                m.var_v_glljqk[(g, l1, l2, j, q, k)]
                >= m.var_w_sns_glljqk[(g, l1, l2, j, q, k)]
            )

        else:  # directed arc, arrival
            if j in m.set_J_stt_ds[(g, l1, l2)]:
                raise ValueError(
                    "Static losses for directed arcs should be mo"
                    + "delled as being in the departure or in the"
                    + " arrival nodes."
                )

            if j in m.set_J_pre[(g, l1, l2)]:
                # pre-existing arc

                return (
                    m.var_v_glljqk[(g, l1, l2, j, q, k)]
                    >= m.param_w_pre_glljqk[(g, l1, l2, j, q, k)]
                )

            else:
                # new arc

                return (
                    m.var_v_glljqk[(g, l1, l2, j, q, k)]
                    >= m.var_w_glljqk[(g, l1, l2, j, q, k)]
                )

    model.constr_static_losses_downstream_nom = pyo.Constraint(
        model.set_GLLJ_static,
        model.set_QK,
        rule=rule_constr_static_losses_downstream_nom,
    )

    # reverse direction

    def rule_constr_static_losses_downstream_und_rev(m, g, l1, l2, j, q, k):
        if j in m.set_J_stt_ds[(g, l2, l1)] or j in m.set_J_stt_dep[(g, l2, l1)]:
            # departure node or downstream

            return (
                m.var_v_glljqk[(g, l2, l1, j, q, k)]
                >= m.var_w_sns_glljqk[(g, l2, l1, j, q, k)]
            )

        else:
            return pyo.Constraint.Skip

    model.constr_static_losses_downstream_und_rev = pyo.Constraint(
        model.set_GLLJ_static_und,
        model.set_QK,
        rule=rule_constr_static_losses_downstream_und_rev,
    )

    # *************************************************************************

    # static losses modulated by flow sense for pre-existing arcs

    # TODO: replace this constraint by reformulating the node balance

    def rule_constr_static_losses_sense_pre_nom(m, g, l1, l2, j, q, k):
        return (
            m.var_w_sns_glljqk[(g, l1, l2, j, q, k)]
            == m.var_zeta_sns_glljqk[(g, l1, l2, j, q, k)]
            * m.param_w_pre_glljqk[(g, l1, l2, j, q, k)]
        )

    model.constr_static_losses_sense_pre_nom = pyo.Constraint(
        model.set_GLLJ_static_und_pre,
        model.set_QK,
        rule=rule_constr_static_losses_sense_pre_nom,
    )

    def rule_constr_static_losses_sense_pre_rev(m, g, l1, l2, j, q, k):
        return (
            m.var_w_sns_glljqk[(g, l2, l1, j, q, k)]
            == m.var_zeta_sns_glljqk[(g, l2, l1, j, q, k)]
            * m.param_w_pre_glljqk[(g, l1, l2, j, q, k)]
        )

    model.constr_static_losses_sense_pre_rev = pyo.Constraint(
        model.set_GLLJ_static_und_pre,
        model.set_QK,
        rule=rule_constr_static_losses_sense_pre_rev,
    )

    # static losses modulated by the flow sense for new undirected arcs

    # general upper bound

    def rule_constr_static_losses_sense_new_ub_nom(m, g, l1, l2, j, q, k):
        if j in m.set_J_col[(g, l1, l2)]:
            return (
                m.var_w_sns_glljqk[(g, l1, l2, j, q, k)]
                <= max(
                    m.param_w_new_glljhqk[(g, l1, l2, j, h, q, k)]
                    for t, h in m.set_TH
                    if (g, l1, l2, j) in m.set_GLLJ_col_t[t]
                )
                * m.var_zeta_sns_glljqk[(g, l1, l2, j, q, k)]
            )

        else:
            return (
                m.var_w_sns_glljqk[(g, l1, l2, j, q, k)]
                <= max(
                    m.param_w_new_glljhqk[(g, l1, l2, j, h, q, k)]
                    for h in m.set_H_gllj[(g, l1, l2, j)]
                )
                * m.var_zeta_sns_glljqk[(g, l1, l2, j, q, k)]
            )

    model.constr_static_losses_sense_new_ub_nom = pyo.Constraint(
        model.set_GLLJ_static_und_new,
        model.set_QK,
        rule=rule_constr_static_losses_sense_new_ub_nom,
    )

    def rule_constr_static_losses_sense_new_ub_rev(m, g, l1, l2, j, q, k):
        if j in m.set_J_col[(g, l1, l2)]:
            # group arc
            return (
                m.var_w_sns_glljqk[(g, l2, l1, j, q, k)]
                <= max(
                    m.param_w_new_glljhqk[(g, l1, l2, j, h, q, k)]
                    for t, h in m.set_TH
                    if (g, l1, l2, j) in m.set_GLLJ_col_t[t]
                )
                * m.var_zeta_sns_glljqk[(g, l2, l1, j, q, k)]
            )
        else:  # individual arc
            return (
                m.var_w_sns_glljqk[(g, l2, l1, j, q, k)]
                <= max(
                    m.param_w_new_glljhqk[(g, l1, l2, j, h, q, k)]
                    for h in m.set_H_gllj[(g, l1, l2, j)]
                )
                * m.var_zeta_sns_glljqk[(g, l2, l1, j, q, k)]
            )

    model.constr_static_losses_sense_new_ub_rev = pyo.Constraint(
        model.set_GLLJ_static_und_new,
        model.set_QK,
        rule=rule_constr_static_losses_sense_new_ub_rev,
    )

    # equation, part 1

    def rule_constr_static_losses_sense_new_eq1_nom(m, g, l1, l2, j, q, k):
        return (
            m.var_w_sns_glljqk[(g, l1, l2, j, q, k)]
            <= m.var_w_glljqk[(g, l1, l2, j, q, k)]
        )

    model.constr_static_losses_sense_new_eq1_nom = pyo.Constraint(
        model.set_GLLJ_static_und_new,
        model.set_QK,
        rule=rule_constr_static_losses_sense_new_eq1_nom,
    )

    def rule_constr_static_losses_sense_new_eq1_rev(m, g, l1, l2, j, q, k):
        return (
            m.var_w_sns_glljqk[(g, l2, l1, j, q, k)]
            <= m.var_w_glljqk[(g, l1, l2, j, q, k)]
        )

    model.constr_static_losses_sense_new_eq1_rev = pyo.Constraint(
        model.set_GLLJ_static_und_new,
        model.set_QK,
        rule=rule_constr_static_losses_sense_new_eq1_rev,
    )

    # equation, part 2

    def rule_constr_static_losses_sense_new_eq2_nom(m, g, l1, l2, j, q, k):
        if j in m.set_J_col[(g, l1, l2)]:
            # arc group
            return m.var_w_sns_glljqk[(g, l1, l2, j, q, k)] >= m.var_w_glljqk[
                (g, l1, l2, j, q, k)
            ] - m.var_zeta_sns_glljqk[(g, l2, l1, j, q, k)] * max(
                m.param_w_new_glljhqk[(g, l1, l2, j, h, q, k)]
                for t, h in m.set_TH
                if (g, l1, l2, j) in m.set_GLLJ_col_t[t]
            )
        else:  # individual
            return m.var_w_sns_glljqk[(g, l1, l2, j, q, k)] >= m.var_w_glljqk[
                (g, l1, l2, j, q, k)
            ] - m.var_zeta_sns_glljqk[(g, l2, l1, j, q, k)] * max(
                m.param_w_new_glljhqk[(g, l1, l2, j, h, q, k)]
                for h in m.set_H_gllj[(g, l1, l2, j)]
            )

    model.constr_static_losses_sense_new_eq2_nom = pyo.Constraint(
        model.set_GLLJ_static_und_new,
        model.set_QK,
        rule=rule_constr_static_losses_sense_new_eq2_nom,
    )

    def rule_constr_static_losses_sense_new_eq2_rev(m, g, l1, l2, j, q, k):
        if j in m.set_J_col[(g, l1, l2)]:
            # arc group
            return m.var_w_sns_glljqk[(g, l2, l1, j, q, k)] >= m.var_w_glljqk[
                (g, l1, l2, j, q, k)
            ] - m.var_zeta_sns_glljqk[(g, l1, l2, j, q, k)] * max(
                m.param_w_new_glljhqk[(g, l1, l2, j, h, q, k)]
                for t, h in m.set_TH
                if (g, l1, l2, j) in m.set_GLLJ_col_t[t]
            )
        else:  # individual group
            return m.var_w_sns_glljqk[(g, l2, l1, j, q, k)] >= m.var_w_glljqk[
                (g, l1, l2, j, q, k)
            ] - m.var_zeta_sns_glljqk[(g, l1, l2, j, q, k)] * max(
                m.param_w_new_glljhqk[(g, l1, l2, j, h, q, k)]
                for h in m.set_H_gllj[(g, l1, l2, j)]
            )

    model.constr_static_losses_sense_new_eq2_rev = pyo.Constraint(
        model.set_GLLJ_static_und_new,
        model.set_QK,
        rule=rule_constr_static_losses_sense_new_eq2_rev,
    )

    # *************************************************************************
    # *************************************************************************

    # groups of arcs

    # *************************************************************************
    # *************************************************************************

    # maximum flow amplitude per time interval, for arcs in arc groups

    # nominal direction, directed and undirected arcs

    def rule_constr_arc_group_max_flow_time_nom(m, t, g, l1, l2, j, q, k):
        return m.var_v_glljqk[(g, l1, l2, j, q, k)] <= m.param_f_amp_v_glljqk[
            (g, l1, l2, j, q, k)
        ] * m.var_v_amp_t[t] - (
            m.var_w_glljqk[(g, l1, l2, j, q, k)]
            if (  # j in m.set_J_stt[(g,l1,l2)] and
                j in m.set_J_stt_dep[(g, l1, l2)] or j in m.set_J_stt_us[(g, l1, l2)]
            )
            else 0
        )

    model.constr_arc_group_max_flow_time_nom = pyo.Constraint(
        model.set_TGLLJ,  # new directed and undirected arcs
        model.set_QK,  # all time intervals
        rule=rule_constr_arc_group_max_flow_time_nom,
    )

    # reverse direction, for undirected arcs

    def rule_constr_arc_group_max_flow_time_rev(m, t, g, l1, l2, j, q, k):
        if j in m.set_J_und[(g, l1, l2)]:
            return m.var_v_glljqk[(g, l2, l1, j, q, k)] <= m.param_f_amp_v_glljqk[
                (g, l1, l2, j, q, k)
            ] * m.var_v_amp_t[t] - (
                m.var_w_glljqk[(g, l1, l2, j, q, k)]
                if (  # j in m.set_J_stt[(g,l1,l2)] and
                    j in m.set_J_stt_arr[(g, l1, l2)]
                    or j in m.set_J_stt_us[(g, l1, l2)]
                )
                else 0
            )
        else:
            return pyo.Constraint.Skip

    model.constr_arc_group_max_flow_time_rev = pyo.Constraint(
        model.set_TGLLJ,  # new and undirected
        model.set_QK,  # all time intervals
        rule=rule_constr_arc_group_max_flow_time_rev,
    )

    # *************************************************************************

    # maximum nominal flow amplitude for arc groups

    def rule_constr_arc_group_max_nominal_flow_amplitude(m, t):
        return m.var_v_amp_t[t] <= sum(
            m.var_delta_arc_inv_th[(t, h)] * m.param_v_amp_max_th[(t, h)]
            for h in m.set_H_t[t]
        )

    model.constr_arc_group_max_nominal_flow_amplitude = pyo.Constraint(
        model.set_T,  # all arc groups
        rule=rule_constr_arc_group_max_nominal_flow_amplitude,
    )

    # *************************************************************************

    # one option per arc group

    def rule_constr_one_arc_option_per_arc_group(m, t):
        if t in m.set_T_mdt:
            return (
                sum(m.var_delta_arc_inv_th[(t, h)] for h in m.set_H_t[t]) == 1
            )  # the arcs are all mandatory
        else:
            return (
                sum(m.var_delta_arc_inv_th[(t, h)] for h in m.set_H_t[t]) <= 1
            )  # the arcs are all optional

    model.constr_one_arc_option_per_arc_group = pyo.Constraint(
        model.set_T, rule=rule_constr_one_arc_option_per_arc_group  # all arc groups
    )

    # *************************************************************************

    # SOS1 constraints for arc group selection
    def rule_constr_arc_group_sos1(m, t):
        var_list = [
            m.var_delta_arc_inv_th[(t, h)]
            for h in m.set_H_t[t]
            ]
        weight_list = [
            m.param_arc_inv_sos1_weights_th[(t, h)]
            for h in m.set_H_t[t]
            ]
        return (var_list, weight_list)
    model.constr_arc_group_sos1 = pyo.SOSConstraint(
        model.set_T_sos1,
        rule=rule_constr_arc_group_sos1,
        sos=1
    )

    # *************************************************************************

    # one flow direction per time interval, for undirected arcs within groups

    def rule_constr_one_sns_per_time_interval_group(m, t, g, l1, l2, j, q, k):
        if j in m.set_J_int[(g, l1, l2)]:
            # using interface
            return (
                m.var_zeta_sns_glljqk[(g, l1, l2, j, q, k)]
                + m.var_zeta_sns_glljqk[(g, l2, l1, j, q, k)]
                == m.var_xi_arc_inv_t[t]
            )
        elif j in m.set_J_mdt[(g, l1, l2)]:
            # mandatory group
            return (
                m.var_zeta_sns_glljqk[(g, l1, l2, j, q, k)]
                + m.var_zeta_sns_glljqk[(g, l2, l1, j, q, k)]
                == 1
            )
        else:
            # optional and not using interface
            return m.var_zeta_sns_glljqk[(g, l1, l2, j, q, k)] + m.var_zeta_sns_glljqk[
                (g, l2, l1, j, q, k)
            ] == sum(m.var_delta_arc_inv_th[(t, h)] for h in m.set_H_t[t])

    model.constr_one_sns_per_time_interval_group = pyo.Constraint(
        model.set_TGLLJ_und,  # once per undirected arc
        model.set_QK,  # once per time interval
        rule=rule_constr_one_sns_per_time_interval_group,
    )

    # *************************************************************************

    # no flow except in the flow direction, for new undirected arcs in groups

    # nominal direction

    def rule_constr_no_flow_except_in_sns_group_nom(m, t, g, l1, l2, j, q, k):
        return m.var_v_glljqk[(g, l1, l2, j, q, k)] <= m.var_zeta_sns_glljqk[
            (g, l1, l2, j, q, k)
        ] * (
            max(
                m.param_f_amp_v_glljqk[(g, l1, l2, j, q, k)]
                * m.param_v_amp_max_th[(t, h)]
                - (
                    m.param_w_new_glljhqk[(g, l1, l2, j, h, q, k)]
                    if (
                        (
                            (g, l1, l2) in m.set_J_stt_dep
                            and j in m.set_J_stt_dep[(g, l1, l2)]
                        )
                        or (
                            (g, l1, l2) in m.set_J_stt_us
                            and j in m.set_J_stt_us[(g, l1, l2)]
                        )
                    )
                    else 0
                )
                for h in m.set_H_t[t]
            )
        )

    model.constr_no_flow_except_in_sns_group_nom = pyo.Constraint(
        model.set_TGLLJ_und,  # new undirected arcs in groups
        model.set_QK,  # once per time interval
        rule=rule_constr_no_flow_except_in_sns_group_nom,
    )

    # reverse direction

    def rule_constr_no_flow_except_in_sns_group_rev(m, t, g, l1, l2, j, q, k):
        return m.var_v_glljqk[(g, l2, l1, j, q, k)] <= m.var_zeta_sns_glljqk[
            (g, l2, l1, j, q, k)
        ] * (
            max(
                m.param_f_amp_v_glljqk[(g, l1, l2, j, q, k)]
                * m.param_v_amp_max_th[(t, h)]
                - (
                    m.param_w_new_glljhqk[(g, l1, l2, j, h, q, k)]
                    if (
                        (
                            (g, l1, l2) in m.set_J_stt_arr
                            and j in m.set_J_stt_arr[(g, l1, l2)]
                        )
                        or (
                            (g, l1, l2) in m.set_J_stt_us
                            and j in m.set_J_stt_us[(g, l1, l2)]
                        )
                    )
                    else 0
                )
                for h in m.set_H_t[t]
            )
        )

    model.constr_no_flow_except_in_sns_group_rev = pyo.Constraint(
        model.set_TGLLJ_und,  # new undirected arcs in groups
        model.set_QK,  # once per time interval
        rule=rule_constr_no_flow_except_in_sns_group_rev,
    )

    # *************************************************************************

    # interface equations for arc groups

    def rule_constr_arc_group_interfaces(m, t):
        return m.var_xi_arc_inv_t[t] == sum(
            m.var_delta_arc_inv_th[(t, h)] for h in m.set_H_t[t]
        )

    model.constr_arc_group_interfaces = pyo.Constraint(
        model.set_T_int, rule=rule_constr_arc_group_interfaces
    )

    # *************************************************************************
    # *************************************************************************

    # converters

    # *************************************************************************
    # *************************************************************************

    # input signal limits for dimensionable inputs

    def rule_constr_u_limit_dim(m, i, m_i, q, k):
        return (
            m.var_u_imqk[(i, m_i, q, k)]
            <= m.var_u_amp_im[(i, m_i)] * m.param_f_amp_u_imqk[(i, m_i, q, k)]
        )

    model.constr_u_limit_dim = pyo.Constraint(
        model.set_IM_dim, model.set_QK, rule=rule_constr_u_limit_dim
    )

    # nominal input amplitude limit for dimensionable inputs

    def rule_constr_u_amp_ub(m, i, m_i):
        return (
            m.var_u_amp_im[(i, m_i)]
            <= m.var_cvt_inv_i[i] * m.param_u_amp_max_im[(i, m_i)]
        )

    model.constr_u_amp_ub = pyo.Constraint(model.set_IM_dim, rule=rule_constr_u_amp_ub)

    # fixed upper limits

    def rule_constr_u_fix_limits(m, i, m_i, q, k):
        # if we need to know the lim input signal (e.g., for the obj. func.)

        if i in m.set_I_new:
            # new converter

            return (
                m.var_u_imqk[(i, m_i, q, k)]
                <= m.param_u_ub_imqk[(i, m_i, q, k)] * m.var_cvt_inv_i[i]
            )

            return m.var_u_imqk[(i, m_i, q, k)] <= m.var_cvt_inv_i[i]

        else:
            # pre-existing

            return m.var_u_imqk[(i, m_i, q, k)] <= m.param_u_ub_imqk[(i, m_i, q, k)]

    model.constr_u_fix_limits = pyo.Constraint(
        model.set_IM_fix, model.set_QK, rule=rule_constr_u_fix_limits
    )

    # input limits for binary inputs

    def rule_constr_u_bin_limits(m, i, m_i, q, k):
        if i in m.set_I_new:
            # binary variables

            return m.var_u_imqk[(i, m_i, q, k)] <= m.var_cvt_inv_i[i]

        else:
            return pyo.Constraint.Skip

    model.constr_u_bin_limits = pyo.Constraint(
        model.set_IM_bin, model.set_QK, rule=rule_constr_u_bin_limits
    )

    # *************************************************************************

    # outputs

    # output equations

    def rule_constr_output_equations(m, i, r, q, k):
        return (
            m.var_y_irqk[(i, r, k)]
            == sum(
                m.param_c_eq_y_irnqk[(i, r, n_i, q, k)] * m.var_x_inqk[(i, n_i, q, k)]
                for n_i in m.set_N[i]
            )
            + sum(
                m.param_d_eq_y_irmqk[(i, r, m_i, q, k)] * m.var_u_imqk[(i, m_i, q, k)]
                for m_i in m.set_M[i]
            )
            + m.param_e_eq_y_irqk[(i, r, q, k)]
        )

    model.constr_output_equations = pyo.Constraint(
        model.set_IR, model.set_QK, rule=rule_constr_output_equations
    )

    # positive amplitude limit for output variables

    def rule_constr_y_vars_have_pos_amp_limits(m, i, r, q, k):
        return m.var_y_irqk[(i, r, q, k)] <= (
            m.var_y_amp_pos_ir[(i, r)] * m.param_f_amp_y_pos_irqk[(i, r, q, k)]
        )

    model.constr_y_vars_have_pos_amp_limits = pyo.Constraint(
        model.set_IR_dim_pos, model.set_QK, rule=rule_constr_y_vars_have_pos_amp_limits
    )

    # negative amplitude limit for output variables

    def rule_constr_y_vars_have_neg_amp_limits(m, i, r, q, k):
        return m.var_y_irqk[(i, r, q, k)] >= (
            -m.var_y_amp_neg_ir[(i, r)] * m.param_f_amp_y_neg_irqk[(i, r, q, k)]
        )

    model.constr_y_vars_have_neg_amp_limits = pyo.Constraint(
        model.set_IR_dim_neg, model.set_QK, rule=rule_constr_y_vars_have_neg_amp_limits
    )

    # positive amplitude limit must be zero unless the system is installed

    def rule_constr_y_amp_pos_zero_if_cvt_not_selected(m, i, r):
        return m.var_y_amp_pos_ir[(i, r)] <= (
            m.var_cvt_inv_i[i] * m.param_y_amp_pos_ir[(i, r)]
        )

    model.constr_y_amp_pos_zero_if_cvt_not_newected = pyo.Constraint(
        model.set_IR_dim_pos, rule=rule_constr_y_amp_pos_zero_if_cvt_not_selected
    )

    # negative amplitude limit must be zero unless the system is installed

    def rule_constr_y_amp_neg_zero_if_cvt_not_selected(m, i, r):
        return m.var_y_amp_neg_ir[(i, r)] <= (
            m.var_cvt_inv_i[i] * m.param_y_amp_neg_ir[(i, r)]
        )

    model.constr_y_amp_neg_zero_if_cvt_not_selected = pyo.Constraint(
        model.set_IR_dim_neg, rule=rule_constr_y_amp_neg_zero_if_cvt_not_selected
    )

    # the positive and negative amplitudes must match

    def rule_constr_y_amp_pos_neg_match(m, i, r):
        return m.var_y_amp_pos_ir[(i, r)] == m.var_y_amp_neg_ir[(i, r)]

    model.constr_y_amp_pos_neg_match = pyo.Constraint(
        model.set_IR_dim_eq, rule=rule_constr_y_amp_pos_neg_match
    )

    # *************************************************************************

    # states

    def rule_constr_state_equations(m, i, n, q, k):
        return (
            m.var_x_inqk[(i, n, q, k)]
            == sum(
                m.param_a_eq_x_innqk[(i, n, n_star, q, k)]
                * (
                    m.var_x_inqk[(i, n_star, q, k - 1)]
                    if k != 0
                    else m.param_x_inq0[(i, n, q)]
                )
                for n_star in m.set_N[i]
            )
            + sum(
                m.param_b_eq_x_inmqk[(i, n, m_i, q, k)] * m.var_u_imqk[(i, m_i, q, k)]
                for m_i in m.set_M[i]
            )
            + m.param_e_eq_x_inqk[(i, n, q, k)]
        )

    model.constr_state_equations = pyo.Constraint(
        model.set_IN, model.set_QK, rule=rule_constr_state_equations
    )

    # positive amplitude limit for state variables

    def rule_constr_x_vars_have_pos_amp_limits(m, i, n, q, k):
        return m.var_x_inqk[(i, n, q, k)] <= (
            m.var_x_amp_pos_in[(i, n)] * m.param_f_amp_x_pos_inqk[(i, n, q, k)]
        )

    model.constr_x_vars_have_pos_amp_limits = pyo.Constraint(
        model.set_IN_dim_pos, model.set_QK, rule=rule_constr_x_vars_have_pos_amp_limits
    )

    # negative amplitude limit for state variables

    def rule_constr_x_vars_have_neg_amp_limits(m, i, n, q, k):
        return m.var_x_inqk[(i, n, q, k)] >= (
            -m.var_y_amp_neg_in[(i, n)] * m.param_f_amp_x_neg_inqk[(i, n, q, k)]
        )

    model.constr_x_vars_have_neg_amp_limits = pyo.Constraint(
        model.set_IN_dim_neg, model.set_QK, rule=rule_constr_x_vars_have_neg_amp_limits
    )

    # positive amplitude limit must be zero unless the system is installed

    def rule_constr_x_amp_pos_zero_if_cvt_not_selected(m, i, n):
        return m.var_x_amp_pos_in[(i, n)] <= (
            m.var_cvt_inv_i[i] * m.param_x_amp_pos_in[(i, n)]
        )

    model.constr_x_amp_pos_zero_if_cvt_not_selected = pyo.Constraint(
        model.set_IN_dim_pos, rule=rule_constr_x_amp_pos_zero_if_cvt_not_selected
    )

    # negative amplitude limit must be zero unless the system is installed

    def rule_constr_x_amp_neg_zero_if_cvt_not_selected(m, i, n):
        return m.var_x_amp_neg_in[(i, n)] <= (
            m.var_cvt_inv_i[i] * m.param_x_amp_neg_in[(i, n)]
        )

    model.constr_x_amp_neg_zero_if_cvt_not_selected = pyo.Constraint(
        model.set_IN_dim_neg, rule=rule_constr_x_amp_neg_zero_if_cvt_not_selected
    )

    # the positive and negative amplitudes must match

    def rule_constr_x_amp_pos_neg_match(m, i, n):
        return m.var_x_amp_pos_in[(i, n)] == m.var_x_amp_neg_in[(i, n)]

    model.constr_x_amp_pos_neg_match = pyo.Constraint(
        model.set_IN_dim_eq, rule=rule_constr_x_amp_pos_neg_match
    )

    # *************************************************************************
    # *************************************************************************

    return model

    # *************************************************************************
    # *************************************************************************


# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
