# imports

# standard libraries

from numbers import Real

import uuid

# import math

from math import inf

# local libraries, external

# import numpy as np

import networkx as nx
from ...data.gis.identify import get_edges_involving_node
from ...data.gis.identify import find_unconnected_nodes
# from .resource import are_prices_time_invariant, ResourcePrice

# *****************************************************************************
# *****************************************************************************

class Arcs:
    """A class for arc technologies in a network."""

    def __init__(
        self,
        name: str,  # human-readable name
        efficiency: dict,  # keys: (q, k)
        efficiency_reverse: dict,  # keys: (q, k)
        static_loss: dict,  # keys: (h, q, k)
        capacity: tuple,  # one value per option (h)
        minimum_cost: tuple,  # one value per option (h)
        specific_capacity_cost: float,  # one value
        capacity_is_instantaneous: bool,  # one value
        validate: bool,
    ):  # one number
        # initialise
        self.name = name
        self.efficiency = efficiency
        self.efficiency_reverse = efficiency_reverse
        self.static_loss = static_loss
        self.capacity = capacity
        self.minimum_cost = minimum_cost
        self.specific_capacity_cost = specific_capacity_cost
        self.capacity_is_instantaneous = capacity_is_instantaneous

        # results (for simulation or post-optimisation analysis)
        self.options_selected = [False for element in self.capacity]

        # validate
        if validate:
            Arcs.validate(self)

    # *************************************************************************
    # *************************************************************************

    def number_options(self):
        "Return the number of arc options."

        return len(self.minimum_cost)

    # *************************************************************************
    # *************************************************************************

    def is_isotropic(self, reverse_none_means_isotropic: bool = True) -> bool:
        """Returns True if the efficiency values do not change with sense."""

        # both efficiency and efficiency_reverse are None

        if type(self.efficiency) == type(None):
            # the efficiency for the nominal sense is None
            # this means it is 1 every time and that the efficiency for the
            # reverse sense can only be 1 too (efficiency_reverse == None)

            return True  # does not change with the flow sense

        elif type(self.efficiency_reverse) == type(None):
            # the efficiency for the nominal sense is not None
            # the efficiency for the reverse sense is None:

            # two possible interpretations

            if reverse_none_means_isotropic:
                # None means no information and therefore it is isotropic

                return True

            else:  # None means that there are no proportional losses
                # therefore, it is only isotropic if there are no proportional
                # losses for the nominal flow sense too

                return tuple(self.efficiency.values()) == tuple(
                    1 for _ in range(len(self.efficiency))
                )

        else:
            # neither efficiency nor efficiency_reverse are None: all values
            # must match in either sense for it to be isotropic

            for key, value in self.efficiency.items():
                if self.efficiency[key] != self.efficiency_reverse[key]:
                    # one pair does not match: not isotropic

                    return False

            # all pairs match: it is istropic (they must have the same size)

            return True

    # *************************************************************************
    # *************************************************************************

    def has_been_selected(self) -> bool:
        """Returns True if an option has been selected and False otherwise."""

        return True in self.options_selected

    # *************************************************************************
    # *************************************************************************

    def is_infinite_capacity(self) -> bool:
        """Returns True if there is one capacity and it is infinite."""

        return len(self.capacity) == 1 and self.capacity == (inf,)

    # *************************************************************************
    # *************************************************************************

    def has_constant_efficiency(self) -> bool:
        """Returns True if the arc has a constant efficiency."""

        if self.has_proportional_losses():
            # has proportional losses
            if self.is_isotropic():
                # is isotropic
                if len(set(self.efficiency.values())) == 1:
                    # is isotropic and has constant eta = True
                    return True
                else:
                    # is isotropic but does not have constant eta = False
                    return False
            else:  # is not isotropic
                return False  # not isotropic = efficiency can change
        else:
            # no proportional losses: always equal to 1 = has constant eta
            return True

    # *************************************************************************
    # *************************************************************************

    def has_proportional_losses(self) -> bool:
        """Returns False if the efficiency is always one, otherwise True."""

        if type(self.efficiency) == type(None):
            # None means the efficiency is always one (default)

            return False  # no proportional losses

        else:  # efficiency is not None, reverse_efficiency may or not be too
            if type(self.efficiency_reverse) == type(None):
                # Return True if at least one efficiency value is not one

                return tuple(self.efficiency.values()) != tuple(
                    1 for _ in range(len(self.efficiency))
                )

            else:
                # Return True if at least one efficiency value is not one
                # Return False if all efficiency values are one

                if tuple(self.efficiency.values()) != tuple(
                    1 for _ in range(len(self.efficiency))
                ):
                    # the efficiency values are different than one: False

                    return True

                else:  # the efficiency values are one
                    # if the reverse efficiency values are not one, return True

                    return tuple(self.efficiency_reverse.values()) != tuple(
                        1 for _ in range(len(self.efficiency_reverse))
                    )

    # *************************************************************************
    # *************************************************************************

    def has_static_losses(self) -> bool:
        """Returns False if the static losses are always zero, otherwise True."""

        if type(self.static_loss) == type(None):
            # None means the static losses are always zero (default)

            return False  # no static losses

        else:
            # Return True if at least one static loss value is not zero

            return tuple(self.static_loss.values()) != tuple(
                0 for _ in range(len(self.static_loss))
            )

    # *************************************************************************
    # *************************************************************************

    def validate_sizes(
        self, number_options: int, number_scenarios: int, number_intervals: list
    ):
        # min cost and capacity: one per option

        if (
            len(self.minimum_cost) != number_options
            or len(self.capacity) != number_options
        ):
            raise ValueError(
                "The minimum cost or the capacity values are inconsistent with"
                + " the number of options."
            )

        # efficiency: scenario and interval

        if type(self.efficiency) != type(None):
            if len(self.efficiency) != sum(number_intervals):
                raise ValueError(
                    "The efficiency values are inconsistent with the number of"
                    + " scenarios and intervals."
                )

        # reverse efficiency: scenario and interval

        if type(self.efficiency_reverse) != type(None):
            if len(self.efficiency_reverse) != sum(number_intervals):
                raise ValueError(
                    "The reverse efficiency values are inconsistent with the "
                    + "number of scenarios and intervals."
                )

        # static loss: option, scenario and interval

        if type(self.static_loss) != type(None):
            if len(self.static_loss) != sum(number_intervals) * number_options:
                raise ValueError(
                    "The static loss values are inconsistent with the number "
                    + "of options, scenarios and intervals."
                )

    # *************************************************************************
    # *************************************************************************

    def validate(self):
        # *********************************************************************

        # check the types

        # the name should be hashable

        try:
            assert type(hash(self.name)) == int

        except TypeError:
            # not hashable

            raise TypeError("The name attribute is not hashable.")

        # efficiency

        if type(self.efficiency) != dict and type(self.efficiency) != type(None):
            # efficiency is not a dict nor a None

            raise TypeError("The efficiency should be given as a dict or None.")

        elif type(self.efficiency) == type(None):
            # efficiency is a None: the reverse efficiency must be too

            if type(self.efficiency_reverse) != type(None):
                # must be None if normal efficiency is None

                raise TypeError(
                    "The reverse efficiency has to match the nominal"
                    + " one when there are no proportional losses."
                )

        else:
            # efficiency is dict: the reverse eff. has to be a None or dict

            # must be given as dict or None

            if type(self.efficiency_reverse) != dict and type(
                self.efficiency_reverse
            ) != type(None):
                raise TypeError(
                    "The reverse efficiency should be given as a dict or None."
                )

        # static loss

        if type(self.static_loss) != dict and type(self.static_loss) != type(None):
            raise TypeError("The static losses should be given as a dict or None.")

        # capacity

        if type(self.capacity) != tuple:
            raise TypeError("The capacity should be given as a tuple.")

        # minimum cost

        if type(self.minimum_cost) != tuple:
            raise TypeError("The minimum cost values should be given as a tuple.")

        # specific capacity cost

        if not isinstance(self.specific_capacity_cost, Real):
            raise TypeError(
                "The specific capacity cost was not given as a numeric type."
            )

        # capacity_is_instantaneous

        if type(self.capacity_is_instantaneous) != bool:
            raise TypeError(
                "The information about capacities being instantaneous or not"
                + " should be given as a boolean variable."
            )

        # *********************************************************************

        # the number of techs. in capacity and minimum_cost should be the same

        if len(self.capacity) != len(self.minimum_cost):
            raise ValueError(
                "The number of capacity and minimum cost entries must match."
            )

        # there should be at least one option

        if len(self.capacity) == 0:
            raise ValueError(
                "No capacity and minimum cost values were provided. At le"
                + "ast one option should be provided."
            )

        # *********************************************************************

        # if efficiency is not None (i.e., the arc has proportional losses)

        if type(self.efficiency) != type(None):
            # efficiency is a dict

            if len(self.efficiency) == 0:
                raise ValueError(
                    "No efficiency values were provided. There should be one"
                    + " value per scenario and time interval."
                )

            for key, value in self.efficiency.items():
                if type(key) != tuple:
                    raise TypeError(
                        "The efficiency dict keys must be (scenario, interval)"
                        + " tuples."
                    )

                if len(key) != 2:
                    raise ValueError(
                        "The efficiency dict keys must be tuples of size 2."
                    )

                if not isinstance(value, Real):
                    raise TypeError(
                        "Efficiency values must be provided as numeric types."
                    )

                if value <= 0:
                    raise ValueError("Efficiency values must be positive.")

            # check if the reverse efficiency is not None

            if type(self.efficiency_reverse) != type(None):
                # it is not None, check the reverse efficiency too

                if len(self.efficiency_reverse) == 0:
                    raise ValueError(
                        "No efficiency values were provided. There should be "
                        + "one value per scenario and time interval."
                    )

                for key, value in self.efficiency_reverse.items():
                    if key not in self.efficiency.keys():
                        raise ValueError(
                            "The keys for the efficiency dicts do not match."
                        )

                    # if so, the key is a tuple of size 2 (checked before)

                    # if type(key) != tuple:

                    #     raise TypeError(
                    #         'The efficiency dict keys must be (scenario, '+
                    #         'interval) tuples.')

                    # if len(key) != 2:

                    #     raise ValueError(
                    #         'The efficiency dict keys must be tuples of '+
                    #         'size 2.')

                    if not isinstance(value, Real):
                        raise TypeError(
                            "Efficiency values must be provided as numeric " + "types."
                        )

                    if value <= 0:
                        raise ValueError("Efficiency values must be positive.")

        # if static loss is not None (i.e., the arc has static losses)

        if type(self.static_loss) != type(None):
            # static loss is a dict

            if len(self.static_loss) == 0:
                raise ValueError(
                    "No static loss values were provided. There should be one"
                    + " value per option, scenario and time interval."
                )

            for key, value in self.static_loss.items():
                if type(key) != tuple:
                    raise TypeError(
                        "The static loss dict keys must be (option, scenario,"
                        + " interval) tuples."
                    )

                if len(key) != 3:
                    raise ValueError(
                        "The static loss dict keys must be tuples of size 3."
                    )

                if not isinstance(value, Real):
                    raise TypeError(
                        "Static loss values must be provided as numeric types."
                    )

                if value < 0:
                    raise ValueError("Static loss values cannot be negative.")

        # capacity is a tuple

        for element in self.capacity:
            if not isinstance(element, Real):
                raise TypeError("Capacity values must be provided as numeric types.")

            if element <= 0:
                raise ValueError("Capacity values must be positive.")

        # minimum_cost is a tuple

        for element in self.minimum_cost:
            if not isinstance(element, Real):
                raise TypeError(
                    "Minimum cost values must be provided as numeric types."
                )

            if element < 0:
                raise ValueError("Minimum cost values must be positive or zero.")

        # *********************************************************************
        # *********************************************************************

    # *************************************************************************
    # *************************************************************************


# *****************************************************************************
# *****************************************************************************


class ArcsWithoutProportionalLosses(Arcs):
    """A class for arc technologies without proportional losses."""

    def __init__(
        self,
        name: str,  # human-readable name
        static_loss: dict,  # keys: (h, q, k)
        capacity: tuple,  # one value per option (h)
        minimum_cost: tuple,  # one value per option (h)
        specific_capacity_cost: float,  # one value
        capacity_is_instantaneous: bool,  # one value
        validate: bool = False,
    ):  # one number
        Arcs.__init__(
            self,
            name=name,
            efficiency=None,
            efficiency_reverse=None,
            static_loss=static_loss,
            capacity=capacity,
            minimum_cost=minimum_cost,
            specific_capacity_cost=specific_capacity_cost,
            capacity_is_instantaneous=capacity_is_instantaneous,
            validate=validate,
        )


# *****************************************************************************
# *****************************************************************************


class ArcsWithoutStaticLosses(Arcs):
    """A class for arc technologies without static losses."""

    def __init__(
        self,
        name: str,  # human-readable name
        efficiency: dict,  # keys: (q, k)
        efficiency_reverse: dict,  # keys: (q, k)
        capacity: tuple,  # one value per option (h)
        minimum_cost: tuple,  # one value per option (h)
        specific_capacity_cost: float,  # one value
        capacity_is_instantaneous: bool,  # one value
        validate: bool = False,
    ):  # one number
        Arcs.__init__(
            self,
            name=name,
            efficiency=efficiency,
            efficiency_reverse=efficiency_reverse,
            static_loss=None,
            capacity=capacity,
            minimum_cost=minimum_cost,
            specific_capacity_cost=specific_capacity_cost,
            capacity_is_instantaneous=capacity_is_instantaneous,
            validate=validate,
        )


# *****************************************************************************
# *****************************************************************************


class ArcsWithoutLosses(Arcs):
    """A class for arc technologies without losses."""

    def __init__(
        self,
        name: str,  # human-readable name
        capacity: tuple,  # one value per option (h)
        minimum_cost: tuple,  # one value per option (h)
        specific_capacity_cost: float,  # one value
        capacity_is_instantaneous: bool,  # one value
        validate: bool = False,
    ):  # one number
        Arcs.__init__(
            self,
            name=name,
            efficiency=None,
            efficiency_reverse=None,
            static_loss=None,
            capacity=capacity,
            minimum_cost=minimum_cost,
            specific_capacity_cost=specific_capacity_cost,
            capacity_is_instantaneous=capacity_is_instantaneous,
            validate=validate,
        )


# *****************************************************************************
# *****************************************************************************


class Network(nx.MultiDiGraph):
    # directed arcs represent streets
    # the nodes are the start and end points of those streets

    KEY_NODE_TYPE = "node_type"
    KEY_NODE_TYPE_IMP = "imp"
    KEY_NODE_TYPE_EXP = "exp"
    KEY_NODE_TYPE_SOURCE_SINK = "source_sink"
    KEY_NODE_TYPE_WAY = "way"
    KEY_NODE_TYPES = [
        KEY_NODE_TYPE_IMP,
        KEY_NODE_TYPE_EXP,
        KEY_NODE_TYPE_SOURCE_SINK,
        KEY_NODE_TYPE_WAY,
    ]
    KEY_NODE_PRICES = "prices"
    # KEY_NODE_PRICES = 'prices'
    # KEY_NODE_VOLUMES = 'volumes'
    KEY_NODE_BASE_FLOW = "base_flow"
    KEY_NODE_PRICES_TIME_INVARIANT = "prices_invariant"

    KEY_ARC_TECH = "technology"
    KEY_ARC_UND = "undirected"

    KEY_ARC_TECH_EFFICIENCY = "efficiency"
    KEY_ARC_TECH_EFFICIENCY_REVERSE = "efficiency_reverse"
    KEY_ARC_TECH_CAPACITY = "capacity"
    KEY_ARC_TECH_MIN_COST = "minimum_cost"
    KEY_ARC_TECH_SPEC_CAP_COST = "specific_capacity_cost"
    KEY_ARC_TECH_OPTIONS_SELECTED = "options_selected"
    KEY_ARC_TECH_CAPACITY_INSTANTANEOUS = "capacity_is_instantaneous"
    KEY_ARC_TECH_STATIC_LOSS = "static_loss"

    KEY_ARC_TECH_ATTR = (
        KEY_ARC_TECH_EFFICIENCY,
        KEY_ARC_TECH_EFFICIENCY_REVERSE,
        KEY_ARC_TECH_CAPACITY,
        KEY_ARC_TECH_MIN_COST,
        KEY_ARC_TECH_SPEC_CAP_COST,
        KEY_ARC_TECH_OPTIONS_SELECTED,
        KEY_ARC_TECH_CAPACITY_INSTANTANEOUS,
        KEY_ARC_TECH_STATIC_LOSS,
    )
    
    NET_TYPE_HYBRID = 0
    NET_TYPE_TREE = 1
    NET_TYPE_REV_TREE = 2
    
    NET_TYPES = (
        NET_TYPE_HYBRID,
        NET_TYPE_TREE,
        NET_TYPE_REV_TREE
        )

    def __init__(self, network_type = NET_TYPE_HYBRID, **kwargs):
    
        # initialise the node type
        self.import_nodes = set()
        self.export_nodes = set()
        self.source_sink_nodes = set()
        self.waypoint_nodes = set()

        # declare variables for the nodes without directed arc limitations
        if network_type not in self.NET_TYPES:
            raise ValueError('Unknown network type.')
        self.network_type = network_type
        # nodes without incoming directed arcs limitations
        self.nodes_w_in_dir_arc_limitations = dict()
        # nodes without outgoing directed arcs limitations
        self.nodes_w_out_dir_arc_limitations = dict()
        
        # run base class init routine
        nx.MultiDiGraph.__init__(self, **kwargs)
        
        # process the input data
        for node_key in self.nodes():
            self._process_node_data(node_key, data=self.nodes[node_key])
        
        # # set up initial nodes if the network is not hybrid
        # if self.network_type != self.NET_TYPE_HYBRID:
        #     for node_key in self.nodes():
        #         self._set_up_node(node_key)
                

    # *************************************************************************
    # *************************************************************************
    
    def _set_up_node(self, node_key, max_number_in_arcs: int = None, max_number_out_arcs: int = None):
        
        if self.should_be_tree_network():
            # nodes have to be part of a tree: one incoming arc per node at most
            self.nodes_w_in_dir_arc_limitations[node_key] = 1
        elif self.should_be_reverse_tree_network():
            # nodes have to be part of a reverse tree: one outgoing arc per node at most
            self.nodes_w_out_dir_arc_limitations[node_key] = 1
        else: 
            # nodes have no peculiar restrictions or they are defined 1 by 1
            if type(max_number_in_arcs) != type(None):
                self.nodes_w_in_dir_arc_limitations[node_key] = max_number_in_arcs
            if type(max_number_out_arcs) != type(None):
                self.nodes_w_out_dir_arc_limitations[node_key] = max_number_out_arcs

    # *************************************************************************
    # *************************************************************************
    
    def should_be_tree_network(self) -> bool:
        return self.network_type == self.NET_TYPE_TREE

    # *************************************************************************
    # *************************************************************************
    
    def should_be_reverse_tree_network(self) -> bool:
        return self.network_type == self.NET_TYPE_REV_TREE

    # *************************************************************************
    # *************************************************************************
    
    def _process_node_data(self, node_key, data: dict, **kwargs):
        "Check the data, determine the node type and update the structures."
        
        # find out whicht type of node it is
        # no data: waypoint
        # prices: transshipment (imp/exp)
        # flows: source/sink
        if type(data) == type(None):
            self.waypoint_nodes.add(node_key)
        elif type(data) == dict:
            # transshipment or source/sink
            if self.KEY_NODE_BASE_FLOW in data and len(data) == 1:
                # source/sink
                self.source_sink_nodes.add(node_key)
            elif self.KEY_NODE_PRICES in data and self.KEY_NODE_TYPE in data and data[self.KEY_NODE_TYPE] == self.KEY_NODE_TYPE_IMP:
                # import node
                self.import_nodes.add(node_key)
            elif self.KEY_NODE_PRICES in data and self.KEY_NODE_TYPE in data and data[self.KEY_NODE_TYPE] == self.KEY_NODE_TYPE_EXP:
                # export node
                self.export_nodes.add(node_key)
            elif self.KEY_NODE_PRICES not in data:
                # waypoint node
                self.waypoint_nodes.add(node_key)
            else:
                # error
                raise TypeError('Invalid input data combination.')
        else:
            raise TypeError('Invalid type for node data.')
        # set up the node
        self._set_up_node(node_key, **kwargs)
    
    # TODO: automatically identify import and export nodes (without defining them explicitly)
        
    # *************************************************************************
    # *************************************************************************
    
    def is_export_node(self, node_key) -> bool:
        "Returns True if the key matches that of an export node."
        return node_key in self.export_nodes
        
    def is_import_node(self, node_key) -> bool:
        "Returns True if the key matches that of an import node."
        return node_key in self.import_nodes
        
    def is_waypoint_node(self, node_key) -> bool:
        "Returns True if the key matches that of an waypoint node."
        return node_key in self.waypoint_nodes
        
    def is_source_sink_node(self, node_key) -> bool:
        "Returns True if the key matches that of an source or sink node."
        return node_key in self.source_sink_nodes

    # *************************************************************************
    # *************************************************************************
    
    def _reset_node_type(self, node_key):
        
        if self.is_export_node(node_key):
            # export node
            self.export_nodes.remove(node_key)
        elif self.is_import_node(node_key):
            # import node
            self.import_nodes.remove(node_key)
        elif self.is_source_sink_node(node_key):
            # source/sink node
            self.source_sink_nodes.remove(node_key)
        elif self.is_waypoint_node(node_key):
            # has to be a waypoint node
            self.waypoint_nodes.remove(node_key)
        # No need to reset node but this could mean something is up
        # else:
        #     raise ValueError('Unknown node type.')

    # *************************************************************************
    # *************************************************************************
    
    # TODO: use a decorator function to prevent the original method(s) from being used inappropriately
    
    def add_node(self, node_key, **kwargs):
        # check if the node can be added and add it
        self._handle_node(node_key, **kwargs)

    # *************************************************************************
    # *************************************************************************
            
    def modify_node(self, node_key, **kwargs):
        if not self.has_node(node_key):
            raise ValueError('The node indicated does not exist.')
        self._handle_node(node_key, **kwargs)

    # *************************************************************************
    # *************************************************************************
    
    # TODO: automatically check if node already exists and implications when "adding" one
    
    # def add_nodes(self, node_key_data: list):
        
    #     # process the input data
    #     for entry in node_key_data:
    #         if type(entry) != tuple :
    #             raise ValueError('The input must be a list of tuples.')
    #         # self._handle_node(entry[0], **entry[1])
    #         self._process_node_data(entry[0], entry[1])
    #     # add the nodes
    #     nx.MultiDiGraph.add_nodes_from(self, node_key_data)

    # *************************************************************************
    # *************************************************************************
    
    def add_nodes_from(self, nodes_for_adding, **kwargs):
        
        # input formats:
        # 1) container of node keys
        # 2) container of tuples
        # process the input data
        for entry in nodes_for_adding:
            if type(entry) == tuple and len(entry) == 2 and type(entry[1]) == dict:
                # option 2
                # update the dict
                new_dict = kwargs.copy()
                new_dict.update(entry[1])
                self._handle_node(entry[0], **new_dict)
            else:
                # option 1
                self._handle_node(entry, **kwargs)

    # *************************************************************************
    # *************************************************************************
    
    def modify_nodes_from(self, nodes_for_adding, **kwargs):
        
        # input formats:
        # 1) container of node keys
        # 2) container of tuples
        # process the input data
        for entry in nodes_for_adding:
            if type(entry) == tuple and len(entry) == 2 and type(entry[1]) == dict:
                # option 2
                new_dict = kwargs.copy()
                new_dict.update(entry[1])
                if not self.has_node(entry[0]):
                    raise ValueError('The node indicated does not exist.')
                self._handle_node(entry[0], **new_dict)
            else:
                # option 1
                if not self.has_node(entry):
                    raise ValueError('The node indicated does not exist.')
                self._handle_node(entry, **kwargs)

    # *************************************************************************
    # *************************************************************************
        
    def _handle_node(self, node_key, **kwargs):
        
        # node has to exist
        # the changes have to be compatible with the arcs involving the node
        # - has outgoing arcs: cannot be an export node
        # - has incoming arcs: cannot be an import node
        # - has no arcs: can be anything
        # the structures have to be updated accordingly
        
        if self.has_node(node_key):
            # node exists
            if type(kwargs) != type(None):
                # has data, check if the node type changes
                if self.KEY_NODE_TYPE in kwargs:
                    # node type is in the dict
                    _edges = get_edges_involving_node(self, node_key, include_self_loops=False)
                    # the node type is specified, possible change of node
                    if kwargs[self.KEY_NODE_TYPE] == self.KEY_NODE_TYPE_EXP and not self.is_export_node(node_key):
                        # change to an export node: it cannot have outgoing arcs
                        for _edge in _edges:
                            if _edge[0] == node_key:
                                # outgoing arc, raise error
                                raise ValueError(
                                    'A node with outgoing arcs cannot be an '
                                    'export node.'
                                    )
                        # change the node type
                    elif kwargs[self.KEY_NODE_TYPE] == self.KEY_NODE_TYPE_IMP and not self.is_import_node(node_key):
                        # change to an import node: it cannot have incoming arcs
                        for _edge in _edges:
                            if _edge[1] == node_key:
                                # incoming arc, raise error
                                raise ValueError(
                                    'A node with incoming arcs cannot be an '
                                    'import node.'
                                    )
                    # else:
                    #     raise ValueError('Unknown option.')
                # everything seems to be okay: reset node data
                self._reset_node_type(node_key)
            else:
                # no data: waypoint node, clear node data
                self._reset_node_type(node_key)
                keys = (self.KEY_NODE_BASE_FLOW, self.KEY_NODE_PRICES)
                for key in keys:
                    if key in self.nodes[node_key]:
                        self.nodes[node_key].pop(key)
            # the changes seem okay
            self._process_node_data(node_key, kwargs)
            nx.MultiDiGraph.add_node(self, node_key, **kwargs)
            
        else:
            # process the input data
            self._process_node_data(node_key, kwargs)
            # add the node
            nx.MultiDiGraph.add_node(self, node_key, **kwargs)

    # *************************************************************************
    # *************************************************************************

    # add a new import node

    def add_import_node(self, node_key, prices: dict):
        node_dict = {
            self.KEY_NODE_TYPE: self.KEY_NODE_TYPE_IMP,
            self.KEY_NODE_PRICES: prices
        }
        self.add_node(node_key, **node_dict)

    # *************************************************************************
    # *************************************************************************

    # add a new export node

    def add_export_node(self, node_key, prices: dict):
        node_dict = {
            self.KEY_NODE_TYPE: self.KEY_NODE_TYPE_EXP,
            self.KEY_NODE_PRICES: prices
        }
        self.add_node(node_key, **node_dict)

    # *************************************************************************
    # *************************************************************************

    # add a new supply/demand node

    def add_source_sink_node(self, node_key, base_flow: dict, **kwargs):
        node_dict = {
            self.KEY_NODE_BASE_FLOW: base_flow,
        }
        self.add_node(node_key, **node_dict)

    # *************************************************************************
    # *************************************************************************

    # add a new waypoint node

    def add_waypoint_node(self, node_key):
        self.add_node(node_key)

    # *************************************************************************
    # *************************************************************************

    # # modify an existing network node

    # def modify_network_node(self, node_key, node_data: dict, **kwargs):
    #     """
    #     Modifies a node in the network object.

    #     Parameters
    #     ----------
    #     node_key : hashable-type
    #         The key that identifies the node.
    #     node_data : dict
    #         A dictionary with the data that one wishes to change in the object.

    #     Raises
    #     ------
    #     ValueError
    #         Errors are raised if the node does not exist in the network object,
    #         and if the node changed has arcs that are incompatible with its new
    #         version, namely in terms of incoming and outgoing arcs.

    #     Returns
    #     -------
    #     None.

    #     """

    #     if self.has_node(node_key):
    #         # check if there will be changes to the type of node

    #         if (
    #             self.KEY_NODE_TYPE in node_data
    #             and self.KEY_NODE_TYPE in self.nodes[node_key]
    #         ):
    #             if (
    #                 node_data[self.KEY_NODE_TYPE]
    #                 != self.nodes[node_key][self.KEY_NODE_TYPE]
    #             ):
    #                 # the node type changed: check if final node is imp./exp.

    #                 # to export nodes

    #                 if node_data[self.KEY_NODE_TYPE] == self.KEY_NODE_TYPE_EXP:
    #                     # export nodes cannot have outgoing arcs
    #                     # check if there are outgoing arcs involving this node

    #                     number_out_arcs = len(
    #                         tuple(
    #                             arc_key
    #                             for arc_key in self.edges(keys=True)
    #                             if arc_key[0] == node_key  # is source
    #                         )
    #                     )

    #                     if number_out_arcs > 0:
    #                         raise ValueError(
    #                             "A node with outgoing arcs cannot be changed"
    #                             + " into an export node, since export nodes "
    #                             + " cannot have outgoing arcs."
    #                         )

    #                 # to import nodes

    #                 if node_data[self.KEY_NODE_TYPE] == self.KEY_NODE_TYPE_IMP:
    #                     # import nodes cannot have incoming arcs
    #                     # check if there are incoming arcs involving this node

    #                     number_in_arcs = len(
    #                         tuple(
    #                             arc_key
    #                             for arc_key in self.edges(keys=True)
    #                             if arc_key[1] == node_key  # is destination
    #                         )
    #                     )

    #                     if number_in_arcs > 0:
    #                         raise ValueError(
    #                             "A node with incoming arcs cannot be changed"
    #                             + " into an import node, since import nodes "
    #                             + " cannot have incoming arcs."
    #                         )

    #         # all good

    #         self.add_node(node_key, **node_data)
    #         self._set_up_node(node_key, **kwargs)

    #     else:
    #         raise ValueError("No such node was found.")

    # *************************************************************************
    # *************************************************************************

    # # identify importing nodes

    # def identify_import_nodes(self):
    #     self.import_nodes = tuple(
    #         node_key
    #         for node_key in self.nodes
    #         if self.KEY_NODE_TYPE in self.nodes[node_key]
    #         if (self.nodes[node_key][self.KEY_NODE_TYPE] == self.KEY_NODE_TYPE_IMP)
    #     )

    # # *************************************************************************
    # # *************************************************************************

    # # identify exporting nodes

    # def identify_export_nodes(self):
    #     self.export_nodes = tuple(
    #         node_key
    #         for node_key in self.nodes
    #         if self.KEY_NODE_TYPE in self.nodes[node_key]
    #         if (self.nodes[node_key][self.KEY_NODE_TYPE] == self.KEY_NODE_TYPE_EXP)
    #     )

    # # *************************************************************************
    # # *************************************************************************

    # # identify waypoint nodes

    # def identify_waypoint_nodes(self):
    #     self.waypoint_nodes = tuple(
    #         node_key
    #         for node_key in self.nodes
    #         if self.KEY_NODE_TYPE in self.nodes[node_key]
    #         if (self.nodes[node_key][self.KEY_NODE_TYPE] == self.KEY_NODE_TYPE_WAY)
    #     )

    # # *************************************************************************
    # # *************************************************************************

    # # identify source sink nodes

    # def identify_source_sink_nodes(self):
    #     self.source_sink_nodes = tuple(
    #         node_key
    #         for node_key in self.nodes
    #         if self.KEY_NODE_TYPE in self.nodes[node_key]
    #         if (
    #             self.nodes[node_key][self.KEY_NODE_TYPE]
    #             == self.KEY_NODE_TYPE_SOURCE_SINK
    #         )
    #     )

    # *************************************************************************
    # *************************************************************************

    # verify if everything is alright

    def validate(self):
        # check each arc

        for arc_key in self.edges(keys=True):
            # directed arcs cannot start in an export node

            if arc_key[0] in self.export_nodes:
                raise ValueError("Directed arcs cannot start in an export node.")

            # directed arcs cannot end on an import node

            if arc_key[1] in self.import_nodes:
                raise ValueError("Directed arcs cannot end in import node.")

            # import-export nodes cannot have static losses

            if (
                arc_key[0] in self.import_nodes
                and arc_key[1] in self.export_nodes
                and self.edges[arc_key][self.KEY_ARC_TECH].has_static_losses()
            ):
                raise ValueError(
                    "Directed arcs between import and export nodes cannot have"
                    + " static losses."
                )

            # undirected arcs cannot involve import nor export nodes

            if self.arc_is_undirected(arc_key):
                if arc_key[0] in self.import_nodes or arc_key[1] in self.export_nodes:
                    raise ValueError(
                        "Undirected arcs cannot involve import nor " + "export nodes."
                    )

    # *************************************************************************
    # *************************************************************************

    # # identify node types

    # def identify_node_types(self):
    #     "Identifies the node type for each node in the network objects."

    #     # identify import nodes

    #     self.identify_import_nodes()

    #     # identify export nodes

    #     self.identify_export_nodes()

    #     # identify source/sink nodes

    #     self.identify_source_sink_nodes()

    #     # identify waypoint nodes

    #     self.identify_waypoint_nodes()

    #     # validate

    #     self.validate()

    # *************************************************************************
    # *************************************************************************

    def add_directed_arc(self, node_key_a, node_key_b, arcs: Arcs):
        # check if the arc ends in an import node
        if node_key_b in self.import_nodes:
            raise ValueError("Directed arcs cannot end in an import node.")

        # check if the arc starts in an export node
        if node_key_a in self.export_nodes:
            raise ValueError("Directed arcs cannot start in an export node.")

        # check the arc is between import and export nodes
        if node_key_a in self.import_nodes and node_key_b in self.export_nodes:
            # it is between import and export nodes
            # check if it involves static losses
            if arcs.has_static_losses():
                raise ValueError(
                    "Arcs between import and export nodes cannot have static "
                    + "losses."
                )

        # add a new arc
        return self.add_edge(
            node_key_a, node_key_b, **{self.KEY_ARC_TECH: arcs, self.KEY_ARC_UND: False}
        )

    # *************************************************************************
    # *************************************************************************

    def add_undirected_arc(self, node_key_a, node_key_b, arcs: Arcs):
        # check if the arc links import or export nodes

        if (
            node_key_a in self.import_nodes
            or node_key_a in self.export_nodes
            or node_key_b in self.import_nodes
            or node_key_b in self.export_nodes
        ):
            raise ValueError("Undirected arcs cannot involve import or export nodes.")

        # add a new arc (undirected arcs require unique keys)

        return self.add_edge(
            node_key_a,
            node_key_b,
            key=self.get_pseudo_unique_arc_key(node_key_a, node_key_b),
            **{self.KEY_ARC_TECH: arcs, self.KEY_ARC_UND: True}
        )

    # *************************************************************************
    # *************************************************************************

    def get_pseudo_unique_arc_key(
        self, node_key_start, node_key_end, max_iterations: int = 10
    ):
        """
        Generate a pseudo-unique arc key for a pair of nodes.

        They key cannot exist in either direction. If such a key cannot be fou-
        nd, an error will be raised.

        Parameters
        ----------
        node_key_start : hashable-type
            The key for the start node.
        node_key_end : hashable-type
            the key for the end node.
        max_iterations : int, optional
            The maximum number of iterations. The default is 10.

        Returns
        -------
        hashable-type
            The pseudo-unique arc key.

        """
        try:
            keys_ab = self._adj[node_key_start][node_key_end].keys()
        except KeyError:
            # no arcs in the ab direction
            keys_ab = ((),)
        try:
            keys_ba = self._adj[node_key_end][node_key_start].keys()
        except KeyError:
            # no arcs in the ba direction
            keys_ba = ((),)

        # try to generate a unique key
        iteration = 0
        while iteration < max_iterations:
            new_key = str(uuid.uuid4())
            if new_key not in keys_ab and new_key not in keys_ba:
                return new_key
            iteration += 1
        raise ValueError("No unique arc key could be produced.")

    # *************************************************************************
    # *************************************************************************

    def modify_network_arc(self, node_key_a, node_key_b, arc_key_ab, data_dict: dict):
        # TODO: raise an error if the arc becomes inconsistent with the nodes it links and the data it has
        # a directed arc can only change to undirected if it does not involve import or export nodes

        # modify the arc

        return self.add_edge(node_key_a, node_key_b, key=arc_key_ab, **data_dict)

    # *************************************************************************
    # *************************************************************************

    def add_infinite_capacity_arc(
        self, node_key_a, node_key_b, efficiency: dict, static_loss: dict
    ):
        """
        Adds an infinite capacity arc, which must be directed and pre-existing.

        Pre-existing arcs are not subject to optimisation but constrain it.

        Parameters
        ----------
        node_key_a : hashable-type
            The node key of the arc's start node.
        node_key_b : hashable-type
            The node key of the arc's end node.
        efficiency : dict
            A dict containing the arc efficiency values for each interval and
            scenario using (scenario, interval) tuples as keys.
        static_loss: dict
            A dict containing the static loss values for each option, interval
            and scenario using (option, scenario, interval) as keys.

        Returns
        -------
        None.

        """

        # prepare the arc technology object

        arcs = Arcs(
            name=str(node_key_a) + str(node_key_b),
            efficiency=efficiency,
            efficiency_reverse=None,  # should be ignored downstream
            static_loss=static_loss,
            capacity=[inf],  # one capacity only, infinity
            minimum_cost=[0],  # pre-existing arcs have no costs
            specific_capacity_cost=0,  # no specific capacity costs
            capacity_is_instantaneous=False,
            validate=False,
        )

        # identify it as having been preselected

        arcs.options_selected[0] = True

        # add the arc

        return self.add_directed_arc(node_key_a, node_key_b, arcs=arcs)

    # *************************************************************************
    # *************************************************************************

    def add_preexisting_directed_arc(
        self,
        node_key_a,
        node_key_b,
        efficiency: dict,
        static_loss: dict,
        capacity: float or int,
        capacity_is_instantaneous: bool,
    ):
        """
        Adds a pre-existing directed arc to the network.

        re-existing arcs are not subject to optimisation but constrain it.

        Parameters
        ----------
        node_key_a : hashable-type
            The node key of the arc's start node.
        node_key_b : hashable-type
            The node key of the arc's end node.
        efficiency : dict
            A dict containing the arc efficiency values for each interval and
            scenario using (scenario, interval) tuples as keys.
        static_loss: dict
            A dict containing the static loss values for each option, interval
            and scenario using (option, scenario, interval) as keys.
        capacity : float or int
            The arc capacity. It must be positive. It can also be infinite, in
            which case it should be set to math.inf (infinity).
        capacity_is_instantaneous : bool
            If True, the capacity is instantaneous. If False, the capacity is
            expressed in relation to the average time interval duration.

        Returns
        -------
        None.

        """

        # prepare the arc technology object

        arcs = Arcs(
            name=str(node_key_a) + str(node_key_b),
            efficiency=efficiency,
            efficiency_reverse=None,  # should be ignored downstream
            static_loss=static_loss,
            capacity=[capacity],  # one capacity only
            minimum_cost=[0],  # pre-existing arcs have no costs
            specific_capacity_cost=0,  # no specific capacity costs
            capacity_is_instantaneous=capacity_is_instantaneous,
            validate=False,
        )

        # identify it as having been preselected

        arcs.options_selected[0] = True

        # add the arc

        return self.add_directed_arc(node_key_a, node_key_b, arcs=arcs)

    # *************************************************************************
    # *************************************************************************

    def add_preexisting_undirected_arc(
        self,
        node_key_a,
        node_key_b,
        efficiency: dict,
        efficiency_reverse: dict,
        static_loss: dict,
        capacity: float or int,
        capacity_is_instantaneous: bool,
    ):
        """
        Adds a pre-existing undirected arc.

        Pre-existing arcs are not subject to optimisation but constrain it. The
        difference with regard to directed arcs is that undirected arcs cannot
        have an infinite capacity.

        Parameters
        ----------
        node_key_a : hashable-type
            The node key of the arc's start node.
        node_key_b : hashable-type
            The node key of the arc's end node.
        efficiency : dict
            A dict containing the arc efficiency values for each interval and
            scenario using (scenario, interval) tuples as keys.
        efficiency_reverse : dict
            A dict containing the reverse arc efficiency values for each
            interval and scenario using (scenario, interval) tuples as keys.
        static_loss: dict
            A dict containing the static loss values for each option, interval
            and scenario using (option, scenario, interval) as keys.
        capacity : float or int
            The arc capacity. It must be positive and finite.
        capacity_is_instantaneous : bool
            If True, the capacity is instantaneous. If False, the capacity is
            expressed in relation to the average time interval duration.

        Returns
        -------
        None.

        """

        # prepare the arc technology object

        arcs = Arcs(
            name=str(node_key_a) + str(node_key_b),
            efficiency=efficiency,
            efficiency_reverse=efficiency_reverse,
            static_loss=static_loss,
            capacity=[capacity],
            minimum_cost=[0],
            specific_capacity_cost=0,
            capacity_is_instantaneous=False,
            validate=False,
        )

        # identify it as having been preselected

        arcs.options_selected = [True]

        # add an undirected arc

        return self.add_undirected_arc(
            node_key_a=node_key_a, node_key_b=node_key_b, arcs=arcs
        )

    # *************************************************************************
    # *************************************************************************

    def arc_is_undirected(self, arc_key) -> bool:
        "Returns True if the arc is undirected and False otherwise."

        if self.has_edge(*arc_key):
            if self.KEY_ARC_UND in self.edges[arc_key]:
                # arc has undirected attribute

                return self.edges[arc_key][self.KEY_ARC_UND]

            else:  # arc has no undirected attribute
                return False

        else:  # arc does not exist
            return False

    # *************************************************************************
    # *************************************************************************

    def has_tree_topology(self) -> bool:
        """
        Returns True if the network has a tree topology and False otherwise. It
        only considers arcs that are selected and connected nodes.

        It relies on networkx.is_tree method.

        Returns
        -------
        bool
            If True, the network has a tree topology. If False, it does not.

        """

        # obtain network view

        network_view = self.copy(as_view=False)

        # remove arcs that were not selected

        network_view.remove_edges_from(
            [
                arc_key
                for arc_key in self.edges(keys=True)
                if True not in self.edges[arc_key]["technology"].options_selected
            ]
        )

        # remove unconnected nodes

        network_view.remove_nodes_from(find_unconnected_nodes(network_view))

        if network_view.number_of_nodes() == 0:
            # trivial solution

            return False

        else:
            # use networkx method for tree detection

            return nx.is_tree(network_view)

    # *************************************************************************
    # *************************************************************************
    
    def has_selected_antiparallel_arcs(self) -> bool:
        "Returns True if any two nodes have selected arcs in both directions."
        return len(self.find_selected_antiparallel_arcs()) != 0

    # *************************************************************************
    # *************************************************************************
    
    def find_selected_antiparallel_arcs(self) -> list:
        """Returns True if any two nodes have (selected) forward and reverse arcs."""
        
        # check the existence of forward and reverse arcs in the same segment
        arcs = [  # get the arcs selected
            arc_key[0:2]
            for arc_key in self.edges(keys=True)
            if True in self.edges[arc_key][Network.KEY_ARC_TECH].options_selected
        ]
        arcs = [  # get the selected arcs that exist both ways
            arc_key
            for arc_key in arcs
            if (arc_key[1], arc_key[0]) in arcs
        ]
        return arcs

# *****************************************************************************
# *****************************************************************************
