# imports

# standard libraries
from numbers import Real
from statistics import mean
import math

# local libraries, external
import pyomo.environ as pyo

# local libraries, internal
from .model import create_model
from ...solvers.interface import SolverInterface
from .network import Network, Arcs
from .system import EnergySystem
from .resource import ResourcePrice
from .time import EconomicTimeFrame
from .time import entries_are_invariant
from .blocks.prices import NODE_PRICE_OTHER, NODE_PRICES

# *****************************************************************************
# *****************************************************************************

# TODO: allow users to define how fixed components in the objective function
# are handled (using a variable equal to one or by excluding them altogether)

# TODO: create a log

class InfrastructurePlanningProblem(EnergySystem):
    """A class for optimisation of infrastructure planning problems."""

    # *************************************************************************
    # *************************************************************************

    SOS1_ARC_WEIGHTS_NONE = None
    SOS1_ARC_WEIGHTS_COST = 1
    SOS1_ARC_WEIGHTS_CAP = 2
    SOS1_ARC_WEIGHTS_SPEC_COST = 3
    SOS1_ARC_WEIGHTS_SPEC_CAP = 4
    SOS1_ARC_WEIGHTS = (
        SOS1_ARC_WEIGHTS_NONE,
        SOS1_ARC_WEIGHTS_COST,
        SOS1_ARC_WEIGHTS_CAP,
        SOS1_ARC_WEIGHTS_SPEC_COST,
        SOS1_ARC_WEIGHTS_SPEC_CAP,
    )

    SOS1_SENSE_OFFSET = 1
    SOS1_SENSE_WEIGHT_NOMINAL_HIGHER = 1
    SOS1_SENSE_WEIGHT_REVERSE_HIGHER = 2
    SOS1_SENSE_WEIGHTS = [
        SOS1_SENSE_WEIGHT_NOMINAL_HIGHER,
        SOS1_SENSE_WEIGHT_REVERSE_HIGHER,
    ]

    STATIC_LOSS_MODE_DEP = 1
    STATIC_LOSS_MODE_ARR = 2
    STATIC_LOSS_MODE_US = 3
    STATIC_LOSS_MODE_DS = 4

    STATIC_LOSS_MODES = (
        STATIC_LOSS_MODE_DEP,
        STATIC_LOSS_MODE_ARR,
        STATIC_LOSS_MODE_US,
        STATIC_LOSS_MODE_DS,
    )

    # *************************************************************************
    # *************************************************************************

    def __init__(
        self,
        # discount_rates: dict,  # key: assessment; value: list
        time_frame: EconomicTimeFrame,
        # reporting_periods: dict,  # key: assessment; value: periods
        # time_intervals: dict,  # key: assessment; value: intervals
        time_weights: dict = None,  # key: assessment, period, interval; value: weight
        normalised_time_interval_duration: dict = None,
        assessment_weights: dict = None,
        networks: dict = None,
        converters: dict = None,
        prepare_model: bool = True,
        validate_inputs: bool = True,
        node_price_model = NODE_PRICE_OTHER,
        use_prices_block: bool = True #False
    ):  # TODO: switch to False when everything is more mature
        # *********************************************************************

        # if validate_inputs:
        #     # validate the inputs

        #     (
        #         self.assessment_keys,
        #         self.number_assessments,
        #         self.number_reporting_periods,
        #         self.number_time_intervals,
        #     ) = self._validate_inputs(
        #         discount_rates=discount_rates,
        #         reporting_periods=time_frame.reporting_periods,
        #         time_intervals=time_frame.time_intervals,
        #         time_weights=time_weights,
        #         normalised_time_interval_duration=(normalised_time_interval_duration),
        #         assessment_weights=assessment_weights,
        #     )

        # else:  # skip validation
        #     self.time_frame.assessments = tuple(discount_rates.keys())

        #     self.number_assessments = len(self.time_frame.assessments)

        #     self.number_reporting_periods = {
        #         q: len(time_frame.reporting_periods[q]) for q in self.time_frame.assessments
        #     }

        #     self.number_time_intervals = {
        #         q: len(time_frame.time_intervals[q]) for q in self.time_frame.assessments
        #     }
        
        # *********************************************************************
        # *********************************************************************
        
        # options
        self.use_prices_block = use_prices_block
        self.node_price_model = node_price_model
        
        # *********************************************************************
        # *********************************************************************

        # initialise

        # self.discount_rates = dict(discount_rates)
        
        self.time_frame = time_frame

        # self.reporting_periods = dict(reporting_periods)

        # self.time_intervals = dict(time_intervals)

        self.average_time_interval = {
            q: mean(self.time_frame.time_interval_durations[q]) for q in self.time_frame.assessments
        }

        self.normalised_time_interval_duration = {
            (q, k): duration / self.average_time_interval[q]
            for q in self.time_frame.assessments
            for k, duration in enumerate(self.time_frame.time_interval_durations[q])
        }

        # relation between reporting periods and time intervals

        if type(time_weights) != dict:
            self.time_weights = None  # default values will be used

            # self.time_weights = {
            #     (q,p,k): 1
            #     for q in self.time_frame.assessments
            #     for p in self.reporting_periods[q]
            #     for k in self.time_intervals[q]
            #     }

        else:
            # TODO: validate

            # non-default values

            self.time_weights = dict(time_weights)

        # weight of each assessment

        if type(assessment_weights) != dict:
            self.assessment_weights = None

        else:
            # TODO: validate

            self.assessment_weights = dict(assessment_weights)

        # *********************************************************************

        # identify the type of problem

        # TODO: develop method to automatically identify the type of problem

        self.optimisation_problem_type = SolverInterface.PROBLEM_MILP

        # *********************************************************************

        # initialise dynamic systems and networks objects

        EnergySystem.__init__(self, networks=networks, converters=converters)

        # *********************************************************************

        # modelling options

        # dict to indicate which arcs should be decided upon using SOS1
        # keys: (g,u,v,j) tuples for selectable arcs using sos
        # values: sos weights for each arc option
        # default outcome (by leaving it empty): do not use them/False

        self.use_sos1_arc_inv = {}

        # list to flag if real arc invest. variables should be used if possible
        # elements: (g,u,v,j) tuples for selectable arcs using sos
        # default outcome (by leaving it empty): do not use them/False

        self.use_real_arc_inv_variables_if_possible = []

        # dict to indicate which arcs should have their sense decided via SOS1
        # keys: (g,u,v,j) tuples for undirected arcs using sos
        # values: sos weights for each arc option
        # default outcome (by leaving it empty): do not use them/False

        self.use_sos1_flow_sense = {}

        # list to flag if real flow sense variables should be used if possible
        # elements: (g,u,v,j) tuples for undirected arcs
        # default outcome (by leaving it empty): do not use them/False

        self.use_real_sense_variables_if_possible = []

        # ... to indicate which arcs are to be modelled using interface variab.
        # elements: (g,u,v,j) tuples for selectable arcs using sos
        # default outcome (by leaving it empty): do not use them/False

        self.use_arc_interface_variables = []

        # *********************************************************************

        # groups

        # dict for groups of arcs
        # keys: arc group key
        # content: GLLJ tuples for the groups
        # default outcome (by leaving it empty): no groups

        self.arc_groups = {}

        # number of options for each group

        self.groups_number_options = {}

        # mandatory groups

        self.groups_mdt = {}

        # groups using interfaces

        self.groups_int = {}

        # groups using sos for arc selection

        self.groups_arc_sos1 = {}

        # groups using nnr for arc selection

        self.groups_arc_nnr = {}

        # *********************************************************************

        # static losses

        # dict to list the arcs whose static losses are placed in the start no.

        self.static_losses_departure_node = {}

        # dict to list the arcs whose static losses are placed in the end node

        self.static_losses_arrival_node = {}

        # dict to list the arcs whose static losses are placed upstream

        self.static_losses_upstream = {}

        # dict to list the arcs whose static losses are placed downstream

        self.static_losses_downstream = {}

        # *********************************************************************

        if prepare_model:
            self.prepare()

    # *************************************************************************
    # *************************************************************************

    def _validate_inputs(
        self,
        discount_rates: dict,  # key: assessment; value: list
        reporting_periods: dict,  # key: assessment; value: periods
        time_intervals: dict,  # key: assessment; value: intervals
        time_weights: dict,  # key: assessment, period, interval; value: weight
        normalised_time_interval_duration: dict,  # key: assessment, interval; value: ratio
        assessment_weights: dict,  # key: assessment; value: weight
        **kwargs
    ) -> tuple:
        # types

        if type(discount_rates) != dict:
            raise TypeError(
                "Discount rates must be provided as a dictionary "
                + "whose keys represent assessments."
            )

        if type(reporting_periods) != dict:
            raise TypeError(
                "Reporting periods must be provided as a dictionary "
                + "whose keys represent assessments."
            )

        if type(time_intervals) != dict:
            raise TypeError(
                "Time intervals must be provided as a dictionary "
                + "whose keys represent assessments."
            )

        # sizes

        assessment_keys = tuple(discount_rates.keys())

        number_assessments = len(assessment_keys)

        if number_assessments == 0:
            raise ValueError(
                "The problem only makes sense with at least one assessment."
            )

        # use the first assessment key to determine the number of reporting p.

        total_number_reporting_periods = len(discount_rates[assessment_keys[0]])

        for q in assessment_keys:
            i_qk = discount_rates[q]

            if type(i_qk) != tuple:
                raise TypeError(
                    "The discount rates for a given assesssment "
                    + "should be provided as a tuple."
                )

            number_i = len(i_qk)

            if number_i == 0:
                raise ValueError(
                    "There should be at least one discount rate " + "per assessment."
                )

            # the number of discount rates should be the same

            if number_i != total_number_reporting_periods:
                raise ValueError(
                    "The discount rates provided are inconsistent"
                    " among the different assessments."
                )

        # check if all the keys for reporting_periods are for assessments

        for key, value in reporting_periods.items():
            if key not in assessment_keys:
                raise ValueError("Unknown assessment key: " + str(key) + ".")

            if type(value) != tuple:
                raise TypeError(
                    "The reporting periods for a given assessment "
                    + "have to provided as a tuple."
                )

        # check if all the keys for time_intervals are for assessments

        for key, value in time_intervals.items():
            if key not in assessment_keys:
                raise ValueError("Unknown assessment key: " + str(key) + ".")

            if type(value) != tuple:
                raise TypeError(
                    "The time intervals for a given assessment "
                    + "have to provided as a tuple."
                )

            # time interval durations have to be positive reals

            for duration in value:
                if isinstance(duration, Real):
                    # time intervals (durations) must be positive

                    if duration <= 0:
                        raise ValueError("Time interval durations have to be positive.")

                else:
                    raise ValueError("Time interval durations have to be real-valued.")

        # all assessments must be included in reporting_periods and time_interv

        # checking the size will suffice, since the keys are valid by now

        if len(reporting_periods.keys()) != number_assessments:
            raise ValueError("There is missing data about (some) assessments.")

        if len(time_intervals.keys()) != number_assessments:
            raise ValueError("There is missing data about (some) assessments.")

        # number of reporting periods and time intervals

        number_reporting_periods = {
            q: len(reporting_periods[q]) for q in assessment_keys
        }

        number_time_intervals = {q: len(time_intervals[q]) for q in assessment_keys}

        # all reporting periods have to be covered by the assessments

        set_reporting_periods = set()

        for q in assessment_keys:
            set_reporting_periods = set_reporting_periods.union(reporting_periods[q])

        if len(set_reporting_periods) != total_number_reporting_periods:
            raise ValueError(
                "Not all reporting periods for which there are discount rates"
                + " are covered."
            )

        # reporting periods have to form a sequence (among all assessments)

        # intervals have to be sequential (within each scenario)

        # return useful data

        return (
            assessment_keys,
            number_assessments,
            number_reporting_periods,
            number_time_intervals,
        )

    # *************************************************************************
    # *************************************************************************

    def reset_arc_groups(self):
        "Clears the dictionaries for the groups of arcs under consideration."

        # groups

        self.arc_groups = {}

        # number of options for each group

        self.groups_number_options = {}

        # mandatory groups

        self.groups_mdt = {}

        # groups using interfaces

        self.groups_int = {}

        # groups using sos1 for arc selection

        self.groups_arc_sos1 = {}

        # groups using nnr for arc selection

        self.groups_arc_nnr = {}

    # *************************************************************************
    # *************************************************************************

    def create_arc_group(
        self,
        gllj_tuples: tuple,
        mandatory: bool = False,
        use_sos1: bool = False,
        use_interface: bool = False,
        use_nnr_variables_if_possible: bool = False,
        sos1_weight_method: int = SOS1_ARC_WEIGHTS_NONE
    ) -> int:
        """
        Create a group of arcs whose invesment is to be decided collectively.

        Parameters
        ----------
        gllj_tuples : tuple
            The tuples identifying every arc in the group.

        Raises
        ------
        ValueError
            This error is raised if the network or the arc are not recognised.

        Returns
        -------
        int
            Returns the key that identifies the arc group, an integer.

        """
        
        # make sure there is at least one arc

        if len(gllj_tuples) < 2:
            raise ValueError(
                "At least two arcs need to be identified to create a group."
            )
        
        for arc_number, gllj in enumerate(gllj_tuples):
            # does the network exist?

            if gllj[0] in self.networks:
                # does the arc exist?

                if self.networks[gllj[0]].has_edge(*gllj[1:]):
                    # the arc exists, check if it has the same number of arc
                    # options as previous arcs

                    if arc_number == 0:
                        # first iteration, determine the number of options

                        number_options = (
                            self.networks[gllj[0]]
                            .edges[gllj[1:]][Network.KEY_ARC_TECH]
                            .number_options()
                        )

                        continue

                    if (
                        number_options
                        != self.networks[gllj[0]]
                        .edges[gllj[1:]][Network.KEY_ARC_TECH]
                        .number_options()
                    ):
                        raise ValueError(
                            "The number of options is not consistent among the"
                            + " arcs."
                        )

                else:
                    # the arc does not exist

                    raise ValueError("The arc " + str(gllj) + " does not exist.")

            else:  # no, it does not
                raise ValueError("The network " + str(gllj[0]) + " does not exist.")

        # everything checks out: add the group

        new_t = len(self.arc_groups)

        self.arc_groups[new_t] = tuple(gllj_tuples)

        # number of options for each group

        self.groups_number_options[new_t] = number_options

        # mandatory groups

        self.groups_mdt[new_t] = mandatory

        # groups using interfaces

        self.groups_int[new_t] = use_interface
        # TODO: allow users to set the weights
        # groups using sos1 for arc selection
        if use_sos1:
            self.groups_arc_sos1[new_t] = [i for i in range(number_options)]

        # groups using nnr for arc selection

        self.groups_arc_nnr[new_t] = use_nnr_variables_if_possible

        # return the key

        return new_t

    # *************************************************************************
    # *************************************************************************

    def clear_static_loss_configurations(self):
        """
        Clears all configurations for arcs with static losses.

        Returns
        -------
        None.

        """

        self.static_losses_arrival_node = {}

        self.static_losses_departure_node = {}

        self.static_losses_downstream = {}

        self.static_losses_upstream = {}

    # *************************************************************************
    # *************************************************************************

    def place_static_losses(self, network_key, arc_key, mode):
        """
        Defines the mode according to which a given arc with static losses is
        to have them represented.

        There are four options available:
            1) in the flow departure node (directed or undirected arcs);
            2) in the flow arrival node (directed or undirected arcs);
            3) upstream (undirected arcs only);
            4) downstream (undirected arcs only).

        Parameters
        ----------
        network_key : hashable-type
            The key to the network object.
        arc_key : hashable-type
            The key to the arc in the specified network.
        mode : hashable-type
            The mode according to which the static losses are to be modelled.
            It must be selected from among those in STATIC_LOSS_MODES, namely:
                1) STATIC_LOSS_MODE_DEP
                2) STATIC_LOSS_MODE_ARR
                3) STATIC_LOSS_MODE_US
                4) STATIC_LOSS_MODE_DS

        Raises
        ------
        ValueError
            This error is raised if the arc or the network are not recognised,
            if the arc is not directed or involves an export node, or if the
            static loss modelling mode is not recognised.

        Returns
        -------
        None.

        """

        if mode not in self.STATIC_LOSS_MODES:
            raise ValueError("Unknown static loss modelling mode.")

        if network_key in self.networks and len(arc_key) == 3:
            # check if the arc exists

            if self.networks[network_key].has_edge(*arc_key):
                # the arc exists, now check if it has static losses

                if (
                    not self.networks[network_key]
                    .edges[arc_key][Network.KEY_ARC_TECH]
                    .has_static_losses()
                ):
                    # no static losses, do nothing

                    return

                # *************************************************************

                # departure node

                if mode == self.STATIC_LOSS_MODE_DEP:
                    # if it exists, it cannot link import and export nodes

                    # if it exists, it cannot start on an import node

                    if arc_key[0] in self.networks[network_key].import_nodes:
                        # raise ValueError(
                        #     'Import nodes cannot be used to place losses in.')

                        self.place_static_losses(
                            network_key, arc_key, mode=self.STATIC_LOSS_MODE_ARR
                        )

                    # if everything is okay, update the dict

                    if (
                        network_key,
                        *arc_key[0:2],
                    ) not in self.static_losses_departure_node:
                        # previous entries exist

                        self.static_losses_departure_node[
                            (network_key, *arc_key[0:2])
                        ] = [arc_key[2]]

                    else:  # new entry
                        self.static_losses_departure_node[
                            (network_key, *arc_key[0:2])
                        ].append(arc_key[2])

                # *************************************************************

                # arrival node

                if mode == self.STATIC_LOSS_MODE_ARR:
                    # if it exists, it cannot link import and export nodes

                    # if it exists, it cannot end on an export node

                    if arc_key[1] in self.networks[network_key].export_nodes:
                        # raise ValueError(
                        #     'Export nodes cannot be used to place losses in.')

                        self.place_static_losses(
                            network_key, arc_key, mode=self.STATIC_LOSS_MODE_DEP
                        )

                    # if everything is okay, update the dict

                    if (
                        network_key,
                        *arc_key[0:2],
                    ) not in self.static_losses_arrival_node:
                        # previous entries exist

                        self.static_losses_arrival_node[
                            (network_key, *arc_key[0:2])
                        ] = [arc_key[2]]

                    else:  # new entry
                        self.static_losses_arrival_node[
                            (network_key, *arc_key[0:2])
                        ].append(arc_key[2])

                # *************************************************************

                # upstream

                if mode == self.STATIC_LOSS_MODE_US:
                    # if it exists, it cannot be directed

                    if not self.networks[network_key].edges[arc_key][
                        Network.KEY_ARC_UND
                    ]:
                        # if the arc is directed, use STATIC_LOSS_MODE_DEP mode

                        self.place_static_losses(
                            network_key, arc_key, mode=self.STATIC_LOSS_MODE_DEP
                        )

                        return  # done

                    # if everything is okay, update the dict

                    if (network_key, *arc_key[0:2]) not in self.static_losses_upstream:
                        # previous entries exist

                        self.static_losses_upstream[(network_key, *arc_key[0:2])] = [
                            arc_key[2]
                        ]

                    else:  # new entry
                        self.static_losses_upstream[
                            (network_key, *arc_key[0:2])
                        ].append(arc_key[2])

                # *************************************************************

                # downstream

                if mode == self.STATIC_LOSS_MODE_DS:
                    # if it exists, it cannot be directed

                    if not self.networks[network_key].edges[arc_key][
                        Network.KEY_ARC_UND
                    ]:
                        # if the arc is directed, use STATIC_LOSS_MODE_ARR mode

                        self.place_static_losses(
                            network_key, arc_key, mode=self.STATIC_LOSS_MODE_ARR
                        )

                        return  # done

                    # if everything is okay, update the dict

                    if (
                        network_key,
                        *arc_key[0:2],
                    ) not in self.static_losses_downstream:
                        # previous entries exist

                        self.static_losses_downstream[(network_key, *arc_key[0:2])] = [
                            arc_key[2]
                        ]

                    else:  # new entry
                        self.static_losses_downstream[
                            (network_key, *arc_key[0:2])
                        ].append(arc_key[2])

                # *************************************************************

            else:  # the arc does not exist
                raise ValueError(
                    "The arc key used does not match any arc in the network."
                )

        else:  # something is up with the network key or arc key length
            raise ValueError(
                "Either the network key provided is incorrect or the arc key "
                + "lacks the proper size."
            )

    # *************************************************************************
    # *************************************************************************

    def place_static_losses_departure_node(self):
        """
        Configures the problem instances so that arcs will have their static
        losses placed in the departure node or, if not possible, in the arrival
        node.

        Returns
        -------
        None.

        """

        # all arcs except those ending in export nodes: departure
        # arcs ending in export nodes: arrival

        # clear all configurations

        self.clear_static_loss_configurations()

        for network_key, network in self.networks.items():
            for arc_key in network.edges(keys=True):
                # if it has static losses

                if network.edges[arc_key][Network.KEY_ARC_TECH].has_static_losses():
                    # if the arc is undirected

                    if network.arc_is_undirected(arc_key):
                        self.place_static_losses(
                            network_key, arc_key, mode=self.STATIC_LOSS_MODE_DEP
                        )

                    elif arc_key[0] not in network.import_nodes:
                        # directed and the departure node is not an import one
                        # >> departure

                        self.place_static_losses(
                            network_key, arc_key, mode=self.STATIC_LOSS_MODE_DEP
                        )

                    elif arc_key[1] not in network.export_nodes:
                        # directed and the arrival node is not an export one
                        # though the departure node is an import one
                        # >> arrival

                        self.place_static_losses(
                            network_key, arc_key, mode=self.STATIC_LOSS_MODE_ARR
                        )

                    else:
                        raise ValueError(
                            "Static losses cannot exist directly between "
                            + "import and export nodes. Use an intermediate arc"
                            + " to obtain the same outcome."
                        )

    # *************************************************************************
    # *************************************************************************

    def place_static_losses_arrival_node(self):
        """
        Configures the problem instances so that arcs will have their static
        losses placed in the arrival node or, if not possible, in the departure
        node.

        Returns
        -------
        None.

        """

        # all arcs except those starting in import nodes: arrival
        # arcs starting in import nodes: departure

        # clear all configurations

        self.clear_static_loss_configurations()

        for network_key, network in self.networks.items():
            for arc_key in network.edges(keys=True):
                # if it has static losses

                if network.edges[arc_key][Network.KEY_ARC_TECH].has_static_losses():
                    # if the arc is undirected

                    if network.arc_is_undirected(arc_key):
                        # arrival

                        self.place_static_losses(
                            network_key, arc_key, mode=self.STATIC_LOSS_MODE_ARR
                        )

                    elif arc_key[1] not in network.export_nodes:
                        # directed and the arrival node is not an export one
                        # >> arrival

                        self.place_static_losses(
                            network_key, arc_key, mode=self.STATIC_LOSS_MODE_ARR
                        )

                    elif arc_key[0] not in network.import_nodes:
                        # directed and the departure node is not an import one
                        # though the arrival node is an export one
                        # >> departure

                        self.place_static_losses(
                            network_key, arc_key, mode=self.STATIC_LOSS_MODE_DEP
                        )

                    else:
                        raise ValueError(
                            "Static losses cannot exist directly between "
                            + "import and export nodes. Use an intermediate arc"
                            + " to obtain the same outcome."
                        )

    # *************************************************************************
    # *************************************************************************

    def place_static_losses_upstream(self):
        """
        Configures the problem instances so that arcs will have their static
        losses placed upstream relative to the flow sense or, if not possible,
        namely if it is a directed arc, in the opposite way. Note that arcs be-
        tween import and exports are assumed to have been excluded.

        Returns
        -------
        None.

        """

        # no import-export arcs with static losses are allowed

        # clear all configurations

        self.clear_static_loss_configurations()

        # for each arc

        for network_key, network in self.networks.items():
            for arc_key in network.edges(keys=True):
                if network.edges[arc_key][Network.KEY_ARC_TECH].has_static_losses():
                    # if the arc is undirected

                    if network.arc_is_undirected(arc_key):
                        self.place_static_losses(
                            network_key, arc_key, mode=self.STATIC_LOSS_MODE_US
                        )

                    elif arc_key[0] not in network.import_nodes:
                        # directed and the departure node is not an import one
                        # >> departure

                        self.place_static_losses(
                            network_key, arc_key, mode=self.STATIC_LOSS_MODE_DEP
                        )

                    elif arc_key[1] not in network.export_nodes:
                        # directed and the arrival node is not an export one
                        # though the departure node is an import one
                        # >> arrival

                        self.place_static_losses(
                            network_key, arc_key, mode=self.STATIC_LOSS_MODE_ARR
                        )

                    else:
                        raise ValueError(
                            "Static losses cannot exist directly between "
                            + "import and export nodes. Use an intermediate arc"
                            + " to obtain the same outcome."
                        )

    # *************************************************************************
    # *************************************************************************

    def place_static_losses_downstream(self):
        """
        Configures the problem instances so that arcs will have their static
        losses placed downstream relative to the flow sense or, when not possi-
        ble, namely if it is a directed arc, in the opposite way. Note that any
        arcs between import and exports are assumed to have been excluded.

        Returns
        -------
        None.

        """

        # clear all configurations

        self.clear_static_loss_configurations()

        # for each arc

        for network_key, network in self.networks.items():
            for arc_key in network.edges(keys=True):
                if network.edges[arc_key][Network.KEY_ARC_TECH].has_static_losses():
                    # if the arc is undirected

                    if network.arc_is_undirected(arc_key):
                        self.place_static_losses(
                            network_key, arc_key, mode=self.STATIC_LOSS_MODE_DS
                        )

                    elif arc_key[1] not in network.export_nodes:
                        # directed and the arrival node is not an export one
                        # >> arrival

                        self.place_static_losses(
                            network_key, arc_key, mode=self.STATIC_LOSS_MODE_ARR
                        )

                    elif arc_key[0] not in network.import_nodes:
                        # directed and the departure node is not an import one
                        # though the arrival node is an export one
                        # >> departure

                        self.place_static_losses(
                            network_key, arc_key, mode=self.STATIC_LOSS_MODE_DEP
                        )

                    else:
                        raise ValueError(
                            "Static losses cannot exist directly between "
                            + "import and export nodes. Use an intermediate arc"
                            + " to obtain the same outcome."
                        )

    # # *************************************************************************

    # def favour_placing_fixed_directed_arc_losses_upstream(self):
    #     """
    #     Sets up the problem to place fixed arc losses upstream.

    #     Returns
    #     -------
    #     None.

    #     """

    #     # initialise the object (if empty, they are all placed upstream)

    #     self.place_fixed_arc_losses_downstream_if_possible = []

    #     # check for exceptions

    #     # for each network

    #     for network_key in self.networks:

    #         # for each arc

    #         for arc_key in self.networks[network_key].edges(keys=True):

    #             # if undirected, continue

    #             if self.networks[network_key].edges[arc_key][
    #                     Network.KEY_ARC_UND]:

    #                 continue

    #             # if directed, make sure it has fixed losses

    #             if not self.networks[network_key].edges[arc_key][
    #                     Network.KEY_ARC_TECH].has_static_losses():

    #             # if not isinstance(self.networks[network_key].edges[arc_key][
    #             #         Network.KEY_ARC_TECH],
    #             #         ArcsWithStaticLosses):

    #                 # if it does not have fixed losses, continue

    #                 continue

    #             # if directed and with fixed losses, check if it involves an
    #             # import arc

    #             if arc_key[0] in self.networks[network_key].import_nodes:

    #                 self.place_fixed_arc_losses_downstream_if_possible.append(
    #                     (network_key,*arc_key)
    #                     )

    #             # *************************************************************

    #         # *****************************************************************

    # # *************************************************************************
    # # *************************************************************************

    # def do_not_place_fixed_losses_downstream(self, network_key, arc_key):
    #     """
    #     Unflags an arc technology from having its fixed losses put downstream.

    #     Parameters
    #     ----------
    #     network_key : hashable-type
    #         The key to the network object.
    #     arc_key : hashable-type
    #         The key to the arc in the specified network.

    #     Returns
    #     -------
    #     None.

    #     """

    #     # check if the tuple exists

    #     if ((network_key, *arc_key) in
    #         self.place_fixed_arc_losses_downstream_if_possible):

    #         # if it does, pop it

    #         self.place_fixed_arc_losses_downstream_if_possible.pop(
    #             (network_key, *arc_key)
    #             )

    #     # if it does not, ignore

    # # *************************************************************************
    # # *************************************************************************

    # def place_fixed_losses_downstream(self, network_key, arc_key):
    #     """
    #     Flags an arc technology as having its fixed losses downstream.

    #     The arc technology must be an instance of ArcsWithStaticLosses
    #     and the end node cannot be an export node (export arcs are excluded).

    #     Parameters
    #     ----------
    #     network_key : hashable-type
    #         The key to the network object.
    #     arc_key : hashable-type
    #         The key to the arc in the specified network.

    #     Raises
    #     ------
    #     ValueError
    #         This error is raised if the arc or the network are not recognised,
    #         or if the arc is not directed or involves an export node.

    #     Returns
    #     -------
    #     None.

    #     """

    #     # the keys must be in the correct format and point to existing objects
    #     # is not possible with export arcs
    #     # is not possible between import nodes and export nodes
    #     # it will be redundant for import arcs, since there is no other way

    #     if network_key in self.networks and len(arc_key) == 3:

    #         # check if the arc exists

    #         if self.networks[network_key].has_edge(*arc_key):

    #             # if it exists, it cannot be undirected

    #             if self.networks[network_key].edges[arc_key][
    #                     Network.KEY_ARC_UND]:

    #                 raise ValueError(
    #                     'The arc selected is not directed.')

    #             # if it exists, it cannot be between import and export nodes

    #             # if it exists, it cannot be an export arc (export node as end)

    #             if arc_key[2] in self.networks[network_key].export_nodes:

    #                 raise ValueError(
    #                     'Export nodes cannot be used to place losses in.')

    #             # if everything is okay, add the (g,u,v,j) tuple to the list
    #             # if it is not already there, otherwise ignore

    #             if ((network_key,*arc_key) not in
    #                 self.place_fixed_arc_losses_downstream_if_possible):

    #                 self.place_fixed_arc_losses_downstream_if_possible.append(
    #                     (network_key,*arc_key)
    #                     )

    #         else: # the arc does not exist

    #             raise ValueError(
    #                 'The arc key used does not match any arc in the network.')

    #     else: # something is up with the network key or arc key length

    #         raise ValueError(
    #             'Either the network key provided is incorrect or the arc key '+
    #             'lacks the proper size.')

    # *************************************************************************
    # *************************************************************************

    def use_interface_variables_for_arc_selection(self, network_key, arc_key):
        """
        Flags a given arc on a given network as requiring interface variables.

        Interface variables are a device introduced to evaluate whether it is
        advantageous to use special ordered sets of type 1 to decide which arcs
        to install while simultaneously using another special ordered set of
        type 1 for flow sense selection.

        Parameters
        ----------
        network_key : hashable-type
            The key to the network object.
        arc_key : hashable-type
            The key to the arc in the specified network.

        Returns
        -------
        None.

        """

        # check if the network and arc exist, including the format

        if network_key in self.networks and len(arc_key) == 3:
            # check if the arc exists

            if self.networks[network_key].has_edge(*arc_key):
                # if it exists, it cannot be preselected

                if (
                    self.networks[network_key]
                    .edges[arc_key][Network.KEY_ARC_TECH]
                    .has_been_selected()
                ):
                    raise ValueError("The arc selected has been preselected.")

                # if so, check if the tuple is already on the list

                if (network_key, *arc_key) not in self.use_arc_interface_variables:
                    # if it is not, add it

                    self.use_arc_interface_variables.append((network_key, *arc_key))

                # if it is, ignore

            else:  # the arc does not exist
                raise ValueError(
                    "The arc key used does not match any arc in the network."
                )

        else:  # something is up with the network key or arc key length
            raise ValueError(
                "Either the network key provided is incorrect or the arc key "
                + "lacks the proper size."
            )

    # *************************************************************************
    # *************************************************************************

    def do_not_use_interface_variables_for_arc_selection(self, network_key, arc_key):
        """
        Sets up the problem to not use interface variables on a given arc.

        Parameters
        ----------
        network_key : hashable-type
            The key to the network object.
        arc_key : hashable-type
            The key to the arc in the specified network.

        Returns
        -------
        None.

        """

        # check if the tuple exists

        if (network_key, *arc_key) in self.use_arc_interface_variables:
            # if it does, pop it
            self.use_arc_interface_variables.remove((network_key, *arc_key))

        # if it does not, ignore

    # *************************************************************************
    # *************************************************************************

    def do_not_use_sos1_for_arc_selection(self, network_key, arc_key):
        """
        Sets up the problem to not use SOS1 on a given arc\'s selection.

        Parameters
        ----------
        network_key : hashable-type
            The key to the network object.
        arc_key : hashable-type
            The key to the arc in the specified network.

        Returns
        -------
        None.

        """

        # check if the tuple exists

        if (network_key, *arc_key) in self.use_sos1_arc_inv:
            # if it does, pop it

            self.use_sos1_arc_inv.pop((network_key, *arc_key))

        # if it does not, ignore

    # *************************************************************************
    # *************************************************************************

    def do_not_use_sos1_for_flow_sense(self, network_key, arc_key):
        """
        Sets up the problem to not use SOS1 to determine flow senses in an arc.

        Parameters
        ----------
        network_key : hashable-type
            The key to the network object.
        arc_key : hashable-type
            The key to the arc in the specified network.

        Returns
        -------
        None.

        """

        # check if the tuple exists

        if (network_key, *arc_key) in self.use_sos1_flow_sense:
            # if it does, pop it

            self.use_sos1_flow_sense.pop((network_key, *arc_key))

        # if it does not, ignore

    # *************************************************************************
    # *************************************************************************

    def use_sos1_for_flow_senses(
        self,
        network_key,
        arc_key,
        use_real_variables_if_possible: bool = True,
        use_interface_variables: bool = True,
        sos1_weight_method: int = SOS1_SENSE_WEIGHT_NOMINAL_HIGHER,
    ):
        """
        Configure the model to use special ordered sets of type 1 (SOS1) for
        selecting the flow sense in a given undirected arc and network.

        Several methods can be used to determine the weights.

        The model can also be configured to avoid using strictly binary variab-
        les, if possible, and to use interface variables to separate arc selec-
        tion from flow sense determination. The former means using non-negative
        real variables for the sense variables and the latter introduces inter-
        mediate variables and equations to separate the two decisions.

        Parameters
        ----------
        network_key : hashable-type
            The key to the network object.
        arc_key : hashable-type
            The key to the arc in the specified network.
        use_real_variables_if_possible : bool, optional
            If True, instances will use real variables for selecting the flow
            sense on this arc, if possible. If False, they will not be used.
            The default is True.
        use_interface_variables : bool, optional
            If True, instances will use interface variables to attempt to de-
            couple the flow sense determination from the arc selection. If Fal-
            se, they will not be used. The default is True.
        sos1_weight_method : int, optional
            The code representing the method used to specify the SOS weights.
            The default is SOS1_SENSE_WEIGHT_NOMINAL_HIGHER, which uses one for
            the nominal direction and zero for the reverse direction. The other
            option is SOS1_SENSE_WEIGHT_REVERSE_HIGHER, which switches values.

        Raises
        ------
        ValueError
            This error is raised if the method is not recognised or if the arc
            is not undirected, since that means the flow sense is known.

        Returns
        -------
        None.

        """

        if sos1_weight_method not in self.SOS1_SENSE_WEIGHTS:
            raise ValueError(
                "The method to determine the SOS weights was not recognised."
            )

        if network_key in self.networks and len(arc_key) == 3:
            # check if the arc exists

            if self.networks[network_key].has_edge(*arc_key):
                # if it exists, it cannot be directed

                if not self.networks[network_key].edges[arc_key][Network.KEY_ARC_UND]:
                    raise ValueError("The arc selected is not undirected.")

                # if so, create a new dict entry or update the old one

                # key: (g,u,v,j)
                # value: sos weights

                self.use_sos1_flow_sense[(network_key, *arc_key)] = sos1_weight_method

                # if so, update the preference for the type of variables used

                if use_real_variables_if_possible:
                    # use real variables if possible

                    if (
                        network_key,
                        *arc_key,
                    ) not in self.use_real_sense_variables_if_possible:
                        # the arc had not been flagged: flag it

                        self.use_real_sense_variables_if_possible.append(
                            (network_key, *arc_key)
                        )

                    # the arc was already flagged: do nothing

                else:  # do not use real variables
                    # check if already exists

                    if (
                        network_key,
                        *arc_key,
                    ) in self.use_real_sense_variables_if_possible:
                        # the arc was already flagged: unflag it

                        self.use_real_sense_variables_if_possible.remove(
                            (network_key, *arc_key)
                        )

                    # the arc was not flagged before: do nothing

                # interface variables

                if use_interface_variables:
                    # use interface variables

                    if (network_key, *arc_key) not in self.use_arc_interface_variables:
                        # the arc had not been flagged: flag it

                        self.use_arc_interface_variables.append((network_key, *arc_key))

                    # the arc was already flagged: do nothing

                else:  # do not use interface variables
                    # check if already exists

                    if (network_key, *arc_key) in self.use_arc_interface_variables:
                        # the arc was already flagged: unflag it

                        self.use_arc_interface_variables.remove((network_key, *arc_key))

            else:  # the arc does not exist
                raise ValueError(
                    "The arc key used does not match any arc in the network."
                )

        else:  # something is up with the network key or arc key length
            raise ValueError(
                "Either the network key provided is incorrect or the arc key "
                + "lacks the proper size."
            )

    # *************************************************************************
    # *************************************************************************

    def use_sos1_for_arc_selection(
        self,
        network_key,
        arc_key,
        use_real_variables_if_possible: bool = True,
        sos1_weight_method: int = SOS1_ARC_WEIGHTS_NONE,
    ):
        """
        Configure the model to use special ordered sets of type 1 (SOS1) for
        selecting which option to invest in a given arc and network.

        Several methods can be used to determine the weights.

        The model can also be configured to avoid using strictly binary variab-
        les, if possible. This means using non-negative real variables, and is
        only possible if the arc options are selectable yet mandatory.

        Parameters
        ----------
        network_key : hashable-type
            The key to the network object.
        arc_key : hashable-type
            The key to the arc in the specified network.
        use_real_variables_if_possible : bool, optional
            If True, instances will use real variables for selecting this arc,
            if possible. If False, they will not be used. The default is True.
        sos1_weight_method : int, optional
            The code representing the method used to specify the SOS1 weights.
            The default is SOS1_ARC_WEIGHTS_NONE, which means no weights will
            be used. All valid methods can be found under SOS1_ARC_WEIGHTS.

        Raises
        ------
        ValueError
            This error is raised if the method is not recognised or if the arc
            has been preselected, since that means no optimisation is needed.

        Returns
        -------
        None.

        """

        if sos1_weight_method not in self.SOS1_ARC_WEIGHTS:
            raise ValueError(
                "The method to determine the SOS1 weights was not recognised."
            )

        if network_key in self.networks and len(arc_key) == 3:
            # check if the arc exists

            if self.networks[network_key].has_edge(*arc_key):
                # if it exists, it cannot be preselected

                if (
                    self.networks[network_key]
                    .edges[arc_key][Network.KEY_ARC_TECH]
                    .has_been_selected()
                ):
                    raise ValueError("The arc selected has been preselected.")

                # if it exists, there must be more than one option

                if (
                    self.networks[network_key]
                    .edges[arc_key][Network.KEY_ARC_TECH]
                    .number_options()
                    <= 1
                ):
                    return  # makes no sense, skip

                # if so, create a new dict entry or update the old one

                # key: (g,u,v,j)
                # value: sos1 weights

                self.use_sos1_arc_inv[
                    (network_key, *arc_key)
                ] = self.compute_arc_sos1_weights(
                    self.networks[network_key].edges[arc_key][Network.KEY_ARC_TECH],
                    method=sos1_weight_method,
                )

                # if so, update the preference for the type of variables used

                if use_real_variables_if_possible:
                    # use real variables if possible

                    if (
                        network_key,
                        *arc_key,
                    ) not in self.use_real_arc_inv_variables_if_possible:
                        # the arc had not been flagged: flag it

                        self.use_real_arc_inv_variables_if_possible.append(
                            (network_key, *arc_key)
                        )

                    # the arc was already flagged: do nothing

                else:  # do not use real variables
                    # check if already exists

                    if (
                        network_key,
                        *arc_key,
                    ) in self.use_real_arc_inv_variables_if_possible:
                        # the arc was already flagged: unflag it

                        self.use_real_arc_inv_variables_if_possible.remove(
                            (network_key, *arc_key)
                        )

                    # the arc was not flagged before: do nothing

            else:  # the arc does not exist
                raise ValueError(
                    "The arc key used does not match any arc in the network."
                )

        else:  # something is up with the network key or arc key length
            raise ValueError(
                "Either the network key provided is incorrect or the arc key "
                + "lacks the proper size."
            )

    # *************************************************************************
    # *************************************************************************

    def compute_arc_sos1_weights(
        self,
        arcs: Arcs,
        method: int = SOS1_ARC_WEIGHTS_CAP,
        verify_weights: bool = False,
    ) -> tuple:
        if method not in self.SOS1_ARC_WEIGHTS:
            raise ValueError(
                "The method to determine the SOS1 weights was not recognised."
            )
        if method == self.SOS1_ARC_WEIGHTS_CAP:
            sos1_weights = tuple(capacity for capacity in arcs.capacity)
        elif method == self.SOS1_ARC_WEIGHTS_COST:
            sos1_weights = tuple(cost for cost in arcs.minimum_cost)
        elif method == self.SOS1_ARC_WEIGHTS_SPEC_CAP:
            sos1_weights = tuple(
                cap / cost for cap, cost in zip(arcs.capacity, arcs.minimum_cost)
            )
        elif method == self.SOS1_ARC_WEIGHTS_SPEC_COST:
            sos1_weights = tuple(
                cost / cap for cap, cost in zip(arcs.capacity, arcs.minimum_cost)
            )
        else:  # SOS1_ARC_WEIGHTS_NONE
            return None

        # make sure they are unique
        if verify_weights:
            for weight in sos1_weights:
                if sos1_weights.count(weight) >= 2:  # TODO: reach this point
                    raise ValueError(
                        "These arcs cannot be considered using "
                        + "special ordered sets of type 1 (SOS1),"
                        + " since some weights are not unique."
                    )
        return sos1_weights

    # *************************************************************************
    # *************************************************************************

    def prepare(self, name = None):
        """Sets up the problem model with which instances can be built."""
        # create pyomo model (AbstractModel)
        self.model = create_model(
            name, 
            use_prices_block=self.use_prices_block,
            node_price_model=self.node_price_model
            )

    # *************************************************************************
    # *************************************************************************

    def instantiate(
        self, pyomo_dict: dict = None, initialise_ancillary_sets: bool = False
    ):
        """Instantiates the model using the data available or provided."""

        # check if the (abstract) model already exists

        try:
            assert type(self.model) == pyo.AbstractModel

        except AttributeError:
            raise ValueError("The problem model does not exist.")

        except AssertionError:
            raise TypeError("The problem model is not a pyomo AbstractModel.")

        if type(pyomo_dict) == dict:
            # use external dict to create pyomo model instance (ConcreteModel)

            self.instance = self.model.create_instance(pyomo_dict)

        else:
            # make pre-instantiation preparations

            # generate pyomo-ready dictionary

            pyomo_dict = self.create_pyomo_dictionary(
                include_ancillary_sets=initialise_ancillary_sets
            )

            # create pyomo model instance (ConcreteModel)

            self.instance = self.model.create_instance(pyomo_dict)

    # *************************************************************************
    # *************************************************************************

    def optimise(
        self,
        solver_name: str,
        solver_options: dict,
        output_options: dict,
        print_solver_output: bool = True,
    ):
        """Optimise the problem using pyomo and the specifications provided."""

        # check if the instance already exists

        try:
            assert type(self.instance) == pyo.ConcreteModel

        except AttributeError:
            raise ValueError("The problem instance does not exist.")

        except AssertionError:
            raise TypeError("The problem instance is not a pyomo ConcreteModel.")

        # ensure compatibility between problem and solver

        self.assert_solver_problem_compatibility(solver_name)

        # configure common solver interface

        self.solver_interface = SolverInterface(
            solver_name=solver_name, **solver_options
        )

        # get the solver handler

        self.solver_handler = self.solver_interface.get_solver_handler(**solver_options)

        # solve

        self.results = self.solver_handler.solve(self.instance, tee=print_solver_output)

        # post-optimisation

        if self.solver_interface.was_optimisation_sucessful(
            self.results, self.optimisation_problem_type
        ):
            # if successful, import results

            self.import_results()

            return True

        return False

        # TODO: reach this statement, perhaps using a small solver budget

    # *************************************************************************
    # *************************************************************************

    def create_pyomo_dictionary(self, include_ancillary_sets: bool = False) -> dict:
        """Returns a dictionary with which the pyomo model can be instantiated."""

        print("creating the dictionary...")

        # *********************************************************************
        # *********************************************************************
        # *********************************************************************
        # *********************************************************************

        # sets

        # *********************************************************************
        # *********************************************************************
        # *********************************************************************
        # *********************************************************************

        # time

        # set of representative periods

        set_Q = tuple(self.time_frame.assessments) # tuple(self.assessment_keys)

        # set of representative periods

        set_P_q = {q: tuple(p for p in self.time_frame.reporting_periods[q]) for q in set_Q}

        # set of time intervals
        
        set_K_q = {
            q: tuple(k for k in self.time_frame.time_intervals[q]) for q in set_Q
        }
        
        # set of (q,p) tuples

        set_QP = [(q, p) for q in set_Q for p in set_P_q[q]]

        # set of (q,k) tuples

        set_QK = [(q, k) for q in set_Q for k in set_K_q[q]]

        # set of (q,p,k) tuples

        set_QPK = [(q, p, k) for (q, p) in set_QP for k in set_K_q[q]]

        # *********************************************************************

        # set of networks

        set_G = [g for g in self.networks.keys()]

        # set of nodes per grid

        set_L = {
            g: tuple(l for l in self.networks[g].nodes) for g in self.networks.keys()
        }

        # set of importing nodes per grid

        set_L_imp = {
            g: tuple(
                l for l in self.networks[g].nodes if l in self.networks[g].import_nodes
            )
            for g in self.networks.keys()
        }

        # set of exporting nodes per grid

        set_L_exp = {
            g: tuple(
                l for l in self.networks[g].nodes if l in self.networks[g].export_nodes
            )
            for g in self.networks.keys()
        }

        set_L_max_in_g = {
            g: tuple(self.networks[g].nodes_w_in_dir_arc_limitations.keys())
            for g in self.networks.keys()
            }

        set_L_max_out_g = {
            g: tuple(self.networks[g].nodes_w_out_dir_arc_limitations.keys())
            for g in self.networks.keys()
            }

        set_GL = tuple((g, l) for g in set_G for l in set_L[g])

        set_GL_imp = tuple((g, l) for (g, l) in set_GL if l in set_L_imp[g])

        set_GL_exp = tuple((g, l) for (g, l) in set_GL if l in set_L_exp[g])

        set_GL_exp_imp = tuple(
            (g, l) for (g, l) in set_GL if (l in set_L_imp[g] or l in set_L_exp[g])
        )

        set_GL_not_exp_imp = tuple(gl for gl in set_GL if gl not in set_GL_exp_imp)

        set_GLL = tuple(
            (g, l1, l2)
            for g in set_G
            for l1 in set_L[g]
            if l1 not in set_L_exp[g]
            for l2 in set_L[g]
            if l2 not in set_L_imp[g]
            if l1 != l2
        )

        # *********************************************************************

        # set of price segments
        set_S = {
            (g, l, q, p, k): tuple(
                s
                for s in range(
                    self.networks[g]
                    .nodes[l][Network.KEY_NODE_PRICES][(q, p, k)]
                    .number_segments()
                )
            )
            for (g, l) in set_GL_exp_imp
            for (q, p, k) in set_QPK
        }
        
        # set of GLKS tuples
        set_GLQPKS = tuple(
            (*glqpk, s) for glqpk, s_tuple in set_S.items() for s in s_tuple
        )

        set_GLQPKS_exp = tuple(
            glqpks for glqpks in set_GLQPKS if glqpks[1] in set_L_exp[glqpks[0]]
        )

        set_GLQPKS_imp = tuple(
            glqpks for glqpks in set_GLQPKS if glqpks[1] in set_L_imp[glqpks[0]]
        )

        # *********************************************************************

        # dynamic systems

        # set of converters

        set_I = [key for key in self.converters.keys()]

        # set of optional converters

        set_I_new = (
            self.optional_converters
        )  # or [key for key in self.optional_converters]

        # *********************************************************************

        # inputs

        # set of inputs

        set_M = {
            converter_key: [m for m in range(converter.number_inputs)]
            for converter_key, converter in self.converters.items()
        }

        # set of inputs that need to be modelled as binary variables

        set_M_bin = {
            converter_key: converter.binary_inputs
            for converter_key, converter in self.converters.items()
        }

        # set of input signals whose amplitude is to be dimensioned

        set_M_dim = {
            converter_key: converter.dimensionable_inputs
            for converter_key, converter in self.converters.items()
            if converter_key in self.optional_converters
        }

        # set of inputs whose amplitude has a cost

        set_M_cost = {
            converter_key: converter.amplitude_penalised_inputs
            for converter_key, converter in self.converters.items()
            for converter_key in self.optional_converters
        }

        # set of inputs that can cause externalities

        set_M_ext = {
            converter_key: converter.externality_inducing_inputs
            for converter_key, converter in self.converters.items()
        }

        # *********************************************************************

        # set of outputs

        set_R = {
            converter_key: [r for r in range(converter.number_outputs)]
            for converter_key, converter in self.converters.items()
        }

        # set of outputs whose amplitude is to be dimensioned

        set_R_dim = {
            converter_key: converter.dimensionable_outputs
            for converter_key, converter in self.converters.items()
            if converter_key in self.optional_converters
        }

        # set of outputs whose amplitude has a cost

        set_R_cost = {
            converter_key: converter.amplitude_penalised_outputs
            for converter_key, converter in self.converters.items()
            for converter_key in self.optional_converters
        }

        # set of outputs inducing externalities

        set_R_ext = {
            converter_key: converter.externality_inducing_outputs
            for converter_key, converter in self.converters.items()
        }

        # *********************************************************************

        # states

        # set of states

        set_N = {
            converter_key: [n for n in range(converter.number_states)]
            for converter_key, converter in self.converters.items()
        }

        # set of state nodes whose upper limit is dimensionable, per system

        set_N_dim_up = {}

        # set of state nodes whose lower limit is dimensionable, per system

        set_N_dim_lo = {}

        # set of state nodes whose upper limit can induce costs, per system

        set_N_cost_up = {}

        # set of state nodes whose lower limit can induce costs, per system

        set_N_cost_lo = {}

        # states inducing externalities

        set_N_ext = {
            converter_key: converter.externality_inducing_states
            for converter_key, converter in self.converters.items()
        }

        # *********************************************************************

        # arcs

        # set of arcs

        set_J = {
            (g, u, v): [
                j for j, _ in self.networks[g]._adj[u][v].items()  # the arc key
            ]
            for g in self.networks.keys()
            for u, v in self.networks[g].edges()
        }

        # set of undirected arcs

        set_J_und = {
            (g, u, v): [
                j  # all arcs that are undirected
                for j in set_J[(g, u, v)]
                if self.networks[g].arc_is_undirected(arc_key=(u, v, j))
            ]
            for (g, u, v) in set_J
        }

        # TODO: replace this with validation rule in the pyomo model, asap

        for g, u, v in set_J_und:
            for j in set_J_und[(g, u, v)]:
                if u in set_L_imp[g] or v in set_L_imp[g]:
                    raise ValueError("Undirected arcs cannot involve import nodes.")

                if u in set_L_exp[g] or v in set_L_exp[g]:
                    raise ValueError("Undirected arcs cannot involve export nodes.")

        # set of predefined arcs

        set_J_pre = {
            (g, u, v): [
                j
                for j in set_J[(g, u, v)]
                # only if it has not been preselected
                if True
                in self.networks[g]
                .edges[(u, v, j)][Network.KEY_ARC_TECH]
                .options_selected
            ]
            for (g, u, v) in set_J
        }

        # set of predefined infinite capacity arcs

        set_J_pre_inf = {
            (g, u, v): [
                j
                for j in set_J_pre[(g, u, v)]
                if j not in set_J_und[(g, u, v)]  # can only be directed
                # the capacity must be set to math.inf
                if self.networks[g]
                .edges[(u, v, j)][Network.KEY_ARC_TECH]
                .capacity[
                    self.networks[g]
                    .edges[(u, v, j)][Network.KEY_ARC_TECH]
                    .options_selected.index(True)
                ]
                == math.inf
            ]
            for (g, u, v) in set_J_pre
        }

        # sets of collectively-selected groups

        set_GLLJ_col = set(
            arc_tuple
            for group_of_arcs in self.arc_groups.values()
            for arc_tuple in group_of_arcs
        )

        set_J_col = {
            (g, u, v): [
                j
                for j in set_J[(g, u, v)]
                if j not in set_J_pre[(g, u, v)]  # new
                if (g, u, v, j) in set_GLLJ_col
            ]  # can only appear once
            for (g, u, v) in set_J
        }

        # set of selectable yet mandatory arcs

        set_J_mdt = {
            (g, u, v): [
                j
                for j in set_J[(g, u, v)]
                if j not in set_J_pre[(g, u, v)]  # new
                if j not in set_J_col[(g, u, v)]  # individual
                if (g, u, v, j) in self.mandatory_arcs
            ]
            for (g, u, v) in set_J
            # if (g,u,v) not in set_J_pre
        }

        # arcs using interfaces (new and optional arcs only)

        set_J_int = {
            (g, u, v): [
                j  # all arcs that are to use interface variables
                for j in set_J[(g, u, v)]
                if j not in set_J_pre[(g, u, v)]  # new
                if j not in set_J_mdt[(g, u, v)]  # optional
                if j not in set_J_col[(g, u, v)]  # individual
                if (g, u, v, j) in self.use_arc_interface_variables
            ]
            for (g, u, v) in set_J
        }

        # set of selectable arcs whose investment decisions are modelled as sos

        set_J_arc_sos1 = {
            (g, u, v): [
                j
                for j in set_J[(g, u, v)]
                if j not in set_J_pre[(g, u, v)]  # new
                if j not in set_J_col[(g, u, v)]  # individual
                if (g, u, v, j) in self.use_sos1_arc_inv
                # if isinstance(self.networks[g].edges[(u,v,j)][
                #         Network.KEY_ARC_TECH], ArcsSOS)
            ]
            for (g, u, v) in set_J
            # if (g,u,v) not in set_J_pre
        }

        # set of selectable arcs whose investment decisions are modelled as nnr

        set_J_arc_nnr = {
            (g, u, v): [
                j
                for j in set_J_arc_sos1[(g, u, v)]
                if (g, u, v) in set_J_mdt
                if j in set_J_mdt[(g, u, v)]
                if ((g, u, v, j) in self.use_real_arc_inv_variables_if_possible)
            ]
            for (g, u, v) in set_J_arc_sos1
            # if (g,u,v) in set_J_mdt
        }

        # set of undirected arcs whose flow sense is modelled using sos1

        set_J_sns_sos1 = {
            (g, u, v): [
                j
                for j in set_J_und[(g, u, v)]
                if (g, u, v, j) in self.use_sos1_flow_sense
            ]
            for (g, u, v) in set_J_und
        }

        # set of undirected arcs whose flow sense is modelled using nnr

        set_J_sns_nnr = {
            (g, u, v): [
                j
                for j in set_J_sns_sos1[(g, u, v)]
                if (g, u, v, j) in self.use_real_sense_variables_if_possible
            ]
            for (g, u, v) in set_J_sns_sos1
        }

        # set of arcs with fixed losses
        # note: there cannot be arcs with fixed losses between import and ex-
        # port nodes, as the nodes cannot be used to place the losses in

        set_J_stt = {
            (g, u, v): [
                j
                for j in set_J[(g, u, v)]
                if self.networks[g]
                .edges[(u, v, j)][Network.KEY_ARC_TECH]
                .has_static_losses()
            ]
            for (g, u, v) in set_J
        }

        # TODO: replace this with validation rule in the pyomo model, asap

        for u in self.networks[g].import_nodes:
            for v in self.networks[g].export_nodes:
                if (g, u, v) in set_J_stt and len(set_J_stt[(g, u, v)]) != 0:
                    raise ValueError(
                        "There cannot be arcs with fixed losses between import"
                        + " and export nodes."
                    )

        # for (g,u,v) in set_J_stt:

        #     for j in set_J_stt[(g,u,v)]:

        #         if (u in self.networks[g].import_nodes and
        #             v in self.networks[g].export_nodes):

        #             raise ValueError(
        #                 'There cannot be arcs with fixed losses between import'
        #                 +' and export nodes.'
        #                 )

        # directed arcs:
        # 1) in the start node (also upstream)
        # 2) in the end node (also downstream)
        # undirected arcs:
        # 1) in the (conventional) start node
        # 2) in the (conventional) end node
        # 3) upstream
        # 4) downstream
        # sets:
        # static_a: directed and undirected
        # static_b: directed and undirected
        # static_us: undirected
        # static_ds: undirected

        # set of (directed) arcs with static losses in the departure node

        set_J_stt_dep = {
            (g, u, v): [
                j
                for j in set_J_stt[(g, u, v)]
                if (g, u, v) in self.static_losses_departure_node
                if j in self.static_losses_departure_node[(g, u, v)]
            ]
            for (g, u, v) in set_J_stt
            # if u in self.networks[g].import_nodes
            # if v not in self.networks[g].export_nodes
        }

        # TODO: replace this with validation rule in the pyomo model, asap

        for g, u, v in set_J_stt_dep:
            for j in set_J_stt_dep[(g, u, v)]:
                if u in self.networks[g].import_nodes:
                    raise ValueError(
                        "Static losses cannot be placed upstream if the start "
                        + "node is an import node."
                    )

        # set of (directed) arcs with static losses in the arrival node
        # note: export arcs cannot have fixed losses placed downstream; if ex-
        # port arcs are found in place_fixed_arc_losses_downstream_if_possible
        # they will not be modelled as having their losses upstream (default)

        set_J_stt_arr = {
            (g, u, v): [
                j
                for j in set_J_stt[(g, u, v)]
                if (g, u, v) in self.static_losses_arrival_node
                if j in self.static_losses_arrival_node[(g, u, v)]
            ]
            for (g, u, v) in set_J_stt
            # if u in self.networks[g].import_nodes
            # if v not in self.networks[g].export_nodes
        }

        # TODO: replace this with validation rule in the pyomo model, asap

        for g, u, v in set_J_stt_arr:
            for j in set_J_stt_arr[(g, u, v)]:
                if v in self.networks[g].export_nodes:
                    raise ValueError(
                        "Static losses cannot be placed downstream if the end "
                        + "node is an export node."
                    )

        # set of (directed) arcs with static losses upstream

        set_J_stt_us = {
            (g, u, v): [
                j
                for j in set_J_stt[(g, u, v)]
                if (g, u, v) in self.static_losses_upstream
                if j in self.static_losses_upstream[(g, u, v)]
            ]
            for (g, u, v) in set_J_stt
            # if u in self.networks[g].import_nodes
            # if v not in self.networks[g].export_nodes
        }

        # TODO: replace this with validation rule in the pyomo model, asap

        for g, u, v in set_J_stt_us:
            for j in set_J_stt_us[(g, u, v)]:
                if (
                    u in self.networks[g].import_nodes
                    or v in self.networks[g].export_nodes
                ):
                    raise ValueError(
                        "Undirected arcs cannot connect import or export nodes" + "."
                    )

        # set of (directed) arcs with static losses downstream

        set_J_stt_ds = {
            (g, u, v): [
                j
                for j in set_J_stt[(g, u, v)]
                if (g, u, v) in self.static_losses_downstream
                if j in self.static_losses_downstream[(g, u, v)]
            ]
            for (g, u, v) in set_J_stt
            # if u in self.networks[g].import_nodes
            # if v not in self.networks[g].export_nodes
        }

        # TODO: replace this with validation rule in the pyomo model, asap

        for g, u, v in set_J_stt_ds:
            for j in set_J_stt_ds[(g, u, v)]:
                if (
                    u in self.networks[g].import_nodes
                    or v in self.networks[g].export_nodes
                ):
                    raise ValueError(
                        "Undirected arcs cannot connect import or export nodes" + "."
                    )

        # TODO: make sure the stt sets are all mutually-exclusive

        for guv, j in set_J_stt.items():
            # a and b, a and us, a and ds, b and us, b and ds, us and ds

            if (
                (j in set_J_stt_dep[guv] and j in set_J_stt_arr[guv])
                or (j in set_J_stt_dep[guv] and j in set_J_stt_us[guv])
                or (j in set_J_stt_dep[guv] and j in set_J_stt_ds[guv])
                or (j in set_J_stt_arr[guv] and j in set_J_stt_us[guv])
                or (j in set_J_stt_arr[guv] and j in set_J_stt_ds[guv])
                or (j in set_J_stt_us[guv] and j in set_J_stt_ds[guv])
            ):
                raise ValueError(
                    "Each arc with static losses can only be modelled one way."
                )

        # set of options for a given arc

        set_H_gllj = {
            (g, u, v, j): tuple(
                h
                for h in range(
                    len(
                        self.networks[g].edges[(u, v, j)][Network.KEY_ARC_TECH].capacity
                    )
                )
            )
            for g in self.networks.keys()
            for u, v in self.networks[g].edges()
            for j in set_J[(g, u, v)]
            if j not in set_J_pre[(g, u, v)]  # new
            if j not in set_J_col[(g, u, v)]  # individual
        }

        # *********************************************************************
        # *********************************************************************
        # *********************************************************************

        # ancillary sets

        set_GLLJ = tuple((g, u, v, j) for (g, u, v) in set_J for j in set_J[(g, u, v)])

        set_GLLJ_und = tuple(
            (g, u, v, j)
            for (g, u, v, j) in set_GLLJ
            for (g, u, v) in set_J_und
            if j in set_J_und[(g, u, v)]
        )

        set_GLLJ_und_ext = set((g, v, u, j) for (g, u, v, j) in set_GLLJ_und)

        set_GLLJ_new = tuple(
            (g, l1, l2, j)
            for (g, l1, l2, j) in set_GLLJ
            if j not in set_J_pre[(g, l1, l2)]
        )

        set_GLLJ_int = tuple(
            (g, u, v, j) for (g, u, v) in set_J_int for j in set_J_int[(g, u, v)]
        )

        set_GLLJ_static = tuple(
            (g, l1, l2, j) for (g, l1, l2) in set_J_stt for j in set_J_stt[(g, l1, l2)]
        )

        set_GLLJ_static_pre = tuple(
            (g, l1, l2, j)
            for (g, l1, l2, j) in set_GLLJ_static
            if j in set_J_pre[(g, l1, l2)]
        )

        set_GLLJ_static_new = tuple(
            (g, l1, l2, j)
            for (g, l1, l2, j) in set_GLLJ_static
            if j not in set_J_pre[(g, l1, l2)]
        )

        # red sets

        set_GLLJ_red = set(set_GLLJ)
        set_GLLJ_red.update(set_GLLJ_und_ext)

        set_GLLJ_und_red = set(set_GLLJ_und)
        set_GLLJ_und_red.update(set_GLLJ_und_ext)

        set_GLLJ_static_und_red = set(
            gllj for gllj in set_GLLJ_und if gllj in set_GLLJ_static
        )

        set_GLLJ_static_und_red.update(
            tuple((g, v, u, j) for (g, u, v, j) in set_GLLJ_static_und_red)
        )

        # *********************************************************************
        # *********************************************************************

        # set of arc groups

        set_T = tuple(t for t in self.arc_groups.keys())

        # set of mandatory arc groups

        set_T_mdt = tuple(t for t in set_T if self.groups_mdt[t])

        # set of arc groups requiring interface variables

        set_T_int = tuple(t for t in set_T if self.groups_int[t])

        # set of arc groups relying on SOS1

        set_T_sos1 = tuple(t for t in set_T if t in self.groups_arc_sos1)

        # set of arg groups relying on non-negative real variables

        set_T_nnr = tuple(t for t in set_T if self.groups_arc_nnr[t])

        # set of arc groups relying on binary variables

        set_T_bin = tuple(t for t in set_T if t not in set_T_nnr)

        # set of arcs in the various arc groups

        # set_GLLJ_col_t =  {
        #     t: self.arc_groups[t]
        #     for t in set_T
        #     }

        set_GLLJ_col_t = self.arc_groups

        # set of arc options for arc groups

        set_H_t = {
            t: tuple(h for h in range(self.groups_number_options[t])) for t in set_T
        }

        # *********************************************************************
        # *********************************************************************
        # *********************************************************************
        # *********************************************************************

        # parameters

        # *********************************************************************
        # *********************************************************************
        # *********************************************************************
        # *********************************************************************

        # objective function

        # *********************************************************************
        # *********************************************************************

        # assessment weight

        if type(self.assessment_weights) != dict:
            param_c_wgt_q = {}

        else:
            param_c_wgt_q = self.assessment_weights

        # weight of each time interval k within the period p

        if type(self.time_weights) != dict:
            param_c_time_qpk = {}  # rely on default values

        else:
            # param_c_time_qpk = {
            #     (q, p, k): self.time_weights[(q,p,k)]
            #     for (q,p) in set_QP
            #     for k in set_K_q[q]
            #     }

            param_c_time_qpk = self.time_weights

        # discount factors for each period

        param_c_df_qp = {
            (q, p): self.time_frame.discount_factor(q, p)
            # (q, p): discount_factor(self.discount_rates[q][0 : p + 1])
            for (q, p) in set_QP
        }

        # prices

        param_p_glqpks = {
            (g, l, q, p, k, s): self.networks[g]
            .nodes[l][Network.KEY_NODE_PRICES][(q, p, k)]
            .prices[s]
            for (g, l, q, p, k) in set_S
            for s in set_S[(g, l, q, p, k)]
        }

        # price function convexity

        param_price_function_is_convex = {
            (g, l, q, p, k): (
                self.networks[g].nodes[l][Network.KEY_NODE_PRICES][(q, p, k)].price_monotonically_increasing_with_volume()
                if l in set_L_imp[g] else
                self.networks[g].nodes[l][Network.KEY_NODE_PRICES][(q, p, k)].price_monotonically_decreasing_with_volume()
                )
            for (g, l, q, p, k) in set_S
        }

        # maximum resource volume per segment (infinity is the default)
        
        param_v_max_glqpks = {
            (g, l, q, p, k, s): self.networks[g]
            .nodes[l][Network.KEY_NODE_PRICES][(q, p, k)]
            .volumes[s]
            for (g, l, q, p, k) in set_S
            for s in set_S[(g, l, q, p, k)]
            # skip if not numeric, since that means infinity (=default)
            if isinstance(
                self.networks[g]
                .nodes[l][Network.KEY_NODE_PRICES][(q, p, k)]
                .volumes[s],
                Real,
            )
        }

        # *********************************************************************
        # *********************************************************************

        # converters

        # *********************************************************************
        # *********************************************************************

        # objective function

        # *********************************************************************

        # minimum cost

        param_c_cvt_i = {
            converter_key: converter.minimum_cost
            for converter_key, converter in self.converters.items()
        }

        # input specific amplitude costs

        param_c_cvt_u_im = {
            (converter_key, m): converter.specific_input_amplitude_cost[m]
            for converter_key, converter in self.converters.items()
            if converter_key in self.optional_converters
            for m in converter.amplitude_penalised_inputs
        }

        # state specific amplitude costs

        param_c_cvt_x_in = {
            (converter_key, n): converter.specific_state_amplitude_cost[n]
            for converter_key, converter in self.converters.items()
            if converter_key in self.optional_converters
            for n in converter.amplitude_penalised_states
        }

        # output specific amplitude costs

        param_c_cvt_y_ir = {
            (converter_key, r): converter.specific_output_amplitude_cost[r]
            for converter_key, converter in self.converters.items()
            if converter_key in self.optional_converters
            for r in converter.amplitude_penalised_outputs
        }

        # input specific externality costs

        param_c_ext_u_imqk = {
            (converter_key, m, q, k): converter.specific_input_externality_cost[
                (m, q, k)
            ]
            for converter_key, converter in self.converters.items()
            for m in converter.externality_inducing_inputs
            for (q, k) in set_QK
        }

        # state specific externality costs

        param_c_ext_x_inqk = {
            (converter_key, n, q, k): converter.specific_state_externality_cost[
                (n, q, k)
            ]
            for converter_key, converter in self.converters.items()
            for n in converter.externality_inducing_states
            for (q, k) in set_QK
        }

        # output specific externality costs

        param_c_ext_y_irqk = {
            (converter_key, r, q, k): converter.specific_output_externality_cost[
                (r, q, k)
            ]
            for converter_key, converter in self.converters.items()
            for r in converter.externality_inducing_outputs
            for (q, k) in set_QK
        }

        # *********************************************************************

        # inputs

        # *********************************************************************

        # upper bound for input signals

        param_u_ub_imqk = {
            (converter_key, m, q, k): converter.u_upper_bounds[k]
            for converter_key, converter in self.converters.items()
            for m in converter.bounded_inputs
            # if m not in converter.u_is_dimensionable # cannot be dimensionable
            # if m not in converter.u_is_binary # cannot be binary
            for (q, k) in set_QK
        }

        # maximum amplitude for inputs

        param_u_lim_max_im = {
            (converter_key, m): converter.u_max_amplitude[m]
            for converter_key, converter in self.converters.items()
            if converter_key in self.optional_converters
            for m in converter.dimensionable_inputs
        }

        # u rated adjustment coefficient (1 if the time intervals are regular)

        param_f_amp_u_imqk = {}

        # if self.intraperiod_time_intervals_are_regular:

        #     # an empty dictionary is enough because the default value is 1

        #     param_f_amp_u_imqk = {}

        # else:

        #     # TODO: test irregular time intervals

        #     param_f_amp_u_imqk = {}

        # *********************************************************************

        # outputs

        # *********************************************************************

        # upper bounds for outputs (default: none)

        param_y_ub_irqk = {
            # (converter_key, r, q, k): converter.u_upper_bounds[k]
            # for converter_key, converter in self.converters.items()
            # for r in converter.bounded_outputs
            # for k in set_K
        }

        # lower bounds for outputs (default: 0)

        param_y_lb_irqk = {
            # (converter_key, r, q, k): converter.u_upper_bounds[k]
            # for converter_key, converter in self.converters.items()
            # for r in converter.bounded_inputs
            # for k in set_K
        }

        # maximum positive amplitude for inputs

        param_y_pos_lim_max_ir = {
            # (converter_key, m): converter.u_max_amplitude[m]
            # for converter_key, converter in self.converters.items()
            # if converter_key in self.optional_converters
            # for m in converter.dimensionable_inputs
        }

        # minimum positive amplitude for inputs

        param_y_pos_lim_min_ir = {
            # (converter_key, m): converter.u_min_amplitude[m]
            # for converter_key, converter in self.converters.items()
            # if converter_key in self.optional_converters
            # for m in converter.dimensionable_inputs
            # # minimum amplitude limits are not mandatory, the default is zero
            # if (converter.u_min_amplitude[m] != 0 or
            #     converter.u_min_amplitude[m] != None)
        }

        # maximum negative amplitude for inputs

        param_y_neg_lim_max_ir = {
            # (converter_key, m): converter.u_max_amplitude[m]
            # for converter_key, converter in self.converters.items()
            # if converter_key in self.optional_converters
            # for m in converter.dimensionable_inputs
        }

        # minimum negative amplitude for inputs

        param_y_neg_lim_min_ir = {
            # (converter_key, m): converter.u_min_amplitude[m]
            # for converter_key, converter in self.converters.items()
            # if converter_key in self.optional_converters
            # for m in converter.dimensionable_inputs
            # # minimum amplitude limits are not mandatory, the default is zero
            # if (converter.u_min_amplitude[m] != 0 or
            #     converter.u_min_amplitude[m] != None)
        }

        # u rated adjustment coefficient (1 if the time intervals are regular)

        param_f_amp_y_irqk = {}

        # if self.intraperiod_time_intervals_are_regular:

        #     # an empty dictionary is enough because the default value is 1

        #     param_f_amp_y_irqk = {}

        # else:

        #     # TODO: test irregular time intervals

        #     param_f_amp_y_irqk = {}

        # output equations: C matrix coefficients
        param_c_eq_y_irnqk = {
            (cvt_key, *rnqk): c_rnqk
            for cvt_key, cvt in self.converters.items()
            for rnqk, c_rnqk in cvt.c_rnqk.items()
        }

        # output equations: D matrix coefficients
        param_d_eq_y_irmqk = {
            (cvt_key, *rmqk): d_rmqk
            for cvt_key, cvt in self.converters.items()
            for rmqk, d_rmqk in cvt.d_rmqk.items()
        }

        # output equations: constant term
        param_e_eq_y_irqk = {
            (cvt_key, *rqk): e_rqk
            for cvt_key, cvt in self.converters.items()
            for rqk, e_rqk in cvt.e_rqk.items()
        }

        # *********************************************************************

        # states
        # state equations: A matrix coefficients
        param_a_eq_x_innqk = {
            (cvt_key, *nnqk): a_nnqk
            for cvt_key, cvt in self.converters.items()
            for nnqk, a_nnqk in cvt.a_nnqk.items()
        }

        # state equations: B matrix coefficients
        param_b_eq_x_inmqk = {
            (cvt_key, *nmqk): b_nmqk
            for cvt_key, cvt in self.converters.items()
            for nmqk, b_nmqk in cvt.b_nmqk.items()
        }

        # state equations: constant term
        param_e_eq_x_inqk = {
            (cvt_key, *nqk): e_nqk
            for cvt_key, cvt in self.converters.items()
            for nqk, e_nqk in cvt.e_nqk.items()
        }

        # initial states
        param_x_inq0 = {}

        # upper bounds for states (default: none)
        param_x_ub_inqk = {}

        # lower bounds for states (default: none)
        param_x_lb_inqk = {}

        # maximum positive amplitude for inputs

        param_x_pos_lim_max_in = {}

        # minimum positive amplitude for inputs

        param_x_pos_lim_min_in = {}

        # maximum negative amplitude for inputs

        param_x_neg_lim_max_in = {}

        # minimum negative amplitude for inputs

        param_x_neg_lim_min_in = {}

        # u rated adjustment coefficient (1 if the time intervals are regular)

        param_f_amp_x_inqk = {}

        # if self.intraperiod_time_intervals_are_regular:

        #     # an empty dictionary is enough because the default value is 1

        #     param_f_amp_x_inqk = {}

        # else:

        #     # TODO: test irregular time intervals

        #     param_f_amp_x_inqk = {}

        # *********************************************************************
        # *********************************************************************

        # nodes

        # *********************************************************************
        # *********************************************************************

        # static balance for nodes (default: 0)

        param_v_base_glqk = {
            (g, l, q, k): self.networks[g].nodes[l][Network.KEY_NODE_BASE_FLOW][(q, k)]
            if l not in self.networks[g].waypoint_nodes
            else 0
            for (g, l) in set_GL_not_exp_imp
            # for g in self.networks.keys()
            # for l in self.networks[g].nodes
            # if l not in self.networks[g].import_nodes
            # if l not in self.networks[g].export_nodes
            # if l not in self.networks[g].waypoint_nodes # relies on default (0)
            for (q, k) in set_QK
        }

        # maximum number of parallel arcs (default in pyomo: 1)

        param_max_number_parallel_arcs = {
            (g, u, v): self.max_number_parallel_arcs[(g, u, v)]
            for (g, u, v) in self.max_number_parallel_arcs
            if ((g, u, v) in set_J or (g, v, u) in set_J)
        }

        # effect of converter inputs on networks

        param_a_nw_glimqk = {
            # (grid, node_key, converter_key, m, k): a_glmk
            # for converter_key, converter in self.converters.items()
            # for (grid, node_key, m, k), a_glmk in dynsys.a_glmk.items()
        }

        # effect of converter outputs on networks

        param_a_nw_glirqk = {
            # (grid, node_key, dynsys_key, m, k): a_glmk
            # for dynsys_key, dynsys in self.dynamic_systems.items()
            # for (grid, node_key, m, k), a_glmk in dynsys.a_glmk.items()
        }

        # *********************************************************************
        # *********************************************************************

        # arcs

        # *********************************************************************
        # *********************************************************************

        # nominal flow amplitude limit adjustment coefficient (1 if the time intervals are regular)

        # instantaneous, regular: f = time interval duration
        # instantaneous, irregular: f = time interval duration
        # not instantaneous, regular: f = time interval duration / reference time interval duration
        # not instantaneous, irregular: f = time interval duration / reference time interval duration

        # note: if capacities are instantaneous, that means they still need to
        # be adjusted to consider the actual time interval durations, i.e., the
        # quantities increase with time step duration

        # note: non-instantaneous capacities should be determined using the re-
        # ference time interval duration since that causes the standard case to
        # be simplified to 1 (all time steps are equal to the reference one)

        param_f_amp_v_glljqk = {}

        param_f_amp_v_glljqk.update(
            {
                (g, u, v, j, q, k): (
                    self.time_frame.time_interval_durations[q][k]  # instantaneous
                    if self.networks[g]
                    .edges[(u, v, j)][Network.KEY_ARC_TECH]
                    .capacity_is_instantaneous
                    else self.normalised_time_interval_duration[  # non-instantaneous: modulate
                        (q, k)
                    ]
                )
                for (g, u, v, j) in set_GLLJ_new
                for (q, k) in set_QK  # for each time interval
            }
        )

        # transmission efficiency

        reverse_none_means_isotropic = True

        param_eta_glljqk = {
            (g, u, v, j, q, k): self.networks[g]
            .edges[(u, v, j)][Network.KEY_ARC_TECH]
            .efficiency[(q, k)]
            if self.networks[g]
            .edges[(u, v, j)][Network.KEY_ARC_TECH]
            .has_proportional_losses()
            else 1
            for (g, u, v) in set_J
            for j in set_J[(g, u, v)]
            for (q, k) in set_QK
        }

        # reverse direction

        param_eta_glljqk.update(
            {
                (g, v, u, j, q, k): (
                    self.networks[g]
                    .edges[(u, v, j)][Network.KEY_ARC_TECH]
                    .efficiency_reverse[(q, k)]
                    if not self.networks[g]
                    .edges[(u, v, j)][Network.KEY_ARC_TECH]
                    .is_isotropic(
                        reverse_none_means_isotropic=reverse_none_means_isotropic
                    )
                    else self.networks[g]
                    .edges[(u, v, j)][Network.KEY_ARC_TECH]
                    .efficiency[(q, k)]
                )
                if self.networks[g]
                .edges[(u, v, j)][Network.KEY_ARC_TECH]
                .has_proportional_losses()
                else 1
                for (g, u, v, j, q, k) in param_eta_glljqk
                if j in set_J_und[(g, u, v)]
            }
        )

        # set_GLLJ_red = tuple(set(
        #     (g, u, v, j) for (g, u, v, j, k) in param_eta_glljk.keys()
        #     ))

        # cost of installing arcs

        param_c_arc_min_glljh = {
            (g, u, v, j, h): self.networks[g]
            .edges[(u, v, j)][Network.KEY_ARC_TECH]
            .minimum_cost[h]
            for (g, u, v) in set_J
            for j in set_J[(g, u, v)]
            if j not in set_J_pre[(g, u, v)]  # new
            if j not in set_J_col[(g, u, v)]  # individual
            for h in set_H_gllj[(g, u, v, j)]
        }

        set_GLLJH_sgl = tuple(param_c_arc_min_glljh.keys())

        # arc capacities (pre-defined)

        param_v_ub_glljqk = {}

        # nominal direction

        param_v_ub_glljqk.update(
            {
                (g, u, v, j, q, k): self.networks[g]
                .edges[(u, v, j)][Network.KEY_ARC_TECH]
                .capacity[
                    self.networks[g]
                    .edges[(u, v, j)][Network.KEY_ARC_TECH]
                    .options_selected.index(True)
                ]
                - (
                    self.networks[g]
                    .edges[(u, v, j)][Network.KEY_ARC_TECH]
                    .static_loss[
                        self.networks[g]
                        .edges[(u, v, j)][Network.KEY_ARC_TECH]
                        .options_selected.index(True),
                        q,
                        k,
                    ]
                    if (
                        ((g, u, v) in set_J_stt_dep and j in set_J_stt_dep[(g, u, v)])
                        or ((g, u, v) in set_J_stt_us and j in set_J_stt_us[(g, u, v)])
                    )
                    else 0
                )
                for (g, u, v) in set_J_pre
                for j in set_J_pre[(g, u, v)]
                if j not in set_J_pre_inf[(g, u, v)]
                for (q, k) in set_QK
            }
        )

        # reverse direction

        param_v_ub_glljqk.update(
            {
                (g, v, u, j, q, k): self.networks[g]
                .edges[(u, v, j)][Network.KEY_ARC_TECH]
                .capacity[
                    self.networks[g]
                    .edges[(u, v, j)][Network.KEY_ARC_TECH]
                    .options_selected.index(True)
                ]
                - (
                    self.networks[g]
                    .edges[(u, v, j)][Network.KEY_ARC_TECH]
                    .static_loss[
                        self.networks[g]
                        .edges[(u, v, j)][Network.KEY_ARC_TECH]
                        .options_selected.index(True),
                        q,
                        k,
                    ]
                    if (
                        ((g, u, v) in set_J_stt_arr and j in set_J_stt_arr[(g, u, v)])
                        or ((g, u, v) in set_J_stt_us and j in set_J_stt_us[(g, u, v)])
                    )
                    else 0
                )
                for (g, u, v) in set_J_pre
                for j in set_J_pre[(g, u, v)]
                if j not in set_J_pre_inf[(g, u, v)]
                if j in set_J_und[(g, u, v)]
                for (q, k) in set_QK
            }
        )

        # set_GLLJ_pre_fin_red = tuple(param_v_ub_glljk.keys())

        set_GLLJ_pre_fin_red = tuple(
            set((g, u, v, j) for (g, u, v, j, q, k) in param_v_ub_glljqk.keys())
        )

        # maximum nominal arc flow amplitude (i.e. capacity)

        param_v_amp_max_glljh = {
            (g, u, v, j, h): self.networks[g]
            .edges[(u, v, j)][Network.KEY_ARC_TECH]
            .capacity[h]
            for (g, u, v) in set_J
            for j in set_J[(g, u, v)]
            if j not in set_J_pre[(g, u, v)]
            if j not in set_J_col[(g, u, v)]
            for h in set_H_gllj[(g, u, v, j)]
        }

        # specific capacity cost

        param_c_arc_var_gllj = {
            (g, u, v, j): self.networks[g]
            .edges[(u, v, j)][Network.KEY_ARC_TECH]
            .specific_capacity_cost
            for (g, u, v) in set_J
            for j in set_J[(g, u, v)]
            if j not in set_J_pre[(g, u, v)]
            if j not in set_J_col[(g, u, v)]  # individual
        }

        set_GLLJ_sgl = tuple(param_c_arc_var_gllj.keys())

        # *********************************************************************

        # sos1 weights for arc selection

        param_arc_inv_sos1_weights_glljh = {
            (g, u, v, j, h): self.use_sos1_arc_inv[(g, u, v, j)][h]
            if self.use_sos1_arc_inv[(g, u, v, j)] is not None
            else h
            for (g, u, v) in set_J_arc_sos1
            for j in set_J_arc_sos1[(g, u, v)]
            # otherwise, go through each option
            for h in set_H_gllj[(g, u, v, j)]
        }

        set_GLLJ_arc_inv_sos1 = tuple(
            (g, u, v, j)
            for (g, u, v) in set_J_arc_sos1
            for j in set_J_arc_sos1[(g, u, v)]
            if (g, u, v, j) in self.use_sos1_arc_inv
        )

        set_GLLJH_arc_inv_sos1 = tuple(param_arc_inv_sos1_weights_glljh.keys())

        # sos1 weights for flow sense determination

        param_arc_sns_sos1_weights_glljqk = {
            (g, u, v, j, q, k): 1 + self.SOS1_SENSE_OFFSET
            if (
                self.use_sos1_flow_sense[(g, u, v, j)]
                == self.SOS1_SENSE_WEIGHT_NOMINAL_HIGHER
            )
            else self.SOS1_SENSE_OFFSET
            for (g, u, v) in set_J_sns_sos1
            for j in set_J_sns_sos1[(g, u, v)]
            # otherwise, go through each option
            for (q, k) in set_QK
        }

        param_arc_sns_sos1_weights_glljqk.update(
            {
                (g, v, u, j, q, k): self.SOS1_SENSE_OFFSET
                if (
                    self.use_sos1_flow_sense[(g, u, v, j)]
                    == self.SOS1_SENSE_WEIGHT_NOMINAL_HIGHER
                )
                else 1 + self.SOS1_SENSE_OFFSET
                for (g, u, v, j, q, k) in param_arc_sns_sos1_weights_glljqk
            }
        )

        set_GLLJQK_und_sns_sos1_red = tuple(param_arc_sns_sos1_weights_glljqk.keys())

        # *********************************************************************
        # *********************************************************************

        # parameters for groups of arcs

        param_c_arc_min_th = {
            (t, h): sum(
                self.networks[g].edges[(u, v, j)][Network.KEY_ARC_TECH].minimum_cost[h]
                for (g, u, v, j) in set_GLLJ_col_t[t]
            )
            for t in set_T
            for h in set_H_t[t]
        }

        param_c_arc_var_t = {
            t: sum(
                self.networks[g]
                .edges[(u, v, j)][Network.KEY_ARC_TECH]
                .specific_capacity_cost
                for (g, u, v, j) in set_GLLJ_col_t[t]
            )
            for t in set_T
        }

        param_v_amp_max_th = {  # TODO: check if this is the case
            (t, h): mean(
                self.networks[g].edges[(u, v, j)][Network.KEY_ARC_TECH].capacity[h]
                for (g, u, v, j) in set_GLLJ_col_t[t]
            )
            for t in set_T
            for h in set_H_t[t]
        }

        param_arc_inv_sos1_weights_th = {
            (t, h): self.groups_arc_sos1[t][h]
            # (t, h): mean(
            #     (self.use_sos1_arc_inv[(g,u,v,j)][h]
            #      if self.use_sos1_arc_inv[(g,u,v,j)] is not None else h)
            #     for (g,u,v,j) in set_GLLJ_col_t[t]
            #     )
            for t in set_T_sos1
            for h in set_H_t[t]
        }

        # *********************************************************************

        # fixed arc losses for preselected arcs

        # param_w_pre_glljk = {
        #     (g, u, v, j, k):
        #         self.networks[g].edges[(u, v, j)][
        #             Network.KEY_ARC_TECH].static_loss[(
        #                 self.networks[g].edges[(u, v, j)][
        #                     Network.KEY_ARC_TECH].options_selected.index(True)
        #                 ,0,k)]
        #     for (g, u, v) in set_J_stt
        #     for j in set_J_stt[(g,u,v)]
        #     if j in set_J_pre[(g,u,v)]
        #     for k in set_K
        #     }

        param_w_pre_glljqk = {
            (g, u, v, j, q, k): self.networks[g]
            .edges[(u, v, j)][Network.KEY_ARC_TECH]
            .static_loss[
                (
                    self.networks[g]
                    .edges[(u, v, j)][Network.KEY_ARC_TECH]
                    .options_selected.index(True),
                    q,
                    k,
                )
            ]
            for (g, u, v, j) in set_GLLJ_static_pre
            # for (g, u, v) in set_J_stt
            # for j in set_J_stt[(g,u,v)]
            # if j in set_J_pre[(g,u,v)]
            for (q, k) in set_QK
        }

        # fixed arc losses for selectable arcs

        # param_w_new_glljhk = {
        #     (g, u, v, j, h, k):
        #         self.networks[g].edges[(u, v, j)][
        #             Network.KEY_ARC_TECH].static_loss[(h, 0, k)]
        #     for (g, u, v) in set_J_stt
        #     for j in set_J_stt[(g,u,v)]
        #     if j not in set_J_pre[(g,u,v)]
        #     for h in set_H_gllj[(g, u, v, j)]
        #     for k in set_K
        #     }

        param_w_new_glljhqk = {
            (g, u, v, j, h, q, k): self.networks[g]
            .edges[(u, v, j)][Network.KEY_ARC_TECH]
            .static_loss[(h, q, k)]
            for (g, u, v, j) in set_GLLJ_static_new
            if j not in set_J_col[(g, u, v)]
            # for j in set_J_stt[(g,u,v)]
            # if j not in set_J_pre[(g,u,v)]
            for h in set_H_gllj[(g, u, v, j)]
            for (q, k) in set_QK
        }

        param_w_new_glljhqk.update(
            {
                (g, u, v, j, h, q, k): self.networks[g]
                .edges[(u, v, j)][Network.KEY_ARC_TECH]
                .static_loss[(h, q, k)]
                for t in set_T
                for (g, u, v, j) in set_GLLJ_col_t[t]
                if j in set_J_stt[(g, u, v)]
                # if (g, u, v, j) in set_GLLJ_static_new
                for h in set_H_t[t]
                for (q, k) in set_QK
            }
        )

        set_GLLJH_static_new = tuple(
            set((g, u, v, j, h) for (g, u, v, j, h, q, k) in param_w_new_glljhqk.keys())
        )

        # *********************************************************************
        # *********************************************************************
        # *********************************************************************
        # *********************************************************************

        # produce a dictionary with the data for the problem
        
        # TODO: built the dict as a function of the price submodel

        data_dict = {
            None: {
                # *****************************************************************
                # *****************************************************************
                # sets
                # *****************************************************************
                # *****************************************************************
                # common sets
                "set_Q": {None: set_Q},
                "set_P_q": set_P_q,
                "set_K_q": set_K_q,
                "set_QP": set_QP,
                "set_QK": set_QK,
                "set_QPK": set_QPK,
                # network sets
                "set_G": {None: set_G},
                "set_L": set_L,
                "set_L_imp": set_L_imp,
                "set_L_exp": set_L_exp,
                "set_L_max_in_g": set_L_max_in_g,
                'set_L_max_out_g': set_L_max_out_g,
                "set_GL": set_GL,
                "set_GL_exp": set_GL_exp,
                "set_GL_imp": set_GL_imp,
                "set_GL_exp_imp": set_GL_exp_imp,
                "set_GL_not_exp_imp": set_GL_not_exp_imp,
                "set_GLL": set_GLL,
                # *****************************************************************
                # tariff
                "set_S": set_S,
                "set_GLQPKS": set_GLQPKS,
                "set_GLQPKS_exp": set_GLQPKS_exp,
                "set_GLQPKS_imp": set_GLQPKS_imp,
                # *****************************************************************
                # converter sets
                "set_I": set_I,
                "set_I_new": set_I_new,
                # input sets
                "set_M": set_M,
                "set_M_bin": set_M_bin,
                "set_M_dim": set_M_dim,
                "set_M_cost": set_M_cost,
                "set_M_ext": set_M_ext,
                # state sets
                "set_N": set_N,
                "set_N_dim_up": set_N_dim_up,
                "set_N_dim_lo": set_N_dim_lo,
                "set_N_cost_up": set_N_cost_up,
                "set_N_cost_lo": set_N_cost_lo,
                "set_N_ext": set_N_ext,
                # output sets
                "set_R": set_R,
                "set_R_dim": set_R_dim,
                "set_R_cost": set_R_cost,
                "set_R_ext": set_R_ext,
                # *****************************************************************
                # arc related sets
                "set_J": set_J,
                "set_J_und": set_J_und,  # if j not here, then it is directed
                "set_J_pre": set_J_pre,  # if j not here, then it is selectable
                "set_J_pre_inf": set_J_pre_inf,  # if j not here, then it is fin cap
                "set_J_col": set_J_col,
                "set_J_mdt": set_J_mdt,  # if j not here, then it is optional
                "set_J_arc_sos1": set_J_arc_sos1,  # option
                "set_J_arc_nnr": set_J_arc_nnr,  # if j not here, then it uses binary vars.
                "set_J_int": set_J_int,  # if j not here, no int. are needed
                "set_J_sns_sos1": set_J_sns_sos1,  # option
                "set_J_sns_nnr": set_J_sns_nnr,  # the alternative is bin. vars.
                "set_J_stt": set_J_stt,
                "set_J_stt_dep": set_J_stt_dep,
                "set_J_stt_arr": set_J_stt_arr,
                "set_J_stt_us": set_J_stt_us,
                "set_J_stt_ds": set_J_stt_ds,
                "set_H_gllj": set_H_gllj,
                # *****************************************************************
                # groups of arcs
                "set_T": set_T,
                # set of mandatory arc groups
                "set_T_mdt": set_T_mdt,
                # set of arc groups requiring interface variables
                "set_T_int": set_T_int,
                # set of arc groups relying on SOS1
                "set_T_sos1": set_T_sos1,
                # set of arg groups relying on non-negative real variables
                "set_T_nnr": set_T_nnr,
                # set of arc groups relying on binary variables
                "set_T_bin": set_T_bin,
                # set of arcs in the various arc groups
                "set_GLLJ_col_t": set_GLLJ_col_t,
                # set of arc options for arc groups
                "set_H_t": set_H_t,
                # # set of (t,h) tuples
                # 'set_TH': set_TH,
                # # set of (t,g,l1,l2,j) tuples
                # 'set_TGLLJ': set_TGLLJ,
                # # set of TH tuples for groups using SOS1
                # 'set_TH_sos1': set_TH_sos1,
                # # set of t-indexed TH tuples for groups of arcs relying on SOS1
                # 'set_TH_sos1_t': set_TH_sos1_t,
                # minimum cost of a group of arcs
                "param_c_arc_min_th": param_c_arc_min_th,
                # unit flow amplitude cost
                "param_c_arc_var_t": param_c_arc_var_t,
                # maximum nominal amplitude
                "param_v_amp_max_th": param_v_amp_max_th,
                # sos1 weights for arc group options
                "param_arc_inv_sos1_weights_th": param_arc_inv_sos1_weights_th,
                # *****************************************************************
                # *****************************************************************
                # ancillary sets
                # *****************************************************************
                # *****************************************************************
                "set_GLLJ": set_GLLJ,
                "set_GLLJ_und": set_GLLJ_und,
                "set_GLLJ_new": set_GLLJ_new,
                "set_GLLJ_sgl": set_GLLJ_sgl,
                "set_GLLJ_und_ext": set_GLLJ_und_ext,
                "set_GLLJ_und_red": set_GLLJ_und_red,
                "set_GLLJ_red": set_GLLJ_red,
                "set_GLLJH_sgl": set_GLLJH_sgl,
                "set_GLLJ_pre_fin_red": set_GLLJ_pre_fin_red,
                "set_GLLJ_arc_inv_sos1": set_GLLJ_arc_inv_sos1,
                "set_GLLJH_arc_inv_sos1": set_GLLJH_arc_inv_sos1,
                "set_GLLJQK_und_sns_sos1_red": set_GLLJQK_und_sns_sos1_red,
                "set_GLLJ_int": set_GLLJ_int,
                "set_GLLJ_static": set_GLLJ_static,
                "set_GLLJ_static_pre": set_GLLJ_static_pre,
                "set_GLLJ_static_new": set_GLLJ_static_new,
                "set_GLLJH_static_new": set_GLLJH_static_new,
                "set_GLLJ_static_und_red": set_GLLJ_static_und_red,
                # model.set_GLLJ_und_red,
                # model.set_GLLJ_und,
                # model.set_GLLJ_und_new,
                # model.set_GLLJH_static_new,
                # model.set_GLLJ_static_pre,
                # model.set_GLLJ_static_und_red,
                # model.set_GLLJ_static_new,
                # model.set_GLLJ_static_dir,
                # model.set_GLLJ_static_und,
                # model.set_GLLJ_static_und_pre,
                # model.set_GLLJ_static_und_new,
                # *****************************************************************
                # *****************************************************************
                # parameters
                # *****************************************************************
                # *****************************************************************
                # objective function
                "param_c_wgt_q": param_c_wgt_q,
                "param_c_df_qp": param_c_df_qp,
                "param_c_time_qpk": param_c_time_qpk,
                "param_p_glqpks": param_p_glqpks,
                "param_price_function_is_convex": param_price_function_is_convex,
                "param_v_max_glqpks": param_v_max_glqpks,
                # *****************************************************************
                # converters
                # objective function
                "param_c_cvt_i": param_c_cvt_i,
                "param_c_cvt_u_im": param_c_cvt_u_im,
                "param_c_cvt_x_in": param_c_cvt_x_in,
                "param_c_cvt_y_ir": param_c_cvt_y_ir,
                "param_c_ext_u_imqk": param_c_ext_u_imqk,
                "param_c_ext_x_inqk": param_c_ext_x_inqk,
                "param_c_ext_y_irqk": param_c_ext_y_irqk,
                # inputs
                "param_u_ub_imqk": param_u_ub_imqk,
                "param_u_lim_max_im": param_u_lim_max_im,
                "param_f_amp_u_imqk": param_f_amp_u_imqk,
                # states
                "param_a_eq_x_innqk": param_a_eq_x_innqk,
                "param_b_eq_x_inmqk": param_b_eq_x_inmqk,
                "param_e_eq_x_inqk": param_e_eq_x_inqk,
                "param_x_inq0": param_x_inq0,
                "param_x_ub_inqk": param_x_ub_inqk,
                "param_x_lb_inqk": param_x_lb_inqk,
                "param_x_pos_lim_max_in": param_x_pos_lim_max_in,
                "param_x_pos_lim_min_in": param_x_pos_lim_min_in,
                "param_x_neg_lim_max_in": param_x_neg_lim_max_in,
                "param_x_neg_lim_min_in": param_x_neg_lim_min_in,
                "param_f_amp_x_inqk": param_f_amp_x_inqk,
                # outputs
                "param_c_eq_y_irnqk": param_c_eq_y_irnqk,
                "param_d_eq_y_irmqk": param_d_eq_y_irmqk,
                "param_e_eq_y_irqk": param_e_eq_y_irqk,
                "param_y_ub_irqk": param_y_ub_irqk,
                "param_y_lb_irqk": param_y_lb_irqk,
                "param_y_pos_lim_max_ir": param_y_pos_lim_max_ir,
                "param_y_pos_lim_min_ir": param_y_pos_lim_min_ir,
                "param_y_neg_lim_max_ir": param_y_neg_lim_max_ir,
                "param_y_neg_lim_min_ir": param_y_neg_lim_min_ir,
                "param_f_amp_y_irqk": param_f_amp_y_irqk,
                # *****************************************************************
                # network
                "param_max_number_parallel_arcs": param_max_number_parallel_arcs,
                "param_v_base_glqk": param_v_base_glqk,
                "param_a_nw_glimqk": param_a_nw_glimqk,
                "param_a_nw_glirqk": param_a_nw_glirqk,
                # *****************************************************************
                # arc parameters
                "param_eta_glljqk": param_eta_glljqk,
                "param_v_ub_glljqk": param_v_ub_glljqk,
                "param_v_amp_max_glljh": param_v_amp_max_glljh,
                "param_c_arc_min_glljh": param_c_arc_min_glljh,
                "param_c_arc_var_gllj": param_c_arc_var_gllj,
                "param_f_amp_v_glljqk": param_f_amp_v_glljqk,
                # fixed losses
                "param_w_pre_glljqk": param_w_pre_glljqk,
                "param_w_new_glljhqk": param_w_new_glljhqk,
                # sos1 weights for arc selection
                "param_arc_inv_sos1_weights_glljh": param_arc_inv_sos1_weights_glljh,
                # sos1 weights for flow sense determination
                "param_arc_sns_sos1_weights_glljqk": param_arc_sns_sos1_weights_glljqk,
                # *****************************************************************
                # *****************************************************************
                "block_prices": {
                    (*gl,*qpk): {
                        'param_price_function_is_convex': {None: param_price_function_is_convex[(*gl,*qpk)]}, 
                        'set_S': {None: set_S[(*gl,*qpk)]}, 
                        'param_p_s': {s: param_p_glqpks[(*gl,*qpk,s)] for s in set_S[(*gl,*qpk)]}, 
                        'param_v_max_s': {s: param_v_max_glqpks[(*gl,*qpk,s)] for s in set_S[(*gl,*qpk)] if (*gl,*qpk,s) in param_v_max_glqpks},
                        }
                    for gl in set_GL_exp_imp
                    for qpk in set_QPK
                    }
            }
        }

        print("dictionary created...")

        return data_dict

    # *************************************************************************
    # *************************************************************************

    def assert_solver_problem_compatibility(self, solver_name):
        return SolverInterface.problem_and_solver_are_compatible(
            solver_name, self.optimisation_problem_type
        )

    # *************************************************************************
    # *************************************************************************
    
    # TODO: do it for converters too

    def import_results(
        self,
        solved_instance: pyo.ConcreteModel = None,
        integrality_tolerance: float = 0.01,
    ):
        # no solved_instance has been provided, load from the class instance

        if solved_instance == None:
            solved_instance = self.instance

        # *********************************************************************

        # upload investment results, namely the NPV, NCFs and DFs

        # # net present value

        # self.net_present_value = pyo.value(solved_instance.obj_f)

        # # discount factors

        # self.discount_factors = [
        #     solved_instance.param_c_df_p[p]
        #     for p in solved_instance.set_P]

        # # net cash flows (for each time period)

        # self.net_cash_flows = [
        #     pyo.value(solved_instance.var_ncf_p[p])
        #     for p in solved_instance.set_P]

        # # sanity check

        # temp_npv = Investment.npv(self.discount_rates, self.net_cash_flows)

        # if abs(temp_npv - self.net_present_value) > npv_tolerance:

        #     Warning('The NPV results are not consistent with the input data.')

        # *********************************************************************

        # for each arc subject to optimisation

        for g, l1, l2, j, h in solved_instance.set_GLLJH_sgl:
            # find the network and the arc

            # convert the decision variable into a boolean (due to integrality tol.)

            decision = bool(
                round(
                    pyo.value(
                        solved_instance.var_delta_arc_inv_glljh[(g, l1, l2, j, h)]
                    )
                )
            )

            # *****************************************************************

            # update the object

            self.networks[g].edges[(l1, l2, j)][Network.KEY_ARC_TECH].options_selected[
                h
            ] = decision

            # *****************************************************************

        # *********************************************************************

        # for each arc group

        for t, h in solved_instance.set_TH:
            decision = bool(
                round(pyo.value(solved_instance.var_delta_arc_inv_th[(t, h)]))
            )

            for g, l1, l2, j in solved_instance.set_GLLJ_col_t[t]:
                self.networks[g].edges[(l1, l2, j)][
                    Network.KEY_ARC_TECH
                ].options_selected[h] = decision

            # *****************************************************************

        # *********************************************************************

    # *************************************************************************
    # *************************************************************************
    
    def peak_total_assessments(self) -> dict:
        """Returns the peak+total compatible assessments and the respective intervals."""
        
        # output: a dict with q keys and k values
        
        # conditions:
        # 1) prices within each period have to be invariant in time and volume
        # - how to quickly determine this?
        # 2) there can only be import or export nodes, not both
        # 3) the range of potential flows through each arc can be represented
        # using a subset of the intervals, for at least one assessment
        # - arc efficiency is constant in time, for each arc and assessment
        # - the variation of static arc losses over time correlates with the
        # variation of the maximum potential flow through the arc over time
        # - there are no converters under consideration

        # *********************************************************************
        
        # initialise the set of peak total assessments
        set_q = set(self.time_frame.assessments)
        # initialise the output dictionary
        out_dict = {}
        
        # *********************************************************************
        
        # there can only be import or export nodes, not both

        # check each network
        for net_key, net in self.networks.items():
            if len(set(net.import_nodes).intersection(net.export_nodes)) >= 1:
                # return an empty dict
                return out_dict
            
        # *********************************************************************
        
        # there can be no converters (might be relaxed)
        if len(self.converters) != 0:
            # there are converters
            return out_dict
        
        # *********************************************************************
        
        # prices have to be invariant per period and assessment ((q,p) tuple)
        
        # for each assessment
        for q in self.time_frame.assessments:
            # assume it is compatible
            q_meets_criteria = True
            # build the (q,p,k) tuples for this assessment and respective periods
            qpk_qp_dict = {
                (q,p): tuple(
                    (q,p,k) for k in self.time_frame.time_intervals[q]
                    )
                for p in self.time_frame.reporting_periods[q]
                }
            # check each network
            for net_key, net in self.networks.items():
                # check import node prices
                for imp_node_key in net.import_nodes:
                    # check if the node prices are invariant within each (q,p)
                    for qp, qpk_tuples in qpk_qp_dict.items():
                        if not entries_are_invariant(
                                net.nodes[imp_node_key][Network.KEY_NODE_PRICES],
                                qpk_tuples
                                ):
                            # entries are not time invariant
                            set_q.remove(q)
                            q_meets_criteria = False
                            break # out of qpk_qp_dict.items() for loop
                        # prices are time invariant, check if they are volume
                        # invariant as well: checking one entry will be enough
                        if not net.nodes[imp_node_key][Network.KEY_NODE_PRICES][
                            qpk_tuples[0]
                        ].is_volume_invariant():
                            # it is not volume invariant
                            set_q.remove(q)
                            q_meets_criteria = False
                            break # out of qpk_qp_dict.items() for loop
                    if not q_meets_criteria:
                        break # out of net.import_nodes for loop
                if not q_meets_criteria:
                    break # out of networks.items() for loop
                        
                # check export node prices
                for exp_node_key in net.export_nodes:
                    # check if the node prices are invariant within each (q,p)
                    for qp, qpk_tuples in qpk_qp_dict.items():
                        if not entries_are_invariant(
                                net.nodes[exp_node_key][Network.KEY_NODE_PRICES],
                                qpk_tuples
                                ):
                            # entries are not time invariant
                            set_q.remove(q)
                            q_meets_criteria = False
                            break # out of qpk_qp_dict.items() for loop
                        # prices are time invariant, check if they are volume
                        # invariant as well: checking one entry will be enough
                        if not net.nodes[exp_node_key][Network.KEY_NODE_PRICES][
                            qpk_tuples[0]
                        ].is_volume_invariant():
                            # it is not volume invariant
                            set_q.remove(q)
                            q_meets_criteria = False
                            break # out of qpk_qp_dict.items() for loop
                    if not q_meets_criteria:
                        break # out of net.export_nodes for loop
                if not q_meets_criteria:
                    break # out of networks.items() for loop                
        
        # *********************************************************************
        
        # for each arc and assessment, the efficiency must be time invariant
           
        # arc efficiencies have to be time invariant per assessment
        # for each assessment
        for q in self.time_frame.assessments:
            # if it is still under consideration
            if q not in set_q:
                continue
            # assume it is compatible
            q_meets_criteria = True
            # build the (q,k) tuples for this assessment
            qk_tuples = tuple((q,k) for k in self.time_frame.time_intervals[q])
            for key, net in self.networks.items():
                # check each edge
                for edge_key in net.edges(keys=True):
                    if net.edges[edge_key][Network.KEY_ARC_TECH].has_proportional_losses():
                        if not entries_are_invariant(
                                net.edges[edge_key][Network.KEY_ARC_TECH].efficiency,
                                qk_tuples
                                ):
                            # entries are not time invariant, remove q from consideration
                            set_q.remove(q)
                            q_meets_criteria = False
                            break # break out of net.edges() for loop
                if not q_meets_criteria:
                    break # break out of self.networks.items():
                
        # *********************************************************************
        
        # find the critical intervals for each compatible assessment
        
        # TODO: revise this method to consider static losses
        
        # for each assessment still under consideration
        for q in set_q:
            # assume it is compatible
            q_meets_criteria = True
            # build the (q,k) tuples for this assessment
            qk_tuples = tuple((q,k) for k in self.time_frame.time_intervals[q])
            # initialise the intervals for this assessment
            set_k = set()
            # for each network
            for key, net in self.networks.items():
                # for each source/sink node
                for node_key in net.source_sink_nodes:
                    # identify the nonzero extremes (minima first)
                    static_flow_sorted = sorted(
                        ((v, k) 
                         for (_, k), v in net.nodes[node_key][Network.KEY_NODE_BASE_FLOW].items()
                         if v != 0 # there should be at least one non-zero entry
                         )
                        )
                    if static_flow_sorted[0][0] < 0:
                        # negative minima: add the interval to the set
                        set_k.add(static_flow_sorted[0][1])
                        # add the other extreme to the list if it is positive
                        if static_flow_sorted[-1][0] > 0:
                            # positive maxima: add it to the set too
                            set_k.add(static_flow_sorted[-1][1])
                        # if not positive: do not add it to the set
                    else:
                        # minima is positive: add the maxima to the set
                        set_k.add(static_flow_sorted[-1][1])
                # # for each edge
                # for edge_key in net.edges(keys=True):
                #     # check for static losses
                #     if net.edges[edge_key][Network.KEY_ARC_TECH].has_static_losses():
                #         # it has static losses, check each option
                #         for h in range(
                #             net.edges[edge_key][
                #                 Network.KEY_ARC_TECH
                #                 ].number_options()
                #         ):
                #             # sort the static losses to identify the extremes
                #             # the maximum appears last
                #             static_loss_sorted = sorted(
                #                 ((net.edges[edge_key][Network.KEY_ARC_TECH].static_loss[(h, q, k)], k)
                #                  for q, k in qk_tuples
                #                  )
                #                 )
                #             # add the interval for the maxima to the set
                #             set_k.add(static_loss_sorted[-1][1])
                # *************************************************************
            # all checks passed so far: move on to the next
            out_dict[q] = list(set_k)
            
        # *********************************************************************
        
        # return statement
        return out_dict

    # *************************************************************************
    # *************************************************************************

    def has_peak_total_assessments(self) -> bool:
        """Returns True if there are peak+total assessments in the problem."""
        return len(self.peak_total_assessments()) >= 1
    
    # *************************************************************************
    # *************************************************************************
    
    def simplify_peak_total_assessments(self, qk_dict: dict=None):
        """Simplifies the problem in the presence of peak+total assessments."""
        
        # *********************************************************************
        # *********************************************************************
        
        # check if the qk dict was defined
        if type(qk_dict) == type(None):
            # no tuples were defined, identify them automatically
            qk_dict = self.peak_total_assessments()
        
        # *********************************************************************
        # *********************************************************************
        
        # critical assessments
        q_ref_tuples = tuple(qk_dict.keys())
        qk_ref_tuples = tuple(
            (q, k) for q in q_ref_tuples for k in qk_dict[q]
            )
        # total assessments: one assessment per reference assessment
        qk_total_tuples = tuple(
            ('total-'+str(q), 0) for q in q_ref_tuples
            )
        q_total_tuples = tuple(qk[0] for qk in qk_total_tuples)
        # peak assessments: one assessment only, uses all ref intervals
        q_peak = 'peak'
        qk_peak_tuples = tuple(
            (q_peak, i)
            for i, qk in enumerate(qk_ref_tuples)
            )
        
        # *********************************************************************
        # *********************************************************************
        
        # prepare the data
        
        # for each network
        for key, net in self.networks.items():
            # for each arc
            for arc_key in net.edges(keys=True):
                
                # static losses
                if net.edges[arc_key][Network.KEY_ARC_TECH].has_static_losses():
                    # compute the total static losses
                    new_data = {
                        (h, *qk_total): sum(
                            net.edges[arc_key][Network.KEY_ARC_TECH].static_loss[
                                (h, q_ref, k)
                            ]
                            for k in self.time_frame.time_intervals[q_ref]
                        )
                        for q_ref, qk_total in zip(q_ref_tuples, qk_total_tuples)
                        for h in range(
                            net.edges[arc_key][
                                Network.KEY_ARC_TECH
                                ].number_options()
                        )
                    }
                    # add data for the peak assessment
                    new_data.update({
                        (h, *qk_peak): net.edges[arc_key][Network.KEY_ARC_TECH].static_loss[
                            (h, *qk_ref)]
                        for qk_peak, qk_ref in zip(qk_peak_tuples, qk_ref_tuples)
                        for h in range(
                            net.edges[arc_key][
                                Network.KEY_ARC_TECH
                                ].number_options()
                        )
                        })
                    # remove dict entries associated with the reference assessments
                    for qk in qk_ref_tuples:
                        for h in range(
                            net.edges[arc_key][
                                Network.KEY_ARC_TECH
                                ].number_options()
                        ):
                            net.edges[arc_key][Network.KEY_ARC_TECH].static_loss.pop((h, *qk))
                    # update the static loss dict with the new data
                    net.edges[arc_key][Network.KEY_ARC_TECH].static_loss.update(new_data)
                # the arc has no static losses, continue
                
                # proportional losses
                if net.edges[arc_key][Network.KEY_ARC_TECH].has_proportional_losses():
                    # get the efficiency values for the total assessments
                    # the efficiency should be constant so one interval will do
                    efficiency_dict = {
                        qk_total: (
                            net.edges[arc_key][Network.KEY_ARC_TECH].efficiency[
                                (q_ref, qk_dict[q_ref][0])
                            ]
                        )
                        for qk_total, q_ref in zip(qk_total_tuples, q_ref_tuples)
                    }
                    # add data for the peak assessment
                    efficiency_dict.update({
                        qk_peak: net.edges[arc_key][Network.KEY_ARC_TECH].efficiency[
                            qk_ref_tuples[0]
                        ]
                        for qk_peak in qk_peak_tuples
                        })
                    # remove dict entries associated with the reference assessments
                    for qk in qk_ref_tuples:
                        net.edges[arc_key][Network.KEY_ARC_TECH].efficiency.pop(qk)
                    # update the efficiency dict to include the total assessme.
                    net.edges[arc_key][Network.KEY_ARC_TECH].efficiency.update(
                        efficiency_dict
                    )
            
            # handle import/export nodes
            for node_key in set(net.import_nodes).union(net.export_nodes):
                # import/export node: determine price (invariant)
                total_price = {
                    (qk_total[0], p, qk_total[1]): (
                        net.nodes[node_key][Network.KEY_NODE_PRICES][(q_ref, p, qk_dict[q_ref][0])]
                    )
                    for q_ref, qk_total in zip(q_ref_tuples, qk_total_tuples)
                    for p in self.time_frame.reporting_periods[q_ref]
                }
                # no data for the peak assessment
                # remove dict entries associated with the reference assessments
                for qk in qk_ref_tuples:
                    for p in self.time_frame.reporting_periods[qk[0]]:
                        net.nodes[node_key][Network.KEY_NODE_PRICES].pop((qk[0],p,qk[1]))
                # update the price dict
                net.nodes[node_key][Network.KEY_NODE_PRICES].update(total_price)
            
            # handle source/sink nodes
            for node_key in net.source_sink_nodes:
                # determine the flow volume for the total assessment
                new_data = {
                    qk_total: sum(
                        net.nodes[node_key][Network.KEY_NODE_BASE_FLOW][(q_ref, k)]
                        for k in range(
                                self.time_frame.number_time_intervals(q_ref)
                                )
                    )
                    for q_ref, qk_total in zip(q_ref_tuples, qk_total_tuples)
                }
                # add data for the peak assessment
                new_data.update({
                    qk_peak: net.nodes[node_key][Network.KEY_NODE_BASE_FLOW][qk_ref]
                    for qk_peak, qk_ref in zip(qk_peak_tuples, qk_ref_tuples)
                    })
                # remove dict entries associated with the reference assessments
                for qk in qk_ref_tuples:
                    net.nodes[node_key][Network.KEY_NODE_BASE_FLOW].pop(qk)
                # update the dict for the base flow needs
                net.nodes[node_key][Network.KEY_NODE_BASE_FLOW].update(new_data)

        # *********************************************************************
        # *********************************************************************
        
        # define the new time frame
        # - 1 total assessment per P&T assessment (1 interval)
        # - 1 peak assessment (1 or more intervals)
        # - non P&T assessments (stay the same)
                
        # discount rates:
        # - non P&T assessments stay the same
        # - total assessments use the discount rates from P&T assessments
        # - peak assessments have no discount rates
        discount_rates_q = {
            q: [self.time_frame.discount_rate(q, p)
                for p in self.time_frame.reporting_periods[q]]
            for q in self.time_frame.assessments 
            if q not in q_ref_tuples
            }
        discount_rates_q.update({
            q_total: [self.time_frame.discount_rate(q_ref, p)
                      for p in self.time_frame.reporting_periods[q_ref]]
            for q_total, q_ref in zip(q_total_tuples, q_ref_tuples)
            })
        discount_rates_q.update({q_peak: []})
        
        # time intervals:
        # - intervals for non-P&T assessments stay the same
        # - total assessments have a single interval numbered 0
        # - peak intervals are numbered consecutively
        time_intervals = {
            q: self.time_frame.time_intervals[q]
            for q in self.time_frame.assessments 
            if q not in q_ref_tuples
            }
        time_intervals.update({
            q_total: [k_total]
            for q_total, k_total in qk_total_tuples
            })
        time_intervals.update({
            q_peak: [k for _, k in qk_peak_tuples]
            })
        
        # time interval durations:
        # - intervals for non-P&T assessments stay the same
        # - total assessments represent the entire duration of a P&T assessment
        # - peak intervals stay the same as in the respective P&T assessments
        time_interval_durations = {
            q: self.time_frame.time_interval_durations[q]
            for q in self.time_frame.assessments 
            if q not in q_ref_tuples
            }
        time_interval_durations.update({
            q_total: [sum(self.time_frame.time_interval_durations[q_ref])]
            for q_ref, q_total in zip(q_ref_tuples, q_total_tuples)
            })
        time_interval_durations.update({
            q_peak: [self.time_frame.time_interval_durations[q_ref][k_ref]
                     for q_ref, k_ref in qk_ref_tuples
                     ]
            })
        
        # reporting periods:
        # - non P&T assessments stay the same
        # - total assessments cover the same periods as the respective P&T assessment
        # - peak assessments do not (need to) cover any period
        reporting_periods = {
            q: self.time_frame.reporting_periods[q]
            for q in self.time_frame.assessments 
            if q not in q_ref_tuples
            }
        reporting_periods.update({
            q_total: self.time_frame.reporting_periods[q_ref]
            for q_ref, q_total in zip(q_ref_tuples, q_total_tuples)
            })
        reporting_periods.update({q_peak: []})
        
        # reporting period durations:
        reporting_period_durations = {
            q: self.time_frame.reporting_period_durations[q]
            for q in self.time_frame.assessments 
            if q not in q_ref_tuples
            }
        reporting_period_durations.update({
            q_total: self.time_frame.reporting_period_durations[q_ref]
            for q_ref, q_total in zip(q_ref_tuples, q_total_tuples)
            })
        reporting_period_durations.update({q_peak: []})
        
        # - non-P&T assessments are not affected
        # - peak assessments get a weight of 1 (possibly through omission)
        # - total assessments receive the weight of the respective reference assessment
        
        # self.param_c_wgt_q
        new_assessment_weights = {
            q: self.assessment_weights[q] if (type(self.assessment_weights) == dict and q in self.assessment_weights) else 1
            for q in self.time_frame.assessments # original assessments
            if q not in q_ref_tuples # non-P&T
            #if q != q_peak
            }
        new_assessment_weights.update({
            q_total: self.assessment_weights[q_ref] if (type(self.assessment_weights) == dict and q_ref in self.assessment_weights) else 1
            for q_total, q_ref in zip(q_total_tuples, q_ref_tuples)
            })
        new_assessment_weights[q_peak] = 1
        self.assessment_weights = new_assessment_weights
        
        # create the object        
        self.time_frame = EconomicTimeFrame(
            discount_rates_q=discount_rates_q,
            reporting_periods=reporting_periods,
            reporting_period_durations=reporting_period_durations,
            time_intervals=time_intervals,
            time_interval_durations=time_interval_durations
            )
        
        # TODO: include regular time steps as a condition

        # average time interval
        self.average_time_interval = {
            q: mean(self.time_frame.time_interval_durations[q]) 
            for q in self.time_frame.assessments
        }
        
        # normalised time interval durations
        # - for non P&T assessments, factors are calculated using the average
        # - for peak assessments, factors are calculated using the average
        # - for total assessments, factors are calculated using the total dura-
        # tion of the intervals composing the assessment. Why? By adding up the
        # flow needs of multiple intervals, amplitudes also need to be adjusted
        self.normalised_time_interval_duration = {
            (q,k): dt/self.average_time_interval[q] if q not in q_total_tuples else dt/self.average_time_interval[q_peak]
            for q in self.time_frame.assessments
            for k, dt in zip(
                    self.time_frame.time_intervals[q],
                    self.time_frame.time_interval_durations[q]
                    )
        }
        
        # *********************************************************************
        # *********************************************************************
        
    # *************************************************************************
    # *************************************************************************

# *****************************************************************************
# *****************************************************************************
