# *****************************************************************************
# *****************************************************************************

from numbers import Real

# *****************************************************************************
# *****************************************************************************

# TODO: change name to ResourceTariff


class ResourcePrice:
    """A class for piece-wise linear resource prices in network problems."""

    def __init__(
            self, 
            prices: list or int, 
            volumes: list = None
            ):
        # how do we keep the size of the object as small as possible
        # if the tariff is time-invariant, how can information be stored?
        # - a flag

        # accepted inputs:
        # 1) prices and values as lists with matching sizes
        # - all elements in the lists need to be numeric
        # - all prices need to be non-negative
        # - all volumes need to be positive
        # 2) price as a list, volume as None
        # - all elements in the list need to be numeric and non-negative

        (self._number_segments, self.prices, self.volumes) = self.validate_list(
            prices, volumes
        )

        self._volume_invariant = self.is_volume_invariant()

        # *********************************************************************
        # *********************************************************************

    # *************************************************************************
    # *************************************************************************

    def validate_list(self, prices, volumes):
        """Validates the inputs provided in list format."""

        if type(volumes) == list:
            # ensure prices are also provided as a list

            if type(prices) != list:
                raise TypeError(
                    "The prices need to be provided as a list, if volumes are" + " too."
                )

            # prices and volumes are lists

            number_segments = len(volumes)

            if number_segments != len(prices):
                raise ValueError("The number of prices and volumes are inconsistent.")

            # check the elements

            for segment_index in range(number_segments):
                price = prices[segment_index]

                volume = volumes[segment_index]

                # price

                try:
                    if price < 0:
                        raise ValueError("The prices provided were not all positive.")

                except TypeError:
                    raise TypeError(
                        "The prices were not all provided as numeric types."
                    )

                # volume

                if type(volume) == type(None):
                    # if None, move to the next iteration

                    if segment_index == number_segments - 1:
                        # the last segment is None: unlimited volume

                        continue

                    else:
                        # intermediate segment:

                        raise ValueError(
                            "The intermediate segments cannot have volume " + "limits."
                        )

                else:
                    # if not None, make sure it is positive

                    try:
                        if volume <= 0:
                            raise ValueError(
                                "The volumes provided were not all positive."
                            )

                    except TypeError:
                        raise TypeError(
                            "The volumes were not all provided as numeric " + "types."
                        )

            # done

            return number_segments, prices, volumes

        elif type(volumes) == type(None) or isinstance(volumes, Real):
            # the prices must be numeric and positive

            if not isinstance(prices, Real):
                raise TypeError("The prices were not all provided as numeric types.")

            if prices < 0:
                raise ValueError("The prices provided were not all positive.")

            # the number of segments must be 1

            number_segments = 1

            # done

            return number_segments, [prices], [volumes]

        else:
            raise TypeError("Unrecognised type for volumes.")

    # *************************************************************************
    # *************************************************************************

    def price_monotonically_increasing_with_volume(self) -> bool:
        # check if we are to focus on one specific time interval or not

        # if there is only one segment, return false

        if self._number_segments == 1:
            return True

        # check one interval:

        for segment_index in range(self._number_segments - 1):
            # check consecutive segments

            if self.prices[segment_index] > self.prices[segment_index + 1]:
                # prices decreased

                return False

        # otherwise

        return True

    # *************************************************************************
    # *************************************************************************

    def price_monotonically_decreasing_with_volume(self) -> bool:
        # check if we are to focus on one specific time interval or not

        # if there is only one segment, return false

        if self._number_segments == 1:
            return True

        # check one interval:

        for segment_index in range(self._number_segments - 1):
            # check consecutive segments

            if self.prices[segment_index] < self.prices[segment_index + 1]:
                # prices increased

                return False

        # otherwise

        return True

    # *************************************************************************
    # *************************************************************************

    def is_volume_capped(self) -> bool:
        return not (type(self.volumes[-1]) == type(None))

    # *************************************************************************
    # *************************************************************************

    def is_volume_invariant(self) -> bool:
        """Returns True if the prices are the same regardless of the volume."""
        # is a list
        if len(set(self.prices)) == 1:
            return True  # a list with the same price >> volume invariant
        else:
            return False  # a list with more than 1 price >> volume sensitive

    # *************************************************************************
    # *************************************************************************

    def number_segments(self, redo: bool = False) -> int:
        """Returns the number of price segments."""
        if hasattr(self, "_number_segments") and not redo:
            return self._number_segments
        return len(self.prices)

    # *************************************************************************
    # *************************************************************************
    
    def __eq__(self, o) -> bool:
        """Returns True if a given ResourcePrice is equivalent to another."""
        return hash(self) == hash(o)
    
    def __hash__(self):
        return hash(
            tuple((
                self.number_segments(),
                tuple(self.prices),
                tuple(self.volumes)
                ))
            )

    # *************************************************************************
    # *************************************************************************

# *****************************************************************************
# *****************************************************************************

# TODO: method to determine if qpk-keyed dict is time-invariant per q and p
def are_prices_time_invariant(resource_prices_qpk: dict) -> bool:
    """Returns True if all prices are identical per time interval."""
    # check if there is only one or no (q,p,k) entry
    if len(resource_prices_qpk) <= 1:
        return True  # only one or no entry = is time invariant
    # check if the entries for the same period and assessment are time invariant
    entries_qp = set([qpk[0:2] for qpk in resource_prices_qpk])
    qpk_qp = {
        qp: [qpk for qpk in resource_prices_qpk if qp == qpk[0:2]] for qp in entries_qp
    }
    # check if the tariffs per period and assessment are equivalent
    for qp, qpk_list in qpk_qp.items():
        for i in range(len(qpk_list) - 1):
            if not resource_prices_qpk[qpk_list[0]] == resource_prices_qpk[qpk_list[i + 1]]:
                return False
    # all tariffs are equivalent per period and assessment: they are invariant
    return True


# *****************************************************************************
# *****************************************************************************
