# -*- coding: utf-8 -*-
"""
Created on Wed Nov 17 13:04:53 2021

@author: pmede
"""

# standard libraries

# local libraries, external

# local libraries, internal

# *****************************************************************************
# *****************************************************************************

# TODO: consider mentioning that the classes are actually for discretised systems and signals


class Signal:
    """A class for discretised dynamic system signals."""

    def __init__(
        self, number_samples: int, samples: list, lower_bounds: list, upper_bounds: list
    ):
        # validate the inputs

        (
            self.is_fixed,
            self.is_upper_bounded,
            self.is_lower_bounded,
        ) = self.validate_inputs(number_samples, samples, lower_bounds, upper_bounds)

        # initialise the variables

        self.number_samples = number_samples

        self.samples = samples

        self.lower_bounds = lower_bounds

        self.upper_bounds = upper_bounds

        # is it bounded?

        self.is_bounded = self.is_upper_bounded or self.is_lower_bounded

    # *************************************************************************
    # *************************************************************************

    def validate_inputs(
        self, number_samples: int, samples: list, lower_bounds: list, upper_bounds: list
    ):
        # number of samples

        if type(number_samples) != int:
            raise TypeError("The number of samples has to be an integer.")

        if number_samples <= 0:
            raise ValueError("The number of samples has to be a positive intenger.")

        # samples

        if type(samples) != type(None):
            # it is not None

            if type(samples) != list:
                raise TypeError("The samples have to be provided via a list.")

            if len(samples) != number_samples:
                raise ValueError(
                    "The samples are inconsistent with the number of samples"
                    + " specified."
                )

            is_fixed = True

        else:
            is_fixed = False

        # upper bounds

        if type(upper_bounds) != type(None):
            # it is not None

            if type(upper_bounds) != list:
                raise TypeError("The samples have to be provided via a list.")

            if len(upper_bounds) != number_samples:
                raise ValueError(
                    "The number of upper bounds specified is not consistent"
                    + " with the number of samples specified."
                )

            has_upper_bounds = True

        else:
            has_upper_bounds = False

        # lower bounds

        if type(lower_bounds) != type(None):
            # it is not None

            if type(lower_bounds) != list:
                raise TypeError("The samples have to be provided via a list.")

            if len(lower_bounds) != number_samples:
                raise ValueError(
                    "The number of lower bounds specified is not consistent"
                    + " with the number of samples specified."
                )

            has_lower_bounds = True

        else:
            has_lower_bounds = False

        # the bounds need to be coherent

        if has_lower_bounds and has_upper_bounds:
            for ub, lb in zip(upper_bounds, lower_bounds):
                if ub < lb:
                    raise ValueError("The bounds specified are incoherent.")

        # bounds

        return is_fixed, has_upper_bounds, has_lower_bounds

    # *************************************************************************
    # *************************************************************************

    def has_upper_bounds(self) -> bool:
        "Returns True if the signal has predetermined upper bounds."

        if type(self.upper_bounds) == list:
            if len(self.upper_bounds) == self.number_samples:
                # upper bounds exist with correctly sized lists
                # size

                return True

            else:
                raise ValueError(
                    "The number of upper bounds does not match the number of "
                    + "samples."
                )

        # if it is not a list or does not have the correct number of samples:

        return False

    # *************************************************************************
    # *************************************************************************

    def has_lower_bounds(self) -> bool:
        "Returns True if the signal has predetermined lower bounds."

        if type(self.lower_bounds) == list:
            if len(self.lower_bounds) == self.number_samples:
                # lower bounds exist with correctly sized lists
                # size

                return True

            else:
                raise ValueError(
                    "The number of lower bounds does not match the number of "
                    + "samples."
                )

        # if it is not a list or does not have the correct number of samples:

        return False

    # *************************************************************************
    # *************************************************************************

    def is_signal_bounded(self) -> bool:
        "Returns True if the signal has predetermined lower or upper bounds."

        # a signal is bounded if it has upper or lower bounds

        return self.has_lower_bounds() or self.has_upper_bounds()

    # *************************************************************************
    # *************************************************************************

    def is_signal_fixed(self) -> bool:
        "Returns True if the signal is predetermined or has been fixed."

        # a signal is fixed if samples have been provided as a non-empty list

        if type(self.samples) == list:
            # if the samples object is a list, check its size

            if len(self.samples) == self.number_samples:
                return True

            else:
                raise ValueError(
                    "The samples are inconsistent with the number of samples"
                    + " specified."
                )

        else:
            # if it is not a list, return False

            return False

    # *************************************************************************
    # *************************************************************************

    def violates_bounds(self, tolerance: float = 0.0) -> bool:
        "Returns True if the signal violates its bounds or if it has none."

        if self.is_bounded and self.is_fixed:
            if self.is_lower_bounded:
                # check lower bounds

                for u, u_lb in zip(self.samples, self.lower_bounds):
                    if u < u_lb - tolerance:
                        # the lower bound is violated

                        return True

            if self.is_upper_bounded:
                # check upper bounds

                for u, u_ub in zip(self.samples, self.upper_bounds):
                    if u > u_ub + tolerance:
                        # the upper bound is violated

                        return True

            # if none of the bounds was violated, it is not out of bounds

            return False

        else:
            # return False, if it has no bounds

            return False

    # *************************************************************************
    # *************************************************************************

    def set_signal(self, samples: list):
        "Defines a signal using externally-provided samples."

        # check the size

        if type(samples) != list:
            raise TypeError("The signal must be provided as a list.")

        if len(samples) != self.number_samples:
            raise ValueError("An unexpected number of samples was provided.")

        # store the data

        self.samples = samples  # list(samples)

        # update the relevant attributes

        self.is_fixed = True

    # *************************************************************************
    # *************************************************************************

    def is_signal_integer_only(self, integrality_tolerance: float = 0.0) -> bool:
        "Return True if all the samples in the signal are integers."

        if self.is_fixed:
            # if it is fixed, it may or may not be integer only

            return are_elements_integer(
                self.samples, integrality_tolerance=integrality_tolerance
            )

        else:
            # if it is not fixed, it is not integer only

            return False

    # *************************************************************************
    # *************************************************************************

    def is_signal_binary_only(self, integrality_tolerance: float = 0.0) -> bool:
        "Return True if all the samples in the signal are binary."

        if self.is_fixed:
            # if it is fixed, it may or may not be integer only

            return are_elements_binary(
                self.samples, integrality_tolerance=integrality_tolerance
            )

        else:
            # if it is not fixed, it is not integer only

            return False

    # *************************************************************************
    # *************************************************************************


# *****************************************************************************
# *****************************************************************************


def are_elements_integer(
    elements: list, integrality_tolerance: float or int = 0.0
) -> bool:
    "Returns True if all the elements in a list are integer numbers."

    # check the tolerance

    if type(integrality_tolerance) != float and type(integrality_tolerance) != int:
        raise TypeError("The integrality tolerance should be a float or zero.")

    if integrality_tolerance >= 0.5 or integrality_tolerance < 0.0:
        raise ValueError(
            "The integrality tolerance should not be greater than or equal to"
            + " 0.5 for the distinction (i.e., integrality) to make sense."
        )

    # go through each element

    for element in elements:
        # if at least one isn't, it isn't

        if abs(element - round(element)) > integrality_tolerance:
            # it is not integer

            return False

    return True


# *****************************************************************************
# *****************************************************************************


def are_elements_binary(
    elements: list, integrality_tolerance: float or int = None
) -> bool:
    "Returns True if all the elements in a list are binary numbers."

    # no particular integrality tolerance was defined: assume rounding is okay

    if type(integrality_tolerance) == type(None):
        # go through each element

        for element in elements:
            # check if it is an integer

            if round(element) != 1 and round(element) != 0:
                # if not 0 or 1, it is not binary

                return False

        # if this point is reached, all the elements are binary

        return True

    # *************************************************************************

    # a specific integrality tolerance was provided

    if type(integrality_tolerance) != float and type(integrality_tolerance) != int:
        raise TypeError("The integrality tolerance should be a float or zero.")

    if integrality_tolerance >= 0.5 or integrality_tolerance < 0.0:
        raise ValueError(
            "The integrality tolerance should not be greater than or equal to"
            + " 0.5 for the distinction (i.e., integrality) to make sense."
        )

    # go through each element

    for element in elements:
        rounded_element = round(element)

        # check if it is an integer

        if abs(element - rounded_element) > integrality_tolerance:
            # it is not integer

            return False

        # if it is an integer, check if it is zero or one

        if rounded_element != 1 and rounded_element != 0:
            # if not 0 or 1, it is not binary

            return False

    # if this point is reached, all the elements are binary

    return True


# *****************************************************************************
# *****************************************************************************


class FixedSignal(Signal):
    "A class for signals that are predetermined, measured or decided on."

    def __init__(
        self, samples: list, lower_bounds: list = None, upper_bounds: list = None
    ):
        if type(samples) != list:
            raise TypeError("The samples must be provided as a list.")

        Signal.__init__(
            self,
            number_samples=len(samples),
            samples=samples,
            lower_bounds=lower_bounds,
            upper_bounds=upper_bounds,
        )


# *****************************************************************************
# *****************************************************************************


class NonNegativeRealSignal(Signal):
    "A class for non-negative real signals."

    def __init__(
        self,
        number_samples: int,
        samples: list = None,
        lower_bounds: list = None,
        upper_bounds: list = None,
    ):
        if type(number_samples) != int:
            raise TypeError("The number of samples must be an integer.")

        if type(lower_bounds) != type(None):
            # specific lower bounds have been provided

            Signal.__init__(
                self,
                number_samples=number_samples,
                samples=samples,
                lower_bounds=lower_bounds,
                upper_bounds=upper_bounds,
            )

        else:
            # no lower bounds have been provided: zero is the lower bound

            Signal.__init__(
                self,
                number_samples=number_samples,
                samples=samples,
                lower_bounds=[0 for _ in range(number_samples)],
                upper_bounds=upper_bounds,
            )

        # TODO: check bounds before calling the base class init method

        # make sure the bounds are nnr

        if not self.are_bounds_nnr():
            raise ValueError("The bounds are not non-negative real.")

    # *************************************************************************
    # *************************************************************************

    def is_nnr(self, tolerance: float = 0.0) -> bool:
        "Returns True if the samples are all non-negative real."

        if self.is_fixed:
            # if it is fixed, make sure each sample is nnr

            for u in self.samples:
                if u < 0 - tolerance:
                    return False

            # if all bounds are positive or zero, return True

            return True

        else:
            # if it is not fixed, return false

            return False

    # *************************************************************************
    # *************************************************************************

    def are_bounds_nnr(self, tolerance: float = 0.0) -> bool:
        "Returns True if the bounds are non-negative real."

        # if no bounds have been provided, check the internal bounds

        if self.is_lower_bounded:
            # if the signal is bounded, its bounds may or may not be nnr

            # check the lower bounds

            # if it does, make sure each lower bound is positive or zero

            for lb in self.lower_bounds:
                if lb < 0 - tolerance:
                    return False

            # check the upper bounds

            if self.is_upper_bounded:
                # if it has upper bounds, make sure they are positive or zero

                for ub in self.upper_bounds:
                    if ub < 0 - tolerance:
                        return False

                # note: the base class's __init__ method will check the bounds
                # for sanity (i.e., ub >= lb), and so that step is skipped here

            # if both the upper and lower bounds (if any) are nnr, return True

            return True

        else:
            # if the signal is not bounded, its bounds are not nnr

            return False

    # *************************************************************************
    # *************************************************************************


# *****************************************************************************
# *****************************************************************************


class FixedNonNegativeRealSignal(NonNegativeRealSignal):
    "A class for free and controllable signals."

    def __init__(
        self, samples: list, lower_bounds: list = None, upper_bounds: list = None
    ):
        if type(samples) != list:
            raise TypeError("The samples must be provided via a list.")

        NonNegativeRealSignal.__init__(
            self,
            number_samples=len(samples),
            samples=samples,
            lower_bounds=lower_bounds,
            upper_bounds=upper_bounds,
        )

        if not self.is_nnr():
            raise ValueError("At least one sample is not non-negative real.")


# *****************************************************************************
# *****************************************************************************


class FreeSignal(Signal):
    "A class for undetermined signals."

    def __init__(
        self, number_samples: int, lower_bounds: list = None, upper_bounds: list = None
    ):
        Signal.__init__(
            self,
            number_samples=number_samples,
            samples=None,
            lower_bounds=lower_bounds,
            upper_bounds=upper_bounds,
        )


# *****************************************************************************
# *****************************************************************************


class FreeUnboundedSignal(FreeSignal):
    "A class for free and unbounded signals."

    def __init__(self, number_samples: int):
        FreeSignal.__init__(
            self, number_samples=number_samples, lower_bounds=None, upper_bounds=None
        )


# *****************************************************************************
# *****************************************************************************


class BinarySignal(NonNegativeRealSignal):
    "A class for binary signals."

    def __init__(self, number_samples: int):
        if type(number_samples) != int:
            raise TypeError("The number of samples must be an integer.")

        NonNegativeRealSignal.__init__(
            self,
            number_samples=number_samples,
            samples=None,
            lower_bounds=[0 for _ in range(number_samples)],
            upper_bounds=[1 for _ in range(number_samples)],
        )


# *****************************************************************************
# *****************************************************************************


class AmplitudeConstrainedSignal(Signal):
    "A class for amplitude-constrained signals."

    def __init__(
        self,
        number_samples: int,
        max_pos_amp_limit: float or int,
        min_pos_amp_limit: float or int,
        max_neg_amp_limit: float or int,
        min_neg_amp_limit: float or int,
        positive_amplitude: list = None,
        negative_amplitude: list = None,
        samples: list = None,
        lower_bounds: list = None,
        upper_bounds: list = None,
    ):
        # validate the inputs

        self.set_positive_amplitude_limits(max_pos_amp_limit, min_pos_amp_limit)

        self.set_negative_amplitude_limits(max_neg_amp_limit, min_neg_amp_limit)

        # call the base class init method

        Signal.__init__(
            self,
            number_samples=number_samples,
            samples=samples,
            lower_bounds=lower_bounds,
            upper_bounds=upper_bounds,
        )

        # validate the bounds

        self.validate_positive_bounds(
            self.has_max_pos_amp_limit,
            self.max_pos_amp_limit,
            self.is_lower_bounded,
            self.lower_bounds,
        )

        self.validate_negative_bounds(
            self.has_max_neg_amp_limit,
            self.max_neg_amp_limit,
            self.is_upper_bounded,
            self.upper_bounds,
        )

        # set the amplitudes

        self.set_positive_amplitude(positive_amplitude)

        self.set_negative_amplitude(negative_amplitude)

    # *************************************************************************

    def violates_amplitude_limits(
        self, samples: list = None, tolerance: float = 0.0
    ) -> bool:
        """
        Returns True if the samples violate the signal's amplitude limits.

        Parameters
        ----------
        samples : list, optional
            The set of samples to be evaluated. The default is None, which cau-
            ses the function to evaluate the internal samples, if any.
        tolerance : float, optional
            The tolerance with which to compare the limits. The default is 0.0.

        Raises
        ------
        ValueError
            This error is caused when no samples are available for evaluation.

        Returns
        -------
        bool
            If True, the samples violate the limits. If False, they do not.

        """

        # if there are no amplitude limits, return False

        if self.has_pos_amp_limits or self.has_neg_amp_limits:
            # if there are amplitude limits, check each sample

            if type(samples) == list:
                # external samples were provided, check them

                for sample in samples:
                    # positive amplitude limits

                    if self.has_max_pos_amp_limit:
                        if sample > self.max_pos_amp_limit + tolerance:
                            return True

                    if self.has_max_neg_amp_limit:
                        if sample < -self.max_neg_amp_limit - tolerance:
                            return True

            elif self.is_signal_fixed():
                # no external samples were provided, check the internal samples

                for sample in self.samples:
                    # positive amplitude limits

                    if self.has_max_pos_amp_limit:
                        if sample > self.max_pos_amp_limit + tolerance:
                            return True

                    if self.has_max_neg_amp_limit:
                        if sample < -self.max_neg_amp_limit - tolerance:
                            return True
            else:
                # the signal has not been set

                raise ValueError("There are no samples to evaluate.")

            # if no other return statement was reached, return False

            return False

        else:
            # if there are no amplitude limits, return False

            return False

    # *************************************************************************

    def set_positive_amplitude(self, positive_amplitude: float or int = None):
        """
        Sets the positive amplitude.

        Parameters
        ----------
        positive_amplitude : float or int, optional
            The positive amplitude. The default is None.

        Raises
        ------
        TypeError
            This error is caused if invalid types are provided.

        Returns
        -------
        None.

        """

        if (
            type(positive_amplitude) == int
            or isinstance(positive_amplitude, float)
            or type(positive_amplitude) == type(None)
        ):
            self.positive_amplitude = positive_amplitude

        else:
            raise TypeError("Unknown positive amplitude type.")

    # *************************************************************************

    def set_negative_amplitude(self, negative_amplitude: float or int = None):
        """
        Sets the positive amplitude.

        Parameters
        ----------
        negative_amplitude : float or int, optional
            The negative amplitude. The default is None.

        Raises
        ------
        TypeError
            This error is caused if invalid types are provided.

        Returns
        -------
        None.

        """

        if (
            type(negative_amplitude) == int
            or isinstance(negative_amplitude, float)
            or type(negative_amplitude) == type(None)
        ):
            self.negative_amplitude = negative_amplitude

        else:
            raise TypeError("Unknown negative amplitude type.")

    # *************************************************************************

    def set_positive_amplitude_limits(
        self, max_pos_amp_limit: float or int, min_pos_amp_limit: float or int
    ):
        """
        Sets the maximum and minimum positive amplitude limits for the signal.

        The positive amplitude limits refer to the maximum value the signal can
        assume when positive. As such, it will affect feasiblity if the signal
        has lower bounds above the maximum limit.

        Parameters
        ----------
        max_pos_amp_limit : float or int
            The maximum positive amplitude limit, which is a positive number.
        min_pos_amp_limit : float or int
            The minimum positive amplitude limit, which is a positive number.

        Returns
        -------
        None.

        """

        # validate inputs

        (self.has_max_pos_amp_limit, self.has_min_pos_amp_limit) = self.validate_limits(
            max_pos_amp_limit, min_pos_amp_limit
        )

        if self.has_max_pos_amp_limit:
            self.max_pos_amp_limit = max_pos_amp_limit

        else:
            self.max_pos_amp_limit = None

        # minimum amplitude limit

        if self.has_min_pos_amp_limit:
            self.min_pos_amp_limit = min_pos_amp_limit

        else:
            self.min_pos_amp_limit = None

        # update the status

        self.has_pos_amp_limits = (
            self.has_max_pos_amp_limit or self.has_min_pos_amp_limit
        )

    # *************************************************************************

    def set_negative_amplitude_limits(
        self, max_neg_amp_limit: float or int, min_neg_amp_limit: float or int
    ):
        """
        Sets the maximum and minimum negative amplitude limits for the signal.

        The negative amplitude limits refer to the minimum value the signal can
        assume when negative, though it is a positive number. As such, it will
        affect feasiblity if the upper bounds are below what the maximum limit
        allows for (i.e., if the upper bounds are negative and more so).

        Parameters
        ----------
        max_neg_amp_limit : float or int
            The maximum negative amplitude limit, which is a positive number.
        min_neg_amp_limit : float or int
            The minimum negative amplitude limit, which is a positive number.

        Returns
        -------
        None.

        """

        # validate inputs

        (self.has_max_neg_amp_limit, self.has_min_neg_amp_limit) = self.validate_limits(
            max_neg_amp_limit, min_neg_amp_limit
        )

        if self.has_max_neg_amp_limit:
            self.max_neg_amp_limit = max_neg_amp_limit

        else:
            self.max_neg_amp_limit = None

        # minimum amplitude limit

        if self.has_min_neg_amp_limit:
            self.min_neg_amp_limit = min_neg_amp_limit

        else:
            self.min_neg_amp_limit = None

        # update the status

        self.has_neg_amp_limits = (
            self.has_max_neg_amp_limit or self.has_min_neg_amp_limit
        )

    # *************************************************************************

    @staticmethod
    def validate_limits(
        max_amp_limit: float or int, min_amp_limit: float or int
    ) -> tuple:
        """
        Validates maximum and minimum amplitude limits.

        Parameters
        ----------
        max_amp_limit : float or int or None (NoneType)
            The maximum amplitude limit. If None, no maximum limit exists.
        min_amp_limit : float or int or None (NoneType)
            The minimum amplitude limit. If None, no minimum limit exists.

        Raises
        ------
        ValueError
            This error is caused if the amplitude limits are not positive or if
            the limits are not consistent (minimum <= maximum).
        TypeError
            This error is caused if the amplitude limits are non-numeric or not
            None.

        Returns
        -------
        tuple
            A tuple containing two booleans indicating if there are valid and
            coherent maximum (first) and minimum (second) amplitude limits.

        """

        # both amplitudes need to be positive or None

        # max amplitude

        if type(max_amp_limit) == int or isinstance(max_amp_limit, float):
            if max_amp_limit <= 0:
                raise ValueError("The amplitude limits need to be positive.")

            has_max_amp = True

        elif type(max_amp_limit) == type(None):
            has_max_amp = False

        else:  # none of the above
            raise TypeError("Unrecognised amplitude format.")

        # min amplitude

        if type(min_amp_limit) == int or isinstance(min_amp_limit, float):
            if min_amp_limit <= 0:
                raise ValueError("The amplitude limits need to be positive.")

            has_min_amp = True

        elif type(min_amp_limit) == type(None):
            has_min_amp = False

        else:  # none of the above
            raise TypeError("Unrecognised amplitude format.")

        # the minimum amplitude must be lower than the maximum, if both exist

        if has_max_amp and has_min_amp:
            if min_amp_limit >= max_amp_limit:
                raise ValueError(
                    "The maximum amplitude limit is not greater than the " + "minimum."
                )

        return has_max_amp, has_min_amp

    # *************************************************************************

    @staticmethod
    def validate_positive_bounds(
        has_max_pos_amp_limit: bool,
        max_pos_amp_limit: float or int,
        is_lower_bounded: bool,
        lower_bounds: list,
    ):
        """
        Validates positive amplitude limits against any existing bounds.

        Parameters
        ----------
        has_max_pos_amp_limit : bool
            If True, there is a maximum positive amplitude limit.
        max_pos_amp_limit : float or int
            The maximum positive amplitude limit, or None, if no limit exists.
        is_lower_bounded : bool
            If True, there are lower bounds.
        lower_bounds : list
            The list of lower bounds, or None, if no lower bounds exist.

        Raises
        ------
        ValueError
            This error is caused if the bounds and limits are not compatible.

        Returns
        -------
        None.

        """

        # the lower bounds cannot be below the maximum amplitude limit

        if has_max_pos_amp_limit and is_lower_bounded:
            # the signal is lower bounded and has a maximum pos. amp. limit

            min_lower_bound = min(lower_bounds)

            if min_lower_bound > max_pos_amp_limit:
                raise ValueError(
                    "The positive amplitude limits are not compatible with "
                    + " the bounds provided."
                )

    # *************************************************************************

    @staticmethod
    def validate_negative_bounds(
        has_max_neg_amp_limit: bool,
        max_neg_amp_limit: float or int,
        is_upper_bounded: bool,
        upper_bounds: list,
    ):
        """
        Validates negative amplitude limits against any existing bounds.

        Parameters
        ----------
        has_max_neg_amp_limit : bool
            If True, there is a maximum negative amplitude limit.
        max_neg_amp_limit : float or int
            The maximum negative amplitude limit, or None, if no limit exists.
        is_upper_bounded : bool
            If True, there are upper bounds.
        upper_bounds : list
            The list of upper bounds, or None, if no upper bounds exist.

        Raises
        ------
        ValueError
            This error is caused if the bounds and limits are not compatible.

        Returns
        -------
        None.

        """

        # the upper bounds cannot be below the level implied by the maximum ne-
        # gative amplitude limit, since that leads to infeasibility
        # max_upper_bound < -max_neg_amp_limit = infeasibility

        if has_max_neg_amp_limit and is_upper_bounded:
            # the signal is upper bounded and has a maximum neg. amp. limit

            max_upper_bound = max(upper_bounds)

            if max_upper_bound < -max_neg_amp_limit:
                raise ValueError(
                    "The negative amplitude limits are not compatible with "
                    + " the bounds provided."
                )

    # *************************************************************************

    def validate_positive_amplitude(
        self, positive_amplitude: float or int = None, tolerance: float or int = 0.0
    ):
        # without amplitude limits, there is nothing to validate against

        if not self.has_pos_amp_limits:
            raise ValueError("There are no limits against which to validate.")

        else:
            # if there are limits, the amplitudes have to fall within their
            # respective ranges: min amplitude <= amplitude <= max amplitude

            # the method is the same for both amplitudes

            # positive amplitude

            if type(positive_amplitude) == type(None):
                # compare internal positive amplitude

                if self.has_max_pos_amp_limit:
                    # there is a maximum amplitude limit

                    if self.positive_amplitude > self.max_pos_amp_limit + tolerance:
                        # the maximum was exceeded

                        raise ValueError(
                            "The positive amplitude exceeds its tolerated " + "maximum."
                        )

                if self.has_min_pos_amp_limit:
                    # there is a minimum amplitude limit

                    if self.positive_amplitude < self.min_pos_amp_limit - tolerance:
                        # the minimum was not observed

                        raise ValueError(
                            "The positive amplitude is below its tolerated "
                            + "minimum."
                        )

            else:
                # compare external positive amplitude

                if self.has_max_pos_amp_limit:
                    # there is a maximum amplitude limit

                    if positive_amplitude > self.max_pos_amp_limit + tolerance:
                        # the maximum was exceeded

                        raise ValueError(
                            "The positive amplitude exceeds its tolerated " + "maximum."
                        )

                if self.has_min_pos_amp_limit:
                    # there is a minimum amplitude limit

                    if positive_amplitude < self.min_pos_amp_limit - tolerance:
                        # the minimum was not observed

                        raise ValueError(
                            "The positive amplitude is below its tolerated "
                            + "minimum."
                        )

    # *************************************************************************

    def validate_negative_amplitude(
        self, negative_amplitude: float or int = None, tolerance: float or int = 0.0
    ):
        # without amplitude limits, there is nothing to validate against

        if not self.has_neg_amp_limits:
            raise ValueError("There are no limits against which to validate.")

        else:
            # if there are limits, the amplitudes have to fall within their
            # respective ranges: min amplitude <= amplitude <= max amplitude

            # the method is the same for both amplitudes

            # negative amplitude

            if type(negative_amplitude) == type(None):
                # compare internal negative amplitude

                if self.has_max_neg_amp_limit:
                    # there is a maximum amplitude limit

                    if self.negative_amplitude > self.max_neg_amp_limit + tolerance:
                        # the maximum was exceeded

                        raise ValueError(
                            "The negative amplitude exceeds its tolerated " + "maximum."
                        )

                if self.has_min_neg_amp_limit:
                    # there is a minimum amplitude limit

                    if self.negative_amplitude < self.min_neg_amp_limit - tolerance:
                        # the minimum was not observed

                        raise ValueError(
                            "The negative amplitude is below its tolerated "
                            + "minimum."
                        )

            else:
                # compare external negative amplitude

                if self.has_max_neg_amp_limit:
                    # there is a maximum amplitude limit

                    if negative_amplitude > self.max_neg_amp_limit + tolerance:
                        # the maximum was exceeded

                        raise ValueError(
                            "The negative amplitude exceeds its tolerated " + "maximum."
                        )

                if self.has_min_neg_amp_limit:
                    # there is a minimum amplitude limit

                    if negative_amplitude < self.min_neg_amp_limit - tolerance:
                        # the minimum was not observed

                        raise ValueError(
                            "The negative amplitude is below its tolerated "
                            + "minimum."
                        )


# *****************************************************************************
# *****************************************************************************


class AmplitudeConstrainedNNRSignal(AmplitudeConstrainedSignal, NonNegativeRealSignal):
    "A class for non-negative real (NNR) amplitude-constrained signals."

    def __init__(
        self,
        number_samples: int,
        max_pos_amp_limit: float or int,
        min_pos_amp_limit: float or int = None,
        positive_amplitude: float or int = None,
        samples: list = None,
        lower_bounds: list = None,
        upper_bounds: list = None,
    ):
        if type(lower_bounds) != type(None):
            # call the AmplitudeConstrainedSignal class's init method

            AmplitudeConstrainedSignal.__init__(
                self,
                number_samples=number_samples,
                max_pos_amp_limit=max_pos_amp_limit,
                min_pos_amp_limit=min_pos_amp_limit,
                max_neg_amp_limit=None,
                min_neg_amp_limit=None,
                positive_amplitude=positive_amplitude,
                negative_amplitude=None,
                samples=samples,
                lower_bounds=lower_bounds,
                upper_bounds=upper_bounds,
            )

        else:  # no peculiar lower bounds were specified
            # the lower bounds must be set to zero

            # the number of samples should be checked before calling the method

            if type(number_samples) != int:
                # the number of samples was not correctly specified

                raise TypeError("The number of samples must be an integer.")

            # call the AmplitudeConstrainedSignal class's init method

            AmplitudeConstrainedSignal.__init__(
                self,
                number_samples=number_samples,
                max_pos_amp_limit=max_pos_amp_limit,
                min_pos_amp_limit=min_pos_amp_limit,
                max_neg_amp_limit=None,
                min_neg_amp_limit=None,
                positive_amplitude=positive_amplitude,
                negative_amplitude=None,
                samples=samples,
                lower_bounds=[0 for _ in range(number_samples)],
                upper_bounds=upper_bounds,
            )

        # make sure the bounds are nnr

        if not self.are_bounds_nnr():
            raise ValueError("The bounds are not non-negative real.")


# *****************************************************************************
# *****************************************************************************
