# -*- coding: utf-8 -*-

# -*- coding: utf-8 -*-
"""
Created on Wed Nov 17 13:04:53 2021

@author: pmede
"""

# standard libraries

# local libraries, external

# local libraries, internal

from .converter import Converter

from .network import Network

# *****************************************************************************
# *****************************************************************************


class EnergySystem:
    """A class to model energy systems for simulation or optimisation."""

    # instances of this class should contain enough information to simulate or
    # optimise the energy system. In the former case, no decisions can be left
    # out.

    # systems:
    # optional or not
    # decision to install, if not already defined (if not optional)
    # inputs, for every mode and time interval
    # outputs, for every output and time interval
    # states, for every node and time interval

    # networks:
    # flows
    # technologies selected

    # *************************************************************************
    # *************************************************************************

    def __init__(
        self,
        networks: dict = None,
        converters: dict = None,
        optional_converters: list = None,
        selected_converters: list = None,
    ):
        # *********************************************************************
        # *********************************************************************

        # networks should be a dict of nx.MultiDiGraph objects

        if networks != None and type(networks) == dict:
            self.networks = dict(networks)

        else:
            self.networks = {}

        # systems should be a dict of DynamicSystem objects

        if converters != None and type(converters) == dict:
            self.converters = dict(converters)

        else:
            self.converters = {}

        # *********************************************************************
        # *********************************************************************

        # optional converters

        # list to indicate which converters are optional
        # elements: converter keys
        # default outcome (if empty): all converters are mandatory

        if optional_converters != None and type(optional_converters) == list:
            self.optional_converters = list(optional_converters)

        else:
            self.optional_converters = []

        # *********************************************************************
        # *********************************************************************

        # selected converters

        # list to indicate which converters have been selected
        # elements: converter keys
        # default outcome (if empty): no converters have been selected

        if selected_converters != None and type(selected_converters) == list:
            self.selected_converters = list(selected_converters)

        else:
            self.selected_converters = []

        # *********************************************************************
        # *********************************************************************

        # TODO: explain why some of attributes should stay here

        # dict to indicate the maximum number of arcs between two nodes
        # key: (g,u,v) tuple
        # value: int
        # default outcome (leaving it empty): +inf

        self.max_number_parallel_arcs = {}

        # list to indicate which specific arcs are selectable yet mandatory
        # elements: (g,u,v,j) tuples
        # default outcome (by leaving it empty): optional/False
        # Note: it is placed here because it is an implication of the lack of
        # alternatives to meet demand/supply in a given node

        self.mandatory_arcs = []

        # dict indicating which arcs
        # key: (network, node, converter, output, time interval)
        # value: coefficient relative to a gi
        # default outcome (if tuple not present):

        self.network_effects = {}

        # *********************************************************************
        # *********************************************************************

    # *************************************************************************
    # *************************************************************************

    def add_network(self, network_key, network: Network):
        """Add a new network to the energy system."""

        self.networks[network_key] = network

    # *************************************************************************
    # *************************************************************************

    def add_converter(
        self,
        converter_key,
        converter: Converter,
        is_optional: bool,
        is_selected: bool = False,
    ) -> bool:
        """Add a converter to the energy system object or update it."""

        # sanity check: mandatory systems cannot have dimensiona. input signals

        if converter.has_dimensionable_input_signals() and not is_optional:
            # the ds is not supposed to be optional but has dim. input signals

            return False

        # update the dict

        self.converters[converter_key] = converter

        # set whether or not it is optional

        self.set_converter_as_optional_or_not(converter_key, is_optional)

        # set whether or not it has been selected

        self.set_converter_selection_status(converter_key, is_selected)

        return True

    # *************************************************************************
    # *************************************************************************

    def set_converter_as_optional_or_not(self, converter_key, is_optional: bool):
        # if is_optional is True: add key to list, if not already there
        # if is_optional is False: remove key from list, if there already

        if is_optional:
            if converter_key not in self.optional_converters:
                self.optional_converters.append(converter_key)

        else:
            while converter_key in self.optional_converters:
                self.optional_converters.remove(converter_key)

    # *************************************************************************
    # *************************************************************************

    def set_converter_selection_status(self, converter_key, selection_status: bool):
        # if selection_status is True: add key to list, if not already there
        # if selection_status is False: remove key from list, if there already

        if selection_status:
            if converter_key not in self.selected_converters:
                self.selected_converters.append(converter_key)

        else:
            while converter_key in self.selected_converters:
                self.selected_converters.remove(converter_key)

    # *************************************************************************
    # *************************************************************************

    def set_converter_network_effects(
        self, converter_key, network_effects: dict
    ) -> bool:
        # # validate converter key

        # if converter_key not in self.converters.keys():

        #     return False

        # # validate dict entries

        # for dict_tuple in network_effects.keys():

        #     if len(dict_tuple) != 4: # (network, node, output, time interval)

        #         return False

        #     if dict_tuple[0] not in self.networks.keys():

        #         return False

        #     if dict_tuple[1] not in self.networks[dict_tuple[0]].nodes():

        #         return False

        #     if dict_tuple[2] not in range(
        #             self.converters[converter_key].number_outputs):

        #         return False

        #     if dict_tuple[3] not in range(self.number_time_intervals):

        #         return False

        #     # update dict

        #     self.network_effects[
        #         (dict_tuple[0],
        #          dict_tuple[1],
        #          converter_key,
        #          dict_tuple[2],
        #          dict_tuple[3])] = networks_effects[]

        raise NotImplementedError

    # *************************************************************************
    # *************************************************************************

    def make_arc_mandatory(self, network_key, arc_key: tuple):
        """
        Marks an arc as being mandatory in an energy system.

        Parameters
        ----------
        network_key : hashable-type
            The key to the network object.
        arc_key : hashable-type
            The key to the arc in the specified network.

        Raises
        ------
        ValueError
            This error is raised if the key is not recognised or if the tuple
            does not have the right size.

        Returns
        -------
        None.

        """

        if network_key in self.networks and len(arc_key) == 3:
            if self.networks[network_key].has_edge(*arc_key):
                # can only be used for selectable arcs: not preselected

                if (
                    self.networks[network_key]
                    .edges[arc_key][Network.KEY_ARC_TECH]
                    .has_been_selected()
                ):
                    raise ValueError("The arc selected has been preselected.")

                if (network_key, *arc_key) not in self.mandatory_arcs:
                    self.mandatory_arcs.append((network_key, *arc_key))

            else:  # the arc does not exist
                raise ValueError(
                    "The arc key used does not match any arc in the network."
                )

        else:  # something is up with the network key or arc key length
            raise ValueError(
                "Either the network key provided is incorrect or the arc key "
                + "lacks the proper size."
            )

    # *************************************************************************
    # *************************************************************************

    def unmake_arc_mandatory(self, network_key, arc_key: tuple):
        """
        Unmarks an arc from being mandatory in an energy system.

        Parameters
        ----------
        network_key : hashable-type
            The key to the network object.
        arc_key : hashable-type
            The key to the arc in the specified network.

        Raises
        ------
        ValueError
            This error is raised if the key is not recognised or if the tuple
            does not have the right size.

        Returns
        -------
        None.

        """

        if network_key in self.networks and len(arc_key) == 3:
            if (network_key, *arc_key) in self.mandatory_arcs:
                # if it does, pop it

                self.mandatory_arcs.pop((network_key, *arc_key))

        else:  # something is up with the network key or arc key length
            raise ValueError(
                "Either the network key provided is incorrect or the arc key "
                + "lacks the proper size."
            )

    # *************************************************************************
    # *************************************************************************

    def set_maximum_number_parallel_arcs(self, network_key, node_a, node_b, limit: int):
        # TODO: docstring

        if network_key in self.networks:
            if self.networks[network_key].has_edge(node_a, node_b):
                self.max_number_parallel_arcs[(network_key, node_a, node_b)] = limit

    # *************************************************************************
    # *************************************************************************

    # TODO: create a method to identify mandatory nodes


# *****************************************************************************
# *****************************************************************************
