# *****************************************************************************
# *****************************************************************************

from numbers import Real
from ...data.finance.invest import discount_factor as _discount_factor

# *****************************************************************************
# *****************************************************************************

class TimeFrame:
    "A class to model time-dependent problems and systems."

    # why?
    # 1) to ensure a consistent time frame for all time-dependent objects
    # 2) to validate conformity with this time frame

    # input modes:
    # 1) uniform time discretisation
    # - number of reporting periods
    # - duration of the reporting period(s)
    # - number of intervals
    # - duration of the interval(s)
    # 2) customised
    # - dicts

    def __init__(
        self,
        # number_reporting_periods: int,
        # number_time_intervals:int,
        reporting_periods: dict,
        reporting_period_durations: dict,
        time_intervals: dict,
        time_interval_durations: dict,
    ):
        self.assessments = set(time_intervals.keys())

        self.time_intervals = time_intervals
        self.time_interval_durations = time_interval_durations

        self.reporting_periods = reporting_periods
        self.reporting_period_durations = reporting_period_durations

        # check for nonpositive durations
        for q, durations in reporting_period_durations.items():
            for duration in durations:
                if duration <= 0:
                    raise ValueError("Durations need to be positive.")
        for q, durations in time_interval_durations.items():
            for duration in durations:
                if duration <= 0:
                    raise ValueError("Durations need to be positive.")

        # declare internal states
        self.number_assessments(redo=True)
        self.number_reporting_periods(next(iter(self.assessments)), redo=True)
        self.number_time_intervals(next(iter(self.assessments)), redo=True)
        self.qk(redo=True)
        self.qp(redo=True)
        self.qpk(redo=True)

    # *************************************************************************
    # *************************************************************************
    
    def number_assessments(self, redo=False) -> int:
        "Returns the number of assessments under consideration."
        if redo:
            self._number_q = len(self.assessments)
        return self._number_q
    
    def number_time_intervals(self, assessment, redo=False) -> int:
        "Returns the number of intervals for a given assessment."
        if redo:
            self._number_k_q = {
                q: len(self.time_intervals[q])
                for q in self.assessments
                }
        return self._number_k_q[assessment]
    
    def number_reporting_periods(self, assessment, redo=False) -> int:
        "Returns the number of reporting periods for a given assessment."
        if redo:
            self._number_p_q = {
                q: len(self.reporting_periods[q])
                for q in self.assessments
                }
        return self._number_p_q[assessment]

    # *************************************************************************
    # *************************************************************************
    
    def valid_q(self, q_keyed_dict: dict) -> bool:
        "Returns True if all q keys are valid, and False otherwise."
        for q in q_keyed_dict.keys():
            if q not in self.assessments:
                return False
        # all q keys are valid
        return True
        
    def complete_q(self, q_keyed_dict: dict) -> bool:
        "Returns True if all valid q keys are provided."
        for q in self.assessments:
            if q not in q_keyed_dict:
                # the dict keys do not include this q tuple
                return False
        # if all q tuples are valid and complete, return True
        return True

    # *************************************************************************
    # *************************************************************************

    def valid_qpk(self, qpk_keyed_dict: dict) -> bool:
        "Returns True if all (q,p,k) tuple keys are valid, and False otherwise."
        for qpk in qpk_keyed_dict.keys():
            if (
                len(qpk) != 3
                or qpk[0] not in self.assessments
                or qpk[1] not in self.reporting_periods[qpk[0]]
                or qpk[2] not in self.time_intervals[qpk[0]]
            ):
                return False
        # if all (q,p) tuples are valid, return True
        return True

    def complete_qpk(self, qpk_keyed_dict: dict) -> bool:
        "Returns True if all valid (q,p,k) tuple keys are provided."
        for q, _p in self.reporting_periods.items():
            for p in _p:
                for k in self.time_intervals[q]:
                    if (q, p, k) not in qpk_keyed_dict:
                        # the dict keys do not include this (q,p,k) tuple
                        return False
        # if all (q,p,k) tuples are valid and complete, return True
        return True

    def consecutive_qpk(self, qpk_keyed_dict: dict) -> bool:
        "Returns True if all (q,p,k) tuple keys are valid and consecutive."
        # all k intervals have to be consecutive for each (q,p) pair
        set_qp = set(qpk[0:2] for qpk in qpk_keyed_dict.keys())
        for qp in set_qp:
            for k in range(self.number_time_intervals(qp[0])):
                if (*qp,k) not in qpk_keyed_dict:
                    return False
        return True

    # *************************************************************************
    # *************************************************************************

    def valid_qp(self, qp_keyed_dict: dict) -> bool:
        "Returns True if all (q,p) tuple keys are valid, and False otherwise."
        for qp in qp_keyed_dict.keys():
            if (
                len(qp) != 2
                or qp[0] not in self.assessments
                or qp[1] not in self.reporting_periods[qp[0]]
            ):
                return False
        # if all (q,p) tuples are valid, return True
        return True

    def complete_qp(self, qp_keyed_dict: dict) -> bool:
        "Returns True if all valid (q,p) tuple keys are provided."
        for q, _p in self.reporting_periods.items():
            for p in _p:
                if (q, p) not in qp_keyed_dict:
                    # the dict keys do not include this (q,p) tuple
                    return False
        # if all (q,p) tuples are valid and complete, return True
        return True

    def consecutive_qp(self, qp_keyed_dict: dict) -> bool:
        "Returns True if all p entries for each q are consecutive."
        return self.consecutive_qk(qk_keyed_dict=qp_keyed_dict)

    # *************************************************************************
    # *************************************************************************

    def valid_qk(self, qk_keyed_dict: dict) -> bool:
        "Returns True if all (q,k) tuple keys are valid, and False otherwise."
        for qk in qk_keyed_dict.keys():
            if (
                len(qk) != 2
                or qk[0] not in self.assessments
                or qk[1] not in self.time_intervals[qk[0]]
            ):
                return False
        # if all (q,k) tuples are valid, return True
        return True

    def complete_qk(self, qk_keyed_dict: dict):
        "Returns True if all valid (q,k) tuple keys are provided."
        for q, _k in self.time_intervals.items():
            for k in _k:
                if (q, k) not in qk_keyed_dict:
                    # the dict keys do not include this (q,k) tuple
                    return False
        # if all (q,k) tuples are valid and complete, return True
        return True

    def consecutive_qk(self, qk_keyed_dict: dict) -> bool:
        "Returns True if all k entries for each q are consecutive."
        # create q-keyed dict with the respective k-values
        new_dict = {
            q: set(tuple(qk[1] for qk in qk_keyed_dict.keys() if qk[0] == q))
            for q in set(tuple(qk[0] for qk in qk_keyed_dict.keys()))
            }
        # for each key, verify that the ks are consecutive
        for q, k_set in new_dict.items():
            # if not consecutive, return false
            min_k = min(k_set)
            for i in range(len(k_set)-1):
                if min_k+i+1 not in k_set:
                    return False
        # return True if all ks are consecutive
        return True

    # *************************************************************************
    # *************************************************************************

    def qk(self, redo: bool = False):
        "Return all valid (q,k) tuples."
        if redo:
            self._qk = tuple(
                (q, k) for q, _k in self.time_intervals.items() for k in _k
            )
        return self._qk

    def qp(self, redo: bool = False):
        "Return all valid (q,p) tuples."
        if redo:
            self._qp = tuple(
                (q, p) for q, _p in self.reporting_periods.items() for p in _p
            )
        return self._qp

    def qpk(self, redo: bool = False):
        "Return all valid (q,p,k) tuples."
        if redo:
            self._qpk = tuple(
                (q, p, k)
                for q, _p in self.reporting_periods.items()
                for p in _p
                for k in self.time_intervals[q]
            )
        return self._qpk

    # *************************************************************************
    # *************************************************************************
    
    def assessments_overlap(self) -> bool:
        "Returns True if any period is covered by more than one assessment."
        # if there is only one assessment, return False
        if self.number_assessments() == 1:
            return False
        else:
            # if there is more than one assessment, check whether two or more
            # cover the same period
            qs = tuple(self.assessments)
            # for each assessment
            for q1, q2 in zip(qs, qs[1:]):
                # for each period in one assessment (q1)
                for p in self.reporting_periods[q1]:
                    # check if it is covered by the other assessment (q2)
                    if p in self.reporting_periods[q2]:
                        # p is covered by at least two assessments (q1 and q2)
                        return True
        # if no period is covered by more than one assessment, return False
        return False
    
    # *************************************************************************
    # *************************************************************************

# *****************************************************************************
# *****************************************************************************

class EconomicTimeFrame(TimeFrame):
    """A class to define investments within a given time frame."""
    
    def __init__(
            self, 
            discount_rate=None,         # real: same value for all q and p
            discount_rates_p=None,      # list: same values for all q (1 per p)
            discount_rates_q=None,      # dict: key=q, value= list for each p
            **kwargs
            ):
        
        TimeFrame.__init__(
            self, 
            **kwargs
            )
        
        if isinstance(discount_rate, Real):
            # real: same value for all q and p
            self._discount_rates = {
                q: tuple([discount_rate for p in self.reporting_periods[q]])
                for q in self.assessments
                }
        elif (type(discount_rates_p) == tuple or 
              type(discount_rates_p) == list):
            # list: same values for all q (1 per p)
            self._discount_rates = {
                q: tuple(discount_rates_p[p] 
                         for p in self.reporting_periods[q])
                for q in self.assessments
                }
        elif (type(discount_rates_q) == dict and 
              self.complete_q(discount_rates_q)):
            # dict: 1 value per p and q
            self._discount_rates = dict(discount_rates_q)
        else:
            raise TypeError('Unrecognised inputs.')
        
        # TODO: validate the discount rate object
        
    # *************************************************************************
    # *************************************************************************
        
    def discount_rate(self, assessment, reporting_period):
        "Returns the discount rate for a given assessment and period."
        return self._discount_rates[assessment][
            self.reporting_periods[assessment].index(reporting_period)
            ]
    
    # *************************************************************************
    # *************************************************************************
    
    def discount_factor(self, assessment, reporting_period):
        "Returns the discount factor for a given assessment and period."
        
        discount_rates = tuple(
            self.discount_rate(assessment, p)
            for p in self.reporting_periods[assessment]
            if p <= reporting_period
            )
        return _discount_factor(discount_rates)

# *****************************************************************************
# *****************************************************************************

def entries_are_invariant(data: dict, keys: tuple = None) -> bool:
    "Returns True if all (identified) entries in the dict are identical."
    return set(
        value for key, value in data.items()
        ) if type(keys) == type(None) else set(
        value for key, value in data.items() if key in keys
        )

# *****************************************************************************
# *****************************************************************************
