# *****************************************************************************
# *****************************************************************************

# standard

# import uuid

# local, external
import networkx as nx
import pyomo.environ as pyo
from matplotlib import pyplot as plt

# local, internal
from .problem import InfrastructurePlanningProblem
from .network import Network
from .blocks.prices import NODE_PRICE_DELTA, NODE_PRICE_LAMBDA #, NODE_PRICE_OTHER

# *****************************************************************************
# *****************************************************************************

def pre_statistics(ipp: InfrastructurePlanningProblem, 
                   node_keys = None):
    "Returns preliminary problem statistics."
    
    if type(node_keys) == type(None):
        # pick all
        node_keys = tuple(
            (g, node_key)
            for g, net in ipp.networks.items()
            for node_key in net.nodes()
            if Network.KEY_NODE_BASE_FLOW in net.nodes[node_key] 
            )
    
    # aggregate static (end use) demand
    aggregate_static_demand_qk = {
        qk: sum(
            ipp.networks[g].nodes[node_key][Network.KEY_NODE_BASE_FLOW][qk]
            for g, node_key in node_keys
            if ipp.networks[g].nodes[node_key][Network.KEY_NODE_BASE_FLOW][qk] >= 0                      
            ) 
        for qk in ipp.time_frame.qk()
        }
    # aggregate static (primary) supply
    aggregate_static_supply_qk = {
        qk: - sum(
            ipp.networks[g].nodes[node_key][Network.KEY_NODE_BASE_FLOW][qk]
            for g, node_key in node_keys
            if ipp.networks[g].nodes[node_key][Network.KEY_NODE_BASE_FLOW][qk] < 0                      
            ) 
        for qk in ipp.time_frame.qk()
        }
    # static nodal balance
    aggregate_static_balance_qk = {
        qk: aggregate_static_demand_qk[qk]-aggregate_static_supply_qk[qk]
        for qk in ipp.time_frame.qk()
        }    
    
    return (
        aggregate_static_demand_qk,
        aggregate_static_supply_qk,
        aggregate_static_balance_qk
        )

# *****************************************************************************
# *****************************************************************************

def statistics(ipp: InfrastructurePlanningProblem, 
               import_node_keys: tuple = None,
               export_node_keys: tuple = None,
               other_node_keys: tuple = None):
    "Returns flow statistics using the optimisation results."
    
    prices_in_block = ipp.use_prices_block
    
    if type(import_node_keys) == type(None):
        # pick all import nodes
        import_node_keys = tuple(
            (g, l)
            for g in ipp.networks
            for l in ipp.networks[g].import_nodes
            )
        
    if type(export_node_keys) == type(None):
        # pick all export nodes
        export_node_keys = tuple(
            (g, l)
            for g in ipp.networks
            for l in ipp.networks[g].export_nodes
            )
    
    if type(other_node_keys) == type(None):
        # pick all
        other_node_keys = tuple(
            (g, node_key)
            for g, net in ipp.networks.items()
            for node_key in net.nodes()
            if Network.KEY_NODE_BASE_FLOW in net.nodes[node_key] 
            )
    if prices_in_block:
        # imports
        if ipp.node_price_model == NODE_PRICE_DELTA or ipp.node_price_model == NODE_PRICE_LAMBDA:
            # imports
            imports_qpk = {
                qpk: pyo.value( 
                    sum(
                        ipp.instance.block_prices[(g,l_imp,*qpk)].var_trans_flows
                        for g, l_imp in import_node_keys
                        )
                    *ipp.instance.param_c_time_qpk[qpk]
                    )
                for qpk in ipp.time_frame.qpk()
                }
            # exports
            exports_qpk = {
                qpk: pyo.value(
                    sum(
                        ipp.instance.block_prices[(g,l_exp,*qpk)].var_trans_flows
                        for g, l_exp in export_node_keys
                        )
                    *ipp.instance.param_c_time_qpk[qpk]
                    )
                for qpk in ipp.time_frame.qpk()
                }
        else:
            # imports
            imports_qpk = {
                qpk: pyo.value( 
                    sum(
                        ipp.instance.block_prices[(g,l_imp,*qpk)].var_trans_flows_s[s]
                        for g, l_imp in import_node_keys
                        # for g in ipp.networks
                        # for l_imp in ipp.networks[g].import_nodes
                        for s in ipp.instance.block_prices[(g,l_imp,*qpk)].set_S
                        )
                    *ipp.instance.param_c_time_qpk[qpk]
                    )
                for qpk in ipp.time_frame.qpk()
                }
            # exports
            exports_qpk = {
                qpk: pyo.value(
                    sum(
                        ipp.instance.block_prices[(g,l_exp,*qpk)].var_trans_flows_s[s]
                        for g, l_exp in export_node_keys
                        # for g in ipp.networks
                        # for l_exp in ipp.networks[g].export_nodes
                        for s in ipp.instance.block_prices[(g,l_exp,*qpk)].set_S
                        )
                    *ipp.instance.param_c_time_qpk[qpk]
                    )
                for qpk in ipp.time_frame.qpk()
                }
    else:
        # not in a block
        if ipp.node_price_model == NODE_PRICE_DELTA or ipp.node_price_model == NODE_PRICE_LAMBDA:
            # imports
            imports_qpk = {
                qpk: pyo.value( 
                    sum(
                        ipp.instance.var_trans_flows_glqpk[(g,l_imp,*qpk)]
                        for g, l_imp in import_node_keys
                        )
                    *ipp.instance.param_c_time_qpk[qpk]
                    )
                for qpk in ipp.time_frame.qpk()
                }
            
            # exports
            exports_qpk = {
                qpk: pyo.value(
                    sum(
                        ipp.instance.var_trans_flows_glqpk[(g,l_exp,*qpk)]
                        for g, l_exp in export_node_keys
                        )
                    *ipp.instance.param_c_time_qpk[qpk]
                    )
                for qpk in ipp.time_frame.qpk()
                }
        else:
            # imports
            imports_qpk = {
                qpk: pyo.value( 
                    sum(
                        ipp.instance.var_trans_flows_glqpks[(g,l_imp,*qpk, s)]
                        for g, l_imp in import_node_keys
                        # for g in ipp.networks
                        # for l_imp in ipp.networks[g].import_nodes
                        for s in ipp.instance.set_S[(g,l_imp,*qpk)]
                        )
                    *ipp.instance.param_c_time_qpk[qpk]
                    )
                for qpk in ipp.time_frame.qpk()
                }
            
            # exports
            exports_qpk = {
                qpk: pyo.value(
                    sum(
                        ipp.instance.var_trans_flows_glqpks[(g,l_exp,*qpk, s)]
                        for g, l_exp in export_node_keys
                        # for g in ipp.networks
                        # for l_exp in ipp.networks[g].export_nodes
                        for s in ipp.instance.set_S[(g,l_exp,*qpk)]
                        )
                    *ipp.instance.param_c_time_qpk[qpk]
                    )
                for qpk in ipp.time_frame.qpk()
                }
    # balance
    balance_qpk = {
        qpk: imports_qpk[qpk]-exports_qpk[qpk]
        for qpk in ipp.time_frame.qpk()
        }
    # import costs
    import_costs_qpk = {
        qpk: pyo.value( 
            sum(
                ipp.instance.var_ifc_glqpk[(g,l_imp,*qpk)]
                for g, l_imp in import_node_keys
                # for g in ipp.networks
                # for l_imp in ipp.networks[g].import_nodes
                )
            *ipp.instance.param_c_time_qpk[qpk]
            )
        for qpk in ipp.time_frame.qpk()
        }
    # export revenue
    export_revenue_qpk = {
        qpk: pyo.value(
            sum(
                ipp.instance.var_efr_glqpk[(g,l_exp,*qpk)]
                for g, l_exp in export_node_keys
                # for g in ipp.networks
                # for l_exp in ipp.networks[g].export_nodes
                )
            *ipp.instance.param_c_time_qpk[qpk]
            )
        for qpk in ipp.time_frame.qpk()
        }
    # net cash flow
    ncf_qpk = {
        qpk: import_costs_qpk[qpk]-export_revenue_qpk[qpk]
        for qpk in ipp.time_frame.qpk()
        }
    # aggregate static (end use) demand
    aggregate_static_demand_qpk = {
        qpk: pyo.value(
            sum(
                ipp.instance.param_v_base_glqk[(g, l, qpk[0], qpk[2])]
                for g, l in other_node_keys
                # for g in ipp.networks
                # for l in ipp.networks[g].source_sink_nodes
                if ipp.instance.param_v_base_glqk[(g, l, qpk[0], qpk[2])] >= 0
                )   
            *ipp.instance.param_c_time_qpk[qpk]
            )
        for qpk in ipp.time_frame.qpk()
        }
    # aggregate static (primary) supply
    aggregate_static_supply_qpk = {
        qpk: -pyo.value(
            sum(
                ipp.instance.param_v_base_glqk[(g, l, qpk[0], qpk[2])]
                for g, l in other_node_keys
                # for g in ipp.networks
                # for l in ipp.networks[g].source_sink_nodes
                if ipp.instance.param_v_base_glqk[(g, l, qpk[0], qpk[2])] < 0
                ) 
            *ipp.instance.param_c_time_qpk[qpk]
            )
        for qpk in ipp.time_frame.qpk()
        }
    # static nodal balance
    aggregate_static_balance_qpk = {
        qpk: aggregate_static_demand_qpk[qpk]-aggregate_static_supply_qpk[qpk]
        for qpk in ipp.time_frame.qpk()
        }    
    
    return (
        imports_qpk, 
        exports_qpk, 
        balance_qpk, 
        import_costs_qpk, 
        export_revenue_qpk, 
        ncf_qpk, 
        aggregate_static_demand_qpk,
        aggregate_static_supply_qpk,
        aggregate_static_balance_qpk
        )

# *****************************************************************************
# *****************************************************************************


def unused_node_key(network: nx.MultiDiGraph):
    """Returns an unused node key."""
    # if the network has N nodes, checking every natural number up to N+1 will
    # certainly return an unused node_key (worst case: N+1th iteration)
    for i in range(network.number_of_nodes() + 1):
        # check if it exists
        if not network.has_node(i):
            # it doesn't, return it
            return i

# *****************************************************************************
# *****************************************************************************


def describe_mves(obj: pyo.ConcreteModel):
    # describe the multi-vector energy system

    print("******************************************************************")
    print("******************************************************************")

    # mves

    # the number of networks

    print("This MVES consists of " + str(len(obj.set_G)) + " network(s).")

    # nodes

    for g in obj.set_G:
        # *********************************************************************

        # the number of nodes for each network

        print("Network " + str(g) + " has " + str(len(obj.set_L[g])) + " node(s).")
        print("Network " + str(g) + "'s nodes: " + str([l for l in obj.set_L[g]]))

        # the number of import nodes for each network

        print(
            "Network "
            + str(g)
            + " has "
            + str(len(obj.set_L_imp[g]))
            + " import node(s)."
        )
        if len(obj.set_L_imp[g]) != 0:
            print(
                "Network "
                + str(g)
                + "'s import nodes: "
                + str([l for l in obj.set_L_imp[g]])
            )

        # the number of export nodes for each network

        print(
            "Network "
            + str(g)
            + " has "
            + str(len(obj.set_L_exp[g]))
            + " export node(s)."
        )
        if len(obj.set_L_exp[g]) != 0:
            print(
                "Network "
                + str(g)
                + "'s export nodes: "
                + str([l for l in obj.set_L_exp[g]])
            )

        # the number of other nodes for each network

        other_nodes = [
            node
            for node in obj.set_L[g]
            if node not in obj.set_L_exp[g]
            if node not in obj.set_L_imp[g]
        ]

        # number_other_nodes = (
        #     len(obj.set_L[g])-len(obj.set_L_exp[g])-len(obj.set_L_imp[g])
        #     )

        number_other_nodes = len(other_nodes)

        print("Network " + str(g) + " has " + str(number_other_nodes) + " other nodes.")
        if number_other_nodes != 0:
            print("Network " + str(g) + "'s other nodes: " + str(other_nodes))

        # the static flow needs for other nodes

        static_flow_limits = [
            (
                min(
                    [
                        pyo.value(obj.param_v_base_glqk[(g, node, q, k)])
                        for q in obj.set_Q
                        for k in obj.set_K_q[q]
                    ]
                ),
                max(
                    [
                        pyo.value(obj.param_v_base_glqk[(g, node, q, k)])
                        for q in obj.set_Q
                        for k in obj.set_K_q[q]
                    ]
                ),
            )
            for node in other_nodes
        ]

        print("Min. and max. flow needs for other nodes: " + str(static_flow_limits))

        # *********************************************************************

        # arcs

        # the number of arcs under consideration

        potential_directed_arcs = [
            (l1, l2)
            for (gx, l1, l2) in obj.set_GLL
            if gx == g
            if (gx, l1, l2) in obj.set_J
            if len(obj.set_J[(gx, l1, l2)]) != 0
        ]

        print(
            "Network "
            + str(g)
            + " considers "
            + str(len(potential_directed_arcs))
            + " potential directed arcs."
        )
        if number_other_nodes != 0:
            print(
                "Network "
                + str(g)
                + "'s potential directed arcs: "
                + str(potential_directed_arcs)
            )

        # unreacheable nodes

        # *********************************************************************

    # *************************************************************************

    # systems

    # the number of systems

    number_systems = len(obj.set_I)

    number_optional_systems = len(obj.set_I_new)

    if number_systems == 0:
        print("This MVES has no systems under consideration.")

    else:
        if number_systems == 1:
            print("This MVES considers a single system.")

            print("System: " + str(obj.set_I))

        else:
            print("This MVES considers " + str(number_systems) + " systems.")

            print("Systems: " + str(obj.set_I))

        if number_optional_systems == 1:
            if number_systems == 1:
                print("The only system under consideration is also optional.")

            else:
                print("One of the aforementioned systems is optional.")

            print("Optional system: " + str(obj.set_I_new))

        else:
            print(
                "Among these systems, "
                + str(number_optional_systems)
                + " are optional."
            )

            print("Optional systems: " + str(obj.set_I_new))

        # how many have dimensionable state amplitudes

        # how many have dimensionable upper (positive) state limits

        # how many have dimensionable lower (negative) state limits

        # how many have dimensionable input signals

    print("******************************************************************")
    print("******************************************************************")

# *****************************************************************************
# *****************************************************************************

# data structure

# *****************************************************************************
# *****************************************************************************

def plot_networks(
    ipp: InfrastructurePlanningProblem,
    ax=None,
    show=False,
    filepath: str = None,
    filename_radical: str = None,
):
    # TODO: include hints about operational performance in the plot

    # TODO: find a way to plot parallel arcs without overlap

    # NOTE: the code below works (could be used to submit a PR)

    # G=nx.MultiGraph ([(1,2),(1,2),(1,2),(3,1),(3,2)])
    # pos = nx.random_layout(G)
    # nx.draw_networkx_nodes(G, pos, node_color = 'r', node_size = 100, alpha = 1)
    # ax = plt.gca()
    # for e in G.edges:
    #     ax.annotate("",
    #                 xy=pos[e[0]], xycoords='data',
    #                 xytext=pos[e[1]], textcoords='data',
    #                 arrowprops=dict(arrowstyle="->", color="0.5",
    #                                 shrinkA=5, shrinkB=5,
    #                                 patchA=None, patchB=None,
    #                                 connectionstyle="arc3,rad=rrr".replace('rrr',str(0.3*e[2])
    #                                 ),
    #                                 ),
    #                 )
    # plt.axis('off')
    # plt.show()

    # G=nx.MultiDiGraph([(1,2,0),(1,2,1),(1,2,2),(3,1,0),(3,2,0)])
    # pos = nx.random_layout(G)
    # nx.draw_networkx_nodes(G, pos, node_color = 'r', node_size = 100, alpha = 1)
    # ax = plt.gca()
    # for e in G.edges:
    #     ax.annotate("",
    #                 xy=pos[e[0]], xycoords='data',
    #                 xytext=pos[e[1]], textcoords='data',
    #                 arrowprops=dict(arrowstyle="->", color="0.5",
    #                                 shrinkA=5, shrinkB=5,
    #                                 patchA=None, patchB=None,
    #                                 connectionstyle="arc3,rad=rrr".replace('rrr',str(0.3*e[2])
    #                                 ),
    #                                 ),
    #                 )
    # plt.axis('off')
    # plt.show()

    # *************************************************************************
    # *************************************************************************

    plt.rcParams["figure.figsize"] = [7.50, 3.50]
    plt.rcParams["figure.autolayout"] = True

    # *************************************************************************
    # *************************************************************************

    node_shapes = ["o", "s", "d"]

    seed_number = 381

    node_size_offset = 300

    node_size_gain = 100

    arc_width_gain = 5

    # for each network

    for grid_key, grid in ipp.networks.items():
        # create a figure

        fig, ax = plt.subplots()

        # create a plot showing which nodes are connected by which arcs

        # get the position of each node

        # pos = {
        #     node_key:grid.nodes[node_key][Network.STR_DATA_NODE].position
        #     for node_key in grid.nodes()}

        pos = nx.spring_layout(grid, seed=seed_number)

        # peak static balance

        peak_base_flow = {
            node_key: sum(
                grid.nodes[node_key][Network.KEY_NODE_BASE_FLOW][sb_key]
                for sb_key, sb in grid.nodes[node_key][
                    Network.KEY_NODE_BASE_FLOW
                ].items()
            )
            if node_key in grid.source_sink_nodes
            else 0
            for node_key in grid.nodes()
        }

        # max supply/demand rating

        max_node_demand = max([peak_base_flow[node_key] for node_key in grid.nodes()])

        # max_node_demand = max([
        #     sum([abs(sb_glk)
        #          for sb_glk in grid.nodes[node_key][
        #                  Network.KEY_NODE_BASE_FLOW]])
        #     for node_key in grid.nodes()
        #     ])

        # select the node sizes

        node_sizes = {
            node_key: node_size_offset
            + node_size_gain  # offset
            * peak_base_flow[node_key]  # gain
            / max_node_demand
            for node_key in grid.nodes()
        }

        # select the relevant edges

        selected_edges = [
            arc_key  # grid.edges[arc_key]
            for arc_key in grid.edges(keys=True)
            if True in grid.edges[arc_key][Network.KEY_ARC_TECH].options_selected
        ]

        # select the edge width based on the arc capacity

        edge_widths = [
            arc_width_gain
            * grid.edges[arc_key][Network.KEY_ARC_TECH].capacity[option_index]
            / max(grid.edges[arc_key][Network.KEY_ARC_TECH].capacity)
            for arc_key in grid.edges(keys=True)
            for option_index, arc_tech_option in enumerate(
                grid.edges[arc_key][Network.KEY_ARC_TECH].options_selected
            )
            if arc_tech_option
        ]

        if len(selected_edges) != len(edge_widths):
            raise Exception

            pass

        M = len(selected_edges)
        edge_colors = range(2, M + 2)
        edge_alphas = [(5 + i) / (M + 4) for i in range(M)]
        cmap = plt.cm.plasma

        # *********************************************************************

        # plot import nodes

        if len(grid.import_nodes) != 0:
            nodes = nx.draw_networkx_nodes(
                grid,
                pos,
                nodelist=grid.import_nodes,
                node_shape="s",
                node_size=[
                    node_size
                    for node_key, node_size in node_sizes.items()
                    if node_key in grid.import_nodes
                ],
                node_color="indigo",
                label="Import node",
            )

        # plot export nodes

        if len(grid.export_nodes) != 0:
            nodes = nx.draw_networkx_nodes(
                grid,
                pos,
                nodelist=grid.export_nodes,
                node_shape="d",
                node_size=[
                    node_size
                    for node_key, node_size in node_sizes.items()
                    if node_key in grid.export_nodes
                ],
                node_color="orange",
                label="Export node",
            )

        # plot other/normal nodes

        other_nodes = list(grid)

        for node in grid.import_nodes:
            other_nodes.remove(node)

        for node in grid.export_nodes:
            other_nodes.remove(node)

        nodes = nx.draw_networkx_nodes(
            grid,
            pos,
            nodelist=other_nodes,
            node_shape="o",
            node_size=[
                node_size
                for node_key, node_size in node_sizes.items()
                if node_key in other_nodes
            ],
            node_color="thistle",
            label="Normal node",
        )

        # multi-edges

        ax = plt.gca()
        for e in selected_edges:
            ax.annotate(
                "",
                xy=pos[e[1]],
                xycoords="data",
                xytext=pos[e[0]],
                textcoords="data",
                arrowprops=dict(
                    arrowstyle="->",
                    color="0.5",
                    shrinkA=5,
                    shrinkB=5,
                    patchA=None,
                    patchB=None,
                    connectionstyle="arc3,rad=rrr".replace("rrr", str(0.3 * e[2])),
                ),
            )

        # edges = nx.draw_networkx_edges(
        #     grid,
        #     pos,
        #     edgelist=selected_edges,
        #     node_size=[
        #         node_size
        #         for node_key, node_size in node_sizes.items()], # used to determine edge positioning
        #     arrowstyle="->",
        #     arrowsize=10,
        #     edge_color=edge_colors,
        #     edge_cmap=cmap,
        #     width=edge_widths,
        #     #connectionstyle='arc3, rad = 0.1'
        # )

        labels = nx.draw_networkx_labels(grid, pos)

        # # set alpha value for each edge
        # for i in range(M):
        #     edges[i].set_alpha(edge_alphas[i])

        # pc = mpl.collections.PatchCollection(edges, cmap=cmap)
        # pc.set_array(edge_colors)
        # plt.colorbar(pc,ax=ax)

        ax = plt.gca()
        ax.set_axis_off()

        plt.legend()

        if filepath is not None and filename_radical is not None:
            # save file

            plt.savefig(filepath + filename_radical + str(grid_key))

        if show:
            plt.show()

    # G = nx.balanced_tree(3,2,create_using=nx.MultiDiGraph)
    # G.add_edge(1,3)
    # G.add_edge(2,4)
    # G.add_edge(5,1)
    # G.add_edge(1,4,data='savage')
    # G.add_edge(1,4,data='traszq')
    # fig, ax = plt.subplots(figsize=(25,25))
    # pos = nx.spring_layout(G, seed=225)  # Seed for reproducible layout
    # nx.draw(G, pos)
    # plt.show()

    # *************************************************************************
    # *************************************************************************


# *****************************************************************************
# *****************************************************************************

def is_integer(variable: float, integrality_tolerance: float) -> bool:
    """Returns True if a given number qualifies as an integer."""
    if integrality_tolerance >= 0.5:
        raise ValueError(
            "A tolerance greater than or equal to 0.5 "
            + "renders the concept of integrality useless."
        )
    return not (abs(round(variable) - variable) > integrality_tolerance)


# *****************************************************************************
# *****************************************************************************