# *****************************************************************************
# *****************************************************************************

# key features:
# 1) same name and units for the same attributes across different solvers
# 2) same input mode for different solvers
# 3) extract results from different solvers

# example: gurobi and cplex require the warmstart parameter, others don't
# example: some solvers allow for absolute mip gap, others don't
# example: scip needs the executable to be specified, others don't
# example: the results object is configured differently based on the solver

# standard libraries

# from dataclasses import dataclass, field, InitVar

# local

import pyomo.environ as pyo

from pyomo.opt.base.solvers import SolverFactory

from pyomo.opt.results.solver import TerminationCondition, SolverStatus

from pyomo.opt.results.solver import check_optimal_termination

# TODO: get the options for each solver from files

# TODO: provide a framework to handle different units (e.g. time)

# TODO: create a way to identify the solver status independent of each solver


class SolverInterface(object):
    "A class for interfacing with solvers through Pyomo."

    # *************************************************************************
    # *************************************************************************

    SOLVER_GLPK = "glpk"

    SOLVER_CBC = "cbc"

    SOLVER_GUROBI = "gurobi"

    SOLVER_SCIP = "scip"

    SOLVER_CPLEX = "cplex"

    SOLVER_CPLEX_DIRECT = "cplex_direct"

    SOLVERS = [
        SOLVER_CBC,
        SOLVER_GLPK,
        SOLVER_SCIP,
        SOLVER_GUROBI,
        SOLVER_CPLEX,
        SOLVER_CPLEX_DIRECT,
    ]

    # *************************************************************************
    # *************************************************************************

    # topupopt options

    STR_DET_TIME_LIMIT = "det_time_limit"
    STR_TIME_LIMIT = "time_limit"
    STR_ABS_MIP_GAP = "absolute_mip_gap"
    STR_REL_MIP_GAP = "relative_mip_gap"

    # option keywords for specific solvers

    OPTIONS_GLPK = {STR_TIME_LIMIT: "tmlim", STR_REL_MIP_GAP: "mipgap"}

    OPTIONS_CBC = {STR_TIME_LIMIT: "seconds", STR_REL_MIP_GAP: "ratioGap"}

    OPTIONS_SCIP = {
        STR_TIME_LIMIT: "limits/time",
        STR_ABS_MIP_GAP: "limits/absgap",
        STR_REL_MIP_GAP: "limits/gap",
    }

    # source: https://www.gurobi.com/documentation/9.5/refman/parameters.html

    OPTIONS_GUROBI = {
        STR_TIME_LIMIT: "TimeLimit",
        STR_ABS_MIP_GAP: "MIPGapAbs",
        STR_REL_MIP_GAP: "MIPGap",
    }

    # souce: https://www.ibm.com/docs/en/icos/12.8.0.0?topic=cplex-parameters

    OPTIONS_CPLEX = {
        STR_DET_TIME_LIMIT: "dettimelimit",
        STR_TIME_LIMIT: "timelimit",
        # STR_ABS_MIP_GAP: 'absmipgap',
        STR_ABS_MIP_GAP: "mip_tolerances_absmipgap",
        STR_REL_MIP_GAP: "mipgap",
    }

    # static dict with solver options for each supported solver

    OPTIONS_SOLVER = {
        SOLVER_GLPK: OPTIONS_GLPK,
        SOLVER_CBC: OPTIONS_CBC,
        SOLVER_SCIP: OPTIONS_SCIP,
        SOLVER_GUROBI: OPTIONS_GUROBI,
        SOLVER_CPLEX: OPTIONS_CPLEX,
        SOLVER_CPLEX_DIRECT: OPTIONS_CPLEX,
    }

    PYOMO_OPTIONS = ["executable", "warmstart", "solver_io", "tee"]

    # *************************************************************************
    # *************************************************************************

    PROBLEM_LP = "lp"
    PROBLEM_MILP = "milp"
    PROBLEM_QP = "qp"
    PROBLEM_MIQP = "miqp"
    PROBLEM_QCP = "qcp"
    PROBLEM_MIQCP = "miqcp"

    PROBLEM_TYPES = [
        PROBLEM_LP,
        PROBLEM_MILP,
        PROBLEM_QP,
        PROBLEM_MIQP,
        PROBLEM_QCP,
        PROBLEM_MIQCP,
    ]

    # https://scipopt.org/doc/html/WHATPROBLEMS.php

    PROBLEMS_COMPATIBLE_SCIP = [PROBLEM_LP, PROBLEM_MILP]

    PROBLEMS_COMPATIBLE_GLPK = [PROBLEM_LP, PROBLEM_MILP]

    PROBLEMS_COMPATIBLE_CBC = [PROBLEM_LP, PROBLEM_MILP]

    # https://www.gurobi.com/products/gurobi-optimizer/

    PROBLEMS_COMPATIBLE_GUROBI = [
        PROBLEM_LP,
        PROBLEM_MILP,
        PROBLEM_QP,
        PROBLEM_MIQP,
        PROBLEM_QCP,
        PROBLEM_MIQCP,
    ]

    PROBLEMS_COMPATIBLE_CPLEX = [
        PROBLEM_LP,
        PROBLEM_MILP,
        PROBLEM_QP,
        PROBLEM_MIQP,
        PROBLEM_QCP,
        PROBLEM_MIQCP,
    ]

    PROBLEMS_COMPATIBLE_SOLVER = {
        SOLVER_GLPK: PROBLEMS_COMPATIBLE_GLPK,
        SOLVER_SCIP: PROBLEMS_COMPATIBLE_SCIP,
        SOLVER_CBC: PROBLEMS_COMPATIBLE_CBC,
        SOLVER_GUROBI: PROBLEMS_COMPATIBLE_GUROBI,
        SOLVER_CPLEX: PROBLEMS_COMPATIBLE_CPLEX,
        SOLVER_CPLEX_DIRECT: PROBLEMS_COMPATIBLE_CPLEX,
    }

    # *************************************************************************
    # *************************************************************************

    # USER -->> SOLVER_INTERFACE -->> PYOMO --> SOLVER

    # the optimisation status depends on the solver status and the term. criteria

    SOLVER_STATUSES = [
        SolverStatus.aborted,
        SolverStatus.error,
        SolverStatus.ok,
        SolverStatus.unknown,
        SolverStatus.warning,
    ]

    TERMINATION_CRITERIA = [
        # UNKNOWN
        TerminationCondition.unknown,
        # OK
        TerminationCondition.maxTimeLimit,
        TerminationCondition.maxIterations,
        TerminationCondition.minFunctionValue,
        TerminationCondition.minStepLength,
        TerminationCondition.globallyOptimal,
        TerminationCondition.locallyOptimal,
        TerminationCondition.feasible,
        TerminationCondition.optimal,
        TerminationCondition.maxEvaluations,
        TerminationCondition.other,
        # WARNING
        TerminationCondition.unbounded,
        TerminationCondition.infeasible,
        TerminationCondition.infeasibleOrUnbounded,
        TerminationCondition.invalidProblem,
        TerminationCondition.intermediateNonInteger,
        TerminationCondition.noSolution,
        # ERROR
        TerminationCondition.solverFailure,
        TerminationCondition.internalSolverError,
        TerminationCondition.error,
        # ABORTED
        TerminationCondition.userInterrupt,
        TerminationCondition.resourceInterrupt,
        TerminationCondition.licensingProblems,
    ]

    # *************************************************************************
    # *************************************************************************

    def __init__(
        self,
        solver_name: str,
        time_limit: int = None,
        relative_mip_gap: float = None,
        absolute_mip_gap: float = None,
        **kwargs
    ):
        # check if the solver is supported

        if solver_name not in self.SOLVERS:
            raise self.UnknownSolverError(solver_name)

        # init

        self.solver_name = solver_name

        self.time_limit = time_limit

        self.relative_mip_gap = relative_mip_gap

        self.absolute_mip_gap = absolute_mip_gap

        # create SolverFactory

        self.obj = SolverFactory(solver_name, **kwargs)

    # *************************************************************************
    # *************************************************************************

    def get_solver_handler(self, **kwargs):
        # find

        options_dict_format = {
            self.OPTIONS_SOLVER[self.solver_name][key]: value
            for key, value in kwargs.items()
            if key in self.OPTIONS_SOLVER[self.solver_name]
        }

        # parameters

        options_param_format = {
            key: value for key, value in kwargs.items() if key in self.PYOMO_OPTIONS
        }

        if len(options_param_format) == 0:
            return self.get_pyomo_solver_object(
                self.solver_name, options_dict_format=options_dict_format
            )

        else:
            return self.get_pyomo_solver_object(
                self.solver_name,
                options_dict_format=options_dict_format,
                options_param_format=options_param_format,
            )

    # *************************************************************************
    # *************************************************************************

    @staticmethod
    def get_pyomo_solver_object(
        solver_name: str, options_dict_format: dict, options_param_format: dict = None
    ):
        # return an SolverFactory object properly configured as per the solver

        if options_param_format is None:
            opt = pyo.SolverFactory(solver_name)

        else:
            opt = pyo.SolverFactory(solver_name, **options_param_format)

        for key, value in options_dict_format.items():
            opt.options[key] = value

        # return the object

        return opt

    # *************************************************************************
    # *************************************************************************

    @classmethod
    def problem_and_solver_are_compatible(cls, solver_name, problem_type) -> bool:
        if solver_name not in cls.SOLVERS:
            raise cls.UnknownSolverError(solver_name)

        if problem_type not in cls.PROBLEM_TYPES:
            raise cls.UnknownProblemTypeError(problem_type)

        if problem_type in cls.PROBLEMS_COMPATIBLE_SOLVER[solver_name]:
            return True

        else:
            return False

    # *************************************************************************
    # *************************************************************************

    def was_optimisation_sucessful(self, results, problem_type) -> bool:
        solver_status = results.solver.status

        termination_condition = results.solver.termination_condition

        # checks are made within problem_and_solver_are_compatible

        if not self.problem_and_solver_are_compatible(self.solver_name, problem_type):
            # if the problem and solver are not compatible, something is up

            raise self.InconsistentProblemTypeAndSolverError(
                problem_type, self.solver_name
            )

        # check the termination condition and the solver status

        if termination_condition not in self.TERMINATION_CRITERIA:
            raise self.UnknownTerminationConditionError(termination_condition)

        if solver_status not in self.SOLVER_STATUSES:
            raise self.UnknownSolverStatusError(solver_status)

        # *********************************************************************
        # *********************************************************************

        expected_solver_status = TerminationCondition.to_solver_status(
            termination_condition
        )

        if expected_solver_status != solver_status:
            raise self.InconsistentSolverStatusError(
                solver_status, expected_solver_status
            )

        if check_optimal_termination(results):
            return True

        # while not optimal, check if the solution was feasible

        if expected_solver_status == SolverStatus.ok:
            return True

        else:
            return False

    # *************************************************************************
    # *************************************************************************

    class InconsistentProblemTypeAndSolverError(Exception):
        def __init__(self, problem_type, solver_name):
            super().__init__(
                "The problem type ("
                + problem_type
                + ") cannot be handled by solver ("
                + solver_name
                + ")."
            )

    # *************************************************************************
    # *************************************************************************

    class InconsistentSolverStatusError(Exception):
        def __init__(self, solver_status, expected_solver_status):
            super().__init__(
                "The expected SolverStatus ("
                + expected_solver_status
                + ") does not match the one obtained ("
                + solver_status
                + ")."
            )

    # *************************************************************************
    # *************************************************************************

    class UnknownSolverError(Exception):
        def __init__(self, solver_name):
            super().__init__("Unknown solver: " + str(solver_name))

    # *************************************************************************
    # *************************************************************************

    class UnknownProblemTypeError(Exception):
        def __init__(self, problem_type):
            super().__init__("Unknown problem type: " + str(problem_type))

    # *************************************************************************
    # *************************************************************************

    class UnknownTerminationConditionError(Exception):
        def __init__(self, termination_condition):
            super().__init__("Unknown termination condition: " + termination_condition)

    # *************************************************************************
    # *************************************************************************

    class UnknownSolverStatusError(Exception):
        def __init__(self, solver_status):
            super().__init__("Unknown solver status: " + solver_status)


# *****************************************************************************
# *****************************************************************************
