# imports

# standard
import random

# local, external
import numpy as np
from scipy.signal import StateSpace, lsim, dlsim

# local, internal
import src.topupopt.problems.esipp.dynsys as dynsys

# *****************************************************************************
# *****************************************************************************


class TestDynsys:

    # *************************************************************************
    # *************************************************************************

    def test_incorrect_time_step_durations(self):
        integrate_outputs = True

        # *********************************************************************

        # negative time step duration

        # regular time steps (as an int)

        time_step_durations = [-1]

        # get the model

        Aw, min_rel_heat = get_stateless_model_data()

        _, _, _, D = stateless_model(Aw, min_rel_heat)

        number_errors = 0
        try:
            _ = dynsys.StatelessSystem(
                time_interval_durations=time_step_durations[0],
                D=D,
                integrate_outputs=integrate_outputs,
            )
        except ValueError:
            number_errors += 1

        assert number_errors == 1

        # *********************************************************************

        # no time step duration (empty list)

        time_step_durations = []

        # get the model

        Aw, min_rel_heat = get_stateless_model_data()

        _, _, _, D = stateless_model(Aw, min_rel_heat)

        number_errors = 0
        try:
            _ = dynsys.StatelessSystem(
                time_interval_durations=time_step_durations,
                D=D,
                integrate_outputs=integrate_outputs,
            )
        except ValueError:
            number_errors += 1

        assert number_errors == 1

        # *********************************************************************

        # time step duration list with negative or zero time step durations

        time_step_durations = [-1, 3, 0]

        # get the model

        Aw, min_rel_heat = get_stateless_model_data()

        _, _, _, D = stateless_model(Aw, min_rel_heat)

        number_errors = 0
        try:
            _ = dynsys.StatelessSystem(
                time_interval_durations=time_step_durations,
                D=D,
                integrate_outputs=integrate_outputs,
            )
        except ValueError:
            number_errors += 1

        assert number_errors == 1

        # *********************************************************************

        # time step duration list with non-numeric types

        time_step_durations = [None, 3, 3]

        # get the model

        Aw, min_rel_heat = get_stateless_model_data()

        _, _, _, D = stateless_model(Aw, min_rel_heat)

        number_errors = 0
        try:
            _ = dynsys.StatelessSystem(
                time_interval_durations=time_step_durations,
                D=D,
                integrate_outputs=integrate_outputs,
            )
        except TypeError:
            number_errors += 1

        assert number_errors == 1

    # *************************************************************************
    # *************************************************************************

    def test_single_time_step_model_incorrect_inputs(self):
        integrate_y = True

        # regular time steps (as an int)

        time_step_durations = [1]

        # define a sequence of time steps

        t = np.array(
            [sum(time_step_durations[0:i]) for i in range(len(time_step_durations) + 1)]
        )

        # define the inputs

        list_inputs = [-5, 50, 0.1, 1, 5]  # extra input signal to force error

        U = np.array(
            [
                [input_i for dt in t]  # extra time step to force special case
                for input_i in list_inputs
            ]
        )

        # U = np.array([[input_i for _ in range(len(time_step_durations))]
        #               for input_i in list_inputs])

        # get the model

        Ci, Ch, Ria, Rih, Aw, min_rel_heat, x0 = get_multi_ode_model_data()

        A, B, C, D = two_node_model(Ci, Ch, Ria, Rih, Aw, min_rel_heat)

        ds = dynsys.DynamicSystem(
            time_interval_durations=time_step_durations,
            A=A,
            B=B,
            C=C,
            D=D,
            integrate_outputs=integrate_y,
        )

        # define the initial conditions

        number_errors = 0

        try:
            X, Y = ds.simulate(U, x0)
        except ValueError:
            number_errors += 1

        assert number_errors == 1

    # *************************************************************************
    # *************************************************************************

    def test_outputless_system_object(self):
        # regular time steps
        time_step_durations = [1, 1, 1, 1]

        # define a sequence of time steps
        t = np.array(
            [sum(time_step_durations[0:i]) for i in range(len(time_step_durations) + 1)]
        )

        # define the inputs
        list_inputs = [-5, 50, 0.1, 1]
        U = np.array(
            [
                [input_i for _ in range(len(time_step_durations))]
                for input_i in list_inputs
            ]
        )

        # get the model
        Ci, Ch, Ria, Rih, Aw, min_rel_heat, x0 = get_multi_ode_model_data()
        A, B, _, _ = two_node_model(Ci, Ch, Ria, Rih, Aw, min_rel_heat)
        ds = dynsys.OutputlessSystem(
            time_interval_durations=time_step_durations, A=A, B=B
        )

        # define the initial conditions
        X, Y = ds.simulate(U, x0)
        assert Y == None
        assert isinstance(X, np.ndarray)

        # *********************************************************************

        # regular time steps (as an int)
        time_step_durations = [1]

        # define a sequence of time steps
        t = np.array(
            [sum(time_step_durations[0:i]) for i in range(len(time_step_durations) + 1)]
        )

        # define the inputs
        list_inputs = [-5, 50, 0.1, 1]
        U = np.array(
            [
                [input_i for _ in range(len(time_step_durations))]
                for input_i in list_inputs
            ]
        )

        # get the model
        Ci, Ch, Ria, Rih, Aw, min_rel_heat, x0 = get_multi_ode_model_data()
        A, B, _, _ = two_node_model(Ci, Ch, Ria, Rih, Aw, min_rel_heat)

        ds = dynsys.OutputlessSystem(
            time_interval_durations=time_step_durations, A=A, B=B
        )

        # define the initial conditions
        X, Y = ds.simulate(U, x0)
        assert Y == None
        assert isinstance(X, np.ndarray)

        # *********************************************************************

        # irregular time steps
        time_step_durations = [1, 1.5, 0.5, 1]

        # define a sequence of time steps
        t = np.array(
            [sum(time_step_durations[0:i]) for i in range(len(time_step_durations) + 1)]
        )

        # define the inputs
        list_inputs = [-5, 50, 0.1, 1]
        U = np.array(
            [
                [input_i for _ in range(len(time_step_durations))]
                for input_i in list_inputs
            ]
        )

        # get the model
        Ci, Ch, Ria, Rih, Aw, min_rel_heat, x0 = get_multi_ode_model_data()
        A, B, _, _ = two_node_model(Ci, Ch, Ria, Rih, Aw, min_rel_heat)
        ds = dynsys.OutputlessSystem(
            time_interval_durations=time_step_durations, A=A, B=B
        )

        # define the initial conditions
        X, Y = ds.simulate(U, x0)
        assert Y == None
        assert isinstance(X, np.ndarray)

    # *************************************************************************
    # *************************************************************************

    def test_stateless_system_object_integration(self):
        integrate_outputs = True
        method_stateless_system_object(integrate_outputs)

    # *************************************************************************
    # *************************************************************************

    def test_stateless_system_object_no_integration(self):
        integrate_outputs = False
        method_stateless_system_object(integrate_outputs)

    # *************************************************************************
    # *************************************************************************

    def test_incompatible_BD_matrices(self):
        integrate_y = True

        # B and D must have the same number of columns

        time_step_durations = [1, 1, 1]

        A = np.array([[4, 9], [1, 3]])

        B = np.array([[5, 6, 1], [7, 2, 1]])

        C = np.array([[4, 6]])

        D = np.array([[3, 6]])

        # test

        number_errors = 0

        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=A,
                B=B,
                C=C,
                D=D,
                integrate_outputs=integrate_y,
            )
        except ValueError:
            number_errors += 1

        assert number_errors == 1

    # *************************************************************************
    # *************************************************************************

    def test_incompatible_AC_matrices(self):
        integrate_y = True

        # A and C must have the same number of columns

        time_step_durations = [1, 1, 1]

        A = np.array([[4, 9], [1, 3]])

        B = np.array([[5, 6, 1], [7, 2, 1]])

        C = np.array([[4]])

        D = np.array([[3, 6, 3]])

        # test

        number_errors = 0

        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=A,
                B=B,
                C=C,
                D=D,
                integrate_outputs=integrate_y,
            )
        except ValueError:
            number_errors += 1

        assert number_errors == 1

    # *****************************************************************************
    # *****************************************************************************

    def test_incompatible_CD_matrices(self):
        integrate_y = True

        # C and D must have the same number of rows

        time_step_durations = [1, 1, 8]

        A = np.array([[4, 9], [1, 3]])

        B = np.array([[5, 6, 1], [7, 2, 1]])

        C = np.array([[4, 9], [1, 3]])

        D = np.array([[3, 6, 3]])

        # test

        number_errors = 0

        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=A,
                B=B,
                C=C,
                D=D,
                integrate_outputs=integrate_y,
            )
        except ValueError:
            number_errors += 1

        assert number_errors == 1

    # *************************************************************************
    # *************************************************************************

    def test_incompatible_AB_matrices(self):
        integrate_y = True

        # A and B must have the same number of rows

        time_step_durations = [1, 1, 8]

        A = np.array([[4, 9], [1, 3]])

        B = np.array([[3, 6, 3]])

        C = np.array([[2, 4]])

        D = np.array([[5, 6, 1]])

        # test A

        number_errors = 0

        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=A,
                B=B,
                C=C,
                D=D,
                integrate_outputs=integrate_y,
            )
        except ValueError:
            number_errors += 1

        assert number_errors == 1

    # *************************************************************************
    # *************************************************************************

    def test_nonsquare_A_matrices(self):
        integrate_y = True

        # Single non-square A matrix

        time_step_durations = [1, 1, 8]

        A = np.array([[4, 9]])

        B = np.array([[3]])

        C = np.array([[2]])

        D = np.array([[5]])

        # test A

        number_errors = 0

        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=A,
                B=B,
                C=C,
                D=D,
                integrate_outputs=integrate_y,
            )
        except ValueError:
            number_errors += 1

        assert number_errors == 1

        # # multiple matrices, at least one of which is non-square

        # A = [np.array([[4]]), np.array([[1, 9]])]

        # B = [np.array([[3]]), np.array([[4]])]

        # C = [np.array([[2]]), np.array([[2]])]

        # D = [np.array([[8]]), np.array([[7]])]

        # # test A

        # number_errors = 0

        # try:
        #     _ = dynsys.DynamicSystem(
        #         time_interval_durations=time_step_durations,
        #         A=A,
        #         B=B,
        #         C=C,
        #         D=D,
        #         integrate_outputs=integrate_y)
        # except ValueError:
        #     number_errors += 1

        # assert number_errors == 1

    # *************************************************************************
    # *************************************************************************

    def test_nonmatching_time_steps_and_matrices(self):
        integrate_y = True

        # multiple matrices and time intervals, but more time steps than matrices

        time_step_durations = [1, 1, 8]

        A = [np.array([[4]]), np.array([[1]])]

        B = [np.array([[3]]), np.array([[4]]), np.array([[4]])]

        C = [np.array([[2]]), np.array([[2]]), np.array([[2]])]

        D = [np.array([[8]]), np.array([[7]]), np.array([[7]])]

        # test A

        number_errors = 0

        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=A,
                B=B,
                C=C,
                D=D,
                integrate_outputs=integrate_y,
            )
        except ValueError:
            number_errors += 1

        assert number_errors == 1

        # test B

        number_errors = 0

        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=B,
                B=A,
                C=C,
                D=D,
                integrate_outputs=integrate_y,
            )
        except ValueError:
            number_errors += 1

        assert number_errors == 1

        # test C

        number_errors = 0

        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=C,
                B=B,
                C=A,
                D=D,
                integrate_outputs=integrate_y,
            )
        except ValueError:
            number_errors += 1

        assert number_errors == 1

        # test D

        number_errors = 0

        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=D,
                B=B,
                C=C,
                D=A,
                integrate_outputs=integrate_y,
            )
        except ValueError:
            number_errors += 1

        assert number_errors == 1

        # multiple matrices and time intervals, but more matrices than time steps

    # *************************************************************************
    # *************************************************************************

    def test_unrecognised_matrix_formats(self):
        integrate_outputs = False

        # single matrix: matrix as lists of lists

        time_step_durations = 8

        A = 3

        B = np.array([[3]])

        C = np.array([[2]])

        D = np.array([[8]])

        # test A

        number_errors = 0

        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=A,
                B=B,
                C=C,
                D=D,
                integrate_outputs=integrate_outputs,
            )
        except TypeError:
            number_errors += 1

        assert number_errors == 1

        # test B

        number_errors = 0

        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=B,
                B=A,
                C=C,
                D=D,
                integrate_outputs=integrate_outputs,
            )
        except TypeError:
            number_errors += 1

        assert number_errors == 1

        # test C

        number_errors = 0

        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=C,
                B=B,
                C=A,
                D=D,
                integrate_outputs=integrate_outputs,
            )
        except TypeError:
            number_errors += 1

        assert number_errors == 1

        # test D

        number_errors = 0

        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=D,
                B=B,
                C=C,
                D=A,
                integrate_outputs=integrate_outputs,
            )
        except TypeError:
            number_errors += 1

        assert number_errors == 1

        # *********************************************************************

        # multiple matrices: matrix as lists of lists

        time_step_durations = [1, 1]

        A = [[[3]], [[3]]]

        B = [np.array([[3]]), np.array([[4]])]

        C = [np.array([[2]]), np.array([[2]])]

        D = [np.array([[8]]), np.array([[7]])]

        # test A

        number_errors = 0

        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=A,
                B=B,
                C=C,
                D=D,
                integrate_outputs=integrate_outputs,
            )
        except TypeError:
            number_errors += 1

        assert number_errors == 1

        # test B

        number_errors = 0

        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=B,
                B=A,
                C=C,
                D=D,
                integrate_outputs=integrate_outputs,
            )
        except TypeError:
            number_errors += 1

        assert number_errors == 1

        # test C

        number_errors = 0

        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=C,
                B=B,
                C=A,
                D=D,
                integrate_outputs=integrate_outputs,
            )
        except TypeError:
            number_errors += 1

        assert number_errors == 1

        # test D

        number_errors = 0

        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=D,
                B=B,
                C=C,
                D=A,
                integrate_outputs=integrate_outputs,
            )
        except TypeError:
            number_errors += 1

        assert number_errors == 1

        # *********************************************************************
        # *********************************************************************

    # *************************************************************************
    # *************************************************************************

    def test_varying_matrix_sizes_no_integration(self):
        integrate_outputs = False

        time_step_durations = [1, 1]

        A1 = np.array([[1]])
        A2 = np.array([[1, 3], [4, 7]])

        A = [A1, A2]
        B = [np.array([[3]]), np.array([[4]])]
        C = [np.array([[2]]), np.array([[2]])]
        D = [np.array([[8]]), np.array([[7]])]

        # test A
        number_errors = 0
        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=A,
                B=B,
                C=C,
                D=D,
                integrate_outputs=integrate_outputs,
            )
        except ValueError:
            number_errors += 1
        assert number_errors == 1

        # test B
        number_errors = 0
        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=B,
                B=A,
                C=C,
                D=D,
                integrate_outputs=integrate_outputs,
            )
        except ValueError:
            number_errors += 1
        assert number_errors == 1

        # test C
        number_errors = 0
        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=C,
                B=B,
                C=A,
                D=D,
                integrate_outputs=integrate_outputs,
            )
        except ValueError:
            number_errors += 1
        assert number_errors == 1

        # test D
        number_errors = 0
        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=D,
                B=B,
                C=C,
                D=A,
                integrate_outputs=integrate_outputs,
            )
        except ValueError:
            number_errors += 1
        assert number_errors == 1

    # *************************************************************************
    # *************************************************************************

    def test_varying_matrix_sizes_integration(self):
        integrate_outputs = True

        time_step_durations = [1, 1]

        A1 = np.array([[1]])
        A2 = np.array([[1, 3], [4, 7]])

        A = [A1, A2]
        B = [np.array([[3]]), np.array([[4]])]
        C = [np.array([[2]]), np.array([[2]])]
        D = [np.array([[8]]), np.array([[7]])]

        # test A

        number_errors = 0
        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=A,
                B=B,
                C=C,
                D=D,
                integrate_outputs=integrate_outputs,
            )
        except ValueError:
            number_errors += 1
        assert number_errors == 1

        # test B

        number_errors = 0
        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=B,
                B=A,
                C=C,
                D=D,
                integrate_outputs=integrate_outputs,
            )
        except ValueError:
            number_errors += 1
        assert number_errors == 1

        # test C
        number_errors = 0
        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=C,
                B=B,
                C=A,
                D=D,
                integrate_outputs=integrate_outputs,
            )
        except ValueError:
            number_errors += 1
        assert number_errors == 1

        # test D

        number_errors = 0
        try:
            _ = dynsys.DynamicSystem(
                time_interval_durations=time_step_durations,
                A=D,
                B=B,
                C=C,
                D=A,
                integrate_outputs=integrate_outputs,
            )
        except ValueError:
            number_errors += 1
        assert number_errors == 1

    # *************************************************************************
    # *************************************************************************

    def test_dynsys_stateless_multiout(self):
        integrate_outputs = False
        number_outputs = 2
        method_dynsys_stateless_multiout(integrate_outputs, number_outputs)

        integrate_outputs = False
        number_outputs = 1
        method_dynsys_stateless_multiout(integrate_outputs, number_outputs)

        integrate_outputs = True
        number_outputs = 2
        method_dynsys_stateless_multiout(integrate_outputs, number_outputs)

        integrate_outputs = True
        number_outputs = 1
        method_dynsys_stateless_multiout(integrate_outputs, number_outputs)

        # integrate_outputs = False
        # number_outputs = 0
        # method_dynsys_stateless_multiout(integrate_outputs, number_outputs)

        # integrate_outputs = True
        # number_outputs = 0
        # method_dynsys_stateless_multiout(integrate_outputs, number_outputs)

    # *************************************************************************
    # *************************************************************************

    def test_dynsys_multiode_multiout(self):
        integrate_outputs = False
        number_outputs = 0
        method_dynsys_multiode_multiout(integrate_outputs, number_outputs)

        integrate_outputs = True
        number_outputs = 0
        method_dynsys_multiode_multiout(integrate_outputs, number_outputs)

        integrate_outputs = False
        number_outputs = 1
        method_dynsys_multiode_multiout(integrate_outputs, number_outputs)

        integrate_outputs = True
        number_outputs = 1
        method_dynsys_multiode_multiout(integrate_outputs, number_outputs)

        integrate_outputs = False
        number_outputs = 2
        method_dynsys_multiode_multiout(integrate_outputs, number_outputs)

        integrate_outputs = True
        number_outputs = 2
        method_dynsys_multiode_multiout(integrate_outputs, number_outputs)

    # *************************************************************************
    # *************************************************************************

    def test_dynsys_singleode_multiout(self):
        integrate_outputs = True
        number_outputs = 2
        method_dynsys_singleode_multiout(integrate_outputs, number_outputs)

        integrate_outputs = True
        number_outputs = 1
        method_dynsys_singleode_multiout(integrate_outputs, number_outputs)

        integrate_outputs = True
        number_outputs = 0
        method_dynsys_singleode_multiout(integrate_outputs, number_outputs)

        integrate_outputs = False
        number_outputs = 2
        method_dynsys_singleode_multiout(integrate_outputs, number_outputs)

        integrate_outputs = False
        number_outputs = 1
        method_dynsys_singleode_multiout(integrate_outputs, number_outputs)

        integrate_outputs = False
        number_outputs = 0
        method_dynsys_singleode_multiout(integrate_outputs, number_outputs)

    # *************************************************************************
    # *************************************************************************


def method_dynsys_stateless_multiout(integrate_outputs, number_outputs):
    # *************************************************************************
    # *************************************************************************

    # time steps
    list_regular_time_steps = [1, 1, 1, 1]
    list_irregular_time_steps = [1, 1.5, 0.5, 1]
    assert len(list_regular_time_steps) == len(list_irregular_time_steps)

    # define the inputs
    list_inputs = [-5, 50, 0.1, 1]

    # *************************************************************************
    # *************************************************************************

    # generate time varying problem

    list_A_matrices = []
    list_B_matrices = []
    list_C_matrices = []
    list_D_matrices = []
    for dt in list_irregular_time_steps:
        # data
        Aw, min_rel_heat = get_stateless_model_data(relative_amplitude_variation=0.1)
        # matrices
        (A_matrix, B_matrix, C_matrix, D_matrix) = stateless_model(Aw, min_rel_heat)
        if number_outputs == 0:
            D_matrix = None
        elif number_outputs != None:
            D_matrix = D_matrix[0:number_outputs, :]

        list_A_matrices.append(A_matrix)
        list_B_matrices.append(B_matrix)
        list_C_matrices.append(C_matrix)
        list_D_matrices.append(D_matrix)

    if number_outputs == 0:
        list_D_matrices = None

    # *************************************************************************
    # *************************************************************************

    # generate time invariant problem

    # data

    Aw, min_rel_heat = get_stateless_model_data()

    # matrices

    (A_matrix, B_matrix, C_matrix, D_matrix) = stateless_model(Aw, min_rel_heat)

    if number_outputs == 0:
        D_matrix = None
    elif number_outputs != None:
        D_matrix = D_matrix[0:number_outputs, :]

    # *************************************************************************
    # *************************************************************************

    # time invariant model, regular time steps
    x, y = example_dynsys_time_invariant(
        list_regular_time_steps,
        list_inputs,
        integrate_outputs,
        A_matrix,
        B_matrix,
        C_matrix,
        D_matrix,
        x0=None,
    )

    # TODO: check results

    # *************************************************************************
    # *************************************************************************

    # time invariant model, irregular time steps
    x, y = example_dynsys_time_invariant(
        list_irregular_time_steps,
        list_inputs,
        integrate_outputs,
        A_matrix,
        B_matrix,
        C_matrix,
        D_matrix,
        x0=None,
    )

    # TODO: check results


# *****************************************************************************
# *****************************************************************************


def method_stateless_system_object(integrate_outputs: bool):
    # regular time steps
    time_step_durations = [1, 1, 1, 1]

    # define a sequence of time steps
    t = np.array(
        [sum(time_step_durations[0:i]) for i in range(len(time_step_durations) + 1)]
    )

    # define the inputs
    list_inputs = [-5, 50, 0.1, 1]
    U = np.array(
        [[input_i for _ in range(len(time_step_durations))] for input_i in list_inputs]
    )

    # get the model
    Aw, min_rel_heat = get_stateless_model_data()
    _, _, _, D = stateless_model(Aw, min_rel_heat)
    ds = dynsys.StatelessSystem(
        time_interval_durations=time_step_durations,
        D=D,
        integrate_outputs=integrate_outputs,
    )

    # define the initial conditions
    X, Y = ds.simulate(U)
    assert X == None
    assert isinstance(Y, np.ndarray)
    if integrate_outputs:
        assert Y.shape == (2, len(t) - 1)  # two outputs
    else:
        assert Y.shape == (2, len(t))  # two outputs

    # *********************************************************************

    # regular time steps (as an int)
    time_step_durations = [1]

    # define a sequence of time steps
    t = np.array(
        [sum(time_step_durations[0:i]) for i in range(len(time_step_durations) + 1)]
    )

    # define the inputs
    list_inputs = [-5, 50, 0.1, 1]
    U = np.array(
        [[input_i for _ in range(len(time_step_durations))] for input_i in list_inputs]
    )

    # get the model
    Aw, min_rel_heat = get_stateless_model_data()
    _, _, _, D = stateless_model(Aw, min_rel_heat)
    ds = dynsys.StatelessSystem(
        time_interval_durations=time_step_durations[0],
        D=D,
        integrate_outputs=integrate_outputs,
    )

    # define the initial conditions
    X, Y = ds.simulate(U)
    # X, Y = ds.simulate(U[:,:-1]) # to avoid having the second

    assert X == None
    assert isinstance(Y, np.ndarray)
    if integrate_outputs:
        assert Y.shape == (2, len(t) - 1)  # two outputs
    else:
        assert Y.shape == (2, len(t))  # two outputs

    # *********************************************************************

    # irregular time steps
    time_step_durations = [1, 1.5, 0.5, 1]
    # define a sequence of time steps
    t = np.array(
        [sum(time_step_durations[0:i]) for i in range(len(time_step_durations) + 1)]
    )
    # define the inputs
    list_inputs = [-5, 50, 0.1, 1]
    U = np.array(
        [[input_i for _ in range(len(time_step_durations))] for input_i in list_inputs]
    )
    # get the model
    Aw, min_rel_heat = get_stateless_model_data()
    _, _, _, D = stateless_model(Aw, min_rel_heat)
    ds = dynsys.StatelessSystem(
        time_interval_durations=time_step_durations,
        D=D,
        integrate_outputs=integrate_outputs,
    )
    # define the initial conditions
    X, Y = ds.simulate(U)
    assert X == None
    assert isinstance(Y, np.ndarray)
    if integrate_outputs:
        assert Y.shape == (2, len(t) - 1)  # two outputs
    else:
        assert Y.shape == (2, len(t))  # two outputs


# *****************************************************************************
# *****************************************************************************


def method_dynsys_multiode_multiout(integrate_outputs: bool, number_outputs: int):
    # *************************************************************************

    # time steps
    list_regular_time_steps = [1, 1, 1, 1]
    list_irregular_time_steps = [1, 1.5, 0.5, 1]
    assert len(list_regular_time_steps) == len(list_irregular_time_steps)

    # define the inputs
    list_inputs = [-5, 50, 0.1, 1]

    # *************************************************************************

    # generate time varying problem
    list_A_matrices = []
    list_B_matrices = []
    list_C_matrices = []
    list_D_matrices = []

    for dt in list_irregular_time_steps:
        # data
        Ci, Ch, Ria, Rih, Aw, min_rel_heat, x0 = get_multi_ode_model_data(
            relative_amplitude_variation=0.1
        )

        # matrices
        (A_matrix, B_matrix, C_matrix, D_matrix) = two_node_model(
            Ci, Ch, Ria, Rih, Aw, min_rel_heat
        )

        if number_outputs == 0:
            C_matrix = None
            D_matrix = None

        elif number_outputs != None:
            C_matrix = C_matrix[0:number_outputs, :]
            D_matrix = D_matrix[0:number_outputs, :]

        list_A_matrices.append(A_matrix)
        list_B_matrices.append(B_matrix)
        list_C_matrices.append(C_matrix)
        list_D_matrices.append(D_matrix)

    if number_outputs == 0:
        list_C_matrices = None
        list_D_matrices = None

    # *************************************************************************

    # generate time invariant problem

    # data
    Ci, Ch, Ria, Rih, Aw, min_rel_heat, x0 = get_multi_ode_model_data()

    # matrices
    (A_matrix, B_matrix, C_matrix, D_matrix) = two_node_model(
        Ci, Ch, Ria, Rih, Aw, min_rel_heat
    )

    if number_outputs == 0:
        C_matrix = None
        D_matrix = None
    elif number_outputs != None:
        C_matrix = C_matrix[0:number_outputs, :]
        D_matrix = D_matrix[0:number_outputs, :]

    # *************************************************************************

    # time invariant model, regular time steps

    x_scipy, y_scipy = example_scipy_time_invariant_regular_steps(
        list_regular_time_steps,
        list_inputs,
        integrate_outputs,
        A_matrix,
        B_matrix,
        C_matrix,
        D_matrix,
        x0,
    )

    x, y = example_dynsys_time_invariant(
        list_regular_time_steps,
        list_inputs,
        integrate_outputs,
        A_matrix,
        B_matrix,
        C_matrix,
        D_matrix,
        x0,
    )

    for x_scipy_i, x_i in zip(x_scipy, x):
        np.testing.assert_allclose(x_i, x_scipy_i)

    if number_outputs != 0:
        for y_scipy_i, y_i in zip(y_scipy, y):
            np.testing.assert_allclose(y_i, y_scipy_i)

    # *************************************************************************

    # time invariant model, irregular time steps

    x_scipy, y_scipy = example_scipy_time_invariant_irregular_steps(
        list_irregular_time_steps,
        list_inputs,
        integrate_outputs,
        A_matrix,
        B_matrix,
        C_matrix,
        D_matrix,
        x0,
    )

    x, y = example_dynsys_time_invariant(
        list_irregular_time_steps,
        list_inputs,
        integrate_outputs,
        A_matrix,
        B_matrix,
        C_matrix,
        D_matrix,
        x0,
    )

    for x_scipy_i, x_i in zip(x_scipy, x):
        np.testing.assert_allclose(x_i, x_scipy_i)

    if number_outputs != 0:
        for y_scipy_i, y_i in zip(y_scipy, y):
            np.testing.assert_allclose(y_i, y_scipy_i)

    # *************************************************************************

    # time-varying model, regular time steps

    x_scipy, y_scipy = example_scipy_time_varying(
        list_regular_time_steps,
        list_inputs,
        integrate_outputs,
        list_A_matrices,
        list_B_matrices,
        list_C_matrices,
        list_D_matrices,
        x0,
    )

    x, y = example_dynsys_time_varying(
        list_regular_time_steps,
        list_inputs,
        integrate_outputs,
        list_A_matrices,
        list_B_matrices,
        list_C_matrices,
        list_D_matrices,
        x0,
    )

    for x_scipy_i, x_i in zip(x_scipy, x):
        np.testing.assert_allclose(x_i, x_scipy_i)

    if number_outputs != 0:
        for y_scipy_i, y_i in zip(y_scipy, y):
            np.testing.assert_allclose(y_i, y_scipy_i)

    # *************************************************************************

    # time-varying model, irregular time steps

    x_scipy, y_scipy = example_scipy_time_varying(
        list_irregular_time_steps,
        list_inputs,
        integrate_outputs,
        list_A_matrices,
        list_B_matrices,
        list_C_matrices,
        list_D_matrices,
        x0,
    )

    x, y = example_dynsys_time_varying(
        list_irregular_time_steps,
        list_inputs,
        integrate_outputs,
        list_A_matrices,
        list_B_matrices,
        list_C_matrices,
        list_D_matrices,
        x0,
    )

    for x_scipy_i, x_i in zip(x_scipy, x):
        np.testing.assert_allclose(x_i, x_scipy_i)

    if number_outputs != 0:
        for y_scipy_i, y_i in zip(y_scipy, y):
            np.testing.assert_allclose(y_i, y_scipy_i)


# *****************************************************************************
# *****************************************************************************

# test single ODE, multi-output dynamic systems while integrating outputs

#     examples_dynsys_singleode_multiout(True, print_outputs, seed_number)

#     # test single ODE, multi-output dynamic systems without integrating outputs

#     examples_dynsys_singleode_multiout(False, print_outputs, seed_number)

#     # test single-ODE, single-output dynamic systems while integrating outputs

#     examples_dynsys_singleode_multiout(True, print_outputs, seed_number, 1)

#     # test single-ODE, single-output dynamic systems without integrating outputs

#     examples_dynsys_singleode_multiout(False, print_outputs, seed_number, 1)

#     # *************************************************************************

#     # outputless

#     # test single-ODE, outputless dynamic systems while integrating outputs

#     examples_dynsys_singleode_multiout(True, print_outputs, seed_number, 0)

#     # test single-ODE, outputless dynamic systems without integrating outputs

#     examples_dynsys_singleode_multiout(False, print_outputs, seed_number, 0)


def method_dynsys_singleode_multiout(integrate_outputs: bool, number_outputs: int = 2):
    # *************************************************************************

    # time steps
    list_regular_time_steps = [1, 1, 1, 1]
    list_irregular_time_steps = [1, 1.5, 0.5, 1]
    assert len(list_regular_time_steps) == len(list_irregular_time_steps)

    # define the inputs
    list_inputs = [-5, 50, 0.1, 1]

    # *************************************************************************

    # generate time varying problem
    list_A_matrices = []
    list_B_matrices = []
    list_C_matrices = []
    list_D_matrices = []

    for dt in list_irregular_time_steps:
        # data

        Ci, Ria, Aw, min_rel_heat, x0 = get_single_ode_model_data(
            relative_amplitude_variation=0.1
        )

        # matrices

        (A_matrix, B_matrix, C_matrix, D_matrix) = single_node_model(
            Ci, Ria, Aw, min_rel_heat
        )

        if number_outputs == 0:
            C_matrix = None
            D_matrix = None
        elif number_outputs != None:
            C_matrix = C_matrix[0:number_outputs, :]
            D_matrix = D_matrix[0:number_outputs, :]

        list_A_matrices.append(A_matrix)
        list_B_matrices.append(B_matrix)
        list_C_matrices.append(C_matrix)
        list_D_matrices.append(D_matrix)

    if number_outputs == 0:
        list_C_matrices = None
        list_D_matrices = None

    # *************************************************************************

    # generate time invariant problem

    # data
    Ci, Ria, Aw, min_rel_heat, x0 = get_single_ode_model_data()
    # matrices
    (A_matrix, B_matrix, C_matrix, D_matrix) = single_node_model(
        Ci, Ria, Aw, min_rel_heat
    )

    if number_outputs == 0:
        C_matrix = None
        D_matrix = None
    elif number_outputs != None:
        C_matrix = C_matrix[0:number_outputs, :]
        D_matrix = D_matrix[0:number_outputs, :]

    # *************************************************************************

    # time invariant model, regular time steps

    x_scipy, y_scipy = example_scipy_time_invariant_regular_steps(
        list_regular_time_steps,
        list_inputs,
        integrate_outputs,
        A_matrix,
        B_matrix,
        C_matrix,
        D_matrix,
        x0,
    )

    x, y = example_dynsys_time_invariant(
        list_regular_time_steps,
        list_inputs,
        integrate_outputs,
        A_matrix,
        B_matrix,
        C_matrix,
        D_matrix,
        x0,
    )

    for x_scipy_i, x_i in zip(x_scipy, x):
        np.testing.assert_allclose(x_i, x_scipy_i)

    if number_outputs != 0:
        for y_scipy_i, y_i in zip(y_scipy, y):
            np.testing.assert_allclose(y_i, y_scipy_i)

    # *************************************************************************

    # time invariant model, irregular time steps

    x_scipy, y_scipy = example_scipy_time_invariant_irregular_steps(
        list_irregular_time_steps,
        list_inputs,
        integrate_outputs,
        A_matrix,
        B_matrix,
        C_matrix,
        D_matrix,
        x0,
    )

    x, y = example_dynsys_time_invariant(
        list_irregular_time_steps,
        list_inputs,
        integrate_outputs,
        A_matrix,
        B_matrix,
        C_matrix,
        D_matrix,
        x0,
    )

    for x_scipy_i, x_i in zip(x_scipy, x):
        np.testing.assert_allclose(x_i, x_scipy_i)

    if number_outputs != 0:
        for y_scipy_i, y_i in zip(y_scipy, y):
            np.testing.assert_allclose(y_i, y_scipy_i)

    # *************************************************************************

    # time-varying model, regular time steps

    x_scipy, y_scipy = example_scipy_time_varying(
        list_regular_time_steps,
        list_inputs,
        integrate_outputs,
        list_A_matrices,
        list_B_matrices,
        list_C_matrices,
        list_D_matrices,
        x0,
    )

    x, y = example_dynsys_time_varying(
        list_regular_time_steps,
        list_inputs,
        integrate_outputs,
        list_A_matrices,
        list_B_matrices,
        list_C_matrices,
        list_D_matrices,
        x0,
    )

    for x_scipy_i, x_i in zip(x_scipy, x):
        np.testing.assert_allclose(x_i, x_scipy_i)

    if number_outputs != 0:
        for y_scipy_i, y_i in zip(y_scipy, y):
            np.testing.assert_allclose(y_i, y_scipy_i)

    # *************************************************************************

    # time-varying model, irregular time steps

    x_scipy, y_scipy = example_scipy_time_varying(
        list_irregular_time_steps,
        list_inputs,
        integrate_outputs,
        list_A_matrices,
        list_B_matrices,
        list_C_matrices,
        list_D_matrices,
        x0,
    )

    x, y = example_dynsys_time_varying(
        list_irregular_time_steps,
        list_inputs,
        integrate_outputs,
        list_A_matrices,
        list_B_matrices,
        list_C_matrices,
        list_D_matrices,
        x0,
    )

    for x_scipy_i, x_i in zip(x_scipy, x):
        np.testing.assert_allclose(x_i, x_scipy_i)

    if number_outputs != 0:
        for y_scipy_i, y_i in zip(y_scipy, y):
            np.testing.assert_allclose(y_i, y_scipy_i)


# *****************************************************************************
# *****************************************************************************


def example_scipy_time_invariant_regular_steps(
    time_step_durations: list,
    inputs_list: list,
    integrate_outputs: bool,
    A_matrix,
    B_matrix,
    C_matrix,
    D_matrix,
    x0,
):
    # define a sequence of time steps

    t = np.array(
        [sum(time_step_durations[0:i]) for i in range(len(time_step_durations) + 1)]
    )

    # define the inputs

    u = np.array([inputs_list for dt in t])

    if type(C_matrix) == type(None) and type(D_matrix) == type(None):
        ss = StateSpace(
            A_matrix,
            B_matrix,
            np.zeros((1, A_matrix.shape[1])),
            np.zeros((1, B_matrix.shape[1])),
        )

    else:
        ss = StateSpace(A_matrix, B_matrix, C_matrix, D_matrix)

    # simulate the system response

    tout, yout, xout = lsim(ss, U=u, T=t, X0=x0)

    # convert to matrix format if there is only one dimension

    if len(xout.shape) == 1:
        xout = np.array([xout]).T

    if len(yout.shape) == 1:
        yout = np.array([yout]).T

    if integrate_outputs:
        # yout's shape should be: (number of time steps, 2)

        # ignore the first instant

        yout = yout[1:, :]

        yout_m, yout_n = yout.shape

        # multiply each output by the respective time step

        yout = np.array(
            [
                [yout[m, n] * time_step_durations[n] for n in range(yout_n)]
                for m in range(yout_m)
            ]
        )

        # note: the code above should not work when C != 0

    return xout.T, yout.T


# *****************************************************************************
# *****************************************************************************


def example_scipy_time_invariant_irregular_steps(
    time_step_durations: list,
    inputs_list: list,
    integrate_outputs: bool,
    A_matrix,
    B_matrix,
    C_matrix,
    D_matrix,
    x0,
):
    # define a sequence of time steps

    t = np.array(
        [sum(time_step_durations[0:i]) for i in range(len(time_step_durations) + 1)]
    )

    # define the inputs

    u = np.array([inputs_list for _ in t])

    # get the model

    if type(C_matrix) == type(None) and type(D_matrix) == type(None):
        ss = StateSpace(
            A_matrix,
            B_matrix,
            np.zeros((1, A_matrix.shape[1])),
            np.zeros((1, B_matrix.shape[1])),
        )

        yout = np.zeros((1, len(time_step_durations) + 1))

    else:
        ss = StateSpace(A_matrix, B_matrix, C_matrix, D_matrix)

        yout = np.zeros((C_matrix.shape[0], len(time_step_durations) + 1))

    # declare output variables

    xout = np.zeros((A_matrix.shape[0], len(time_step_durations) + 1))

    # yout = np.zeros((C_matrix.shape[0], len(time_step_durations)+1))

    xout[:, 0] = x0

    # initial output

    if not integrate_outputs:
        # do not ignore the first instant

        yout[:, 0] = np.dot(ss.D, u[0, :])

    # else:

    #     yout[:,0] = np.array([0, 0])

    # for each time step

    for i in range(len(time_step_durations)):
        # time step duration

        dt = time_step_durations[i]

        ssd = ss.to_discrete(dt=dt)

        # simulate the system response

        _, ytemp, xtemp = dlsim(
            ssd, u=np.array([u[i, :], u[i + 1, :]]), t=np.array([0, dt]), x0=xout[:, i]
        )

        xout[:, i + 1] = xtemp[1, :]  # np.array([0,0])

        yout[:, i + 1] = ytemp[1, :]  # np.array([0,0])

    if integrate_outputs:
        # yout's shape should be: (2,number of time steps)

        # ignore the first instant

        yout = yout[:, 1:]

        yout_m, yout_n = yout.shape

        # multiply each output by the respective time step

        for m in range(yout_m):
            for n in range(yout_n):
                if n == 0:
                    continue  # ignore the first time interval

                yout[m, n] = yout[m, n] * time_step_durations[n]

    return xout, yout


# *****************************************************************************
# *****************************************************************************


def example_scipy_time_varying(
    time_step_durations: list,
    inputs_list: list,
    integrate_outputs: bool,
    A_matrix,
    B_matrix,
    C_matrix,
    D_matrix,
    x0,
):
    # define a sequence of time steps

    t = np.array(
        [sum(time_step_durations[0:i]) for i in range(len(time_step_durations) + 1)]
    )

    # define the inputs

    u = np.array([inputs_list for dt in t])

    # define the initial conditions

    xout = np.zeros((A_matrix[0].shape[0], len(time_step_durations) + 1))

    xout[:, 0] = x0

    if type(C_matrix) == type(None) and type(D_matrix) == type(None):
        yout = np.zeros((1, len(time_step_durations) + 1))

    else:
        yout = np.zeros((C_matrix[0].shape[0], len(time_step_durations) + 1))

    for i, dt in enumerate(time_step_durations):
        # get the model

        if type(C_matrix) == type(None) and type(D_matrix) == type(None):
            ss = StateSpace(
                A_matrix[i],
                B_matrix[i],
                np.zeros((1, A_matrix[i].shape[1])),
                np.zeros((1, B_matrix[i].shape[1])),
            )

        else:
            ss = StateSpace(A_matrix[i], B_matrix[i], C_matrix[i], D_matrix[i])

        if i == 0 and not integrate_outputs:
            # compute the initial output

            yout[:, 0] = np.dot(ss.D, u[0, :])

        ssd = ss.to_discrete(dt=dt)

        # simulate the system response

        if i == 0:
            _, ytemp, xtemp = dlsim(
                ssd, u=np.array([u[i], u[i]]), t=np.array([0, dt]), x0=x0
            )

        else:
            _, ytemp, xtemp = dlsim(
                ssd, u=np.array([u[i], u[i]]), t=np.array([0, dt]), x0=xtemp[1, :]
            )

        # assert that the sizes match

        xout[:, i + 1] = xtemp[1, :]

        yout[:, i + 1] = ytemp[1, :]

    if integrate_outputs:
        # yout's shape should be: (2,number of time steps)

        # ignore the first instant

        yout = yout[:, 1:]

        yout_m, yout_n = yout.shape

        # multiply each output by the respective time step

        for m in range(yout_m):
            for n in range(yout_n):
                if n == 0:
                    continue  # ignore the first time interval

                yout[m, n] = yout[m, n] * time_step_durations[n]

    return xout, yout


# *****************************************************************************
# *****************************************************************************


def example_dynsys_time_invariant(
    time_step_durations: list,
    inputs_list: list,
    integrate_outputs: bool,
    A_matrix,
    B_matrix,
    C_matrix,
    D_matrix,
    x0,
):
    # define a sequence of time steps

    t = np.array(
        [sum(time_step_durations[0:i]) for i in range(len(time_step_durations) + 1)]
    )

    # define the inputs

    # U = np.array([[input_i for dt in t]
    #               for input_i in inputs_list])

    U = np.array(
        [[input_i for _ in range(len(time_step_durations))] for input_i in inputs_list]
    )

    # get the model

    ds = dynsys.DynamicSystem(
        time_interval_durations=time_step_durations,
        A=A_matrix,
        B=B_matrix,
        C=C_matrix,
        D=D_matrix,
        integrate_outputs=integrate_outputs,
    )

    # define the initial conditions

    X, Y = ds.simulate(U, x0)

    return X, Y


# *****************************************************************************
# *****************************************************************************


def example_dynsys_time_varying(
    time_step_durations: list,
    inputs_list: list,
    integrate_outputs: bool,
    A_matrix,
    B_matrix,
    C_matrix,
    D_matrix,
    x0,
):
    # define a sequence of time steps

    t = np.array(
        [sum(time_step_durations[0:i]) for i in range(len(time_step_durations) + 1)]
    )

    # define the inputs

    # U = np.array([[input_i for dt in t]
    #               for input_i in inputs_list])

    U = np.array(
        [[input_i for _ in range(len(time_step_durations))] for input_i in inputs_list]
    )

    ds = dynsys.DynamicSystem(
        time_interval_durations=time_step_durations,
        A=A_matrix,
        B=B_matrix,
        C=C_matrix,
        D=D_matrix,
        integrate_outputs=integrate_outputs,
    )

    # define the initial conditions

    X, Y = ds.simulate(U, x0)

    return X, Y


# *****************************************************************************
# *****************************************************************************


def get_stateless_model_data(relative_amplitude_variation: float = 0.0):
    mrh_deviation = random.random() - 0.5

    Aw = 6.22  # original: 6.22 m2

    min_rel_heat = 0.2 * (1 + relative_amplitude_variation * mrh_deviation)

    return Aw, min_rel_heat


# *****************************************************************************
# *****************************************************************************


def get_single_ode_model_data(relative_amplitude_variation: float = 0.0):
    # define how the coefficients change

    Ria_deviation = random.random() - 0.5

    # define the (A, B, C and D) matrices
    # A: n*n
    # B: n*m
    # C: r*n
    # D: r*m

    Ci = 1.360 * 3600000
    Ria = (1 + relative_amplitude_variation * Ria_deviation) * 5.31 / 3600000
    Aw = 6.22

    min_rel_heat = 0.2

    x0 = np.array([20])

    return Ci, Ria, Aw, min_rel_heat, x0


# *****************************************************************************
# *****************************************************************************


def get_multi_ode_model_data(relative_amplitude_variation: float = 0.0):
    # define how the coefficients change

    Rih_deviation = random.random() - 0.5

    Ria_deviation = random.random() - 0.5

    # define the (A, B, C and D) matrices
    # A: n*n
    # B: n*m
    # C: r*n
    # D: r*m

    # from Bacher and Madsen (2011): model TiTh

    Ci = 1.360 * 3600000  # original: 1.36 kWh/ºC
    Ch = 0.309 * 3600000  # original: 0.309 kWh/ºC
    Ria = (
        (1 + relative_amplitude_variation * Ria_deviation) * 5.31 / 3600000
    )  # original: 5.31 ºC/kWh
    Rih = (
        (1 + relative_amplitude_variation * Rih_deviation) * 0.639 / 3600000
    )  # original: 0.639 ºC/kWh
    Aw = 6.22  # original: 6.22 m2

    min_rel_heat = 0.2

    x0 = np.array([20, 20])

    return Ci, Ch, Ria, Rih, Aw, min_rel_heat, x0


# *****************************************************************************
# *****************************************************************************


def stateless_model(Aw, min_rel_heat):
    # inputs: Ta, phi_s, phi_h above the minimum, phi_h status
    # outputs: solar irradiance, heat

    d = np.array([[0, Aw, 0, 0], [0, 0, (1 - min_rel_heat), min_rel_heat]])

    return None, None, None, d


# *****************************************************************************
# *****************************************************************************


def single_node_model(Ci, Ria, Aw, min_rel_heat):
    # states: Ti and Th
    # inputs: Ta, phi_s, phi_h above the minimum, phi_h status
    # outputs: solar irradiance, heat

    a = np.array([[-1 / (Ria * Ci)]])
    b = np.array(
        [[1 / (Ci * Ria), Aw / Ci, (1 - min_rel_heat) / Ci, min_rel_heat / Ci]]
    )
    c = np.array([[0], [0]])
    d = np.array([[0, Aw, 0, 0], [0, 0, (1 - min_rel_heat), min_rel_heat]])

    return a, b, c, d


# *****************************************************************************
# *****************************************************************************


def two_node_model(Ci, Ch, Ria, Rih, Aw, min_rel_heat):
    # states: Ti and Th
    # inputs: Ta, phi_s, phi_h above the minimum, phi_h status
    # outputs: solar irradiance, heat

    a = np.array(
        [[-(1 / Rih + 1 / Ria) / Ci, 1 / (Ci * Rih)], [1 / (Ch * Rih), -1 / (Ch * Rih)]]
    )
    b = np.array(
        [
            [1 / (Ci * Ria), Aw / Ci, 0, 0],
            [0, 0, (1 - min_rel_heat) / Ch, min_rel_heat / Ch],
        ]
    )
    c = np.array([[0, 0], [0, 0]])
    d = np.array([[0, Aw, 0, 0], [0, 0, (1 - min_rel_heat), min_rel_heat]])

    return a, b, c, d


# *****************************************************************************
# *****************************************************************************
