# imports

# local, internal
from src.topupopt.problems.esipp.time import TimeFrame
from src.topupopt.problems.esipp.time import EconomicTimeFrame
from math import isclose

# *****************************************************************************
# *****************************************************************************


class TestTimeFrame:
    # *************************************************************************
    # *************************************************************************

    # test problem from table 2.1

    def test_nonpositive_durations(self):

        # negative reporting period durations

        reporting_periods = {0: [0, 1, 2]}
        reporting_period_durations = {0: [1, -1, 1]}
        time_intervals = {0: [0, 1]}
        time_interval_durations = {0: [1, 1]}

        error_raised = True
        try:
            TimeFrame(
                reporting_periods=reporting_periods,
                reporting_period_durations=reporting_period_durations,
                time_intervals=time_intervals,
                time_interval_durations=time_interval_durations,
            )
        except ValueError:
            error_raised = True
        assert error_raised

        # zero time interval durations

        reporting_periods = {0: [0, 1, 2]}
        reporting_period_durations = {0: [1, 1, 1]}
        time_intervals = {0: [0, 1]}
        time_interval_durations = {0: [1, 0]}

        error_raised = True
        try:
            TimeFrame(
                reporting_periods=reporting_periods,
                reporting_period_durations=reporting_period_durations,
                time_intervals=time_intervals,
                time_interval_durations=time_interval_durations,
            )
        except ValueError:
            error_raised = True
        assert error_raised

    # *************************************************************************
    # *************************************************************************

    # test problem from table 2.1

    def test_table_21(self):
        # summary:
        # 1 assessment
        # 3 reporting periods
        # 2 time intervals

        reporting_periods = {0: [0, 1, 2]}
        reporting_period_durations = {
            0: [365 * 24 * 3600, 365 * 24 * 3600, 365 * 24 * 3600]
        }
        time_intervals = {0: [0, 1]}
        time_interval_durations = {0: [1, 1]}

        tf = TimeFrame(
            reporting_periods=reporting_periods,
            reporting_period_durations=reporting_period_durations,
            time_intervals=time_intervals,
            time_interval_durations=time_interval_durations,
        )
        
        assert tf.number_assessments() == 1
        # number of reporting periods
        assert tf.number_reporting_periods(0) == 3
        # number of time intervals
        assert tf.number_time_intervals(0) == 2
        # no overlapping assessments
        assert not tf.assessments_overlap()
        
        # q: valid
        assert tf.valid_q(reporting_periods)
        assert tf.valid_q(time_intervals)
        # q: invalid
        assert not tf.valid_q({1: 0})
        # q: complete
        assert tf.complete_q(reporting_periods)
        assert tf.complete_q(time_intervals)
        # q: incomplete
        assert not tf.complete_q({1: 0})

        # qk: valid
        assert tf.valid_qk(
            {
                (0, 0): 1,
                (0, 1): 1,
            }
        )
        # qk: not valid
        assert not tf.valid_qk(
            {
                (0, 0): 1,
                (0, 2): 1,
            }
        )
        # qk: valid and complete
        assert tf.complete_qk(
            {
                (0, 0): 1,
                (0, 1): 1,
            }
        )
        # qk: valid but not complete
        assert not tf.complete_qk(
            {
                (0, 0): 1,
            }
        )
        # qk: consecutive
        assert tf.consecutive_qk(
            {
                (0, 0): 1,
                (0, 1): 1,
            }
        )
        # qk: not consecutive
        assert not tf.consecutive_qk(
            {
                (0, 0): 1,
                (0, 2): 1,
            }
        )
        

        # qp: valid
        assert tf.valid_qp(
            {
                (0, 0): 1,
            }
        )
        # qp: not valid
        assert not tf.valid_qp({(0, 0): 1, (0, 1): 1, (0, 3): 1})
        # qp: valid and complete
        assert tf.complete_qp({(0, 0): 1, (0, 1): 1, (0, 2): 1})
        # qp: valid but not complete
        assert not tf.complete_qp({(0, 0): 1, (0, 1): 1})
        # qp: consecutive
        assert tf.consecutive_qp(
            {
                (0, 0): 1,
                (0, 1): 1,
            }
        )
        # qp: not consecutive
        assert not tf.consecutive_qp(
            {
                (0, 0): 1,
                (0, 3): 1,
            }
        )

        # qpk: valid
        assert tf.valid_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 1): 1,
            }
        )
        # qpk: not valid
        assert not tf.valid_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 1): 1,
                (0, 1, 0): 1,
                (0, 1, 1): 1,
                (0, 3, 0): 1,
                (0, 3, 1): 1,
            }
        )
        # qpk: valid and complete
        assert tf.complete_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 1): 1,
                (0, 1, 0): 1,
                (0, 1, 1): 1,
                (0, 2, 0): 1,
                (0, 2, 1): 1,
            }
        )
        # qpk: valid but not complete
        assert not tf.complete_qpk(
            {(0, 0, 0): 1, (0, 0, 1): 1, (0, 1, 0): 1, (0, 1, 1): 1}
        )
        # qpk: consecutive
        assert tf.consecutive_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 1): 1,
                (0, 1, 0): 1,
                (0, 1, 1): 1,
                (0, 2, 0): 1,
                (0, 2, 1): 1,
            }
        )
        # qpk: not consecutive
        assert not tf.consecutive_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 1): 1,
                (0, 1, 1): 1,
                (0, 1, 2): 1,
                (0, 2, 0): 1,
                (0, 2, 1): 1,
            }
        )

        qk_dict = {qk: None for qk in tf.qk()}
        qp_dict = {qp: None for qp in tf.qp()}
        qpk_dict = {qpk: None for qpk in tf.qpk()}

        assert tf.complete_qk(qk_dict)
        assert tf.complete_qp(qp_dict)
        assert tf.complete_qpk(qpk_dict)

    # *************************************************************************
    # *************************************************************************

    # test problem from table 2.2

    def test_table_22(self):
        # summary:
        # 1 assessment
        # 3 reporting periods
        # 2 time intervals

        reporting_periods = {0: [0], 1: [1, 2]}
        reporting_period_durations = {
            0: [365 * 24 * 3600],
            1: [365 * 24 * 3600, 365 * 24 * 3600],
        }
        time_intervals = {0: [0, 1], 1: [0, 1]}
        time_interval_durations = {0: [1, 1], 1: [1, 1]}

        tf = TimeFrame(
            reporting_periods=reporting_periods,
            reporting_period_durations=reporting_period_durations,
            time_intervals=time_intervals,
            time_interval_durations=time_interval_durations,
        )
        
        # number of assessments
        assert tf.number_assessments() == 2
        # number of reporting periods
        assert tf.number_reporting_periods(0) == 1
        assert tf.number_reporting_periods(1) == 2
        # number of time intervals
        assert tf.number_time_intervals(0) == 2
        assert tf.number_time_intervals(1) == 2
        # no overlapping assessments
        assert not tf.assessments_overlap()
        
        # q: valid
        assert tf.valid_q(reporting_periods)
        assert tf.valid_q(time_intervals)
        # q: invalid
        assert not tf.valid_q({2: 0})
        # q: complete
        assert tf.complete_q(reporting_periods)
        assert tf.complete_q(time_intervals)
        # q: incomplete
        assert not tf.complete_q({1: 0})

        # qk: valid
        assert tf.valid_qk(
            {
                (0, 0): 1,
                (0, 1): 1,
                (1, 0): 1,
                (1, 1): 1,
            }
        )
        # qk: not valid
        assert not tf.valid_qk(
            {
                (0, 0): 1,
                (0, 1): 1,
                (1, 0): 1,
                (1, 1): 1,
                (1, 2): 1,
            }
        )
        # qk: valid and complete
        assert tf.complete_qk(
            {
                (0, 0): 1,
                (0, 1): 1,
                (1, 0): 1,
                (1, 1): 1,
            }
        )
        # qk: valid but not complete
        assert not tf.complete_qk(
            {
                (0, 0): 1,
                # (0,1): 1,
                (1, 0): 1,
                (1, 1): 1,
            }
        )

        # qp: valid
        assert tf.valid_qp(
            {
                (0, 0): 1,
                (1, 1): 1,
                (1, 2): 1,
            }
        )
        # qp: not valid
        assert not tf.valid_qp(
            {
                (0, 0): 1,
                (1, 1): 1,
                (1, 3): 1,
            }
        )
        # qp: valid and complete
        assert tf.complete_qp(
            {
                (0, 0): 1,
                (1, 1): 1,
                (1, 2): 1,
            }
        )
        # qp: valid but not complete
        assert not tf.complete_qp({(0, 0): 1, (1, 1): 1})

        # qpk: valid
        assert tf.valid_qpk(
            {
                (0, 0, 0): 1,
                (1, 1, 1): 1,
            }
        )
        # qpk: not valid
        assert not tf.valid_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 1): 1,
                (1, 1, 0): 1,
                (1, 1, 1): 1,
                (1, 3, 0): 1,
                (1, 3, 1): 1,
            }
        )
        # qpk: valid and complete
        assert tf.complete_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 1): 1,
                (1, 1, 0): 1,
                (1, 1, 1): 1,
                (1, 2, 0): 1,
                (1, 2, 1): 1,
            }
        )
        # qpk: valid but not complete
        assert not tf.complete_qpk(
            {(0, 0, 0): 1, (0, 0, 1): 1, (1, 1, 0): 1, (1, 1, 1): 1}
        )
        # qpk: consecutive
        assert tf.consecutive_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 1): 1,
                (1, 1, 0): 1,
                (1, 1, 1): 1,
                (1, 2, 0): 1,
                (1, 2, 1): 1,
            }
        )
        # qpk: not consecutive
        assert not tf.consecutive_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 3): 1,
                (1, 1, 0): 1,
                (1, 1, 1): 1,
                (1, 2, 0): 1,
                (1, 2, 1): 1,
            }
        )

        qk_dict = {qk: None for qk in tf.qk()}
        qp_dict = {qp: None for qp in tf.qp()}
        qpk_dict = {qpk: None for qpk in tf.qpk()}

        assert tf.complete_qk(qk_dict)
        assert tf.complete_qp(qp_dict)
        assert tf.complete_qpk(qpk_dict)

    # *************************************************************************
    # *************************************************************************

    def test_table_23(self):
        # Problem with three reporting periods relying on two assessments
        # with two and three intervals: one assessment for the first two periods with two
        # intervals; and another for last period with three intervals.

        reporting_periods = {0: [0, 1], 1: [2]}
        reporting_period_durations = {
            1: [365 * 24 * 3600],
            0: [365 * 24 * 3600, 365 * 24 * 3600],
        }
        time_intervals = {0: [0, 1], 1: [0, 1, 2]}
        time_interval_durations = {0: [1, 1], 1: [1, 1, 1]}

        tf = TimeFrame(
            reporting_periods=reporting_periods,
            reporting_period_durations=reporting_period_durations,
            time_intervals=time_intervals,
            time_interval_durations=time_interval_durations,
        )
        
        # q: valid
        assert tf.valid_q(reporting_periods)
        assert not tf.valid_q({2: 1})
        assert tf.complete_q(reporting_periods)
        assert not tf.complete_q({1: [365 * 24 * 3600]})
        # no overlapping assessments
        assert not tf.assessments_overlap()

        # qk: valid
        assert tf.valid_qk(
            {
                (0, 0): 1,
                (0, 1): 1,
                (1, 0): 1,
                (1, 1): 1,
                (1, 2): 1,
            }
        )
        # qk: not valid
        assert not tf.valid_qk(
            {
                (0, 0): 1,
                (0, 1): 1,
                (1, 0): 1,
                (1, 1): 1,
                (1, 2): 1,
                (1, 3): 1,
            }
        )
        # qk: valid and complete
        assert tf.complete_qk(
            {
                (0, 0): 1,
                (0, 1): 1,
                (1, 0): 1,
                (1, 1): 1,
                (1, 2): 1,
            }
        )
        # qk: valid but not complete
        assert not tf.complete_qk(
            {
                (0, 0): 1,
                # (0,1): 1,
                (1, 0): 1,
                (1, 1): 1,
            }
        )

        # qp: valid
        assert tf.valid_qp(
            {
                (0, 0): 1,
                (0, 1): 1,
                (1, 2): 1,
            }
        )
        # qp: not valid
        assert not tf.valid_qp(
            {
                (0, 0): 1,
                (0, 1): 1,
                (1, 3): 1,
            }
        )
        # qp: valid and complete
        assert tf.complete_qp(
            {
                (0, 0): 1,
                (0, 1): 1,
                (1, 2): 1,
            }
        )
        # qp: valid but not complete
        assert not tf.complete_qp({(0, 0): 1, (1, 1): 1})

        # qpk: valid
        assert tf.valid_qpk(
            {
                (0, 0, 0): 1,
                (0, 1, 1): 1,
                (1, 2, 1): 1,
            }
        )
        # qpk: not valid
        assert not tf.valid_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 1): 1,
                (1, 1, 0): 1,
                (1, 1, 1): 1,
                (1, 3, 0): 1,
                (1, 3, 1): 1,
            }
        )
        # qpk: valid and complete
        assert tf.complete_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 1): 1,
                (0, 1, 0): 1,
                (0, 1, 1): 1,
                (1, 2, 0): 1,
                (1, 2, 1): 1,
                (1, 2, 2): 1,
            }
        )
        # qpk: valid but not complete
        assert not tf.complete_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 1): 1,
                (0, 1, 0): 1,
                (0, 1, 1): 1,
                (1, 2, 0): 1,
                (1, 2, 1): 1,
            }
        )
        # qpk: consecutive
        assert tf.consecutive_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 1): 1,
                (0, 1, 0): 1,
                (0, 1, 1): 1,
                (1, 2, 0): 1,
                (1, 2, 1): 1,
                (1, 2, 2): 1,
            }
        )
        # qpk: not consecutive
        assert not tf.consecutive_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 1): 1,
                (0, 1, 0): 1,
                (0, 1, 1): 1,
                (1, 2, 0): 1,
                (1, 2, 1): 1,
                (1, 2, 3): 1,
            }
        )

        qk_dict = {qk: None for qk in tf.qk()}
        qp_dict = {qp: None for qp in tf.qp()}
        qpk_dict = {qpk: None for qpk in tf.qpk()}

        assert tf.complete_qk(qk_dict)
        assert tf.complete_qp(qp_dict)
        assert tf.complete_qpk(qpk_dict)

    # *************************************************************************
    # *************************************************************************

    def test_table_24(self):
        
        # Problem with three reporting periods relying on two assessments
        # with two intervals: each assessment is used for all three reporting periods.

        reporting_periods = {0: [0, 1, 2], 1: [0, 1, 2]}
        reporting_period_durations = {
            0: [365 * 24 * 3600, 365 * 24 * 3600, 365 * 24 * 3600],
            1: [365 * 24 * 3600, 365 * 24 * 3600, 365 * 24 * 3600],
        }
        time_intervals = {0: [0, 1], 1: [0, 1]}
        time_interval_durations = {0: [1, 1], 1: [1, 1]}

        tf = TimeFrame(
            reporting_periods=reporting_periods,
            reporting_period_durations=reporting_period_durations,
            time_intervals=time_intervals,
            time_interval_durations=time_interval_durations,
        )
        
        # number of assessments
        assert tf.number_assessments() == 2
        # number of reporting periods
        assert tf.number_reporting_periods(0) == 3
        assert tf.number_reporting_periods(1) == 3
        # number of time intervals
        assert tf.number_time_intervals(0) == 2
        assert tf.number_time_intervals(1) == 2
        # assessments overlap
        assert tf.assessments_overlap()
        
        # q: valid
        assert tf.valid_q(reporting_periods)
        assert tf.valid_q(time_intervals)
        assert not tf.valid_q({2: 1})
        assert tf.complete_q(reporting_periods)
        assert tf.complete_q(time_intervals)
        assert not tf.complete_q({1: [365 * 24 * 3600]})

        # qk: valid
        assert tf.valid_qk(
            {
                (0, 0): 1,
                (0, 1): 1,
                (1, 0): 1,
                (1, 1): 1
            }
        )
        # qk: not valid
        assert not tf.valid_qk(
            {
                (0, 0): 1,
                (0, 1): 1,
                (1, 0): 1,
                (1, 1): 1,
                (1, 2): 1,
            }
        )
        # qk: valid and complete
        assert tf.complete_qk(
            {
                (0, 0): 1,
                (0, 1): 1,
                (1, 0): 1,
                (1, 1): 1
            }
        )
        # qk: valid but not complete
        assert not tf.complete_qk(
            {
                (0, 0): 1,
                (1, 0): 1,
                (1, 1): 1,
            }
        )

        # qp: valid
        assert tf.valid_qp(
            {
                (0, 0): 1,
                (0, 1): 1,
                (1, 2): 1,
            }
        )
        # qp: not valid
        assert not tf.valid_qp(
            {
                (0, 0): 1,
                (0, 1): 1,
                (1, 3): 1,
            }
        )
        # qp: valid and complete
        assert tf.complete_qp(
            {
                (0, 0): 1,
                (0, 1): 1,
                (0, 2): 1,
                (1, 0): 1,
                (1, 1): 1,
                (1, 2): 1,
            }
        )
        # qp: valid but not complete
        assert not tf.complete_qp({(0, 0): 1, (1, 1): 1})

        # qpk: valid
        assert tf.valid_qpk(
            {
                (0, 0, 0): 1,
                (0, 1, 1): 1,
                (1, 2, 1): 1,
            }
        )
        # qpk: not valid
        assert not tf.valid_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 1): 1,
                (1, 1, 0): 1,
                (1, 1, 1): 1,
                (1, 3, 0): 1,
                (1, 3, 1): 1,
            }
        )
        # qpk: valid and complete
        assert tf.complete_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 1): 1,
                (0, 1, 0): 1,
                (0, 1, 1): 1,
                (0, 2, 0): 1,
                (0, 2, 1): 1,
                (1, 0, 0): 1,
                (1, 0, 1): 1,
                (1, 1, 0): 1,
                (1, 1, 1): 1,
                (1, 2, 0): 1,
                (1, 2, 1): 1,
            }
        )
        # qpk: valid but not complete
        assert not tf.complete_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 1): 1,
                (0, 1, 0): 1,
                (0, 1, 1): 1,
                (1, 2, 0): 1,
                (1, 2, 1): 1,
            }
        )
        # qpk: consecutive
        assert tf.consecutive_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 1): 1,
                (0, 1, 0): 1,
                (0, 1, 1): 1,
                (0, 2, 0): 1,
                (0, 2, 1): 1,
                (1, 0, 0): 1,
                (1, 0, 1): 1,
                (1, 1, 0): 1,
                (1, 1, 1): 1,
                (1, 2, 0): 1,
                (1, 2, 1): 1,
            }
        )
        # qpk: not consecutive
        assert not tf.consecutive_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 1): 1,
                (0, 1, 0): 1,
                (0, 1, 1): 1,
                (0, 2, 0): 1,
                (0, 2, 2): 1,
                (1, 0, 0): 1,
                (1, 0, 1): 1,
                (1, 1, 0): 1,
                (1, 1, 1): 1,
                (1, 2, 0): 1,
                (1, 2, 1): 1,
            }
        )

        qk_dict = {qk: None for qk in tf.qk()}
        qp_dict = {qp: None for qp in tf.qp()}
        qpk_dict = {qpk: None for qpk in tf.qpk()}

        assert tf.complete_qk(qk_dict)
        assert tf.complete_qp(qp_dict)
        assert tf.complete_qpk(qpk_dict)

    # *************************************************************************
    # *************************************************************************

    def test_table_25(self):
        
        # Problem with three reporting periods relying on six assessments with
        # two intervals: each reporting period is covered by two assessments.

        reporting_periods = {
            0: [0], 
            1: [0], 
            2: [1], 
            3: [1], 
            4: [2], 
            5: [2]
            }
        reporting_period_durations = {
            0: [365 * 24 * 3600],
            1: [365 * 24 * 3600],
            2: [365 * 24 * 3600],
            3: [365 * 24 * 3600],
            4: [365 * 24 * 3600],
            5: [365 * 24 * 3600],
            }
        time_intervals = {
            0: [0, 1], 
            1: [0, 1], 
            2: [0, 1], 
            3: [0, 1], 
            4: [0, 1], 
            5: [0, 1]
            }
        time_interval_durations = {
            0: [1, 1], 
            1: [1, 1], 
            2: [1, 1], 
            3: [1, 1], 
            4: [1, 1], 
            5: [1, 1]
            }

        tf = TimeFrame(
            reporting_periods=reporting_periods,
            reporting_period_durations=reporting_period_durations,
            time_intervals=time_intervals,
            time_interval_durations=time_interval_durations,
        )
        
        # number of assessments
        assert tf.number_assessments() == 6
        # number of reporting periods
        assert tf.number_reporting_periods(0) == 1
        assert tf.number_reporting_periods(1) == 1
        assert tf.number_reporting_periods(2) == 1
        assert tf.number_reporting_periods(3) == 1
        assert tf.number_reporting_periods(4) == 1
        assert tf.number_reporting_periods(5) == 1
        # number of time intervals
        assert tf.number_time_intervals(0) == 2
        assert tf.number_time_intervals(1) == 2
        assert tf.number_time_intervals(2) == 2
        assert tf.number_time_intervals(3) == 2
        assert tf.number_time_intervals(4) == 2
        assert tf.number_time_intervals(5) == 2
        # assessments overlap
        assert tf.assessments_overlap()
        
        # q: valid
        assert tf.valid_q(reporting_periods)
        assert tf.valid_q(time_intervals)
        assert not tf.valid_q({7: 1})
        assert tf.complete_q(reporting_periods)
        assert tf.complete_q(time_intervals)
        assert not tf.complete_q({1: [365 * 24 * 3600]})

        # qk: valid
        assert tf.valid_qk(
            {
                (0, 0): 1,
                (0, 1): 1,
                (1, 0): 1,
                (1, 1): 1
            }
        )
        # qk: not valid
        assert not tf.valid_qk(
            {
                (0, 0): 1,
                (0, 1): 1,
                (1, 0): 1,
                (1, 1): 1,
                (1, 2): 1,
            }
        )
        # qk: valid and complete
        assert tf.complete_qk(
            {
                (0, 0): 1,
                (0, 1): 1,
                (1, 0): 1,
                (1, 1): 1,
                (2, 0): 1,
                (2, 1): 1,
                (3, 0): 1,
                (3, 1): 1,
                (4, 0): 1,
                (4, 1): 1,
                (5, 0): 1,
                (5, 1): 1
            }
        )
        # qk: valid but not complete
        assert not tf.complete_qk(
            {
                #(0, 0): 1,
                (0, 1): 1,
                (1, 0): 1,
                (1, 1): 1,
                (2, 0): 1,
                (2, 1): 1,
                (3, 0): 1,
                (3, 1): 1,
                (4, 0): 1,
                (4, 1): 1,
                (5, 0): 1,
                (5, 1): 1
            }
        )

        # qp: valid
        assert tf.valid_qp(
            {
                (0, 0): 1,
                (1, 0): 1,
                (2, 1): 1,
                (3, 1): 1,
                (4, 2): 1,
                (5, 2): 1,
            }
        )
        # qp: not valid
        assert not tf.valid_qp(
            {
                (0, 0): 1,
                (0, 1): 1,
                (1, 3): 1,
            }
        )
        # qp: valid and complete
        assert tf.complete_qp(
            {
                (0, 0): 1,
                (1, 0): 1,
                (2, 1): 1,
                (3, 1): 1,
                (4, 2): 1,
                (5, 2): 1,
            }
        )
        # qp: valid but not complete
        assert not tf.complete_qp({(0, 0): 1, (1, 1): 1})

        # qpk: valid
        assert tf.valid_qpk(
            {
                (0, 0, 0): 1,
                (1, 0, 0): 1,
                (2, 1, 0): 1,
                (3, 1, 0): 1,
                (4, 2, 0): 1,
                (5, 2, 0): 1,
                (0, 0, 1): 1,
                (1, 0, 1): 1,
                (2, 1, 1): 1,
                (3, 1, 1): 1,
                (4, 2, 1): 1,
                (5, 2, 1): 1,
            }
        )
        # qpk: not valid
        assert not tf.valid_qpk(
            {
                (0, 0, 0): 1,
                (0, 0, 1): 1,
                (1, 1, 0): 1,
                (1, 1, 1): 1,
                (1, 3, 0): 1,
                (1, 3, 1): 1,
            }
        )
        # qpk: valid and complete
        assert tf.complete_qpk(
            {
                (0, 0, 0): 1,
                (1, 0, 0): 1,
                (2, 1, 0): 1,
                (3, 1, 0): 1,
                (4, 2, 0): 1,
                (5, 2, 0): 1,
                (0, 0, 1): 1,
                (1, 0, 1): 1,
                (2, 1, 1): 1,
                (3, 1, 1): 1,
                (4, 2, 1): 1,
                (5, 2, 1): 1,
            }
        )
        # qpk: valid but not complete
        assert not tf.complete_qpk(
            {
                (0, 0, 0): 1,
                (1, 0, 0): 1,
                (2, 1, 0): 1,
                (3, 1, 0): 1,
                (4, 2, 0): 1,
                (5, 2, 0): 1,
                (0, 0, 1): 1,
                (1, 0, 1): 1,
                (2, 1, 1): 1,
                (3, 1, 1): 1,
                # (4, 2, 1): 1,
                # (5, 2, 1): 1,
            }
        )
        # qpk: consecutive
        assert tf.consecutive_qpk(
            {
                (0, 0, 0): 1,
                (1, 0, 0): 1,
                (2, 1, 0): 1,
                (3, 1, 0): 1,
                (4, 2, 0): 1,
                (5, 2, 0): 1,
                (0, 0, 1): 1,
                (1, 0, 1): 1,
                (2, 1, 1): 1,
                (3, 1, 1): 1,
                (4, 2, 1): 1,
                (5, 2, 1): 1,
            }
        )
        # qpk: not consecutive
        assert not tf.consecutive_qpk(
            {
                (0, 0, 0): 1,
                (1, 0, 0): 1,
                (2, 1, 0): 1,
                (3, 1, 0): 1,
                (4, 2, 0): 1,
                (5, 2, 0): 1,
                (0, 0, 2): 1,
                (1, 0, 1): 1,
                (2, 1, 1): 1,
                (3, 1, 1): 1,
                (4, 2, 1): 1,
                (5, 2, 1): 1,
            }
        )

        qk_dict = {qk: None for qk in tf.qk()}
        qp_dict = {qp: None for qp in tf.qp()}
        qpk_dict = {qpk: None for qpk in tf.qpk()}

        assert tf.complete_qk(qk_dict)
        assert tf.complete_qp(qp_dict)
        assert tf.complete_qpk(qpk_dict)

    # *************************************************************************
    # *************************************************************************

    def test_economic_timeframe_mode1(self):
        
        # using a single discount rate for all q and p
        
        reporting_periods = {0: [0, 1], 1: [2]}
        reporting_period_durations = {
            0: [365 * 24 * 3600, 365 * 24 * 3600],
            1: [365 * 24 * 3600],
        }
        time_intervals = {0: [0, 1], 1: [0, 1, 2]}
        time_interval_durations = {0: [1, 1], 1: [1, 1, 1]}

        discount_rate = 3.5/100
        
        etf = EconomicTimeFrame(
            discount_rate=discount_rate, 
            reporting_periods=reporting_periods, 
            reporting_period_durations=reporting_period_durations, 
            time_intervals=time_intervals, 
            time_interval_durations=time_interval_durations
            )
        
        # check the discount rates
        for q, p_list in reporting_periods.items():
            for p in p_list:
                assert etf.discount_rate(q, p) == 3.5/100

    # *************************************************************************
    # *************************************************************************

    def test_economic_timeframe_mode2(self):
        
        # using a list of discount rates for each q
        
        reporting_periods = {0: [0, 1], 1: [2]}
        reporting_period_durations = {
            
            0: [365 * 24 * 3600, 365 * 24 * 3600],
            1: [365 * 24 * 3600],
        }
        time_intervals = {0: [0, 1], 1: [0, 1, 2]}
        time_interval_durations = {0: [1, 1], 1: [1, 1, 1]}

        discount_rates = [3.5/100, 3.6/100, 3.7/100]
        
        etf = EconomicTimeFrame(
            discount_rates_p=discount_rates, 
            reporting_periods=reporting_periods, 
            reporting_period_durations=reporting_period_durations, 
            time_intervals=time_intervals, 
            time_interval_durations=time_interval_durations
            )
        
        # check the discount rates    
        assert etf.discount_rate(0, 0) == 3.5/100
        assert etf.discount_rate(0, 1) == 3.6/100
        assert etf.discount_rate(1, 2) == 3.7/100

    # *************************************************************************
    # *************************************************************************
    
    def test_economic_timeframe_mode3(self):
        
        # using a list of discount rates for each q
        
        reporting_periods = {0: [0, 1, 2], 1: [0, 1, 2]}
        reporting_period_durations = {
            0: [365 * 24 * 3600, 365 * 24 * 3600, 365 * 24 * 3600],
            1: [365 * 24 * 3600, 365 * 24 * 3600, 365 * 24 * 3600],
        }
        time_intervals = {0: [0, 1], 1: [0, 1]}
        time_interval_durations = {0: [1, 1], 1: [1, 1]}
    
        discount_rates = {
            0: [3.5/100, 3.6/100, 3.7/100],
            1: [1.5/100, 1.6/100, 1.7/100],
            }
        
        etf = EconomicTimeFrame(
            discount_rates_q=discount_rates, 
            reporting_periods=reporting_periods, 
            reporting_period_durations=reporting_period_durations, 
            time_intervals=time_intervals, 
            time_interval_durations=time_interval_durations
            )
        
        # check the discount rates
        assert etf.discount_rate(0, 0) == 3.5/100
        assert etf.discount_rate(0, 1) == 3.6/100
        assert etf.discount_rate(0, 2) == 3.7/100
        assert etf.discount_rate(1, 0) == 1.5/100
        assert etf.discount_rate(1, 1) == 1.6/100
        assert etf.discount_rate(1, 2) == 1.7/100

    # *************************************************************************
    # *************************************************************************
    
    def test_discount_factor_constant(self):
        
        discount_rate = 3.5/100
        
        etf = EconomicTimeFrame(
            discount_rate=discount_rate,
            reporting_periods={
                0: [0,1,2,3,4,5,6,7,8,9]
                }, 
            reporting_period_durations={
                0: [1,1,1,1,1,1,1,1,1,1]
                }, 
            time_intervals={
                0: [0]
                },
            time_interval_durations={
                0: [1]
                },
            )

        true_factors = [
            0.966, # 0
            0.934, # 1
            0.902, # 2
            0.871, # 3
            0.842, # 4
            0.814, # 5
            0.786, # 6
            0.759, # 7
            0.734, # 8
            0.709, # 9
        ]
        
        assessment = 0
        factors = [
            etf.discount_factor(assessment, p)
            for p in etf.reporting_periods[assessment]
            ]

        for df, true_df in zip(factors, true_factors):
            assert isclose(df, true_df, abs_tol=0.001)
            
    # *************************************************************************
    # *************************************************************************
    
    def test_discount_factor_single_period(self):
        
        discount_rate = 0
        
        etf = EconomicTimeFrame(
            discount_rate=discount_rate,
            reporting_periods={
                0: [0]
                }, 
            reporting_period_durations={
                0: [1]
                }, 
            time_intervals={
                0: [0]
                },
            time_interval_durations={
                0: [1]
                },
            )

        true_factors = [1]
        
        assessment = 0
        factors = [
            etf.discount_factor(assessment, p)
            for p in etf.reporting_periods[assessment]
            ]

        for df, true_df in zip(factors, true_factors):
            assert isclose(df, true_df, abs_tol=0.001)
            

    # *************************************************************************
    # *************************************************************************
    
    def test_discount_factor_variable(self):

        # variable discount factors
        
        etf = EconomicTimeFrame(
            discount_rates_p=[0.035, 0.05, 0.075, 0.06],
            reporting_periods={
                0: [0,1,2,3]
                }, 
            reporting_period_durations={
                0: [1,1,1,1]
                }, 
            time_intervals={
                0: [0]
                },
            time_interval_durations={
                0: [1]
                },
            )
        true_factors = [0.96618, 0.92017, 0.85598, 0.80753]
        
        assessment = 0
        factors = [
            etf.discount_factor(assessment, p)
            for p in etf.reporting_periods[assessment]
            ]

        for df, true_df in zip(factors, true_factors):
            assert isclose(df, true_df, abs_tol=0.001)
            
    # *************************************************************************
    # *************************************************************************
    
    def test_etf_unrecognised_input(self):

        # define the discount rate using a set
        error_raised = False
        try:
            EconomicTimeFrame(
                discount_rate={0.035},
                reporting_periods={
                    0: [0,1,2,3]
                    }, 
                reporting_period_durations={
                    0: [1,1,1,1]
                    }, 
                time_intervals={
                    0: [0]
                    },
                time_interval_durations={
                    0: [1]
                    },
                )
        except TypeError:
            error_raised = True
        assert error_raised

# *****************************************************************************
# *****************************************************************************
