# imports

# standard

from math import isclose
import random
from statistics import mean

# local, external

from shapely.geometry import Point, LineString
from shapely import length
import networkx as nx
import osmnx as ox

# from osmnx.utils_graph import get_undirected

# local, internal

import src.topupopt.data.gis.osm as osm
import src.topupopt.data.gis.modify as gis_mod
import src.topupopt.data.gis.calculate as gis_calc
import src.topupopt.data.gis.identify as gis_iden
import src.topupopt.problems.esipp.utils as prob_utils

# *****************************************************************************
# *****************************************************************************


class TestGisModify:
    # *************************************************************************
    # *************************************************************************

    def test_create_reverse_edges(self):
        network = nx.MultiDiGraph()

        node_key0 = 0
        node_key0_dict = {osm.KEY_OSMNX_X: 55, osm.KEY_OSMNX_Y: 25}
        node_key1 = 1
        node_key1_dict = {osm.KEY_OSMNX_X: 55.001, osm.KEY_OSMNX_Y: 25.001}

        network.add_node(node_key0, **node_key0_dict)
        network.add_node(node_key1, **node_key1_dict)
        # create a line between node 0 and node 1
        geo_line = LineString(
            [
                (node_key0_dict[osm.KEY_OSMNX_X], node_key0_dict[osm.KEY_OSMNX_Y]),
                (node_key1_dict[osm.KEY_OSMNX_X], node_key1_dict[osm.KEY_OSMNX_Y]),
            ]
        )
        # the same line, reversed
        geo_line_reversed = geo_line.reverse()

        k_no_geo = network.add_edge(
            0,
            1,
            **{
                osm.KEY_OSMNX_LENGTH: 3,
                osm.KEY_OSMNX_REVERSED: False,
                osm.KEY_OSMNX_OSMID: 1,
                osm.KEY_OSMNX_ONEWAY: False,
            }
        )

        k_normal = network.add_edge(
            0,
            1,
            **{
                osm.KEY_OSMNX_LENGTH: 3,
                osm.KEY_OSMNX_ONEWAY: False,
                osm.KEY_OSMNX_OSMID: 1,
                osm.KEY_OSMNX_REVERSED: [True, False],
                osm.KEY_OSMNX_GEOMETRY: geo_line,
            }
        )

        k_reversed = network.add_edge(
            0,
            1,
            **{
                osm.KEY_OSMNX_LENGTH: 3,
                osm.KEY_OSMNX_OSMID: 1,
                osm.KEY_OSMNX_ONEWAY: False,
                osm.KEY_OSMNX_REVERSED: True,
                osm.KEY_OSMNX_GEOMETRY: geo_line_reversed,
            }
        )

        # edge without geometry should be consistent
        edge_key = (node_key0, node_key1, k_no_geo)
        assert gis_iden.is_edge_osmnx_compliant(network, edge_key)
        new_edges = gis_mod.create_reverse_edges(network, edge_keys=[edge_key])
        for new_edge in new_edges:
            assert gis_iden.is_edge_osmnx_compliant(network, new_edge)
        network.remove_edges_from(new_edges)
        # edge with normal geometry should be consistent
        edge_key = (node_key0, node_key1, k_normal)
        assert gis_iden.is_edge_osmnx_compliant(network, edge_key)
        new_edges = gis_mod.create_reverse_edges(network, edge_keys=[edge_key])
        for new_edge in new_edges:
            assert gis_iden.is_edge_osmnx_compliant(network, new_edge)
        network.remove_edges_from(new_edges)
        # edge with reversed geometry should not be consistent
        edge_key = (node_key0, node_key1, k_reversed)
        assert gis_iden.is_edge_osmnx_compliant(network, edge_key)
        new_edges = gis_mod.create_reverse_edges(network, edge_keys=[edge_key])
        for new_edge in new_edges:
            assert gis_iden.is_edge_osmnx_compliant(network, new_edge)
        network.remove_edges_from(new_edges)

        assert network.number_of_edges() == 3
        new_edges = gis_mod.create_reverse_edges(network)
        assert len(new_edges) == 3

        # trigger no edge found error
        error_raised = False
        try:
            edge_key = (node_key0, node_key1, k_no_geo - 1)
            gis_mod.create_reverse_edges(network, [edge_key])
        except ValueError:
            error_raised = True
        assert error_raised

        # trigger valuerror due to incorrect reversed type
        k_error = network.add_edge(
            0,
            1,
            **{
                osm.KEY_OSMNX_LENGTH: 3,
                osm.KEY_OSMNX_REVERSED: {True, False},
                osm.KEY_OSMNX_ONEWAY: False,
                osm.KEY_OSMNX_OSMID: 1,
                osm.KEY_OSMNX_GEOMETRY: geo_line,
            }
        )
        error_raised = False
        try:
            edge_key = (node_key0, node_key1, k_error)
            gis_mod.create_reverse_edges(network, [edge_key])
        except ValueError:
            error_raised = True
        assert error_raised

    # *************************************************************************
    # *************************************************************************

    def test_replace_paths_osmnx(self):
        # get the network
        _net = ox.graph_from_point(
            (55.71654, 9.11728),
            network_type="drive",
            custom_filter='["highway"~"residential|tertiary|unclassified|service"]',
            truncate_by_edge=True,
        )

        # define the settings
        ignore_self_loops = False
        consider_reversed_edges = True

        # find paths
        paths = gis_iden.find_simplifiable_paths(
            _net,
            protected_nodes=[],
            ignore_self_loops=ignore_self_loops,
            consider_reversed_edges=consider_reversed_edges,
        )

        # verify the paths
        for path in paths:
            gis_iden.is_path_straight(
                _net,
                path,
                consider_reversed_edges=consider_reversed_edges,
                ignore_self_loops=ignore_self_loops,
            )

        # modify an edge in one of the paths to have list attributes
        _edge_key = tuple(
            gis_iden.get_edges_from_a_to_b(
                _net,
                paths[0][0],  # first path, first node
                paths[0][1],  # first paht, second node
            )
        )[0]
        _net.add_edge(
            *_edge_key,
            **{
                osm.KEY_OSMNX_ONEWAY: [_net.edges[_edge_key][osm.KEY_OSMNX_ONEWAY]],
                osm.KEY_OSMNX_REVERSED: [_net.edges[_edge_key][osm.KEY_OSMNX_REVERSED]],
            }
        )
        # measure the distances
        true_path_lengths = [gis_calc.node_path_length(_net, path) for path in paths]
        # replace the paths
        new_path_edges = [gis_mod.replace_path(_net, path) for path in paths]
        # compare
        for edge_key, true_length in zip(new_path_edges, true_path_lengths):
            assert isclose(
                _net.edges[edge_key][gis_iden.osm.KEY_OSMNX_LENGTH],
                true_length,
                abs_tol=1e-3,  # 23.400000000000034
            )

    # *************************************************************************
    # *************************************************************************

    def test_replace_nonsimplifiable_path(self):
        _net = nx.MultiDiGraph()

        path = [0, 1]

        error_raised = False
        try:
            gis_mod.replace_path(_net, path)
        except ValueError:
            error_raised = True
        assert error_raised

    # *************************************************************************
    # *************************************************************************

    def example_replace_paths(self, project_the_graph: bool = False):
        # *********************************************************************

        # smallest possible path

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"
        network.graph["simplified"] = False
        # add nodes
        network.add_nodes_from(
            [
                (0, {osm.KEY_OSMNX_Y: 56.00, osm.KEY_OSMNX_X: 12.00}),
                (1, {osm.KEY_OSMNX_Y: 56.01, osm.KEY_OSMNX_X: 12.00}),
                (2, {osm.KEY_OSMNX_Y: 56.02, osm.KEY_OSMNX_X: 12.01}),
            ]
        )
        # add edges
        network.add_edges_from(
            [
                (
                    0,
                    1,
                    0,
                    {
                        osm.KEY_OSMNX_OSMID: 1,
                        osm.KEY_OSMNX_LENGTH: 3,
                        osm.KEY_OSMNX_REVERSED: False,
                        osm.KEY_OSMNX_ONEWAY: False,
                    },
                ),
                (
                    1,
                    2,
                    0,
                    {
                        osm.KEY_OSMNX_OSMID: 2,
                        osm.KEY_OSMNX_LENGTH: 4,
                        osm.KEY_OSMNX_REVERSED: False,
                        osm.KEY_OSMNX_ONEWAY: False,
                    },
                ),
            ]
        )
        lengths = gis_calc.edge_lengths(network, edge_keys=[(0, 1, 0), (1, 2, 0)])
        network.edges[(0, 1, 0)][osm.KEY_OSMNX_LENGTH] = lengths[(0, 1, 0)]
        network.edges[(1, 2, 0)][osm.KEY_OSMNX_LENGTH] = lengths[(1, 2, 0)]
        number_edges = network.number_of_edges()
        if project_the_graph:
            network = ox.project_graph(G=network)
        path = [0, 1, 2]
        new_edge_key = gis_mod.replace_path(network, path=path)
        # a new edge should exist
        assert network.has_edge(*new_edge_key)
        # assert that two edges are removed and one is created
        assert network.number_of_edges() - number_edges == 1 - 2
        # intermediate nodes should not exist any longer
        for node in path[1:-1]:
            assert not network.has_node(node)
        # the distances need to match
        assert isclose(
            sum(lengths.values()),
            network.edges[new_edge_key][osm.KEY_OSMNX_LENGTH],
            abs_tol=1e-3,
        )
        # the new edge needs to have a geometry because it is not simple
        assert osm.KEY_OSMNX_GEOMETRY in network.edges[new_edge_key]
        # the geometry must have 3 points
        assert (
            len(tuple(network.edges[new_edge_key][osm.KEY_OSMNX_GEOMETRY].coords)) == 3
        )
        new_edge_key_lengths = gis_calc.edge_lengths(network, edge_keys=[new_edge_key])
        # the geometry's length needs to match that of the edge's
        assert isclose(
            network.edges[new_edge_key][osm.KEY_OSMNX_LENGTH],
            new_edge_key_lengths[new_edge_key],
            abs_tol=1 if not project_the_graph else 3.861,  # 3.8601244551728087
        )
        # print('hallo1')
        # print(tuple(network.edges[new_edge_key][osm.KEY_OSMNX_GEOMETRY].coords))
        # print(network.edges[new_edge_key][osm.KEY_OSMNX_LENGTH])
        # print(new_edge_key_lengths[new_edge_key])

        # *********************************************************************

        # small path with 3 edges

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"
        network.graph["simplified"] = False
        # add nodes
        network.add_nodes_from(
            [
                (0, {osm.KEY_OSMNX_Y: 56.00, osm.KEY_OSMNX_X: 12.00}),
                (1, {osm.KEY_OSMNX_Y: 56.01, osm.KEY_OSMNX_X: 12.00}),
                (2, {osm.KEY_OSMNX_Y: 56.02, osm.KEY_OSMNX_X: 12.01}),
                (3, {osm.KEY_OSMNX_Y: 56.04, osm.KEY_OSMNX_X: 12.02}),
            ]
        )
        # add edges
        network.add_edges_from(
            [
                (
                    0,
                    1,
                    0,
                    {
                        osm.KEY_OSMNX_OSMID: 1,
                        osm.KEY_OSMNX_LENGTH: 3,
                        osm.KEY_OSMNX_REVERSED: False,
                        osm.KEY_OSMNX_ONEWAY: False,
                    },
                ),
                (
                    1,
                    2,
                    0,
                    {
                        osm.KEY_OSMNX_OSMID: 2,
                        osm.KEY_OSMNX_LENGTH: 4,
                        osm.KEY_OSMNX_REVERSED: False,
                        osm.KEY_OSMNX_ONEWAY: False,
                    },
                ),
                (
                    2,
                    3,
                    0,
                    {
                        osm.KEY_OSMNX_OSMID: 3,
                        osm.KEY_OSMNX_LENGTH: 5,
                        osm.KEY_OSMNX_REVERSED: False,
                        osm.KEY_OSMNX_ONEWAY: False,
                    },
                ),
            ]
        )
        lengths = gis_calc.edge_lengths(
            network, edge_keys=[(0, 1, 0), (1, 2, 0), (2, 3, 0)]
        )
        network.edges[(0, 1, 0)][osm.KEY_OSMNX_LENGTH] = lengths[(0, 1, 0)]
        network.edges[(1, 2, 0)][osm.KEY_OSMNX_LENGTH] = lengths[(1, 2, 0)]
        network.edges[(2, 3, 0)][osm.KEY_OSMNX_LENGTH] = lengths[(2, 3, 0)]
        number_edges = network.number_of_edges()
        if project_the_graph:
            network = ox.project_graph(G=network)
        path = [0, 1, 2, 3]
        new_edge_key = gis_mod.replace_path(network, path=path)
        # a new edge should exist
        assert network.has_edge(*new_edge_key)
        # assert that two edges are removed and one is created
        assert network.number_of_edges() - number_edges == 1 - 3
        # intermediate nodes should not exist any longer
        for node in path[1:-1]:
            assert not network.has_node(node)
        # the distances need to match
        assert isclose(
            sum(lengths.values()),
            network.edges[new_edge_key][osm.KEY_OSMNX_LENGTH],
            abs_tol=1e-3,
        )
        # the new edge needs to have a geometry because it is not simple
        assert osm.KEY_OSMNX_GEOMETRY in network.edges[new_edge_key]
        # the geometry must have 4 points
        assert (
            len(tuple(network.edges[new_edge_key][osm.KEY_OSMNX_GEOMETRY].coords)) == 4
        )
        new_edge_key_lengths = gis_calc.edge_lengths(network, edge_keys=[new_edge_key])
        # the geometry's length needs to match that of the edge's
        assert isclose(
            network.edges[new_edge_key][osm.KEY_OSMNX_LENGTH],
            new_edge_key_lengths[new_edge_key],
            abs_tol=1 if not project_the_graph else 7.33,  # 7.327521377403173
        )
        # print('hallo2')
        # print(tuple(network.edges[new_edge_key][osm.KEY_OSMNX_GEOMETRY].coords))
        # print(network.edges[new_edge_key][osm.KEY_OSMNX_LENGTH])
        # print(new_edge_key_lengths[new_edge_key])

        # # *********************************************************************

        # smallest possible path, but featuring simplified geometries already

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"
        network.graph["simplified"] = True
        # add nodes
        network.add_nodes_from(
            [
                (0, {osm.KEY_OSMNX_Y: 56.00, osm.KEY_OSMNX_X: 12.00}),
                (1, {osm.KEY_OSMNX_Y: 56.01, osm.KEY_OSMNX_X: 12.00}),
                (2, {osm.KEY_OSMNX_Y: 56.02, osm.KEY_OSMNX_X: 12.01}),
            ]
        )
        # add edges
        network.add_edges_from(
            [
                (
                    0,
                    1,
                    0,
                    {
                        osm.KEY_OSMNX_OSMID: 1,
                        osm.KEY_OSMNX_LENGTH: 5,
                        osm.KEY_OSMNX_REVERSED: False,
                        osm.KEY_OSMNX_ONEWAY: False,
                    },
                ),
                (
                    1,
                    2,
                    0,
                    {
                        osm.KEY_OSMNX_OSMID: 2,
                        osm.KEY_OSMNX_LENGTH: 10,
                        osm.KEY_OSMNX_REVERSED: False,
                        osm.KEY_OSMNX_ONEWAY: False,
                    },
                ),
            ]
        )
        # build create geometries
        edge01_geo = LineString([(12.00, 56.00), (11.99, 56.00), (12.00, 56.01)])
        edge12_geo = LineString([(12.00, 56.01), (12.02, 56.00), (12.01, 56.02)])
        network.edges[(0, 1, 0)][osm.KEY_OSMNX_GEOMETRY] = edge01_geo
        network.edges[(1, 2, 0)][osm.KEY_OSMNX_GEOMETRY] = edge12_geo
        lengths = gis_calc.edge_lengths(network, edge_keys=[(0, 1, 0), (1, 2, 0)])
        network.edges[(0, 1, 0)][osm.KEY_OSMNX_LENGTH] = lengths[(0, 1, 0)]
        network.edges[(1, 2, 0)][osm.KEY_OSMNX_LENGTH] = lengths[(1, 2, 0)]
        number_edges = network.number_of_edges()
        if project_the_graph:
            network = ox.project_graph(G=network)
        path = [0, 1, 2]
        new_edge_key = gis_mod.replace_path(network, path=path)
        # a new edge should exist
        assert network.has_edge(*new_edge_key)
        # assert that two edges are removed and one is created
        assert network.number_of_edges() - number_edges == 1 - 2
        # intermediate nodes should not exist any longer
        for node in path[1:-1]:
            assert not network.has_node(node)
        # the distances need to match
        assert isclose(
            sum(lengths.values()),
            network.edges[new_edge_key][osm.KEY_OSMNX_LENGTH],
            abs_tol=1e-3,  # 1 if not project_the_graph else 1e-3
        )
        # the new edge needs to have a geometry because it is not simple
        assert osm.KEY_OSMNX_GEOMETRY in network.edges[new_edge_key]
        # the geometry must have 5 points
        assert (
            len(tuple(network.edges[new_edge_key][osm.KEY_OSMNX_GEOMETRY].coords)) == 5
        )
        new_edge_key_lengths = gis_calc.edge_lengths(network, edge_keys=[new_edge_key])
        # the geometry's length needs to match that of the edge's
        assert isclose(
            network.edges[new_edge_key][osm.KEY_OSMNX_LENGTH],
            new_edge_key_lengths[new_edge_key],
            abs_tol=1 if not project_the_graph else 12.2,  # -12.178460200064364
        )

        # *********************************************************************

        # smallest possible path, but featuring reversed geometries already (#1)

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"
        network.graph["simplified"] = True
        # add nodes
        network.add_nodes_from(
            [
                (0, {osm.KEY_OSMNX_Y: 56.00, osm.KEY_OSMNX_X: 12.00}),
                (1, {osm.KEY_OSMNX_Y: 56.01, osm.KEY_OSMNX_X: 12.00}),
                (2, {osm.KEY_OSMNX_Y: 56.02, osm.KEY_OSMNX_X: 12.01}),
            ]
        )
        # add edges
        network.add_edges_from(
            [
                (
                    0,
                    1,
                    0,
                    {
                        osm.KEY_OSMNX_OSMID: 1,
                        osm.KEY_OSMNX_LENGTH: 3,
                        osm.KEY_OSMNX_REVERSED: False,
                        osm.KEY_OSMNX_ONEWAY: False,
                    },
                ),
                (
                    1,
                    2,
                    0,
                    {
                        osm.KEY_OSMNX_OSMID: 2,
                        osm.KEY_OSMNX_LENGTH: 4,
                        osm.KEY_OSMNX_REVERSED: False,
                        osm.KEY_OSMNX_ONEWAY: False,
                    },
                ),
            ]
        )
        # build create geometries
        edge01_geo = LineString([(12.00, 56.00), (11.99, 56.00), (12.00, 56.01)])
        edge12_geo = LineString([(12.01, 56.02), (12.02, 56.00), (12.00, 56.01)])
        network.edges[(0, 1, 0)][osm.KEY_OSMNX_GEOMETRY] = edge01_geo
        network.edges[(1, 2, 0)][osm.KEY_OSMNX_GEOMETRY] = edge12_geo
        network.edges[(0, 1, 0)][osm.KEY_OSMNX_REVERSED] = False
        network.edges[(1, 2, 0)][osm.KEY_OSMNX_REVERSED] = True
        lengths = gis_calc.edge_lengths(network, edge_keys=[(0, 1, 0), (1, 2, 0)])
        network.edges[(0, 1, 0)][osm.KEY_OSMNX_LENGTH] = lengths[(0, 1, 0)]
        network.edges[(1, 2, 0)][osm.KEY_OSMNX_LENGTH] = lengths[(1, 2, 0)]
        number_edges = network.number_of_edges()
        if project_the_graph:
            network = ox.project_graph(G=network)
        path = [0, 1, 2]
        new_edge_key = gis_mod.replace_path(network, path=path)
        # a new edge should exist
        assert network.has_edge(*new_edge_key)
        # assert that two edges are removed and one is created
        assert network.number_of_edges() - number_edges == 1 - 2
        # intermediate nodes should not exist any longer
        for node in path[1:-1]:
            assert not network.has_node(node)
        # the distances need to match
        assert isclose(
            sum(lengths.values()),
            network.edges[new_edge_key][osm.KEY_OSMNX_LENGTH],
            abs_tol=1e-3,
        )
        # the new edge needs to have a geometry because it is not simple
        assert osm.KEY_OSMNX_GEOMETRY in network.edges[new_edge_key]
        assert (
            len(tuple(network.edges[new_edge_key][osm.KEY_OSMNX_GEOMETRY].coords)) == 5
        )
        new_edge_key_lengths = gis_calc.edge_lengths(network, edge_keys=[new_edge_key])
        # the geometry's length needs to match that of the edge's
        assert isclose(
            network.edges[new_edge_key][osm.KEY_OSMNX_LENGTH],
            new_edge_key_lengths[new_edge_key],
            abs_tol=1 if not project_the_graph else 12.2,  # -12.178460200064364
        )
        # print('hallo4')
        # print(tuple(network.edges[new_edge_key][osm.KEY_OSMNX_GEOMETRY].coords))
        # print(network.edges[new_edge_key][osm.KEY_OSMNX_LENGTH])
        # print(new_edge_key_lengths[new_edge_key])

        # *********************************************************************

        # smallest possible path, but featuring reversed geometries already (#2)

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"
        network.graph["simplified"] = True
        # add nodes
        network.add_nodes_from(
            [
                (0, {osm.KEY_OSMNX_Y: 56.00, osm.KEY_OSMNX_X: 12.00}),
                (1, {osm.KEY_OSMNX_Y: 56.01, osm.KEY_OSMNX_X: 12.00}),
                (2, {osm.KEY_OSMNX_Y: 56.02, osm.KEY_OSMNX_X: 12.01}),
            ]
        )
        # add edges
        network.add_edges_from(
            [
                (
                    0,
                    1,
                    0,
                    {
                        osm.KEY_OSMNX_OSMID: 1,
                        osm.KEY_OSMNX_LENGTH: 3,
                        osm.KEY_OSMNX_REVERSED: False,
                        osm.KEY_OSMNX_ONEWAY: False,
                    },
                ),
                (
                    1,
                    2,
                    0,
                    {
                        osm.KEY_OSMNX_OSMID: 2,
                        osm.KEY_OSMNX_LENGTH: 4,
                        osm.KEY_OSMNX_REVERSED: False,
                        osm.KEY_OSMNX_ONEWAY: False,
                    },
                ),
            ]
        )
        # build create geometries
        edge01_geo = LineString([(12.00, 56.01), (11.99, 56.00), (12.00, 56.00)])
        edge12_geo = LineString([(12.00, 56.01), (12.02, 56.00), (12.01, 56.02)])
        network.edges[(0, 1, 0)][osm.KEY_OSMNX_GEOMETRY] = edge01_geo
        network.edges[(1, 2, 0)][osm.KEY_OSMNX_GEOMETRY] = edge12_geo
        network.edges[(0, 1, 0)][osm.KEY_OSMNX_REVERSED] = True
        network.edges[(1, 2, 0)][osm.KEY_OSMNX_REVERSED] = False
        lengths = gis_calc.edge_lengths(network, edge_keys=[(0, 1, 0), (1, 2, 0)])
        network.edges[(0, 1, 0)][osm.KEY_OSMNX_LENGTH] = lengths[(0, 1, 0)]
        network.edges[(1, 2, 0)][osm.KEY_OSMNX_LENGTH] = lengths[(1, 2, 0)]
        number_edges = network.number_of_edges()
        if project_the_graph:
            network = ox.project_graph(G=network)
        path = [0, 1, 2]
        new_edge_key = gis_mod.replace_path(network, path=path)
        # a new edge should exist
        assert network.has_edge(*new_edge_key)
        # assert that two edges are removed and one is created
        assert network.number_of_edges() - number_edges == 1 - 2
        # intermediate nodes should not exist any longer
        for node in path[1:-1]:
            assert not network.has_node(node)
        # the distances need to match
        assert isclose(
            sum(lengths.values()),
            network.edges[new_edge_key][osm.KEY_OSMNX_LENGTH],
            abs_tol=1e-3,
        )
        # the new edge needs to have a geometry because it is not simple
        assert osm.KEY_OSMNX_GEOMETRY in network.edges[new_edge_key]
        assert (
            len(tuple(network.edges[new_edge_key][osm.KEY_OSMNX_GEOMETRY].coords)) == 5
        )
        new_edge_key_lengths = gis_calc.edge_lengths(network, edge_keys=[new_edge_key])
        # the geometry's length needs to match that of the edge's
        assert isclose(
            network.edges[new_edge_key][osm.KEY_OSMNX_LENGTH],
            new_edge_key_lengths[new_edge_key],
            abs_tol=1 if not project_the_graph else 12.2,  # -12.178460200064364
        )
        # print('hallo5')
        # print(tuple(network.edges[new_edge_key][osm.KEY_OSMNX_GEOMETRY].coords))
        # print(network.edges[new_edge_key][osm.KEY_OSMNX_LENGTH])
        # print(new_edge_key_lengths[new_edge_key])

        # *********************************************************************

        # small path with 3 edges, but featuring reversed geometries already

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"
        network.graph["simplified"] = True
        # add nodes
        network.add_nodes_from(
            [
                (0, {osm.KEY_OSMNX_Y: 56.00, osm.KEY_OSMNX_X: 12.00}),
                (1, {osm.KEY_OSMNX_Y: 56.01, osm.KEY_OSMNX_X: 12.00}),
                (2, {osm.KEY_OSMNX_Y: 56.02, osm.KEY_OSMNX_X: 12.01}),
                (3, {osm.KEY_OSMNX_Y: 56.04, osm.KEY_OSMNX_X: 12.02}),
            ]
        )
        # add edges
        network.add_edges_from(
            [
                (
                    0,
                    1,
                    0,
                    {
                        osm.KEY_OSMNX_OSMID: 1,
                        osm.KEY_OSMNX_LENGTH: 6,
                        osm.KEY_OSMNX_REVERSED: False,
                        osm.KEY_OSMNX_ONEWAY: False,
                    },
                ),
                (
                    1,
                    2,
                    0,
                    {
                        osm.KEY_OSMNX_OSMID: 2,
                        osm.KEY_OSMNX_LENGTH: 7,
                        osm.KEY_OSMNX_REVERSED: False,
                        osm.KEY_OSMNX_ONEWAY: False,
                    },
                ),
                (
                    2,
                    3,
                    0,
                    {
                        osm.KEY_OSMNX_OSMID: 3,
                        osm.KEY_OSMNX_LENGTH: 8,
                        osm.KEY_OSMNX_REVERSED: False,
                        osm.KEY_OSMNX_ONEWAY: False,
                    },
                ),
            ]
        )
        # build create geometries
        edge01_geo = LineString([(12.00, 56.00), (11.99, 56.02), (12.00, 56.01)])
        edge12_geo = LineString([(12.00, 56.01), (12.02, 56.00), (12.01, 56.02)])
        edge23_geo = LineString([(12.01, 56.02), (12.05, 56.10), (12.02, 56.04)])
        network.edges[(0, 1, 0)][osm.KEY_OSMNX_GEOMETRY] = edge01_geo
        network.edges[(1, 2, 0)][osm.KEY_OSMNX_GEOMETRY] = edge12_geo
        network.edges[(2, 3, 0)][osm.KEY_OSMNX_GEOMETRY] = edge23_geo
        network.edges[(0, 1, 0)][osm.KEY_OSMNX_REVERSED] = False
        network.edges[(1, 2, 0)][osm.KEY_OSMNX_REVERSED] = True
        network.edges[(2, 3, 0)][osm.KEY_OSMNX_REVERSED] = False
        lengths = gis_calc.edge_lengths(
            network, edge_keys=[(0, 1, 0), (1, 2, 0), (2, 3, 0)]
        )
        network.edges[(0, 1, 0)][osm.KEY_OSMNX_LENGTH] = lengths[(0, 1, 0)]
        network.edges[(1, 2, 0)][osm.KEY_OSMNX_LENGTH] = lengths[(1, 2, 0)]
        network.edges[(2, 3, 0)][osm.KEY_OSMNX_LENGTH] = lengths[(2, 3, 0)]
        number_edges = network.number_of_edges()
        if project_the_graph:
            network = ox.project_graph(G=network)
        path = [0, 1, 2, 3]
        new_edge_key = gis_mod.replace_path(network, path=path)
        # a new edge should exist
        assert network.has_edge(*new_edge_key)
        # assert that two edges are removed and one is created
        assert network.number_of_edges() - number_edges == 1 - 3
        # intermediate nodes should not exist any longer
        for node in path[1:-1]:
            assert not network.has_node(node)
        # the distances need to match
        assert isclose(
            sum(lengths.values()),
            network.edges[new_edge_key][osm.KEY_OSMNX_LENGTH],
            abs_tol=1e-3,
        )
        # the new edge needs to have a geometry because it is not simple
        assert osm.KEY_OSMNX_GEOMETRY in network.edges[new_edge_key]
        assert (
            len(tuple(network.edges[new_edge_key][osm.KEY_OSMNX_GEOMETRY].coords)) == 7
        )
        new_edge_key_lengths = gis_calc.edge_lengths(network, edge_keys=[new_edge_key])
        # the geometry's length needs to match that of the edge's
        assert isclose(
            network.edges[new_edge_key][osm.KEY_OSMNX_LENGTH],
            new_edge_key_lengths[new_edge_key],
            abs_tol=1 if not project_the_graph else 37.77,  # -37.76434146326574
        )
        # print('hallo6')
        # print(tuple(network.edges[new_edge_key][osm.KEY_OSMNX_GEOMETRY].coords))
        # print(network.edges[new_edge_key][osm.KEY_OSMNX_LENGTH])
        # print(new_edge_key_lengths[new_edge_key])

        # *********************************************************************
        # *********************************************************************

    # *************************************************************************
    # *************************************************************************

    def test_replace_paths_unprojected(self):
        self.example_replace_paths(project_the_graph=False)

        # *********************************************************************
        # *********************************************************************

    # *************************************************************************
    # *************************************************************************

    def test_replace_paths_projected(self):
        self.example_replace_paths(project_the_graph=True)

        # *********************************************************************
        # *********************************************************************

    # *************************************************************************
    # *************************************************************************

    def test_remove_reversed_edges(self):
        # get the network
        network = ox.graph_from_point(
            (55.71654, 9.11728),
            network_type="drive",
            custom_filter='["highway"~"residential|tertiary|unclassified|service"]',
            truncate_by_edge=True,
        )

        _net = network.copy()

        reversed_attribute = False
        removed_edges = gis_mod.remove_reversed_edges(
            network=_net, reversed_attr=reversed_attribute
        )
        # verify that all edges have the correct attribute
        for edge_key in removed_edges:
            # confirm that there is at least one edge in reverse
            reverse_edge_found = False
            for other_edge_key in gis_iden.get_edges_from_a_to_b(
                network, edge_key[1], edge_key[0]
            ):
                if gis_iden.edges_are_in_reverse(
                    network, edge_key, other_edge_key
                ) and _net.has_edge(*other_edge_key):
                    # other_edge_key is a reversed edge and is still on _net
                    reverse_edge_found = True
                    break
            assert reverse_edge_found

        # *********************************************************************

        # repeat for True argument
        _net = network.copy()
        reversed_attribute = True
        removed_edges = gis_mod.remove_reversed_edges(
            network=_net, reversed_attr=reversed_attribute
        )
        # verify that all edges have the correct attribute
        for edge_key in removed_edges:
            # confirm that there is at least one edge in reverse
            reverse_edge_found = False
            for other_edge_key in gis_iden.get_edges_from_a_to_b(
                network, edge_key[1], edge_key[0]
            ):
                if gis_iden.edges_are_in_reverse(
                    network, edge_key, other_edge_key
                ) and _net.has_edge(*other_edge_key):
                    # other_edge_key is a reversed edge and is still on _net
                    reverse_edge_found = True
                    break
            assert reverse_edge_found

    # *************************************************************************
    # *************************************************************************

    def test_remove_self_loops(self):
        # find one self-loop
        network = nx.MultiDiGraph()
        network.add_edges_from([(0, 1, 0), (1, 2, 0), (2, 0, 0), (1, 1, 0)])
        true_selflooping_nodes = [1]
        selflooping_nodes = gis_mod.remove_self_loops(network)
        assert len(selflooping_nodes) == len(true_selflooping_nodes)
        for node_key in selflooping_nodes:
            assert node_key in true_selflooping_nodes

        # find two self-loops
        network = nx.MultiDiGraph()
        network.add_edges_from([(0, 1, 0), (1, 2, 0), (2, 0, 0), (1, 1, 0), (2, 2, 0)])
        true_selflooping_nodes = [1, 2]
        selflooping_nodes = gis_mod.remove_self_loops(network)
        assert len(selflooping_nodes) == len(true_selflooping_nodes)
        for node_key in selflooping_nodes:
            assert node_key in true_selflooping_nodes

        # find no self-loops
        network = nx.MultiDiGraph()
        network.add_edges_from([(0, 1, 0), (1, 2, 0), (2, 0, 0)])
        selflooping_nodes = gis_mod.remove_self_loops(network)
        assert len(selflooping_nodes) == 0

    # *************************************************************************
    # *************************************************************************

    def test_remove_longer_edges(self):
        # simple example
        network = nx.MultiDiGraph()
        network.add_edges_from(
            [
                (0, 1, 0, {"length": 3}),
                (1, 2, 0, {"length": 4}),
                (2, 0, 0, {"length": 5}),
                # additional edges
                (0, 1, 1, {"length": 4}),
                (1, 2, 1, {"length": 5}),
                (2, 0, 1, {"length": 6}),
            ]
        )
        initial_number_edges = network.number_of_edges()
        true_edges_removed = [(0, 1, 1), (1, 2, 1), (2, 0, 1)]
        edges_removed = gis_mod.remove_longer_parallel_edges(network, ignore_edge_directions=False)
        assert len(edges_removed) == len(true_edges_removed)
        for edge_key in edges_removed:
            assert edge_key in true_edges_removed
        assert network.number_of_edges() == initial_number_edges - len(edges_removed)

        # example with more than one alternative
        network = nx.MultiDiGraph()
        network.add_edges_from(
            [
                (0, 1, 0, {"length": 3}),
                (1, 2, 0, {"length": 4}),
                (2, 0, 0, {"length": 5}),
                # additional edges
                (0, 1, 1, {"length": 4}),
                (0, 1, 2, {"length": 5}),
                (0, 1, 3, {"length": 6}),
            ]
        )
        initial_number_edges = network.number_of_edges()
        true_edges_removed = [(0, 1, 1), (0, 1, 2), (0, 1, 3)]
        edges_removed = gis_mod.remove_longer_parallel_edges(network, ignore_edge_directions=False)
        assert len(edges_removed) == len(true_edges_removed)
        for edge_key in edges_removed:
            assert edge_key in true_edges_removed
        assert network.number_of_edges() == initial_number_edges - len(edges_removed)

        # example with opposite edges (that won't be removed)
        network = nx.MultiDiGraph()
        network.add_edges_from(
            [
                (0, 1, 0, {"length": 3}),
                (1, 2, 0, {"length": 4}),
                (2, 0, 0, {"length": 5}),
                # additional edges
                (0, 1, 1, {"length": 4}),
                (1, 2, 1, {"length": 5}),
                (2, 0, 1, {"length": 6}),
                # oppposite edges
                (1, 0, 0, {"length": 7}),
                (2, 1, 0, {"length": 8}),
                (0, 2, 0, {"length": 9}),
            ]
        )
        initial_number_edges = network.number_of_edges()
        true_edges_removed = [(0, 1, 1), (1, 2, 1), (2, 0, 1)]
        edges_removed = gis_mod.remove_longer_parallel_edges(network, ignore_edge_directions=False)
        assert len(edges_removed) == len(true_edges_removed)
        for edge_key in edges_removed:
            assert edge_key in true_edges_removed
        assert network.number_of_edges() == initial_number_edges - len(edges_removed)

        # example with opposite edges (that will be removed)

        network = nx.MultiDiGraph()
        network.add_edges_from(
            [
                (0, 1, 0, {"length": 3}),
                (1, 2, 0, {"length": 4}),
                (2, 0, 0, {"length": 5}),
                # additional edges
                (0, 1, 1, {"length": 4}),
                (1, 2, 1, {"length": 5}),
                (2, 0, 1, {"length": 6}),
                # oppposite edges
                (1, 0, 0, {"length": 7}),
                (2, 1, 0, {"length": 8}),
                (0, 2, 0, {"length": 9}),
            ]
        )
        initial_number_edges = network.number_of_edges()
        true_edges_removed = [
            (0, 1, 1),
            (1, 2, 1),
            (2, 0, 1),
            (1, 0, 0),
            (2, 1, 0),
            (0, 2, 0),
        ]
        edges_removed = gis_mod.remove_longer_parallel_edges(network, ignore_edge_directions=True)
        assert len(edges_removed) == len(true_edges_removed)
        for edge_key in edges_removed:
            assert edge_key in true_edges_removed
        assert network.number_of_edges() == initial_number_edges - len(edges_removed)
        
        # example using protected edges

        network = nx.MultiDiGraph()
        network.add_edges_from(
            [
                (0, 1, 0, {"length": 3}),
                (1, 2, 0, {"length": 4}),
                (2, 0, 0, {"length": 5}),
                # additional edges
                (0, 1, 1, {"length": 4}),
                (1, 2, 1, {"length": 5}),
                (2, 0, 1, {"length": 6}),
                # oppposite edges
                (1, 0, 0, {"length": 7}),
                (2, 1, 0, {"length": 8}),
                (0, 2, 0, {"length": 9}),
            ]
        )
        initial_number_edges = network.number_of_edges()
        true_edges_removed = [
            (0, 1, 1),
            (1, 2, 1),
            #(2, 0, 1), # protected
            (1, 0, 0),
            (2, 1, 0),
            (0, 2, 0),
        ]
        edges_removed = gis_mod.remove_longer_parallel_edges(
            network, 
            ignore_edge_directions=True,
            protected_edges=[(0, 1, 0), (2, 0, 1)]
            )
        assert len(edges_removed) == len(true_edges_removed)
        for edge_key in edges_removed:
            assert edge_key in true_edges_removed
        assert network.number_of_edges() == initial_number_edges - len(edges_removed)

        # test using non-integers as node keys
        network = nx.MultiDiGraph()
        network.add_edges_from(
            [
                (0, "a", 0, {"length": 3}),
                ("a", "b", 0, {"length": 4}),
                ("b", 0, 0, {"length": 5}),
                # additional edges
                (0, "a", 1, {"length": 4}),
                ("a", "b", 1, {"length": 5}),
                ("b", 0, 1, {"length": 6}),
                # oppposite edges
                ("a", 0, 0, {"length": 7}),
                ("b", "a", 0, {"length": 8}),
                (0, "b", 0, {"length": 9}),
            ]
        )
        initial_number_edges = network.number_of_edges()
        true_edges_removed = [
            (0, "a", 1),
            ("a", "b", 1),
            ("b", 0, 1),
            ("a", 0, 0),
            ("b", "a", 0),
            (0, "b", 0),
        ]
        edges_removed = gis_mod.remove_longer_parallel_edges(network, ignore_edge_directions=True)
        assert len(edges_removed) == len(true_edges_removed)
        for edge_key in edges_removed:
            assert edge_key in true_edges_removed
        assert network.number_of_edges() == initial_number_edges - len(edges_removed)

    # *************************************************************************
    # *************************************************************************

    def test_remove_dead_ends(self):
        # *********************************************************************

        # example without dead ends
        network = nx.MultiDiGraph()
        nodes_removed = gis_mod.remove_dead_ends(network)
        assert len(nodes_removed) == 0

        # *********************************************************************

        # minimal example
        network = nx.MultiDiGraph()
        network.add_edges_from(
            [
                (0, 1, 0),
                (1, 2, 0),
                (2, 0, 0),
                # removable edges
                (3, 0, 0),
                (4, 1, 0),
                (5, 2, 0),
            ]
        )
        nodes_removed = gis_mod.remove_dead_ends(network)
        true_nodes_removed = [3, 4, 5]
        for node_key in true_nodes_removed:
            assert node_key in nodes_removed
        assert len(nodes_removed) == len(true_nodes_removed)

        # *********************************************************************

        # example requiring multiple iterations

        network = nx.MultiDiGraph()

        network.add_edges_from(
            [
                (0, 1, 0),
                (1, 2, 0),
                (2, 0, 0),
                # removable edges
                (3, 0, 0),
                (4, 1, 0),
                (5, 2, 0),
                (3, 6, 0),
                (4, 7, 0),
                (5, 8, 0),
            ]
        )

        nodes_removed = gis_mod.remove_dead_ends(network, max_iterations=2)

        true_nodes_removed = [3, 4, 5, 6, 7, 8]

        for node_key in true_nodes_removed:
            assert node_key in nodes_removed

        assert len(nodes_removed) == len(true_nodes_removed)

        # *********************************************************************

        # example with self-loops

        network = nx.MultiDiGraph()

        network.add_edges_from(
            [
                (0, 1, 0),
                (1, 2, 0),
                (2, 0, 0),
                # removable edges
                (3, 0, 0),
                (4, 1, 0),
                (5, 2, 0),
                (3, 6, 0),
                (4, 7, 0),
                (5, 8, 0),
                (8, 8, 0),
                (4, 4, 0),
            ]
        )

        nodes_removed = gis_mod.remove_dead_ends(network, max_iterations=2)

        true_nodes_removed = [3, 4, 5, 6, 7, 8]

        for node_key in true_nodes_removed:
            assert node_key in nodes_removed

        assert len(nodes_removed) == len(true_nodes_removed)

        # *********************************************************************

        # example with excluded nodes

        network = nx.MultiDiGraph()

        network.add_edges_from(
            [
                (0, 1, 0),
                (1, 2, 0),
                (2, 0, 0),
                # removable edges
                (3, 0, 0),
                (3, 6, 0),  # 3 and 6 can be removed
                (4, 1, 0),
                (4, 7, 0),  # only 7 can be removed
                (5, 2, 0),
                (5, 8, 0),  # 8 cannot be removed, nor can 5
            ]
        )

        nodes_removed = gis_mod.remove_dead_ends(
            network, protected_nodes=[4, 8], max_iterations=2
        )

        true_nodes_removed = [3, 6, 7]

        for node_key in true_nodes_removed:
            assert node_key in nodes_removed

        assert len(nodes_removed) == len(true_nodes_removed)

        # *********************************************************************

        # example with forward and reverse edges

        network = nx.MultiDiGraph()

        network.add_edges_from(
            [
                (0, 1, 0),
                (1, 2, 0),
                (2, 0, 0),
                # removable edges
                (3, 0, 0),
                (0, 3, 0),
                (3, 6, 0),  # branch has forward and reverse edges
                (4, 1, 0),
                (1, 4, 0),
                (4, 7, 0),  # branch has forward and reverse edges
                (5, 2, 0),
                (5, 8, 0),  # branch not affected
            ]
        )

        nodes_removed = gis_mod.remove_dead_ends(network, max_iterations=2)

        true_nodes_removed = [3, 4, 5, 6, 7, 8]

        for node_key in true_nodes_removed:
            assert node_key in nodes_removed

        assert len(nodes_removed) == len(true_nodes_removed)

    # *************************************************************************
    # *************************************************************************

    def test_remove_dead_ends_osmnx(self):
        # test removing dead ends without restrictions

        # get the network
        network = ox.graph_from_point(
            (55.71654, 9.11728),
            network_type="drive",
            custom_filter='["highway"~"residential|tertiary|unclassified|service"]',
            truncate_by_edge=True,
        )
        # copy the network
        _net = network.copy()
        # identify all the dead end nodes
        dead_end_nodes = tuple(
            node_key
            for node_key in _net.nodes()
            if len(tuple(gis_iden.neighbours(network, node_key))) == 1
            )
        share_keeper_dead_end_nodes = 0.25
        keeper_dead_end_nodes = list(dead_end_nodes[0:round(len(dead_end_nodes)*share_keeper_dead_end_nodes)])

        max_iterations = 5

        nodes_removed = gis_mod.remove_dead_ends(
            _net, protected_nodes=keeper_dead_end_nodes, max_iterations=max_iterations
        )
        # the nodes removed cannot include keeper nodes
        for node_key in nodes_removed:
            assert node_key not in keeper_dead_end_nodes
        
        for node_key in _net.nodes():
            if node_key in keeper_dead_end_nodes:
                # dead end node that was not meant to be removed
                continue
            # any other node cannot be a dead end node
            assert node_key not in dead_end_nodes
            # ensure that they have at least two neighbours
            assert len(tuple(gis_iden.neighbours(_net, node_key))) >= 2

    # *************************************************************************
    # *************************************************************************

    def example_connect_points_to_edges_osmnx(
        self,
        _net: nx.MultiDiGraph,
        use_two_edges=False,
        store_unsimplified_geo: bool = False,
        project_network: bool = False,
    ):
        # test using projected graph

        # _net = network.copy()
        all_node_keys = tuple(_net.nodes())
        all_edge_keys = tuple(_net.edges(keys=True))
        number_edges = len(all_edge_keys)
        number_nodes = len(all_node_keys)

        # create three random nodes
        number_new_points = 3
        number_subgroup_points = (
            int(number_nodes / 10) if number_nodes >= 20 else 2
        )  # should be at least 2 points
        node_keys = []
        for i in range(number_new_points):
            # define the sampling points
            sampling_point_node_keys = [
                all_node_keys[random.randint(0, number_nodes - 1)]
                for j in range(number_subgroup_points)
            ]
            # define the x and y coordinates
            x = mean(_net.nodes[node_key]["x"] for node_key in sampling_point_node_keys)
            y = mean(_net.nodes[node_key]["y"] for node_key in sampling_point_node_keys)
            node_key = prob_utils.unused_node_key(_net)
            # add the node
            _net.add_node(node_key, x=x, y=y)
            # store the key
            node_keys.append(node_key)

        # pick random edges to which to connect each new point
        edge_keys = [
            all_edge_keys[random.randint(0, number_edges - 1)]
            for i in range(number_new_points)
        ]
        # record paths and distances
        all_paths_ab = {
            edge_key: list(
                nx.all_simple_edge_paths(
                    _net,
                    edge_key[0],
                    edge_key[1],
                    cutoff=2 + len(node_keys),  # only one extra node
                )
            )
            for edge_key in edge_keys
        }
        original_path_lengths_ab = {
            edge_key: [
                gis_calc.edge_path_length(_net, path) for path in all_paths_ab[edge_key]
            ]
            for edge_key in edge_keys
        }
        initial_number_edges = _net.number_of_edges()

        if project_network:
            _net = ox.project_graph(G=_net)

        # connect them
        mod_net, _, _, _ = gis_mod.connect_nodes_to_edges(
            _net,
            node_keys,
            edge_keys,
            store_unsimplified_geometries=store_unsimplified_geo,
            use_one_edge_per_direction=use_two_edges,
        )

        #  check the changes: all
        self.check_split_recreate_connect(
            network=mod_net,
            edge_keys=edge_keys,
            node_keys=node_keys,
            all_paths_ab=all_paths_ab,
            original_path_lengths_ab=original_path_lengths_ab,
            abs_tol=0.925,  # 0.29327665321937957, 0.9249539991553775
        )

        # there should be at least one extra edge per node

        assert mod_net.number_of_edges() >= initial_number_edges + len(node_keys)

        # *********************************************************************
        # *********************************************************************

    # *************************************************************************
    # *************************************************************************

    def test_connect_points_to_edges_osmnx_default(self):
        # get the network
        network = ox.graph_from_point(
            (55.71654, 9.11728),
            network_type="drive",
            custom_filter=('["highway"~"residential|tertiary|unclassified|service"]'),
            truncate_by_edge=True,
        )
        # find one edge and create a reversed version w/ inconsistent geometry
        # to cover a few more lines of tests in check_split_recreate_connect
        for edge_key in network.edges(keys=True):
            break
        edge_dict = network.get_edge_data(*edge_key)
        network.add_edge(edge_key[1], edge_key[0], **edge_dict)
        # try method
        self.example_connect_points_to_edges_osmnx(
            _net=network,
            use_two_edges=False,
            store_unsimplified_geo=False,
            project_network=False,
        )

    # *************************************************************************
    # *************************************************************************

    def test_connect_points_to_edges_osmnx_2edges(self):
        # get the network
        network = ox.graph_from_point(
            (55.71654, 9.11728),
            network_type="drive",
            custom_filter=('["highway"~"residential|tertiary|unclassified|service"]'),
            truncate_by_edge=True,
        )
        # find one edge and create a reversed version w/ inconsistent geometry
        # to cover a few more lines of tests in check_split_recreate_connect
        for edge_key in network.edges(keys=True):
            break
        edge_dict = network.get_edge_data(*edge_key)
        network.add_edge(edge_key[1], edge_key[0], **edge_dict)
        # try method
        self.example_connect_points_to_edges_osmnx(
            _net=network,
            use_two_edges=True,
            store_unsimplified_geo=False,
            project_network=False,
        )

    # *************************************************************************
    # *************************************************************************

    def test_connect_points_to_edges_osmnx_unsimplified(self):
        # get the network
        network = ox.graph_from_point(
            (55.71654, 9.11728),
            network_type="drive",
            custom_filter=('["highway"~"residential|tertiary|unclassified|service"]'),
            truncate_by_edge=True,
        )
        # try method
        self.example_connect_points_to_edges_osmnx(
            _net=network,
            use_two_edges=False,
            store_unsimplified_geo=True,
            project_network=False,
        )

    # *************************************************************************
    # *************************************************************************

    def test_connect_points_to_edges_osmnx_projected(self):
        # get the network
        network = ox.graph_from_point(
            (55.71654, 9.11728),
            network_type="drive",
            custom_filter=('["highway"~"residential|tertiary|unclassified|service"]'),
            truncate_by_edge=True,
        )
        # try method
        self.example_connect_points_to_edges_osmnx(
            _net=network,
            use_two_edges=False,
            store_unsimplified_geo=False,
            project_network=True,
        )

    # *************************************************************************
    # *************************************************************************

    def check_split_recreate_connect(
        self,
        network,
        edge_keys,
        node_keys,
        all_paths_ab,
        original_path_lengths_ab,
        abs_tol: float = 2e-3,
    ):
        new_paths_ab = {
            edge_key: list(
                nx.all_simple_edge_paths(
                    network,
                    edge_key[0],
                    edge_key[1],
                    cutoff=2 + len(node_keys),  # keep it as short as possible
                )
            )
            for edge_key in edge_keys
        }
        # how to check?
        # for each edge and node pair
        for edge_key, node_key in zip(edge_keys, node_keys):
            # there must be a path between the original start and end nodes
            assert nx.has_path(network, edge_key[0], edge_key[1])
            # the nodes must be connected to the start node
            assert nx.has_path(network, edge_key[0], node_key)
            # length from beginning to end must be roughly the same
            for new_path in new_paths_ab[edge_key]:
                # exclude new paths with self-loops
                no_self_loops = True
                for edge_in_path in new_path:
                    if edge_in_path[0] == edge_in_path[1]:
                        no_self_loops = False
                        break
                if not no_self_loops:
                    # there are self loops, skip path
                    continue
                if new_path in all_paths_ab[edge_key]:
                    # old path: it must have the same length (the edge is unchanged)
                    assert isclose(
                        gis_calc.edge_path_length(network, new_path),
                        original_path_lengths_ab[edge_key][
                            all_paths_ab[edge_key].index(new_path)
                        ],
                        abs_tol=abs_tol,
                    )
                else:
                    # new path:must have the same length (with intermediate points)
                    assert isclose(
                        gis_calc.edge_path_length(network, new_path),
                        original_path_lengths_ab[edge_key][
                            all_paths_ab[edge_key].index([edge_key])
                        ],
                        abs_tol=abs_tol,
                    )

        # each edge with a geometry must have the correct start and end points
        for edge_key in network.edges(keys=True):
            # if gis_mod.KEY_OSMNX_GEOMETRY
            if osm.KEY_OSMNX_GEOMETRY in network.edges[edge_key]:
                # has geometry, now check if the start/end points add up
                start_x = network.nodes[edge_key[0]][osm.KEY_OSMNX_X]
                start_y = network.nodes[edge_key[0]][osm.KEY_OSMNX_Y]
                end_x = network.nodes[edge_key[1]][osm.KEY_OSMNX_X]
                end_y = network.nodes[edge_key[1]][osm.KEY_OSMNX_Y]
                coords = tuple(network.edges[edge_key][osm.KEY_OSMNX_GEOMETRY].coords)

                if gis_iden.is_edge_consistent_with_geometry(network, edge_key):
                    # no reversed attr or not reversed
                    assert start_x == coords[0][0]
                    assert start_y == coords[0][1]
                    assert end_x == coords[-1][0]
                    assert end_y == coords[-1][1]
                else:
                    # reversed attr and reversed
                    assert start_x == coords[-1][0]
                    assert start_y == coords[-1][1]
                    assert end_x == coords[0][0]
                    assert end_y == coords[0][1]

        # *********************************************************************
        # *********************************************************************

    # *************************************************************************
    # *************************************************************************

    def test_connect_points_to_edges(self):
        # single edge, one intermediate point

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"
        line = LineString([(0, 0), (0, 1)])
        edge_key = (0, 1, 0)
        network.add_node(0, x=0, y=0)
        network.add_node(1, x=0, y=1)
        network.add_node(2, x=0.4, y=0.6)
        network.add_edge(
            *edge_key,
            geometry=line,
            length=gis_calc.great_circle_distance_along_path(line),
            undirected=False
        )
        node_keys = [2]
        edge_keys = [edge_key]
        all_paths_ab = {
            edge_key: list(nx.all_simple_edge_paths(network, edge_key[0], edge_key[1]))
            for edge_key in edge_keys
        }
        original_path_lengths_ab = {
            edge_key: [
                gis_calc.edge_path_length(network, path)
                for path in all_paths_ab[edge_key]
            ]
            for edge_key in edge_keys
        }
        initial_number_edges = network.number_of_edges()
        network, _, _, _ = gis_mod.connect_nodes_to_edges(
            network=network, node_keys=node_keys, edge_keys=edge_keys
        )

        # the original edge must no longer exist
        assert not network.has_edge(*edge_keys[0])
        # edge balance: one edge less, two replacements, plus one per node
        assert network.number_of_edges() - initial_number_edges == 2 - 1 + 1
        # make sure the edge was correctly recreated
        self.check_split_recreate_connect(
            network,
            edge_keys,
            node_keys,
            all_paths_ab,
            original_path_lengths_ab,
            abs_tol=1e-3,
        )
        # make sure the number of edges is the expected one

        # *********************************************************************

        # single edge, one point (start)

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"
        line = LineString([(0, 0), (0, 1)])
        edge_key = (0, 1, 0)
        network.add_node(0, x=0, y=0)
        network.add_node(1, x=0, y=1)
        network.add_node(2, x=0.3, y=0)
        network.add_edge(
            *edge_key,
            geometry=line,
            length=gis_calc.great_circle_distance_along_path(line),
            undirected=False
        )
        node_keys = [2]
        edge_keys = [edge_key]
        all_paths_ab = {
            edge_key: list(nx.all_simple_edge_paths(network, edge_key[0], edge_key[1]))
            for edge_key in edge_keys
        }
        original_path_lengths_ab = {
            edge_key: [
                gis_calc.edge_path_length(network, path)
                for path in all_paths_ab[edge_key]
            ]
            for edge_key in edge_keys
        }
        initial_number_edges = network.number_of_edges()
        network, _, _, _ = gis_mod.connect_nodes_to_edges(
            network=network, node_keys=node_keys, edge_keys=edge_keys
        )

        # the original edge must no longer exist
        assert network.has_edge(*edge_keys[0])
        # edge balance: one extra edge
        assert network.number_of_edges() - initial_number_edges == 1
        self.check_split_recreate_connect(
            network,
            edge_keys,
            node_keys,
            all_paths_ab,
            original_path_lengths_ab,
            abs_tol=1e-3,
        )

        # *********************************************************************

        # single edge, one point (end)

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"
        line = LineString([(0, 0), (0, 1)])
        edge_key = (0, 1, 0)
        network.add_node(0, x=0, y=0)
        network.add_node(1, x=0, y=1)
        network.add_node(2, x=0.6, y=1)
        network.add_edge(
            *edge_key,
            geometry=line,
            length=gis_calc.great_circle_distance_along_path(line),
            undirected=False
        )
        node_keys = [2]
        edge_keys = [edge_key]
        all_paths_ab = {
            edge_key: list(nx.all_simple_edge_paths(network, edge_key[0], edge_key[1]))
            for edge_key in edge_keys
        }
        original_path_lengths_ab = {
            edge_key: [
                gis_calc.edge_path_length(network, path)
                for path in all_paths_ab[edge_key]
            ]
            for edge_key in edge_keys
        }
        initial_number_edges = network.number_of_edges()
        network, _, _, _ = gis_mod.connect_nodes_to_edges(
            network=network, node_keys=node_keys, edge_keys=edge_keys
        )

        # the original edge should still exist
        assert network.has_edge(*edge_keys[0])
        # edge balance: one extra
        assert network.number_of_edges() - initial_number_edges == 1
        self.check_split_recreate_connect(
            network,
            edge_keys,
            node_keys,
            all_paths_ab,
            original_path_lengths_ab,
            abs_tol=1e-3,
        )

        # *********************************************************************

        # single edge, one point of each kind

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"
        line = LineString([(0, 0), (0, 1)])
        edge_key = (0, 1, 0)
        network.add_node(0, x=0, y=0)
        network.add_node(1, x=0, y=1)
        network.add_node(2, x=0.3, y=0)  # start
        network.add_node(3, x=0.5, y=0.5)
        network.add_node(4, x=0.7, y=1)  # end
        network.add_edge(
            *edge_key,
            geometry=line,
            length=gis_calc.great_circle_distance_along_path(line),
            undirected=False
        )
        node_keys = [2, 3, 4]
        edge_keys = [edge_key, edge_key, edge_key]
        all_paths_ab = {
            edge_key: list(nx.all_simple_edge_paths(network, edge_key[0], edge_key[1]))
            for edge_key in edge_keys
        }
        original_path_lengths_ab = {
            edge_key: [
                gis_calc.edge_path_length(network, path)
                for path in all_paths_ab[edge_key]
            ]
            for edge_key in edge_keys
        }
        initial_number_edges = network.number_of_edges()
        network, _, _, _ = gis_mod.connect_nodes_to_edges(
            network=network, node_keys=node_keys, edge_keys=edge_keys
        )

        # the original edge must no longer exist
        assert not network.has_edge(*edge_keys[0])
        # edge balance: one edge less, two replacements, plus one per node
        assert network.number_of_edges() - initial_number_edges == 2 - 1 + 3
        self.check_split_recreate_connect(
            network,
            edge_keys,
            node_keys,
            all_paths_ab,
            original_path_lengths_ab,
            abs_tol=1e-3,
        )

        # *********************************************************************

        # test multiple nodes closer to the same edge

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"
        line = LineString([(0, 0), (0, 1)])
        edge_key = (0, 1, 0)
        network.add_node(0, x=0, y=0)
        network.add_node(1, x=0, y=1)
        network.add_node(2, x=-0.3, y=0.2)
        network.add_node(3, x=0.5, y=0.5)
        network.add_node(4, x=-0.7, y=0.8)
        network.add_edge(
            *edge_key,
            geometry=line,
            length=gis_calc.great_circle_distance_along_path(line),
            undirected=False
        )
        node_keys = [2, 3, 4]
        edge_keys = [edge_key, edge_key, edge_key]
        all_paths_ab = {
            edge_key: list(nx.all_simple_edge_paths(network, edge_key[0], edge_key[1]))
            for edge_key in edge_keys
        }
        original_path_lengths_ab = {
            edge_key: [
                gis_calc.edge_path_length(network, path)
                for path in all_paths_ab[edge_key]
            ]
            for edge_key in edge_keys
        }

        initial_number_edges = network.number_of_edges()
        network, _, _, _ = gis_mod.connect_nodes_to_edges(
            network=network, node_keys=node_keys, edge_keys=edge_keys
        )

        # the original edge must no longer exist
        assert not network.has_edge(*edge_keys[0])
        # edge balance: one edge less, four replacements, plus one per node
        assert network.number_of_edges() - initial_number_edges == 4 - 1 + 3
        self.check_split_recreate_connect(
            network,
            edge_keys,
            node_keys,
            all_paths_ab,
            original_path_lengths_ab,
            abs_tol=1e-3,
        )

        # *********************************************************************

        # test edge geometries with three points

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"
        line = LineString([(0, 0), (0, 0.5), (0, 1)])
        edge_key = (0, 1, 0)
        network.add_node(0, x=0, y=0)
        network.add_node(1, x=0, y=1)
        network.add_node(2, x=0.1, y=0.25)  # closer to the first segment (split)
        network.add_node(3, x=-0.1, y=0.75)  # closer to the second segment (split)
        network.add_node(4, x=0, y=-0.1)  # closer to the start point (no split)
        network.add_node(5, x=0.1, y=0.5)  # closer to the middle point (split)
        network.add_node(6, x=0, y=1.1)  # closer to the end point (no split)
        network.add_edge(
            *edge_key,
            geometry=line,
            length=gis_calc.great_circle_distance_along_path(line),
            undirected=False
        )
        node_keys = [2, 3, 4, 5, 6]
        edge_keys = [edge_key, edge_key, edge_key, edge_key, edge_key]
        all_paths_ab = {
            edge_key: list(nx.all_simple_edge_paths(network, edge_key[0], edge_key[1]))
            for edge_key in edge_keys
        }
        original_path_lengths_ab = {
            edge_key: [
                gis_calc.edge_path_length(network, path)
                for path in all_paths_ab[edge_key]
            ]
            for edge_key in edge_keys
        }

        initial_number_edges = network.number_of_edges()
        network, _, _, _ = gis_mod.connect_nodes_to_edges(
            network=network, node_keys=node_keys, edge_keys=edge_keys
        )

        # the original edge must no longer exist
        assert not network.has_edge(*edge_keys[0])
        # make sure everything adds up
        self.check_split_recreate_connect(
            network,
            edge_keys,
            node_keys,
            all_paths_ab,
            original_path_lengths_ab,
            abs_tol=1e-3,
        )
        # edge balance: one edge less, four replacements, plus one per node
        assert network.number_of_edges() - initial_number_edges == 4 - 1 + 5

        # *********************************************************************

        # test nodes closer to the same point on an edge

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"
        line = LineString([(0, 0), (0, 0.5), (0, 1)])
        edge_key = (0, 1, 0)
        network.add_node(0, x=0, y=0)
        network.add_node(1, x=0, y=1)
        network.add_node(2, x=-0.1, y=0.5)  # closer to the middle point (split)
        network.add_node(5, x=0.1, y=0.5)  # closer to the middle point (split)
        network.add_edge(
            *edge_key,
            geometry=line,
            length=gis_calc.great_circle_distance_along_path(line),
            undirected=False
        )
        node_keys = [2, 5]
        edge_keys = [edge_key, edge_key]
        all_paths_ab = {
            edge_key: list(nx.all_simple_edge_paths(network, edge_key[0], edge_key[1]))
            for edge_key in edge_keys
        }
        original_path_lengths_ab = {
            edge_key: [
                gis_calc.edge_path_length(network, path)
                for path in all_paths_ab[edge_key]
            ]
            for edge_key in edge_keys
        }

        initial_number_edges = network.number_of_edges()
        network, _, _, _ = gis_mod.connect_nodes_to_edges(
            network=network, node_keys=node_keys, edge_keys=edge_keys
        )

        # the original edge must no longer exist
        assert not network.has_edge(*edge_keys[0])
        # make sure everything adds up
        self.check_split_recreate_connect(
            network,
            edge_keys,
            node_keys,
            all_paths_ab,
            original_path_lengths_ab,
            abs_tol=1e-3,
        )
        # edge balance: one edge less, two replacements, plus one per node
        assert network.number_of_edges() - initial_number_edges == 2 - 1 + 2

        # *********************************************************************

        # test connecting points to two different edges

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"
        line1 = LineString([(0, 0), (0, 0.5), (0, 1)])
        edge_key1 = (0, 1, 0)
        network.add_node(0, x=0, y=0)
        network.add_node(1, x=0, y=1)
        line2 = LineString([(1, 0), (1, 0.5), (1, 1)])
        edge_key2 = (2, 3, 0)
        network.add_node(2, x=1, y=0)
        network.add_node(3, x=1, y=1)
        # unconnected nodes
        network.add_node(4, x=-0.5, y=0.5)  # closer to the first edge
        network.add_node(5, x=1.5, y=0.5)  # closer to the second edge
        network.add_edge(
            *edge_key1,
            geometry=line1,
            length=gis_calc.great_circle_distance_along_path(line1),
            undirected=False
        )
        network.add_edge(
            *edge_key2,
            geometry=line2,
            length=gis_calc.great_circle_distance_along_path(line2),
            undirected=False
        )
        node_keys = [4, 5]
        edge_keys = [edge_key1, edge_key2]
        all_paths_ab = {
            edge_key: list(nx.all_simple_edge_paths(network, edge_key[0], edge_key[1]))
            for edge_key in edge_keys
        }
        original_path_lengths_ab = {
            edge_key: [
                gis_calc.edge_path_length(network, path)
                for path in all_paths_ab[edge_key]
            ]
            for edge_key in edge_keys
        }

        initial_number_edges = network.number_of_edges()
        network, _, _, _ = gis_mod.connect_nodes_to_edges(
            network=network, node_keys=node_keys, edge_keys=edge_keys
        )

        # the original edges should no longer exist
        assert not network.has_edge(*edge_keys[0])
        assert not network.has_edge(*edge_keys[1])
        # make sure everything adds up
        self.check_split_recreate_connect(
            network,
            edge_keys,
            node_keys,
            all_paths_ab,
            original_path_lengths_ab,
            abs_tol=1e-3,
        )
        # edge balance: two fewer edges, four extra, plus one per node
        assert network.number_of_edges() - initial_number_edges == 4 - 2 + 2

        # *********************************************************************

        # test connecting a point that is already connected

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"
        line1 = LineString([(0, 0), (0, 0.5), (0, 1)])
        edge_key1 = (0, 1, 0)
        network.add_node(0, x=0, y=0)
        network.add_node(1, x=0, y=1)
        network.add_edge(
            *edge_key1,
            geometry=line1,
            length=gis_calc.great_circle_distance_along_path(line1),
            undirected=False
        )
        node_keys = [1]
        edge_keys = [edge_key1]
        all_paths_ab = {
            edge_key: list(nx.all_simple_edge_paths(network, edge_key[0], edge_key[1]))
            for edge_key in edge_keys
        }
        original_path_lengths_ab = {
            edge_key: [
                gis_calc.edge_path_length(network, path)
                for path in all_paths_ab[edge_key]
            ]
            for edge_key in edge_keys
        }

        initial_number_edges = network.number_of_edges()
        network, _, _, _ = gis_mod.connect_nodes_to_edges(
            network=network, node_keys=node_keys, edge_keys=edge_keys
        )

        # the original edge still exists
        assert network.has_edge(*edge_keys[0])
        # make sure everything adds up
        self.check_split_recreate_connect(
            network,
            edge_keys,
            node_keys,
            all_paths_ab,
            original_path_lengths_ab,
            abs_tol=1e-3,
        )
        # edge balance: two fewer edges, four extra, plus one per node
        assert network.number_of_edges() - initial_number_edges == 0

    # *************************************************************************
    # *************************************************************************

    def test_recreate_edges_01(self):
        # create network

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"

        # create and add simple edge to the network

        line = LineString([(0, 0), (0, 1)])

        edge_key = (0, 1, 0)

        network.add_edge(*edge_key, geometry=line)
        network.add_node(0, x=0, y=0)
        network.add_node(1, x=0, y=1)

        # recreate the edge using an intermediate point

        connection_node_keys_per_edge, recreated_edges = gis_mod.recreate_edges(
            network=network, points={edge_key: [Point(0, 0.5)]}
        )

        # verify if the edge was recreated
        assert len(recreated_edges) == 1
        assert edge_key in recreated_edges
        # verify that the connection nodes exist
        assert len(connection_node_keys_per_edge[edge_key]) == 1
        assert network.has_node(connection_node_keys_per_edge[edge_key][0])
        # verify that the replacing edges exist
        assert network.has_edge(
            u=edge_key[0], v=connection_node_keys_per_edge[edge_key][0]
        )
        assert network.has_edge(
            u=connection_node_keys_per_edge[edge_key][0],
            v=edge_key[1],
        )
        # make sure the geometries make sense
        assert isclose(
            length(
                network.edges[
                    (edge_key[0], connection_node_keys_per_edge[edge_key][0], 0)
                ]["geometry"]
            )
            + length(
                network.edges[
                    (connection_node_keys_per_edge[edge_key][0], edge_key[1], 0)
                ]["geometry"]
            ),
            length(network.edges[edge_key]["geometry"]),
            abs_tol=1e-3,
        )
        # verify the geometries
        edge_01 = (edge_key[0], connection_node_keys_per_edge[edge_key][0], 0)
        edge_02 = (connection_node_keys_per_edge[edge_key][0], edge_key[1], 0)
        assert tuple(network.edges[edge_01]["geometry"].coords) == ((0, 0), (0, 0.5))
        assert tuple(network.edges[edge_02]["geometry"].coords) == ((0, 0.5), (0, 1))

    # *************************************************************************
    # *************************************************************************

    def test_recreate_edges_02(self):
        # test using multiple segments and points

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"

        # create and add simple edge to the network
        line = LineString([(0, 0), (0, 0.5), (0, 1)])
        edge_key = (0, 1, 0)
        network.add_edge(*edge_key, geometry=line)
        network.add_node(0, x=0, y=0)
        network.add_node(1, x=0, y=1)

        # recreate the edge using intermediate points

        connection_node_keys_per_edge, recreated_edges = gis_mod.recreate_edges(
            network=network, points={edge_key: [Point(0, 0.2), Point(0, 0.8)]}
        )

        # verify if the edge was recreated
        assert len(recreated_edges) == 1
        assert edge_key in recreated_edges
        # verify that the connection nodes exist
        assert len(connection_node_keys_per_edge[edge_key]) == 2
        assert network.has_node(connection_node_keys_per_edge[edge_key][0])
        assert network.has_node(connection_node_keys_per_edge[edge_key][1])
        # verify that the replacing edges exist
        assert network.has_edge(
            u=edge_key[0], v=connection_node_keys_per_edge[edge_key][0]
        )
        assert network.has_edge(
            u=connection_node_keys_per_edge[edge_key][0],
            v=connection_node_keys_per_edge[edge_key][1],
        )
        assert network.has_edge(
            u=connection_node_keys_per_edge[edge_key][1],
            v=edge_key[1],
        )
        # make sure the geometries make sense
        assert isclose(
            length(
                network.edges[
                    (edge_key[0], connection_node_keys_per_edge[edge_key][0], 0)
                ]["geometry"]
            )
            + length(
                network.edges[
                    (
                        connection_node_keys_per_edge[edge_key][0],
                        connection_node_keys_per_edge[edge_key][1],
                        0,
                    )
                ]["geometry"]
            )
            + length(
                network.edges[
                    (connection_node_keys_per_edge[edge_key][1], edge_key[1], 0)
                ]["geometry"]
            ),
            length(network.edges[edge_key]["geometry"]),
            abs_tol=1e-3,
        )

        # verify the geometries
        edge_01 = (edge_key[0], connection_node_keys_per_edge[edge_key][0], 0)
        edge_02 = (
            connection_node_keys_per_edge[edge_key][0],
            connection_node_keys_per_edge[edge_key][1],
            0,
        )
        edge_03 = (connection_node_keys_per_edge[edge_key][1], edge_key[1], 0)
        assert tuple(network.edges[edge_01]["geometry"].coords) == ((0, 0), (0, 0.2))
        assert tuple(network.edges[edge_02]["geometry"].coords) == (
            (0, 0.2),
            (0, 0.5),
            (0, 0.8),
        )
        assert tuple(network.edges[edge_03]["geometry"].coords) == ((0, 0.8), (0, 1))

        # *********************************************************************

    # *************************************************************************
    # *************************************************************************

    def test_recreate_edges_03(self):
        # test using equidistant points

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"

        # create and add simple edge to the network

        line = LineString([(0, 0), (0, 0.5), (0, 1)])
        edge_key = (0, 1, 0)
        network.add_edge(*edge_key, geometry=line)
        network.add_node(0, x=0, y=0)
        network.add_node(1, x=0, y=1)

        # recreate the edge using an intermediate point

        connection_node_keys_per_edge, recreated_edges = gis_mod.recreate_edges(
            network=network, points={edge_key: [Point(0, 0.25), Point(0, 0.75)]}
        )

        # verify if the edge was recreated
        assert len(recreated_edges) == 1
        assert edge_key in recreated_edges
        # verify that the connection nodes exist
        assert len(connection_node_keys_per_edge[edge_key]) == 2
        assert network.has_node(connection_node_keys_per_edge[edge_key][0])
        assert network.has_node(connection_node_keys_per_edge[edge_key][1])
        # verify that the replacing edges exist
        assert network.has_edge(
            u=edge_key[0], v=connection_node_keys_per_edge[edge_key][0]
        )
        assert network.has_edge(
            u=connection_node_keys_per_edge[edge_key][0],
            v=connection_node_keys_per_edge[edge_key][1],
        )
        assert network.has_edge(
            u=connection_node_keys_per_edge[edge_key][1],
            v=edge_key[1],
        )
        # make sure the geometries make sense
        assert isclose(
            length(
                network.edges[
                    (edge_key[0], connection_node_keys_per_edge[edge_key][0], 0)
                ]["geometry"]
            )
            + length(
                network.edges[
                    (
                        connection_node_keys_per_edge[edge_key][0],
                        connection_node_keys_per_edge[edge_key][1],
                        0,
                    )
                ]["geometry"]
            )
            + length(
                network.edges[
                    (connection_node_keys_per_edge[edge_key][1], edge_key[1], 0)
                ]["geometry"]
            ),
            length(network.edges[edge_key]["geometry"]),
            abs_tol=1e-3,
        )

        # verify the geometries
        edge_01 = (edge_key[0], connection_node_keys_per_edge[edge_key][0], 0)
        edge_02 = (
            connection_node_keys_per_edge[edge_key][0],
            connection_node_keys_per_edge[edge_key][1],
            0,
        )
        edge_03 = (connection_node_keys_per_edge[edge_key][1], edge_key[1], 0)
        assert tuple(network.edges[edge_01]["geometry"].coords) == ((0, 0), (0, 0.25))
        assert tuple(network.edges[edge_02]["geometry"].coords) == (
            (0, 0.25),
            (0, 0.5),
            (0, 0.75),
        )
        assert tuple(network.edges[edge_03]["geometry"].coords) == ((0, 0.75), (0, 1))

        # *********************************************************************

    # *************************************************************************
    # *************************************************************************

    def test_recreate_edges_04(self):
        # test using start points

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"
        line = LineString([(0, 0), (0, 1)])
        edge_key = (0, 1, 0)
        network.add_edge(*edge_key, geometry=line)
        network.add_node(0, x=0, y=0)
        network.add_node(1, x=0, y=1)
        connection_node_keys_per_edge, recreated_edges = gis_mod.recreate_edges(
            network=network, points={edge_key: [Point(0, 0)]}
        )

        # verify that the edge was not recreated
        assert len(recreated_edges) == 0  # it should not be recreated
        assert edge_key not in recreated_edges
        # verify that the connection nodes exist
        assert len(connection_node_keys_per_edge[edge_key]) == 1
        assert network.has_node(connection_node_keys_per_edge[edge_key][0])
        assert edge_key[0] == connection_node_keys_per_edge[edge_key][0]
        # verify that there are not replacing edges
        assert not network.has_edge(
            u=edge_key[0], v=connection_node_keys_per_edge[edge_key][0]
        )
        assert network.has_edge(
            u=connection_node_keys_per_edge[edge_key][0],
            v=edge_key[1],
        )

        # *********************************************************************

    # *************************************************************************
    # *************************************************************************

    def test_recreate_edges_05(self):
        # test using end points

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"
        line = LineString([(0, 0), (0, 1)])
        edge_key = (0, 1, 0)
        network.add_edge(*edge_key, geometry=line)
        network.add_node(0, x=0, y=0)
        network.add_node(1, x=0, y=1)
        connection_node_keys_per_edge, recreated_edges = gis_mod.recreate_edges(
            network=network, points={edge_key: [Point(0, 1)]}
        )

        # verify that the edge was not recreated
        assert len(recreated_edges) == 0  # it should not be recreated
        assert edge_key not in recreated_edges
        # verify that the connection nodes exist
        assert len(connection_node_keys_per_edge[edge_key]) == 1
        assert network.has_node(connection_node_keys_per_edge[edge_key][0])
        assert edge_key[1] == connection_node_keys_per_edge[edge_key][0]
        # verify that there are no replacing edges
        assert network.has_edge(
            u=edge_key[0], v=connection_node_keys_per_edge[edge_key][0]
        )
        assert not network.has_edge(
            u=connection_node_keys_per_edge[edge_key][0],
            v=edge_key[1],
        )

        # *********************************************************************

    # *************************************************************************
    # *************************************************************************

    def test_recreate_edges_06(self):
        # test using multiple start points

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"
        line = LineString([(0, 0), (0, 1)])
        edge_key = (0, 1, 0)
        network.add_edge(*edge_key, geometry=line)
        network.add_node(0, x=0, y=0)
        network.add_node(1, x=0, y=1)
        connection_node_keys_per_edge, recreated_edges = gis_mod.recreate_edges(
            network=network, points={edge_key: [Point(0, 0), Point(0, 0)]}
        )

        # verify that the edge was not recreated
        assert len(recreated_edges) == 0  # it should not be recreated
        assert edge_key not in recreated_edges
        # verify that the connection nodes exist
        assert len(connection_node_keys_per_edge[edge_key]) == 2
        # verify that both nodes exist
        assert network.has_node(connection_node_keys_per_edge[edge_key][0])
        assert network.has_node(connection_node_keys_per_edge[edge_key][1])
        # verify that both nodes are the start nodes
        assert edge_key[0] == connection_node_keys_per_edge[edge_key][0]
        assert edge_key[0] == connection_node_keys_per_edge[edge_key][1]
        # verify that there are not replacing edges
        assert not network.has_edge(
            u=edge_key[0], v=connection_node_keys_per_edge[edge_key][0]
        )
        assert not network.has_edge(
            u=edge_key[0], v=connection_node_keys_per_edge[edge_key][1]
        )
        assert network.has_edge(
            u=connection_node_keys_per_edge[edge_key][0],
            v=edge_key[1],
        )
        assert network.has_edge(
            u=connection_node_keys_per_edge[edge_key][1],
            v=edge_key[1],
        )

        # *********************************************************************

    # *************************************************************************
    # *************************************************************************

    def test_recreate_edges_07(self):
        # test using multiple end points

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"
        line = LineString([(0, 0), (0, 1)])
        edge_key = (0, 1, 0)
        network.add_edge(*edge_key, geometry=line)
        network.add_node(0, x=0, y=0)
        network.add_node(1, x=0, y=1)
        connection_node_keys_per_edge, recreated_edges = gis_mod.recreate_edges(
            network=network, points={edge_key: [Point(0, 1), Point(0, 1)]}
        )

        # verify that the edge was not recreated
        assert len(recreated_edges) == 0  # it should not be recreated
        assert edge_key not in recreated_edges
        # verify that the connection nodes exist
        assert len(connection_node_keys_per_edge[edge_key]) == 2
        assert network.has_node(connection_node_keys_per_edge[edge_key][0])
        assert network.has_node(connection_node_keys_per_edge[edge_key][1])
        assert edge_key[1] == connection_node_keys_per_edge[edge_key][0]
        assert edge_key[1] == connection_node_keys_per_edge[edge_key][1]
        # verify that there are no replacing edges
        assert network.has_edge(
            u=edge_key[0], v=connection_node_keys_per_edge[edge_key][0]
        )
        assert network.has_edge(
            u=edge_key[0], v=connection_node_keys_per_edge[edge_key][1]
        )
        assert not network.has_edge(
            u=connection_node_keys_per_edge[edge_key][0],
            v=edge_key[1],
        )
        assert not network.has_edge(
            u=connection_node_keys_per_edge[edge_key][1],
            v=edge_key[1],
        )

        # *********************************************************************

    def test_recreate_edges_08(self):
        # test using geometries that do not match the edge declaration

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"

        # create and add simple edge to the network

        line = LineString([(0, 1), (0, 0.5), (0, 0)])

        edge_key = (0, 1, 0)
        edge_dict = {
            "geometry": line,
            "reversed": False,
            "oneway": True,
            "osmid": 3,
            "length": length(line),
        }
        network.add_edge(*edge_key, **edge_dict)
        network.add_node(0, x=0, y=0)
        network.add_node(1, x=0, y=1)

        # recreate the edge using intermediate points

        connection_node_keys_per_edge, recreated_edges = gis_mod.recreate_edges(
            network=network,
            points={edge_key: [Point(0, 0), Point(0, 0.2), Point(0, 0.8), Point(0, 1)]},
        )

        # verify if the edge was recreated
        assert len(recreated_edges) == 1
        assert edge_key in recreated_edges
        # verify that the connection nodes exist
        assert len(connection_node_keys_per_edge[edge_key]) == 4
        assert network.has_node(connection_node_keys_per_edge[edge_key][0])
        assert network.has_node(connection_node_keys_per_edge[edge_key][1])
        assert network.has_node(connection_node_keys_per_edge[edge_key][2])
        assert network.has_node(connection_node_keys_per_edge[edge_key][3])
        # verify that the replacing edges exist
        # the first connection node is the start node
        assert edge_key[0] == connection_node_keys_per_edge[edge_key][0]
        # the last connection node is the end node
        assert edge_key[1] == connection_node_keys_per_edge[edge_key][3]
        # the start node connects to the second point
        assert network.has_edge(
            u=edge_key[0], v=connection_node_keys_per_edge[edge_key][1]
        )
        # the second point connects to the third
        assert network.has_edge(
            u=connection_node_keys_per_edge[edge_key][1],
            v=connection_node_keys_per_edge[edge_key][2],
        )
        # the third point connects to the end node
        assert network.has_edge(
            u=connection_node_keys_per_edge[edge_key][2], v=edge_key[1]
        )
        # make sure the geometries make sense
        assert network.edges[
            (edge_key[0], connection_node_keys_per_edge[edge_key][1], 0)
        ]["geometry"] == LineString([(0, 0), (0, 0.2)])
        assert network.edges[
            (
                connection_node_keys_per_edge[edge_key][1],
                connection_node_keys_per_edge[edge_key][2],
                0,
            )
        ]["geometry"] == LineString([(0, 0.2), (0, 0.5), (0, 0.8)])
        assert network.edges[
            (connection_node_keys_per_edge[edge_key][2], edge_key[1], 0)
        ]["geometry"] == LineString([(0, 0.8), (0, 1)])
        # make sure the lengths add up
        assert isclose(
            network.edges[(edge_key[0], connection_node_keys_per_edge[edge_key][1], 0)][
                "length"
            ],
            gis_calc.great_circle_distance_along_path(LineString([(0, 0), (0, 0.2)])),
            abs_tol=2e-3,
        )
        assert isclose(
            network.edges[
                (
                    connection_node_keys_per_edge[edge_key][1],
                    connection_node_keys_per_edge[edge_key][2],
                    0,
                )
            ]["length"],
            gis_calc.great_circle_distance_along_path(
                LineString([(0, 0.2), (0, 0.5), (0, 0.8)])
            ),
            abs_tol=2e-3,
        )
        assert isclose(
            network.edges[(connection_node_keys_per_edge[edge_key][2], edge_key[1], 0)][
                "length"
            ],
            gis_calc.great_circle_distance_along_path(LineString([(0, 0.8), (0, 1)])),
            abs_tol=2e-3,
        )
        # make sure the edges have the original 'reversed' value
        for key in osm.KEYS_OSMNX_EDGES_ESSENTIAL:
            if key == osm.KEY_OSMNX_GEOMETRY or key == osm.KEY_OSMNX_LENGTH:
                continue
            assert (
                network.edges[
                    (edge_key[0], connection_node_keys_per_edge[edge_key][1], 0)
                ][key]
                == edge_dict[key]
            )
            assert (
                network.edges[
                    (
                        connection_node_keys_per_edge[edge_key][1],
                        connection_node_keys_per_edge[edge_key][2],
                        0,
                    )
                ][key]
                == edge_dict[key]
            )
            assert (
                network.edges[
                    (connection_node_keys_per_edge[edge_key][2], edge_key[1], 0)
                ][key]
                == edge_dict[key]
            )

        # verify the geometries
        edge_01 = (edge_key[0], connection_node_keys_per_edge[edge_key][1], 0)
        edge_02 = (
            connection_node_keys_per_edge[edge_key][1],
            connection_node_keys_per_edge[edge_key][2],
            0,
        )
        edge_03 = (connection_node_keys_per_edge[edge_key][2], edge_key[1], 0)
        assert tuple(network.edges[edge_01]["geometry"].coords) == ((0, 0), (0, 0.2))
        assert tuple(network.edges[edge_02]["geometry"].coords) == (
            (0, 0.2),
            (0, 0.5),
            (0, 0.8),
        )
        assert tuple(network.edges[edge_03]["geometry"].coords) == ((0, 0.8), (0, 1))
        # *********************************************************************

    def test_recreate_edges_09(self):
        # test using reversed geometries and multiple start points

        network = nx.MultiDiGraph()
        network.graph["crs"] = "EPSG:4326"

        # create and add simple edge to the network

        line = LineString([(0, 1), (0, 0)])

        edge_key = (0, 1, 0)
        edge_dict = {
            "geometry": line,
            "reversed": True,
            "oneway": True,
            "osmid": 3,
            "length": length(line),
        }
        network.add_edge(*edge_key, **edge_dict)
        network.add_node(0, x=0, y=0)
        network.add_node(1, x=0, y=1)
        # edge_key = (0,1,1)
        # network.add_edge(
        #     *edge_key,
        #     geometry=line.reverse(), #gis_mod.reverse_linestring(line),
        #     reversed=False,
        #     oneway=False,
        #     osmid=2,
        #     length=length(line)
        #     )
        # edge_key = (0,1,0)

        # recreate the edge using an intermediate point

        connection_node_keys_per_edge, recreated_edges = gis_mod.recreate_edges(
            network=network,
            points={
                edge_key: [
                    Point(0, 0.2),
                    Point(0, 0),
                    Point(0, 0.5),
                    Point(0, 0.8),
                    Point(0, 0),
                ]
            },
        )

        # verify if the edge was recreated
        assert len(recreated_edges) == 1
        assert edge_key in recreated_edges
        # verify that the connection nodes exist
        assert len(connection_node_keys_per_edge[edge_key]) == 5
        assert network.has_node(connection_node_keys_per_edge[edge_key][0])
        assert network.has_node(connection_node_keys_per_edge[edge_key][1])
        assert network.has_node(connection_node_keys_per_edge[edge_key][2])
        assert network.has_node(connection_node_keys_per_edge[edge_key][3])
        assert network.has_node(connection_node_keys_per_edge[edge_key][4])
        # verify that the replacing edges exist
        # the second connection node is the start node
        assert edge_key[0] == connection_node_keys_per_edge[edge_key][1]
        # the last connection node is the start node too
        assert edge_key[0] == connection_node_keys_per_edge[edge_key][4]
        # the start node connects to the first point
        assert network.has_edge(
            u=edge_key[0], v=connection_node_keys_per_edge[edge_key][0]
        )
        # the first point connects to the third
        assert network.has_edge(
            u=connection_node_keys_per_edge[edge_key][0],
            v=connection_node_keys_per_edge[edge_key][2],
        )
        # the third point connects to the fourth node
        assert network.has_edge(
            u=connection_node_keys_per_edge[edge_key][2],
            v=connection_node_keys_per_edge[edge_key][3],
        )
        # the fourth point connects to the end node
        assert network.has_edge(
            u=connection_node_keys_per_edge[edge_key][3], v=edge_key[1]
        )
        # make sure the geometries make sense
        assert network.edges[
            (edge_key[0], connection_node_keys_per_edge[edge_key][0], 0)
        ]["geometry"] == LineString([(0, 0), (0, 0.2)])
        assert network.edges[
            (
                connection_node_keys_per_edge[edge_key][0],
                connection_node_keys_per_edge[edge_key][2],
                0,
            )
        ]["geometry"] == LineString([(0, 0.2), (0, 0.5)])
        assert network.edges[
            (
                connection_node_keys_per_edge[edge_key][2],
                connection_node_keys_per_edge[edge_key][3],
                0,
            )
        ]["geometry"] == LineString([(0, 0.5), (0, 0.8)])
        assert network.edges[
            (connection_node_keys_per_edge[edge_key][3], edge_key[1], 0)
        ]["geometry"] == LineString([(0, 0.8), (0, 1)])
        # make sure the lengths add up
        assert isclose(
            network.edges[(edge_key[0], connection_node_keys_per_edge[edge_key][0], 0)][
                "length"
            ],
            gis_calc.great_circle_distance_along_path(LineString([(0, 0), (0, 0.2)])),
            abs_tol=2e-3,
        )
        assert isclose(
            network.edges[
                (
                    connection_node_keys_per_edge[edge_key][0],
                    connection_node_keys_per_edge[edge_key][2],
                    0,
                )
            ]["length"],
            gis_calc.great_circle_distance_along_path(LineString([(0, 0.2), (0, 0.5)])),
            abs_tol=2e-3,
        )
        assert isclose(
            network.edges[
                (
                    connection_node_keys_per_edge[edge_key][2],
                    connection_node_keys_per_edge[edge_key][3],
                    0,
                )
            ]["length"],
            gis_calc.great_circle_distance_along_path(LineString([(0, 0.5), (0, 0.8)])),
            abs_tol=2e-3,
        )
        assert isclose(
            network.edges[(connection_node_keys_per_edge[edge_key][3], edge_key[1], 0)][
                "length"
            ],
            gis_calc.great_circle_distance_along_path(LineString([(0, 0.8), (0, 1)])),
            abs_tol=2e-3,
        )
        # verify the geometries
        edge_01 = (edge_key[0], connection_node_keys_per_edge[edge_key][0], 0)
        edge_02 = (
            connection_node_keys_per_edge[edge_key][0],
            connection_node_keys_per_edge[edge_key][2],
            0,
        )
        edge_03 = (
            connection_node_keys_per_edge[edge_key][2],
            connection_node_keys_per_edge[edge_key][3],
            0,
        )
        edge_04 = (connection_node_keys_per_edge[edge_key][3], edge_key[1], 0)
        assert tuple(network.edges[edge_01]["geometry"].coords) == ((0, 0), (0, 0.2))
        assert tuple(network.edges[edge_02]["geometry"].coords) == ((0, 0.2), (0, 0.5))
        assert tuple(network.edges[edge_03]["geometry"].coords) == ((0, 0.5), (0, 0.8))
        assert tuple(network.edges[edge_04]["geometry"].coords) == ((0, 0.8), (0, 1))

        # make sure the edges have the original 'reversed' value
        for key in osm.KEYS_OSMNX_EDGES_ESSENTIAL:
            if key == osm.KEY_OSMNX_GEOMETRY or key == osm.KEY_OSMNX_LENGTH:
                continue
            assert (
                network.edges[
                    (edge_key[0], connection_node_keys_per_edge[edge_key][0], 0)
                ][key]
                == edge_dict[key]
            )
            assert (
                network.edges[
                    (
                        connection_node_keys_per_edge[edge_key][0],
                        connection_node_keys_per_edge[edge_key][2],
                        0,
                    )
                ][key]
                == edge_dict[key]
            )
            assert (
                network.edges[
                    (
                        connection_node_keys_per_edge[edge_key][2],
                        connection_node_keys_per_edge[edge_key][3],
                        0,
                    )
                ][key]
                == edge_dict[key]
            )
            assert (
                network.edges[
                    (connection_node_keys_per_edge[edge_key][3], edge_key[1], 0)
                ][key]
                == edge_dict[key]
            )

    # *************************************************************************
    # *************************************************************************

    def test_split_linestring(self):
        # *********************************************************************
        # *********************************************************************

        # simple line split by a point

        line_coords = tuple([(0.0, 0.0), (0.0, 1.0)])
        line = LineString(line_coords)

        segments, close_to_start, close_to_end = gis_mod.split_linestring(
            line=line, points=[Point(0, 0.5)]
        )

        assert len(segments) == 2
        assert repr(tuple(segments[0].coords)) == repr(((0.0, 0.0), (0.0, 0.5)))
        assert repr(tuple(segments[1].coords)) == repr(((0.0, 0.5), (0.0, 1.0)))

        # simple line split by a point slightly off the line

        line_coords = tuple([(0.0, 0.0), (0.0, 1.0)])
        line = LineString(line_coords)

        segments, _, _ = gis_mod.split_linestring(line=line, points=[Point(0.1, 0.5)])

        assert len(segments) == 2
        assert repr(tuple(segments[0].coords)) == repr(((0.0, 0.0), (0.1, 0.5)))
        assert repr(tuple(segments[1].coords)) == repr(((0.1, 0.5), (0.0, 1.0)))

        # *********************************************************************
        # *********************************************************************

        # simple line split by two points

        line_coords = tuple([(0.0, 0.0), (0.0, 1.0), (0.0, 2.0)])
        line = LineString(line_coords)

        segments, _, _ = gis_mod.split_linestring(
            line=line, points=[Point(0, 0.5), Point(0, 1.5)]
        )

        assert len(segments) == 3
        assert repr(tuple(segments[0].coords)) == repr(((0.0, 0.0), (0.0, 0.5)))
        assert repr(tuple(segments[1].coords)) == repr(
            ((0.0, 0.5), (0.0, 1.0), (0.0, 1.5))
        )
        assert repr(tuple(segments[2].coords)) == repr(((0.0, 1.5), (0.0, 2.0)))

        # simple line split by two points slightly off the line

        line_coords = tuple([(0.0, 0.0), (0.0, 1.0), (0.0, 2.0)])
        line = LineString(line_coords)

        segments, _, _ = gis_mod.split_linestring(
            line=line, points=[Point(-0.1, 0.5), Point(0.2, 1.5)]
        )

        assert len(segments) == 3
        assert repr(tuple(segments[0].coords)) == repr(((0.0, 0.0), (-0.1, 0.5)))
        assert repr(tuple(segments[1].coords)) == repr(
            ((-0.1, 0.5), (0.0, 1.0), (0.2, 1.5))
        )
        assert repr(tuple(segments[2].coords)) == repr(((0.2, 1.5), (0.0, 2.0)))

        # *********************************************************************
        # *********************************************************************

        # simple line split by a point matching the start point

        line_coords = tuple([(0.0, 0.0), (0.0, 1.0)])
        line = LineString(line_coords)

        segments, _start, _end = gis_mod.split_linestring(
            line=line, points=[Point(0.0, 0.0)]
        )

        assert len(segments) == 0
        assert len(_start) == 1
        assert len(_end) == 0
        assert repr(_start) == repr([0])
        assert repr(_end) == repr([])

        # *********************************************************************
        # *********************************************************************

        # simple line split by a point matching the end point

        line_coords = tuple([(0.0, 0.0), (0.0, 1.0)])
        line = LineString(line_coords)

        segments, _start, _end = gis_mod.split_linestring(
            line=line, points=[Point(0.0, 1.0)]
        )

        assert len(segments) == 0
        assert len(_start) == 0
        assert len(_end) == 1
        assert repr(_start) == repr([])
        assert repr(_end) == repr([0])

        # *********************************************************************
        # *********************************************************************

        # simple line split by two points, both within the same segment (first)

        line_coords = tuple([(0.0, 0.0), (0.0, 1.0), (0.0, 2.0)])
        line = LineString(line_coords)

        segments, _, _ = gis_mod.split_linestring(
            line=line, points=[Point(0, 0.25), Point(0, 0.75)]
        )

        assert len(segments) == 3
        assert repr(tuple(segments[0].coords)) == repr(((0.0, 0.0), (0.0, 0.25)))
        assert repr(tuple(segments[1].coords)) == repr(((0.0, 0.25), (0.0, 0.75)))
        assert repr(tuple(segments[2].coords)) == repr(
            ((0.0, 0.75), (0.0, 1.0), (0.0, 2.0))
        )

        # simple line split by two points, both within the same segment (last)

        line_coords = tuple([(0.0, 0.0), (0.0, 1.0), (0.0, 2.0)])
        line = LineString(line_coords)

        segments, _, _ = gis_mod.split_linestring(
            line=line, points=[Point(0.0, 1.25), Point(0.0, 1.75)]
        )

        assert len(segments) == 3
        assert repr(tuple(segments[0].coords)) == repr(
            ((0.0, 0.0), (0.0, 1.0), (0.0, 1.25))
        )
        assert repr(tuple(segments[1].coords)) == repr(((0.0, 1.25), (0.0, 1.75)))
        assert repr(tuple(segments[2].coords)) == repr(((0.0, 1.75), (0.0, 2.0)))

        # *********************************************************************
        # *********************************************************************

        # simple line split by two points, all within the same segment (first)

        line_coords = tuple([(0.0, 0.0), (0.0, 1.0), (0.0, 2.0)])
        line = LineString(line_coords)

        segments, _, _ = gis_mod.split_linestring(
            line=line, points=[Point(0, 0.25), Point(0, 0.5), Point(0, 0.75)]
        )

        assert len(segments) == 4
        assert repr(tuple(segments[0].coords)) == repr(((0.0, 0.0), (0.0, 0.25)))
        assert repr(tuple(segments[1].coords)) == repr(((0.0, 0.25), (0.0, 0.5)))
        assert repr(tuple(segments[2].coords)) == repr(((0.0, 0.5), (0.0, 0.75)))
        assert repr(tuple(segments[3].coords)) == repr(
            ((0.0, 0.75), (0.0, 1.0), (0.0, 2.0))
        )

        # simple line split by two points, all within the same segment (last)

        line_coords = tuple([(0.0, 0.0), (0.0, 1.0), (0.0, 2.0)])
        line = LineString(line_coords)

        segments, _, _ = gis_mod.split_linestring(
            line=line, points=[Point(0.0, 1.25), Point(0.0, 1.5), Point(0.0, 1.75)]
        )

        assert len(segments) == 4
        assert repr(tuple(segments[0].coords)) == repr(
            ((0.0, 0.0), (0.0, 1.0), (0.0, 1.25))
        )
        assert repr(tuple(segments[1].coords)) == repr(((0.0, 1.25), (0.0, 1.5)))
        assert repr(tuple(segments[2].coords)) == repr(((0.0, 1.5), (0.0, 1.75)))
        assert repr(tuple(segments[3].coords)) == repr(((0.0, 1.75), (0.0, 2.0)))

        # *********************************************************************
        # *********************************************************************

        # simple line split by two points, both within the same segment (intermed.)

        line_coords = tuple([(0.0, 0.0), (0.0, 0.5), (0.0, 1.5), (0.0, 2.0)])
        line = LineString(line_coords)

        segments, _, _ = gis_mod.split_linestring(
            line=line, points=[Point(0, 0.75), Point(0, 1.25)]
        )

        assert len(segments) == 3
        assert repr(tuple(segments[0].coords)) == repr(
            ((0.0, 0.0), (0.0, 0.5), (0.0, 0.75))
        )
        assert repr(tuple(segments[1].coords)) == repr(((0.0, 0.75), (0.0, 1.25)))
        assert repr(tuple(segments[2].coords)) == repr(
            ((0.0, 1.25), (0.0, 1.5), (0.0, 2.0))
        )

        # simple line split by three points, all within the same segment (interme.)

        line_coords = tuple([(0.0, 0.0), (0.0, 0.5), (0.0, 1.5), (0.0, 2.0)])
        line = LineString(line_coords)

        segments, _, _ = gis_mod.split_linestring(
            line=line, points=[Point(0.0, 0.75), Point(0.0, 1.0), Point(0.0, 1.25)]
        )

        assert len(segments) == 4
        assert repr(tuple(segments[0].coords)) == repr(
            ((0.0, 0.0), (0.0, 0.5), (0.0, 0.75))
        )
        assert repr(tuple(segments[1].coords)) == repr(((0.0, 0.75), (0.0, 1.0)))
        assert repr(tuple(segments[2].coords)) == repr(((0.0, 1.0), (0.0, 1.25)))
        assert repr(tuple(segments[3].coords)) == repr(
            ((0.0, 1.25), (0.0, 1.5), (0.0, 2.0))
        )

        # *********************************************************************
        # *********************************************************************

    # *************************************************************************
    # *************************************************************************

    def test_merge_points_into_linestring(self):
        # *********************************************************************
        # *********************************************************************

        # simple line

        line_coords = tuple([(0.0, 0.0), (0.0, 1.0)])
        line = LineString(line_coords)

        _line, close_to_start, close_to_end = gis_mod.merge_points_into_linestring(
            line, points=[Point(0, 0.5)]
        )

        assert repr(tuple(_line.coords)) == repr(((0.0, 0.0), (0.0, 0.5), (0.0, 1.0)))
        assert repr(close_to_start) == repr([])
        assert repr(close_to_end) == repr([])

        # *********************************************************************
        # *********************************************************************

        # redundant points
        # expected result: the points are ignored

        line_coords = tuple([(1.0, 1.0), (2.0, 2.0), (3.0, 0.0)])
        line = LineString(line_coords)

        (new_line, close_to_start, close_to_end) = gis_mod.merge_points_into_linestring(
            line, points=[Point(1, 1), Point(2, 2), Point(3, 0)]
        )

        assert repr(line_coords) == repr(tuple(new_line.coords))
        assert repr(close_to_start) == repr([0])
        assert repr(close_to_end) == repr([2])

        # redundant points, different order

        (new_line, close_to_start, close_to_end) = gis_mod.merge_points_into_linestring(
            line, points=[Point(3, 0), Point(2, 2), Point(1, 1)]
        )

        assert repr(line_coords) == repr(tuple(new_line.coords))
        assert repr(close_to_start) == repr([2])
        assert repr(close_to_end) == repr([0])

        # redundant points, yet another order

        (new_line, close_to_start, close_to_end) = gis_mod.merge_points_into_linestring(
            line, points=[Point(2, 2), Point(3, 0), Point(1, 1)]
        )

        assert repr(line_coords) == repr(tuple(new_line.coords))
        assert repr(close_to_start) == repr([2])
        assert repr(close_to_end) == repr([1])

        # *********************************************************************
        # *********************************************************************

        # new points, directly on the line
        # expected result: the new points appear on the geometry

        (new_line, close_to_start, close_to_end) = gis_mod.merge_points_into_linestring(
            line,
            points=[Point(1.2, 1.2), Point(3, 0), Point(1.5, 1.5)],
        )

        assert repr(
            ((1.0, 1.0), (1.2, 1.2), (1.5, 1.5), (2.0, 2.0), (3.0, 0.0))
        ) == repr(tuple(new_line.coords))
        assert repr(close_to_start) == repr([])
        assert repr(close_to_end) == repr([1])

        # *********************************************************************
        # *********************************************************************

        # new points, extending beyond the line
        # expected result: the line is NOT extended

        (new_line, close_to_start, close_to_end) = gis_mod.merge_points_into_linestring(
            line,
            points=[Point(0.5, 0.5), Point(3, 0), Point(4, -2)],
        )

        assert repr(tuple(new_line.coords)) == repr(line_coords)
        assert repr(close_to_start) == repr([0])
        assert repr(close_to_end) == repr([1, 2])

        # *********************************************************************
        # *********************************************************************

        # new points, not on the line
        # expected result: the new points appear on the geometry

        (new_line, close_to_start, close_to_end) = gis_mod.merge_points_into_linestring(
            line,
            points=[Point(0.5, 0.5), Point(1.5, 1.75), Point(2.5, 2.5), Point(4, -2)],
        )

        assert repr(
            ((1.0, 1.0), (1.5, 1.75), (2.0, 2.0), (2.5, 2.5), (3.0, 0.0))
        ) == repr(tuple(new_line.coords))
        assert repr(close_to_start) == repr([0])
        assert repr(close_to_end) == repr([3])

        # new points, not on the line, different order
        # expected result: the new points appear on the geometry

        (new_line, close_to_start, close_to_end) = gis_mod.merge_points_into_linestring(
            line,
            points=[Point(2.5, 2.5), Point(0.5, 0.5), Point(4, -2), Point(1.5, 1.75)],
        )

        assert repr(
            ((1.0, 1.0), (1.5, 1.75), (2.0, 2.0), (2.5, 2.5), (3.0, 0.0))
        ) == repr(tuple(new_line.coords))
        assert repr(close_to_start) == repr([1])
        assert repr(close_to_end) == repr([2])

        # *********************************************************************
        # *********************************************************************

        # new points, equidistant from start and end points, preference former

        line_coords = tuple([(0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0)])
        line = LineString(line_coords)

        (new_line, close_to_start, close_to_end) = gis_mod.merge_points_into_linestring(
            line,
            points=[Point(0.5, 0.0)],
        )

        assert repr(((0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0))) == repr(
            tuple(new_line.coords)
        )
        assert repr(close_to_start) == repr([0])
        assert repr(close_to_end) == repr([])

        # new points, equidistant from start and end points, preference latter

        (new_line, close_to_start, close_to_end) = gis_mod.merge_points_into_linestring(
            line, points=[Point(0.5, 0.0)], use_start_point_equidistant=False
        )

        assert repr(((0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0))) == repr(
            tuple(new_line.coords)
        )
        assert repr(close_to_start) == repr([])
        assert repr(close_to_end) == repr([0])

        # *********************************************************************
        # *********************************************************************

        # new points, equidistant from multiple segments

        # two segments

        line_coords = tuple([(0.0, 0.0), (0.0, 1.0), (1.0, 1.0)])
        line = LineString(line_coords)

        (new_line, close_to_start, close_to_end) = gis_mod.merge_points_into_linestring(
            line,
            points=[Point(0.5, 0.5)],
        )
        assert repr(((0.0, 0.0), (0.5, 0.5), (0.0, 1.0), (1.0, 1.0))) == repr(
            tuple(new_line.coords)
        )
        assert repr(close_to_start) == repr([])
        assert repr(close_to_end) == repr([])

        # three segments

        line_coords = tuple([(0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0)])
        line = LineString(line_coords)

        (new_line, close_to_start, close_to_end) = gis_mod.merge_points_into_linestring(
            line,
            points=[Point(0.5, 0.5)],
        )
        assert repr(
            ((0.0, 0.0), (0.5, 0.5), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0))
        ) == repr(tuple(new_line.coords))
        assert repr(close_to_start) == repr([])
        assert repr(close_to_end) == repr([])

        # four segments

        line_coords = tuple(
            [(0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0)]
        )
        line = LineString(line_coords)

        (new_line, close_to_start, close_to_end) = gis_mod.merge_points_into_linestring(
            line,
            points=[Point(0.5, 0.5)],
        )
        assert repr(
            ((0.0, 0.0), (0.5, 0.5), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0))
        ) == repr(tuple(new_line.coords))
        assert repr(close_to_start) == repr([])
        assert repr(close_to_end) == repr([])

        # *********************************************************************
        # *********************************************************************

    def test_merge_points_into_linestring_wo_fixed_extremities(self):
        # simple line
        line_coords = tuple([(0.0, 0.0), (0.0, 1.0)])
        line = LineString(line_coords)

        error_raised = False
        try:
            (
                _line,
                close_to_start,
                close_to_end,
            ) = gis_mod.merge_points_into_linestring(
                line, points=[Point(0, 0.5)], fixed_extremities=False
            )
        except NotImplementedError:
            error_raised = True
        assert error_raised

    # *************************************************************************
    # *************************************************************************

    def test_modify_roundabouts_unprojected(self):
        # network should be a OSM-nx formatted graph
        network = ox.graph_from_point(
            (55.71654, 9.11728),
            network_type="drive",
            custom_filter='["highway"~"residential|tertiary|unclassified|service"]',
            truncate_by_edge=True,
        )
        # find all roundabouts
        roundabouts = gis_iden.find_roundabouts(network)
        # confirm they are roundabouts
        for roundabout in roundabouts:
            assert gis_iden.is_roundabout(network, roundabout)
        # insert fake roundabout
        roundabouts.append([0, 1, 2])
        # modify the roundabouts
        node_replacements = gis_mod.transform_roundabouts_into_crossroads(
            network, roundabouts
        )
        # make sure the fake roundabout was detected
        assert type(node_replacements[-1]) == type(None)
        # TODO: test the conversion itself

    # *************************************************************************
    # *************************************************************************

    def test_modify_roundabouts_projected(self):
        # network should be a OSM-nx formatted graph
        network = ox.graph_from_point(
            (55.71654, 9.11728),
            network_type="drive",
            custom_filter='["highway"~"residential|tertiary|unclassified|service"]',
            truncate_by_edge=True,
        )
        # project the network
        network = ox.project_graph(G=network)
        # find all roundabouts
        roundabouts = gis_iden.find_roundabouts(network)
        # confirm they are roundabouts
        for roundabout in roundabouts:
            assert gis_iden.is_roundabout(network, roundabout)
        # insert fake roundabout
        roundabouts.append([0, 1, 2])
        # modify the roundabouts
        node_replacements = gis_mod.transform_roundabouts_into_crossroads(
            network, roundabouts
        )
        # make sure the fake roundabout was detected
        assert type(node_replacements[-1]) == type(None)
        # TODO: test the conversion itself


# *****************************************************************************
# *****************************************************************************
