# imports

# standard
import sys
from ast import literal_eval
import random

# local, external

from geopandas import GeoDataFrame
from pandas import concat, MultiIndex, Series
import networkx as nx
import osmnx as ox
from shapely.geometry import Point, LineString
from shapely import intersects

# local, internal

import src.topupopt.data.gis.identify as gis_iden
import src.topupopt.data.gis.utils as gis_utils
import src.topupopt.data.gis.osm as _osm

ox.settings.use_cache = True

# *****************************************************************************
# *****************************************************************************


class TestGisUtils:
    # *************************************************************************
    # *************************************************************************

    def test_examples(self):
        # test io

        self.example_io_geodataframe(preserve_original_gdf=True, identify_columns=False)

        self.example_io_geodataframe(
            preserve_original_gdf=False, identify_columns=False
        )

        # example_io_geodataframe(
        #     preserve_original_gdf=True, identify_columns=True
        #     )

        # example_io_geodataframe(
        #     preserve_original_gdf=False, identify_columns=True
        #     )

        # TODO: handle GeoJSON files

        # example_io_geodataframe(preserve_original_gdf=True,
        #                         file_extension='.json')

        # example_io_geodataframe(preserve_original_gdf=False,
        #                         file_extension='.json')

    # *************************************************************************
    # *************************************************************************

    def test_identifying_entrances_simple_no_driveway(self):
        # no driveway, all nodes

        self.example_identify_entrances_simple_no_driveway(
            AB_right_BC_wrong=True,
            create_reversed_edges=False,
            focus_on_node_P_only=False,
        )

        self.example_identify_entrances_simple_no_driveway(
            AB_right_BC_wrong=False,
            create_reversed_edges=False,
            focus_on_node_P_only=False,
        )

        self.example_identify_entrances_simple_no_driveway(
            AB_right_BC_wrong=True,
            create_reversed_edges=True,
            focus_on_node_P_only=False,
        )

        self.example_identify_entrances_simple_no_driveway(
            AB_right_BC_wrong=False,
            create_reversed_edges=True,
            focus_on_node_P_only=False,
        )

        # no driveway, all nodes, multiple addresses per edge

        self.example_identify_entrances_simple_no_driveway(
            AB_right_BC_wrong=True,
            create_reversed_edges=False,
            focus_on_node_P_only=False,
            use_multiple_addresses=True,
        )

        self.example_identify_entrances_simple_no_driveway(
            AB_right_BC_wrong=False,
            create_reversed_edges=False,
            focus_on_node_P_only=False,
            use_multiple_addresses=True,
        )

        self.example_identify_entrances_simple_no_driveway(
            AB_right_BC_wrong=True,
            create_reversed_edges=True,
            focus_on_node_P_only=False,
            use_multiple_addresses=True,
        )

        self.example_identify_entrances_simple_no_driveway(
            AB_right_BC_wrong=False,
            create_reversed_edges=True,
            focus_on_node_P_only=False,
            use_multiple_addresses=True,
        )

        # no driveway, all nodes, revert projection

        self.example_identify_entrances_simple_no_driveway(
            AB_right_BC_wrong=False,
            create_reversed_edges=True,
            focus_on_node_P_only=False,
            revert_to_original_crs=True,
        )

        # no driveway, limited selection of nodes

        self.example_identify_entrances_simple_no_driveway(
            AB_right_BC_wrong=True,
            create_reversed_edges=False,
            focus_on_node_P_only=True,
        )

        self.example_identify_entrances_simple_no_driveway(
            AB_right_BC_wrong=False,
            create_reversed_edges=False,
            focus_on_node_P_only=True,
        )

        self.example_identify_entrances_simple_no_driveway(
            AB_right_BC_wrong=True,
            create_reversed_edges=True,
            focus_on_node_P_only=True,
        )

        self.example_identify_entrances_simple_no_driveway(
            AB_right_BC_wrong=False,
            create_reversed_edges=True,
            focus_on_node_P_only=True,
        )

        # driveway, all nodes

        self.example_identify_entrances_simple_driveway(
            AB_right_BC_wrong=True, create_reversed_edges=False
        )

        self.example_identify_entrances_simple_driveway(
            AB_right_BC_wrong=False, create_reversed_edges=False
        )

        self.example_identify_entrances_simple_driveway(
            AB_right_BC_wrong=True, create_reversed_edges=True
        )

        self.example_identify_entrances_simple_driveway(
            AB_right_BC_wrong=False, create_reversed_edges=True
        )

        # driveway, all nodes, multiple addresses per edge

        self.example_identify_entrances_simple_driveway(
            AB_right_BC_wrong=True,
            create_reversed_edges=False,
            use_multiple_addresses=True,
        )

        self.example_identify_entrances_simple_driveway(
            AB_right_BC_wrong=False,
            create_reversed_edges=False,
            use_multiple_addresses=True,
        )

        self.example_identify_entrances_simple_driveway(
            AB_right_BC_wrong=True,
            create_reversed_edges=True,
            use_multiple_addresses=True,
        )

        self.example_identify_entrances_simple_driveway(
            AB_right_BC_wrong=False,
            create_reversed_edges=True,
            use_multiple_addresses=True,
        )

        # driveway, limited selection

        self.example_identify_entrances_simple_driveway(
            AB_right_BC_wrong=True,
            create_reversed_edges=False,
            focus_on_node_P_only=True,
        )

        self.example_identify_entrances_simple_driveway(
            AB_right_BC_wrong=False,
            create_reversed_edges=False,
            focus_on_node_P_only=True,
        )

        self.example_identify_entrances_simple_driveway(
            AB_right_BC_wrong=True,
            create_reversed_edges=True,
            focus_on_node_P_only=True,
        )

        self.example_identify_entrances_simple_driveway(
            AB_right_BC_wrong=False,
            create_reversed_edges=True,
            focus_on_node_P_only=True,
        )

        # driveway variation

        self.example_identify_entrances_simple_driveway(
            AB_right_BC_wrong=True,
            create_reversed_edges=False,
            focus_on_node_P_only=False,
            BD_with_name=False,
            BD_right_address=False,
        )

        self.example_identify_entrances_simple_driveway(
            AB_right_BC_wrong=True,
            create_reversed_edges=False,
            focus_on_node_P_only=False,
            BD_with_name=True,
            BD_right_address=False,
        )

        self.example_identify_entrances_simple_driveway(
            AB_right_BC_wrong=False,
            create_reversed_edges=False,
            focus_on_node_P_only=False,
            BD_with_name=True,
            BD_right_address=False,
        )

        self.example_identify_entrances_simple_driveway(
            AB_right_BC_wrong=True,
            create_reversed_edges=False,
            focus_on_node_P_only=False,
            BD_with_name=True,
            BD_right_address=True,
        )

        # special cases

        self.example_identify_entrances_simple_special(
            create_reversed_edges=False,
            revert_to_original_crs=False,
            focus_on_node_P_only=False,
        )

        self.example_identify_entrances_simple_special(
            create_reversed_edges=True,
            revert_to_original_crs=False,
            focus_on_node_P_only=False,
        )

        self.example_identify_entrances_simple_special(
            create_reversed_edges=False,
            revert_to_original_crs=False,
            focus_on_node_P_only=True,
        )

        self.example_identify_entrances_simple_special(
            create_reversed_edges=True,
            revert_to_original_crs=False,
            focus_on_node_P_only=True,
        )

        # no matching street name in the entire network

        self.example_identify_entrances_simple_special(
            create_reversed_edges=False,
            revert_to_original_crs=False,
            focus_on_node_P_only=False,
            CE_wrong=True,
        )

        # TODO: test a case with multiple parallel edges

    # *************************************************************************
    # *************************************************************************

    def test_generating_node_containers(self):
        # test generating containers

        self.example_generate_node_container(False, False)

        self.example_generate_node_container(True, False)

        self.example_generate_node_container(False, True)

        self.example_generate_node_container(True, True)

        self.example_generate_node_container(False, False, True)

    # *************************************************************************
    # *************************************************************************

    def get_node_gdf_A(
        self,
        right_address: str = "right",
        wrong_address: str = "wrong",
        country_code: str = _osm.KEY_COUNTRY_DK,
    ):
        # 4 nodes: A, B, C and P

        # node A

        osmid_A = "A"

        xy_A = (0, 0)

        geo_A = Point(xy_A)

        node_uid_A = 5123

        address_A = None

        # node B

        osmid_B = "B"

        xy_B = (1, 0)

        geo_B = Point(xy_B)

        node_uid_B = 1844

        address_B = None

        # node C

        osmid_C = "C"

        xy_C = (2, 0)

        geo_C = Point(xy_C)

        node_uid_C = 1845

        address_C = None

        # node P

        osmid_P = "P"

        xy_P = (0.5, 1)

        geo_P = Point(xy_P)

        node_uid_P = 9475

        address_P = right_address

        # **************************************************************************

        # geodataframe: should have 'addr:street', 'osak:identifier' and index

        gdf = GeoDataFrame(
            data={
                _osm.KEY_OSM_STREET: [
                    address_A,  # A
                    address_B,  # B
                    address_C,  # C
                    address_P,
                ],  # P
                _osm.KEY_OSM_BUILDING_ENTRANCE_ID[country_code]: [
                    node_uid_A,  # A
                    node_uid_B,  # B
                    node_uid_C,  # C
                    node_uid_P,
                ],  # P
                _osm.KEY_OSMNX_ELEMENT_TYPE: ["node", "node", "node", "node"],
                _osm.KEY_OSMNX_OSMID: [osmid_A, osmid_B, osmid_C, osmid_P],
            },
            geometry=[geo_A, geo_B, geo_C, geo_P],
        )

        gdf.set_index(
            [_osm.KEY_OSMNX_ELEMENT_TYPE, _osm.KEY_OSMNX_OSMID], drop=True, inplace=True
        )

        return gdf

        # **************************************************************************

    # *****************************************************************************
    # *****************************************************************************

    def get_node_gdf_B(
        self,
        right_address: str = "right",
        wrong_address: str = "wrong",
        country_code: str = _osm.KEY_COUNTRY_DK,
    ):
        # **************************************************************************

        gdf = self.get_node_gdf_A(
            right_address=right_address,
            wrong_address=wrong_address,
            country_code=country_code,
        )

        # add another node D closer to P than A, B and C

        osmid_D = "D"

        xy_D = (0.75, 1)

        geo_D = Point(xy_D)

        node_uid_D = 8842

        # address_D = None

        gdf_node_D = GeoDataFrame(
            {
                _osm.KEY_OSMNX_GEOMETRY: [geo_D],
                _osm.KEY_OSM_BUILDING_ENTRANCE_ID[country_code]: [node_uid_D],
                # _osm.KEY_OSM_STREET: [address_D],# P
                # _osm.KEY_OSMNX_ELEMENT_TYPE: ['node'],
                # _osm.KEY_OSMNX_OSMID: [osmid_D]
            },
            # index=[('node', osmid_D)],
            index=MultiIndex.from_tuples(
                [("node", osmid_D)],
                names=[_osm.KEY_OSMNX_ELEMENT_TYPE, _osm.KEY_OSMNX_OSMID],
            ),
        )

        # **************************************************************************

        gdf = concat([gdf, gdf_node_D])

        return gdf

        # **************************************************************************

    # *****************************************************************************
    # *****************************************************************************

    def get_node_gdf_C(
        self,
        right_address: str = "right",
        wrong_address: str = "wrong",
        country_code: str = _osm.KEY_COUNTRY_DK,
    ):
        # **************************************************************************

        gdf = self.get_node_gdf_B(
            right_address=right_address,
            wrong_address=wrong_address,
            country_code=country_code,
        )

        # add another node E, east of C

        osmid_E = "E"

        xy_E = (3, 0)

        geo_E = Point(xy_E)

        node_uid_E = 9173

        # address_E = right_address

        gdf_node_E = GeoDataFrame(
            {
                _osm.KEY_OSMNX_GEOMETRY: [geo_E],
                _osm.KEY_OSM_BUILDING_ENTRANCE_ID[country_code]: [node_uid_E],
                # _osm.KEY_OSM_STREET: [address_E],
            },
            # index=[('node', osmid_E)]
            index=MultiIndex.from_tuples(
                [("node", osmid_E)],
                names=[_osm.KEY_OSMNX_ELEMENT_TYPE, _osm.KEY_OSMNX_OSMID],
            ),
        )

        # **************************************************************************

        gdf = concat([gdf, gdf_node_E])

        return gdf

        # **************************************************************************

    # *****************************************************************************
    # *****************************************************************************

    def get_network_A(
        self,
        gdf: GeoDataFrame,
        right_address: str = "right",
        wrong_address: str = "wrong",
        AB_right_BC_wrong: bool = True,
        country_code: str = _osm.KEY_COUNTRY_DK,
        use_multiple_addresses: bool = False,
    ):
        # **************************************************************************

        # create network

        (
            node_keys,
            node_data_container,
            node_key_to_gdf_index_dict,
        ) = gis_utils.prepare_node_data_from_geodataframe(
            include_geometry=True, gdf=gdf
        )

        network = nx.MultiDiGraph()

        network.graph["crs"] = "EPSG:4326"

        network.add_nodes_from(node_data_container)

        # **************************************************************************

        # two edges: AB and BC

        node_key_A = "A"
        node_key_B = "B"
        node_key_C = "C"

        # edge AB

        geo_AB = LineString(
            [
                (
                    network.nodes[node_key_A][_osm.KEY_OSMNX_X],
                    network.nodes[node_key_A][_osm.KEY_OSMNX_Y],
                ),
                (
                    network.nodes[node_key_B][_osm.KEY_OSMNX_X],
                    network.nodes[node_key_B][_osm.KEY_OSMNX_Y],
                ),
            ]
        )

        length_AB = network.nodes[node_key_A][_osm.KEY_OSMNX_GEOMETRY].distance(
            network.nodes[node_key_B][_osm.KEY_OSMNX_GEOMETRY]
        )

        network.add_edge(
            node_key_A,
            node_key_B,
            **{
                _osm.KEY_OSMNX_GEOMETRY: geo_AB,
                _osm.KEY_OSMNX_LENGTH: length_AB,
                _osm.KEY_OSMNX_NAME: (
                    ["HZ", (right_address if AB_right_BC_wrong else wrong_address)]
                    if use_multiple_addresses
                    else (right_address if AB_right_BC_wrong else wrong_address)
                ),
            }
        )

        # edge BC

        geo_BC = LineString(
            [
                (
                    network.nodes[node_key_B][_osm.KEY_OSMNX_X],
                    network.nodes[node_key_B][_osm.KEY_OSMNX_Y],
                ),
                (
                    network.nodes[node_key_C][_osm.KEY_OSMNX_X],
                    network.nodes[node_key_C][_osm.KEY_OSMNX_Y],
                ),
            ]
        )

        length_BC = network.nodes[node_key_B][_osm.KEY_OSMNX_GEOMETRY].distance(
            network.nodes[node_key_C][_osm.KEY_OSMNX_GEOMETRY]
        )

        network.add_edge(
            node_key_B,
            node_key_C,
            **{
                _osm.KEY_OSMNX_GEOMETRY: geo_BC,
                _osm.KEY_OSMNX_LENGTH: length_BC,
                _osm.KEY_OSMNX_NAME: (
                    [(wrong_address if AB_right_BC_wrong else right_address), "UQ"]
                    if use_multiple_addresses
                    else (wrong_address if AB_right_BC_wrong else right_address)
                ),
            }
        )

        # **************************************************************************

        return network, node_keys, node_key_to_gdf_index_dict

        # **************************************************************************

    # *****************************************************************************
    # *****************************************************************************

    def get_network_B(
        self,
        gdf: GeoDataFrame,
        right_address: str = "right",
        wrong_address: str = "wrong",
        AB_right_BC_wrong: bool = True,
        BD_with_name: bool = False,
        BD_right_address: bool = False,
        country_code: str = _osm.KEY_COUNTRY_DK,
        use_multiple_addresses: bool = False,
    ):
        # **************************************************************************

        # create network

        network, node_keys, node_key_to_gdf_index_dict = self.get_network_A(
            gdf=gdf,
            right_address=right_address,
            wrong_address=wrong_address,
            country_code=country_code,
            AB_right_BC_wrong=AB_right_BC_wrong,
            use_multiple_addresses=use_multiple_addresses,
        )

        # add nameless BD edge

        node_key_B = "B"
        node_key_D = "D"

        # edge AB

        geo_BD = LineString(
            [
                (
                    network.nodes[node_key_B][_osm.KEY_OSMNX_X],
                    network.nodes[node_key_B][_osm.KEY_OSMNX_Y],
                ),
                (
                    network.nodes[node_key_D][_osm.KEY_OSMNX_X],
                    network.nodes[node_key_D][_osm.KEY_OSMNX_Y],
                ),
            ]
        )

        length_BD = network.nodes[node_key_B][_osm.KEY_OSMNX_GEOMETRY].distance(
            network.nodes[node_key_D][_osm.KEY_OSMNX_GEOMETRY]
        )

        BD_dict = {
            _osm.KEY_OSMNX_GEOMETRY: geo_BD,
            _osm.KEY_OSMNX_LENGTH: length_BD,
            # _osm.KEY_OSMNX_NAME: ( # no name for BD
            #    right_address if AB_right_BC_wrong else
            #    wrong_address)
        }

        if BD_with_name:
            BD_dict[_osm.KEY_OSMNX_NAME] = (
                right_address if BD_right_address else wrong_address
            )

        network.add_edge(node_key_B, node_key_D, **BD_dict)

        # **************************************************************************

        return network, node_keys, node_key_to_gdf_index_dict

        # **************************************************************************

    # *****************************************************************************
    # *****************************************************************************

    def get_network_C(
        self,
        gdf: GeoDataFrame,
        right_address: str = "right",
        wrong_address: str = "wrong",
        country_code: str = _osm.KEY_COUNTRY_DK,
        CE_wrong: bool = False,
        use_multiple_addresses: bool = False,
    ):
        # **************************************************************************

        # create network

        network, node_keys, node_key_to_gdf_index_dict = self.get_network_B(
            gdf=gdf,
            right_address=wrong_address,
            wrong_address=wrong_address,
            country_code=country_code,
            AB_right_BC_wrong=True,
        )

        # add a CE edge with the right name

        node_key_C = "C"
        node_key_E = "E"

        # edge AB

        geo_CE = LineString(
            [
                (
                    network.nodes[node_key_C][_osm.KEY_OSMNX_X],
                    network.nodes[node_key_C][_osm.KEY_OSMNX_Y],
                ),
                (
                    network.nodes[node_key_E][_osm.KEY_OSMNX_X],
                    network.nodes[node_key_E][_osm.KEY_OSMNX_Y],
                ),
            ]
        )

        length_CE = network.nodes[node_key_C][_osm.KEY_OSMNX_GEOMETRY].distance(
            network.nodes[node_key_E][_osm.KEY_OSMNX_GEOMETRY]
        )

        network.add_edge(
            node_key_C,
            node_key_E,
            **{
                _osm.KEY_OSMNX_GEOMETRY: geo_CE,
                _osm.KEY_OSMNX_LENGTH: length_CE,
                _osm.KEY_OSMNX_NAME: (wrong_address if CE_wrong else right_address),
            }
        )

        # **************************************************************************

        return network, node_keys, node_key_to_gdf_index_dict

        # **************************************************************************

    # *****************************************************************************
    # *****************************************************************************

    def example_identify_entrances_simple_special(
        self,
        create_reversed_edges: bool = False,
        revert_to_original_crs: bool = False,
        focus_on_node_P_only: bool = False,
        CE_wrong: bool = False,
    ):
        # get problem details

        country_code = _osm.KEY_COUNTRY_DK

        gdf = self.get_node_gdf_C(country_code=country_code)

        network, node_keys, node_key_to_gdf_index_dict = self.get_network_C(
            gdf=gdf, country_code=country_code, CE_wrong=CE_wrong
        )

        # create reverse edges

        if create_reversed_edges:
            previous_edge_keys = list(edge_key for edge_key in network.edges(keys=True))

            for edge_key in previous_edge_keys:
                edge_dict = network.get_edge_data(
                    u=edge_key[0], v=edge_key[1], key=edge_key[2]
                )

                network.add_edge(
                    u_for_edge=edge_key[1], v_for_edge=edge_key[0], **edge_dict
                )

        # find out which is the closest edge

        if focus_on_node_P_only:
            nearest_edge_keys, _, _ = gis_utils.identify_building_entrance_edges(
                gdf=gdf,
                gdf_street_column=_osm.KEY_OSM_STREET,
                network=network,
                node_key_to_gdf_index_dict={"P": node_key_to_gdf_index_dict["P"]},
                crs=None,
                revert_to_original_crs=revert_to_original_crs,
            )

        else:
            nearest_edge_keys, _, _ = gis_utils.identify_building_entrance_edges(
                gdf=gdf,
                gdf_street_column=_osm.KEY_OSM_STREET,
                network=network,
                node_key_to_gdf_index_dict=node_key_to_gdf_index_dict,
                crs=None,
                revert_to_original_crs=revert_to_original_crs,
            )

        # validate the outcome

        if CE_wrong:
            # no edges with the right address, the closest edge should be selected

            assert ("B", "D", 0) == nearest_edge_keys["P"] or (
                "D",
                "B",
                0,
            ) == nearest_edge_keys["P"]

        else:
            # CE has the right address, it should be selected

            assert ("C", "E", 0) == nearest_edge_keys["P"] or (
                "E",
                "C",
                0,
            ) == nearest_edge_keys["P"]

        # **************************************************************************

    # *****************************************************************************
    # *****************************************************************************

    def example_identify_entrances_simple_driveway(
        self,
        AB_right_BC_wrong: bool = True,
        create_reversed_edges: bool = False,
        revert_to_original_crs: bool = False,
        focus_on_node_P_only: bool = False,
        BD_with_name: bool = False,
        BD_right_address: bool = False,
        use_multiple_addresses: bool = False,
    ):
        # get problem details

        country_code = _osm.KEY_COUNTRY_DK

        gdf = self.get_node_gdf_B(country_code=country_code)

        network, node_keys, node_key_to_gdf_index_dict = self.get_network_B(
            gdf=gdf,
            country_code=country_code,
            BD_with_name=BD_with_name,
            BD_right_address=BD_right_address,
            AB_right_BC_wrong=AB_right_BC_wrong,
            use_multiple_addresses=use_multiple_addresses,
        )

        # create reverse edges

        if create_reversed_edges:
            previous_edge_keys = list(edge_key for edge_key in network.edges(keys=True))

            for edge_key in previous_edge_keys:
                edge_dict = network.get_edge_data(
                    u=edge_key[0], v=edge_key[1], key=edge_key[2]
                )

                network.add_edge(
                    u_for_edge=edge_key[1], v_for_edge=edge_key[0], **edge_dict
                )

        # find out which is the closest edge

        if focus_on_node_P_only:
            nearest_edge_keys, _, _ = gis_utils.identify_building_entrance_edges(
                gdf=gdf,
                gdf_street_column=_osm.KEY_OSM_STREET,
                network=network,
                node_key_to_gdf_index_dict={"P": node_key_to_gdf_index_dict["P"]},
                crs=None,
                revert_to_original_crs=revert_to_original_crs,
            )

        else:
            nearest_edge_keys, _, _ = gis_utils.identify_building_entrance_edges(
                gdf=gdf,
                gdf_street_column=_osm.KEY_OSM_STREET,
                network=network,
                node_key_to_gdf_index_dict=node_key_to_gdf_index_dict,
                crs=None,
                revert_to_original_crs=revert_to_original_crs,
            )

        # validate the outcome

        if BD_with_name and not BD_right_address:
            if AB_right_BC_wrong:
                assert ("A", "B", 0) == nearest_edge_keys["P"] or (
                    "B",
                    "A",
                    0,
                ) == nearest_edge_keys["P"]

            else:
                assert ("B", "C", 0) == nearest_edge_keys["P"] or (
                    "C",
                    "B",
                    0,
                ) == nearest_edge_keys["P"]

        # elif BD_with_name and BD_right_address:

        #     assert (('B','D', 0) == nearest_edge_keys['P'] or
        #             ('D','B', 0) == nearest_edge_keys['P'])

        else:
            assert ("B", "D", 0) == nearest_edge_keys["P"] or (
                "D",
                "B",
                0,
            ) == nearest_edge_keys["P"]

        # **************************************************************************

    # *****************************************************************************
    # *****************************************************************************

    def example_identify_entrances_simple_no_driveway(
        self,
        AB_right_BC_wrong: bool = True,
        create_reversed_edges: bool = False,
        focus_on_node_P_only: bool = False,
        revert_to_original_crs: bool = False,
        use_multiple_addresses: bool = False,
    ):
        # get problem details

        country_code = _osm.KEY_COUNTRY_DK

        gdf = self.get_node_gdf_A(country_code=country_code)

        network, node_keys, node_key_to_gdf_index_dict = self.get_network_A(
            gdf=gdf,
            country_code=country_code,
            AB_right_BC_wrong=AB_right_BC_wrong,
            use_multiple_addresses=use_multiple_addresses,
        )

        # create reverse edges

        if create_reversed_edges:
            previous_edge_keys = list(edge_key for edge_key in network.edges(keys=True))

            for edge_key in previous_edge_keys:
                edge_dict = network.get_edge_data(
                    u=edge_key[0], v=edge_key[1], key=edge_key[2]
                )

                network.add_edge(
                    u_for_edge=edge_key[1], v_for_edge=edge_key[0], **edge_dict
                )

        # find out which is the closest edge

        if focus_on_node_P_only:
            nearest_edge_keys, _, _ = gis_utils.identify_building_entrance_edges(
                gdf=gdf,
                gdf_street_column=_osm.KEY_OSM_STREET,
                network=network,
                node_key_to_gdf_index_dict={"P": node_key_to_gdf_index_dict["P"]},
                crs=None,
                revert_to_original_crs=revert_to_original_crs,
            )

        else:
            nearest_edge_keys, _, _ = gis_utils.identify_building_entrance_edges(
                gdf=gdf,
                gdf_street_column=_osm.KEY_OSM_STREET,
                network=network,
                node_key_to_gdf_index_dict=node_key_to_gdf_index_dict,
                crs=None,
                revert_to_original_crs=revert_to_original_crs,
            )

        # validate the outcome

        if AB_right_BC_wrong:
            # the closest edge should be AB

            assert ("A", "B", 0) == nearest_edge_keys["P"]

        else:
            # the closest edge should be BC

            assert ("B", "C", 0) == nearest_edge_keys["P"]

    # *************************************************************************
    # *************************************************************************

    def test_identify_entrances_simple_no_driveway_closest(self):
        # get problem details

        country_code = _osm.KEY_COUNTRY_DK

        gdf = self.get_node_gdf_A(country_code=country_code)

        network, node_keys, node_key_to_gdf_index_dict = self.get_network_A(
            gdf=gdf, country_code=country_code
        )
        # find out which is the closest edge
        nearest_edge_keys, network = gis_iden.identify_edge_closest_to_node(
            network, node_keys=["P"]
        )

        # the closest edge should be AB

        assert ("A", "B", 0) in nearest_edge_keys

        assert len(nearest_edge_keys) == 1

    # *************************************************************************
    # *************************************************************************

    # test generating a node data container (to create nodes in a network object)

    def example_generate_node_container(
        self,
        include_geometry: bool = False,
        include_street_column: bool = False,
        use_id_as_node_key: bool = False,
    ):
        # get problem details

        country_code = _osm.KEY_COUNTRY_DK

        gdf = self.get_node_gdf_A(country_code=country_code)

        # prepare node data

        (
            node_keys,
            node_data_container,
            _,
        ) = gis_utils.prepare_node_data_from_geodataframe(
            gdf=gdf,
            node_key_column=(
                _osm.KEY_OSM_BUILDING_ENTRANCE_ID[country_code]
                if use_id_as_node_key
                else None
            ),
            include_columns=([_osm.KEY_OSM_STREET] if include_street_column else None),
            include_geometry=include_geometry,
        )

        # node key to gdf index

        node_key_to_gdf_index_dict = {
            node_key: (
                "node",
                gdf[
                    gdf[_osm.KEY_OSM_BUILDING_ENTRANCE_ID[country_code]] == node_key
                ].index[0][1]
                if use_id_as_node_key
                else node_key,
            )
            for node_key in node_keys
        }

        # add nodes to new network

        network = nx.MultiDiGraph()

        network.graph["crs"] = "EPSG:4326"

        network.add_nodes_from(node_data_container)

        # verify the data

        for node_key in node_keys:
            assert network.has_node(node_key)

            gdf_index = node_key_to_gdf_index_dict[node_key]

            assert (
                network.nodes[node_key][_osm.KEY_OSMNX_X]
                == gdf.loc[gdf_index][gis_utils.KEY_GPD_GEOMETRY].x
            )

            assert (
                network.nodes[node_key][_osm.KEY_OSMNX_Y]
                == gdf.loc[gdf_index][gis_utils.KEY_GPD_GEOMETRY].y
            )

            if include_geometry:
                assert (
                    network.nodes[node_key][_osm.KEY_OSMNX_GEOMETRY]
                    == gdf.loc[gdf_index][gis_utils.KEY_GPD_GEOMETRY]
                )

            if include_street_column:
                assert (
                    network.nodes[node_key][_osm.KEY_OSM_STREET]
                    == gdf.loc[gdf_index][_osm.KEY_OSM_STREET]
                )

    # *************************************************************************
    # *************************************************************************

    def test_node_container_errors(self):
        self.example_node_container_error()

    # trigger ValueError by using an index that differs from the osmnx-provided one

    def example_node_container_error(self, country_code: str = "dk"):
        # get problem details

        country_code = _osm.KEY_COUNTRY_DK

        gdf = self.get_node_gdf_A(country_code=country_code)

        # modify index to trigger error

        gdf.set_index(_osm.KEY_OSM_BUILDING_ENTRANCE_ID[country_code], inplace=True)

        # trigger the error

        error_raised = False
        try:
            (
                node_keys,
                node_data_container,
                _,
            ) = gis_utils.prepare_node_data_from_geodataframe(gdf=gdf)
        except ValueError:
            error_raised = True
        assert error_raised

    # *****************************************************************************
    # *****************************************************************************

    # test the counting of occurrences in a geodataframe

    def test_occurrences(self):
        gdf = GeoDataFrame(
            data={
                "column_A": [1, 2, 2, 3, 4],
                "column_B": [5.46, 5.46, 7, 7, 7.3],
                "column_C": ["a", "a", "a", "a", "a"],
                "column_D": ["a", "b", "c", "d", "e"],
                "column_E": ["hello", "goodbye", None, "hello", None],
            },
            geometry=[Point(0, 1), Point(4, 5), Point(2, 3), Point(4, 6), Point(7, 2)],
        )

        solution = {
            "column_A": {1: 1, 2: 2, 3: 1, 4: 1},
            "column_B": {5.46: 2, 7: 2, 7.3: 1},
            "column_C": {"a": 5},
            "column_D": {"a": 1, "b": 1, "c": 1, "d": 1, "e": 1},
            "column_E": {"hello": 2, "goodbye": 1, None: 2},
        }

        # all elements

        for column in solution:
            assert gis_utils.count_ocurrences(gdf, column) == solution[column]

        # specific ones

        assert gis_utils.count_ocurrences(gdf, "column_A", [1, 2]) == {1: 1, 2: 2}

        assert gis_utils.count_ocurrences(gdf, "column_A", [10]) == {10: 0}

        assert gis_utils.count_ocurrences(gdf, "column_B", [7, 7.3]) == {7: 2, 7.3: 1}

        assert gis_utils.count_ocurrences(gdf, "column_C", ["a"]) == {"a": 5}

        assert gis_utils.count_ocurrences(gdf, "column_C", ["b"]) == {"b": 0}

        assert gis_utils.count_ocurrences(gdf, "column_D", ["b"]) == {"b": 1}

        assert gis_utils.count_ocurrences(gdf, "column_E", ["hello"]) == {"hello": 2}

        assert gis_utils.count_ocurrences(gdf, "column_E", [None]) == {None: 2}

    # *****************************************************************************
    # *****************************************************************************

    # test creating osmnx-like geodataframes for nodes

    def test_create_osmnx_gdf(self):
        # method for basic gdf compliance verification

        def verify_osmnx_gdf(gdf: GeoDataFrame, extra_column_names: list = None):
            # index format

            assert type(gdf.index) == MultiIndex

            assert len(gdf.index.names) == 2

            assert _osm.KEY_OSMNX_OSMID in gdf.index.names

            assert _osm.KEY_OSMNX_ELEMENT_TYPE in gdf.index.names

            # geometry column

            assert _osm.KEY_OSMNX_GEOMETRY in gdf.columns

            # extra columns

            if type(extra_column_names) != type(None):
                for extra_column_name in extra_column_names:
                    assert extra_column_name in gdf.columns

            # the elements

            for index in gdf.index:
                # must be a node

                assert "node" in index[0]  # first position of multi-index

                # must have point geometry

                assert type(gdf.loc[index][_osm.KEY_OSMNX_GEOMETRY]) == Point

        # test gdf

        gdf_example = gis_utils.GeoDataFrame(
            {
                _osm.KEY_OSMNX_GEOMETRY: [Point(152, 546)],
            },
            index=MultiIndex.from_tuples(
                [("node", "badfnbjiadbnd")],
                names=[_osm.KEY_OSMNX_ELEMENT_TYPE, _osm.KEY_OSMNX_OSMID],
            ),
        )

        verify_osmnx_gdf(gdf_example)

        # single node

        _latitude = 23
        _longitude = -12

        gdf_single = gis_utils.create_node_geodataframe(
            longitudes=(_longitude,), latitudes=(_latitude,)
        )

        verify_osmnx_gdf(gdf_single)

        # single node, using a specific key

        mynodekey = "mynodekeyishere"

        gdf_single = gis_utils.create_node_geodataframe(
            longitudes=(_longitude,), latitudes=(_latitude,), osmids=(mynodekey,)
        )

        verify_osmnx_gdf(gdf_single)

        assert gdf_single.index[0][1] == mynodekey

        # single node, with extra columns

        gdf_single = gis_utils.create_node_geodataframe(
            longitudes=(_longitude,),
            latitudes=(_latitude,),
            osmids=(mynodekey,),
            long=(_longitude,),
            lat=(_latitude,),
        )

        verify_osmnx_gdf(gdf_single, extra_column_names=("long", "lat"))

        assert gdf_single.index[0][1] == mynodekey
        assert gdf_single.iloc[0]["long"] == _longitude
        assert gdf_single.iloc[0]["lat"] == _latitude

        # **************************************************************************

        # multiple nodes

        _latitudes = (23, 45, 73)
        _longitudes = (-12, 33, 24)

        gdf_multi = gis_utils.create_node_geodataframe(
            longitudes=_longitudes, latitudes=_latitudes
        )

        verify_osmnx_gdf(gdf_multi)

        # multiple nodes and specific keys

        _osmids = (54, "a4h4", 44.323)

        gdf_multi = gis_utils.create_node_geodataframe(
            longitudes=_longitudes, latitudes=_latitudes, osmids=_osmids
        )

        verify_osmnx_gdf(gdf_multi)

        for i in range(len(gdf_multi)):
            assert gdf_multi.index[i][1] == _osmids[i]

        # multiple nodes and extra columns

        gdf_multi = gis_utils.create_node_geodataframe(
            longitudes=_longitudes,
            latitudes=_latitudes,
            osmids=_osmids,
            long=_longitudes,
            lat=_latitudes,
        )

        verify_osmnx_gdf(gdf_multi, extra_column_names=("long", "lat"))

        for i in range(len(gdf_multi)):
            assert gdf_multi.index[i][1] == _osmids[i]
            assert gdf_multi.iloc[i]["long"] == _longitudes[i]
            assert gdf_multi.iloc[i]["lat"] == _latitudes[i]

        # **************************************************************************

        # trigger errors

        # mismatched longitudes and latitudes

        error_raised = False
        try:
            _ = gis_utils.create_node_geodataframe(
                longitudes=(_longitude, 528), latitudes=(_latitude,)
            )
        except ValueError:
            error_raised = True
        assert error_raised

        # mismatched longitudes/latitudes and osmids

        error_raised = False
        try:
            _ = gis_utils.create_node_geodataframe(
                longitudes=(_longitude, 528),
                latitudes=(_latitude, 92),
                osmids=(59, 482, 135),
            )
        except ValueError:
            error_raised = True
        assert error_raised

    # *************************************************************************
    # *************************************************************************

    # TODO: test plotting using cached data

    # *************************************************************************
    # *************************************************************************

    # test writing a GeoDataFrame with containers

    def example_io_geodataframe(
        self,
        preserve_original_gdf: bool = True,
        identify_columns: bool = False,
        file_extension: str = ".gpkg",
    ):
        # **************************************************************************
        # **************************************************************************

        filename = "tests/mygdffile" + file_extension

        # print('bing')
        # print('preserve_original_gdf:'+str(preserve_original_gdf))
        # print('identify_columns:'+str(identify_columns))
        # print('file_extension:'+str(file_extension))

        def verify_gdf_conformity(gdf, new_gdf, preserve_original_gdf):
            # verify conformity
            # print(gdf)
            # print(new_gdf)

            # for each column in the original gdf

            for column in gdf.columns:
                # assert that the column exists or that it is a merged 1 (no need)

                assert column in new_gdf.columns or gis_utils.RKW_GPKG == column

                # packed column

                if gis_utils.RKW_GPKG == column:
                    # duplicates

                    # if the original was preserved, there should be no packed col.
                    # hence, it cannot have been preserved

                    assert not preserve_original_gdf

                    # for each key in the packed column
                    # print(gdf.columns)
                    # print(gdf[column])

                    for index in gdf.index:
                        contents_dict = literal_eval(gdf.loc[(index, column)])

                        for new_gdf_column in contents_dict.keys():
                            assert new_gdf_column in new_gdf.columns
                            # print(new_gdf_column)
                            # print("......................................................")
                            # #print(index)
                            # print(gdf[column].dtype)
                            # print(new_gdf[new_gdf_column].dtype)
                            # print(contents_dict[new_gdf_column])
                            # print(new_gdf.loc[(index, new_gdf_column)])
                            # print(type(contents_dict[new_gdf_column]))
                            # print(type(new_gdf.loc[(index, new_gdf_column)]))
                            # print(repr(contents_dict[new_gdf_column]))
                            # print(repr(new_gdf.loc[(index, new_gdf_column)]))

                            if new_gdf_column in special_columns:
                                # the contents are containers: use literal_eval

                                assert repr(
                                    literal_eval(contents_dict[new_gdf_column])
                                ) == repr(new_gdf.loc[(index, new_gdf_column)])

                            else:  # the contents are not containers
                                # direct comparison
                                # TODO: reach this statement
                                assert repr(contents_dict[new_gdf_column]) == repr(
                                    new_gdf.loc[(index, new_gdf_column)]
                                )

                    continue

                # ******************************************************************
                # ******************************************************************

                # non-packed column

                for index in gdf.index:
                    if preserve_original_gdf:
                        # the original gdf has been preserved

                        # print("......................................................")
                        # print(gdf[column].dtype)
                        # print(new_gdf[column].dtype)
                        # print(gdf.loc[(index, column)])
                        # print(new_gdf.loc[(index, column)])
                        # print(type(gdf.loc[(index, column)]))
                        # print(type(new_gdf.loc[(index, column)]))
                        # print(repr(gdf.loc[(index, column)]))
                        # print(repr(new_gdf.loc[(index, column)]))

                        # the types should match

                        assert type(gdf.loc[(index, column)]) == type(
                            new_gdf.loc[(index, column)]
                        )

                        # sets require special behaviour

                        if (
                            type(gdf.loc[(index, column)]) == set
                            or column == gis_utils.KEY_GPD_GEOMETRY
                        ):
                            # sets are non-ordered:
                            # repr() may reveal different results

                            assert (
                                gdf.loc[(index, column)] == new_gdf.loc[(index, column)]
                            )

                        else:  # standard
                            assert repr(gdf.loc[(index, column)]) == repr(
                                new_gdf.loc[(index, column)]
                            )

                    else:  # the original gdf has not been preserved
                        # print("......................................................")
                        # print(gdf.columns)
                        # print(gdf[column].dtype)
                        # print(new_gdf[column].dtype)
                        # print(gdf[column].loc[index])
                        # print(new_gdf[column].loc[index])
                        # print(type(gdf[column].loc[index]))
                        # print(type(new_gdf[column].loc[index]))
                        # print(repr(gdf[column].loc[index]))
                        # print(repr(new_gdf[column].loc[index]))

                        if column == gis_utils.KEY_GPD_GEOMETRY:
                            # assert (
                            #     gdf[column].loc[index] ==
                            #     new_gdf[column].loc[index]
                            #     )

                            assert (
                                gdf.loc[(index, column)] == new_gdf.loc[(index, column)]
                            )

                        elif column in special_columns:
                            assert repr(literal_eval(gdf.loc[(index, column)])) == repr(
                                new_gdf.loc[(index, column)]
                            )

                        else:
                            assert repr(gdf.loc[(index, column)]) == repr(
                                new_gdf.loc[(index, column)]
                            )

        # **************************************************************************
        # **************************************************************************

        # TODO: test methods without specifying the columns

        # **************************************************************************
        # **************************************************************************

        # gdf object with simple index, undeclared

        gdf = GeoDataFrame(
            {
                "id": [1, 2, 3],
                "mymymy": [None, None, 1.23],
                "another_id": [53.4, 54.4, 55.4],
                "not_another_id": ["you", "they", "us"],
                "column_of_lists": [list([0, 1, 2]), list([3, 4, 5]), list([6, 7, 8])],
                "column_of_tuples": [
                    tuple([-1, -2, -3]),
                    tuple([-4, -5, -6]),
                    tuple([-7, -8, -9]),
                ],
                "column_of_sets": [
                    set([-1, -2, -3]),
                    set([-4, -5, -6]),
                    set([-7, -8, -9]),
                ],
                "column_of_dicts": [
                    {1: 34, 6: "a", 5: 46.32},
                    {"a": 575, 4: [], 3: (2, 3)},
                    {(4, 5): 3, 4: {2: 5}, 3: 4},
                ],
                "column_of_strs": [
                    "set([-1,-2,-3])",
                    "set([-4,-5,-6])",
                    "set([-7,-8,-9])",
                ],
                "another_id2": ["hello", 53.4, None],  # requires special handling
                "another_id3": [53.4, "hello", None],  # requires special handling
                "yet_another_id": [None, None, None],  # requires special handling
            },
            geometry=[
                LineString([(3, 2), (7, 7)]),
                LineString([(3, 7), (7, 2)]),
                LineString([(6, 2), (6, 6)]),
            ],
        )

        # identify the columns that require special treatment

        if identify_columns:
            # find the columns automatically

            special_columns = None  # TODO: reach this statement

        else:
            special_columns = (
                "column_of_lists",
                "column_of_tuples",
                "column_of_sets",
                "column_of_dicts",
                #'column_of_strs' # can be omitted
                "another_id2",
                "another_id3",
                "yet_another_id",
            )

            # find the columns automatically

            set_packable_columns = gis_utils.find_gpkg_packable_columns(gdf)

            # make sure the columns can be identified

            for packable_column in set_packable_columns:
                assert packable_column in special_columns

        # write file

        gis_utils.write_gdf_file(
            gdf=gdf,
            filename=filename,
            columns_to_pack=special_columns,
            preserve_original=preserve_original_gdf,
        )

        new_gdf = gis_utils.read_gdf_file(
            filename=filename, packed_columns=special_columns
        )

        # verify conformity

        verify_gdf_conformity(gdf, new_gdf, preserve_original_gdf)

        # **************************************************************************
        # **************************************************************************

        # gdf object with simple index, declared

        gdf = GeoDataFrame(
            {
                "id": [1, 2, 3],
                "column_of_lists": [list([0, 1, 2]), list([3, 4, 5]), list([6, 7, 8])],
                "column_of_tuples": [
                    tuple([-1, -2, -3]),
                    tuple([-4, -5, -6]),
                    tuple([-7, -8, -9]),
                ],
                "column_of_sets": [
                    set([-1, -2, -3]),
                    set([-4, -5, -6]),
                    set([-7, -8, -9]),
                ],
                "column_of_dicts": [
                    {1: 34, 6: "a", 5: 46.32},
                    {"a": 575, 4: [], 3: (2, 3)},
                    {(4, 5): 3, 4: {2: 5}, 3: 4},
                ],
                "column_of_strs": [
                    "set([-1,-2,-3])",
                    "set([-4,-5,-6])",
                    "set([-7,-8,-9])",
                ],
            },
            geometry=[
                LineString([(3, 2), (7, 7)]),
                LineString([(3, 7), (7, 2)]),
                LineString([(6, 2), (6, 6)]),
            ],
            index=["a", "b", "c"],  # index is declared
        )

        # identify the columns that require special treatment

        if identify_columns:
            # find the columns automatically

            special_columns = None  # TODO: reach this statement

        else:
            special_columns = (
                "column_of_lists",
                "column_of_tuples",
                "column_of_sets",
                "column_of_dicts",
                "column_of_strs",
            )

            # find the columns automatically

            set_packable_columns = gis_utils.find_gpkg_packable_columns(gdf)

            # make sure the columns can be identified

            for packable_column in set_packable_columns:
                assert packable_column in special_columns

        # write file

        gis_utils.write_gdf_file(
            gdf=gdf,
            filename=filename,
            columns_to_pack=special_columns,
            preserve_original=preserve_original_gdf,
        )

        new_gdf = gis_utils.read_gdf_file(
            filename=filename, packed_columns=special_columns, index="index"
        )  # index has to be specified

        # verify conformity

        verify_gdf_conformity(gdf, new_gdf, preserve_original_gdf)

        # **************************************************************************
        # **************************************************************************

        # gdf object with multi-index, declared

        gdf = gis_utils.GeoDataFrame(
            {
                "other_column": [1, 2, 3],
                "column_a": ["nadbnppadfb", "agasdgnp", "adfgdn"],
                "column_b": [12517.4247, 0.54673, 0.3723],
                "column_c": [(1, 2, 3, 4), (5, 6, 7, 8), (44, 1247)],
                "column_d": [{"beans": "cheese", "lollipops": "dentist"}, {}, {1: 3}],
                "column_e": [[1, 2, 3], [4.5, 4.6, 4.7], [9.0, 10.0, 11.0]],
                "column_f": [{4, 5, 6}, {5.64435, 0.7545, 1.4634}, {"a", "b", "c"}],
                "geometry": [Point(12, 55), Point(2, 4), Point(3, 6)],
            },
            index=MultiIndex.from_tuples(
                [("a", 124), ("b", 754), ("c", 234)], names=["index1", "index2"]
            ),
        )

        # identify the columns that require special treatment

        if identify_columns:
            # find the columns automatically

            special_columns = None  # TODO: reach this statement

        else:
            special_columns = ("column_c", "column_d", "column_e", "column_f")

            # find the columns automatically

            set_packable_columns = gis_utils.find_gpkg_packable_columns(gdf)

            # make sure the columns can be identified

            for packable_column in set_packable_columns:
                assert packable_column in special_columns

        # write file

        gis_utils.write_gdf_file(
            gdf=gdf,
            filename=filename,
            columns_to_pack=special_columns,
            preserve_original=preserve_original_gdf,
        )

        new_gdf = gis_utils.read_gdf_file(
            filename=filename,
            packed_columns=special_columns,
            index=["index1", "index2"],
        )

        # verify conformity

        verify_gdf_conformity(gdf, new_gdf, preserve_original_gdf)

        # **************************************************************************
        # **************************************************************************

        # gdf with column names matching in lower case (not good for .gpkg files)

        gdf = GeoDataFrame(
            {
                "id": [1, 2, 3],
                "mymymy": [901.1, 53.4, None],
                "another_id": [53.4, 54.4, 55.4],
                "not_another_id": ["you", "they", "us"],
                "abc": [list([0, 1, 2]), list([3, 4, 5]), list([6, 7, 8])],
                "ABC": [tuple([-1, -2, -3]), tuple([-4, -5, -6]), tuple([-7, -8, -9])],
                "Abc": ["here", "there", "nowhere"],
                "aBc": [(1, 2, 3, 4), [5, 6, 7, 8], {9, 10, 11, 12}],
                "aBC": [53.643, {3: 6, 7: "goodbye"}, None],
                "ABc": [None, None, None],
                "mymymy2": ["hello", 53.4, None],  # requires special handling
                "yet_another_id": [None, None, None],  # requires special handling
            },
            geometry=[
                LineString([(3, 2), (7, 7)]),
                LineString([(3, 7), (7, 2)]),
                LineString([(6, 2), (6, 6)]),
            ],
        )

        # identify the columns that require special treatment

        if identify_columns:
            # find the columns automatically

            special_columns = None  # TODO: reach this statement

        else:
            special_columns = (
                "abc",
                "ABC",
                "Abc",  # no containers but has the same lowercase name as others
                "aBc",
                "aBC",
                "ABc",
                # special cases
                "mymymy2",
                "yet_another_id",
            )

            # find the columns automatically

            set_packable_columns = gis_utils.find_gpkg_packable_columns(gdf)

            # make sure the columns can be identified

            for packable_column in set_packable_columns:
                assert packable_column in special_columns

        # write file

        gis_utils.write_gdf_file(
            gdf=gdf,
            filename=filename,
            columns_to_pack=special_columns,
            preserve_original=preserve_original_gdf,
        )

        new_gdf = gis_utils.read_gdf_file(
            filename=filename, packed_columns=special_columns
        )

        # verify conformity

        verify_gdf_conformity(gdf, new_gdf, preserve_original_gdf)

        # **************************************************************************
        # **************************************************************************

        # TODO: force the methods to throw errors with non-primitive types

        # *********************************************************************
        # *********************************************************************

    # *************************************************************************
    # *************************************************************************

    def test_gpkg_write_errors(self):
        filename_gpkg: str = "test.gpkg"

        type_status = {
            int: True,
            str: True,
            float: True,
            bytes: False,
            dict: True,  # works but comes out incorrectly
            set: False,
            tuple: False,
            list: False,
            type(None): True,  # works but comes out incorrectly
        }

        for a_type, a_status in type_status.items():
            if a_type == int:
                data = [1, 2]

            elif a_type == str:
                data = ["hello", "goodbye"]

            elif a_type == float:
                data = [3.4, 6.7]

            elif a_type == bytes:
                data = [b"\x04\x00", b"\x00\x00\x00\x00\x00\x00\x00\x00\x04\x00"]

            elif a_type == dict:
                data = [{0: 1}, {"a": 53.46}]

            elif a_type == set:
                data = [{0, 1}, {"a", 53.46}]

            elif a_type == tuple:
                data = [(0, 1), ("a", 53.46)]

            elif a_type == list:
                data = [list((0, 1)), list(("a", 53.46))]

            elif a_type == type(None):
                data = [None, None]

            # create gdf

            gdf = GeoDataFrame(
                {
                    "data": data,
                    "geometry": [Point(1, 2), Point(3, 4)],
                },
                index=["a", "b"],
            )

            # verify the status

            if a_status:
                # compatible: no errors are expected

                gdf.to_file(filename_gpkg)

            else:  # incompatible: errors are expected
                error_raised = False
                try:
                    gdf.to_file(filename_gpkg)
                except Exception:
                    error_raised = True
                assert error_raised

        # *********************************************************************
        # *********************************************************************

    # *************************************************************************
    # *************************************************************************

    def test_discrete_plot_gdf(self):
        # *********************************************************************
        # *********************************************************************

        G = ox.graph_from_point(
            (55.71654, 9.11728),
            network_type="drive",
            custom_filter='["highway"~"residential|tertiary|unclassified|service"]',
            truncate_by_edge=True,
        )

        # *********************************************************************
        # *********************************************************************

        gdf = ox.utils_graph.graph_to_gdfs(G, edges=False)  # nodes only

        # *********************************************************************
        # *********************************************************************

        # add add random discrete element to gdf

        column = "discrete_category_column"

        offset = random.randint(0, int(1e3))

        number_options = 10

        category = {
            idx: random.randint(0, number_options - 1) + offset for idx in gdf.index
        }

        set_categories = set(category.values())

        category_to_label = {cat: "label for " + str(cat) for cat in set_categories}

        # create column

        gdf[column] = Series(data=category, index=gdf.index)

        # *********************************************************************
        # *********************************************************************

        gis_utils.plot_discrete_attributes(
            gdf, column=column, category_to_label=category_to_label
        )

        # *********************************************************************
        # *********************************************************************

    # *************************************************************************
    # *************************************************************************

    def test_convert_edge_path(self):
        # create network

        network = nx.MultiDiGraph()

        # define and add edges

        list_edges = [
            (0, 1),
            (1, 1),
            (1, 2),
            (2, 3),
            (3, 4),
            (4, 5),
            (5, 6),
            (7, 6),
            (8, 7),
            (9, 8),
            (6, 7),
            (7, 8),
            (8, 9),
        ]

        network.add_edges_from(list_edges)

        # *********************************************************************
        # *********************************************************************

        allow_reversed_edges = False

        edge_paths = [
            [(0, 1)],
            [(0, 1), (1, 2)],
            [(1, 2), (2, 3)],
            [(0, 1), (1, 2), (2, 3)],
            [(0, 1), (1, 1), (1, 2)],  # self loop
            [(6, 7), (7, 8), (8, 9)],  # should work
        ]

        expected_node_paths = [
            [0, 1],
            [0, 1, 2],
            [1, 2, 3],
            [0, 1, 2, 3],
            [0, 1, 2],
            [6, 7, 8, 9],
        ]

        for edge_index, edge_path in enumerate(edge_paths):
            assert (
                gis_utils.convert_edge_path(
                    network, edge_path, allow_reversed_edges=allow_reversed_edges
                )
                == expected_node_paths[edge_index]
            )

        # invalid edge paths

        invalid_edge_paths = [
            [(7, 6), (8, 7), (9, 8)],  # all reversed, should fail
            [(7, 6), (7, 8), (8, 9)],  # first reversed, should fail
            [(6, 7), (8, 7), (8, 9)],  # second reversed, should fail
            [(6, 7), (7, 8), (9, 8)],  # third reversed, should fail
        ]
        for edge_path in invalid_edge_paths:
            error_raised = False
            try:
                gis_utils.convert_edge_path(
                    network, edge_path, allow_reversed_edges=allow_reversed_edges
                )
            except ValueError:
                error_raised = True
            assert error_raised

        # *********************************************************************
        # *********************************************************************

        allow_reversed_edges = True

        edge_paths = [
            [(0, 1)],
            [(0, 1), (1, 2)],
            [(1, 2), (2, 3)],
            [(0, 1), (1, 2), (2, 3)],
            [(0, 1), (1, 1), (1, 2)],  # self loop
            [(6, 7), (7, 8), (8, 9)],  # should work
            [(7, 6), (8, 7), (9, 8)],  # all reversed, should fail
            [(7, 6), (7, 8), (8, 9)],  # first reversed, should fail
            [(6, 7), (8, 7), (8, 9)],  # second reversed, should fail
            [(6, 7), (7, 8), (9, 8)],  # third reversed, should fail
        ]

        expected_node_paths = [
            [0, 1],
            [0, 1, 2],
            [1, 2, 3],
            [0, 1, 2, 3],
            [0, 1, 2],
            [6, 7, 8, 9],
            [6, 7, 8, 9],
            [6, 7, 8, 9],
            [6, 7, 8, 9],
            [6, 7, 8, 9],
        ]

        for edge_index, edge_path in enumerate(edge_paths):
            assert (
                gis_utils.convert_edge_path(
                    network, edge_path, allow_reversed_edges=allow_reversed_edges
                )
                == expected_node_paths[edge_index]
            )

    # *************************************************************************
    # *************************************************************************

    def test_get_directed(self):
        # get a network

        network = ox.graph_from_point(
            (55.71654, 9.11728),
            network_type="drive",
            custom_filter='["highway"~"residential|tertiary|unclassified|service"]',
            truncate_by_edge=True,
        )

        # convert to undirected
        undirected_network = ox.get_undirected(network)

        # convert to directed
        directed_network = gis_utils.get_directed(undirected_network)

        # make sure the same nodes exist on both objects
        for node_key in network.nodes():
            assert node_key in directed_network.nodes()
        assert network.number_of_nodes() == directed_network.number_of_nodes()

        # assert that all edges on the directed network exist on the undirected one
        assert network.number_of_edges() >= directed_network.number_of_edges()
        # for each edge in the directed graph object
        for edge_key in directed_network.edges(keys=True):
            # make sure at least one suitable edge exists
            assert network.has_edge(*edge_key)
            # cycle through suitable edges on the original network until the one is
            # found that has all the matching attributes and content
            edge_dict = directed_network.edges[edge_key]
            for other_edge_key in gis_iden.get_edges_from_a_to_b(
                network, edge_key[0], edge_key[1]
            ):
                # check all attributes
                number_matching_attributes = 0
                for edge_attr, edge_data in edge_dict.items():
                    try:
                        assert edge_data == network.edges[other_edge_key][edge_attr]
                    except AssertionError:
                        # the data is different
                        continue
                    except KeyError:
                        # the attribute does not exist
                        continue
                    number_matching_attributes += 1

                if number_matching_attributes == len(edge_dict):
                    # a compatible edge was found, break
                    break
            assert number_matching_attributes == len(edge_dict)

    # *************************************************************************
    # *************************************************************************

    def test_simplify_network_manual(self):
        
        # seed number
        seed_number = random.randint(0, int(1e5))
        random.seed(seed_number)
        # define the network graph
        network = nx.MultiDiGraph()
        
        network.add_edges_from([
            (1,2,0,{
                _osm.KEY_OSMNX_OSMID: 1,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (1,2,0,{
                _osm.KEY_OSMNX_OSMID: 2,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (2,3,0,{
                _osm.KEY_OSMNX_OSMID: 3,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (3,3,0,{
                _osm.KEY_OSMNX_OSMID: 4,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (3,4,0,{
                _osm.KEY_OSMNX_OSMID: 5,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (2,5,0,{
                _osm.KEY_OSMNX_OSMID: 6,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            
            (5,6,0,{
                _osm.KEY_OSMNX_OSMID: 7,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (6,7,0,{
                _osm.KEY_OSMNX_OSMID: 8,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (5,8,0,{
                _osm.KEY_OSMNX_OSMID: 1,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            
            (8,9,0,{
                _osm.KEY_OSMNX_OSMID: 9,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            
            (9,10,0,{
                _osm.KEY_OSMNX_OSMID: 10,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            
            (10,12,0,{
                _osm.KEY_OSMNX_OSMID: 11,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (12,11,0,{
                _osm.KEY_OSMNX_OSMID: 12,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (8,11,0,{
                _osm.KEY_OSMNX_OSMID: 13,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (8,13,0,{
                _osm.KEY_OSMNX_OSMID: 14,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (13,14,0,{
                _osm.KEY_OSMNX_OSMID: 15,
                _osm.KEY_OSMNX_LENGTH: 3,
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (14,14,0,{
                _osm.KEY_OSMNX_OSMID: 16,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (14,15,0,{
                _osm.KEY_OSMNX_OSMID: 17,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (13,16,0,{
                _osm.KEY_OSMNX_OSMID: 18,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (16,17,0,{
                _osm.KEY_OSMNX_OSMID: 19,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (17,18,0,{
                _osm.KEY_OSMNX_OSMID: 20,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (16,19,0,{
                _osm.KEY_OSMNX_OSMID: 21,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (19,20,0,{
                _osm.KEY_OSMNX_OSMID: 22,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (20,22,0,{
                _osm.KEY_OSMNX_OSMID: 23,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (19,21,0,{
                _osm.KEY_OSMNX_OSMID: 24,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (21,22,0,{
                _osm.KEY_OSMNX_OSMID: 25,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (21,19,0,{
                _osm.KEY_OSMNX_OSMID: 26,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (21,19,1,{
                _osm.KEY_OSMNX_OSMID: 27,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (19,22,0,{
                _osm.KEY_OSMNX_OSMID: 28,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (19,23,0,{
                _osm.KEY_OSMNX_OSMID: 29,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (23,27,0,{
                _osm.KEY_OSMNX_OSMID: 30,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (27,24,0,{
                _osm.KEY_OSMNX_OSMID: 31,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (27,25,0,{
                _osm.KEY_OSMNX_OSMID: 32,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (27,26,0,{
                _osm.KEY_OSMNX_OSMID: 33,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (23,28,0,{
                _osm.KEY_OSMNX_OSMID: 34,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (28,29,0,{
                _osm.KEY_OSMNX_OSMID: 35,
                _osm.KEY_OSMNX_LENGTH: 0.1,
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (29,32,0,{
                _osm.KEY_OSMNX_OSMID: 36,
                _osm.KEY_OSMNX_LENGTH: 0.1+0.1,
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (28,30,0,{
                _osm.KEY_OSMNX_OSMID: 37,
                _osm.KEY_OSMNX_LENGTH: 0.1,
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (30,32,0,{
                _osm.KEY_OSMNX_OSMID: 38,
                _osm.KEY_OSMNX_LENGTH: 0.1+0.2,
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (32,31,0,{
                _osm.KEY_OSMNX_OSMID: 39,
                _osm.KEY_OSMNX_LENGTH: 0.1,
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (31,28,0,{
                _osm.KEY_OSMNX_OSMID: 40,
                _osm.KEY_OSMNX_LENGTH: 0.1+0.3,
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            # branch for protected edge (33,34)
            
            (23,33,0,{
                _osm.KEY_OSMNX_OSMID: 41,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (33,34,0,{
                _osm.KEY_OSMNX_OSMID: 42,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (34,35,0,{
                _osm.KEY_OSMNX_OSMID: 43,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            # branch for protected edge (36,37)
            (23,36,0,{
                _osm.KEY_OSMNX_OSMID: 44,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (36,37,0,{
                _osm.KEY_OSMNX_OSMID: 45,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
            
            (37,38,0,{
                _osm.KEY_OSMNX_OSMID: 46,
                _osm.KEY_OSMNX_LENGTH: random.random(),
                _osm.KEY_OSMNX_REVERSED: False,
                _osm.KEY_OSMNX_ONEWAY: False,
                },
            ),
        ])
        # add node data
        for i, node_key in enumerate(network.nodes()):
            network.nodes[node_key][_osm.KEY_OSMNX_X] = i+random.random()
            network.nodes[node_key][_osm.KEY_OSMNX_Y] = i+1+random.random()
        # define the keepers
        protected_nodes = [1, 2, 5, 8, 13, 16, 19, 4, 18, 22, 23, 24, 25, 26, 32, 35]
        # identify the original nodes
        node_keys = tuple(network.nodes())        
        # try simplifying it
        gis_utils.simplify_network(
            network, 
            protected_nodes=protected_nodes,
            protected_edges=[(33,34,0),(36,37,0)]
            )
        # protected nodes must still exist
        for node_key in protected_nodes:
            assert network.has_node(node_key)
        # there cannot be any self loop on the network
        assert len(tuple(gis_iden.find_self_loops(network))) == 0
        # there cannot be any simplifiable path
        assert len(gis_iden.find_simplifiable_paths(network, protected_nodes)) == 0
        # there cannot be any parallel arcs
        for edge_key in network.edges(keys=True):
            assert len(tuple(gis_iden.get_edges_between_two_nodes(network, *edge_key[0:2]))) == 1
        # there cannot be dead ends
        for node_key in node_keys:
            # nodes that no longer exist cannot be checked
            if not network.has_node(node_key):
                continue
            # some nodes can be dead ends, if they are protected
            if node_key in protected_nodes:
                continue
            # all other nodes cannot be dead ends
            assert len(tuple(gis_iden.neighbours(network, node_key))) >= 1
    
        # *********************************************************************
        # *********************************************************************
        
        # final configuration
        # there should be an edge between 2 and 4
        assert network.has_edge(2, 4) or network.has_edge(4, 2)
        # node 3 should not exist
        assert not network.has_node(3)
        # nodes 6 and 7 should not exist
        assert not network.has_node(6) and not network.has_node(7)
        # nodes 9, 10, 11, and 12 should not exist either
        assert (
            not network.has_node(9) and 
            not network.has_node(10) and 
            not network.has_node(11) and 
            not network.has_node(12)
            )
        # nodes 14 and 15 should not exist
        assert not network.has_node(14) and not network.has_node(15)
        
        # there should be an edge between 16 and 18
        assert network.has_edge(16, 18) or network.has_edge(18, 16)
        # node 17 should not exist
        assert not network.has_node(17)
        
        # there should be an edge between 19 and 22
        assert network.has_edge(19, 22) or network.has_edge(22, 19)
        # node 20 should not exist
        assert not network.has_node(20)
        # node 21 should not exist
        assert not network.has_node(21)
        
        # nodes 27, 33, 34 and 35 should exist
        assert network.has_node(27)
        assert network.has_node(33)
        assert network.has_node(34)
        assert network.has_node(35)
        # there should be an edge between 33 and 34
        assert network.has_edge(33, 34) or network.has_edge(34, 33)
        # there should be an edge between 34 and 35
        assert network.has_edge(34, 35) or network.has_edge(35, 34)
        
        # nodes 28, 29, 30, and 31 should not exist
        assert not network.has_node(28)
        assert not network.has_node(29)
        assert not network.has_node(30)
        assert not network.has_node(31)        
        # there should be an edge between 23 and 32
        assert network.has_edge(23, 32) or network.has_edge(32, 23)
        
        # nodes 36 and 37 should exist, node 38 should not
        assert network.has_node(36)
        assert network.has_node(37)
        assert not network.has_node(38)
        # the edge (36,37) should exist 
        assert network.has_edge(36, 37) or network.has_edge(37, 36)

    # *************************************************************************
    # *************************************************************************

    def test_simplify_network_osmnx(self):
        
        # BUG: there is a bug here, therefore we are printing the seed number
        # it seems to occur if all the nodes around 1106295281 are simplified
        # seed: 5785034948163332129
#         for edge_key in network.edges(keys=True):
# >           assert len(tuple(gis_iden.get_edges_between_two_nodes(network, *edge_key[0:2]))) == 1
# E           assert 2 == 1
# E            +  where 2 = len(((1106295281, 1106295315, 0), (1106295315, 1106295281, 0)))
# E            +    where ((1106295281, 1106295315, 0), (1106295315, 1106295281, 0)) = tuple([(1106295281, 1106295315, 0), (1106295315, 1106295281, 0)])
# E            +      where [(1106295281, 1106295315, 0), (1106295315, 1106295281, 0)] = <function get_edges_between_two_nodes at 0x7f8d07bc9ee0>(<networkx.classes.multidigraph.MultiDiGraph object at 0x7f8d00d17410>, *(1106295281, 1106295315))
# E            +        where <function get_edges_between_two_nodes at 0x7f8d07bc9ee0> = gis_iden.get_edges_between_two_nodes
        seed = random.randrange(sys.maxsize)
        random.seed(seed)
        print("Seed was:", seed)
        
        # get a network
        network = ox.graph_from_point(
            (55.71654, 9.11728),
            network_type="drive",
            custom_filter='["highway"~"residential|tertiary|unclassified|service"]',
            truncate_by_edge=True,
        )
        # protect some nodes
        node_keys = tuple(network.nodes())
        share_nodes_protected = 0.25
        number_nodes_protected = round(len(node_keys)*share_nodes_protected)
        protected_nodes = [
            node_keys[random.randint(0, len(node_keys) - 1)]
            for i in range(number_nodes_protected)
        ]
        # protected_nodes.append(1106295281)
        # assert 1 == 0
        # E        +  where 1 = len([[317212013, 1106295281, 1106295315]])
        # E        +    where [[317212013, 1106295281, 1106295315]] = <function find_simplifiable_paths at 0x7f9ac0cf22a0>(<networkx.classes.multidigraph.MultiDiGraph object at 0x7f9ab2c7b350>, [115838, 317195115, 5031764839, 317811652, 5076232388, 615539272, ...])
        # E        +      where <function find_simplifiable_paths at 0x7f9ac0cf22a0> = gis_iden.find_simplifiable_paths
        # try simplifying it
        gis_utils.simplify_network(network, protected_nodes=protected_nodes)
        # protected nodes must still exist
        for node_key in protected_nodes:
            assert network.has_node(node_key)
        # there cannot be any self loop on the network
        assert len(tuple(gis_iden.find_self_loops(network))) == 0
        # there cannot be any simplifiable path
        assert len(gis_iden.find_simplifiable_paths(network, protected_nodes)) == 0
        # there cannot be any parallel arcs
        for edge_key in network.edges(keys=True):
            assert len(tuple(gis_iden.get_edges_between_two_nodes(network, *edge_key[0:2]))) == 1
        # there cannot be dead ends
        for node_key in node_keys:
            # nodes that no longer exist cannot be checked
            if not network.has_node(node_key):
                continue
            # some nodes can be dead ends, if they are protected
            if node_key in protected_nodes:
                continue
            # all other nodes cannot be dead ends
            assert len(tuple(gis_iden.neighbours(network, node_key))) >= 1

    # *************************************************************************
    # *************************************************************************
    
    def test_create_edge_geometry(self):
        
        G = nx.MultiDiGraph()
        edge_key = (0, 1)
        G.add_node(0, x=0, y=0)
        G.add_node(1, x=1, y=1)
        k1 = G.add_edge(0, 1, length=2**0.5)
        first_geo = gis_utils.create_edge_geometry(G, edge_key)
        edge_key = (2, 3)
        G.add_node(2, x=0, y=1)
        G.add_node(3, x=1, y=0)
        k2 = G.add_edge(2, 3, length=2**0.5)
        second_geo = gis_utils.create_edge_geometry(G, edge_key)
        assert intersects(first_geo, second_geo)
        
        # test finding overlapping edges
        ol_edges = gis_utils.find_overlapping_edges(G)
        assert len(ol_edges) == 1
        assert ((0,1,k1),(2,3,k2)) in ol_edges
        # no overlapping edges if there is only one node
        ol_edges = gis_utils.find_overlapping_edges(G, excluded_edges=[(0,1,k1)])
        assert len(ol_edges) == 0
        
        # add fifth node
        G.add_node(4, x=0.5, y=1)
        k3 = G.add_edge(3, 4, length=(0.5**2+1**2)**0.5)
        ol_edges = gis_utils.find_overlapping_edges(G)
        assert len(ol_edges) == 2
        assert ((0,1,k1),(2,3,k2)) in ol_edges
        assert ((0,1,k1),(3,4,k3)) in ol_edges
        # fewer overlapping edges
        ol_edges = gis_utils.find_overlapping_edges(G, excluded_edges=[(2,3,k2)])
        assert len(ol_edges) == 1
        assert ((0,1,k1),(3,4,k3)) in ol_edges
        
    # *************************************************************************
    # *************************************************************************

# *****************************************************************************
# *****************************************************************************
